#include "pmu-events/pmu-events.h"
struct pmu_event pme_broadwellde[] = {
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "bdwde metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "bdwde metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely consumed by program instructions",
	.topic = "bdwde metrics",
	.metric_expr = "min( 1 , idq.mite_uops / ( uops_retired.retire_slots / inst_retired.any * 16 * ( icache.hit + icache.misses ) / 4.0 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "Frontend",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded Icache; or Uop Cache)",
	.topic = "bdwde metrics",
	.metric_expr = "idq.dsb_uops / ( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops )",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB; Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "bdwde metrics",
	.metric_expr = "1 / inst_retired.any / cycles",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active. This is called 'Clockticks' in VTune",
	.topic = "bdwde metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots",
	.topic = "bdwde metrics",
	.metric_expr = "4*( cpu_clk_unhalted.thread_any / 2 ) if #smt_on else cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "bdwde metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "bdwde metrics",
	.metric_expr = "inst_retired.any / ( cpu_clk_unhalted.thread_any / 2 ) if #smt_on else cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "bdwde metrics",
	.metric_expr = "uops_executed.thread / ( cpu@uops_executed.core\\,cmask\\=1@ / 2) if #smt_on else uops_executed.cycles_ge_1_uop_exec",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Average Branch Address Clear Cost (fraction of cycles)",
	.topic = "bdwde metrics",
	.metric_expr = "2* ( rs_events.empty_cycles - icache.ifdata_stall  - ( 14 * itlb_misses.stlb_hit + cpu@itlb_misses.walk_duration\\,cmask\\=1@ + 7* itlb_misses.walk_completed ) ) / rs_events.empty_end",
	.metric_name = "BAClear_Cost",
	.metric_group = "Unknown_Branches",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "bdwde metrics",
	.metric_expr = "( cpu_clk_unhalted.thread_any / 2 ) if #smt_on else cpu_clk_unhalted.thread",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads",
	.topic = "bdwde metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_uops_retired.l1_miss + mem_load_uops_retired.hit_lfb )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least 1 such miss)",
	.topic = "bdwde metrics",
	.metric_expr = "l1d_pend_miss.pending / ( cpu@l1d_pend_miss.pending_cycles\\,any\\=1@ / 2) if #smt_on else l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "bdwde metrics",
	.metric_expr = "( cpu@itlb_misses.walk_duration\\,cmask\\=1@ + cpu@dtlb_load_misses.walk_duration\\,cmask\\=1@ + cpu@dtlb_store_misses.walk_duration\\,cmask\\=1@ + 7*(dtlb_store_misses.walk_completed+dtlb_load_misses.walk_completed+itlb_misses.walk_completed)) / ( cpu_clk_unhalted.thread_any / 2 ) if #smt_on else cycles",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Average CPU Utilization",
	.topic = "bdwde metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "bdwde metrics",
	.metric_expr = "( 1*( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2* fp_arith_inst_retired.128b_packed_double + 4*( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8* fp_arith_inst_retired.256b_packed_single ) / 1000000000 / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "bdwde metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "bdwde metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "bdwde metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "C3 residency percent per core",
	.topic = "bdwde metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "bdwde metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "bdwde metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "bdwde metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "bdwde metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "bdwde metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "bdwde metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "l2_rqsts.demand_data_rd_miss",
	.event = "umask=0x21,period=200003,event=0x24",
	.desc = "Demand Data Read miss L2, no rejects",
	.topic = "cache",
	.long_desc = "This event counts the number of demand Data Read requests that miss L2 cache. Only not rejected loads are counted",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x22,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x24,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_demand_miss",
	.event = "umask=0x27,period=200003,event=0x24",
	.desc = "Demand requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.l2_pf_miss",
	.event = "umask=0x30,period=200003,event=0x24",
	.desc = "L2 prefetch requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of requests from the L2 hardware prefetchers that miss L2 cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0x3f,period=200003,event=0x24",
	.desc = "All requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0x41,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of demand Data Read requests that hit L2 cache. Only not rejected loads are counted",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x42,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0x44,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
},
{
	.name = "l2_rqsts.l2_pf_hit",
	.event = "umask=0x50,period=200003,event=0x24",
	.desc = "L2 prefetch requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of requests from the L2 hardware prefetchers that hit L2 cache. L3 prefetch new types",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0xe1,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
	.long_desc = "This event counts the number of demand Data Read requests (including requests from L1D hardware prefetchers). These loads may hit or miss L2 cache. Only non rejected loads are counted",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xe2,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the total number of RFO (read for ownership) requests to L2 cache. L2 RFO requests include both L1D demand RFO misses as well as L1D RFO prefetches",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0xe4,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "This event counts the total number of L2 code requests",
},
{
	.name = "l2_rqsts.all_demand_references",
	.event = "umask=0xe7,period=200003,event=0x24",
	.desc = "Demand requests to L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xf8,period=200003,event=0x24",
	.desc = "Requests from L2 hardware prefetchers",
	.topic = "cache",
	.long_desc = "This event counts the total number of requests from the L2 hardware prefetchers",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200003,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
},
{
	.name = "l2_demand_rqsts.wb_hit",
	.event = "umask=0x50,period=200003,event=0x27",
	.desc = "Not rejected writebacks that hit L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of WB requests that hit L2 cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed L3",
	.topic = "cache",
	.long_desc = "This event counts core-originated cacheable demand requests that miss the last level cache (LLC). Demand requests include loads, RFOs, and hardware prefetches from L1D, and instruction fetches from IFU",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to L3",
	.topic = "cache",
	.long_desc = "This event counts core-originated cacheable demand requests that refer to the last level cache (LLC). Demand requests include loads, RFOs, and hardware prefetches from L1D, and instruction fetches from IFU",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss oustandings duration in cycles",
	.topic = "cache",
	.long_desc = "This event counts duration of L1D miss outstanding, that is each cycle number of Fill Buffers (FB) outstanding required by Demand Reads. FB either is held by demand loads, or it is held by non-demand loads and gets hit at least once by demand. The valid outstanding interval is defined until the FB deallocation by one of the following ways: from FB allocation, if FB is allocated by demand; from the demand Hit FB, if it is allocated by hardware or software prefetch.\nNote: In the L1D, a Demand Read contains cacheable or noncacheable demand loads, including ones causing cache-line splits and reads due to page walks resulted from any request type",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
	.long_desc = "This event counts duration of L1D miss outstanding in cycles",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
	.topic = "cache",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "This event counts L1D data line replacements including opportunistic replacements, and replacements that require stall-for-replace or block-for-replace",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding Demand Data Read transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor. See the corresponding Umask under OFFCORE_REQUESTS.\nNote: A prefetch promoted to Demand is counted from the promotion point  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts cycles when offcore outstanding Demand Data Read transactions are present in the super queue (SQ). A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation)  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,cmask=6,period=2000003,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue  Spec update: BDM76",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding Code Reads transactions in the super queue every cycle. The Offcore outstanding state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding RFO store transactions in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding RFO (store) transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding demand rfo Reads transactions in the super queue every cycle. The Offcore outstanding state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding cacheable Core Data Read transactions in the super queue every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts cycles when offcore outstanding cacheable Core Data Read transactions are present in the super queue. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "lock_cycles.cache_lock_duration",
	.event = "umask=0x2,period=2000003,event=0x63",
	.desc = "Cycles when L1D is locked",
	.topic = "cache",
	.long_desc = "This event counts the number of cycles when the L1D is locked. It is a superset of the 0x1 mask (BUS_LOCK_CLOCKS.BUS_LOCK_DURATION)",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
	.long_desc = "This event counts the Demand Data Read requests sent to uncore. Use it in conjunction with OFFCORE_REQUESTS_OUTSTANDING to determine average latency in the uncore",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "This event counts both cacheable and noncachaeble code read requests",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "This event counts the demand RFO (read for ownership) requests including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "This event counts the demand and prefetch data reads. All Core Data Reads include cacheable Demands and L2 prefetchers (not L3 prefetchers). Counting also covers reads due to page walks resulted from any request type",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Offcore requests buffer cannot take more entries for this thread core",
	.topic = "cache",
	.long_desc = "This event counts the number of cases when the offcore requests buffer cannot take more entries for the core. This can happen when the superqueue does not contain eligible entries, or when L1D writeback pending FIFO requests is full.\nNote: Writeback pending FIFO has six entries",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load uops that miss the STLB. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts load uops with true STLB miss retired to the architected path. True STLB miss is an uop triggering page walk that gets completed without blocks, and later gets retired. This page walk can end up with or without a fault  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store uops that miss the STLB. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts store uops true STLB miss retired to the architected path. True STLB miss is an uop triggering page walk that gets completed without blocks, and later gets retired. This page walk can end up with or without a fault  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load uops with locked access. (Precise Event - PEBS)  Supports address when precise.  Spec update: BDM35 (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts load uops with locked access retired to the architected path  Supports address when precise.  Spec update: BDM35 (Precise event)",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load uops that split across a cacheline boundary.(Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts line-splitted load uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K)  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts line-splitted store uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K)  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load uops. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts load uops retired to the architected path with a filter on bits 0 and 1 applied.\nNote: This event ?ounts AVX-256bit load/store double-pump memory uops as a single uop at retirement. This event also counts SW prefetches  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts store uops retired to the architected path with a filter on bits 0 and 1 applied.\nNote: This event ?ounts AVX-256bit load/store double-pump memory uops as a single uop at retirement  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load uops with L1 cache hits as data sources. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data source were hits in the nearest-level (L1) cache.\nNote: Only two data-sources of L1/FB are applicable for AVX-256bit  even though the corresponding AVX load could be serviced by a deeper level in the memory hierarchy. Data source is reported for the Low-half load. This event also counts SW prefetches independent of the actual data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load uops with L2 cache hits as data sources. (Precise Event - PEBS)  Supports address when precise.  Spec update: BDM35 (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were hits in the mid-level (L2) cache  Supports address when precise.  Spec update: BDM35 (Precise event)",
},
{
	.name = "mem_load_uops_retired.l3_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Hit in last-level (L3) cache. Excludes Unknown data-source. (Precise Event - PEBS)  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were data hits in the last-level (L3) cache without snoops required  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load uops misses in L1 cache as data sources. Uses PEBS  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were misses in the nearest-level (L1) cache. Counting excludes unknown and UC data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Retired load uops with L2 cache misses as data sources. Uses PEBS  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were misses in the mid-level (L2) cache. Counting excludes unknown and UC data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l3_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Miss in last-level (L3) cache. Excludes Unknown data-source. (Precise Event - PEBS)  Supports address when precise.  Spec update: BDM100, BDE70 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.hit_lfb",
	.event = "umask=0x40,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources were load uops missed L1 but hit FB due to preceding miss to the same cache line with data not ready. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were load uops missed L1 but hit a fill buffer due to a preceding miss to the same cache line with the data not ready.\nNote: Only two data-sources of L1/FB are applicable for AVX-256bit  even though the corresponding AVX load could be serviced by a deeper level in the memory hierarchy. Data source is reported for the Low-half load  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 hit and cross-core snoop missed in on-pkg core cache. (Precise Event - PEBS)  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were L3 Hit and a cross-core snoop missed in the on-pkg core cache  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 and cross-core snoop hits in on-pkg core cache. (Precise Event - PEBS)  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were L3 hit and a cross-core snoop hit in the on-pkg core cache  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were HitM responses from shared L3. (Precise Event - PEBS)  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were HitM responses from a core on same socket (shared L3)  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load uops which data sources were hits in L3 without snoops required. (Precise Event - PEBS)  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were hits in the last-level (L3) cache without snoops required  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_l3_miss_retired.local_dram",
	.event = "umask=0x1,period=100007,event=0xd3",
	.desc = "(null)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops where the data came from local DRAM. This does not include hardware prefetches. This is a precise event  Supports address when precise.  Spec update: BDE70, BDM100",
},
{
	.name = "mem_load_uops_l3_miss_retired.remote_dram",
	.event = "umask=0x4,period=100007,event=0xd3",
	.desc = "Retired load uop whose Data Source was: remote DRAM either Snoop not needed or Snoop Miss (RspI) (Precise Event)  Supports address when precise.  Spec update: BDE70",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_miss_retired.remote_hitm",
	.event = "umask=0x10,period=100007,event=0xd3",
	.desc = "Retired load uop whose Data Source was: Remote cache HITM (Precise Event)  Supports address when precise.  Spec update: BDE70",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_miss_retired.remote_fwd",
	.event = "umask=0x20,period=100007,event=0xd3",
	.desc = "Retired load uop whose Data Source was: forwarded from remote cache (Precise Event)  Supports address when precise.  Spec update: BDE70",
	.topic = "cache",
},
{
	.name = "l2_trans.demand_data_rd",
	.event = "umask=0x1,period=200003,event=0xf0",
	.desc = "Demand Data Read requests that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts Demand Data Read requests that access L2 cache, including rejects",
},
{
	.name = "l2_trans.rfo",
	.event = "umask=0x2,period=200003,event=0xf0",
	.desc = "RFO requests that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts Read for Ownership (RFO) requests that access L2 cache",
},
{
	.name = "l2_trans.code_rd",
	.event = "umask=0x4,period=200003,event=0xf0",
	.desc = "L2 cache accesses when fetching instructions",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache accesses when fetching instructions",
},
{
	.name = "l2_trans.all_pf",
	.event = "umask=0x8,period=200003,event=0xf0",
	.desc = "L2 or L3 HW prefetches that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L2 or L3 HW prefetches that access L2 cache including rejects",
},
{
	.name = "l2_trans.l1d_wb",
	.event = "umask=0x10,period=200003,event=0xf0",
	.desc = "L1D writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L1D writebacks that access L2 cache",
},
{
	.name = "l2_trans.l2_fill",
	.event = "umask=0x20,period=200003,event=0xf0",
	.desc = "L2 fill requests that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L2 fill requests that access L2 cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L2 writebacks that access L2 cache",
},
{
	.name = "l2_trans.all_requests",
	.event = "umask=0x80,period=200003,event=0xf0",
	.desc = "Transactions accessing L2 pipe",
	.topic = "cache",
	.long_desc = "This event counts transactions that access the L2 pipe including snoops, pagewalks, and so on",
},
{
	.name = "l2_lines_in.i",
	.event = "umask=0x1,period=100003,event=0xf1",
	.desc = "L2 cache lines in I state filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines in the Invalidate state filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_in.s",
	.event = "umask=0x2,period=100003,event=0xf1",
	.desc = "L2 cache lines in S state filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines in the Shared state filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_in.e",
	.event = "umask=0x4,period=100003,event=0xf1",
	.desc = "L2 cache lines in E state filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines in the Exclusive state filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x7,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x5,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by demand",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Split locks in SQ",
	.topic = "cache",
	.long_desc = "This event counts the number of split locks in the super queue",
},
{
	.name = "other_assists.avx_to_sse",
	.event = "umask=0x8,period=100003,event=0xc1",
	.desc = "Number of transitions from AVX-256 to legacy SSE when penalty applicable  Spec update: BDM30",
	.topic = "floating point",
	.long_desc = "This event counts the number of transitions from AVX-256 to legacy SSE when penalty is applicable  Spec update: BDM30",
},
{
	.name = "other_assists.sse_to_avx",
	.event = "umask=0x10,period=100003,event=0xc1",
	.desc = "Number of transitions from SSE to AVX-256 when penalty applicable  Spec update: BDM30",
	.topic = "floating point",
	.long_desc = "This event counts the number of transitions from legacy SSE to AVX-256 when penalty is applicable  Spec update: BDM30",
},
{
	.name = "fp_arith_inst_retired.scalar_double",
	.event = "umask=0x1,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar double precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.scalar_single",
	.event = "umask=0x2,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar single precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.scalar",
	.event = "umask=0x3,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar floating-point instructions retired. Applies to SSE* and AVX* scalar, double and single precision floating-point: ADD SUB MUL DIV MIN MAX RSQRT RCP SQRT FM(N)ADD/SUB. FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_double",
	.event = "umask=0x4,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed double precision floating-point instructions retired.  Each count represents 2 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_single",
	.event = "umask=0x8,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed single precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_double",
	.event = "umask=0x10,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed double precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.double",
	.event = "umask=0x15,period=2000006,event=0xc7",
	.desc = "Number of SSE/AVX computational double precision floating-point instructions retired. Applies to SSE* and AVX*scalar, double and single precision floating-point: ADD SUB MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element.  ?",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_single",
	.event = "umask=0x20,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed single precision floating-point instructions retired.  Each count represents 8 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.single",
	.event = "umask=0x2a,period=2000005,event=0xc7",
	.desc = "Number of SSE/AVX computational single precision floating-point instructions retired. Applies to SSE* and AVX*scalar, double and single precision floating-point: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element. ?",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.packed",
	.event = "umask=0x3c,period=2000004,event=0xc7",
	.desc = "Number of SSE/AVX computational packed floating-point instructions retired. Applies to SSE* and AVX*, packed, double and single precision floating-point: ADD SUB MUL DIV MIN MAX RSQRT RCP SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_assist.x87_output",
	.event = "umask=0x2,period=100003,event=0xca",
	.desc = "Number of X87 assists due to output value",
	.topic = "floating point",
	.long_desc = "This event counts the number of x87 floating point (FP) micro-code assist (numeric overflow/underflow, inexact result) when the output value (destination register) is invalid",
},
{
	.name = "fp_assist.x87_input",
	.event = "umask=0x4,period=100003,event=0xca",
	.desc = "Number of X87 assists due to input value",
	.topic = "floating point",
	.long_desc = "This event counts x87 floating point (FP) micro-code assist (invalid operation, denormal operand, SNaN operand) when the input value (one of the source operands to an FP instruction) is invalid",
},
{
	.name = "fp_assist.simd_output",
	.event = "umask=0x8,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to Output values",
	.topic = "floating point",
	.long_desc = "This event counts the number of SSE* floating point (FP) micro-code assist (numeric overflow/underflow) when the output value (destination register) is invalid. Counting covers only cases involving penalties that require micro-code assist intervention",
},
{
	.name = "fp_assist.simd_input",
	.event = "umask=0x10,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to input values",
	.topic = "floating point",
	.long_desc = "This event counts any input SSE* FP assist - invalid operation, denormal operand, dividing by zero, SNaN operand. Counting includes only cases involving penalties that required micro-code assist intervention",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,cmask=1,period=100003,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
	.long_desc = "This event counts cycles with any input and output SSE or x87 FP assist. If an input and output assist are detected on the same cycle the event increments by 1",
},
{
	.name = "idq.empty",
	.event = "umask=0x2,period=2000003,event=0x79",
	.desc = "Instruction Decode Queue (IDQ) empty cycles",
	.topic = "frontend",
	.long_desc = "This counts the number of cycles that the instruction decoder queue is empty and can indicate that the application may be bound in the front end.  It does not determine whether there are uops being delivered to the Alloc stage since uops can be delivered by bypass skipping the Instruction Decode Queue (IDQ) when it is empty",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_dsb_uops",
	.event = "umask=0x10,period=2000003,event=0x79",
	.desc = "Uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_dsb_occur",
	.event = "edge=1,umask=0x10,cmask=1,period=2000003,event=0x79",
	.desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the number of deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while the Microcode Sequencer (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles 4  uops were  delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles  uops were  delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while the Microcode Sequenser (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles 4  uops were  delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles  uops were delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the total number of uops delivered to Instruction Decode Queue (IDQ) while the Microcode Sequenser (MS) is busy. Counting includes uops that may bypass the IDQ. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequenser (MS) is busy. Counting includes uops that may bypass the IDQ. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "idq.ms_switches",
	.event = "edge=1,umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
},
{
	.name = "idq.mite_all_uops",
	.event = "umask=0x3c,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=2000003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
	.topic = "frontend",
	.long_desc = "This event counts the number of both cacheable and noncacheable Instruction Cache, Streaming Buffer and Victim Cache Reads including UC fetches",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Misses. Includes Uncacheable accesses",
	.topic = "frontend",
	.long_desc = "This event counts the number of instruction cache, streaming buffer and victim cache misses. Counting includes UC accesses",
},
{
	.name = "icache.ifdata_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction-cache miss",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which the demand fetch waits for data (wfdM104H) from L2 or iSB (opportunistic hit)",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops not delivered to Resource Allocation Table (RAT) per thread adding 4  x when Resource Allocation Table (RAT) is not stalled and Instruction Decode Queue (IDQ) delivers x uops to Resource Allocation Table (RAT) (where x belongs to {0,1,2,3}). Counting does not cover cases when:\n a. IDQ-Resource Allocation Table (RAT) pipe serves the other thread;\n b. Resource Allocation Table (RAT) is stalled for the thread (including uop drops and clear BE conditions); \n c. Instruction Decode Queue (IDQ) delivers four uops",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,cmask=4,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "This event counts, on the per-thread basis, cycles when no uops are delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core =4",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,cmask=3,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "This event counts, on the per-thread basis, cycles when less than 1 uop is  delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core >=3",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,cmask=2,period=2000003,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "This event counts Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles. These cycles do not include uops routed through because of the switch itself, for example, when Instruction Decode Queue (IDQ) pre-allocation is unavailable, or Instruction Decode Queue (IDQ) is full. SBD-to-MITE switch true penalty cycles happen after the merge mux (MM) receives Decode Stream Buffer (DSB) Sync-indication until receiving the first MITE uop. \nMM is placed before Instruction Decode Queue (IDQ) to merge uops being fed from the MITE and Decode Stream Buffer (DSB) paths. Decode Stream Buffer (DSB) inserts the Sync-indication whenever a Decode Stream Buffer (DSB)-to-MITE switch occurs.\nPenalty: A Decode Stream Buffer (DSB) hit followed by a Decode Stream Buffer (DSB) miss can cost up to six cycles in which no uops are delivered to the IDQ. Most often, such switches from the Decode Stream Buffer (DSB) to the legacy pipeline cost 02 cycles",
},
{
	.name = "misalign_mem_ref.loads",
	.event = "umask=0x1,period=2000003,event=0x5",
	.desc = "Speculative cache line split load uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "This event counts speculative cache-line split load uops dispatched to the L1 cache",
},
{
	.name = "misalign_mem_ref.stores",
	.event = "umask=0x2,period=2000003,event=0x5",
	.desc = "Speculative cache line split STA uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "This event counts speculative cache line split store-address (STA) uops dispatched to the L1 cache",
},
{
	.name = "tx_mem.abort_conflict",
	.event = "umask=0x1,period=2000003,event=0x54",
	.desc = "Number of times a TSX line had a cache conflict",
	.topic = "memory",
	.long_desc = "Number of times a TSX line had a cache conflict",
},
{
	.name = "tx_mem.abort_capacity_write",
	.event = "umask=0x2,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to an evicted line caused by a transaction overflow",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to an evicted line caused by a transaction overflow",
},
{
	.name = "tx_mem.abort_hle_store_to_elided_lock",
	.event = "umask=0x4,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to a non-release/commit store to lock",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to a non-release/commit store to lock",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_not_empty",
	.event = "umask=0x8,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_mismatch",
	.event = "umask=0x10,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_unsupported_alignment",
	.event = "umask=0x20,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
},
{
	.name = "tx_mem.hle_elision_buffer_full",
	.event = "umask=0x40,period=2000003,event=0x54",
	.desc = "Number of times we could not allocate Lock Buffer",
	.topic = "memory",
	.long_desc = "Number of times we could not allocate Lock Buffer",
},
{
	.name = "tx_exec.misc1",
	.event = "umask=0x1,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions that may cause a transactional abort was executed. Since this is the count of execution, it may not always cause a transactional abort",
	.topic = "memory",
},
{
	.name = "tx_exec.misc2",
	.event = "umask=0x2,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions (e.g., vzeroupper) that may cause a transactional abort was executed inside a transactional region",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by  a vzeroupper instruction",
},
{
	.name = "tx_exec.misc3",
	.event = "umask=0x4,period=2000003,event=0x5d",
	.desc = "Counts the number of times an instruction execution caused the transactional nest count supported to be exceeded",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by a nest count that is too deep",
},
{
	.name = "tx_exec.misc4",
	.event = "umask=0x8,period=2000003,event=0x5d",
	.desc = "Counts the number of times a XBEGIN instruction was executed inside an HLE transactional region",
	.topic = "memory",
	.long_desc = "RTM region detected inside HLE",
},
{
	.name = "tx_exec.misc5",
	.event = "umask=0x10,period=2000003,event=0x5d",
	.desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
	.topic = "memory",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts",
	.topic = "memory",
	.long_desc = "This event counts the number of memory ordering Machine Clears detected. Memory Ordering Machine Clears can result from one of the following:\n1. memory disambiguation,\n2. external snoop, or\n3. cross SMT-HW-thread snoop (stores) hitting load buffer",
},
{
	.name = "hle_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc8",
	.desc = "Number of times we entered an HLE region; does not count nested transactions",
	.topic = "memory",
	.long_desc = "Number of times we entered an HLE region\n does not count nested transactions",
},
{
	.name = "hle_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc8",
	.desc = "Number of times HLE commit succeeded",
	.topic = "memory",
	.long_desc = "Number of times HLE commit succeeded",
},
{
	.name = "hle_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc8",
	.desc = "Number of times HLE abort was triggered (PEBS) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times HLE abort was triggered (PEBS) (Precise event)",
},
{
	.name = "hle_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an HLE abort was attributed to a Memory condition (See TSX_Memory event for additional details)",
},
{
	.name = "hle_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to uncommon conditions",
	.topic = "memory",
	.long_desc = "Number of times the TSX watchdog signaled an HLE abort",
},
{
	.name = "hle_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Number of times a disallowed operation caused an HLE abort",
},
{
	.name = "hle_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times HLE caused a fault",
},
{
	.name = "hle_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to none of the previous 4 categories (e.g. interrupts)",
	.topic = "memory",
	.long_desc = "Number of times HLE aborted and was not due to the abort conditions in subevents 3-6",
},
{
	.name = "rtm_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc9",
	.desc = "Number of times we entered an RTM region; does not count nested transactions",
	.topic = "memory",
	.long_desc = "Number of times we entered an RTM region\n does not count nested transactions",
},
{
	.name = "rtm_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc9",
	.desc = "Number of times RTM commit succeeded",
	.topic = "memory",
	.long_desc = "Number of times RTM commit succeeded",
},
{
	.name = "rtm_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc9",
	.desc = "Number of times RTM abort was triggered (PEBS) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times RTM abort was triggered (PEBS) (Precise event)",
},
{
	.name = "rtm_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an RTM abort was attributed to a Memory condition (See TSX_Memory event for additional details)",
},
{
	.name = "rtm_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times the TSX watchdog signaled an RTM abort",
},
{
	.name = "rtm_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Number of times a disallowed operation caused an RTM abort",
},
{
	.name = "rtm_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times a RTM caused a fault",
},
{
	.name = "rtm_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
	.topic = "memory",
	.long_desc = "Number of times RTM aborted and was not due to the abort conditions in subevents 3-6",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Loads with latency value being above 4  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "This event counts loads with latency value being above four  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Loads with latency value being above 8  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "This event counts loads with latency value being above eight  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Loads with latency value being above 16  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "This event counts loads with latency value being above 16  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Loads with latency value being above 32  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "This event counts loads with latency value being above 32  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Loads with latency value being above 64  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "This event counts loads with latency value being above 64  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Loads with latency value being above 128  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "This event counts loads with latency value being above 128  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Loads with latency value being above 256  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "This event counts loads with latency value being above 256  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Loads with latency value being above 512  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "This event counts loads with latency value being above 512  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "cpl_cycles.ring0",
	.event = "umask=0x1,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when the thread is in ring 0",
	.topic = "other",
	.long_desc = "This event counts the unhalted core cycles during which the thread is in the ring 0 privileged mode",
},
{
	.name = "cpl_cycles.ring0_trans",
	.event = "edge=1,umask=0x1,cmask=1,period=100007,event=0x5c",
	.desc = "Number of intervals between processor halts while thread is in ring 0",
	.topic = "other",
	.long_desc = "This event counts when there is a transition from ring 1,2 or 3 to ring0",
},
{
	.name = "cpl_cycles.ring123",
	.event = "umask=0x2,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when thread is in rings 1, 2, or 3",
	.topic = "other",
	.long_desc = "This event counts unhalted core cycles during which the thread is in rings 1, 2, or 3",
},
{
	.name = "lock_cycles.split_lock_uc_lock_duration",
	.event = "umask=0x1,period=2000003,event=0x63",
	.desc = "Cycles when L1 and L2 are locked due to UC or split lock",
	.topic = "other",
	.long_desc = "This event counts cycles in which the L1 and L2 are locked due to a UC lock or split lock. A lock is asserted in case of locked memory access, due to noncacheable memory, locked operation that spans two cache lines, or a page walk from the noncacheable page table. L1D and L2 locks have a very high performance penalty and it is highly recommended to avoid such access",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, this event counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers. \nNotes: INST_RETIRED.ANY is counted by a designated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. INST_RETIRED.ANY_P is counted by a programmable counter and it is an architectural performance event. \nCounting: Faulting executions of GETSEC/VM entry/VM Exit/MWait will not count as retired instructions",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state. This event has a constant ratio with the CPU_CLK_UNHALTED.REF_XCLK event. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. \nNote: On all current platforms this event stops counting during 'throttling (TM)' states duty off periods the processor is 'halted'.  This event is clocked by base clock (100 Mhz) on Sandy Bridge. The counter update is done at a lower clock rate then the core clock the overflow status bit for this counter may appear 'sticky'.  After the counter has overflowed and software clears the overflow status bit and resets the counter to less than MAX. The reset value to the counter is not clocked immediately so the overflow status bit will flip 'high (1)' and generate another PMI (if enabled) after which the reset value gets clocked into the counter. Therefore, software will get the interrupt, read the overflow status bit '1 for bit 34 while the counter value is less than MAX. Software should ignore this case",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Cases when loads get true Block-on-Store blocking code preventing store forwarding",
	.topic = "pipeline",
	.long_desc = "This event counts how many times the load operation got the true Block-on-Store blocking code preventing store forwarding. This includes cases when:\n - preceding store conflicts with the load (incomplete overlap);\n - store forwarding is impossible due to u-arch limitations;\n - preceding lock RMW operations are not forwarded;\n - store has the no-forward bit set (uncacheable/page-split/masked stores);\n - all-blocking stores are used (mostly, fences and port I/O);\nand others.\nThe most common case is a load blocked due to its address range overlapping with a preceding smaller uncompleted store. Note: This event does not take into account cases of out-of-SW-control (for example, SbTailHit), unknown physical STA, and cases of blocking loads on store due to being non-WB memory type or a lock. These cases are covered by other events.\nSee the table of not supported store forwards in the Optimization Guide",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "This event counts the number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare",
	.topic = "pipeline",
	.long_desc = "This event counts false dependencies in MOB when the partial comparison upon loose net check and dependency was resolved by the Enhanced Loose net mechanism. This may not result in high performance penalties. Loose net checks can fail when loads and stores are 4k aliased",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x3,cmask=1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for this thread (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "Cycles checkpoints in Resource Allocation Table (RAT) are recovering from JEClear or machine clear",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x3,any=1,cmask=1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "int_misc.rat_stall_cycles",
	.event = "umask=0x8,period=2000003,event=0xd",
	.desc = "Cycles when Resource Allocation Table (RAT) external stall is sent to Instruction Decode Queue (IDQ) for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts the number of cycles during which Resource Allocation Table (RAT) external stall is sent to Instruction Decode Queue (IDQ) for the current thread. This also includes the cycles during which the Allocator is serving another thread",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of Uops issued by the Resource Allocation Table (RAT) to the reservation station (RS)",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which the Resource Allocation Table (RAT) does not issue any Uops to the reservation station (RS) for the current thread",
},
{
	.name = "uops_issued.flags_merge",
	.event = "umask=0x10,period=2000003,event=0xe",
	.desc = "Number of flags-merge uops being allocated. Such uops considered perf sensitive; added by GSR u-arch",
	.topic = "pipeline",
	.long_desc = "Number of flags-merge uops being allocated. Such uops considered perf sensitive\n added by GSR u-arch",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
},
{
	.name = "uops_issued.single_mul",
	.event = "umask=0x40,period=2000003,event=0xe",
	.desc = "Number of Multiply packed/scalar single precision uops allocated",
	.topic = "pipeline",
},
{
	.name = "arith.fpu_div_active",
	.event = "umask=0x1,period=2000003,event=0x14",
	.desc = "Cycles when divider is busy executing divide operations",
	.topic = "pipeline",
	.long_desc = "This event counts the number of the divide operations executed. Uses edge-detect and a cmask value of 1 on ARITH.FPU_DIV_ACTIVE to get the number of the divide operations executed",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This is an architectural event that counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling. For this reason, this event may have a changing ratio with regards to wall clock time",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "This is a fixed-frequency event programmed to general counters. It counts when the core is unhalted at 100 Mhz",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "This event counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the software prefetch. It can also be incremented by some lock instructions. So it should only be used with profiling so that the locks can be excluded by asm inspection of the nearby instructions",
},
{
	.name = "load_hit_pre.hw_pf",
	.event = "umask=0x2,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for hardware prefetch",
	.topic = "pipeline",
	.long_desc = "This event counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the hardware prefetch",
},
{
	.name = "move_elimination.int_eliminated",
	.event = "umask=0x1,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_eliminated",
	.event = "umask=0x2,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.int_not_eliminated",
	.event = "umask=0x4,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_not_eliminated",
	.event = "umask=0x8,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which the reservation station (RS) is empty for the thread.\nNote: In ST-mode, not active thread should drive 0. This is usually caused by severely costly branch mispredictions, or allocator/FE issues",
},
{
	.name = "rs_events.empty_end",
	.event = "edge=1,inv=1,umask=0x1,cmask=1,period=200003,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
	.long_desc = "This event counts stalls occured due to changing prefix length (66, 67 or REX.W when they change the length of the decoded instruction). Occurrences counting is proportional to the number of prefixes in a 16B-line. This may result in the following penalties: three-cycle penalty for each LCP in a 16-byte chunk",
},
{
	.name = "br_inst_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x88",
	.desc = "Not taken macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts not taken macro-conditional branch instructions",
},
{
	.name = "br_inst_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired macro-conditional branch instructions",
},
{
	.name = "br_inst_exec.taken_direct_jump",
	.event = "umask=0x82,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired macro-conditional branch instructions excluding calls and indirect branches",
},
{
	.name = "br_inst_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired indirect branches excluding calls and return branches",
},
{
	.name = "br_inst_exec.taken_indirect_near_return",
	.event = "umask=0x88,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired indirect branches that have a return mnemonic",
},
{
	.name = "br_inst_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x88",
	.desc = "Taken speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect calls",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired indirect calls including both register and memory indirect",
},
{
	.name = "br_inst_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x88",
	.desc = "Speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired macro-conditional branch instructions",
},
{
	.name = "br_inst_exec.all_direct_jmp",
	.event = "umask=0xc2,period=200003,event=0x88",
	.desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired macro-unconditional branch instructions, excluding calls and indirects",
},
{
	.name = "br_inst_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x88",
	.desc = "Speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired indirect branches excluding calls and return branches",
},
{
	.name = "br_inst_exec.all_indirect_near_return",
	.event = "umask=0xc8,period=200003,event=0x88",
	.desc = "Speculative and retired indirect return branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired indirect branches that have a return mnemonic",
},
{
	.name = "br_inst_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x88",
	.desc = "Speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x88",
	.desc = "Speculative and retired  branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired branch instructions",
},
{
	.name = "br_misp_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x89",
	.desc = "Not taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts not taken speculative and retired mispredicted macro conditional branch instructions",
},
{
	.name = "br_misp_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired mispredicted macro conditional branch instructions",
},
{
	.name = "br_misp_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.taken_return_near",
	.event = "umask=0x88,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired mispredicted indirect branches that have a return mnemonic",
},
{
	.name = "br_misp_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired mispredicted macro conditional branch instructions",
},
{
	.name = "br_misp_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x89",
	.desc = "Mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired mispredicted branch instructions",
},
{
	.name = "uop_dispatches_cancelled.simd_prf",
	.event = "umask=0x3,period=2000003,event=0xa0",
	.desc = "Micro-op dispatches cancelled due to insufficient SIMD physical register file read ports",
	.topic = "pipeline",
	.long_desc = "This event counts the number of micro-operations cancelled after they were dispatched from the scheduler to the execution units when the total number of physical register read ports across all dispatch ports exceeds the read bandwidth of the physical register file.  The SIMD_PRF subevent applies to the following instructions: VDPPS, DPPS, VPCMPESTRI, PCMPESTRI, VPCMPESTRM, PCMPESTRM, VFMADD*, VFMADDSUB*, VFMSUB*, VMSUBADD*, VFNMADD*, VFNMSUB*.  See the Broadwell Optimization Guide for more information",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_executed_port.port_0_core",
	.event = "umask=0x1,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_executed_port.port_1_core",
	.event = "umask=0x2,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 2",
},
{
	.name = "uops_executed_port.port_2_core",
	.event = "umask=0x4,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 2",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 2",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 3",
},
{
	.name = "uops_executed_port.port_3_core",
	.event = "umask=0x8,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 3",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 3",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 4",
},
{
	.name = "uops_executed_port.port_4_core",
	.event = "umask=0x10,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 4",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_executed_port.port_5_core",
	.event = "umask=0x20,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_dispatched_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_executed_port.port_6_core",
	.event = "umask=0x40,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 6",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_dispatched_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 7",
},
{
	.name = "uops_executed_port.port_7_core",
	.event = "umask=0x80,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 7",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 7",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
	.long_desc = "This event counts resource-related stall cycles. Reasons for stalls can be as follows:\n - *any* u-arch structure got full (LB, SB, RS, ROB, BOB, LM, Physical Register Reclaim Table (PRRT), or Physical History Table (PHT) slots)\n - *any* u-arch structure got empty (like INT/SIMD FreeLists)\n - FPU control word (FPCW), MXCSR\nand others. This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "resource_stalls.rs",
	.event = "umask=0x4,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no eligible RS entry available",
	.topic = "pipeline",
	.long_desc = "This event counts stall cycles caused by absence of eligible entries in the reservation station (RS). This may result from RS overflow, or from RS deallocation because of the RS array Write Port allocation scheme (each RS entry has two write ports instead of four. As a result, empty entries could not be used, although RS is not really full). This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "This event counts stall cycles caused by the store buffer (SB) overflow (excluding draining from synch). This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "resource_stalls.rob",
	.event = "umask=0x10,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to re-order buffer full",
	.topic = "pipeline",
	.long_desc = "This event counts ROB full stall cycles. This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "cycle_activity.cycles_l2_pending",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles the CPU has at least one pending  demand* load request missing the L2 cache",
},
{
	.name = "cycle_activity.cycles_l2_miss",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_ldm_pending",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles the CPU has at least one pending  demand load request (that is cycles with non-completed load waiting for its data from memory subsystem)",
},
{
	.name = "cycle_activity.cycles_mem_any",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_no_execute",
	.event = "umask=0x4,cmask=4,period=2000003,event=0xa3",
	.desc = "This event increments by 1 for every cycle where there was no execute for this thread",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port",
},
{
	.name = "cycle_activity.stalls_total",
	.event = "umask=0x4,cmask=4,period=2000003,event=0xa3",
	.desc = "Total execution stalls",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_pending",
	.event = "umask=0x5,cmask=5,period=2000003,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port, while there was at least one pending demand* load request missing the L2 cache.(as a footprint) * includes also L1 HW prefetch requests that may or may not be required by demands",
},
{
	.name = "cycle_activity.stalls_l2_miss",
	.event = "umask=0x5,cmask=5,period=2000003,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_ldm_pending",
	.event = "umask=0x6,cmask=6,period=2000003,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port, while there was at least one pending demand load request",
},
{
	.name = "cycle_activity.stalls_mem_any",
	.event = "umask=0x6,cmask=6,period=2000003,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_pending",
	.event = "umask=0x8,cmask=8,period=2000003,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles the CPU has at least one pending  demand load request missing the L1 data cache",
},
{
	.name = "cycle_activity.cycles_l1d_miss",
	.event = "umask=0x8,cmask=8,period=2000003,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_pending",
	.event = "umask=0xc,cmask=12,period=2000003,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port, while there was at least one pending demand load request missing the L1 data cache",
},
{
	.name = "cycle_activity.stalls_l1d_miss",
	.event = "umask=0xc,cmask=12,period=2000003,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "uops_executed.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Counts the number of uops to be executed per-thread each cycle",
	.topic = "pipeline",
	.long_desc = "Number of uops to be executed per-thread each cycle",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which no uops were dispatched from the Reservation Station (RS) per thread",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core",
	.topic = "pipeline",
	.long_desc = "Number of uops executed from any thread",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter   - architectural event  Spec update: BDM61",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions (EOMs) retired. Counting covers macro-fused instructions individually (that is, increments by two)  Spec update: BDM61",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: BDM11, BDM55 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts instructions retired  Spec update: BDM11, BDM55 (Must be precise)",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000003,event=0xc0",
	.desc = "FP operations  retired. X87 FP operations that have no exceptions:",
	.topic = "pipeline",
	.long_desc = "This event counts FP operations retired. For X87 FP operations that have no exceptions counting also includes flows that have several X87, or flows that use X87 uops in the exception handling",
},
{
	.name = "other_assists.any_wb_assist",
	.event = "umask=0x40,period=100003,event=0xc1",
	.desc = "Number of times any microcode assist is invoked by HW upon uop writeback",
	.topic = "pipeline",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Actually retired uops. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts all actually retired uops. Counting increments by two for micro-fused uops, and by one for macro-fused and other uops. Maximal increment value for one cycle is eight  Supports address when precise (Precise event)",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
	.long_desc = "This event counts cycles without actually retired uops",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,cmask=10,period=2000003,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
	.long_desc = "Number of cycles using always true condition (uops_ret < 16) applied to non PEBS uops retired event",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts the number of retirement slots used (Precise event)",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=2000003,event=0xc3",
	.desc = "Cycles there was a Nuke. Account for both thread-specific and All Thread Nukes",
	.topic = "pipeline",
	.long_desc = "This event counts both thread-specific (TS) and all-thread (AT) nukes",
},
{
	.name = "machine_clears.count",
	.event = "edge=1,umask=0x1,cmask=1,period=100003,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "This event counts self-modifying code (SMC) detected, which causes a machine clear",
},
{
	.name = "machine_clears.maskmov",
	.event = "umask=0x20,period=100003,event=0xc3",
	.desc = "This event counts the number of executed Intel AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
	.topic = "pipeline",
	.long_desc = "Maskmov false fault - counts number of time ucode passes through Maskmov flow due to instruction's mask being 0 while the flow was completed without raising a fault",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts all (macro) branch instructions retired",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts conditional branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts both direct and indirect near call instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_call_r3",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect macro near call instructions retired (captured in ring 3). (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts both direct and indirect macro near call instructions retired (captured in ring 3) (Precise event)",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired. (Precise Event - PEBS)  Spec update: BDW98 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_INST_RETIRED.ALL_BRANCHES that counts all (macro) branch instructions retired  Spec update: BDW98 (Must be precise)",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts return instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Not taken branch instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts not taken branch instructions retired",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts taken branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Far branch instructions retired  Spec update: BDW98",
	.topic = "pipeline",
	.long_desc = "This event counts far branch instructions retired  Spec update: BDW98",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts all mispredicted macro branch instructions retired",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts mispredicted conditional branch instructions retired (Precise event)",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired. (Precise Event - PEBS) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_MISP_RETIRED.ALL_BRANCHES that counts all mispredicted macro branch instructions retired (Must be precise)",
},
{
	.name = "br_misp_retired.ret",
	.event = "umask=0x8,period=100007,event=0xc5",
	.desc = "This event counts the number of mispredicted ret instructions retired.(Precise Event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts mispredicted return instructions retired",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "number of near branch instructions retired that were mispredicted and taken. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "Number of near branch instructions retired that were mispredicted and taken. (Precise Event - PEBS) (Precise event)",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Count cases of saving new LBR",
	.topic = "pipeline",
	.long_desc = "This event counts cases of saving new LBR records by hardware. This assumes proper enabling of LBRs and takes into account LBR filtering done by the LBR_SELECT register",
},
{
	.name = "baclears.any",
	.event = "umask=0x1f,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
},
{
	.name = "unc_c_clockticks",
	.event = "event=0",
	.desc = "Uncore cache clock ticks. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_lookup.any",
	.event = "umask=0x11,event=0x34,filter_state=0x1",
	.desc = "All LLC Misses (code+ data rd + data wr - including demand and prefetch). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_victims.m_state",
	.event = "umask=0x1,event=0x37",
	.desc = "M line evictions from LLC (writebacks to memory). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.data_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x182",
	.desc = "LLC misses - demand and prefetch data reads - excludes LLC prefetches. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.uncacheable",
	.event = "umask=0x3,event=0x35,filter_opc=0x187",
	.desc = "LLC misses - Uncacheable reads (from cpu) . Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x187,filter_nc=1",
	.desc = "MMIO reads. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x18f,filter_nc=1",
	.desc = "MMIO writes. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.rfo_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x190",
	.desc = "LLC prefetch misses for RFO. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.code_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x191",
	.desc = "LLC prefetch misses for code reads. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.data_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x192",
	.desc = "LLC prefetch misses for data reads. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x19e",
	.desc = "LLC misses for PCIe read current. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x1c8",
	.desc = "ItoM write misses (as part of fast string memcpy stores) + PCIe full line writes. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_non_snoop_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x1c8,filter_tid=0x3e",
	.desc = "PCIe write misses (full cache line). Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_ns_partial_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x180,filter_tid=0x3e",
	.desc = "PCIe writes (partial cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "llc_references.code_llc_prefetch",
	.event = "umask=0x1,event=0x35,filter_opc=0x181",
	.desc = "L2 demand and L2 prefetch code references to LLC. Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_full",
	.event = "umask=0x1,event=0x35,filter_opc=0x18c",
	.desc = "Streaming stores (full cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_partial",
	.event = "umask=0x1,event=0x35,filter_opc=0x18d",
	.desc = "Streaming stores (partial cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_read",
	.event = "umask=0x1,event=0x35,filter_opc=0x19e",
	.desc = "PCIe read current. Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x1c8,filter_tid=0x3e",
	.desc = "PCIe write references (full cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_tor_occupancy.llc_data_read",
	.event = "umask=0x3,event=0x36,filter_opc=0x182",
	.desc = "Occupancy counter for LLC data reads (demand and L2 prefetch). Derived from unc_c_tor_occupancy.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads",
	.event = "umask=0x3,event=0x1",
	.desc = "read requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads_local",
	.event = "umask=0x1,event=0x1",
	.desc = "read requests to local home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads_remote",
	.event = "umask=0x2,event=0x1",
	.desc = "read requests to remote home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes",
	.event = "umask=0xC,event=0x1",
	.desc = "write requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes_local",
	.event = "umask=0x4,event=0x1",
	.desc = "write requests to local home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes_remote",
	.event = "umask=0x8,event=0x1",
	.desc = "write requests to remote home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspcnflct",
	.event = "umask=0x40,event=0x21",
	.desc = "Conflict requests (requests for same address from multiple agents simultaneously). Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rsp_fwd_wb",
	.event = "umask=0x20,event=0x21",
	.desc = "M line forwarded from remote cache along with writeback to memory. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspifwd",
	.event = "umask=0x4,event=0x21",
	.desc = "M line forwarded from remote cache with no writeback to memory. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rsps",
	.event = "umask=0x2,event=0x21",
	.desc = "Shared line response from remote cache. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspsfwd",
	.event = "umask=0x8,event=0x21",
	.desc = "Shared line forwarded from remote cache. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_read",
	.event = "umask=0x3,event=0x4",
	.desc = "read requests to memory controller. Derived from unc_m_cas_count.rd. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_write",
	.event = "umask=0xC,event=0x4",
	.desc = "write requests to memory controller. Derived from unc_m_cas_count.wr. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_dclockticks",
	.event = "event=0",
	.desc = "Memory controller clock ticks. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_power_channel_ppd",
	.event = "event=0x85",
	.desc = "Cycles where DRAM ranks are in power down (CKE) mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_channel_ppd / unc_m_dclockticks) * 100.",
	.metric_name = "power_channel_ppd %",
},
{
	.name = "unc_m_power_critical_throttle_cycles",
	.event = "event=0x86",
	.desc = "Cycles all ranks are in critical thermal throttle. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_critical_throttle_cycles / unc_m_dclockticks) * 100.",
	.metric_name = "power_critical_throttle_cycles %",
},
{
	.name = "unc_m_power_self_refresh",
	.event = "event=0x43",
	.desc = "Cycles Memory is in self refresh power mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_self_refresh / unc_m_dclockticks) * 100.",
	.metric_name = "power_self_refresh %",
},
{
	.name = "unc_m_pre_count.page_miss",
	.event = "umask=0x1,event=0x2",
	.desc = "Pre-charges due to page misses. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.rd",
	.event = "umask=0x4,event=0x2",
	.desc = "Pre-charge for reads. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.wr",
	.event = "umask=0x8,event=0x2",
	.desc = "Pre-charge for writes. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_p_clockticks",
	.event = "event=0",
	.desc = "PCU clock ticks. Use to get percentages of PCU cycles events. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
},
{
	.name = "unc_p_power_state_occupancy.cores_c0",
	.event = "event=0x80,occ_sel=1",
	.desc = "This is an occupancy event that tracks the number of cores that are in C0.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c0 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c0 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c3",
	.event = "event=0x80,occ_sel=2",
	.desc = "This is an occupancy event that tracks the number of cores that are in C3.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c3 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c3 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c6",
	.event = "event=0x80,occ_sel=3",
	.desc = "This is an occupancy event that tracks the number of cores that are in C6.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events . Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c6 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c6 %",
},
{
	.name = "unc_p_prochot_external_cycles",
	.event = "event=0xa",
	.desc = "Counts the number of cycles that we are in external PROCHOT mode.  This mode is triggered when a sensor off the die determines that something off-die (like DRAM) is too hot and must throttle to avoid damaging the chip. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_prochot_external_cycles / unc_p_clockticks) * 100.",
	.metric_name = "prochot_external_cycles %",
},
{
	.name = "unc_p_freq_max_limit_thermal_cycles",
	.event = "event=0x4",
	.desc = "Counts the number of cycles when temperature is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_limit_thermal_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_limit_thermal_cycles %",
},
{
	.name = "unc_p_freq_max_os_cycles",
	.event = "event=0x6",
	.desc = "Counts the number of cycles when the OS is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_os_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_os_cycles %",
},
{
	.name = "unc_p_freq_max_power_cycles",
	.event = "event=0x5",
	.desc = "Counts the number of cycles when power is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_power_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_power_cycles %",
},
{
	.name = "unc_p_freq_trans_cycles",
	.event = "event=0x74",
	.desc = "Counts the number of cycles when current is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_trans_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_trans_cycles %",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause page walks of any page size (4K/2M/4M/1G)  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (4K)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause a completed page walk (4K page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (2M/4M)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed_1g",
	.event = "umask=0x8,period=2000003,event=0x8",
	.desc = "Load miss in all TLB levels causes a page walk that completes. (1G)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause a completed page walk (1G  page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes of any page size  Spec update: BDM69",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_duration",
	.event = "umask=0x10,period=2000003,event=0x8",
	.desc = "Cycles when PMH is busy with page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of cycles while PMH is busy with the page walk  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.stlb_hit_4k",
	.event = "umask=0x20,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit_2m",
	.event = "umask=0x40,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x60,period=2000003,event=0x8",
	.desc = "Load operations that miss the first DTLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause page walks of any page size (4K/2M/4M/1G)  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (4K)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (4K page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks (2M/4M)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks (1G)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (1G  page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks  Spec update: BDM69",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x49",
	.desc = "Cycles when PMH is busy with page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of cycles while PMH is busy with the page walk  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x49",
	.desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Cycle count for an Extended Page table walk",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles for an extended page table walk. The Extended Page directory cache differs from standard TLB caches by the operating system that use it. Virtual machine operating systems use the extended page directory cache, while guest operating systems use the standard TLB caches",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause page walks of any page size (4K/2M/4M/1G)  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (4K)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (4K page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (2M/4M)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x85",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (1G)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (1G  page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x85",
	.desc = "Misses in all ITLB levels that cause completed page walks  Spec update: BDM69",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Cycles when PMH is busy with page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of cycles while PMH is busy with the page walk  Spec update: BDM69",
},
{
	.name = "itlb_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x85",
	.desc = "Core misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x85",
	.desc = "Code misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x85",
	.desc = "Operations that miss the first ITLB level but hit the second and do not cause any page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of flushes of the big or small ITLB pages. Counting include both TLB Flush (covering all sets) and TLB Set Clear (set-specific)",
},
{
	.name = "page_walker_loads.dtlb_l1",
	.event = "umask=0x11,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L1+FB  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.dtlb_l2",
	.event = "umask=0x12,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L2  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.dtlb_l3",
	.event = "umask=0x14,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L3 + XSNP  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.dtlb_memory",
	.event = "umask=0x18,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in Memory  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.itlb_l1",
	.event = "umask=0x21,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L1+FB  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.itlb_l2",
	.event = "umask=0x22,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L2  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.itlb_l3",
	.event = "umask=0x24,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L3 + XSNP  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of any STLB flush attempts (such as entire, VPID, PCID, InvPage, CR3 write, and so on)",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_broadwellx[] = {
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "bdx metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "bdx metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "bdx metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "bdx metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "bdx metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "bdx metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "bdx metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "bdx metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "bdx metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Instruction per taken branch",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_taken",
	.metric_name = "IpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Branch instructions per taken branch",
	.topic = "bdx metrics",
	.metric_expr = "br_inst_retired.all_branches / br_inst_retired.near_taken",
	.metric_name = "BpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "bdx metrics",
	.metric_expr = "min( 1 , idq.mite_uops / ( (uops_retired.retire_slots / inst_retired.any) * 16 * ( icache.hit + icache.misses ) / 4.0 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "bdx metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops ) )",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "bdx metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "bdx metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "bdx metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "bdx metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Instructions per Load (lower number means loads are more frequent)",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_loads",
	.metric_name = "IpL",
	.metric_group = "Instruction_Type;L1_Bound",
},
{
	.desc = "Instructions per Store",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_stores",
	.metric_name = "IpS",
	.metric_group = "Instruction_Type;Store_Bound",
},
{
	.desc = "Instructions per Branch",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.all_branches",
	.metric_name = "IpB",
	.metric_group = "Branches;Instruction_Type;Port_5;Port_6",
},
{
	.desc = "Instruction per (near) call",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_call",
	.metric_name = "IpCall",
	.metric_group = "Branches",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "bdx metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * fp_arith_inst_retired.256b_packed_single )) / cycles",
	.metric_name = "FLOPc",
	.metric_group = "FLOPS",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "bdx metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * fp_arith_inst_retired.256b_packed_single )) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "FLOPc_SMT",
	.metric_group = "FLOPS_SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "bdx metrics",
	.metric_expr = "uops_executed.thread / (( cpu@uops_executed.core\\,cmask\\=1@ / 2 ) if #smt_on else uops_executed.cycles_ge_1_uop_exec)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "bdx metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) * (12 * ( br_misp_retired.all_branches + machine_clears.count + baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) ) * (4 * cycles) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "bdx metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) * (12 * ( br_misp_retired.all_branches + machine_clears.count + baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) ) * (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost_SMT",
	.metric_group = "Branch_Mispredicts_SMT",
},
{
	.desc = "Number of Instructions per non-speculative Branch Misprediction (JEClear)",
	.topic = "bdx metrics",
	.metric_expr = "inst_retired.any / br_misp_retired.all_branches",
	.metric_name = "IpMispredict",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "bdx metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads (in core cycles)",
	.topic = "bdx metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_uops_retired.l1_miss + mem_load_uops_retired.hit_lfb )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least one such miss. Per-thread)",
	.topic = "bdx metrics",
	.metric_expr = "l1d_pend_miss.pending / l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "bdx metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration + 7 * ( dtlb_store_misses.walk_completed + dtlb_load_misses.walk_completed + itlb_misses.walk_completed ) ) / ( 2 * cycles )",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "bdx metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration + 7 * ( dtlb_store_misses.walk_completed + dtlb_load_misses.walk_completed + itlb_misses.walk_completed ) ) / ( 2 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )) )",
	.metric_name = "Page_Walks_Utilization_SMT",
	.metric_group = "TLB_SMT",
},
{
	.desc = "Average data fill bandwidth to the L1 data cache [GB / sec]",
	.topic = "bdx metrics",
	.metric_expr = "64 * l1d.replacement / 1000000000 / duration_time",
	.metric_name = "L1D_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average data fill bandwidth to the L2 cache [GB / sec]",
	.topic = "bdx metrics",
	.metric_expr = "64 * l2_lines_in.all / 1000000000 / duration_time",
	.metric_name = "L2_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "bdx metrics",
	.metric_expr = "64 * longest_lat_cache.miss / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "L1 cache true misses per kilo instruction for retired demand loads",
	.topic = "bdx metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l1_miss / inst_retired.any",
	.metric_name = "L1MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache true misses per kilo instruction for retired demand loads",
	.topic = "bdx metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache misses per kilo instruction for all request types (including speculative)",
	.topic = "bdx metrics",
	.metric_expr = "1000 * l2_rqsts.miss / inst_retired.any",
	.metric_name = "L2MPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache hits per kilo instruction for all request types (including speculative)",
	.topic = "bdx metrics",
	.metric_expr = "1000 * ( l2_rqsts.references - l2_rqsts.miss ) / inst_retired.any",
	.metric_name = "L2HPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L3 cache true misses per kilo instruction for retired demand loads",
	.topic = "bdx metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l3_miss / inst_retired.any",
	.metric_name = "L3MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "Average CPU Utilization",
	.topic = "bdx metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "bdx metrics",
	.metric_expr = "( (( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * fp_arith_inst_retired.256b_packed_single )) / 1000000000 ) / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "bdx metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "bdx metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "bdx metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "bdx metrics",
	.metric_expr = "( 64 * ( uncore_imc@cas_count_read@ + uncore_imc@cas_count_write@ ) / 1000000000 ) / duration_time",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average latency of data read request to external memory (in nanoseconds). Accounts for demand loads and L1/L2 prefetches",
	.topic = "bdx metrics",
	.metric_expr = "1000000000 * ( cbox@event\\=0x36\\,umask\\=0x3\\,filter_opc\\=0x182@ / cbox@event\\=0x35\\,umask\\=0x3\\,filter_opc\\=0x182@ ) / ( cbox_0@event\\=0x0@ / duration_time )",
	.metric_name = "DRAM_Read_Latency",
	.metric_group = "Memory_Lat",
},
{
	.desc = "Average number of parallel data read requests to external memory. Accounts for demand loads and L1/L2 prefetches",
	.topic = "bdx metrics",
	.metric_expr = "cbox@event\\=0x36\\,umask\\=0x3\\,filter_opc\\=0x182@ / cbox@event\\=0x36\\,umask\\=0x3\\,filter_opc\\=0x182\\,thresh\\=1@",
	.metric_name = "DRAM_Parallel_Reads",
	.metric_group = "Memory_BW",
},
{
	.desc = "Socket actual clocks when any core is active on that socket",
	.topic = "bdx metrics",
	.metric_expr = "cbox_0@event\\=0x0@",
	.metric_name = "Socket_CLKS",
	.metric_group = "",
},
{
	.desc = "C3 residency percent per core",
	.topic = "bdx metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "bdx metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "bdx metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "bdx metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "bdx metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "bdx metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "bdx metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "l2_rqsts.demand_data_rd_miss",
	.event = "umask=0x21,period=200003,event=0x24",
	.desc = "Demand Data Read miss L2, no rejects",
	.topic = "cache",
	.long_desc = "This event counts the number of demand Data Read requests that miss L2 cache. Only not rejected loads are counted",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x22,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x24,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_demand_miss",
	.event = "umask=0x27,period=200003,event=0x24",
	.desc = "Demand requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.l2_pf_miss",
	.event = "umask=0x30,period=200003,event=0x24",
	.desc = "L2 prefetch requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of requests from the L2 hardware prefetchers that miss L2 cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0x3f,period=200003,event=0x24",
	.desc = "All requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0xc1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests, initiated by load instructions, that hit L2 cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0xc2,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0xc4,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
},
{
	.name = "l2_rqsts.l2_pf_hit",
	.event = "umask=0xd0,period=200003,event=0x24",
	.desc = "L2 prefetch requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of requests from the L2 hardware prefetchers that hit L2 cache. L3 prefetch new types",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0xe1,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
	.long_desc = "This event counts the number of demand Data Read requests (including requests from L1D hardware prefetchers). These loads may hit or miss L2 cache. Only non rejected loads are counted",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xe2,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the total number of RFO (read for ownership) requests to L2 cache. L2 RFO requests include both L1D demand RFO misses as well as L1D RFO prefetches",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0xe4,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "This event counts the total number of L2 code requests",
},
{
	.name = "l2_rqsts.all_demand_references",
	.event = "umask=0xe7,period=200003,event=0x24",
	.desc = "Demand requests to L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xf8,period=200003,event=0x24",
	.desc = "Requests from L2 hardware prefetchers",
	.topic = "cache",
	.long_desc = "This event counts the total number of requests from the L2 hardware prefetchers",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200003,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
},
{
	.name = "l2_demand_rqsts.wb_hit",
	.event = "umask=0x50,period=200003,event=0x27",
	.desc = "Not rejected writebacks that hit L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of WB requests that hit L2 cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed L3",
	.topic = "cache",
	.long_desc = "This event counts core-originated cacheable demand requests that miss the last level cache (LLC). Demand requests include loads, RFOs, and hardware prefetches from L1D, and instruction fetches from IFU",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to L3",
	.topic = "cache",
	.long_desc = "This event counts core-originated cacheable demand requests that refer to the last level cache (LLC). Demand requests include loads, RFOs, and hardware prefetches from L1D, and instruction fetches from IFU",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss oustandings duration in cycles",
	.topic = "cache",
	.long_desc = "This event counts duration of L1D miss outstanding, that is each cycle number of Fill Buffers (FB) outstanding required by Demand Reads. FB either is held by demand loads, or it is held by non-demand loads and gets hit at least once by demand. The valid outstanding interval is defined until the FB deallocation by one of the following ways: from FB allocation, if FB is allocated by demand; from the demand Hit FB, if it is allocated by hardware or software prefetch.\nNote: In the L1D, a Demand Read contains cacheable or noncacheable demand loads, including ones causing cache-line splits and reads due to page walks resulted from any request type",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
	.long_desc = "This event counts duration of L1D miss outstanding in cycles",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
	.topic = "cache",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "This event counts L1D data line replacements including opportunistic replacements, and replacements that require stall-for-replace or block-for-replace",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding Demand Data Read transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor. See the corresponding Umask under OFFCORE_REQUESTS.\nNote: A prefetch promoted to Demand is counted from the promotion point  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts cycles when offcore outstanding Demand Data Read transactions are present in the super queue (SQ). A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation)  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,cmask=6,period=2000003,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue  Spec update: BDM76",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding Code Reads transactions in the super queue every cycle. The Offcore outstanding state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding RFO store transactions in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding RFO (store) transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding demand rfo Reads transactions in the super queue every cycle. The Offcore outstanding state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding cacheable Core Data Read transactions in the super queue every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts cycles when offcore outstanding cacheable Core Data Read transactions are present in the super queue. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "lock_cycles.cache_lock_duration",
	.event = "umask=0x2,period=2000003,event=0x63",
	.desc = "Cycles when L1D is locked",
	.topic = "cache",
	.long_desc = "This event counts the number of cycles when the L1D is locked. It is a superset of the 0x1 mask (BUS_LOCK_CLOCKS.BUS_LOCK_DURATION)",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
	.long_desc = "This event counts the Demand Data Read requests sent to uncore. Use it in conjunction with OFFCORE_REQUESTS_OUTSTANDING to determine average latency in the uncore",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "This event counts both cacheable and noncachaeble code read requests",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "This event counts the demand RFO (read for ownership) requests including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "This event counts the demand and prefetch data reads. All Core Data Reads include cacheable Demands and L2 prefetchers (not L3 prefetchers). Counting also covers reads due to page walks resulted from any request type",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Offcore requests buffer cannot take more entries for this thread core",
	.topic = "cache",
	.long_desc = "This event counts the number of cases when the offcore requests buffer cannot take more entries for the core. This can happen when the superqueue does not contain eligible entries, or when L1D writeback pending FIFO requests is full.\nNote: Writeback pending FIFO has six entries",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load uops that miss the STLB  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts load uops with true STLB miss retired to the architected path. True STLB miss is an uop triggering page walk that gets completed without blocks, and later gets retired. This page walk can end up with or without a fault  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store uops that miss the STLB  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts store uops with true STLB miss retired to the architected path. True STLB miss is an uop triggering page walk that gets completed without blocks, and later gets retired. This page walk can end up with or without a fault  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load uops with locked access  Supports address when precise.  Spec update: BDM35 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts load uops with locked access retired to the architected path  Supports address when precise.  Spec update: BDM35 (Precise event)",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load uops that split across a cacheline boundary  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts line-splitted load uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K)  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts line-splitted store uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K)  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load uops  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts load uops retired to the architected path with a filter on bits 0 and 1 applied.\nNote: This event counts AVX-256bit load/store double-pump memory uops as a single uop at retirement. This event also counts SW prefetches  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store uops  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts store uops retired to the architected path with a filter on bits 0 and 1 applied.\nNote: This event counts AVX-256bit load/store double-pump memory uops as a single uop at retirement  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load uops with L1 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were hits in the nearest-level (L1) cache.\nNote: Only two data-sources of L1/FB are applicable for AVX-256bit  even though the corresponding AVX load could be serviced by a deeper level in the memory hierarchy. Data source is reported for the Low-half load. This event also counts SW prefetches independent of the actual data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load uops with L2 cache hits as data sources  Supports address when precise.  Spec update: BDM35 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were hits in the mid-level (L2) cache  Supports address when precise.  Spec update: BDM35 (Precise event)",
},
{
	.name = "mem_load_uops_retired.l3_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load uops which data sources were data hits in L3 without snoops required  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were data hits in the last-level (L3) cache without snoops required  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load uops misses in L1 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were misses in the nearest-level (L1) cache. Counting excludes unknown and UC data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Miss in mid-level (L2) cache. Excludes Unknown data-source  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were misses in the mid-level (L2) cache. Counting excludes unknown and UC data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l3_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Miss in last-level (L3) cache. Excludes Unknown data-source  Supports address when precise.  Spec update: BDM100, BDE70 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.hit_lfb",
	.event = "umask=0x40,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources were load uops missed L1 but hit FB due to preceding miss to the same cache line with data not ready  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were load uops missed L1 but hit a fill buffer due to a preceding miss to the same cache line with the data not ready.\nNote: Only two data-sources of L1/FB are applicable for AVX-256bit  even though the corresponding AVX load could be serviced by a deeper level in the memory hierarchy. Data source is reported for the Low-half load  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 hit and cross-core snoop missed in on-pkg core cache  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were L3 Hit and a cross-core snoop missed in the on-pkg core cache  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were L3 hit and a cross-core snoop hit in the on-pkg core cache  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were HitM responses from shared L3  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were HitM responses from a core on same socket (shared L3)  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load uops which data sources were hits in L3 without snoops required  Supports address when precise.  Spec update: BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops which data sources were hits in the last-level (L3) cache without snoops required  Supports address when precise.  Spec update: BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_l3_miss_retired.local_dram",
	.event = "umask=0x1,period=100007,event=0xd3",
	.desc = "Data from local DRAM either Snoop not needed or Snoop Miss (RspI)  Supports address when precise.  Spec update: BDE70, BDM100 (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load uop whose Data Source was: local DRAM either Snoop not needed or Snoop Miss (RspI)  Supports address when precise.  Spec update: BDE70, BDM100 (Precise event)",
},
{
	.name = "mem_load_uops_l3_miss_retired.remote_dram",
	.event = "umask=0x4,period=100007,event=0xd3",
	.desc = "Retired load uop whose Data Source was: remote DRAM either Snoop not needed or Snoop Miss (RspI)  Supports address when precise.  Spec update: BDE70 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_miss_retired.remote_hitm",
	.event = "umask=0x10,period=100007,event=0xd3",
	.desc = "Retired load uop whose Data Source was: Remote cache HITM  Supports address when precise.  Spec update: BDE70 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_miss_retired.remote_fwd",
	.event = "umask=0x20,period=100007,event=0xd3",
	.desc = "Retired load uop whose Data Source was: forwarded from remote cache  Supports address when precise.  Spec update: BDE70 (Precise event)",
	.topic = "cache",
},
{
	.name = "l2_trans.demand_data_rd",
	.event = "umask=0x1,period=200003,event=0xf0",
	.desc = "Demand Data Read requests that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts Demand Data Read requests that access L2 cache, including rejects",
},
{
	.name = "l2_trans.rfo",
	.event = "umask=0x2,period=200003,event=0xf0",
	.desc = "RFO requests that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts Read for Ownership (RFO) requests that access L2 cache",
},
{
	.name = "l2_trans.code_rd",
	.event = "umask=0x4,period=200003,event=0xf0",
	.desc = "L2 cache accesses when fetching instructions",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache accesses when fetching instructions",
},
{
	.name = "l2_trans.all_pf",
	.event = "umask=0x8,period=200003,event=0xf0",
	.desc = "L2 or L3 HW prefetches that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L2 or L3 HW prefetches that access L2 cache including rejects",
},
{
	.name = "l2_trans.l1d_wb",
	.event = "umask=0x10,period=200003,event=0xf0",
	.desc = "L1D writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L1D writebacks that access L2 cache",
},
{
	.name = "l2_trans.l2_fill",
	.event = "umask=0x20,period=200003,event=0xf0",
	.desc = "L2 fill requests that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L2 fill requests that access L2 cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L2 writebacks that access L2 cache",
},
{
	.name = "l2_trans.all_requests",
	.event = "umask=0x80,period=200003,event=0xf0",
	.desc = "Transactions accessing L2 pipe",
	.topic = "cache",
	.long_desc = "This event counts transactions that access the L2 pipe including snoops, pagewalks, and so on",
},
{
	.name = "l2_lines_in.i",
	.event = "umask=0x1,period=100003,event=0xf1",
	.desc = "L2 cache lines in I state filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines in the Invalidate state filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_in.s",
	.event = "umask=0x2,period=100003,event=0xf1",
	.desc = "L2 cache lines in S state filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines in the Shared state filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_in.e",
	.event = "umask=0x4,period=100003,event=0xf1",
	.desc = "L2 cache lines in E state filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines in the Exclusive state filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x7,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x5,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by demand",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Split locks in SQ",
	.topic = "cache",
	.long_desc = "This event counts the number of split locks in the super queue",
},
{
	.name = "offcore_response.all_requests.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C8FFF",
	.desc = "Counts all requests hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all requests hit in the L3",
},
{
	.name = "offcore_response.all_reads.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C07F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.all_reads.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C07F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_code_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0244",
	.desc = "Counts all demand & prefetch code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0122",
	.desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.all_rfo.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0122",
	.desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0091",
	.desc = "Counts all demand & prefetch data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0091",
	.desc = "Counts all demand & prefetch data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads hit in the L3",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads hit in the L3",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs hit in the L3",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0002",
	.desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0002",
	.desc = "Counts all demand data writes (RFOs) hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) hit in the L3",
},
{
	.name = "other_assists.avx_to_sse",
	.event = "umask=0x8,period=100003,event=0xc1",
	.desc = "Number of transitions from AVX-256 to legacy SSE when penalty applicable  Spec update: BDM30",
	.topic = "floating point",
	.long_desc = "This event counts the number of transitions from AVX-256 to legacy SSE when penalty is applicable  Spec update: BDM30",
},
{
	.name = "other_assists.sse_to_avx",
	.event = "umask=0x10,period=100003,event=0xc1",
	.desc = "Number of transitions from SSE to AVX-256 when penalty applicable  Spec update: BDM30",
	.topic = "floating point",
	.long_desc = "This event counts the number of transitions from legacy SSE to AVX-256 when penalty is applicable  Spec update: BDM30",
},
{
	.name = "fp_arith_inst_retired.scalar_double",
	.event = "umask=0x1,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar double precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.scalar_single",
	.event = "umask=0x2,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar single precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.scalar",
	.event = "umask=0x3,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar floating-point instructions retired. Applies to SSE* and AVX* scalar, double and single precision floating-point: ADD SUB MUL DIV MIN MAX RSQRT RCP SQRT FM(N)ADD/SUB. FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element. (RSQRT for single precision?)",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_double",
	.event = "umask=0x4,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed double precision floating-point instructions retired.  Each count represents 2 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_single",
	.event = "umask=0x8,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed single precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT RSQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_double",
	.event = "umask=0x10,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed double precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.double",
	.event = "umask=0x15,period=2000006,event=0xc7",
	.desc = "Number of SSE/AVX computational double precision floating-point instructions retired. Applies to SSE* and AVX*scalar, double and single precision floating-point: ADD SUB MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_single",
	.event = "umask=0x20,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed single precision floating-point instructions retired.  Each count represents 8 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT RSQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.single",
	.event = "umask=0x2a,period=2000005,event=0xc7",
	.desc = "Number of SSE/AVX computational single precision floating-point instructions retired. Applies to SSE* and AVX*scalar, double and single precision floating-point: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.packed",
	.event = "umask=0x3c,period=2000004,event=0xc7",
	.desc = "Number of SSE/AVX computational packed floating-point instructions retired. Applies to SSE* and AVX*, packed, double and single precision floating-point: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element. (RSQRT for single-precision?)",
	.topic = "floating point",
},
{
	.name = "fp_assist.x87_output",
	.event = "umask=0x2,period=100003,event=0xca",
	.desc = "Number of X87 assists due to output value",
	.topic = "floating point",
	.long_desc = "This event counts the number of x87 floating point (FP) micro-code assist (numeric overflow/underflow, inexact result) when the output value (destination register) is invalid",
},
{
	.name = "fp_assist.x87_input",
	.event = "umask=0x4,period=100003,event=0xca",
	.desc = "Number of X87 assists due to input value",
	.topic = "floating point",
	.long_desc = "This event counts x87 floating point (FP) micro-code assist (invalid operation, denormal operand, SNaN operand) when the input value (one of the source operands to an FP instruction) is invalid",
},
{
	.name = "fp_assist.simd_output",
	.event = "umask=0x8,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to Output values",
	.topic = "floating point",
	.long_desc = "This event counts the number of SSE* floating point (FP) micro-code assist (numeric overflow/underflow) when the output value (destination register) is invalid. Counting covers only cases involving penalties that require micro-code assist intervention",
},
{
	.name = "fp_assist.simd_input",
	.event = "umask=0x10,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to input values",
	.topic = "floating point",
	.long_desc = "This event counts any input SSE* FP assist - invalid operation, denormal operand, dividing by zero, SNaN operand. Counting includes only cases involving penalties that required micro-code assist intervention",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,cmask=1,period=100003,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
	.long_desc = "This event counts cycles with any input and output SSE or x87 FP assist. If an input and output assist are detected on the same cycle the event increments by 1",
},
{
	.name = "idq.empty",
	.event = "umask=0x2,period=2000003,event=0x79",
	.desc = "Instruction Decode Queue (IDQ) empty cycles",
	.topic = "frontend",
	.long_desc = "This counts the number of cycles that the instruction decoder queue is empty and can indicate that the application may be bound in the front end.  It does not determine whether there are uops being delivered to the Alloc stage since uops can be delivered by bypass skipping the Instruction Decode Queue (IDQ) when it is empty",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_dsb_uops",
	.event = "umask=0x10,period=2000003,event=0x79",
	.desc = "Uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_dsb_occur",
	.event = "edge=1,umask=0x10,cmask=1,period=2000003,event=0x79",
	.desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the number of deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while the Microcode Sequencer (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles 4  uops were  delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles  uops were  delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while the Microcode Sequenser (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles 4  uops were  delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles  uops were delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the total number of uops delivered to Instruction Decode Queue (IDQ) while the Microcode Sequenser (MS) is busy. Counting includes uops that may bypass the IDQ. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequenser (MS) is busy. Counting includes uops that may bypass the IDQ. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "idq.ms_switches",
	.event = "edge=1,umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
},
{
	.name = "idq.mite_all_uops",
	.event = "umask=0x3c,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=2000003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
	.topic = "frontend",
	.long_desc = "This event counts the number of both cacheable and noncacheable Instruction Cache, Streaming Buffer and Victim Cache Reads including UC fetches",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Misses. Includes Uncacheable accesses",
	.topic = "frontend",
	.long_desc = "This event counts the number of instruction cache, streaming buffer and victim cache misses. Counting includes UC accesses",
},
{
	.name = "icache.ifdata_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction-cache miss",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which the demand fetch waits for data (wfdM104H) from L2 or iSB (opportunistic hit)",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops not delivered to Resource Allocation Table (RAT) per thread adding 4  x when Resource Allocation Table (RAT) is not stalled and Instruction Decode Queue (IDQ) delivers x uops to Resource Allocation Table (RAT) (where x belongs to {0,1,2,3}). Counting does not cover cases when:\n a. IDQ-Resource Allocation Table (RAT) pipe serves the other thread;\n b. Resource Allocation Table (RAT) is stalled for the thread (including uop drops and clear BE conditions); \n c. Instruction Decode Queue (IDQ) delivers four uops",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,cmask=4,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "This event counts, on the per-thread basis, cycles when no uops are delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core =4",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,cmask=3,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "This event counts, on the per-thread basis, cycles when less than 1 uop is  delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core >=3",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,cmask=2,period=2000003,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "This event counts Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles. These cycles do not include uops routed through because of the switch itself, for example, when Instruction Decode Queue (IDQ) pre-allocation is unavailable, or Instruction Decode Queue (IDQ) is full. SBD-to-MITE switch true penalty cycles happen after the merge mux (MM) receives Decode Stream Buffer (DSB) Sync-indication until receiving the first MITE uop. \nMM is placed before Instruction Decode Queue (IDQ) to merge uops being fed from the MITE and Decode Stream Buffer (DSB) paths. Decode Stream Buffer (DSB) inserts the Sync-indication whenever a Decode Stream Buffer (DSB)-to-MITE switch occurs.\nPenalty: A Decode Stream Buffer (DSB) hit followed by a Decode Stream Buffer (DSB) miss can cost up to six cycles in which no uops are delivered to the IDQ. Most often, such switches from the Decode Stream Buffer (DSB) to the legacy pipeline cost 02 cycles",
},
{
	.name = "misalign_mem_ref.loads",
	.event = "umask=0x1,period=2000003,event=0x5",
	.desc = "Speculative cache line split load uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "This event counts speculative cache-line split load uops dispatched to the L1 cache",
},
{
	.name = "misalign_mem_ref.stores",
	.event = "umask=0x2,period=2000003,event=0x5",
	.desc = "Speculative cache line split STA uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "This event counts speculative cache line split store-address (STA) uops dispatched to the L1 cache",
},
{
	.name = "tx_mem.abort_conflict",
	.event = "umask=0x1,period=2000003,event=0x54",
	.desc = "Number of times a TSX line had a cache conflict",
	.topic = "memory",
	.long_desc = "Number of times a TSX line had a cache conflict",
},
{
	.name = "tx_mem.abort_capacity_write",
	.event = "umask=0x2,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to an evicted line caused by a transaction overflow",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to an evicted line caused by a transaction overflow",
},
{
	.name = "tx_mem.abort_hle_store_to_elided_lock",
	.event = "umask=0x4,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to a non-release/commit store to lock",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to a non-release/commit store to lock",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_not_empty",
	.event = "umask=0x8,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_mismatch",
	.event = "umask=0x10,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_unsupported_alignment",
	.event = "umask=0x20,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
},
{
	.name = "tx_mem.hle_elision_buffer_full",
	.event = "umask=0x40,period=2000003,event=0x54",
	.desc = "Number of times we could not allocate Lock Buffer",
	.topic = "memory",
	.long_desc = "Number of times we could not allocate Lock Buffer",
},
{
	.name = "tx_exec.misc1",
	.event = "umask=0x1,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions that may cause a transactional abort was executed. Since this is the count of execution, it may not always cause a transactional abort",
	.topic = "memory",
},
{
	.name = "tx_exec.misc2",
	.event = "umask=0x2,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions (e.g., vzeroupper) that may cause a transactional abort was executed inside a transactional region",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by  a vzeroupper instruction",
},
{
	.name = "tx_exec.misc3",
	.event = "umask=0x4,period=2000003,event=0x5d",
	.desc = "Counts the number of times an instruction execution caused the transactional nest count supported to be exceeded",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by a nest count that is too deep",
},
{
	.name = "tx_exec.misc4",
	.event = "umask=0x8,period=2000003,event=0x5d",
	.desc = "Counts the number of times a XBEGIN instruction was executed inside an HLE transactional region",
	.topic = "memory",
	.long_desc = "RTM region detected inside HLE",
},
{
	.name = "tx_exec.misc5",
	.event = "umask=0x10,period=2000003,event=0x5d",
	.desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
	.topic = "memory",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts",
	.topic = "memory",
	.long_desc = "This event counts the number of memory ordering Machine Clears detected. Memory Ordering Machine Clears can result from one of the following:\n1. memory disambiguation,\n2. external snoop, or\n3. cross SMT-HW-thread snoop (stores) hitting load buffer",
},
{
	.name = "hle_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc8",
	.desc = "Number of times we entered an HLE region; does not count nested transactions",
	.topic = "memory",
	.long_desc = "Number of times we entered an HLE region\n does not count nested transactions",
},
{
	.name = "hle_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc8",
	.desc = "Number of times HLE commit succeeded",
	.topic = "memory",
	.long_desc = "Number of times HLE commit succeeded",
},
{
	.name = "hle_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc8",
	.desc = "Number of times HLE abort was triggered (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times HLE abort was triggered (Precise event)",
},
{
	.name = "hle_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an HLE abort was attributed to a Memory condition (See TSX_Memory event for additional details)",
},
{
	.name = "hle_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to uncommon conditions",
	.topic = "memory",
	.long_desc = "Number of times the TSX watchdog signaled an HLE abort",
},
{
	.name = "hle_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Number of times a disallowed operation caused an HLE abort",
},
{
	.name = "hle_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times HLE caused a fault",
},
{
	.name = "hle_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to none of the previous 4 categories (e.g. interrupts)",
	.topic = "memory",
	.long_desc = "Number of times HLE aborted and was not due to the abort conditions in subevents 3-6",
},
{
	.name = "rtm_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc9",
	.desc = "Number of times we entered an RTM region; does not count nested transactions",
	.topic = "memory",
	.long_desc = "Number of times we entered an RTM region\n does not count nested transactions",
},
{
	.name = "rtm_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc9",
	.desc = "Number of times RTM commit succeeded",
	.topic = "memory",
	.long_desc = "Number of times RTM commit succeeded",
},
{
	.name = "rtm_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc9",
	.desc = "Number of times RTM abort was triggered (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times RTM abort was triggered  (Precise event)",
},
{
	.name = "rtm_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an RTM abort was attributed to a Memory condition (See TSX_Memory event for additional details)",
},
{
	.name = "rtm_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times the TSX watchdog signaled an RTM abort",
},
{
	.name = "rtm_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Number of times a disallowed operation caused an RTM abort",
},
{
	.name = "rtm_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times a RTM caused a fault",
},
{
	.name = "rtm_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
	.topic = "memory",
	.long_desc = "Number of times RTM aborted and was not due to the abort conditions in subevents 3-6",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Randomly selected loads with latency value being above 4  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above four  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Randomly selected loads with latency value being above 8  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above eight  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Randomly selected loads with latency value being above 16  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 16  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Randomly selected loads with latency value being above 32  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 32  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Randomly selected loads with latency value being above 64  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 64  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Randomly selected loads with latency value being above 128  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 128  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Randomly selected loads with latency value being above 256  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 256  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Randomly selected loads with latency value being above 512  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 512  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "offcore_response.all_requests.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC08FFF",
	.desc = "Counts all requests miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all requests miss in the L3",
},
{
	.name = "offcore_response.all_reads.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x087FC007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and clean or shared data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and clean or shared data is transferred from remote cache",
},
{
	.name = "offcore_response.all_reads.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the modified data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the modified data is transferred from remote cache",
},
{
	.name = "offcore_response.all_reads.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063BC007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the data is returned from remote dram",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the data is returned from remote dram",
},
{
	.name = "offcore_response.all_reads.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x06040007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_reads.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss in the L3",
},
{
	.name = "offcore_response.all_code_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000244",
	.desc = "Counts all demand & prefetch code reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch code reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00244",
	.desc = "Counts all demand & prefetch code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch code reads miss in the L3",
},
{
	.name = "offcore_response.all_rfo.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000122",
	.desc = "Counts all demand & prefetch RFOs miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_rfo.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00122",
	.desc = "Counts all demand & prefetch RFOs miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs miss in the L3",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x087FC00091",
	.desc = "Counts all demand & prefetch data reads miss the L3 and clean or shared data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss the L3 and clean or shared data is transferred from remote cache",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00091",
	.desc = "Counts all demand & prefetch data reads miss the L3 and the modified data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss the L3 and the modified data is transferred from remote cache",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063BC00091",
	.desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from remote dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from remote dram",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000091",
	.desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00091",
	.desc = "Counts all demand & prefetch data reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss in the L3",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads miss in the L3",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs miss in the L3",
},
{
	.name = "offcore_response.demand_rfo.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00002",
	.desc = "Counts all demand data writes (RFOs) miss the L3 and the modified data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) miss the L3 and the modified data is transferred from remote cache",
},
{
	.name = "offcore_response.demand_rfo.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00002",
	.desc = "Counts all demand data writes (RFOs) miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) miss in the L3",
},
{
	.name = "cpl_cycles.ring0",
	.event = "umask=0x1,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when the thread is in ring 0",
	.topic = "other",
	.long_desc = "This event counts the unhalted core cycles during which the thread is in the ring 0 privileged mode",
},
{
	.name = "cpl_cycles.ring0_trans",
	.event = "edge=1,umask=0x1,cmask=1,period=100007,event=0x5c",
	.desc = "Number of intervals between processor halts while thread is in ring 0",
	.topic = "other",
	.long_desc = "This event counts when there is a transition from ring 1,2 or 3 to ring0",
},
{
	.name = "cpl_cycles.ring123",
	.event = "umask=0x2,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when thread is in rings 1, 2, or 3",
	.topic = "other",
	.long_desc = "This event counts unhalted core cycles during which the thread is in rings 1, 2, or 3",
},
{
	.name = "lock_cycles.split_lock_uc_lock_duration",
	.event = "umask=0x1,period=2000003,event=0x63",
	.desc = "Cycles when L1 and L2 are locked due to UC or split lock",
	.topic = "other",
	.long_desc = "This event counts cycles in which the L1 and L2 are locked due to a UC lock or split lock. A lock is asserted in case of locked memory access, due to noncacheable memory, locked operation that spans two cache lines, or a page walk from the noncacheable page table. L1D and L2 locks have a very high performance penalty and it is highly recommended to avoid such access",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, this event counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers. \nNotes: INST_RETIRED.ANY is counted by a designated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. INST_RETIRED.ANY_P is counted by a programmable counter and it is an architectural performance event. \nCounting: Faulting executions of GETSEC/VM entry/VM Exit/MWait will not count as retired instructions",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state. This event has a constant ratio with the CPU_CLK_UNHALTED.REF_XCLK event. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. \nNote: On all current platforms this event stops counting during 'throttling (TM)' states duty off periods the processor is 'halted'.  This event is clocked by base clock (100 Mhz) on Sandy Bridge. The counter update is done at a lower clock rate then the core clock the overflow status bit for this counter may appear 'sticky'.  After the counter has overflowed and software clears the overflow status bit and resets the counter to less than MAX. The reset value to the counter is not clocked immediately so the overflow status bit will flip 'high (1)' and generate another PMI (if enabled) after which the reset value gets clocked into the counter. Therefore, software will get the interrupt, read the overflow status bit '1 for bit 34 while the counter value is less than MAX. Software should ignore this case",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Cases when loads get true Block-on-Store blocking code preventing store forwarding",
	.topic = "pipeline",
	.long_desc = "This event counts how many times the load operation got the true Block-on-Store blocking code preventing store forwarding. This includes cases when:\n - preceding store conflicts with the load (incomplete overlap);\n - store forwarding is impossible due to u-arch limitations;\n - preceding lock RMW operations are not forwarded;\n - store has the no-forward bit set (uncacheable/page-split/masked stores);\n - all-blocking stores are used (mostly, fences and port I/O);\nand others.\nThe most common case is a load blocked due to its address range overlapping with a preceding smaller uncompleted store. Note: This event does not take into account cases of out-of-SW-control (for example, SbTailHit), unknown physical STA, and cases of blocking loads on store due to being non-WB memory type or a lock. These cases are covered by other events.\nSee the table of not supported store forwards in the Optimization Guide",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "This event counts the number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare",
	.topic = "pipeline",
	.long_desc = "This event counts false dependencies in MOB when the partial comparison upon loose net check and dependency was resolved by the Enhanced Loose net mechanism. This may not result in high performance penalties. Loose net checks can fail when loads and stores are 4k aliased",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x3,cmask=1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for this thread (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "Cycles checkpoints in Resource Allocation Table (RAT) are recovering from JEClear or machine clear",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x3,any=1,cmask=1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "int_misc.rat_stall_cycles",
	.event = "umask=0x8,period=2000003,event=0xd",
	.desc = "Cycles when Resource Allocation Table (RAT) external stall is sent to Instruction Decode Queue (IDQ) for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts the number of cycles during which Resource Allocation Table (RAT) external stall is sent to Instruction Decode Queue (IDQ) for the current thread. This also includes the cycles during which the Allocator is serving another thread",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of Uops issued by the Resource Allocation Table (RAT) to the reservation station (RS)",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which the Resource Allocation Table (RAT) does not issue any Uops to the reservation station (RS) for the current thread",
},
{
	.name = "uops_issued.flags_merge",
	.event = "umask=0x10,period=2000003,event=0xe",
	.desc = "Number of flags-merge uops being allocated. Such uops considered perf sensitive; added by GSR u-arch",
	.topic = "pipeline",
	.long_desc = "Number of flags-merge uops being allocated. Such uops considered perf sensitive\n added by GSR u-arch",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
},
{
	.name = "uops_issued.single_mul",
	.event = "umask=0x40,period=2000003,event=0xe",
	.desc = "Number of Multiply packed/scalar single precision uops allocated",
	.topic = "pipeline",
},
{
	.name = "arith.fpu_div_active",
	.event = "umask=0x1,period=2000003,event=0x14",
	.desc = "Cycles when divider is busy executing divide operations",
	.topic = "pipeline",
	.long_desc = "This event counts the number of the divide operations executed. Uses edge-detect and a cmask value of 1 on ARITH.FPU_DIV_ACTIVE to get the number of the divide operations executed",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This is an architectural event that counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling. For this reason, this event may have a changing ratio with regards to wall clock time",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "This is a fixed-frequency event programmed to general counters. It counts when the core is unhalted at 100 Mhz",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "This event counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the software prefetch. It can also be incremented by some lock instructions. So it should only be used with profiling so that the locks can be excluded by asm inspection of the nearby instructions",
},
{
	.name = "load_hit_pre.hw_pf",
	.event = "umask=0x2,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for hardware prefetch",
	.topic = "pipeline",
	.long_desc = "This event counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the hardware prefetch",
},
{
	.name = "move_elimination.int_eliminated",
	.event = "umask=0x1,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_eliminated",
	.event = "umask=0x2,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.int_not_eliminated",
	.event = "umask=0x4,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_not_eliminated",
	.event = "umask=0x8,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which the reservation station (RS) is empty for the thread.\nNote: In ST-mode, not active thread should drive 0. This is usually caused by severely costly branch mispredictions, or allocator/FE issues",
},
{
	.name = "rs_events.empty_end",
	.event = "edge=1,inv=1,umask=0x1,cmask=1,period=200003,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
	.long_desc = "This event counts stalls occured due to changing prefix length (66, 67 or REX.W when they change the length of the decoded instruction). Occurrences counting is proportional to the number of prefixes in a 16B-line. This may result in the following penalties: three-cycle penalty for each LCP in a 16-byte chunk",
},
{
	.name = "br_inst_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x88",
	.desc = "Not taken macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts not taken macro-conditional branch instructions",
},
{
	.name = "br_inst_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired macro-conditional branch instructions",
},
{
	.name = "br_inst_exec.taken_direct_jump",
	.event = "umask=0x82,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired macro-conditional branch instructions excluding calls and indirect branches",
},
{
	.name = "br_inst_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired indirect branches excluding calls and return branches",
},
{
	.name = "br_inst_exec.taken_indirect_near_return",
	.event = "umask=0x88,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired indirect branches that have a return mnemonic",
},
{
	.name = "br_inst_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x88",
	.desc = "Taken speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect calls",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired indirect calls including both register and memory indirect",
},
{
	.name = "br_inst_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x88",
	.desc = "Speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired macro-conditional branch instructions",
},
{
	.name = "br_inst_exec.all_direct_jmp",
	.event = "umask=0xc2,period=200003,event=0x88",
	.desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired macro-unconditional branch instructions, excluding calls and indirects",
},
{
	.name = "br_inst_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x88",
	.desc = "Speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired indirect branches excluding calls and return branches",
},
{
	.name = "br_inst_exec.all_indirect_near_return",
	.event = "umask=0xc8,period=200003,event=0x88",
	.desc = "Speculative and retired indirect return branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired indirect branches that have a return mnemonic",
},
{
	.name = "br_inst_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x88",
	.desc = "Speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x88",
	.desc = "Speculative and retired  branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired branch instructions",
},
{
	.name = "br_misp_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x89",
	.desc = "Not taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts not taken speculative and retired mispredicted macro conditional branch instructions",
},
{
	.name = "br_misp_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired mispredicted macro conditional branch instructions",
},
{
	.name = "br_misp_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.taken_return_near",
	.event = "umask=0x88,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired mispredicted indirect branches that have a return mnemonic",
},
{
	.name = "br_misp_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired mispredicted macro conditional branch instructions",
},
{
	.name = "br_misp_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x89",
	.desc = "Mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired mispredicted branch instructions",
},
{
	.name = "uop_dispatches_cancelled.simd_prf",
	.event = "umask=0x3,period=2000003,event=0xa0",
	.desc = "Micro-op dispatches cancelled due to insufficient SIMD physical register file read ports",
	.topic = "pipeline",
	.long_desc = "This event counts the number of micro-operations cancelled after they were dispatched from the scheduler to the execution units when the total number of physical register read ports across all dispatch ports exceeds the read bandwidth of the physical register file.  The SIMD_PRF subevent applies to the following instructions: VDPPS, DPPS, VPCMPESTRI, PCMPESTRI, VPCMPESTRM, PCMPESTRM, VFMADD*, VFMADDSUB*, VFMSUB*, VMSUBADD*, VFNMADD*, VFNMSUB*.  See the Broadwell Optimization Guide for more information",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_executed_port.port_0_core",
	.event = "umask=0x1,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_executed_port.port_1_core",
	.event = "umask=0x2,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 2",
},
{
	.name = "uops_executed_port.port_2_core",
	.event = "umask=0x4,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 2",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 2",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 3",
},
{
	.name = "uops_executed_port.port_3_core",
	.event = "umask=0x8,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 3",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 3",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 4",
},
{
	.name = "uops_executed_port.port_4_core",
	.event = "umask=0x10,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 4",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_executed_port.port_5_core",
	.event = "umask=0x20,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_dispatched_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_executed_port.port_6_core",
	.event = "umask=0x40,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 6",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_dispatched_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 7",
},
{
	.name = "uops_executed_port.port_7_core",
	.event = "umask=0x80,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 7",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 7",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
	.long_desc = "This event counts resource-related stall cycles",
},
{
	.name = "resource_stalls.rs",
	.event = "umask=0x4,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no eligible RS entry available",
	.topic = "pipeline",
	.long_desc = "This event counts stall cycles caused by absence of eligible entries in the reservation station (RS). This may result from RS overflow, or from RS deallocation because of the RS array Write Port allocation scheme (each RS entry has two write ports instead of four. As a result, empty entries could not be used, although RS is not really full). This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "This event counts stall cycles caused by the store buffer (SB) overflow (excluding draining from synch). This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "resource_stalls.rob",
	.event = "umask=0x10,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to re-order buffer full",
	.topic = "pipeline",
	.long_desc = "This event counts ROB full stall cycles. This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "cycle_activity.cycles_l2_pending",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles the CPU has at least one pending  demand* load request missing the L2 cache",
},
{
	.name = "cycle_activity.cycles_l2_miss",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_ldm_pending",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles the CPU has at least one pending  demand load request (that is cycles with non-completed load waiting for its data from memory subsystem)",
},
{
	.name = "cycle_activity.cycles_mem_any",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_no_execute",
	.event = "umask=0x4,cmask=4,period=2000003,event=0xa3",
	.desc = "This event increments by 1 for every cycle where there was no execute for this thread",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port",
},
{
	.name = "cycle_activity.stalls_total",
	.event = "umask=0x4,cmask=4,period=2000003,event=0xa3",
	.desc = "Total execution stalls",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_pending",
	.event = "umask=0x5,cmask=5,period=2000003,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port, while there was at least one pending demand* load request missing the L2 cache.(as a footprint) * includes also L1 HW prefetch requests that may or may not be required by demands",
},
{
	.name = "cycle_activity.stalls_l2_miss",
	.event = "umask=0x5,cmask=5,period=2000003,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_ldm_pending",
	.event = "umask=0x6,cmask=6,period=2000003,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port, while there was at least one pending demand load request",
},
{
	.name = "cycle_activity.stalls_mem_any",
	.event = "umask=0x6,cmask=6,period=2000003,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_pending",
	.event = "umask=0x8,cmask=8,period=2000003,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles the CPU has at least one pending  demand load request missing the L1 data cache",
},
{
	.name = "cycle_activity.cycles_l1d_miss",
	.event = "umask=0x8,cmask=8,period=2000003,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_pending",
	.event = "umask=0xc,cmask=12,period=2000003,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port, while there was at least one pending demand load request missing the L1 data cache",
},
{
	.name = "cycle_activity.stalls_l1d_miss",
	.event = "umask=0xc,cmask=12,period=2000003,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "uops_executed.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Counts the number of uops to be executed per-thread each cycle",
	.topic = "pipeline",
	.long_desc = "Number of uops to be executed per-thread each cycle",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which no uops were dispatched from the Reservation Station (RS) per thread",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core",
	.topic = "pipeline",
	.long_desc = "Number of uops executed from any thread",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter   - architectural event  Spec update: BDM61",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions (EOMs) retired. Counting covers macro-fused instructions individually (that is, increments by two)  Spec update: BDM61",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: BDM11, BDM55 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts instructions retired  Spec update: BDM11, BDM55 (Must be precise)",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000003,event=0xc0",
	.desc = "FP operations  retired. X87 FP operations that have no exceptions:",
	.topic = "pipeline",
	.long_desc = "This event counts FP operations retired. For X87 FP operations that have no exceptions counting also includes flows that have several X87, or flows that use X87 uops in the exception handling",
},
{
	.name = "other_assists.any_wb_assist",
	.event = "umask=0x40,period=100003,event=0xc1",
	.desc = "Number of times any microcode assist is invoked by HW upon uop writeback",
	.topic = "pipeline",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Actually retired uops  Supports address when precise (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts all actually retired uops. Counting increments by two for micro-fused uops, and by one for macro-fused and other uops. Maximal increment value for one cycle is eight  Supports address when precise (Precise event)",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
	.long_desc = "This event counts cycles without actually retired uops",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,cmask=10,period=2000003,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
	.long_desc = "Number of cycles using always true condition (uops_ret < 16) applied to non PEBS uops retired event",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of retirement slots used (Precise event)",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=2000003,event=0xc3",
	.desc = "Cycles there was a Nuke. Account for both thread-specific and All Thread Nukes",
	.topic = "pipeline",
	.long_desc = "This event counts both thread-specific (TS) and all-thread (AT) nukes",
},
{
	.name = "machine_clears.count",
	.event = "edge=1,umask=0x1,cmask=1,period=100003,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "This event counts self-modifying code (SMC) detected, which causes a machine clear",
},
{
	.name = "machine_clears.maskmov",
	.event = "umask=0x20,period=100003,event=0xc3",
	.desc = "This event counts the number of executed Intel AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
	.topic = "pipeline",
	.long_desc = "Maskmov false fault - counts number of time ucode passes through Maskmov flow due to instruction's mask being 0 while the flow was completed without raising a fault",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts all (macro) branch instructions retired",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts conditional branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts both direct and indirect near call instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_call_r3",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect macro near call instructions retired (captured in ring 3) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts both direct and indirect macro near call instructions retired (captured in ring 3) (Precise event)",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired. (Precise Event - PEBS)  Spec update: BDW98 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_INST_RETIRED.ALL_BRANCHES that counts all (macro) branch instructions retired  Spec update: BDW98 (Must be precise)",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts return instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Not taken branch instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts not taken branch instructions retired",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts taken branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Far branch instructions retired  Spec update: BDW98",
	.topic = "pipeline",
	.long_desc = "This event counts far branch instructions retired  Spec update: BDW98",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts all mispredicted macro branch instructions retired",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts mispredicted conditional branch instructions retired (Precise event)",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired. (Precise Event - PEBS) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_MISP_RETIRED.ALL_BRANCHES that counts all mispredicted macro branch instructions retired (Must be precise)",
},
{
	.name = "br_misp_retired.ret",
	.event = "umask=0x8,period=100007,event=0xc5",
	.desc = "This event counts the number of mispredicted ret instructions retired. Non PEBS (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts mispredicted return instructions retired (Precise event)",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "number of near branch instructions retired that were mispredicted and taken (Precise event)",
	.topic = "pipeline",
	.long_desc = "Number of near branch instructions retired that were mispredicted and taken (Precise event)",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Count cases of saving new LBR",
	.topic = "pipeline",
	.long_desc = "This event counts cases of saving new LBR records by hardware. This assumes proper enabling of LBRs and takes into account LBR filtering done by the LBR_SELECT register",
},
{
	.name = "baclears.any",
	.event = "umask=0x1f,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
},
{
	.name = "unc_c_clockticks",
	.event = "event=0",
	.desc = "Uncore cache clock ticks. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_lookup.any",
	.event = "umask=0x11,event=0x34,filter_state=0x1",
	.desc = "All LLC Misses (code+ data rd + data wr - including demand and prefetch). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_victims.m_state",
	.event = "umask=0x1,event=0x37",
	.desc = "M line evictions from LLC (writebacks to memory). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.data_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x182",
	.desc = "LLC misses - demand and prefetch data reads - excludes LLC prefetches. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.uncacheable",
	.event = "umask=0x3,event=0x35,filter_opc=0x187",
	.desc = "LLC misses - Uncacheable reads (from cpu) . Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x187,filter_nc=1",
	.desc = "MMIO reads. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x18f,filter_nc=1",
	.desc = "MMIO writes. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.rfo_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x190",
	.desc = "LLC prefetch misses for RFO. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.code_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x191",
	.desc = "LLC prefetch misses for code reads. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.data_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x192",
	.desc = "LLC prefetch misses for data reads. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x19e",
	.desc = "LLC misses for PCIe read current. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x1c8",
	.desc = "ItoM write misses (as part of fast string memcpy stores) + PCIe full line writes. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_non_snoop_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x1c8,filter_tid=0x3e",
	.desc = "PCIe write misses (full cache line). Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_ns_partial_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x180,filter_tid=0x3e",
	.desc = "PCIe writes (partial cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "llc_references.code_llc_prefetch",
	.event = "umask=0x1,event=0x35,filter_opc=0x181",
	.desc = "L2 demand and L2 prefetch code references to LLC. Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_full",
	.event = "umask=0x1,event=0x35,filter_opc=0x18c",
	.desc = "Streaming stores (full cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_partial",
	.event = "umask=0x1,event=0x35,filter_opc=0x18d",
	.desc = "Streaming stores (partial cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_read",
	.event = "umask=0x1,event=0x35,filter_opc=0x19e",
	.desc = "PCIe read current. Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x1c8,filter_tid=0x3e",
	.desc = "PCIe write references (full cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_tor_occupancy.llc_data_read",
	.event = "umask=0x3,event=0x36,filter_opc=0x182",
	.desc = "Occupancy counter for LLC data reads (demand and L2 prefetch). Derived from unc_c_tor_occupancy.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads",
	.event = "umask=0x3,event=0x1",
	.desc = "read requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads_local",
	.event = "umask=0x1,event=0x1",
	.desc = "read requests to local home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads_remote",
	.event = "umask=0x2,event=0x1",
	.desc = "read requests to remote home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes",
	.event = "umask=0xC,event=0x1",
	.desc = "write requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes_local",
	.event = "umask=0x4,event=0x1",
	.desc = "write requests to local home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes_remote",
	.event = "umask=0x8,event=0x1",
	.desc = "write requests to remote home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspcnflct",
	.event = "umask=0x40,event=0x21",
	.desc = "Conflict requests (requests for same address from multiple agents simultaneously). Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rsp_fwd_wb",
	.event = "umask=0x20,event=0x21",
	.desc = "M line forwarded from remote cache along with writeback to memory. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspifwd",
	.event = "umask=0x4,event=0x21",
	.desc = "M line forwarded from remote cache with no writeback to memory. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rsps",
	.event = "umask=0x2,event=0x21",
	.desc = "Shared line response from remote cache. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspsfwd",
	.event = "umask=0x8,event=0x21",
	.desc = "Shared line forwarded from remote cache. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_q_clockticks",
	.event = "event=0x14",
	.desc = "QPI clock ticks. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.perpkg = "1",
},
{
	.name = "qpi_data_bandwidth_tx",
	.event = "umask=0x2,event=0",
	.desc = "Number of data flits transmitted . Derived from unc_q_txl_flits_g0.data. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.unit = "8Bytes",
	.perpkg = "1",
},
{
	.name = "qpi_ctl_bandwidth_tx",
	.event = "umask=0x4,event=0",
	.desc = "Number of non data (control) flits transmitted . Derived from unc_q_txl_flits_g0.non_data. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.unit = "8Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_read",
	.event = "umask=0x3,event=0x4",
	.desc = "read requests to memory controller. Derived from unc_m_cas_count.rd. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_write",
	.event = "umask=0xC,event=0x4",
	.desc = "write requests to memory controller. Derived from unc_m_cas_count.wr. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_clockticks",
	.event = "event=0",
	.desc = "Memory controller clock ticks. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_power_channel_ppd",
	.event = "event=0x85",
	.desc = "Cycles where DRAM ranks are in power down (CKE) mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_channel_ppd / unc_m_clockticks) * 100.",
	.metric_name = "power_channel_ppd %",
},
{
	.name = "unc_m_power_critical_throttle_cycles",
	.event = "event=0x86",
	.desc = "Cycles all ranks are in critical thermal throttle. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_critical_throttle_cycles / unc_m_clockticks) * 100.",
	.metric_name = "power_critical_throttle_cycles %",
},
{
	.name = "unc_m_power_self_refresh",
	.event = "event=0x43",
	.desc = "Cycles Memory is in self refresh power mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_self_refresh / unc_m_clockticks) * 100.",
	.metric_name = "power_self_refresh %",
},
{
	.name = "unc_m_pre_count.page_miss",
	.event = "umask=0x1,event=0x2",
	.desc = "Pre-charges due to page misses. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.rd",
	.event = "umask=0x4,event=0x2",
	.desc = "Pre-charge for reads. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.wr",
	.event = "umask=0x8,event=0x2",
	.desc = "Pre-charge for writes. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_p_clockticks",
	.event = "event=0",
	.desc = "PCU clock ticks. Use to get percentages of PCU cycles events. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
},
{
	.name = "unc_p_power_state_occupancy.cores_c0",
	.event = "event=0x80,occ_sel=1",
	.desc = "This is an occupancy event that tracks the number of cores that are in C0.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c0 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c0 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c3",
	.event = "event=0x80,occ_sel=2",
	.desc = "This is an occupancy event that tracks the number of cores that are in C3.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c3 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c3 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c6",
	.event = "event=0x80,occ_sel=3",
	.desc = "This is an occupancy event that tracks the number of cores that are in C6.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events . Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c6 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c6 %",
},
{
	.name = "unc_p_prochot_external_cycles",
	.event = "event=0xa",
	.desc = "Counts the number of cycles that we are in external PROCHOT mode.  This mode is triggered when a sensor off the die determines that something off-die (like DRAM) is too hot and must throttle to avoid damaging the chip. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_prochot_external_cycles / unc_p_clockticks) * 100.",
	.metric_name = "prochot_external_cycles %",
},
{
	.name = "unc_p_freq_max_limit_thermal_cycles",
	.event = "event=0x4",
	.desc = "Counts the number of cycles when temperature is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_limit_thermal_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_limit_thermal_cycles %",
},
{
	.name = "unc_p_freq_max_os_cycles",
	.event = "event=0x6",
	.desc = "Counts the number of cycles when the OS is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_os_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_os_cycles %",
},
{
	.name = "unc_p_freq_max_power_cycles",
	.event = "event=0x5",
	.desc = "Counts the number of cycles when power is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_power_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_power_cycles %",
},
{
	.name = "unc_p_freq_trans_cycles",
	.event = "event=0x74",
	.desc = "Counts the number of cycles when current is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_trans_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_trans_cycles %",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause page walks of any page size (4K/2M/4M/1G)  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (4K)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause a completed page walk (4K page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (2M/4M)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed_1g",
	.event = "umask=0x8,period=2000003,event=0x8",
	.desc = "Load miss in all TLB levels causes a page walk that completes. (1G)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause a completed page walk (1G  page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes of any page size  Spec update: BDM69",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_duration",
	.event = "umask=0x10,period=2000003,event=0x8",
	.desc = "Cycles when PMH is busy with page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of cycles while PMH is busy with the page walk  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.stlb_hit_4k",
	.event = "umask=0x20,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit_2m",
	.event = "umask=0x40,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x60,period=2000003,event=0x8",
	.desc = "Load operations that miss the first DTLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause page walks of any page size (4K/2M/4M/1G)  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (4K)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (4K page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks (2M/4M)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks (1G)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (1G  page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks  Spec update: BDM69",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x49",
	.desc = "Cycles when PMH is busy with page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of cycles while PMH is busy with the page walk  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x49",
	.desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Cycle count for an Extended Page table walk",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles for an extended page table walk. The Extended Page directory cache differs from standard TLB caches by the operating system that use it. Virtual machine operating systems use the extended page directory cache, while guest operating systems use the standard TLB caches",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause page walks of any page size (4K/2M/4M/1G)  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (4K)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (4K page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (2M/4M)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x85",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (1G)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (1G  page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x85",
	.desc = "Misses in all ITLB levels that cause completed page walks  Spec update: BDM69",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Cycles when PMH is busy with page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of cycles while PMH is busy with the page walk  Spec update: BDM69",
},
{
	.name = "itlb_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x85",
	.desc = "Core misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x85",
	.desc = "Code misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x85",
	.desc = "Operations that miss the first ITLB level but hit the second and do not cause any page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of flushes of the big or small ITLB pages. Counting include both TLB Flush (covering all sets) and TLB Set Clear (set-specific)",
},
{
	.name = "page_walker_loads.dtlb_l1",
	.event = "umask=0x11,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L1+FB  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.dtlb_l2",
	.event = "umask=0x12,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L2  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.dtlb_l3",
	.event = "umask=0x14,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L3 + XSNP  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.dtlb_memory",
	.event = "umask=0x18,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in Memory  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.itlb_l1",
	.event = "umask=0x21,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L1+FB  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.itlb_l2",
	.event = "umask=0x22,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L2  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.itlb_l3",
	.event = "umask=0x24,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L3 + XSNP  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of any STLB flush attempts (such as entire, VPID, PCID, InvPage, CR3 write, and so on)",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_cascadelakex[] = {
{
	.name = "l2_rqsts.demand_data_rd_miss",
	.event = "umask=0x21,period=200003,event=0x24",
	.desc = "Demand Data Read miss L2, no rejects",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests that miss L2 cache. Only not rejected loads are counted",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x22,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts the RFO (Read-for-Ownership) requests that miss L2 cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x24,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
	.long_desc = "Counts L2 cache misses when fetching instructions",
},
{
	.name = "l2_rqsts.all_demand_miss",
	.event = "umask=0x27,period=200003,event=0x24",
	.desc = "Demand requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Demand requests that miss L2 cache",
},
{
	.name = "l2_rqsts.pf_miss",
	.event = "umask=0x38,period=200003,event=0x24",
	.desc = "Requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that miss L2 cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0x3f,period=200003,event=0x24",
	.desc = "All requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "All requests that miss L2 cache",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0xc1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests, initiated by load instructions, that hit L2 cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0xc2,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the RFO (Read-for-Ownership) requests that hit L2 cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0xc4,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
	.long_desc = "Counts L2 cache hits when fetching instructions, code reads",
},
{
	.name = "l2_rqsts.pf_hit",
	.event = "umask=0xd8,period=200003,event=0x24",
	.desc = "Requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that hit L2 cache",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0xe1,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests (including requests from L1D hardware prefetchers). These loads may hit or miss L2 cache. Only non rejected loads are counted",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xe2,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "Counts the total number of RFO (read for ownership) requests to L2 cache. L2 RFO requests include both L1D demand RFO misses as well as L1D RFO prefetches",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0xe4,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "Counts the total number of L2 code requests",
},
{
	.name = "l2_rqsts.all_demand_references",
	.event = "umask=0xe7,period=200003,event=0x24",
	.desc = "Demand requests to L2 cache",
	.topic = "cache",
	.long_desc = "Demand requests to L2 cache",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xf8,period=200003,event=0x24",
	.desc = "Requests from the L1/L2/L3 hardware prefetchers or Load software prefetches",
	.topic = "cache",
	.long_desc = "Counts the total number of requests from the L2 hardware prefetchers",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200003,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
	.long_desc = "All L2 requests",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed L3  Spec update: SKL057",
	.topic = "cache",
	.long_desc = "Counts core-originated cacheable requests that miss the L3 cache (Longest Latency cache). Requests include data and code reads, Reads-for-Ownership (RFOs), speculative accesses and hardware prefetches from L1 and L2. It does not include all misses to the L3  Spec update: SKL057",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to L3  Spec update: SKL057",
	.topic = "cache",
	.long_desc = "Counts core-originated cacheable requests to the  L3 cache (Longest Latency cache). Requests include data and code reads, Reads-for-Ownership (RFOs), speculative accesses and hardware prefetches from L1 and L2.  It does not include all accesses to the L3  Spec update: SKL057",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
	.long_desc = "Counts duration of L1D miss outstanding in cycles",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss outstandings duration in cycles",
	.topic = "cache",
	.long_desc = "Counts duration of L1D miss outstanding, that is each cycle number of Fill Buffers (FB) outstanding required by Demand Reads. FB either is held by demand loads, or it is held by non-demand loads and gets hit at least once by demand. The valid outstanding interval is defined until the FB deallocation by one of the following ways: from FB allocation, if FB is allocated by demand from the demand Hit FB, if it is allocated by hardware or software prefetch.Note: In the L1D, a Demand Read contains cacheable or noncacheable demand loads, including ones causing cache-line splits and reads due to page walks resulted from any request type",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,event=0x48",
	.desc = "Number of times a request needed a FB entry but there was no entry available for it. That is the FB unavailability was dominant reason for blocking the request. A request includes cacheable/uncacheable demands that is load, store or SW prefetch",
	.topic = "cache",
	.long_desc = "Number of times a request needed a FB (Fill Buffer) entry but there was no entry available for it. A request includes cacheable/uncacheable demands that are load, store or SW prefetch instructions",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "Counts L1D data line replacements including opportunistic replacements, and replacements that require stall-for-replace or block-for-replace",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts cycles when offcore outstanding Demand Data Read transactions are present in the super queue (SQ). A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation)",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding Demand Data Read transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor. See the corresponding Umask under OFFCORE_REQUESTS.Note: A prefetch promoted to Demand is counted from the promotion point",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,cmask=6,period=2000003,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding Code Reads transactions in the SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding Code Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_code_rd",
	.event = "umask=0x2,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles with offcore outstanding Code Reads transactions in the SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding Code Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding RFO (store) transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles with offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding demand rfo Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts cycles when offcore outstanding cacheable Core Data Read transactions are present in the super queue. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding cacheable Core Data Read transactions in the super queue every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
	.long_desc = "Counts the Demand Data Read requests sent to uncore. Use it in conjunction with OFFCORE_REQUESTS_OUTSTANDING to determine average latency in the uncore",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "Counts both cacheable and non-cacheable code read requests",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "Counts the demand RFO (read for ownership) requests including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts the demand and prefetch data reads. All Core Data Reads include cacheable 'Demands' and L2 prefetchers (not L3 prefetchers). Counting also covers reads due to page walks resulted from any request type",
},
{
	.name = "offcore_requests.all_requests",
	.event = "umask=0x80,period=100003,event=0xb0",
	.desc = "Any memory transaction that reached the SQ",
	.topic = "cache",
	.long_desc = "Counts memory transactions reached the super queue including requests initiated by the core, all L3 prefetches, page walks, etc.",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Offcore requests buffer cannot take more entries for this thread core",
	.topic = "cache",
	.long_desc = "Counts the number of cases when the offcore requests buffer cannot take more entries for the core. This can happen when the superqueue does not contain eligible entries, or when L1D writeback pending FIFO requests is full.Note: Writeback pending FIFO has six entries",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "mem_inst_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load instructions that miss the STLB  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store instructions that miss the STLB  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load instructions with locked access  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load instructions that split across a cacheline boundary  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions that split across a cacheline boundary  Supports address when precise (Precise event)",
},
{
	.name = "mem_inst_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store instructions that split across a cacheline boundary  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired store instructions that split across a cacheline boundary  Supports address when precise (Precise event)",
},
{
	.name = "mem_inst_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load instructions  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store instructions  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load instructions with L1 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that hit in the L1 data cache. This event includes all SW prefetches and lock instructions regardless of the data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load instructions with L2 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions with L2 cache hits as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l3_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load instructions with L3 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that hit in the L3 cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load instructions missed L1 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that missed in the L1 cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Retired load instructions missed L2 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions missed L2 cache as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l3_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Retired load instructions missed L3 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that missed in the L3 cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.fb_hit",
	.event = "umask=0x40,period=100007,event=0xd1",
	.desc = "Retired load instructions which data sources were load missed L1 but hit FB due to preceding miss to the same cache line with data not ready  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop was load missed in L1 but hit FB (Fill Buffers) due to preceding miss to the same cache line with data not ready  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.local_pmm",
	.event = "umask=0x80,period=100003,event=0xd1",
	.desc = "Retired load instructions with local Intel\u00ae Optane\u2122 DC persistent memory as the data source where the data request missed all caches. Precise event  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with local Intel\u00ae Optane\u2122 DC persistent memory as the data source and the data request missed L3 (AppDirect or Memory Mode) and DRAM cache(Memory Mode). Precise event  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load instructions which data sources were L3 hit and cross-core snoop missed in on-pkg core cache  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load instructions which data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load instructions which data sources were HitM responses from shared L3  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources were HitM responses from shared L3  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load instructions which data sources were hits in L3 without snoops required  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources were hits in L3 without snoops required  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_miss_retired.local_dram",
	.event = "umask=0x1,period=100007,event=0xd3",
	.desc = "Retired load instructions which data sources missed L3 but serviced from local dram  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources missed L3 but serviced from local DRAM  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_miss_retired.remote_dram",
	.event = "umask=0x2,period=100007,event=0xd3",
	.desc = "Retired load instructions which data sources missed L3 but serviced from remote dram  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_l3_miss_retired.remote_hitm",
	.event = "umask=0x4,period=100007,event=0xd3",
	.desc = "Retired load instructions whose data sources was remote HITM  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions whose data sources was remote HITM  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_miss_retired.remote_fwd",
	.event = "umask=0x8,period=100007,event=0xd3",
	.desc = "Retired load instructions whose data sources was forwarded from a remote cache  Supports address when precise",
	.topic = "cache",
	.long_desc = "Retired load instructions whose data sources was forwarded from a remote cache  Supports address when precise",
},
{
	.name = "mem_load_l3_miss_retired.remote_pmm",
	.event = "umask=0x10,period=100007,event=0xd3",
	.desc = "Retired load instructions with remote Intel\u00ae Optane\u2122 DC persistent memory as the data source where the data request missed all caches. Precise event  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with remote Intel\u00ae Optane\u2122 DC persistent memory as the data source and the data request missed L3 (AppDirect or Memory Mode) and DRAM cache(Memory Mode). Precise event  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_misc_retired.uc",
	.event = "umask=0x4,period=100007,event=0xd4",
	.desc = "Retired instructions with at least 1 uncacheable load or lock  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "Counts L2 writebacks that access L2 cache",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x1f,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "Counts the number of L2 cache lines filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_out.silent",
	.event = "umask=0x1,period=200003,event=0xf2",
	.desc = "Counts the number of lines that are silently dropped by L2 cache when triggered by an L2 cache fill. These lines are typically in Shared state. A non-threaded event",
	.topic = "cache",
},
{
	.name = "l2_lines_out.non_silent",
	.event = "umask=0x2,period=200003,event=0xf2",
	.desc = "Counts the number of lines that are evicted by L2 cache when triggered by an L2 cache fill. Those lines can be either in modified state or clean state. Modified lines may either be written back to L3 or directly written to memory and not allocated in L3.  Clean lines may either be allocated in L3 or dropped",
	.topic = "cache",
	.long_desc = "Counts the number of lines that are evicted by L2 cache when triggered by an L2 cache fill. Those lines can be either in modified state or clean state. Modified lines may either be written back to L3 or directly written to memory and not allocated in L3.  Clean lines may either be allocated in L3 or dropped",
},
{
	.name = "l2_lines_out.useless_pref",
	.event = "umask=0x4,period=200003,event=0xf2",
	.desc = "This event is deprecated. Refer to new event L2_LINES_OUT.USELESS_HWPF",
	.topic = "cache",
},
{
	.name = "l2_lines_out.useless_hwpf",
	.event = "umask=0x4,period=200003,event=0xf2",
	.desc = "Counts the number of lines that have been hardware prefetched but not used and now evicted by L2 cache",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Number of cache line split locks sent to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of cache line split locks sent to the uncore",
},
{
	.name = "offcore_response:request=demand_data_rd:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080028000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100028000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200028000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400028000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800028000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000028000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80028000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080048000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100048000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200048000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400048000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800048000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000048000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80048000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080088000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100088000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200088000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400088000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800088000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000088000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80088000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080108000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100108000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200108000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400108000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800108000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000108000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80108000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080208000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100208000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200208000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400208000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800208000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000208000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80208000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C8000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C8000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C8000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C8000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C8000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C8000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C8000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00800207F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.SUPPLIER_NONE.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01000207F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.SUPPLIER_NONE.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02000207F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.SUPPLIER_NONE.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04000207F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.SUPPLIER_NONE.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08000207F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.SUPPLIER_NONE.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10000207F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.SUPPLIER_NONE.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F800207F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.SUPPLIER_NONE.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00800407F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_M.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01000407F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_M.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02000407F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_M.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04000407F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_M.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08000407F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_M.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10000407F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_M.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F800407F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_M.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00800807F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_E.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01000807F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_E.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02000807F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_E.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04000807F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_E.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08000807F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_E.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10000807F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_E.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F800807F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_E.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00801007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_S.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01001007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_S.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02001007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_S.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04001007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_S.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08001007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_S.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10001007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_S.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F801007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_S.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00802007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_F.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01002007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_F.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02002007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_F.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04002007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_F.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08002007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_F.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10002007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_F.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F802007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT_F.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C07F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C07F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C07F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT.SNOOP_MISS",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C07F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT.HIT_OTHER_CORE_NO_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C07F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT.HIT_OTHER_CORE_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C07F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT.HITM_OTHER_CORE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C07F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000018000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C8000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C8000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00000107F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.ANY_RESPONSE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C07F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT.NO_SNOOP_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C07F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_HIT.SNOOP_HIT_WITH_FWD",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100408000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080408000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01004007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.PMM_HIT_LOCAL_PMM.SNOOP_NOT_NEEDED",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00804007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.PMM_HIT_LOCAL_PMM.SNOOP_NONE",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80408000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F804007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.PMM_HIT_LOCAL_PMM.ANY_SNOOP",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "clx metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "clx metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "clx metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "clx metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "clx metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "clx metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "clx metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "clx metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "clx metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Instruction per taken branch",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_taken",
	.metric_name = "IpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Branch instructions per taken branch",
	.topic = "clx metrics",
	.metric_expr = "br_inst_retired.all_branches / br_inst_retired.near_taken",
	.metric_name = "BpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "clx metrics",
	.metric_expr = "min( 1 , uops_issued.any / ( (uops_retired.retire_slots / inst_retired.any) * 64 * ( icache_64b.iftag_hit + icache_64b.iftag_miss ) / 4.1 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "clx metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops ))",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "clx metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "clx metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "clx metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "clx metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Instructions per Load (lower number means loads are more frequent)",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any / mem_inst_retired.all_loads",
	.metric_name = "IpL",
	.metric_group = "Instruction_Type;L1_Bound",
},
{
	.desc = "Instructions per Store",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any / mem_inst_retired.all_stores",
	.metric_name = "IpS",
	.metric_group = "Instruction_Type;Store_Bound",
},
{
	.desc = "Instructions per Branch",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.all_branches",
	.metric_name = "IpB",
	.metric_group = "Branches;Instruction_Type;Port_5;Port_6",
},
{
	.desc = "Instruction per (near) call",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_call",
	.metric_name = "IpCall",
	.metric_group = "Branches",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "clx metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * ( fp_arith_inst_retired.256b_packed_single + fp_arith_inst_retired.512b_packed_double ) + 16 * fp_arith_inst_retired.512b_packed_single )) / cycles",
	.metric_name = "FLOPc",
	.metric_group = "FLOPS",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "clx metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * ( fp_arith_inst_retired.256b_packed_single + fp_arith_inst_retired.512b_packed_double ) + 16 * fp_arith_inst_retired.512b_packed_single )) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "FLOPc_SMT",
	.metric_group = "FLOPS_SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "clx metrics",
	.metric_expr = "uops_executed.thread / (( uops_executed.core_cycles_ge_1 / 2 ) if #smt_on else uops_executed.core_cycles_ge_1)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "clx metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) * (( int_misc.clear_resteer_cycles + 9 * baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) ) * (4 * cycles) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "clx metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) * (( int_misc.clear_resteer_cycles + 9 * baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) ) * (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost_SMT",
	.metric_group = "Branch_Mispredicts_SMT",
},
{
	.desc = "Number of Instructions per non-speculative Branch Misprediction (JEClear)",
	.topic = "clx metrics",
	.metric_expr = "inst_retired.any / br_misp_retired.all_branches",
	.metric_name = "IpMispredict",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "clx metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads (in core cycles)",
	.topic = "clx metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_retired.l1_miss + mem_load_retired.fb_hit )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least one such miss. Per-thread)",
	.topic = "clx metrics",
	.metric_expr = "l1d_pend_miss.pending / l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "clx metrics",
	.metric_expr = "( itlb_misses.walk_pending + dtlb_load_misses.walk_pending + dtlb_store_misses.walk_pending + ept.walk_pending ) / ( 2 * cycles )",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "clx metrics",
	.metric_expr = "( itlb_misses.walk_pending + dtlb_load_misses.walk_pending + dtlb_store_misses.walk_pending + ept.walk_pending ) / ( 2 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )) )",
	.metric_name = "Page_Walks_Utilization_SMT",
	.metric_group = "TLB_SMT",
},
{
	.desc = "Average data fill bandwidth to the L1 data cache [GB / sec]",
	.topic = "clx metrics",
	.metric_expr = "64 * l1d.replacement / 1000000000 / duration_time",
	.metric_name = "L1D_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average data fill bandwidth to the L2 cache [GB / sec]",
	.topic = "clx metrics",
	.metric_expr = "64 * l2_lines_in.all / 1000000000 / duration_time",
	.metric_name = "L2_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "clx metrics",
	.metric_expr = "64 * longest_lat_cache.miss / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "clx metrics",
	.metric_expr = "64 * offcore_requests.all_requests / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Access_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "L1 cache true misses per kilo instruction for retired demand loads",
	.topic = "clx metrics",
	.metric_expr = "1000 * mem_load_retired.l1_miss / inst_retired.any",
	.metric_name = "L1MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache true misses per kilo instruction for retired demand loads",
	.topic = "clx metrics",
	.metric_expr = "1000 * mem_load_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache misses per kilo instruction for all request types (including speculative)",
	.topic = "clx metrics",
	.metric_expr = "1000 * l2_rqsts.miss / inst_retired.any",
	.metric_name = "L2MPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache hits per kilo instruction for all request types (including speculative)",
	.topic = "clx metrics",
	.metric_expr = "1000 * ( l2_rqsts.references - l2_rqsts.miss ) / inst_retired.any",
	.metric_name = "L2HPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L3 cache true misses per kilo instruction for retired demand loads",
	.topic = "clx metrics",
	.metric_expr = "1000 * mem_load_retired.l3_miss / inst_retired.any",
	.metric_name = "L3MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "Average CPU Utilization",
	.topic = "clx metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "clx metrics",
	.metric_expr = "( (( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * ( fp_arith_inst_retired.256b_packed_single + fp_arith_inst_retired.512b_packed_double ) + 16 * fp_arith_inst_retired.512b_packed_single )) / 1000000000 ) / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "clx metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "clx metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "clx metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "clx metrics",
	.metric_expr = "( 64 * ( uncore_imc@cas_count_read@ + uncore_imc@cas_count_write@ ) / 1000000000 ) / duration_time",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average latency of data read request to external memory (in nanoseconds). Accounts for demand loads and L1/L2 prefetches",
	.topic = "clx metrics",
	.metric_expr = "1000000000 * ( cha@event\\=0x36\\\\\\,umask\\=0x21\\\\\\,config\\=0x40433@ / cha@event\\=0x35\\\\\\,umask\\=0x21\\\\\\,config\\=0x40433@ ) / ( cha_0@event\\=0x0@ / duration_time )",
	.metric_name = "DRAM_Read_Latency",
	.metric_group = "Memory_Lat",
},
{
	.desc = "Average number of parallel data read requests to external memory. Accounts for demand loads and L1/L2 prefetches",
	.topic = "clx metrics",
	.metric_expr = "cha@event\\=0x36\\\\\\,umask\\=0x21\\\\\\,config\\=0x40433@ / cha@event\\=0x36\\\\\\,umask\\=0x21\\\\\\,thresh\\=1\\\\\\,config\\=0x40433@",
	.metric_name = "DRAM_Parallel_Reads",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average latency of data read request to external 3D X-Point memory [in nanoseconds]. Accounts for demand loads and L1/L2 data-read prefetches",
	.topic = "clx metrics",
	.metric_expr = "( 1000000000 * ( imc@event\\=0xe0\\\\\\,umask\\=0x1@ / imc@event\\=0xe3@ ) / imc_0@event\\=0x0@ ) if 1 if 1 == 1 else 0 else 0",
	.metric_name = "MEM_PMM_Read_Latency",
	.metric_group = "Memory_Lat",
},
{
	.desc = "Average 3DXP Memory Bandwidth Use for reads [GB / sec]",
	.topic = "clx metrics",
	.metric_expr = "( ( 64 * imc@event\\=0xe3@ / 1000000000 ) / duration_time ) if 1 if 1 == 1 else 0 else 0",
	.metric_name = "PMM_Read_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average 3DXP Memory Bandwidth Use for Writes [GB / sec]",
	.topic = "clx metrics",
	.metric_expr = "( ( 64 * imc@event\\=0xe7@ / 1000000000 ) / duration_time ) if 1 if 1 == 1 else 0 else 0",
	.metric_name = "PMM_Write_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Socket actual clocks when any core is active on that socket",
	.topic = "clx metrics",
	.metric_expr = "cha_0@event\\=0x0@",
	.metric_name = "Socket_CLKS",
	.metric_group = "",
},
{
	.desc = "C3 residency percent per core",
	.topic = "clx metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "clx metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "clx metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "clx metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "clx metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "clx metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "clx metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "fp_arith_inst_retired.scalar_double",
	.event = "umask=0x1,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 1 computation. Applies to SSE* and AVX* scalar double precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.scalar_single",
	.event = "umask=0x2,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 1 computation. Applies to SSE* and AVX* scalar single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_double",
	.event = "umask=0x4,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 2 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT RSQRT14 RCP14 DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_single",
	.event = "umask=0x8,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 4 computation operations, one for each element.  Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 4 calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_double",
	.event = "umask=0x10,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 4 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 4 calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_single",
	.event = "umask=0x20,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 8 computation operations, one for each element.  Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 8 calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.512b_packed_double",
	.event = "umask=0x40,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 512-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 8 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 8 calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.512b_packed_single",
	.event = "umask=0x80,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 512-bit packed single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 16 computation operations, one for each element.  Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 16 calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,cmask=1,period=100003,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
	.long_desc = "Counts cycles with any input and output SSE or x87 FP assist. If an input and output assist are detected on the same cycle the event increments by 1",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may 'bypass' the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Counts the number of uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles uops were delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Count includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles 4 uops were delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Count includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts the number of uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles uops were delivered to the Instruction Decode Queue (IDQ) from the MITE (legacy decode pipeline) path. Counting includes uops that may 'bypass' the IDQ. During these cycles uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles 4 uops were delivered to the Instruction Decode Queue (IDQ) from the MITE (legacy decode pipeline) path. Counting includes uops that may 'bypass' the IDQ. During these cycles uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may 'bypass' the IDQ. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts the total number of uops delivered by the Microcode Sequencer (MS). Any instruction over 4 uops will be delivered by the MS. Some instructions such as transcendentals may additionally generate uops from the MS",
},
{
	.name = "idq.ms_switches",
	.event = "edge=1,umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
	.long_desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
},
{
	.name = "icache_16b.ifdata_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction cache miss",
	.topic = "frontend",
	.long_desc = "Cycles where a code line fetch is stalled due to an L1 instruction cache miss. The legacy decode pipeline works at a 16 Byte granularity",
},
{
	.name = "icache_64b.iftag_hit",
	.event = "umask=0x1,period=200003,event=0x83",
	.desc = "Instruction fetch tag lookups that hit in the instruction cache (L1I). Counts at 64-byte cache-line granularity",
	.topic = "frontend",
},
{
	.name = "icache_64b.iftag_miss",
	.event = "umask=0x2,period=200003,event=0x83",
	.desc = "Instruction fetch tag lookups that miss in the instruction cache (L1I). Counts at 64-byte cache-line granularity",
	.topic = "frontend",
},
{
	.name = "icache_64b.iftag_stall",
	.event = "umask=0x4,period=200003,event=0x83",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction cache tag miss",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
	.long_desc = "Cycles with less than 3 uops delivered by the front-end",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,cmask=2,period=2000003,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
	.long_desc = "Cycles with less than 2 uops delivered by the front-end",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,cmask=3,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts, on the per-thread basis, cycles when less than 1 uop is delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core >= 3",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,cmask=4,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts, on the per-thread basis, cycles when no uops are delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core =4",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts the number of uops not delivered to Resource Allocation Table (RAT) per thread adding 4  x when Resource Allocation Table (RAT) is not stalled and Instruction Decode Queue (IDQ) delivers x uops to Resource Allocation Table (RAT) (where x belongs to {0,1,2,3}). Counting does not cover cases when: a. IDQ-Resource Allocation Table (RAT) pipe serves the other thread. b. Resource Allocation Table (RAT) is stalled for the thread (including uop drops and clear BE conditions).  c. Instruction Decode Queue (IDQ) delivers four uops",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "Counts Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles. These cycles do not include uops routed through because of the switch itself, for example, when Instruction Decode Queue (IDQ) pre-allocation is unavailable, or Instruction Decode Queue (IDQ) is full. SBD-to-MITE switch true penalty cycles happen after the merge mux (MM) receives Decode Stream Buffer (DSB) Sync-indication until receiving the first MITE uop. MM is placed before Instruction Decode Queue (IDQ) to merge uops being fed from the MITE and Decode Stream Buffer (DSB) paths. Decode Stream Buffer (DSB) inserts the Sync-indication whenever a Decode Stream Buffer (DSB)-to-MITE switch occurs.Penalty: A Decode Stream Buffer (DSB) hit followed by a Decode Stream Buffer (DSB) miss can cost up to six cycles in which no uops are delivered to the IDQ. Most often, such switches from the Decode Stream Buffer (DSB) to the legacy pipeline cost 02 cycles",
},
{
	.name = "frontend_retired.latency_ge_4",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x400406",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 4 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_2",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x200206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 2 bubble-slots for a period of 2 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_2",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x400206",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 2 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.stlb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x15",
	.desc = "Retired Instructions who experienced STLB (2nd level TLB) true miss (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced STLB (2nd level TLB) true miss (Precise event)",
},
{
	.name = "frontend_retired.itlb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x14",
	.desc = "Retired Instructions who experienced iTLB true miss (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced iTLB (Instruction TLB) true miss (Precise event)",
},
{
	.name = "frontend_retired.l2_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x13",
	.desc = "Retired Instructions who experienced Instruction L2 Cache true miss (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.l1i_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x12",
	.desc = "Retired Instructions who experienced Instruction L1 Cache true miss (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.dsb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x11",
	.desc = "Retired Instructions who experienced decode stream buffer (DSB - the decoded instruction-cache) miss (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced DSB (Decode stream buffer i.e. the decoded instruction-cache) miss (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_3",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x300206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 3 bubble-slots for a period of 2 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_1",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x100206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 1 bubble-slot for a period of 2 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after the front-end had at least 1 bubble-slot for a period of 2 cycles. A bubble-slot is an empty issue-pipeline slot while there was no RAT stall (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_512",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x420006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 512 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_256",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x410006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 256 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_128",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x408006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 128 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_64",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x404006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 64 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_32",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x402006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 32 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 32 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_16",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x401006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 16 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 16 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_8",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x400806",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 8 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 8 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "tx_mem.abort_conflict",
	.event = "umask=0x1,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data conflict on a transactionally accessed address",
	.topic = "memory",
	.long_desc = "Number of times a TSX line had a cache conflict",
},
{
	.name = "tx_mem.abort_capacity",
	.event = "umask=0x2,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data capacity limitation for transactional reads or writes",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_store_to_elided_lock",
	.event = "umask=0x4,period=2000003,event=0x54",
	.desc = "Number of times a HLE transactional region aborted due to a non XRELEASE prefixed instruction writing to an elided lock in the elision buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to a non-release/commit store to lock",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_not_empty",
	.event = "umask=0x8,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to NoAllocatedElisionBuffer being non-zero",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_mismatch",
	.event = "umask=0x10,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to XRELEASE lock not satisfying the address and value requirements in the elision buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_unsupported_alignment",
	.event = "umask=0x20,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to an unsupported read alignment from the elision buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
},
{
	.name = "tx_mem.hle_elision_buffer_full",
	.event = "umask=0x40,period=2000003,event=0x54",
	.desc = "Number of times HLE lock could not be elided due to ElisionBufferAvailable being zero",
	.topic = "memory",
	.long_desc = "Number of times we could not allocate Lock Buffer",
},
{
	.name = "tx_exec.misc1",
	.event = "umask=0x1,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions that may cause a transactional abort was executed. Since this is the count of execution, it may not always cause a transactional abort",
	.topic = "memory",
},
{
	.name = "tx_exec.misc2",
	.event = "umask=0x2,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions (e.g., vzeroupper) that may cause a transactional abort was executed inside a transactional region",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by a vzeroupper instruction",
},
{
	.name = "tx_exec.misc3",
	.event = "umask=0x4,period=2000003,event=0x5d",
	.desc = "Counts the number of times an instruction execution caused the transactional nest count supported to be exceeded",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by a nest count that is too deep",
},
{
	.name = "tx_exec.misc4",
	.event = "umask=0x8,period=2000003,event=0x5d",
	.desc = "Counts the number of times a XBEGIN instruction was executed inside an HLE transactional region",
	.topic = "memory",
	.long_desc = "RTM region detected inside HLE",
},
{
	.name = "tx_exec.misc5",
	.event = "umask=0x10,period=2000003,event=0x5d",
	.desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
	.topic = "memory",
	.long_desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
},
{
	.name = "offcore_requests_outstanding.l3_miss_demand_data_rd",
	.event = "umask=0x10,period=2000003,event=0x60",
	.desc = "Counts number of Offcore outstanding Demand Data Read requests that miss L3 cache in the superQ every cycle",
	.topic = "memory",
},
{
	.name = "offcore_requests_outstanding.l3_miss_demand_data_rd_ge_6",
	.event = "umask=0x10,cmask=6,period=2000003,event=0x60",
	.desc = "Cycles with at least 6 Demand Data Read requests that miss L3 cache in the superQ",
	.topic = "memory",
},
{
	.name = "offcore_requests_outstanding.cycles_with_l3_miss_demand_data_rd",
	.event = "umask=0x10,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles with at least 1 Demand Data Read requests who miss L3 cache in the superQ",
	.topic = "memory",
},
{
	.name = "cycle_activity.cycles_l3_miss",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xa3",
	.desc = "Cycles while L3 cache miss demand load is outstanding",
	.topic = "memory",
},
{
	.name = "cycle_activity.stalls_l3_miss",
	.event = "umask=0x6,cmask=6,period=2000003,event=0xa3",
	.desc = "Execution stalls while L3 cache miss demand load is outstanding",
	.topic = "memory",
},
{
	.name = "offcore_requests.l3_miss_demand_data_rd",
	.event = "umask=0x10,period=100003,event=0xb0",
	.desc = "Demand Data Read requests who miss L3 cache",
	.topic = "memory",
	.long_desc = "Demand Data Read requests who miss L3 cache",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts  Spec update: SKL089",
	.topic = "memory",
	.long_desc = "Counts the number of memory ordering Machine Clears detected. Memory Ordering Machine Clears can result from one of the following:a. memory disambiguation,b. external snoop, orc. cross SMT-HW-thread snoop (stores) hitting load buffer  Spec update: SKL089",
},
{
	.name = "hle_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution started",
	.topic = "memory",
	.long_desc = "Number of times we entered an HLE region. Does not count nested transactions",
},
{
	.name = "hle_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution successfully committed",
	.topic = "memory",
	.long_desc = "Number of times HLE commit succeeded",
},
{
	.name = "hle_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times HLE abort was triggered (Precise event)",
},
{
	.name = "hle_retired.aborted_mem",
	.event = "umask=0x8,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_timer",
	.event = "umask=0x10,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to hardware timer expiration",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_unfriendly",
	.event = "umask=0x20,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions and certain unfriendly events (such as AD assists etc.)",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_memtype",
	.event = "umask=0x40,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times an HLE execution aborted due to incompatible memory type",
},
{
	.name = "hle_retired.aborted_events",
	.event = "umask=0x80,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to unfriendly events (such as interrupts)",
	.topic = "memory",
},
{
	.name = "rtm_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution started",
	.topic = "memory",
	.long_desc = "Number of times we entered an RTM region. Does not count nested transactions",
},
{
	.name = "rtm_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution successfully committed",
	.topic = "memory",
	.long_desc = "Number of times RTM commit succeeded",
},
{
	.name = "rtm_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times RTM abort was triggered (Precise event)",
},
{
	.name = "rtm_retired.aborted_mem",
	.event = "umask=0x8,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
},
{
	.name = "rtm_retired.aborted_timer",
	.event = "umask=0x10,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to uncommon conditions",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_unfriendly",
	.event = "umask=0x20,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
},
{
	.name = "rtm_retired.aborted_memtype",
	.event = "umask=0x40,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to incompatible memory type",
},
{
	.name = "rtm_retired.aborted_events",
	.event = "umask=0x80,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 512 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 512 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 256 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 256 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 128 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 128 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 64 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 64 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 32 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 32 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 16 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 16 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 8 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 8 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 4 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 4 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00840007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_LOCAL_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01040007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_LOCAL_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02040007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_LOCAL_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04040007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08040007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_LOCAL_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10040007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_LOCAL_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F840007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_LOCAL_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00900007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01100007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_REMOTE_HOP1_DRAM.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02100007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_REMOTE_HOP1_DRAM.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04100007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08100007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_REMOTE_HOP1_DRAM.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10100007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_REMOTE_HOP1_DRAM.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F900007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_REMOTE_HOP1_DRAM.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC0007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS.SNOOP_NONE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C0007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS.NO_SNOOP_NEEDED",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C0007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS.SNOOP_MISS",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C0007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS.HIT_OTHER_CORE_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C0007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS.HIT_OTHER_CORE_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C0007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS.HITM_OTHER_CORE",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC0007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS.ANY_SNOOP",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B808000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604008000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B8007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_REMOTE_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x06040007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS_LOCAL_DRAM.SNOOP_MISS_OR_NO_FWD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC08000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS.REMOTE_HITM",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_data_rd:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00001",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_DATA_RD.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_rfo:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00002",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_RFO.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=demand_code_rd:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00004",
	.desc = "\nThis event is deprecated. Refer to new event OCR.DEMAND_CODE_RD.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_data_rd:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00010",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_DATA_RD.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l2_rfo:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00020",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L2_RFO.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_data_rd:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00080",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_DATA_RD.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l3_rfo:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00100",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L3_RFO.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=pf_l1d_and_sw:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00400",
	.desc = "\nThis event is deprecated. Refer to new event OCR.PF_L1D_AND_SW.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=other:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC08000",
	.desc = "\nThis event is deprecated. Refer to new event OCR.OTHER.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_data_rd:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00490",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_DATA_RD.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_pf_rfo:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00120",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_PF_RFO.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_data_rd:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00491",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_DATA_RD.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_rfo:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00122",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_RFO.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response:request=all_reads:response=l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC007F7",
	.desc = "\nThis event is deprecated. Refer to new event OCR.ALL_READS.L3_MISS.REMOTE_HIT_FORWARD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000001",
	.desc = "Counts demand data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000004",
	.desc = "Counts all demand code reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000004",
	.desc = "Counts all demand code reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90008000",
	.desc = "Counts any other requests  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC008000",
	.desc = "Counts any other requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C008000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C008000",
	.desc = "Counts any other requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C008000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C008000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C008000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC008000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000490",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000490",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000490",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000490",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000490",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000490",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000490",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000490",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000490",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000490",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000120",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000120",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000120",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000120",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000120",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000120",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000120",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000120",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000120",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000120",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000491",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000491",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000491",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000491",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000491",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000491",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000491",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000491",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000491",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000491",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000122",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000122",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0804000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0090000122",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0110000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0210000122",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0410000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0810000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1010000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F90000122",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000122",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000122",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C000122",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C000122",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000122",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00840007F7",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_local_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01040007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02040007F7",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_local_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04040007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_local_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08040007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_local_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10040007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F840007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_remote_hop1_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00900007F7",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_remote_hop1_dram.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01100007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_remote_hop1_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02100007F7",
	.desc = "TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_remote_hop1_dram.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04100007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_remote_hop1_dram.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08100007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_remote_hop1_dram.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10100007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_remote_hop1_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F900007F7",
	.desc = "TBD  TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC0007F7",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C0007F7",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C0007F7",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C0007F7",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083C0007F7",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C0007F7",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC0007F7",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800004",
	.desc = "Counts all demand code reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000004",
	.desc = "Counts all demand code reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B808000",
	.desc = "Counts any other requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604008000",
	.desc = "Counts any other requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B8007F7",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x06040007F7",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00004",
	.desc = "Counts all demand code reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC08000",
	.desc = "Counts any other requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC007F7",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00004",
	.desc = "Counts all demand code reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC08000",
	.desc = "Counts any other requests TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC007F7",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "memory_disambiguation.history_reset",
	.event = "umask=0x1,period=2000003,event=0x9",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_power.lvl0_turbo_license",
	.event = "umask=0x7,period=200003,event=0x28",
	.desc = "Core cycles where the core was running in a manner where Turbo may be clipped to the Non-AVX turbo schedule",
	.topic = "other",
	.long_desc = "Core cycles where the core was running with power-delivery for baseline license level 0.  This includes non-AVX codes, SSE, AVX 128-bit, and low-current AVX 256-bit codes",
},
{
	.name = "core_power.lvl1_turbo_license",
	.event = "umask=0x18,period=200003,event=0x28",
	.desc = "Core cycles where the core was running in a manner where Turbo may be clipped to the AVX2 turbo schedule",
	.topic = "other",
	.long_desc = "Core cycles where the core was running with power-delivery for license level 1.  This includes high current AVX 256-bit instructions as well as low current AVX 512-bit instructions",
},
{
	.name = "core_power.lvl2_turbo_license",
	.event = "umask=0x20,period=200003,event=0x28",
	.desc = "Core cycles where the core was running in a manner where Turbo may be clipped to the AVX512 turbo schedule",
	.topic = "other",
	.long_desc = "Core cycles where the core was running with power-delivery for license level 2 (introduced in Skylake Server michroarchtecture).  This includes high current AVX 512-bit instructions",
},
{
	.name = "core_power.throttle",
	.event = "umask=0x40,period=200003,event=0x28",
	.desc = "Core cycles the core was throttled due to a pending power level request",
	.topic = "other",
	.long_desc = "Core cycles the out-of-order engine was throttled due to a pending power level request",
},
{
	.name = "sw_prefetch_access.nta",
	.event = "umask=0x1,period=2000003,event=0x32",
	.desc = "Number of PREFETCHNTA instructions executed",
	.topic = "other",
},
{
	.name = "sw_prefetch_access.t0",
	.event = "umask=0x2,period=2000003,event=0x32",
	.desc = "Number of PREFETCHT0 instructions executed",
	.topic = "other",
},
{
	.name = "sw_prefetch_access.t1_t2",
	.event = "umask=0x4,period=2000003,event=0x32",
	.desc = "Number of PREFETCHT1 or PREFETCHT2 instructions executed",
	.topic = "other",
},
{
	.name = "sw_prefetch_access.prefetchw",
	.event = "umask=0x8,period=2000003,event=0x32",
	.desc = "Number of PREFETCHW instructions executed",
	.topic = "other",
},
{
	.name = "hw_interrupts.received",
	.event = "umask=0x1,period=203,event=0xcb",
	.desc = "Number of hardware interrupts received by the processor",
	.topic = "other",
	.long_desc = "Counts the number of hardware interruptions received by the processor",
},
{
	.name = "core_snoop_response.rsp_ihiti",
	.event = "umask=0x1,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_ihitfse",
	.event = "umask=0x2,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_shitfse",
	.event = "umask=0x4,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_sfwdm",
	.event = "umask=0x8,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_ifwdm",
	.event = "umask=0x10,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_ifwdfe",
	.event = "umask=0x20,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_sfwdfe",
	.event = "umask=0x40,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "idi_misc.wb_upgrade",
	.event = "umask=0x2,period=100003,event=0xfe",
	.desc = "Counts number of cache lines that are allocated and written back to L3 with the intention that they are more likely to be reused shortly",
	.topic = "other",
	.long_desc = "Counts number of cache lines that are allocated and written back to L3 with the intention that they are more likely to be reused shortly",
},
{
	.name = "idi_misc.wb_downgrade",
	.event = "umask=0x4,period=100003,event=0xfe",
	.desc = "Counts number of cache lines that are dropped and not written back to L3 as they are deemed to be less likely to be reused shortly",
	.topic = "other",
	.long_desc = "Counts number of cache lines that are dropped and not written back to L3 as they are deemed to be less likely to be reused shortly",
},
{
	.name = "ocr.demand_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200001",
	.desc = "Counts demand data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0001",
	.desc = "Counts demand data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0001",
	.desc = "Counts demand data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200002",
	.desc = "Counts all demand data writes (RFOs)  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200004",
	.desc = "Counts all demand code reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0004",
	.desc = "Counts all demand code reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0004",
	.desc = "Counts all demand code reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0004",
	.desc = "Counts all demand code reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200010",
	.desc = "Counts prefetch (that bring data to L2) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080028000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100028000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200028000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400028000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800028000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000028000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80028000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080048000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100048000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200048000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400048000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800048000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000048000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80048000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080088000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100088000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200088000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400088000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800088000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000088000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80088000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080108000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100108000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200108000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400108000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800108000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000108000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80108000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080208000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100208000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200208000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400208000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800208000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000208000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80208000",
	.desc = "Counts any other requests  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C8000",
	.desc = "Counts any other requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C8000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C8000",
	.desc = "Counts any other requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C8000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C8000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C8000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C8000",
	.desc = "Counts any other requests TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200490",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0490",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0490",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0490",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0490",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0490",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0490",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0490",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200120",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0120",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0120",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0120",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0120",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0120",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0120",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0120",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200491",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0491",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0491",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0491",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0491",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0491",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0491",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0491",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800020122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800040122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80040122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800080122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80080122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800100122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80100122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080200122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100200122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200200122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400200122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0800200122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000200122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80200122",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0122",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0122",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0122",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0122",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C0122",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0122",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0122",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00800207F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.supplier_none.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01000207F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02000207F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.supplier_none.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04000207F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.supplier_none.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08000207F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.supplier_none.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10000207F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F800207F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00800407F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_m.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01000407F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02000407F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_m.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04000407F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_m.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08000407F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_m.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10000407F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F800407F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00800807F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_e.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01000807F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02000807F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_e.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04000807F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_e.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08000807F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_e.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10000807F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F800807F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00801007F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_s.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01001007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02001007F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_s.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04001007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_s.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08001007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_s.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10001007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F801007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_f.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00802007F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_f.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01002007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_f.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02002007F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_f.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04002007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_f.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08002007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_f.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10002007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit_f.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F802007F7",
	.desc = "TBD  TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C07F7",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C07F7",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C07F7",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C07F7",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit.hit_other_core_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08003C07F7",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C07F7",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C07F7",
	.desc = "TBD TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010001",
	.desc = "Counts demand data reads have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0001",
	.desc = "Counts demand data reads hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0001",
	.desc = "Counts demand data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010002",
	.desc = "Counts all demand data writes (RFOs) have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0002",
	.desc = "Counts all demand data writes (RFOs) hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010004",
	.desc = "Counts all demand code reads have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0004",
	.desc = "Counts all demand code reads hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0004",
	.desc = "Counts all demand code reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010010",
	.desc = "Counts prefetch (that bring data to L2) data reads have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000018000",
	.desc = "Counts any other requests have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C8000",
	.desc = "Counts any other requests hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C8000",
	.desc = "Counts any other requests",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010490",
	.desc = "TBD have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0490",
	.desc = "TBD hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0490",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010120",
	.desc = "TBD have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0120",
	.desc = "TBD hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0120",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010491",
	.desc = "TBD have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0491",
	.desc = "TBD hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0491",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010122",
	.desc = "TBD have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0122",
	.desc = "TBD hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C0122",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00000107F7",
	.desc = "TBD have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C07F7",
	.desc = "TBD hit in the L3 and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x08007C07F7",
	.desc = "TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400001",
	.desc = "Counts demand data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400001",
	.desc = "Counts demand data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400004",
	.desc = "Counts all demand code reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400004",
	.desc = "Counts all demand code reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100408000",
	.desc = "Counts any other requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080408000",
	.desc = "Counts any other requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400490",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400490",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400120",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400120",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400491",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400491",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400122",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400122",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.pmm_hit_local_pmm.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01004007F7",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.pmm_hit_local_pmm.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00804007F7",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400001",
	.desc = "Counts demand data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_code_rd.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400004",
	.desc = "Counts all demand code reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_data_rd.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l2_rfo.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_data_rd.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l3_rfo.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.pf_l1d_and_sw.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.other.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80408000",
	.desc = "Counts any other requests TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_data_rd.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400490",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_pf_rfo.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400120",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_data_rd.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400491",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_rfo.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80400122",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.all_reads.pmm_hit_local_pmm.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F804007F7",
	.desc = "TBD TBD",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, Counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers. Notes: INST_RETIRED.ANY is counted by a designated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. INST_RETIRED.ANY_P is counted by a programmable counter and it is an architectural performance event. Counting: Faulting executions of GETSEC/VM entry/VM Exit/MWait will not count as retired instructions",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state. This event has a constant ratio with the CPU_CLK_UNHALTED.REF_XCLK event. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. Note: On all current platforms this event stops counting during 'throttling (TM)' states duty off periods the processor is 'halted'.  The counter update is done at a lower clock rate then the core clock the overflow status bit for this counter may appear 'sticky'.  After the counter has overflowed and software clears the overflow status bit and resets the counter to less than MAX. The reset value to the counter is not clocked immediately so the overflow status bit will flip 'high (1)' and generate another PMI (if enabled) after which the reset value gets clocked into the counter. Therefore, software will get the interrupt, read the overflow status bit '1 for bit 34 while the counter value is less than MAX. Software should ignore this case",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Loads blocked by overlapping with store buffer that cannot be forwarded ",
	.topic = "pipeline",
	.long_desc = "Counts how many times the load operation got the true Block-on-Store blocking code preventing store forwarding. This includes cases when:a. preceding store conflicts with the load (incomplete overlap),b. store forwarding is impossible due to u-arch limitations,c. preceding lock RMW operations are not forwarded,d. store has the no-forward bit set (uncacheable/page-split/masked stores),e. all-blocking stores are used (mostly, fences and port I/O), and others.The most common case is a load blocked due to its address range overlapping with a preceding smaller uncompleted store. Note: This event does not take into account cases of out-of-SW-control (for example, SbTailHit), unknown physical STA, and cases of blocking loads on store due to being non-WB memory type or a lock. These cases are covered by other events. See the table of not supported store forwards in the Optimization Guide",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
	.long_desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare on address",
	.topic = "pipeline",
	.long_desc = "Counts false dependencies in MOB when the partial comparison upon loose net check and dependency was resolved by the Enhanced Loose net mechanism. This may not result in high performance penalties. Loose net checks can fail when loads and stores are 4k aliased",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for this thread (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "Core cycles the Resource allocator was stalled due to recovery from an earlier branch misprediction or machine clear event",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "int_misc.clear_resteer_cycles",
	.event = "umask=0x80,period=2000003,event=0xd",
	.desc = "Cycles the issue-stage is waiting for front-end to fetch from resteered path following branch misprediction or machine clear events",
	.topic = "pipeline",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which the Resource Allocation Table (RAT) does not issue any Uops to the reservation station (RS) for the current thread",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "Counts the number of uops that the Resource Allocation Table (RAT) issues to the Reservation Station (RS)",
},
{
	.name = "uops_issued.vector_width_mismatch",
	.event = "umask=0x2,period=2000003,event=0xe",
	.desc = "Uops inserted at issue-stage in order to preserve upper bits of vector registers",
	.topic = "pipeline",
	.long_desc = "Counts the number of Blend Uops issued by the Resource Allocation Table (RAT) to the reservation station (RS) in order to preserve upper bits of vector registers. Starting with the Skylake microarchitecture, these Blend uops are needed since every Intel SSE instruction executed in Dirty Upper State needs to preserve bits 128-255 of the destination register. For more information, refer to Mixing Intel AVX and Intel SSE Code section of the Optimization Guide",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
},
{
	.name = "arith.divider_active",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x14",
	.desc = "Cycles when divide unit is busy executing divide or square root operations. Accounts for integer and floating-point operations",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This is an architectural event that counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling. For this reason, this event may have a changing ratio with regards to wall clock time",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ring0_trans",
	.event = "edge=1,umask=0x0,cmask=1,period=100007,event=0x3c",
	.desc = "Counts when there is a transition from ring 1, 2 or 3 to ring 0",
	.topic = "pipeline",
	.long_desc = "Counts when the Current Privilege Level (CPL) transitions from ring 1, 2 or 3 to ring 0 (Kernel)",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=25003,event=0x3c",
	.desc = "Core crystal clock cycles when the thread is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=25003,event=0x3c",
	.desc = "Core crystal clock cycles when at least one thread on the physical core is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=25003,event=0x3c",
	.desc = "Core crystal clock cycles when at least one thread on the physical core is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=25003,event=0x3c",
	.desc = "Core crystal clock cycles when the thread is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=25003,event=0x3c",
	.desc = "Core crystal clock cycles when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=25003,event=0x3c",
	.desc = "Core crystal clock cycles when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Demand load dispatches that hit L1D fill buffer (FB) allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "Counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the software prefetch. It can also be incremented by some lock instructions. So it should only be used with profiling so that the locks can be excluded by ASM (Assembly File) inspection of the nearby instructions",
},
{
	.name = "partial_rat_stalls.scoreboard",
	.event = "umask=0x1,period=2000003,event=0x59",
	.desc = "Cycles where the pipeline is stalled due to serializing operations",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which the microcode scoreboard stalls happen",
},
{
	.name = "rs_events.empty_end",
	.event = "edge=1,inv=1,umask=0x1,cmask=1,period=2000003,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
	.long_desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate front-end Latency Bound issues",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which the reservation station (RS) is empty for the thread.; Note: In ST-mode, not active thread should drive 0. This is usually caused by severely costly branch mispredictions, or allocator/FE issues",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
	.long_desc = "Counts cycles that the Instruction Length decoder (ILD) stalls occurred due to dynamically changing prefix length of the decoded instruction (by operand size prefix instruction 0x66, address size prefix instruction 0x67 or REX.W for Intel64). Count is proportional to the number of prefixes in a 16B-line. This may result in a three-cycle penalty for each LCP (Length changing prefix) in a 16-byte chunk",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 2",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 3",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 4",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_dispatched_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_dispatched_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 7",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
	.long_desc = "Counts resource-related stall cycles",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "Counts allocation stall cycles caused by the store buffer (SB) being full. This counts cycles that the pipeline back-end blocked uop delivery from the front-end",
},
{
	.name = "cycle_activity.cycles_l2_miss",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_total",
	.event = "umask=0x4,cmask=4,period=2000003,event=0xa3",
	.desc = "Total execution stalls",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_miss",
	.event = "umask=0x5,cmask=5,period=2000003,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_miss",
	.event = "umask=0x8,cmask=8,period=2000003,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_miss",
	.event = "umask=0xc,cmask=12,period=2000003,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_mem_any",
	.event = "umask=0x10,cmask=16,period=2000003,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_mem_any",
	.event = "umask=0x14,cmask=20,period=2000003,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "exe_activity.exe_bound_0_ports",
	.event = "umask=0x1,period=2000003,event=0xa6",
	.desc = "Cycles where no uops were executed, the Reservation Station was not empty, the Store Buffer was full and there was no outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which no uops were executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.1_ports_util",
	.event = "umask=0x2,period=2000003,event=0xa6",
	.desc = "Cycles total of 1 uop is executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which a total of 1 uop was executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.2_ports_util",
	.event = "umask=0x4,period=2000003,event=0xa6",
	.desc = "Cycles total of 2 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which a total of 2 uops were executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.3_ports_util",
	.event = "umask=0x8,period=2000003,event=0xa6",
	.desc = "Cycles total of 3 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Cycles total of 3 uops are executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.4_ports_util",
	.event = "umask=0x10,period=2000003,event=0xa6",
	.desc = "Cycles total of 4 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Cycles total of 4 uops are executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.bound_on_stores",
	.event = "umask=0x40,period=2000003,event=0xa6",
	.desc = "Cycles where the Store Buffer was full and no outstanding load",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
	.long_desc = "Number of uops delivered to the back-end by the LSD(Loop Stream Detector)",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Counts the cycles when 4 uops are delivered by the LSD (Loop-stream detector)",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Counts the cycles when at least one uop is delivered by the LSD (Loop-stream detector)",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 4 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 3 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 2 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 1 uop was executed per-thread",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which no uops were dispatched from the Reservation Station (RS) per thread",
},
{
	.name = "uops_executed.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Counts the number of uops to be executed per-thread each cycle",
	.topic = "pipeline",
	.long_desc = "Number of uops to be executed per-thread each cycle",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core",
	.topic = "pipeline",
	.long_desc = "Number of uops executed from any thread",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.x87",
	.event = "umask=0x10,period=2000003,event=0xb1",
	.desc = "Counts the number of x87 uops dispatched",
	.topic = "pipeline",
	.long_desc = "Counts the number of x87 uops executed",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter - architectural event  Spec update: SKL091, SKL044",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions (EOMs) retired. Counting covers macro-fused instructions individually (that is, increments by two)  Spec update: SKL091, SKL044",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: SKL091, SKL044 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "A version of INST_RETIRED that allows for a more unbiased distribution of samples across instructions retired. It utilizes the Precise Distribution of Instructions Retired (PDIR) feature to mitigate some bias in how retired instructions get sampled  Spec update: SKL091, SKL044 (Must be precise)",
},
{
	.name = "inst_retired.total_cycles_ps",
	.event = "inv=1,umask=0x1,cmask=10,period=2000003,event=0xc0",
	.desc = "Number of cycles using always true condition applied to  PEBS instructions retired event  Spec update: SKL091, SKL044 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Number of cycles using an always true condition applied to  PEBS instructions retired event. (inst_ret< 16)  Spec update: SKL091, SKL044 (Must be precise)",
},
{
	.name = "other_assists.any",
	.event = "umask=0x3f,period=100003,event=0xc1",
	.desc = "Number of times a microcode assist is invoked by HW other than FP-assist. Examples include AD (page Access Dirty) and AVX* related assists",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x2,cmask=10,period=2000003,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
	.long_desc = "Number of cycles using always true condition (uops_ret < 16) applied to non PEBS uops retired event",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x2,cmask=1,period=2000003,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
	.long_desc = "This event counts cycles without actually retired uops",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used",
	.topic = "pipeline",
	.long_desc = "Counts the retirement slots used",
},
{
	.name = "machine_clears.count",
	.event = "edge=1,umask=0x1,cmask=1,period=100003,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "Counts self-modifying code (SMC) detected, which causes a machine clear",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired  Spec update: SKL091",
	.topic = "pipeline",
	.long_desc = "Counts all (macro) branch instructions retired  Spec update: SKL091",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts conditional branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts both direct and indirect near call instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired  Spec update: SKL091 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_INST_RETIRED.ALL_BRANCHES that counts all (macro) branch instructions retired  Spec update: SKL091 (Must be precise)",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts return instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Not taken branch instructions retired  Spec update: SKL091",
	.topic = "pipeline",
	.long_desc = "This event counts not taken branch instructions retired  Spec update: SKL091",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts taken branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Far branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts far branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Counts all the retired branch instructions that were mispredicted by the processor. A branch misprediction occurs when the processor incorrectly predicts the destination of the branch.  When the misprediction is discovered at execution, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts mispredicted conditional branch instructions retired (Precise event)",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=400009,event=0xc5",
	.desc = "Mispredicted direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts both taken and not taken retired mispredicted direct and indirect near calls, including both register and memory indirect (Precise event)",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_MISP_RETIRED.ALL_BRANCHES that counts all mispredicted macro branch instructions retired (Must be precise)",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "Number of near branch instructions retired that were mispredicted and taken (Precise event)",
	.topic = "pipeline",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Increments whenever there is an update to the LBR array",
	.topic = "pipeline",
	.long_desc = "Increments when an entry is added to the Last Branch Record (LBR) array (or removed from the array in case of RETURNs in call stack mode). The event requires LBR enable via IA32_DEBUGCTL MSR and branch type selection via MSR_LBR_SELECT",
},
{
	.name = "rob_misc_events.pause_inst",
	.event = "umask=0x40,period=2000003,event=0xcc",
	.desc = "Number of retired PAUSE instructions (that do not end up with a VMExit to the VMM; TSX aborted Instructions may be counted). This event is not supported on first SKL and KBL products",
	.topic = "pipeline",
},
{
	.name = "baclears.any",
	.event = "umask=0x1,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
	.long_desc = "Counts the number of times the front-end is resteered when it finds a branch instruction in a fetch line. This occurs for the first time a branch instruction is fetched or when the branch is not tracked by the BPU (Branch Prediction Unit) anymore",
},
{
	.name = "llc_misses.mem_read",
	.event = "umask=0x3,event=0x4",
	.desc = "read requests to memory controller. Derived from unc_m_cas_count.rd. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_write",
	.event = "umask=0xC,event=0x4",
	.desc = "write requests to memory controller. Derived from unc_m_cas_count.wr. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_clockticks",
	.event = "event=0",
	.desc = "Memory controller clock ticks. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_power_channel_ppd",
	.event = "event=0x85",
	.desc = "Cycles where DRAM ranks are in power down (CKE) mode+C37. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_channel_ppd / unc_m_clockticks) * 100.",
	.metric_name = "power_channel_ppd %",
},
{
	.name = "unc_m_power_self_refresh",
	.event = "event=0x43",
	.desc = "Cycles Memory is in self refresh power mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_self_refresh / unc_m_clockticks) * 100.",
	.metric_name = "power_self_refresh %",
},
{
	.name = "unc_m_pre_count.page_miss",
	.event = "umask=0x1,event=0x2",
	.desc = "Pre-charges due to page misses. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.rd",
	.event = "umask=0x4,event=0x2",
	.desc = "Pre-charge for reads. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.wr",
	.event = "umask=0x8,event=0x2",
	.desc = "Pre-charge for writes. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pmm_bandwidth.read",
	.event = "event=0xe3",
	.desc = "Intel Optane DC persistent memory bandwidth read (MB/sec). Derived from unc_m_pmm_rpq_inserts. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "6.103515625E-5MB/sec",
	.perpkg = "1",
},
{
	.name = "unc_m_pmm_bandwidth.write",
	.event = "event=0xe7",
	.desc = "Intel Optane DC persistent memory bandwidth write (MB/sec). Derived from unc_m_pmm_wpq_inserts. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "6.103515625E-5MB/sec",
	.perpkg = "1",
},
{
	.name = "unc_m_pmm_bandwidth.total",
	.event = "event=0xe3",
	.desc = "Intel Optane DC persistent memory bandwidth total (MB/sec). Derived from unc_m_pmm_rpq_inserts. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "6.103515625E-5MB/sec",
	.perpkg = "1",
	.metric_expr = "unc_m_pmm_rpq_inserts + unc_m_pmm_wpq_inserts",
	.metric_name = "UNC_M_PMM_BANDWIDTH.TOTAL",
},
{
	.name = "unc_m_pmm_read_latency",
	.event = "umask=0x1,event=0xe0",
	.desc = "Intel Optane DC persistent memory read latency (ns). Derived from unc_m_pmm_rpq_occupancy.all. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "6000000000ns",
	.perpkg = "1",
	.metric_expr = "unc_m_pmm_rpq_occupancy.all / unc_m_pmm_rpq_inserts / unc_m_clockticks",
	.metric_name = "UNC_M_PMM_READ_LATENCY",
},
{
	.name = "unc_cha_clockticks",
	.event = "event=0",
	.desc = "Uncore cache clock ticks. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_misses.uncacheable",
	.event = "umask=0x21,event=0x35,config1=0x40e33",
	.desc = "LLC misses - Uncacheable reads (from cpu) . Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_read",
	.event = "umask=0x21,event=0x35,config1=0x40040e33",
	.desc = "MMIO reads. Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_write",
	.event = "umask=0x21,event=0x35,config1=0x40041e33",
	.desc = "MMIO writes. Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_full",
	.event = "umask=0x21,event=0x35,config1=0x41833",
	.desc = "Streaming stores (full cache line). Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_partial",
	.event = "umask=0x21,event=0x35,config1=0x41a33",
	.desc = "Streaming stores (partial cache line). Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.reads",
	.event = "umask=0x03,event=0x50",
	.desc = "read requests from home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.reads_local",
	.event = "umask=0x01,event=0x50",
	.desc = "read requests from local home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.reads_remote",
	.event = "umask=0x02,event=0x50",
	.desc = "read requests from remote home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.writes",
	.event = "umask=0x0C,event=0x50",
	.desc = "write requests from home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.writes_local",
	.event = "umask=0x04,event=0x50",
	.desc = "write requests from local home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.writes_remote",
	.event = "umask=0x08,event=0x50",
	.desc = "write requests from remote home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "upi_data_bandwidth_tx",
	.event = "umask=0x0F,event=0x2",
	.desc = "UPI interconnect send bandwidth for payload. Derived from unc_upi_txl_flits.all_data. Unit: uncore_upi ",
	.topic = "uncore other",
	.pmu = "uncore_upi",
	.unit = "7.11E-06Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_read",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x04,event=0x83,ch_mask=0x1f",
	.desc = "PCI Express bandwidth reading at IIO. Derived from unc_iio_data_req_of_cpu.mem_read.part0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_read.part0 + unc_iio_data_req_of_cpu.mem_read.part1 + unc_iio_data_req_of_cpu.mem_read.part2 + unc_iio_data_req_of_cpu.mem_read.part3",
	.metric_name = "LLC_MISSES.PCIE_READ",
},
{
	.name = "llc_misses.pcie_write",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x01,event=0x83,ch_mask=0x1f",
	.desc = "PCI Express bandwidth writing at IIO. Derived from unc_iio_data_req_of_cpu.mem_write.part0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_write.part0 +unc_iio_data_req_of_cpu.mem_write.part1 +unc_iio_data_req_of_cpu.mem_write.part2 +unc_iio_data_req_of_cpu.mem_write.part3",
	.metric_name = "LLC_MISSES.PCIE_WRITE",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_write.part0 +unc_iio_data_req_of_cpu.mem_write.part1 +unc_iio_data_req_of_cpu.mem_write.part2 +unc_iio_data_req_of_cpu.mem_write.part3",
	.metric_name = "LLC_MISSES.PCIE_WRITE",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 1. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 2. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 3. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_read.part0 + unc_iio_data_req_of_cpu.mem_read.part1 + unc_iio_data_req_of_cpu.mem_read.part2 + unc_iio_data_req_of_cpu.mem_read.part3",
	.metric_name = "LLC_MISSES.PCIE_READ",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 1. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 2. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 3. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts demand data loads that caused a page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels, but the walk need not have completed",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x8",
	.desc = "Page walk completed due to a demand data load to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Page walk completed due to a demand data load to a 2M/4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 2M/4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_1g",
	.event = "umask=0x8,period=2000003,event=0x8",
	.desc = "Page walk completed due to a demand data load to a 1G page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x8",
	.desc = "Load miss in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts demand data loads that caused a completed page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels. The page walk can end with or without a fault",
},
{
	.name = "dtlb_load_misses.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x8",
	.desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a load. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a load. EPT page walk duration are excluded in Skylake microarchitecture",
},
{
	.name = "dtlb_load_misses.walk_active",
	.event = "umask=0x10,cmask=1,period=100003,event=0x8",
	.desc = "Cycles when at least one PMH is busy with a page walk for a load. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts cycles when at least one PMH (Page Miss Handler) is busy with a page walk for a load",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x20,period=2000003,event=0x8",
	.desc = "Loads that miss the DTLB and hit the STLB",
	.topic = "virtual memory",
	.long_desc = "Counts loads that miss the DTLB (Data TLB) and hit the STLB (Second level TLB)",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts demand data stores that caused a page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels, but the walk need not have completed",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 2M/4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 2M/4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 1G page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 1G pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x49",
	.desc = "Store misses in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts demand data stores that caused a completed page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels. The page walk can end with or without a fault",
},
{
	.name = "dtlb_store_misses.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x49",
	.desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a store. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a store. EPT page walk duration are excluded in Skylake microarchitecture",
},
{
	.name = "dtlb_store_misses.walk_active",
	.event = "umask=0x10,cmask=1,period=100003,event=0x49",
	.desc = "Cycles when at least one PMH is busy with a page walk for a store. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts cycles when at least one PMH (Page Miss Handler) is busy with a page walk for a store",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x20,period=100003,event=0x49",
	.desc = "Stores that miss the DTLB and hit the STLB",
	.topic = "virtual memory",
	.long_desc = "Stores that miss the DTLB (Data TLB) and hit the STLB (2nd Level TLB)",
},
{
	.name = "ept.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Counts 1 per cycle for each PMH that is busy with a EPT (Extended Page Table) walk for any request type",
	.topic = "virtual memory",
	.long_desc = "Counts cycles for each PMH (Page Miss Handler) that is busy with an EPT (Extended Page Table) walk for any request type",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts page walks of any page size (4K/2M/4M/1G) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB, but the walk need not have completed",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (4K)",
	.topic = "virtual memory",
	.long_desc = "Counts completed page walks (4K page size) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB. The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Counts code misses in all ITLB levels that caused a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (1G)",
	.topic = "virtual memory",
	.long_desc = "Counts store misses in all DTLB levels that cause a completed page walk (1G page size). The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts completed page walks (2M and 4M page sizes) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB. The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_pending",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Counts 1 per cycle for each PMH that is busy with a page walk for an instruction fetch request. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts 1 per cycle for each PMH (Page Miss Handler) that is busy with a page walk for an instruction fetch request. EPT page walk duration are excluded in Skylake michroarchitecture",
},
{
	.name = "itlb_misses.walk_active",
	.event = "umask=0x10,cmask=1,period=100003,event=0x85",
	.desc = "Cycles when at least one PMH is busy with a page walk for code (instruction fetch) request. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Cycles when at least one PMH is busy with a page walk for code (instruction fetch) request. EPT page walk duration are excluded in Skylake microarchitecture",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x20,period=100003,event=0x85",
	.desc = "Instruction fetch requests that miss the ITLB and hit the STLB",
	.topic = "virtual memory",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "Counts the number of flushes of the big or small ITLB pages. Counting include both TLB Flush (covering all sets) and TLB Set Clear (set-specific)",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "Counts the number of DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "Counts the number of any STLB flush attempts (such as entire, VPID, PCID, InvPage, CR3 write, etc.)",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_goldmont[] = {
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=200003,event=0x2e",
	.desc = "L2 cache request misses",
	.topic = "cache",
	.long_desc = "Counts memory requests originating from the core that miss in the L2 cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=200003,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
	.long_desc = "Counts memory requests originating from the core that reference a cache line in the L2 cache",
},
{
	.name = "l2_reject_xq.all",
	.event = "umask=0x0,period=200003,event=0x30",
	.desc = "Requests rejected by the XQ",
	.topic = "cache",
	.long_desc = "Counts the number of demand and prefetch transactions that the L2 XQ rejects due to a full or near full condition which likely indicates back pressure from the intra-die interconnect (IDI) fabric. The XQ may reject transactions from the L2Q (non-cacheable requests), L2 misses and L2 write-back victims",
},
{
	.name = "core_reject_l2q.all",
	.event = "umask=0x0,period=200003,event=0x31",
	.desc = "Requests rejected by the L2Q",
	.topic = "cache",
	.long_desc = "Counts the number of demand and L1 prefetcher requests rejected by the L2Q due to a full or nearly full condition which likely indicates back pressure from L2Q. It also counts requests that would have gone directly to the XQ, but are rejected due to a full or nearly full condition, indicating back pressure from the IDI link. The L2Q may also reject transactions from a core to ensure fairness between cores, or to delay a core's dirty eviction when the address conflicts with incoming external snoops",
},
{
	.name = "dl1.dirty_eviction",
	.event = "umask=0x1,period=200003,event=0x51",
	.desc = "L1 Cache evictions for dirty data",
	.topic = "cache",
	.long_desc = "Counts when a modified (dirty) cache line is evicted from the data L1 cache and needs to be written back to memory.  No count will occur if the evicted line is clean, and hence does not require a writeback",
},
{
	.name = "fetch_stall.icache_fill_pending_cycles",
	.event = "umask=0x2,period=200003,event=0x86",
	.desc = "Cycles code-fetch stalled due to an outstanding ICache miss",
	.topic = "cache",
	.long_desc = "Counts cycles that fetch is stalled due to an outstanding ICache miss. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes due to an ICache miss.  Note: this event is not the same as the total number of cycles spent retrieving instruction cache lines from the memory hierarchy",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100007,event=0xb7",
	.desc = "Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=200003,event=0xd0",
	.desc = "Locked load uops retired (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts locked memory uops retired.  This includes regular locks and bus locks. (To specifically count bus locks only, see the Offcore response event.)  A locked access is one with a lock prefix, or an exchange to memory.  See the SDM for a complete description of which memory load accesses are locks  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=200003,event=0xd0",
	.desc = "Load uops retired that split a cache-line (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired where the data requested spans a 64 byte cache line boundary  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=200003,event=0xd0",
	.desc = "Stores uops retired that split a cache-line (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts store uops retired where the data requested spans a 64 byte cache line boundary  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.split",
	.event = "umask=0x43,period=200003,event=0xd0",
	.desc = "Memory uops retired that split a cache-line (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts memory uops retired where the data requested spans a 64 byte cache line boundary  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=200003,event=0xd0",
	.desc = "Load uops retired (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts the number of load uops retired  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=200003,event=0xd0",
	.desc = "Store uops retired (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts the number of store uops retired  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.all",
	.event = "umask=0x83,period=200003,event=0xd0",
	.desc = "Memory uops retired (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts the number of memory uops retired that is either a loads or a store or both  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=200003,event=0xd1",
	.desc = "Load uops retired that hit L1 data cache (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired that hit the L1 data cache  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=200003,event=0xd1",
	.desc = "Load uops retired that hit L2 (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired that hit in the L2 cache  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=200003,event=0xd1",
	.desc = "Load uops retired that missed L1 data cache (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired that miss the L1 data cache  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=200003,event=0xd1",
	.desc = "Load uops retired that missed L2 (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired that miss in the L2 cache  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.hitm",
	.event = "umask=0x20,period=200003,event=0xd1",
	.desc = "Memory uop retired where cross core or cross module HITM occurred (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired where the cache line containing the data was in the modified state of another core or modules cache (HITM).  More specifically, this means that when the load address was checked by other caching agents (typically another processor) in the system, one of those caching agents indicated that they had a dirty copy of the data.  Loads that obtain a HITM response incur greater latency than most is typical for a load.  In addition, since HITM indicates that some other processor had this data in its cache, it implies that the data was shared between processors, or potentially was a lock or semaphore value.  This event is useful for locating sharing, false sharing, and contended locks  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.wcb_hit",
	.event = "umask=0x40,period=200003,event=0xd1",
	.desc = "Loads retired that hit WCB (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts memory load uops retired where the data is retrieved from the WCB (or fill buffer), indicating that the load found its data while that data was in the process of being brought into the L1 cache.  Typically a load will receive this indication when some other load or prefetch missed the L1 cache and was in the process of retrieving the cache line containing the data, but that process had not yet finished (and written the data back to the cache). For example, consider load X and Y, both referencing the same cache line that is not in the L1 cache.  If load X misses cache first, it obtains and WCB (or fill buffer) and begins the process of requesting the data.  When load Y requests the data, it will either hit the WCB, or the L1 cache, depending on exactly what time the request to Y occurs  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.dram_hit",
	.event = "umask=0x80,period=200003,event=0xd1",
	.desc = "Loads retired that came from DRAM (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts memory load uops retired where the data is retrieved from DRAM.  Event is counted at retirement, so the speculative loads are ignored.  A memory load can hit (or miss) the L1 cache, hit (or miss) the L2 cache, hit DRAM, hit in the WCB or receive a HITM response  Supports address when precise (Must be precise)",
},
{
	.name = "offcore_response.any_read.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x36000032b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_read.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x10000032b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_read.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x04000032b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_read.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x02000032b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_read.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00000432b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600003091",
	.desc = "Counts data reads (demand & prefetch) that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000003091",
	.desc = "Counts data reads (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400003091",
	.desc = "Counts data reads (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200003091",
	.desc = "Counts data reads (demand & prefetch) that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000043091",
	.desc = "Counts data reads (demand & prefetch) that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600003010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000003010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400003010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200003010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000043010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000008000",
	.desc = "Counts requests to the uncore subsystem that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400008000",
	.desc = "Counts requests to the uncore subsystem that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200008000",
	.desc = "Counts requests to the uncore subsystem that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000048000",
	.desc = "Counts requests to the uncore subsystem that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000018000",
	.desc = "Counts requests to the uncore subsystem that have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem that have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.streaming_stores.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600004800",
	.desc = "Counts any data writes to uncacheable write combining (USWC) memory region  that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts any data writes to uncacheable write combining (USWC) memory region  that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.streaming_stores.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000044800",
	.desc = "Counts any data writes to uncacheable write combining (USWC) memory region  that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts any data writes to uncacheable write combining (USWC) memory region  that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.partial_streaming_stores.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600004000",
	.desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.partial_streaming_stores.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000004000",
	.desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.partial_streaming_stores.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400004000",
	.desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.partial_streaming_stores.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200004000",
	.desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.partial_streaming_stores.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000044000",
	.desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts partial cache line data writes to uncacheable write combining (USWC) memory region  that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600002000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000002000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400002000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200002000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000042000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600001000",
	.desc = "Counts data cache lines requests by software prefetch instructions that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000001000",
	.desc = "Counts data cache lines requests by software prefetch instructions that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400001000",
	.desc = "Counts data cache lines requests by software prefetch instructions that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200001000",
	.desc = "Counts data cache lines requests by software prefetch instructions that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000041000",
	.desc = "Counts data cache lines requests by software prefetch instructions that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.bus_locks.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010400",
	.desc = "Counts bus lock and split lock requests that have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts bus lock and split lock requests that have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.partial_writes.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000100",
	.desc = "Counts the number of demand write requests (RFO) generated by a write to partial data cache line, including the writes to uncacheable (UC) and write through (WT), and write protected (WP) types of memory that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of demand write requests (RFO) generated by a write to partial data cache line, including the writes to uncacheable (UC) and write through (WT), and write protected (WP) types of memory that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.partial_reads.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000080",
	.desc = "Counts demand data partial reads, including data in uncacheable (UC) or uncacheable write combining (USWC) memory types that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand data partial reads, including data in uncacheable (UC) or uncacheable write combining (USWC) memory types that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that are outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that are outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that are outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that are outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000001",
	.desc = "Counts demand cacheable data reads of full cache lines that are outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines that are outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x3600000001",
	.desc = "Counts demand cacheable data reads of full cache lines that miss the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines that miss the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000001",
	.desc = "Counts demand cacheable data reads of full cache lines that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines that miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000001",
	.desc = "Counts demand cacheable data reads of full cache lines that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines that miss the L2 cache with a snoop hit in the other processor module, no data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000001",
	.desc = "Counts demand cacheable data reads of full cache lines that true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines that true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040001",
	.desc = "Counts demand cacheable data reads of full cache lines that hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines that hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=200003,event=0x80",
	.desc = "References per ICache line that are available in the ICache (hit). This event counts differently than Intel processors based on Silvermont microarchitecture",
	.topic = "frontend",
	.long_desc = "Counts requests to the Instruction Cache (ICache) for one or more bytes in an ICache Line and that cache line is in the ICache (hit).  The event strives to count on a cache line basis, so that multiple accesses which hit in a single cache line count as one ICACHE.HIT.  Specifically, the event counts when straight line code crosses the cache line boundary, or when a branch target is to a new line, and that cache line is in the ICache. This event counts differently than Intel processors based on Silvermont microarchitecture",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "References per ICache line that are not available in the ICache (miss). This event counts differently than Intel processors based on Silvermont microarchitecture",
	.topic = "frontend",
	.long_desc = "Counts requests to the Instruction Cache (ICache)  for one or more bytes in an ICache Line and that cache line is not in the ICache (miss).  The event strives to count on a cache line basis, so that multiple accesses which miss in a single cache line count as one ICACHE.MISS.  Specifically, the event counts when straight line code crosses the cache line boundary, or when a branch target is to a new line, and that cache line is not in the ICache. This event counts differently than Intel processors based on Silvermont microarchitecture",
},
{
	.name = "icache.accesses",
	.event = "umask=0x3,period=200003,event=0x80",
	.desc = "References per ICache line. This event counts differently than Intel processors based on Silvermont microarchitecture",
	.topic = "frontend",
	.long_desc = "Counts requests to the Instruction Cache (ICache) for one or more bytes in an ICache Line.  The event strives to count on a cache line basis, so that multiple fetches to a single cache line count as one ICACHE.ACCESS.  Specifically, the event counts when accesses from straight line code crosses the cache line boundary, or when a branch target is to a new line.\r\nThis event counts differently than Intel processors based on Silvermont microarchitecture",
},
{
	.name = "ms_decoded.ms_entry",
	.event = "umask=0x1,period=200003,event=0xe7",
	.desc = "MS decode starts",
	.topic = "frontend",
	.long_desc = "Counts the number of times the Microcode Sequencer (MS) starts a flow of uops from the MSROM. It does not count every time a uop is read from the MSROM.  The most common case that this counts is when a micro-coded instruction is encountered by the front end of the machine.  Other cases include when an instruction encounters a fault, trap, or microcode assist of any sort that initiates a flow of uops.  The event will count MS startups for uops that are speculative, and subsequently cleared by branch mispredict or a machine clear",
},
{
	.name = "decode_restriction.predecode_wrong",
	.event = "umask=0x1,period=200003,event=0xe9",
	.desc = "Decode restrictions due to predicting wrong instruction length",
	.topic = "frontend",
	.long_desc = "Counts the number of times the prediction (from the predecode cache) for instruction length is incorrect",
},
{
	.name = "misalign_mem_ref.load_page_split",
	.event = "umask=0x2,period=200003,event=0x13",
	.desc = "Load uops that split a page (Precise event capable) (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts when a memory load of a uop spans a page boundary (a split) is retired (Must be precise)",
},
{
	.name = "misalign_mem_ref.store_page_split",
	.event = "umask=0x4,period=200003,event=0x13",
	.desc = "Store uops that split a page (Precise event capable) (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts when a memory store of a uop spans a page boundary (a split) is retired (Must be precise)",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=200003,event=0xc3",
	.desc = "Machine clears due to memory ordering issue",
	.topic = "memory",
	.long_desc = "Counts machine clears due to memory ordering issues.  This occurs when a snoop request happens and the machine is uncertain if memory ordering will be preserved as another core is in the process of modifying the data",
},
{
	.name = "fetch_stall.all",
	.event = "umask=0x0,period=200003,event=0x86",
	.desc = "Cycles code-fetch stalled due to any reason",
	.topic = "other",
	.long_desc = "Counts cycles that fetch is stalled due to any reason. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes.  This will include cycles due to an ITLB miss, ICache miss and other events",
},
{
	.name = "fetch_stall.itlb_fill_pending_cycles",
	.event = "umask=0x1,period=200003,event=0x86",
	.desc = "Cycles code-fetch stalled due to an outstanding ITLB miss",
	.topic = "other",
	.long_desc = "Counts cycles that fetch is stalled due to an outstanding ITLB miss. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes due to an ITLB miss.  Note: this event is not the same as page walk cycles to retrieve an instruction translation",
},
{
	.name = "issue_slots_not_consumed.any",
	.event = "umask=0x0,period=200003,event=0xca",
	.desc = "Unfilled issue slots per cycle",
	.topic = "other",
	.long_desc = "Counts the number of issue slots per core cycle that were not consumed by the backend due to either a full resource  in the backend (RESOURCE_FULL) or due to the processor recovering from some event (RECOVERY)",
},
{
	.name = "issue_slots_not_consumed.resource_full",
	.event = "umask=0x1,period=200003,event=0xca",
	.desc = "Unfilled issue slots per cycle because of a full resource in the backend",
	.topic = "other",
	.long_desc = "Counts the number of issue slots per core cycle that were not consumed because of a full resource in the backend.  Including but not limited to resources such as the Re-order Buffer (ROB), reservation stations (RS), load/store buffers, physical registers, or any other needed machine resource that is currently unavailable.   Note that uops must be available for consumption in order for this event to fire.  If a uop is not available (Instruction Queue is empty), this event will not count",
},
{
	.name = "issue_slots_not_consumed.recovery",
	.event = "umask=0x2,period=200003,event=0xca",
	.desc = "Unfilled issue slots per cycle to recover",
	.topic = "other",
	.long_desc = "Counts the number of issue slots per core cycle that were not consumed by the backend because allocation is stalled waiting for a mispredicted jump to retire or other branch-like conditions (e.g. the event is relevant during certain microcode flows).   Counts all issue slots blocked while within this window including slots where uops were not available in the Instruction Queue",
},
{
	.name = "hw_interrupts.received",
	.event = "umask=0x1,period=203,event=0xcb",
	.desc = "Hardware interrupts received",
	.topic = "other",
	.long_desc = "Counts hardware interrupts received by the processor",
},
{
	.name = "hw_interrupts.masked",
	.event = "umask=0x2,period=200003,event=0xcb",
	.desc = "Cycles hardware interrupts are masked",
	.topic = "other",
	.long_desc = "Counts the number of core cycles during which interrupts are masked (disabled). Increments by 1 each core cycle that EFLAGS.IF is 0, regardless of whether interrupts are pending or not",
},
{
	.name = "hw_interrupts.pending_and_masked",
	.event = "umask=0x4,period=200003,event=0xcb",
	.desc = "Cycles pending interrupts are masked",
	.topic = "other",
	.long_desc = "Counts core cycles during which there are pending interrupts, but interrupts are masked (EFLAGS.IF = 0)",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (Fixed event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions that retire execution. For instructions that consist of multiple uops, this event counts the retirement of the last uop of the instruction. The counter continues counting during hardware interrupts, traps, and inside interrupt handlers.  This event uses fixed counter 0.  You cannot collect a PEBs record for this event",
},
{
	.name = "cpu_clk_unhalted.core",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when core is not halted  (Fixed event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of core cycles while the core is not in a halt state.  The core enters the halt state when it is running the HLT instruction. In mobile systems the core frequency may change from time to time. For this reason this event may have a changing ratio with regards to time.  This event uses fixed counter 1.  You cannot collect a PEBs record for this event",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when core is not halted  (Fixed event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of reference cycles that the core is not in a halt state. The core enters the halt state when it is running the HLT instruction.  In mobile systems the core frequency may change from time.  This event is not affected by core frequency changes but counts as if the core is running at the maximum frequency all the time.  This event uses fixed counter 2.  You cannot collect a PEBs record for this event",
},
{
	.name = "ld_blocks.data_unknown",
	.event = "umask=0x1,period=200003,event=0x3",
	.desc = "Loads blocked due to store data not ready (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts a load blocked from using a store forward, but did not occur because the store data was not available at the right time.  The forward might occur subsequently when the data is available (Must be precise)",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=200003,event=0x3",
	.desc = "Loads blocked due to store forward restriction (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts a load blocked from using a store forward because of an address/size mismatch, only one of the loads blocked from each store will be counted (Must be precise)",
},
{
	.name = "ld_blocks.4k_alias",
	.event = "umask=0x4,period=200003,event=0x3",
	.desc = "Loads blocked because address has 4k partial address false dependence (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts loads that block because their address modulo 4K matches a pending store (Must be precise)",
},
{
	.name = "ld_blocks.utlb_miss",
	.event = "umask=0x8,period=200003,event=0x3",
	.desc = "Loads blocked because address in not in the UTLB (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts loads blocked because they are unable to find their physical address in the micro TLB (UTLB) (Must be precise)",
},
{
	.name = "ld_blocks.all_block",
	.event = "umask=0x10,period=200003,event=0x3",
	.desc = "Loads blocked (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts anytime a load that retires is blocked for any reason (Must be precise)",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x0,period=200003,event=0xe",
	.desc = "Uops issued to the back end per cycle",
	.topic = "pipeline",
	.long_desc = "Counts uops issued by the front end and allocated into the back end of the machine.  This event counts uops that retire as well as uops that were speculatively executed but didn't retire. The sort of speculative uops that might be counted includes, but is not limited to those uops issued in the shadow of a miss-predicted branch, those uops that are inserted during an assist (such as for a denormal floating point result), and (previously allocated) uops that might be canceled during a machine clear",
},
{
	.name = "cpu_clk_unhalted.core_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Core cycles when core is not halted",
	.topic = "pipeline",
	.long_desc = "Core cycles when core is not halted.  This event uses a (_P)rogrammable general purpose performance counter",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Reference cycles when core is not halted",
	.topic = "pipeline",
	.long_desc = "Reference cycles when core is not halted.  This event uses a programmable general purpose performance counter",
},
{
	.name = "uops_not_delivered.any",
	.event = "umask=0x0,period=200003,event=0x9c",
	.desc = "Uops requested but not-delivered to the back-end per cycle",
	.topic = "pipeline",
	.long_desc = "This event used to measure front-end inefficiencies. I.e. when front-end of the machine is not delivering uops to the back-end and the back-end has is not stalled. This event can be used to identify if the machine is truly front-end bound.  When this event occurs, it is an indication that the front-end of the machine is operating at less than its theoretical peak performance. Background: We can think of the processor pipeline as being divided into 2 broader parts: Front-end and Back-end. Front-end is responsible for fetching the instruction, decoding into uops in machine understandable format and putting them into a uop queue to be consumed by back end. The back-end then takes these uops, allocates the required resources.  When all resources are ready, uops are executed. If the back-end is not ready to accept uops from the front-end, then we do not want to count these as front-end bottlenecks.  However, whenever we have bottlenecks in the back-end, we will have allocation unit stalls and eventually forcing the front-end to wait until the back-end is ready to receive more uops. This event counts only when back-end is requesting more uops and front-end is not able to provide them. When 3 uops are requested and no uops are delivered, the event counts 3. When 3 are requested, and only 1 is delivered, the event counts 2. When only 2 are delivered, the event counts 1. Alternatively stated, the event will not count if 3 uops are delivered, or if the back end is stalled and not requesting any uops at all.  Counts indicate missed opportunities for the front-end to deliver a uop to the back end. Some examples of conditions that cause front-end efficiencies are: ICache misses, ITLB misses, and decoder restrictions that limit the front-end bandwidth. Known Issues: Some uops require multiple allocation slots.  These uops will not be charged as a front end 'not delivered' opportunity, and will be regarded as a back end problem. For example, the INC instruction has one uop that requires 2 issue slots.  A stream of INC instructions will not count as UOPS_NOT_DELIVERED, even though only one instruction can be issued per clock.  The low uop issue rate for a stream of INC instructions is considered to be a back end issue",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions that retire execution. For instructions that consist of multiple uops, this event counts the retirement of the last uop of the instruction. The event continues counting during hardware interrupts, traps, and inside interrupt handlers.  This is an architectural performance event.  This event uses a (_P)rogrammable general purpose performance counter. *This event is Precise Event capable:  The EventingRIP field in the PEBS record is precise to the address of the instruction which caused the event.  Note: Because PEBS records can be collected only on IA32_PMC0, only one event can use the PEBS facility at a time (Must be precise)",
},
{
	.name = "uops_retired.any",
	.event = "umask=0x0,period=2000003,event=0xc2",
	.desc = "Uops retired (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts uops which retired (Must be precise)",
},
{
	.name = "uops_retired.ms",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "MS uops retired (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts uops retired that are from the complex flows issued by the micro-sequencer (MS).  Counts both the uops from a micro-coded instruction, and the uops that might be generated from a micro-coded assist (Must be precise)",
},
{
	.name = "uops_retired.fpdiv",
	.event = "umask=0x8,period=2000003,event=0xc2",
	.desc = "Floating point divide uops retired. (Precise Event Capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts the number of floating point divide uops retired (Must be precise)",
},
{
	.name = "uops_retired.idiv",
	.event = "umask=0x10,period=2000003,event=0xc2",
	.desc = "Integer divide uops retired. (Precise Event Capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts the number of integer divide uops retired (Must be precise)",
},
{
	.name = "machine_clears.all",
	.event = "umask=0x0,period=200003,event=0xc3",
	.desc = "All machine clears",
	.topic = "pipeline",
	.long_desc = "Counts machine clears for any reason",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x1,period=200003,event=0xc3",
	.desc = "Self-Modifying Code detected",
	.topic = "pipeline",
	.long_desc = "Counts the number of times that the processor detects that a program is writing to a code section and has to perform a machine clear because of that modification.  Self-modifying code (SMC) causes a severe penalty in all Intel\u00ae architecture processors",
},
{
	.name = "machine_clears.fp_assist",
	.event = "umask=0x4,period=200003,event=0xc3",
	.desc = "Machine clears due to FP assists",
	.topic = "pipeline",
	.long_desc = "Counts machine clears due to floating point (FP) operations needing assists.  For instance, if the result was a floating point denormal, the hardware clears the pipeline and reissues uops to produce the correct IEEE compliant denormal result",
},
{
	.name = "machine_clears.disambiguation",
	.event = "umask=0x8,period=200003,event=0xc3",
	.desc = "Machine clears due to memory disambiguation",
	.topic = "pipeline",
	.long_desc = "Counts machine clears due to memory disambiguation.  Memory disambiguation happens when a load which has been issued conflicts with a previous unretired store in the pipeline whose address was not known at issue time, but is later resolved to be the same as the load address",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=200003,event=0xc4",
	.desc = "Retired branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts branch instructions retired for all branch types.  This is an architectural performance event (Must be precise)",
},
{
	.name = "br_inst_retired.jcc",
	.event = "umask=0x7e,period=200003,event=0xc4",
	.desc = "Retired conditional branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts retired Jcc (Jump on Conditional Code/Jump if Condition is Met) branch instructions retired, including both when the branch was taken and when it was not taken (Must be precise)",
},
{
	.name = "br_inst_retired.all_taken_branches",
	.event = "umask=0x80,period=200003,event=0xc4",
	.desc = "Retired taken branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts the number of taken branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0xbf,period=200003,event=0xc4",
	.desc = "Retired far branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts far branch instructions retired.  This includes far jump, far call and return, and Interrupt call and return (Must be precise)",
},
{
	.name = "br_inst_retired.non_return_ind",
	.event = "umask=0xeb,period=200003,event=0xc4",
	.desc = "Retired instructions of near indirect Jmp or call (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near indirect call or near indirect jmp branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.return",
	.event = "umask=0xf7,period=200003,event=0xc4",
	.desc = "Retired near return instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near return branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.call",
	.event = "umask=0xf9,period=200003,event=0xc4",
	.desc = "Retired near call instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near CALL branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.ind_call",
	.event = "umask=0xfb,period=200003,event=0xc4",
	.desc = "Retired near indirect call instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near indirect CALL branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.rel_call",
	.event = "umask=0xfd,period=200003,event=0xc4",
	.desc = "Retired near relative call instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near relative CALL branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.taken_jcc",
	.event = "umask=0xfe,period=200003,event=0xc4",
	.desc = "Retired conditional branch instructions that were taken (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts Jcc (Jump on Conditional Code/Jump if Condition is Met) branch instructions retired that were taken and does not count when the Jcc branch instruction were not taken (Must be precise)",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=200003,event=0xc5",
	.desc = "Retired mispredicted branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted branch instructions retired including all branch types (Must be precise)",
},
{
	.name = "br_misp_retired.jcc",
	.event = "umask=0x7e,period=200003,event=0xc5",
	.desc = "Retired mispredicted conditional branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted retired Jcc (Jump on Conditional Code/Jump if Condition is Met) branch instructions retired, including both when the branch was supposed to be taken and when it was not supposed to be taken (but the processor predicted the opposite condition) (Must be precise)",
},
{
	.name = "br_misp_retired.non_return_ind",
	.event = "umask=0xeb,period=200003,event=0xc5",
	.desc = "Retired mispredicted instructions of near indirect Jmp or near indirect call. (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted branch instructions retired that were near indirect call or near indirect jmp, where the target address taken was not what the processor predicted (Must be precise)",
},
{
	.name = "br_misp_retired.return",
	.event = "umask=0xf7,period=200003,event=0xc5",
	.desc = "Retired mispredicted near return instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted near RET branch instructions retired, where the return address taken was not what the processor predicted (Must be precise)",
},
{
	.name = "br_misp_retired.ind_call",
	.event = "umask=0xfb,period=200003,event=0xc5",
	.desc = "Retired mispredicted near indirect call instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted near indirect CALL branch instructions retired, where the target address taken was not what the processor predicted (Must be precise)",
},
{
	.name = "br_misp_retired.taken_jcc",
	.event = "umask=0xfe,period=200003,event=0xc5",
	.desc = "Retired mispredicted conditional branch instructions that were taken (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted retired Jcc (Jump on Conditional Code/Jump if Condition is Met) branch instructions retired that were supposed to be taken but the processor predicted that it would not be taken (Must be precise)",
},
{
	.name = "cycles_div_busy.all",
	.event = "umask=0x0,period=2000003,event=0xcd",
	.desc = "Cycles a divider is busy",
	.topic = "pipeline",
	.long_desc = "Counts core cycles if either divide unit is busy",
},
{
	.name = "cycles_div_busy.idiv",
	.event = "umask=0x1,period=200003,event=0xcd",
	.desc = "Cycles the integer divide unit is busy",
	.topic = "pipeline",
	.long_desc = "Counts core cycles the integer divide unit is busy",
},
{
	.name = "cycles_div_busy.fpdiv",
	.event = "umask=0x2,period=200003,event=0xcd",
	.desc = "Cycles the FP divide unit is busy",
	.topic = "pipeline",
	.long_desc = "Counts core cycles the floating point divide unit is busy",
},
{
	.name = "baclears.all",
	.event = "umask=0x1,period=200003,event=0xe6",
	.desc = "BACLEARs asserted for any branch type",
	.topic = "pipeline",
	.long_desc = "Counts the number of times a BACLEAR is signaled for any reason, including, but not limited to indirect branch/call,  Jcc (Jump on Conditional Code/Jump if Condition is Met) branch, unconditional branch/call, and returns",
},
{
	.name = "baclears.return",
	.event = "umask=0x8,period=200003,event=0xe6",
	.desc = "BACLEARs asserted for return branch",
	.topic = "pipeline",
	.long_desc = "Counts BACLEARS on return instructions",
},
{
	.name = "baclears.cond",
	.event = "umask=0x10,period=200003,event=0xe6",
	.desc = "BACLEARs asserted for conditional branch",
	.topic = "pipeline",
	.long_desc = "Counts BACLEARS on Jcc (Jump on Conditional Code/Jump if Condition is Met) branches",
},
{
	.name = "page_walks.d_side_cycles",
	.event = "umask=0x1,period=200003,event=0x5",
	.desc = "Duration of D-side page-walks in cycles",
	.topic = "virtual memory",
	.long_desc = "Counts every core cycle when a Data-side (walks due to a data operation) page walk is in progress",
},
{
	.name = "page_walks.i_side_cycles",
	.event = "umask=0x2,period=200003,event=0x5",
	.desc = "Duration of I-side pagewalks in cycles",
	.topic = "virtual memory",
	.long_desc = "Counts every core cycle when a Instruction-side (walks due to an instruction fetch) page walk is in progress",
},
{
	.name = "page_walks.cycles",
	.event = "umask=0x3,period=200003,event=0x5",
	.desc = "Duration of page-walks in cycles",
	.topic = "virtual memory",
	.long_desc = "Counts every core cycle a page-walk is in progress due to either a data memory operation or an instruction fetch",
},
{
	.name = "itlb.miss",
	.event = "umask=0x4,period=200003,event=0x81",
	.desc = "ITLB misses",
	.topic = "virtual memory",
	.long_desc = "Counts the number of times the machine was unable to find a translation in the Instruction Translation Lookaside Buffer (ITLB) for a linear address of an instruction fetch.  It counts when new translation are filled into the ITLB.  The event is speculative in nature, but will not count translations (page walks) that are begun and not finished, or translations that are finished but not filled into the ITLB",
},
{
	.name = "mem_uops_retired.dtlb_miss_loads",
	.event = "umask=0x11,period=200003,event=0xd0",
	.desc = "Load uops retired that missed the DTLB (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "virtual memory",
	.long_desc = "Counts load uops retired that caused a DTLB miss  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.dtlb_miss_stores",
	.event = "umask=0x12,period=200003,event=0xd0",
	.desc = "Store uops retired that missed the DTLB (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "virtual memory",
	.long_desc = "Counts store uops retired that caused a DTLB miss  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.dtlb_miss",
	.event = "umask=0x13,period=200003,event=0xd0",
	.desc = "Memory uops retired that missed the DTLB (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "virtual memory",
	.long_desc = "Counts uops retired that had a DTLB miss on load, store or either.  Note that when two distinct memory operations to the same page miss the DTLB, only one of them will be recorded as a DTLB miss  Supports address when precise (Must be precise)",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_goldmontplus[] = {
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=200003,event=0x2e",
	.desc = "L2 cache request misses",
	.topic = "cache",
	.long_desc = "Counts memory requests originating from the core that miss in the L2 cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=200003,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
	.long_desc = "Counts memory requests originating from the core that reference a cache line in the L2 cache",
},
{
	.name = "l2_reject_xq.all",
	.event = "umask=0x0,period=200003,event=0x30",
	.desc = "Requests rejected by the XQ",
	.topic = "cache",
	.long_desc = "Counts the number of demand and prefetch transactions that the L2 XQ rejects due to a full or near full condition which likely indicates back pressure from the intra-die interconnect (IDI) fabric. The XQ may reject transactions from the L2Q (non-cacheable requests), L2 misses and L2 write-back victims",
},
{
	.name = "core_reject_l2q.all",
	.event = "umask=0x0,period=200003,event=0x31",
	.desc = "Requests rejected by the L2Q",
	.topic = "cache",
	.long_desc = "Counts the number of demand and L1 prefetcher requests rejected by the L2Q due to a full or nearly full condition which likely indicates back pressure from L2Q. It also counts requests that would have gone directly to the XQ, but are rejected due to a full or nearly full condition, indicating back pressure from the IDI link. The L2Q may also reject transactions from a core to insure fairness between cores, or to delay a core's dirty eviction when the address conflicts with incoming external snoops",
},
{
	.name = "dl1.replacement",
	.event = "umask=0x1,period=200003,event=0x51",
	.desc = "L1 Cache evictions for dirty data",
	.topic = "cache",
	.long_desc = "Counts when a modified (dirty) cache line is evicted from the data L1 cache and needs to be written back to memory.  No count will occur if the evicted line is clean, and hence does not require a writeback",
},
{
	.name = "fetch_stall.icache_fill_pending_cycles",
	.event = "umask=0x2,period=200003,event=0x86",
	.desc = "Cycles code-fetch stalled due to an outstanding ICache miss",
	.topic = "cache",
	.long_desc = "Counts cycles that fetch is stalled due to an outstanding ICache miss. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes due to an ICache miss.  Note: this event is not the same as the total number of cycles spent retrieving instruction cache lines from the memory hierarchy",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100007,event=0xb7",
	.desc = "Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=200003,event=0xd0",
	.desc = "Locked load uops retired (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts locked memory uops retired.  This includes regular locks and bus locks. (To specifically count bus locks only, see the Offcore response event.)  A locked access is one with a lock prefix, or an exchange to memory.  See the SDM for a complete description of which memory load accesses are locks  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=200003,event=0xd0",
	.desc = "Load uops retired that split a cache-line (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired where the data requested spans a 64 byte cache line boundary  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=200003,event=0xd0",
	.desc = "Stores uops retired that split a cache-line (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts store uops retired where the data requested spans a 64 byte cache line boundary  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.split",
	.event = "umask=0x43,period=200003,event=0xd0",
	.desc = "Memory uops retired that split a cache-line (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts memory uops retired where the data requested spans a 64 byte cache line boundary  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=200003,event=0xd0",
	.desc = "Load uops retired (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts the number of load uops retired  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=200003,event=0xd0",
	.desc = "Store uops retired (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts the number of store uops retired  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.all",
	.event = "umask=0x83,period=200003,event=0xd0",
	.desc = "Memory uops retired (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts the number of memory uops retired that is either a loads or a store or both  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=200003,event=0xd1",
	.desc = "Load uops retired that hit L1 data cache (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired that hit the L1 data cache  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=200003,event=0xd1",
	.desc = "Load uops retired that hit L2 (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired that hit in the L2 cache  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=200003,event=0xd1",
	.desc = "Load uops retired that missed L1 data cache (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired that miss the L1 data cache  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=200003,event=0xd1",
	.desc = "Load uops retired that missed L2 (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired that miss in the L2 cache  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.hitm",
	.event = "umask=0x20,period=200003,event=0xd1",
	.desc = "Memory uop retired where cross core or cross module HITM occurred (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts load uops retired where the cache line containing the data was in the modified state of another core or modules cache (HITM).  More specifically, this means that when the load address was checked by other caching agents (typically another processor) in the system, one of those caching agents indicated that they had a dirty copy of the data.  Loads that obtain a HITM response incur greater latency than most is typical for a load.  In addition, since HITM indicates that some other processor had this data in its cache, it implies that the data was shared between processors, or potentially was a lock or semaphore value.  This event is useful for locating sharing, false sharing, and contended locks  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.wcb_hit",
	.event = "umask=0x40,period=200003,event=0xd1",
	.desc = "Loads retired that hit WCB (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts memory load uops retired where the data is retrieved from the WCB (or fill buffer), indicating that the load found its data while that data was in the process of being brought into the L1 cache.  Typically a load will receive this indication when some other load or prefetch missed the L1 cache and was in the process of retrieving the cache line containing the data, but that process had not yet finished (and written the data back to the cache). For example, consider load X and Y, both referencing the same cache line that is not in the L1 cache.  If load X misses cache first, it obtains and WCB (or fill buffer) and begins the process of requesting the data.  When load Y requests the data, it will either hit the WCB, or the L1 cache, depending on exactly what time the request to Y occurs  Supports address when precise (Must be precise)",
},
{
	.name = "mem_load_uops_retired.dram_hit",
	.event = "umask=0x80,period=200003,event=0xd1",
	.desc = "Loads retired that came from DRAM (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "cache",
	.long_desc = "Counts memory load uops retired where the data is retrieved from DRAM.  Event is counted at retirement, so the speculative loads are ignored.  A memory load can hit (or miss) the L1 cache, hit (or miss) the L2 cache, hit DRAM, hit in the WCB or receive a HITM response  Supports address when precise (Must be precise)",
},
{
	.name = "offcore_response.demand_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010001",
	.desc = "Counts demand cacheable data reads of full cache lines have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040001",
	.desc = "Counts demand cacheable data reads of full cache lines hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000001",
	.desc = "Counts demand cacheable data reads of full cache lines true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000001",
	.desc = "Counts demand cacheable data reads of full cache lines miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000001",
	.desc = "Counts demand cacheable data reads of full cache lines outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts demand cacheable data reads of full cache lines outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_rfo.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000002",
	.desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts demand reads for ownership (RFO) requests generated by a write to full data cache line outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.demand_code_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000004",
	.desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts demand instruction cacheline and I-side prefetch requests that miss the instruction cache outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.corewb.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000008",
	.desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts the number of writeback transactions caused by L1 or L2 cache evictions outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000010",
	.desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts data cacheline reads generated by hardware L2 cache prefetcher outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l2_rfo.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000020",
	.desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests generated by L2 prefetcher outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.bus_locks.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010400",
	.desc = "Counts bus lock and split lock requests have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts bus lock and split lock requests have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.bus_locks.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040400",
	.desc = "Counts bus lock and split lock requests hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts bus lock and split lock requests hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.bus_locks.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000400",
	.desc = "Counts bus lock and split lock requests true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts bus lock and split lock requests true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.bus_locks.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000400",
	.desc = "Counts bus lock and split lock requests miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts bus lock and split lock requests miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.bus_locks.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000400",
	.desc = "Counts bus lock and split lock requests outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts bus lock and split lock requests outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.full_streaming_stores.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000800",
	.desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts full cache line data writes to uncacheable write combining (USWC) memory region and full cache-line non-temporal writes outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000011000",
	.desc = "Counts data cache lines requests by software prefetch instructions have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000041000",
	.desc = "Counts data cache lines requests by software prefetch instructions hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200001000",
	.desc = "Counts data cache lines requests by software prefetch instructions true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000001000",
	.desc = "Counts data cache lines requests by software prefetch instructions miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.sw_prefetch.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000001000",
	.desc = "Counts data cache lines requests by software prefetch instructions outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts data cache lines requests by software prefetch instructions outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000012000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000042000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200002000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000002000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.pf_l1_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000002000",
	.desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts data cache line reads generated by hardware L1 data cache prefetcher outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.streaming_stores.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000014800",
	.desc = "Counts any data writes to uncacheable write combining (USWC) memory region  have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts any data writes to uncacheable write combining (USWC) memory region  have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.streaming_stores.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000044800",
	.desc = "Counts any data writes to uncacheable write combining (USWC) memory region  hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts any data writes to uncacheable write combining (USWC) memory region  hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.streaming_stores.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200004800",
	.desc = "Counts any data writes to uncacheable write combining (USWC) memory region  true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts any data writes to uncacheable write combining (USWC) memory region  true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.streaming_stores.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000004800",
	.desc = "Counts any data writes to uncacheable write combining (USWC) memory region  miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts any data writes to uncacheable write combining (USWC) memory region  miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.streaming_stores.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000004800",
	.desc = "Counts any data writes to uncacheable write combining (USWC) memory region  outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts any data writes to uncacheable write combining (USWC) memory region  outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000018000",
	.desc = "Counts requests to the uncore subsystem have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000048000",
	.desc = "Counts requests to the uncore subsystem hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200008000",
	.desc = "Counts requests to the uncore subsystem true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000008000",
	.desc = "Counts requests to the uncore subsystem miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_request.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000008000",
	.desc = "Counts requests to the uncore subsystem outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts requests to the uncore subsystem outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000013010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000043010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200003010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000003010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_pf_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000003010",
	.desc = "Counts data reads generated by L1 or L2 prefetchers outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts data reads generated by L1 or L2 prefetchers outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000013091",
	.desc = "Counts data reads (demand & prefetch) have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000043091",
	.desc = "Counts data reads (demand & prefetch) hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200003091",
	.desc = "Counts data reads (demand & prefetch) true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000003091",
	.desc = "Counts data reads (demand & prefetch) miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000003091",
	.desc = "Counts data reads (demand & prefetch) outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts data reads (demand & prefetch) outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000040022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_rfo.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000022",
	.desc = "Counts reads for ownership (RFO) requests (demand & prefetch) outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts reads for ownership (RFO) requests (demand & prefetch) outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_read.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00000132b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) have any transaction responses from the uncore subsystem",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) have any transaction responses from the uncore subsystem. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_read.l2_hit",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00000432b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) hit the L2 cache",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) hit the L2 cache. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_read.l2_miss.snoop_miss_or_no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x02000032b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) true miss for the L2 cache with a snoop miss in the other processor module",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) true miss for the L2 cache with a snoop miss in the other processor module.  Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_read.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x10000032b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) miss the L2 cache with a snoop hit in the other processor module, data forwarding is required",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) miss the L2 cache with a snoop hit in the other processor module, data forwarding is required. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "offcore_response.any_read.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x40000032b7",
	.desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
	.long_desc = "Counts data read, code read, and read for ownership (RFO) requests (demand & prefetch) outstanding, per cycle, from the time of the L2 miss to when any response is received. Requires MSR_OFFCORE_RESP[0,1] to specify request type and response. (duplicated for both MSRs)",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=200003,event=0x80",
	.desc = "References per ICache line that are available in the ICache (hit). This event counts differently than Intel processors based on Silvermont microarchitecture",
	.topic = "frontend",
	.long_desc = "Counts requests to the Instruction Cache (ICache) for one or more bytes in an ICache Line and that cache line is in the ICache (hit).  The event strives to count on a cache line basis, so that multiple accesses which hit in a single cache line count as one ICACHE.HIT.  Specifically, the event counts when straight line code crosses the cache line boundary, or when a branch target is to a new line, and that cache line is in the ICache. This event counts differently than Intel processors based on Silvermont microarchitecture",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "References per ICache line that are not available in the ICache (miss). This event counts differently than Intel processors based on Silvermont microarchitecture",
	.topic = "frontend",
	.long_desc = "Counts requests to the Instruction Cache (ICache)  for one or more bytes in an ICache Line and that cache line is not in the ICache (miss).  The event strives to count on a cache line basis, so that multiple accesses which miss in a single cache line count as one ICACHE.MISS.  Specifically, the event counts when straight line code crosses the cache line boundary, or when a branch target is to a new line, and that cache line is not in the ICache. This event counts differently than Intel processors based on Silvermont microarchitecture",
},
{
	.name = "icache.accesses",
	.event = "umask=0x3,period=200003,event=0x80",
	.desc = "References per ICache line. This event counts differently than Intel processors based on Silvermont microarchitecture",
	.topic = "frontend",
	.long_desc = "Counts requests to the Instruction Cache (ICache) for one or more bytes in an ICache Line.  The event strives to count on a cache line basis, so that multiple fetches to a single cache line count as one ICACHE.ACCESS.  Specifically, the event counts when accesses from straight line code crosses the cache line boundary, or when a branch target is to a new line.\r\nThis event counts differently than Intel processors based on Silvermont microarchitecture",
},
{
	.name = "ms_decoded.ms_entry",
	.event = "umask=0x1,period=200003,event=0xe7",
	.desc = "MS decode starts",
	.topic = "frontend",
	.long_desc = "Counts the number of times the Microcode Sequencer (MS) starts a flow of uops from the MSROM. It does not count every time a uop is read from the MSROM.  The most common case that this counts is when a micro-coded instruction is encountered by the front end of the machine.  Other cases include when an instruction encounters a fault, trap, or microcode assist of any sort that initiates a flow of uops.  The event will count MS startups for uops that are speculative, and subsequently cleared by branch mispredict or a machine clear",
},
{
	.name = "decode_restriction.predecode_wrong",
	.event = "umask=0x1,period=200003,event=0xe9",
	.desc = "Decode restrictions due to predicting wrong instruction length",
	.topic = "frontend",
	.long_desc = "Counts the number of times the prediction (from the predecode cache) for instruction length is incorrect",
},
{
	.name = "misalign_mem_ref.load_page_split",
	.event = "umask=0x2,period=200003,event=0x13",
	.desc = "Load uops that split a page (Precise event capable) (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts when a memory load of a uop spans a page boundary (a split) is retired (Must be precise)",
},
{
	.name = "misalign_mem_ref.store_page_split",
	.event = "umask=0x4,period=200003,event=0x13",
	.desc = "Store uops that split a page (Precise event capable) (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts when a memory store of a uop spans a page boundary (a split) is retired (Must be precise)",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=20003,event=0xc3",
	.desc = "Machine clears due to memory ordering issue",
	.topic = "memory",
	.long_desc = "Counts machine clears due to memory ordering issues.  This occurs when a snoop request happens and the machine is uncertain if memory ordering will be preserved - as another core is in the process of modifying the data",
},
{
	.name = "fetch_stall.all",
	.event = "umask=0x0,period=200003,event=0x86",
	.desc = "Cycles code-fetch stalled due to any reason",
	.topic = "other",
	.long_desc = "Counts cycles that fetch is stalled due to any reason. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes.  This will include cycles due to an ITLB miss, ICache miss and other events",
},
{
	.name = "fetch_stall.itlb_fill_pending_cycles",
	.event = "umask=0x1,period=200003,event=0x86",
	.desc = "Cycles the code-fetch stalls and an ITLB miss is outstanding",
	.topic = "other",
	.long_desc = "Counts cycles that fetch is stalled due to an outstanding ITLB miss. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes due to an ITLB miss.  Note: this event is not the same as page walk cycles to retrieve an instruction translation",
},
{
	.name = "issue_slots_not_consumed.any",
	.event = "umask=0x0,period=200003,event=0xca",
	.desc = "Unfilled issue slots per cycle",
	.topic = "other",
	.long_desc = "Counts the number of issue slots per core cycle that were not consumed by the backend due to either a full resource  in the backend (RESOURCE_FULL) or due to the processor recovering from some event (RECOVERY)",
},
{
	.name = "issue_slots_not_consumed.resource_full",
	.event = "umask=0x1,period=200003,event=0xca",
	.desc = "Unfilled issue slots per cycle because of a full resource in the backend",
	.topic = "other",
	.long_desc = "Counts the number of issue slots per core cycle that were not consumed because of a full resource in the backend.  Including but not limited to resources such as the Re-order Buffer (ROB), reservation stations (RS), load/store buffers, physical registers, or any other needed machine resource that is currently unavailable.   Note that uops must be available for consumption in order for this event to fire.  If a uop is not available (Instruction Queue is empty), this event will not count",
},
{
	.name = "issue_slots_not_consumed.recovery",
	.event = "umask=0x2,period=200003,event=0xca",
	.desc = "Unfilled issue slots per cycle to recover",
	.topic = "other",
	.long_desc = "Counts the number of issue slots per core cycle that were not consumed by the backend because allocation is stalled waiting for a mispredicted jump to retire or other branch-like conditions (e.g. the event is relevant during certain microcode flows).   Counts all issue slots blocked while within this window including slots where uops were not available in the Instruction Queue",
},
{
	.name = "hw_interrupts.received",
	.event = "umask=0x1,period=203,event=0xcb",
	.desc = "Hardware interrupts received",
	.topic = "other",
	.long_desc = "Counts hardware interrupts received by the processor",
},
{
	.name = "hw_interrupts.masked",
	.event = "umask=0x2,period=200003,event=0xcb",
	.desc = "Cycles hardware interrupts are masked",
	.topic = "other",
	.long_desc = "Counts the number of core cycles during which interrupts are masked (disabled). Increments by 1 each core cycle that EFLAGS.IF is 0, regardless of whether interrupts are pending or not",
},
{
	.name = "hw_interrupts.pending_and_masked",
	.event = "umask=0x4,period=200003,event=0xcb",
	.desc = "Cycles pending interrupts are masked",
	.topic = "other",
	.long_desc = "Counts core cycles during which there are pending interrupts, but interrupts are masked (EFLAGS.IF = 0)",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (Fixed event) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions that retire execution. For instructions that consist of multiple uops, this event counts the retirement of the last uop of the instruction. The counter continues counting during hardware interrupts, traps, and inside interrupt handlers.  This event uses fixed counter 0.  You cannot collect a PEBs record for this event (Must be precise)",
},
{
	.name = "cpu_clk_unhalted.core",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when core is not halted  (Fixed event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of core cycles while the core is not in a halt state.  The core enters the halt state when it is running the HLT instruction. In mobile systems the core frequency may change from time to time. For this reason this event may have a changing ratio with regards to time.  This event uses fixed counter 1.  You cannot collect a PEBs record for this event",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when core is not halted  (Fixed event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of reference cycles that the core is not in a halt state. The core enters the halt state when it is running the HLT instruction.  In mobile systems the core frequency may change from time.  This event is not affected by core frequency changes but counts as if the core is running at the maximum frequency all the time.  This event uses fixed counter 2.  You cannot collect a PEBs record for this event",
},
{
	.name = "ld_blocks.data_unknown",
	.event = "umask=0x1,period=200003,event=0x3",
	.desc = "Loads blocked due to store data not ready (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts a load blocked from using a store forward, but did not occur because the store data was not available at the right time.  The forward might occur subsequently when the data is available (Must be precise)",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=200003,event=0x3",
	.desc = "Loads blocked due to store forward restriction (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts a load blocked from using a store forward because of an address/size mismatch, only one of the loads blocked from each store will be counted (Must be precise)",
},
{
	.name = "ld_blocks.4k_alias",
	.event = "umask=0x4,period=200003,event=0x3",
	.desc = "Loads blocked because address has 4k partial address false dependence (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts loads that block because their address modulo 4K matches a pending store (Must be precise)",
},
{
	.name = "ld_blocks.utlb_miss",
	.event = "umask=0x8,period=200003,event=0x3",
	.desc = "Loads blocked because address in not in the UTLB (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts loads blocked because they are unable to find their physical address in the micro TLB (UTLB) (Must be precise)",
},
{
	.name = "ld_blocks.all_block",
	.event = "umask=0x10,period=200003,event=0x3",
	.desc = "Loads blocked (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts anytime a load that retires is blocked for any reason (Must be precise)",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x0,period=200003,event=0xe",
	.desc = "Uops issued to the back end per cycle",
	.topic = "pipeline",
	.long_desc = "Counts uops issued by the front end and allocated into the back end of the machine.  This event counts uops that retire as well as uops that were speculatively executed but didn't retire. The sort of speculative uops that might be counted includes, but is not limited to those uops issued in the shadow of a miss-predicted branch, those uops that are inserted during an assist (such as for a denormal floating point result), and (previously allocated) uops that might be canceled during a machine clear",
},
{
	.name = "cpu_clk_unhalted.core_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Core cycles when core is not halted",
	.topic = "pipeline",
	.long_desc = "Core cycles when core is not halted.  This event uses a (_P)rogrammable general purpose performance counter",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Reference cycles when core is not halted",
	.topic = "pipeline",
	.long_desc = "Reference cycles when core is not halted.  This event uses a (_P)rogrammable general purpose performance counter",
},
{
	.name = "uops_not_delivered.any",
	.event = "umask=0x0,period=200003,event=0x9c",
	.desc = "Uops requested but not-delivered to the back-end per cycle",
	.topic = "pipeline",
	.long_desc = "This event used to measure front-end inefficiencies. I.e. when front-end of the machine is not delivering uops to the back-end and the back-end has is not stalled. This event can be used to identify if the machine is truly front-end bound.  When this event occurs, it is an indication that the front-end of the machine is operating at less than its theoretical peak performance. Background: We can think of the processor pipeline as being divided into 2 broader parts: Front-end and Back-end. Front-end is responsible for fetching the instruction, decoding into uops in machine understandable format and putting them into a uop queue to be consumed by back end. The back-end then takes these uops, allocates the required resources.  When all resources are ready, uops are executed. If the back-end is not ready to accept uops from the front-end, then we do not want to count these as front-end bottlenecks.  However, whenever we have bottlenecks in the back-end, we will have allocation unit stalls and eventually forcing the front-end to wait until the back-end is ready to receive more uops. This event counts only when back-end is requesting more uops and front-end is not able to provide them. When 3 uops are requested and no uops are delivered, the event counts 3. When 3 are requested, and only 1 is delivered, the event counts 2. When only 2 are delivered, the event counts 1. Alternatively stated, the event will not count if 3 uops are delivered, or if the back end is stalled and not requesting any uops at all.  Counts indicate missed opportunities for the front-end to deliver a uop to the back end. Some examples of conditions that cause front-end efficiencies are: ICache misses, ITLB misses, and decoder restrictions that limit the front-end bandwidth. Known Issues: Some uops require multiple allocation slots.  These uops will not be charged as a front end 'not delivered' opportunity, and will be regarded as a back end problem. For example, the INC instruction has one uop that requires 2 issue slots.  A stream of INC instructions will not count as UOPS_NOT_DELIVERED, even though only one instruction can be issued per clock.  The low uop issue rate for a stream of INC instructions is considered to be a back end issue",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions that retire execution. For instructions that consist of multiple uops, this event counts the retirement of the last uop of the instruction. The event continues counting during hardware interrupts, traps, and inside interrupt handlers.  This is an architectural performance event.  This event uses a (_P)rogrammable general purpose performance counter. *This event is Precise Event capable:  The EventingRIP field in the PEBS record is precise to the address of the instruction which caused the event.  Note: Because PEBS records can be collected only on IA32_PMC0, only one event can use the PEBS facility at a time (Must be precise)",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x0,period=2000003,event=0xc0",
	.desc = "Instructions retired - using Reduced Skid PEBS feature (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts INST_RETIRED.ANY using the Reduced Skid PEBS feature that reduces the shadow in which events aren't counted allowing for a more unbiased distribution of samples across instructions retired (Must be precise)",
},
{
	.name = "uops_retired.any",
	.event = "umask=0x0,period=2000003,event=0xc2",
	.desc = "Uops retired (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts uops which retired (Must be precise)",
},
{
	.name = "uops_retired.ms",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "MS uops retired (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts uops retired that are from the complex flows issued by the micro-sequencer (MS).  Counts both the uops from a micro-coded instruction, and the uops that might be generated from a micro-coded assist (Must be precise)",
},
{
	.name = "uops_retired.fpdiv",
	.event = "umask=0x8,period=2000003,event=0xc2",
	.desc = "Floating point divide uops retired (Precise Event Capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts the number of floating point divide uops retired (Must be precise)",
},
{
	.name = "uops_retired.idiv",
	.event = "umask=0x10,period=2000003,event=0xc2",
	.desc = "Integer divide uops retired (Precise Event Capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts the number of integer divide uops retired (Must be precise)",
},
{
	.name = "machine_clears.all",
	.event = "umask=0x0,period=20003,event=0xc3",
	.desc = "All machine clears",
	.topic = "pipeline",
	.long_desc = "Counts machine clears for any reason",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x1,period=20003,event=0xc3",
	.desc = "Self-Modifying Code detected",
	.topic = "pipeline",
	.long_desc = "Counts the number of times that the processor detects that a program is writing to a code section and has to perform a machine clear because of that modification.  Self-modifying code (SMC) causes a severe penalty in all Intel\u00ae architecture processors",
},
{
	.name = "machine_clears.fp_assist",
	.event = "umask=0x4,period=20003,event=0xc3",
	.desc = "Machine clears due to FP assists",
	.topic = "pipeline",
	.long_desc = "Counts machine clears due to floating point (FP) operations needing assists.  For instance, if the result was a floating point denormal, the hardware clears the pipeline and reissues uops to produce the correct IEEE compliant denormal result",
},
{
	.name = "machine_clears.disambiguation",
	.event = "umask=0x8,period=20003,event=0xc3",
	.desc = "Machine clears due to memory disambiguation",
	.topic = "pipeline",
	.long_desc = "Counts machine clears due to memory disambiguation.  Memory disambiguation happens when a load which has been issued conflicts with a previous unretired store in the pipeline whose address was not known at issue time, but is later resolved to be the same as the load address",
},
{
	.name = "machine_clears.page_fault",
	.event = "umask=0x20,period=20003,event=0xc3",
	.desc = "Machines clear due to a page fault",
	.topic = "pipeline",
	.long_desc = "Counts the number of times that the machines clears due to a page fault. Covers both I-side and D-side(Loads/Stores) page faults. A page fault occurs when either page is not present, or an access violation",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=200003,event=0xc4",
	.desc = "Retired branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts branch instructions retired for all branch types.  This is an architectural performance event (Must be precise)",
},
{
	.name = "br_inst_retired.jcc",
	.event = "umask=0x7e,period=200003,event=0xc4",
	.desc = "Retired conditional branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts retired Jcc (Jump on Conditional Code/Jump if Condition is Met) branch instructions retired, including both when the branch was taken and when it was not taken (Must be precise)",
},
{
	.name = "br_inst_retired.all_taken_branches",
	.event = "umask=0x80,period=200003,event=0xc4",
	.desc = "Retired taken branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts the number of taken branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0xbf,period=200003,event=0xc4",
	.desc = "Retired far branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts far branch instructions retired.  This includes far jump, far call and return, and Interrupt call and return (Must be precise)",
},
{
	.name = "br_inst_retired.non_return_ind",
	.event = "umask=0xeb,period=200003,event=0xc4",
	.desc = "Retired instructions of near indirect Jmp or call (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near indirect call or near indirect jmp branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.return",
	.event = "umask=0xf7,period=200003,event=0xc4",
	.desc = "Retired near return instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near return branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.call",
	.event = "umask=0xf9,period=200003,event=0xc4",
	.desc = "Retired near call instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near CALL branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.ind_call",
	.event = "umask=0xfb,period=200003,event=0xc4",
	.desc = "Retired near indirect call instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near indirect CALL branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.rel_call",
	.event = "umask=0xfd,period=200003,event=0xc4",
	.desc = "Retired near relative call instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts near relative CALL branch instructions retired (Must be precise)",
},
{
	.name = "br_inst_retired.taken_jcc",
	.event = "umask=0xfe,period=200003,event=0xc4",
	.desc = "Retired conditional branch instructions that were taken (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts Jcc (Jump on Conditional Code/Jump if Condition is Met) branch instructions retired that were taken and does not count when the Jcc branch instruction were not taken (Must be precise)",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=200003,event=0xc5",
	.desc = "Retired mispredicted branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted branch instructions retired including all branch types (Must be precise)",
},
{
	.name = "br_misp_retired.jcc",
	.event = "umask=0x7e,period=200003,event=0xc5",
	.desc = "Retired mispredicted conditional branch instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted retired Jcc (Jump on Conditional Code/Jump if Condition is Met) branch instructions retired, including both when the branch was supposed to be taken and when it was not supposed to be taken (but the processor predicted the opposite condition) (Must be precise)",
},
{
	.name = "br_misp_retired.non_return_ind",
	.event = "umask=0xeb,period=200003,event=0xc5",
	.desc = "Retired mispredicted instructions of near indirect Jmp or near indirect call (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted branch instructions retired that were near indirect call or near indirect jmp, where the target address taken was not what the processor predicted (Must be precise)",
},
{
	.name = "br_misp_retired.return",
	.event = "umask=0xf7,period=200003,event=0xc5",
	.desc = "Retired mispredicted near return instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted near RET branch instructions retired, where the return address taken was not what the processor predicted (Must be precise)",
},
{
	.name = "br_misp_retired.ind_call",
	.event = "umask=0xfb,period=200003,event=0xc5",
	.desc = "Retired mispredicted near indirect call instructions (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted near indirect CALL branch instructions retired, where the target address taken was not what the processor predicted (Must be precise)",
},
{
	.name = "br_misp_retired.taken_jcc",
	.event = "umask=0xfe,period=200003,event=0xc5",
	.desc = "Retired mispredicted conditional branch instructions that were taken (Precise event capable) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted retired Jcc (Jump on Conditional Code/Jump if Condition is Met) branch instructions retired that were supposed to be taken but the processor predicted that it would not be taken (Must be precise)",
},
{
	.name = "cycles_div_busy.all",
	.event = "umask=0x0,period=2000003,event=0xcd",
	.desc = "Cycles a divider is busy",
	.topic = "pipeline",
	.long_desc = "Counts core cycles if either divide unit is busy",
},
{
	.name = "cycles_div_busy.idiv",
	.event = "umask=0x1,period=200003,event=0xcd",
	.desc = "Cycles the integer divide unit is busy",
	.topic = "pipeline",
	.long_desc = "Counts core cycles the integer divide unit is busy",
},
{
	.name = "cycles_div_busy.fpdiv",
	.event = "umask=0x2,period=200003,event=0xcd",
	.desc = "Cycles the FP divide unit is busy",
	.topic = "pipeline",
	.long_desc = "Counts core cycles the floating point divide unit is busy",
},
{
	.name = "baclears.all",
	.event = "umask=0x1,period=200003,event=0xe6",
	.desc = "BACLEARs asserted for any branch type",
	.topic = "pipeline",
	.long_desc = "Counts the number of times a BACLEAR is signaled for any reason, including, but not limited to indirect branch/call,  Jcc (Jump on Conditional Code/Jump if Condition is Met) branch, unconditional branch/call, and returns",
},
{
	.name = "baclears.return",
	.event = "umask=0x8,period=200003,event=0xe6",
	.desc = "BACLEARs asserted for return branch",
	.topic = "pipeline",
	.long_desc = "Counts BACLEARS on return instructions",
},
{
	.name = "baclears.cond",
	.event = "umask=0x10,period=200003,event=0xe6",
	.desc = "BACLEARs asserted for conditional branch",
	.topic = "pipeline",
	.long_desc = "Counts BACLEARS on Jcc (Jump on Conditional Code/Jump if Condition is Met) branches",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=200003,event=0x8",
	.desc = "Page walk completed due to a demand load to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads (including SW prefetches) whose address translations missed in all TLB levels and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=200003,event=0x8",
	.desc = "Page walk completed due to a demand load to a 2M or 4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads (including SW prefetches) whose address translations missed in all TLB levels and were mapped to 2M or 4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_1gb",
	.event = "umask=0x8,period=200003,event=0x8",
	.desc = "Page walk completed due to a demand load to a 1GB page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads (including SW prefetches) whose address translations missed in all TLB levels and were mapped to 1GB pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_pending",
	.event = "umask=0x10,period=200003,event=0x8",
	.desc = "Page walks outstanding due to a demand load every cycle",
	.topic = "virtual memory",
	.long_desc = "Counts once per cycle for each page walk occurring due to a load (demand data loads or SW prefetches). Includes cycles spent traversing the Extended Page Table (EPT). Average cycles per walk can be calculated by dividing by the number of walks",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 2M or 4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 2M or 4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_1gb",
	.event = "umask=0x8,period=2000003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 1GB page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 1GB pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_pending",
	.event = "umask=0x10,period=200003,event=0x49",
	.desc = "Page walks outstanding due to a demand data store every cycle",
	.topic = "virtual memory",
	.long_desc = "Counts once per cycle for each page walk occurring due to a demand data store. Includes cycles spent traversing the Extended Page Table (EPT). Average cycles per walk can be calculated by dividing by the number of walks",
},
{
	.name = "ept.walk_pending",
	.event = "umask=0x10,period=200003,event=0x4f",
	.desc = "Page walks outstanding due to walking the EPT every cycle",
	.topic = "virtual memory",
	.long_desc = "Counts once per cycle for each page walk only while traversing the Extended Page Table (EPT), and does not count during the rest of the translation.  The EPT is used for translating Guest-Physical Addresses to Physical Addresses for Virtual Machine Monitors (VMMs).  Average cycles per walk can be calculated by dividing the count by number of walks",
},
{
	.name = "itlb.miss",
	.event = "umask=0x4,period=200003,event=0x81",
	.desc = "ITLB misses",
	.topic = "virtual memory",
	.long_desc = "Counts the number of times the machine was unable to find a translation in the Instruction Translation Lookaside Buffer (ITLB) for a linear address of an instruction fetch.  It counts when new translation are filled into the ITLB.  The event is speculative in nature, but will not count translations (page walks) that are begun and not finished, or translations that are finished but not filled into the ITLB",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x85",
	.desc = "Page walk completed due to an instruction fetch in a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to instruction fetches whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x85",
	.desc = "Page walk completed due to an instruction fetch in a 2M or 4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to instruction fetches whose address translations missed in the TLB and were mapped to 2M or 4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "itlb_misses.walk_completed_1gb",
	.event = "umask=0x8,period=2000003,event=0x85",
	.desc = "Page walk completed due to an instruction fetch in a 1GB page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to instruction fetches whose address translations missed in the TLB and were mapped to 1GB pages.  The page walks can end with or without a page fault",
},
{
	.name = "itlb_misses.walk_pending",
	.event = "umask=0x10,period=200003,event=0x85",
	.desc = "Page walks outstanding due to an instruction fetch every cycle",
	.topic = "virtual memory",
	.long_desc = "Counts once per cycle for each page walk occurring due to an instruction fetch. Includes cycles spent traversing the Extended Page Table (EPT). Average cycles per walk can be calculated by dividing by the number of walks",
},
{
	.name = "tlb_flushes.stlb_any",
	.event = "umask=0x20,period=20003,event=0xbd",
	.desc = "STLB flushes",
	.topic = "virtual memory",
	.long_desc = "Counts STLB flushes.  The TLBs are flushed on instructions like INVLPG and MOV to CR3",
},
{
	.name = "mem_uops_retired.dtlb_miss_loads",
	.event = "umask=0x11,period=200003,event=0xd0",
	.desc = "Load uops retired that missed the DTLB (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "virtual memory",
	.long_desc = "Counts load uops retired that caused a DTLB miss  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.dtlb_miss_stores",
	.event = "umask=0x12,period=200003,event=0xd0",
	.desc = "Store uops retired that missed the DTLB (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "virtual memory",
	.long_desc = "Counts store uops retired that caused a DTLB miss  Supports address when precise (Must be precise)",
},
{
	.name = "mem_uops_retired.dtlb_miss",
	.event = "umask=0x13,period=200003,event=0xd0",
	.desc = "Memory uops retired that missed the DTLB (Precise event capable)  Supports address when precise (Must be precise)",
	.topic = "virtual memory",
	.long_desc = "Counts uops retired that had a DTLB miss on load, store or either.  Note that when two distinct memory operations to the same page miss the DTLB, only one of them will be recorded as a DTLB miss  Supports address when precise (Must be precise)",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_haswell[] = {
{
	.name = "l2_rqsts.demand_data_rd_miss",
	.event = "umask=0x21,period=200003,event=0x24",
	.desc = "Demand Data Read miss L2, no rejects  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Demand data read requests that missed L2, no rejects  Spec update: HSD78",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x22,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of store RFO requests that miss the L2 cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x24,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
	.long_desc = "Number of instruction fetches that missed the L2 cache",
},
{
	.name = "l2_rqsts.all_demand_miss",
	.event = "umask=0x27,period=200003,event=0x24",
	.desc = "Demand requests that miss L2 cache  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Demand requests that miss L2 cache  Spec update: HSD78",
},
{
	.name = "l2_rqsts.l2_pf_miss",
	.event = "umask=0x30,period=200003,event=0x24",
	.desc = "L2 prefetch requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests that missed L2",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0x3f,period=200003,event=0x24",
	.desc = "All requests that miss L2 cache  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "All requests that missed L2  Spec update: HSD78",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0xc1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests, initiated by load instructions, that hit L2 cache  Spec update: HSD78",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0xc2,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of store RFO requests that hit the L2 cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0xc4,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
	.long_desc = "Number of instruction fetches that hit the L2 cache",
},
{
	.name = "l2_rqsts.l2_pf_hit",
	.event = "umask=0xd0,period=200003,event=0x24",
	.desc = "L2 prefetch requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests that hit L2",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0xe1,period=200003,event=0x24",
	.desc = "Demand Data Read requests  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Counts any demand and L1 HW prefetch data load requests to L2  Spec update: HSD78",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xe2,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 store RFO requests",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0xe4,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "Counts all L2 code requests",
},
{
	.name = "l2_rqsts.all_demand_references",
	.event = "umask=0xe7,period=200003,event=0x24",
	.desc = "Demand requests to L2 cache  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Demand requests to L2 cache  Spec update: HSD78",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xf8,period=200003,event=0x24",
	.desc = "Requests from L2 hardware prefetchers",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200003,event=0x24",
	.desc = "All L2 requests  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "All requests to L2 cache  Spec update: HSD78",
},
{
	.name = "l2_demand_rqsts.wb_hit",
	.event = "umask=0x50,period=200003,event=0x27",
	.desc = "Not rejected writebacks that hit L2 cache",
	.topic = "cache",
	.long_desc = "Not rejected writebacks that hit L2 cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed L3",
	.topic = "cache",
	.long_desc = "This event counts each cache miss condition for references to the last level cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to L3",
	.topic = "cache",
	.long_desc = "This event counts requests originating from the core that reference a cache line in the last level cache",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss oustandings duration in cycles",
	.topic = "cache",
	.long_desc = "Increments the number of outstanding L1D misses every cycle. Set Cmask = 1 and Edge =1 to count occurrences",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.request_fb_full",
	.event = "umask=0x2,period=2000003,event=0x48",
	.desc = "Number of times a request needed a FB entry but there was no entry available for it. That is the FB unavailability was dominant reason for blocking the request. A request includes cacheable/uncacheable demands that is load, store or SW prefetch. HWP are e",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
	.topic = "cache",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "This event counts when new data lines are brought into the L1 Data cache, which cause other lines to be evicted from the cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue  Spec update: HSD78, HSD62, HSD61",
	.topic = "cache",
	.long_desc = "Offcore outstanding demand data read transactions in SQ to uncore. Set Cmask=1 to count cycles  Spec update: HSD78, HSD62, HSD61",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: HSD78, HSD62, HSD61",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,period=2000003,cmask=6,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue  Spec update: HSD78, HSD62, HSD61",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: HSD62, HSD61",
	.topic = "cache",
	.long_desc = "Offcore outstanding Demand code Read transactions in SQ to uncore. Set Cmask=1 to count cycles  Spec update: HSD62, HSD61",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding RFO store transactions in SuperQueue (SQ), queue to uncore  Spec update: HSD62, HSD61",
	.topic = "cache",
	.long_desc = "Offcore outstanding RFO store transactions in SQ to uncore. Set Cmask=1 to count cycles  Spec update: HSD62, HSD61",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: HSD62, HSD61",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore  Spec update: HSD62, HSD61",
	.topic = "cache",
	.long_desc = "Offcore outstanding cacheable data read transactions in SQ to uncore. Set Cmask=1 to count cycles  Spec update: HSD62, HSD61",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: HSD62, HSD61",
	.topic = "cache",
},
{
	.name = "lock_cycles.cache_lock_duration",
	.event = "umask=0x2,period=2000003,event=0x63",
	.desc = "Cycles when L1D is locked",
	.topic = "cache",
	.long_desc = "Cycles in which the L1D is locked",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Demand data read requests sent to uncore  Spec update: HSD78",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "Demand code read requests sent to uncore",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "Demand RFO read requests sent to uncore, including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "Data read requests sent to uncore (demand and prefetch)",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Offcore requests buffer cannot take more entries for this thread core",
	.topic = "cache",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load uops that miss the STLB. (precise Event)  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store uops that miss the STLB. (precise Event)  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=100003,event=0xd0",
	.desc = "Retired load uops with locked access. (precise Event)  Spec update: HSD76, HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load uops that split across a cacheline boundary. (precise Event)  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts load uops retired which had memory addresses spilt across 2 cache lines. A line split is across 64B cache-lines which may include a page split (4K). This is a precise event  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary. (precise Event)  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts store uops retired which had memory addresses spilt across 2 cache lines. A line split is across 64B cache-lines which may include a page split (4K). This is a precise event  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load uops. (precise Event)  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store uops. (precise Event)  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts all store uops retired. This is a precise event  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load uops with L1 cache hits as data sources  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load uops with L2 cache hits as data sources  Spec update: HSD76, HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l3_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Miss in last-level (L3) cache. Excludes Unknown data-source  Spec update: HSD74, HSD29, HSD25, HSM26, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops in which data sources were data hits in the L3 cache without snoops required. This does not include hardware prefetches. This is a precise event  Spec update: HSD74, HSD29, HSD25, HSM26, HSM30.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load uops misses in L1 cache as data sources  Spec update: HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops in which data sources missed in the L1 cache. This does not include hardware prefetches. This is a precise event  Spec update: HSM30.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Retired load uops with L2 cache misses as data sources  Spec update: HSD29, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l3_miss",
	.event = "umask=0x20,period=100003,event=0xd1",
	.desc = "Miss in last-level (L3) cache. Excludes Unknown data-source  Spec update: HSD74, HSD29, HSD25, HSM26, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.hit_lfb",
	.event = "umask=0x40,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources were load uops missed L1 but hit FB due to preceding miss to the same cache line with data not ready  Spec update: HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 hit and cross-core snoop missed in on-pkg core cache  Spec update: HSD29, HSD25, HSM26, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 and cross-core snoop hits in on-pkg core cache  Spec update: HSD29, HSD25, HSM26, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops that hit in the L3 cache, but required a cross-core snoop which resulted in a HIT in an on-pkg core cache. This does not include hardware prefetches. This is a precise event  Spec update: HSD29, HSD25, HSM26, HSM30.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were HitM responses from shared L3  Spec update: HSD29, HSD25, HSM26, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops that hit in the L3 cache, but required a cross-core snoop which resulted in a HITM (hit modified) in an on-pkg core cache. This does not include hardware prefetches. This is a precise event  Spec update: HSD29, HSD25, HSM26, HSM30.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load uops which data sources were hits in L3 without snoops required  Spec update: HSD74, HSD29, HSD25, HSM26, HSM30.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_miss_retired.local_dram",
	.event = "umask=0x1,period=100003,event=0xd3",
	.desc = "(null)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops where the data came from local DRAM. This does not include hardware prefetches. This is a precise event  Spec update: HSD74, HSD29, HSD25, HSM30.  Supports address when precise",
},
{
	.name = "l2_trans.demand_data_rd",
	.event = "umask=0x1,period=200003,event=0xf0",
	.desc = "Demand Data Read requests that access L2 cache",
	.topic = "cache",
	.long_desc = "Demand data read requests that access L2 cache",
},
{
	.name = "l2_trans.rfo",
	.event = "umask=0x2,period=200003,event=0xf0",
	.desc = "RFO requests that access L2 cache",
	.topic = "cache",
	.long_desc = "RFO requests that access L2 cache",
},
{
	.name = "l2_trans.code_rd",
	.event = "umask=0x4,period=200003,event=0xf0",
	.desc = "L2 cache accesses when fetching instructions",
	.topic = "cache",
	.long_desc = "L2 cache accesses when fetching instructions",
},
{
	.name = "l2_trans.all_pf",
	.event = "umask=0x8,period=200003,event=0xf0",
	.desc = "L2 or L3 HW prefetches that access L2 cache",
	.topic = "cache",
	.long_desc = "Any MLC or L3 HW prefetch accessing L2, including rejects",
},
{
	.name = "l2_trans.l1d_wb",
	.event = "umask=0x10,period=200003,event=0xf0",
	.desc = "L1D writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "L1D writebacks that access L2 cache",
},
{
	.name = "l2_trans.l2_fill",
	.event = "umask=0x20,period=200003,event=0xf0",
	.desc = "L2 fill requests that access L2 cache",
	.topic = "cache",
	.long_desc = "L2 fill requests that access L2 cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "L2 writebacks that access L2 cache",
},
{
	.name = "l2_trans.all_requests",
	.event = "umask=0x80,period=200003,event=0xf0",
	.desc = "Transactions accessing L2 pipe",
	.topic = "cache",
	.long_desc = "Transactions accessing L2 pipe",
},
{
	.name = "l2_lines_in.i",
	.event = "umask=0x1,period=100003,event=0xf1",
	.desc = "L2 cache lines in I state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in I state filling L2",
},
{
	.name = "l2_lines_in.s",
	.event = "umask=0x2,period=100003,event=0xf1",
	.desc = "L2 cache lines in S state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in S state filling L2",
},
{
	.name = "l2_lines_in.e",
	.event = "umask=0x4,period=100003,event=0xf1",
	.desc = "L2 cache lines in E state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in E state filling L2",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x7,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines brought into the L2 cache.  Lines are filled into the L2 cache when there was an L2 miss",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x5,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by demand",
	.topic = "cache",
	.long_desc = "Clean L2 cache lines evicted by demand",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x6,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by demand",
	.topic = "cache",
	.long_desc = "Dirty L2 cache lines evicted by demand",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Split locks in SQ",
	.topic = "cache",
},
{
	.name = "offcore_response.all_requests.l3_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C8FFF",
	.desc = "Counts all requests hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all requests hit in the L3",
},
{
	.name = "offcore_response.all_reads.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C07F7",
	.desc = "hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.all_reads.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C07F7",
	.desc = "hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_code_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0244",
	.desc = "Counts all demand & prefetch code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0122",
	.desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.all_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0122",
	.desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0091",
	.desc = "Counts all demand & prefetch data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0091",
	.desc = "Counts all demand & prefetch data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads hit in the L3",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads hit in the L3",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs hit in the L3",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads hit in the L3",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads hit in the L3",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs hit in the L3",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads hit in the L3",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads hit in the L3",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0004",
	.desc = "Counts all demand code reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand code reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0004",
	.desc = "Counts all demand code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0002",
	.desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0002",
	.desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0001",
	.desc = "Counts demand data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts demand data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0001",
	.desc = "Counts demand data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts demand data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "other_assists.avx_to_sse",
	.event = "umask=0x8,period=100003,event=0xc1",
	.desc = "Number of transitions from AVX-256 to legacy SSE when penalty applicable  Spec update: HSD56, HSM57 (Precise event)",
	.topic = "floating point",
	.long_desc = " Spec update: HSD56, HSM57 (Precise event)",
},
{
	.name = "other_assists.sse_to_avx",
	.event = "umask=0x10,period=100003,event=0xc1",
	.desc = "Number of transitions from legacy SSE to AVX-256 when penalty applicable  Spec update: HSD56, HSM57 (Precise event)",
	.topic = "floating point",
	.long_desc = " Spec update: HSD56, HSM57 (Precise event)",
},
{
	.name = "avx_insts.all",
	.event = "umask=0x7,period=2000003,event=0xc6",
	.desc = "Approximate counts of AVX & AVX2 256-bit instructions, including non-arithmetic instructions, loads, and stores.  May count non-AVX instructions that employ 256-bit operations, including (but not necessarily limited to) rep string instructions that use 256-bit loads and stores for optimized performance, XSAVE* and XRSTOR*, and operations that transition the x87 FPU data registers between x87 and MMX",
	.topic = "floating point",
	.long_desc = "Note that a whole rep string only counts AVX_INST.ALL once",
},
{
	.name = "fp_assist.x87_output",
	.event = "umask=0x2,period=100003,event=0xca",
	.desc = "output - Numeric Overflow, Numeric Underflow, Inexact Result (Precise event)",
	.topic = "floating point",
	.long_desc = "(Precise event)",
},
{
	.name = "fp_assist.x87_input",
	.event = "umask=0x4,period=100003,event=0xca",
	.desc = "input - Invalid Operation, Denormal Operand, SNaN Operand (Precise event)",
	.topic = "floating point",
	.long_desc = "(Precise event)",
},
{
	.name = "fp_assist.simd_output",
	.event = "umask=0x8,period=100003,event=0xca",
	.desc = "SSE* FP micro-code assist when output value is invalid (Precise event)",
	.topic = "floating point",
	.long_desc = "(Precise event)",
},
{
	.name = "fp_assist.simd_input",
	.event = "umask=0x10,period=100003,event=0xca",
	.desc = "Any input SSE* FP Assist (Precise event)",
	.topic = "floating point",
	.long_desc = "(Precise event)",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,period=100003,cmask=1,event=0xca",
	.desc = "Counts any FP_ASSIST umask was incrementing (Precise event)",
	.topic = "floating point",
	.long_desc = "(Precise event)",
},
{
	.name = "idq.empty",
	.event = "umask=0x2,period=2000003,event=0x79",
	.desc = "Instruction Decode Queue (IDQ) empty cycles  Spec update: HSD135",
	.topic = "frontend",
	.long_desc = "Counts cycles the IDQ is empty  Spec update: HSD135",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ from MITE path. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Increment each cycle. # of uops delivered to IDQ from DSB path. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_uops",
	.event = "umask=0x10,period=2000003,event=0x79",
	.desc = "Uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ when MS_busy by DSB. Set Cmask = 1 to count cycles. Add Edge=1 to count # of delivery",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_occur",
	.event = "umask=0x10,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts cycles DSB is delivered four uops. Set Cmask = 4",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts cycles DSB is delivered at least one uops. Set Cmask = 1",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ when MS_busy by MITE. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts cycles MITE is delivered four uops. Set Cmask = 4",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts cycles MITE is delivered at least one uop. Set Cmask = 1",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts uops delivered by the Front-end with the assistance of the microcode sequencer.  Microcode assists are used for complex instructions or scenarios that can't be handled by the standard decoder.  Using other instructions, if possible, will usually improve performance",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which the microcode sequencer assisted the Front-end in delivering uops.  Microcode assists are used for complex instructions or scenarios that can't be handled by the standard decoder.  Using other instructions, if possible, will usually improve performance",
},
{
	.name = "idq.ms_switches",
	.event = "umask=0x30,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
},
{
	.name = "idq.mite_all_uops",
	.event = "umask=0x3c,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Number of uops delivered to IDQ from any path",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=2000003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
	.topic = "frontend",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Misses. Includes Uncacheable accesses",
	.topic = "frontend",
	.long_desc = "This event counts Instruction Cache (ICACHE) misses",
},
{
	.name = "icache.ifetch_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction-cache miss",
	.topic = "frontend",
},
{
	.name = "icache.ifdata_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction-cache miss",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled  Spec update: HSD135",
	.topic = "frontend",
	.long_desc = "This event count the number of undelivered (unallocated) uops from the Front-end to the Resource Allocation Table (RAT) while the Back-end of the processor is not stalled. The Front-end can allocate up to 4 uops per cycle so this event can increment 0-4 times per cycle depending on the number of unallocated uops. This event is counted on a per-core basis  Spec update: HSD135",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=4,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled  Spec update: HSD135",
	.topic = "frontend",
	.long_desc = "This event counts the number cycles during which the Front-end allocated exactly zero uops to the Resource Allocation Table (RAT) while the Back-end of the processor is not stalled.  This event is counted on a per-core basis  Spec update: HSD135",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=3,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled  Spec update: HSD135",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=2,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end  Spec update: HSD135",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end  Spec update: HSD135",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE  Spec update: HSD135",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "hsw metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "hsw metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "hsw metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "hsw metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "hsw metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "hsw metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "hsw metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "hsw metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "hsw metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Instruction per taken branch",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_taken",
	.metric_name = "IpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Branch instructions per taken branch",
	.topic = "hsw metrics",
	.metric_expr = "br_inst_retired.all_branches / br_inst_retired.near_taken",
	.metric_name = "BpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "hsw metrics",
	.metric_expr = "min( 1 , idq.mite_uops / ( (uops_retired.retire_slots / inst_retired.any) * 16 * ( icache.hit + icache.misses ) / 4.0 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "hsw metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops ) )",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "hsw metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "hsw metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "hsw metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "hsw metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Instructions per Load (lower number means loads are more frequent)",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_loads",
	.metric_name = "IpL",
	.metric_group = "Instruction_Type;L1_Bound",
},
{
	.desc = "Instructions per Store",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_stores",
	.metric_name = "IpS",
	.metric_group = "Instruction_Type;Store_Bound",
},
{
	.desc = "Instructions per Branch",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.all_branches",
	.metric_name = "IpB",
	.metric_group = "Branches;Instruction_Type;Port_5;Port_6",
},
{
	.desc = "Instruction per (near) call",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_call",
	.metric_name = "IpCall",
	.metric_group = "Branches",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "hsw metrics",
	.metric_expr = "( uops_executed.core / 2 / (( cpu@uops_executed.core\\,cmask\\=1@ / 2 ) if #smt_on else cpu@uops_executed.core\\,cmask\\=1@) ) if #smt_on else uops_executed.core / (( cpu@uops_executed.core\\,cmask\\=1@ / 2 ) if #smt_on else cpu@uops_executed.core\\,cmask\\=1@)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Number of Instructions per non-speculative Branch Misprediction (JEClear)",
	.topic = "hsw metrics",
	.metric_expr = "inst_retired.any / br_misp_retired.all_branches",
	.metric_name = "IpMispredict",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "hsw metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads (in core cycles)",
	.topic = "hsw metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_uops_retired.l1_miss + mem_load_uops_retired.hit_lfb )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least one such miss. Per-thread)",
	.topic = "hsw metrics",
	.metric_expr = "l1d_pend_miss.pending / l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "hsw metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration ) / cycles",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "hsw metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration ) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "Page_Walks_Utilization_SMT",
	.metric_group = "TLB_SMT",
},
{
	.desc = "Average data fill bandwidth to the L1 data cache [GB / sec]",
	.topic = "hsw metrics",
	.metric_expr = "64 * l1d.replacement / 1000000000 / duration_time",
	.metric_name = "L1D_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average data fill bandwidth to the L2 cache [GB / sec]",
	.topic = "hsw metrics",
	.metric_expr = "64 * l2_lines_in.all / 1000000000 / duration_time",
	.metric_name = "L2_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "hsw metrics",
	.metric_expr = "64 * longest_lat_cache.miss / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "L1 cache true misses per kilo instruction for retired demand loads",
	.topic = "hsw metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l1_miss / inst_retired.any",
	.metric_name = "L1MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache true misses per kilo instruction for retired demand loads",
	.topic = "hsw metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache misses per kilo instruction for all request types (including speculative)",
	.topic = "hsw metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache hits per kilo instruction for all request types (including speculative)",
	.topic = "hsw metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2HPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L3 cache true misses per kilo instruction for retired demand loads",
	.topic = "hsw metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l3_miss / inst_retired.any",
	.metric_name = "L3MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "Average CPU Utilization",
	.topic = "hsw metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "hsw metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "hsw metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "hsw metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "hsw metrics",
	.metric_expr = "64 * ( arb@event\\=0x81\\,umask\\=0x1@ + arb@event\\=0x84\\,umask\\=0x1@ ) / 1000000 / duration_time / 1000",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "C3 residency percent per core",
	.topic = "hsw metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "hsw metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "hsw metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "hsw metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "hsw metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "hsw metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "hsw metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "misalign_mem_ref.loads",
	.event = "umask=0x1,period=2000003,event=0x5",
	.desc = "Speculative cache line split load uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "Speculative cache-line split load uops dispatched to L1D",
},
{
	.name = "misalign_mem_ref.stores",
	.event = "umask=0x2,period=2000003,event=0x5",
	.desc = "Speculative cache line split STA uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "Speculative cache-line split store-address uops dispatched to L1D",
},
{
	.name = "tx_mem.abort_conflict",
	.event = "umask=0x1,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data conflict on a transactionally accessed address",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_capacity_write",
	.event = "umask=0x2,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data capacity limitation for transactional writes",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_store_to_elided_lock",
	.event = "umask=0x4,period=2000003,event=0x54",
	.desc = "Number of times a HLE transactional region aborted due to a non XRELEASE prefixed instruction writing to an elided lock in the elision buffer",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_not_empty",
	.event = "umask=0x8,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to NoAllocatedElisionBuffer being non-zero",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_mismatch",
	.event = "umask=0x10,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to XRELEASE lock not satisfying the address and value requirements in the elision buffer",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_unsupported_alignment",
	.event = "umask=0x20,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to an unsupported read alignment from the elision buffer",
	.topic = "memory",
},
{
	.name = "tx_mem.hle_elision_buffer_full",
	.event = "umask=0x40,period=2000003,event=0x54",
	.desc = "Number of times HLE lock could not be elided due to ElisionBufferAvailable being zero",
	.topic = "memory",
},
{
	.name = "tx_exec.misc1",
	.event = "umask=0x1,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions that may cause a transactional abort was executed. Since this is the count of execution, it may not always cause a transactional abort",
	.topic = "memory",
},
{
	.name = "tx_exec.misc2",
	.event = "umask=0x2,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions (e.g., vzeroupper) that may cause a transactional abort was executed inside a transactional region",
	.topic = "memory",
},
{
	.name = "tx_exec.misc3",
	.event = "umask=0x4,period=2000003,event=0x5d",
	.desc = "Counts the number of times an instruction execution caused the transactional nest count supported to be exceeded",
	.topic = "memory",
},
{
	.name = "tx_exec.misc4",
	.event = "umask=0x8,period=2000003,event=0x5d",
	.desc = "Counts the number of times a XBEGIN instruction was executed inside an HLE transactional region",
	.topic = "memory",
},
{
	.name = "tx_exec.misc5",
	.event = "umask=0x10,period=2000003,event=0x5d",
	.desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
	.topic = "memory",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts",
	.topic = "memory",
	.long_desc = "This event counts the number of memory ordering machine clears detected. Memory ordering machine clears can result from memory address aliasing or snoops from another hardware thread or core to data inflight in the pipeline.  Machine clears can have a significant performance impact if they are happening frequently",
},
{
	.name = "hle_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution started",
	.topic = "memory",
},
{
	.name = "hle_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution successfully committed",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to uncommon conditions",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to incompatible memory type  Spec update: HSD65",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to none of the previous 4 categories (e.g. interrupts)",
	.topic = "memory",
	.long_desc = "Number of times an HLE execution aborted due to none of the previous 4 categories (e.g. interrupts)",
},
{
	.name = "rtm_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution started",
	.topic = "memory",
},
{
	.name = "rtm_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution successfully committed",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
},
{
	.name = "rtm_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to incompatible memory type  Spec update: HSD65",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Randomly selected loads with latency value being above 4  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Randomly selected loads with latency value being above 8  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Randomly selected loads with latency value being above 16  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x20",
	.desc = "Randomly selected loads with latency value being above 32  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Randomly selected loads with latency value being above 64  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Randomly selected loads with latency value being above 128  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Randomly selected loads with latency value being above 256  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Randomly selected loads with latency value being above 512  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "offcore_response.all_requests.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC08FFF",
	.desc = "Counts all requests miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all requests miss in the L3",
},
{
	.name = "offcore_response.all_reads.l3_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01004007F7",
	.desc = "miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_reads.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC007F7",
	.desc = "miss in the L3",
	.topic = "memory",
	.long_desc = "miss in the L3",
},
{
	.name = "offcore_response.all_code_rd.l3_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400244",
	.desc = "Counts all demand & prefetch code reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch code reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_code_rd.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00244",
	.desc = "Counts all demand & prefetch code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch code reads miss in the L3",
},
{
	.name = "offcore_response.all_rfo.l3_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400122",
	.desc = "Counts all demand & prefetch RFOs miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_rfo.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00122",
	.desc = "Counts all demand & prefetch RFOs miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs miss in the L3",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400091",
	.desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00091",
	.desc = "Counts all demand & prefetch data reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss in the L3",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads miss in the L3",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs miss in the L3",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads miss in the L3",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads miss in the L3",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs miss in the L3",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00010",
	.desc = "Counts prefetch (that bring data to L2) data reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads miss in the L3",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400004",
	.desc = "Counts all demand code reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand code reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00004",
	.desc = "Counts all demand code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand code reads miss in the L3",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400002",
	.desc = "Counts all demand data writes (RFOs) miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00002",
	.desc = "Counts all demand data writes (RFOs) miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) miss in the L3",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400001",
	.desc = "Counts demand data reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts demand data reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC00001",
	.desc = "Counts demand data reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts demand data reads miss in the L3",
},
{
	.name = "cpl_cycles.ring0",
	.event = "umask=0x1,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when the thread is in ring 0",
	.topic = "other",
	.long_desc = "Unhalted core cycles when the thread is in ring 0",
},
{
	.name = "cpl_cycles.ring0_trans",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0x5c",
	.desc = "Number of intervals between processor halts while thread is in ring 0",
	.topic = "other",
},
{
	.name = "cpl_cycles.ring123",
	.event = "umask=0x2,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when thread is in rings 1, 2, or 3",
	.topic = "other",
	.long_desc = "Unhalted core cycles when the thread is not in ring 0",
},
{
	.name = "lock_cycles.split_lock_uc_lock_duration",
	.event = "umask=0x1,period=2000003,event=0x63",
	.desc = "Cycles when L1 and L2 are locked due to UC or split lock",
	.topic = "other",
	.long_desc = "Cycles in which the L1D and L2 are locked, due to a UC lock or split lock",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution  Spec update: HSD140, HSD143",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, this event counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers. INST_RETIRED.ANY is counted by a designated fixed counter, leaving the programmable counters available for other events. Faulting executions of GETSEC/VM entry/VM Exit/MWait will not count as retired instructions  Spec update: HSD140, HSD143",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "loads blocked by overlapping with store buffer that cannot be forwarded",
	.topic = "pipeline",
	.long_desc = "This event counts loads that followed a store to the same address, where the data could not be forwarded inside the pipeline from the store to the load.  The most common reason why store forwarding would be blocked is when a load's address range overlaps with a preceding smaller uncompleted store. The penalty for blocked store forwarding is that the load must wait for the store to write its value to the cache before it can be issued",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
	.long_desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare on address",
	.topic = "pipeline",
	.long_desc = "Aliasing occurs when a load is issued after a store and their memory addresses are offset by 4K.  This event counts the number of loads that aliased with a preceding store, resulting in an extended address check in the pipeline which can have a performance impact",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x3,period=2000003,cmask=1,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for this thread (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of cycles spent waiting for a recovery after an event such as a processor nuke, JEClear, assist, hle/rtm abort etc",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x3,any=1,period=2000003,cmask=1,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of uops issued by the Front-end of the pipeline to the Back-end. This event is counted at the allocation stage and will count both retired and non-retired uops",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for all threads",
	.topic = "pipeline",
},
{
	.name = "uops_issued.flags_merge",
	.event = "umask=0x10,period=2000003,event=0xe",
	.desc = "Number of flags-merge uops being allocated. Such uops considered perf sensitive; added by GSR u-arch",
	.topic = "pipeline",
	.long_desc = "Number of flags-merge uops allocated. Such uops add delay",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
	.long_desc = "Number of slow LEA or similar uops allocated. Such uop has 3 sources (for example, 2 sources + immediate) regardless of whether it is a result of LEA instruction or not",
},
{
	.name = "uops_issued.single_mul",
	.event = "umask=0x40,period=2000003,event=0xe",
	.desc = "Number of Multiply packed/scalar single precision uops allocated",
	.topic = "pipeline",
	.long_desc = "Number of multiply packed/scalar single precision uops allocated",
},
{
	.name = "arith.divider_uops",
	.event = "umask=0x2,period=2000003,event=0x14",
	.desc = "Any uop executed by the Divider. (This includes all divide uops, sqrt, ...)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Increments at the frequency of XCLK (100 MHz) when not halted",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the thread is unhalted. (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "Non-SW-prefetch load dispatches that hit fill buffer allocated for S/W prefetch",
},
{
	.name = "load_hit_pre.hw_pf",
	.event = "umask=0x2,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for hardware prefetch",
	.topic = "pipeline",
	.long_desc = "Non-SW-prefetch load dispatches that hit fill buffer allocated for H/W prefetch",
},
{
	.name = "move_elimination.int_eliminated",
	.event = "umask=0x1,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
	.long_desc = "Number of integer move elimination candidate uops that were eliminated",
},
{
	.name = "move_elimination.simd_eliminated",
	.event = "umask=0x2,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
	.long_desc = "Number of SIMD move elimination candidate uops that were eliminated",
},
{
	.name = "move_elimination.int_not_eliminated",
	.event = "umask=0x4,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
	.long_desc = "Number of integer move elimination candidate uops that were not eliminated",
},
{
	.name = "move_elimination.simd_not_eliminated",
	.event = "umask=0x8,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
	.long_desc = "Number of SIMD move elimination candidate uops that were not eliminated",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles when the Reservation Station ( RS ) is empty for the thread. The RS is a structure that buffers allocated micro-ops from the Front-end. If there are many cycles when the RS is empty, it may represent an underflow of instructions delivered from the Front-end",
},
{
	.name = "rs_events.empty_end",
	.event = "inv=1,umask=0x1,edge=1,period=200003,cmask=1,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
	.long_desc = "This event counts cycles where the decoder is stalled on an instruction with a length changing prefix (LCP)",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000003,event=0x87",
	.desc = "Stall cycles because IQ is full",
	.topic = "pipeline",
	.long_desc = "Stall cycles due to IQ is full",
},
{
	.name = "br_inst_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x88",
	.desc = "Not taken macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_direct_jump",
	.event = "umask=0x82,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_near_return",
	.event = "umask=0x88,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches with return mnemonic",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x88",
	.desc = "Taken speculative and retired direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x88",
	.desc = "Speculative and retired macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_jmp",
	.event = "umask=0xc2,period=200003,event=0x88",
	.desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x88",
	.desc = "Speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_indirect_near_return",
	.event = "umask=0xc8,period=200003,event=0x88",
	.desc = "Speculative and retired indirect return branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x88",
	.desc = "Speculative and retired direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x88",
	.desc = "Speculative and retired  branches",
	.topic = "pipeline",
	.long_desc = "Counts all near executed branches (not necessarily retired)",
},
{
	.name = "br_misp_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x89",
	.desc = "Not taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_return_near",
	.event = "umask=0x88,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x89",
	.desc = "Mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Counts all near executed branches (not necessarily retired)",
},
{
	.name = "uops_executed_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 0 in this thread",
},
{
	.name = "uops_executed_port.port_0_core",
	.event = "umask=0x1,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 0",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 1 in this thread",
},
{
	.name = "uops_executed_port.port_1_core",
	.event = "umask=0x2,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 1",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 2 in this thread",
},
{
	.name = "uops_executed_port.port_2_core",
	.event = "umask=0x4,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 2",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 3 in this thread",
},
{
	.name = "uops_executed_port.port_3_core",
	.event = "umask=0x8,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 3",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 4 in this thread",
},
{
	.name = "uops_executed_port.port_4_core",
	.event = "umask=0x10,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 4",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 5 in this thread",
},
{
	.name = "uops_executed_port.port_5_core",
	.event = "umask=0x20,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 5",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 6 in this thread",
},
{
	.name = "uops_executed_port.port_6_core",
	.event = "umask=0x40,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 6",
},
{
	.name = "uops_dispatched_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 7 in this thread",
},
{
	.name = "uops_executed_port.port_7_core",
	.event = "umask=0x80,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 7",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles  Spec update: HSD135",
	.topic = "pipeline",
	.long_desc = "Cycles allocation is stalled due to resource related reason  Spec update: HSD135",
},
{
	.name = "resource_stalls.rs",
	.event = "umask=0x4,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no eligible RS entry available",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which no instructions were allocated because no Store Buffers (SB) were available",
},
{
	.name = "resource_stalls.rob",
	.event = "umask=0x10,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to re-order buffer full",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l2_pending",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Cycles with pending L2 cache miss loads  Spec update: HSD78",
	.topic = "pipeline",
	.long_desc = "Cycles with pending L2 miss loads. Set Cmask=2 to count cycle  Spec update: HSD78",
},
{
	.name = "cycle_activity.cycles_ldm_pending",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Cycles with pending memory loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending memory loads. Set Cmask=2 to count cycle",
},
{
	.name = "cycle_activity.cycles_no_execute",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "This event increments by 1 for every cycle where there was no execute for this thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which no instructions were executed in the execution stage of the pipeline",
},
{
	.name = "cycle_activity.stalls_l2_pending",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Execution stalls due to L2 cache misses",
	.topic = "pipeline",
	.long_desc = "Number of loads missed L2",
},
{
	.name = "cycle_activity.stalls_ldm_pending",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Execution stalls due to memory subsystem",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which no instructions were executed in the execution stage of the pipeline and there were memory instructions pending (waiting for data)",
},
{
	.name = "cycle_activity.cycles_l1d_pending",
	.event = "umask=0x8,period=2000003,cmask=8,event=0xa3",
	.desc = "Cycles with pending L1 cache miss loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending L1 data cache miss loads. Set Cmask=8 to count cycle",
},
{
	.name = "cycle_activity.stalls_l1d_pending",
	.event = "umask=0xc,period=2000003,cmask=12,event=0xa3",
	.desc = "Execution stalls due to L1 data cache misses",
	.topic = "pipeline",
	.long_desc = "Execution stalls due to L1 data cache miss loads. Set Cmask=0CH",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
	.long_desc = "Number of uops delivered by the LSD",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
	.long_desc = "This events counts the cycles where at least one uop was executed. It is counted per thread  Spec update: HSD144, HSD30, HSM31",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
	.long_desc = "This events counts the cycles where at least two uop were executed. It is counted per thread  Spec update: HSD144, HSD30, HSM31",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
	.long_desc = "This events counts the cycles where at least three uop were executed. It is counted per thread  Spec update: HSD144, HSD30, HSM31",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
	.long_desc = "Counts total number of uops to be executed per-core each cycle  Spec update: HSD30, HSM31",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter   - architectural event  Spec update: HSD11, HSD140",
	.topic = "pipeline",
	.long_desc = "Number of instructions at retirement  Spec update: HSD11, HSD140",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: HSD140 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: HSD140 (Must be precise)",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000003,event=0xc0",
	.desc = "FP operations retired. X87 FP operations that have no exceptions: Counts also flows that have several X87 or flows that use X87 uops in the exception handling (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts FP operations retired. For X87 FP operations that have no exceptions counting also includes flows that have several X87, or flows that use X87 uops in the exception handling (Precise event)",
},
{
	.name = "other_assists.any_wb_assist",
	.event = "umask=0x40,period=100003,event=0xc1",
	.desc = "(null)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Actually retired uops  Supports address when precise (Precise event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles no executable uops retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "(Precise event)",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=10,event=0xc2",
	.desc = "Number of cycles using always true condition applied to  PEBS uops retired event (Precise event)",
	.topic = "pipeline",
	.long_desc = "(Precise event)",
},
{
	.name = "uops_retired.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles no executable uops retired on core (Precise event)",
	.topic = "pipeline",
	.long_desc = "(Precise event)",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used (Precise event)",
	.topic = "pipeline",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=2000003,event=0xc3",
	.desc = "Cycles there was a Nuke. Account for both thread-specific and All Thread Nukes",
	.topic = "pipeline",
},
{
	.name = "machine_clears.count",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "This event is incremented when self-modifying code (SMC) is detected, which causes a machine clear.  Machine clears can have a significant performance impact if they are happening frequently",
},
{
	.name = "machine_clears.maskmov",
	.event = "umask=0x20,period=100003,event=0xc3",
	.desc = "This event counts the number of executed Intel AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Branch instructions at retirement",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100003,event=0xc4",
	.desc = "Direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call_r3",
	.event = "umask=0x2,period=100003,event=0xc4",
	.desc = "Direct and indirect macro near call instructions retired (captured in ring 3) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100003,event=0xc4",
	.desc = "Return instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Counts all not taken macro branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "(Precise event)",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100003,event=0xc4",
	.desc = "Counts the number of far branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "(Precise event)",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Mispredicted branch instructions at retirement",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This event counts all mispredicted branch instructions retired. This is a precise event (Must be precise)",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "number of near branch instructions retired that were mispredicted and taken (Precise event)",
	.topic = "pipeline",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Count cases of saving new LBR",
	.topic = "pipeline",
	.long_desc = "Count cases of saving new LBR records by hardware",
},
{
	.name = "baclears.any",
	.event = "umask=0x1f,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
	.long_desc = "Number of front end re-steers due to BPU misprediction",
},
{
	.name = "unc_cbo_xsnp_response.miss_external",
	.event = "umask=0x21,event=0x22",
	.desc = "Unit: uncore_cbox An external snoop misses in some processor core",
	.topic = "uncore",
	.long_desc = "An external snoop misses in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.miss_xcore",
	.event = "umask=0x41,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop initiated by this Cbox due to processor core memory request which misses in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop initiated by this Cbox due to processor core memory request which misses in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.miss_eviction",
	.event = "umask=0x81,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop resulted from L3 Eviction which misses in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop resulted from L3 Eviction which misses in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hit_external",
	.event = "umask=0x24,event=0x22",
	.desc = "Unit: uncore_cbox An external snoop hits a non-modified line in some processor core",
	.topic = "uncore",
	.long_desc = "An external snoop hits a non-modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hit_xcore",
	.event = "umask=0x44,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop initiated by this Cbox due to processor core memory request which hits a non-modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop initiated by this Cbox due to processor core memory request which hits a non-modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hit_eviction",
	.event = "umask=0x84,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop resulted from L3 Eviction which hits a non-modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop resulted from L3 Eviction which hits a non-modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hitm_external",
	.event = "umask=0x28,event=0x22",
	.desc = "Unit: uncore_cbox An external snoop hits a modified line in some processor core",
	.topic = "uncore",
	.long_desc = "An external snoop hits a modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hitm_xcore",
	.event = "umask=0x48,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop initiated by this Cbox due to processor core memory request which hits a modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop initiated by this Cbox due to processor core memory request which hits a modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hitm_eviction",
	.event = "umask=0x88,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop resulted from L3 Eviction which hits a modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop resulted from L3 Eviction which hits a modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_m",
	.event = "umask=0x11,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_m",
	.event = "umask=0x21,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.extsnp_m",
	.event = "umask=0x41,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup external snoop request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup external snoop request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_m",
	.event = "umask=0x81,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_i",
	.event = "umask=0x18,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_i",
	.event = "umask=0x28,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.extsnp_i",
	.event = "umask=0x48,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup external snoop request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup external snoop request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_i",
	.event = "umask=0x88,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_mesi",
	.event = "umask=0x1f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in any MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in any MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_mesi",
	.event = "umask=0x2f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.extsnp_mesi",
	.event = "umask=0x4f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup external snoop request that access cache and found line in MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup external snoop request that access cache and found line in MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_mesi",
	.event = "umask=0x8f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_es",
	.event = "umask=0x86,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.extsnp_es",
	.event = "umask=0x46,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup external snoop request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup external snoop request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_es",
	.event = "umask=0x16,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_es",
	.event = "umask=0x26,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_arb_trk_occupancy.all",
	.event = "umask=0x01,event=0x80",
	.desc = "Unit: uncore_arb Each cycle count number of all Core outgoing valid entries. Such entry is defined as valid from it's allocation till first of IDI0 or DRS0 messages is sent out. Accounts for Coherent and non-coherent traffic",
	.topic = "uncore",
	.long_desc = "Each cycle count number of all Core outgoing valid entries. Such entry is defined as valid from it's allocation till first of IDI0 or DRS0 messages is sent out. Accounts for Coherent and non-coherent traffic",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.all",
	.event = "umask=0x01,event=0x81",
	.desc = "Unit: uncore_arb Total number of Core outgoing entries allocated. Accounts for Coherent and non-coherent traffic",
	.topic = "uncore",
	.long_desc = "Total number of Core outgoing entries allocated. Accounts for Coherent and non-coherent traffic",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.writes",
	.event = "umask=0x20,event=0x81",
	.desc = "Unit: uncore_arb Number of Writes allocated - any write transactions: full/partials writes and evictions",
	.topic = "uncore",
	.long_desc = "Number of Writes allocated - any write transactions: full/partials writes and evictions",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_coh_trk_occupancy.all",
	.event = "umask=0x01,event=0x83",
	.desc = "Unit: uncore_arb Each cycle count number of valid entries in Coherency Tracker queue from allocation till deallocation. Aperture requests (snoops) appear as NC decoded internally and become coherent (snoop L3, access memory)",
	.topic = "uncore",
	.long_desc = "Each cycle count number of valid entries in Coherency Tracker queue from allocation till deallocation. Aperture requests (snoops) appear as NC decoded internally and become coherent (snoop L3, access memory)",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_coh_trk_requests.all",
	.event = "umask=0x01,event=0x84",
	.desc = "Unit: uncore_arb Number of entries allocated. Account for Any type: e.g. Snoop, Core aperture, etc",
	.topic = "uncore",
	.long_desc = "Number of entries allocated. Account for Any type: e.g. Snoop, Core aperture, etc",
	.pmu = "uncore_arb",
},
{
	.name = "unc_clock.socket",
	.event = "umask=0x01,event=0",
	.desc = "Unit: uncore_ncu This 48-bit fixed counter counts the UCLK cycles",
	.topic = "uncore",
	.long_desc = "This 48-bit fixed counter counts the UCLK cycles",
	.pmu = "uncore_ncu",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Misses in all TLB levels that cause a page walk of any page size",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (4K)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to demand load misses that caused 4K page walks in any TLB levels",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to demand load misses that caused 2M/4M page walks in any TLB levels",
},
{
	.name = "dtlb_load_misses.walk_completed_1g",
	.event = "umask=0x8,period=2000003,event=0x8",
	.desc = "Load miss in all TLB levels causes a page walk that completes. (1G)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes of any page size",
	.topic = "virtual memory",
	.long_desc = "Completed page walks in any TLB of any page size due to demand load misses",
},
{
	.name = "dtlb_load_misses.walk_duration",
	.event = "umask=0x10,period=2000003,event=0x8",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles when the  page miss handler (PMH) is servicing page walks caused by DTLB load misses",
},
{
	.name = "dtlb_load_misses.stlb_hit_4k",
	.event = "umask=0x20,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
	.long_desc = "This event counts load operations from a 4K page that miss the first DTLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_load_misses.stlb_hit_2m",
	.event = "umask=0x40,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
	.long_desc = "This event counts load operations from a 2M page that miss the first DTLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x60,period=2000003,event=0x8",
	.desc = "Load operations that miss the first DTLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "Number of cache load STLB hits. No page walk",
},
{
	.name = "dtlb_load_misses.pde_cache_miss",
	.event = "umask=0x80,period=100003,event=0x8",
	.desc = "DTLB demand load misses with low part of linear-to-physical address translation missed",
	.topic = "virtual memory",
	.long_desc = "DTLB demand load misses with low part of linear-to-physical address translation missed",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Miss in all TLB levels causes a page walk of any page size (4K/2M/4M/1G)",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (4K)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to store misses in one or more TLB levels of 4K page structure",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to store misses in one or more TLB levels of 2M/4M page structure",
},
{
	.name = "dtlb_store_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks. (1G)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to store miss in any TLB levels of any page size (4K/2M/4M/1G)",
},
{
	.name = "dtlb_store_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x49",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles when the  page miss handler (PMH) is servicing page walks caused by DTLB store misses",
},
{
	.name = "dtlb_store_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
	.long_desc = "This event counts store operations from a 4K page that miss the first DTLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_store_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
	.long_desc = "This event counts store operations from a 2M page that miss the first DTLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x49",
	.desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_store_misses.pde_cache_miss",
	.event = "umask=0x80,period=100003,event=0x49",
	.desc = "DTLB store misses with low part of linear-to-physical address translation missed",
	.topic = "virtual memory",
	.long_desc = "DTLB store misses with low part of linear-to-physical address translation missed",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Cycle count for an Extended Page table walk",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Misses in ITLB that causes a page walk of any page size",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (4K)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to misses in ITLB 4K page entries",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to misses in ITLB 2M/4M page entries",
},
{
	.name = "itlb_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x85",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (1G)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x85",
	.desc = "Misses in all ITLB levels that cause completed page walks",
	.topic = "virtual memory",
	.long_desc = "Completed page walks in ITLB of any page size",
},
{
	.name = "itlb_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles when the  page miss handler (PMH) is servicing page walks caused by ITLB misses",
},
{
	.name = "itlb_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x85",
	.desc = "Core misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
	.long_desc = "ITLB misses that hit STLB (4K)",
},
{
	.name = "itlb_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x85",
	.desc = "Code misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
	.long_desc = "ITLB misses that hit STLB (2M)",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x85",
	.desc = "Operations that miss the first ITLB level but hit the second and do not cause any page walks",
	.topic = "virtual memory",
	.long_desc = "ITLB misses that hit STLB. No page walk",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100003,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "Counts the number of ITLB flushes, includes 4k/2M/4M pages",
},
{
	.name = "page_walker_loads.dtlb_l1",
	.event = "umask=0x11,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L1+FB",
	.topic = "virtual memory",
	.long_desc = "Number of DTLB page walker loads that hit in the L1+FB",
},
{
	.name = "page_walker_loads.dtlb_l2",
	.event = "umask=0x12,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L2",
	.topic = "virtual memory",
	.long_desc = "Number of DTLB page walker loads that hit in the L2",
},
{
	.name = "page_walker_loads.dtlb_l3",
	.event = "umask=0x14,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L3 + XSNP  Spec update: HSD25",
	.topic = "virtual memory",
	.long_desc = "Number of DTLB page walker loads that hit in the L3  Spec update: HSD25",
},
{
	.name = "page_walker_loads.dtlb_memory",
	.event = "umask=0x18,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in Memory  Spec update: HSD25",
	.topic = "virtual memory",
	.long_desc = "Number of DTLB page walker loads from memory  Spec update: HSD25",
},
{
	.name = "page_walker_loads.itlb_l1",
	.event = "umask=0x21,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L1+FB",
	.topic = "virtual memory",
	.long_desc = "Number of ITLB page walker loads that hit in the L1+FB",
},
{
	.name = "page_walker_loads.itlb_l2",
	.event = "umask=0x22,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L2",
	.topic = "virtual memory",
	.long_desc = "Number of ITLB page walker loads that hit in the L2",
},
{
	.name = "page_walker_loads.itlb_l3",
	.event = "umask=0x24,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L3 + XSNP  Spec update: HSD25",
	.topic = "virtual memory",
	.long_desc = "Number of ITLB page walker loads that hit in the L3  Spec update: HSD25",
},
{
	.name = "page_walker_loads.itlb_memory",
	.event = "umask=0x28,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in Memory  Spec update: HSD25",
	.topic = "virtual memory",
	.long_desc = "Number of ITLB page walker loads from memory  Spec update: HSD25",
},
{
	.name = "page_walker_loads.ept_dtlb_l1",
	.event = "umask=0x41,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the DTLB that hit in the L1 and FB",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_dtlb_l2",
	.event = "umask=0x42,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the DTLB that hit in the L2",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_dtlb_l3",
	.event = "umask=0x44,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the DTLB that hit in the L3",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_dtlb_memory",
	.event = "umask=0x48,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the DTLB that hit in memory",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_itlb_l1",
	.event = "umask=0x81,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the ITLB that hit in the L1 and FB",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_itlb_l2",
	.event = "umask=0x82,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the ITLB that hit in the L2",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_itlb_l3",
	.event = "umask=0x84,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the ITLB that hit in the L2",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_itlb_memory",
	.event = "umask=0x88,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the ITLB that hit in memory",
	.topic = "virtual memory",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100003,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100003,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "Count number of STLB flush attempts",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_haswellx[] = {
{
	.name = "l2_rqsts.demand_data_rd_miss",
	.event = "umask=0x21,period=200003,event=0x24",
	.desc = "Demand Data Read miss L2, no rejects  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Demand data read requests that missed L2, no rejects  Spec update: HSD78",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x22,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of store RFO requests that miss the L2 cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x24,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
	.long_desc = "Number of instruction fetches that missed the L2 cache",
},
{
	.name = "l2_rqsts.all_demand_miss",
	.event = "umask=0x27,period=200003,event=0x24",
	.desc = "Demand requests that miss L2 cache  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Demand requests that miss L2 cache  Spec update: HSD78",
},
{
	.name = "l2_rqsts.l2_pf_miss",
	.event = "umask=0x30,period=200003,event=0x24",
	.desc = "L2 prefetch requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests that missed L2",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0x3f,period=200003,event=0x24",
	.desc = "All requests that miss L2 cache  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "All requests that missed L2  Spec update: HSD78",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0xc1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests, initiated by load instructions, that hit L2 cache  Spec update: HSD78",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0xc2,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of store RFO requests that hit the L2 cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0xc4,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
	.long_desc = "Number of instruction fetches that hit the L2 cache",
},
{
	.name = "l2_rqsts.l2_pf_hit",
	.event = "umask=0xd0,period=200003,event=0x24",
	.desc = "L2 prefetch requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests that hit L2",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0xe1,period=200003,event=0x24",
	.desc = "Demand Data Read requests  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Counts any demand and L1 HW prefetch data load requests to L2  Spec update: HSD78",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xe2,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 store RFO requests",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0xe4,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "Counts all L2 code requests",
},
{
	.name = "l2_rqsts.all_demand_references",
	.event = "umask=0xe7,period=200003,event=0x24",
	.desc = "Demand requests to L2 cache  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Demand requests to L2 cache  Spec update: HSD78",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xf8,period=200003,event=0x24",
	.desc = "Requests from L2 hardware prefetchers",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200003,event=0x24",
	.desc = "All L2 requests  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "All requests to L2 cache  Spec update: HSD78",
},
{
	.name = "l2_demand_rqsts.wb_hit",
	.event = "umask=0x50,period=200003,event=0x27",
	.desc = "Not rejected writebacks that hit L2 cache",
	.topic = "cache",
	.long_desc = "Not rejected writebacks that hit L2 cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed L3",
	.topic = "cache",
	.long_desc = "This event counts each cache miss condition for references to the last level cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to L3",
	.topic = "cache",
	.long_desc = "This event counts requests originating from the core that reference a cache line in the last level cache",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss oustandings duration in cycles",
	.topic = "cache",
	.long_desc = "Increments the number of outstanding L1D misses every cycle. Set Cmask = 1 and Edge =1 to count occurrences",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.request_fb_full",
	.event = "umask=0x2,period=2000003,event=0x48",
	.desc = "Number of times a request needed a FB entry but there was no entry available for it. That is the FB unavailability was dominant reason for blocking the request. A request includes cacheable/uncacheable demands that is load, store or SW prefetch. HWP are e",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
	.topic = "cache",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "This event counts when new data lines are brought into the L1 Data cache, which cause other lines to be evicted from the cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue  Spec update: HSD78, HSD62, HSD61",
	.topic = "cache",
	.long_desc = "Offcore outstanding demand data read transactions in SQ to uncore. Set Cmask=1 to count cycles  Spec update: HSD78, HSD62, HSD61",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: HSD78, HSD62, HSD61",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,cmask=6,period=2000003,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue  Spec update: HSD78, HSD62, HSD61",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: HSD62, HSD61",
	.topic = "cache",
	.long_desc = "Offcore outstanding Demand code Read transactions in SQ to uncore. Set Cmask=1 to count cycles  Spec update: HSD62, HSD61",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding RFO store transactions in SuperQueue (SQ), queue to uncore  Spec update: HSD62, HSD61",
	.topic = "cache",
	.long_desc = "Offcore outstanding RFO store transactions in SQ to uncore. Set Cmask=1 to count cycles  Spec update: HSD62, HSD61",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: HSD62, HSD61",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore  Spec update: HSD62, HSD61",
	.topic = "cache",
	.long_desc = "Offcore outstanding cacheable data read transactions in SQ to uncore. Set Cmask=1 to count cycles  Spec update: HSD62, HSD61",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: HSD62, HSD61",
	.topic = "cache",
},
{
	.name = "lock_cycles.cache_lock_duration",
	.event = "umask=0x2,period=2000003,event=0x63",
	.desc = "Cycles when L1D is locked",
	.topic = "cache",
	.long_desc = "Cycles in which the L1D is locked",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore  Spec update: HSD78",
	.topic = "cache",
	.long_desc = "Demand data read requests sent to uncore  Spec update: HSD78",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "Demand code read requests sent to uncore",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "Demand RFO read requests sent to uncore, including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "Data read requests sent to uncore (demand and prefetch)",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Offcore requests buffer cannot take more entries for this thread core",
	.topic = "cache",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load uops that miss the STLB  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store uops that miss the STLB  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=100003,event=0xd0",
	.desc = "Retired load uops with locked access  Supports address when precise.  Spec update: HSD76, HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load uops that split across a cacheline boundary  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load uops  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store uops  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load uops with L1 cache hits as data sources  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load uops with L2 cache hits as data sources  Supports address when precise.  Spec update: HSD76, HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l3_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load uops which data sources were data hits in L3 without snoops required  Supports address when precise.  Spec update: HSD74, HSD29, HSD25, HSM26, HSM30 (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load uops with L3 cache hits as data sources  Supports address when precise.  Spec update: HSD74, HSD29, HSD25, HSM26, HSM30 (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load uops misses in L1 cache as data sources  Supports address when precise.  Spec update: HSM30 (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load uops missed L1 cache as data sources  Supports address when precise.  Spec update: HSM30 (Precise event)",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Miss in mid-level (L2) cache. Excludes Unknown data-source  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load uops missed L2. Unknown data source excluded  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
},
{
	.name = "mem_load_uops_retired.l3_miss",
	.event = "umask=0x20,period=100003,event=0xd1",
	.desc = "Miss in last-level (L3) cache. Excludes Unknown data-source  Supports address when precise.  Spec update: HSD74, HSD29, HSD25, HSM26, HSM30 (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load uops missed L3. Excludes unknown data source   Supports address when precise.  Spec update: HSD74, HSD29, HSD25, HSM26, HSM30 (Precise event)",
},
{
	.name = "mem_load_uops_retired.hit_lfb",
	.event = "umask=0x40,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources were load uops missed L1 but hit FB due to preceding miss to the same cache line with data not ready  Supports address when precise.  Spec update: HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 hit and cross-core snoop missed in on-pkg core cache  Supports address when precise.  Spec update: HSD29, HSD25, HSM26, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise.  Spec update: HSD29, HSD25, HSM26, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were HitM responses from shared L3  Supports address when precise.  Spec update: HSD29, HSD25, HSM26, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load uops which data sources were hits in L3 without snoops required  Supports address when precise.  Spec update: HSD74, HSD29, HSD25, HSM26, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_miss_retired.local_dram",
	.event = "umask=0x1,period=100003,event=0xd3",
	.desc = "Data from local DRAM either Snoop not needed or Snoop Miss (RspI)  Supports address when precise.  Spec update: HSD74, HSD29, HSD25, HSM30 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops where the data came from local DRAM. This does not include hardware prefetches  Supports address when precise.  Spec update: HSD74, HSD29, HSD25, HSM30 (Precise event)",
},
{
	.name = "mem_load_uops_l3_miss_retired.remote_dram",
	.event = "umask=0x4,period=100003,event=0xd3",
	.desc = "Retired load uop whose Data Source was: remote DRAM either Snoop not needed or Snoop Miss (RspI)  Supports address when precise.  Spec update: HSD29, HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_miss_retired.remote_hitm",
	.event = "umask=0x10,period=100003,event=0xd3",
	.desc = "Retired load uop whose Data Source was: Remote cache HITM  Supports address when precise.  Spec update: HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_l3_miss_retired.remote_fwd",
	.event = "umask=0x20,period=100003,event=0xd3",
	.desc = "Retired load uop whose Data Source was: forwarded from remote cache  Supports address when precise.  Spec update: HSM30 (Precise event)",
	.topic = "cache",
},
{
	.name = "l2_trans.demand_data_rd",
	.event = "umask=0x1,period=200003,event=0xf0",
	.desc = "Demand Data Read requests that access L2 cache",
	.topic = "cache",
	.long_desc = "Demand data read requests that access L2 cache",
},
{
	.name = "l2_trans.rfo",
	.event = "umask=0x2,period=200003,event=0xf0",
	.desc = "RFO requests that access L2 cache",
	.topic = "cache",
	.long_desc = "RFO requests that access L2 cache",
},
{
	.name = "l2_trans.code_rd",
	.event = "umask=0x4,period=200003,event=0xf0",
	.desc = "L2 cache accesses when fetching instructions",
	.topic = "cache",
	.long_desc = "L2 cache accesses when fetching instructions",
},
{
	.name = "l2_trans.all_pf",
	.event = "umask=0x8,period=200003,event=0xf0",
	.desc = "L2 or L3 HW prefetches that access L2 cache",
	.topic = "cache",
	.long_desc = "Any MLC or L3 HW prefetch accessing L2, including rejects",
},
{
	.name = "l2_trans.l1d_wb",
	.event = "umask=0x10,period=200003,event=0xf0",
	.desc = "L1D writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "L1D writebacks that access L2 cache",
},
{
	.name = "l2_trans.l2_fill",
	.event = "umask=0x20,period=200003,event=0xf0",
	.desc = "L2 fill requests that access L2 cache",
	.topic = "cache",
	.long_desc = "L2 fill requests that access L2 cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "L2 writebacks that access L2 cache",
},
{
	.name = "l2_trans.all_requests",
	.event = "umask=0x80,period=200003,event=0xf0",
	.desc = "Transactions accessing L2 pipe",
	.topic = "cache",
	.long_desc = "Transactions accessing L2 pipe",
},
{
	.name = "l2_lines_in.i",
	.event = "umask=0x1,period=100003,event=0xf1",
	.desc = "L2 cache lines in I state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in I state filling L2",
},
{
	.name = "l2_lines_in.s",
	.event = "umask=0x2,period=100003,event=0xf1",
	.desc = "L2 cache lines in S state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in S state filling L2",
},
{
	.name = "l2_lines_in.e",
	.event = "umask=0x4,period=100003,event=0xf1",
	.desc = "L2 cache lines in E state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in E state filling L2",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x7,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines brought into the L2 cache.  Lines are filled into the L2 cache when there was an L2 miss",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x5,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by demand",
	.topic = "cache",
	.long_desc = "Clean L2 cache lines evicted by demand",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x6,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by demand",
	.topic = "cache",
	.long_desc = "Dirty L2 cache lines evicted by demand",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Split locks in SQ",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0001",
	.desc = "Counts demand data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts demand data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0001",
	.desc = "Counts demand data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts demand data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0002",
	.desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0002",
	.desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0004",
	.desc = "Counts all demand code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0004",
	.desc = "Counts all demand code reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand code reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads hit in the L3",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads hit in the L3",
},
{
	.name = "offcore_response.pf_l2_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs hit in the L3",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads hit in the L3",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads hit in the L3",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs hit in the L3",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads hit in the L3",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads hit in the L3",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0091",
	.desc = "Counts all demand & prefetch data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0091",
	.desc = "Counts all demand & prefetch data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.all_rfo.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0122",
	.desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0122",
	.desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.all_code_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0244",
	.desc = "Counts all demand & prefetch code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch code reads hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_reads.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C07F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) hit in the L3 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
},
{
	.name = "offcore_response.all_reads.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C07F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) hit in the L3 and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
},
{
	.name = "offcore_response.all_requests.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C8FFF",
	.desc = "Counts all requests hit in the L3",
	.topic = "cache",
	.long_desc = "Counts all requests hit in the L3",
},
{
	.name = "other_assists.avx_to_sse",
	.event = "umask=0x8,period=100003,event=0xc1",
	.desc = "Number of transitions from AVX-256 to legacy SSE when penalty applicable  Spec update: HSD56, HSM57",
	.topic = "floating point",
},
{
	.name = "other_assists.sse_to_avx",
	.event = "umask=0x10,period=100003,event=0xc1",
	.desc = "Number of transitions from SSE to AVX-256 when penalty applicable  Spec update: HSD56, HSM57",
	.topic = "floating point",
},
{
	.name = "avx_insts.all",
	.event = "umask=0x7,period=2000003,event=0xc6",
	.desc = "Approximate counts of AVX & AVX2 256-bit instructions, including non-arithmetic instructions, loads, and stores.  May count non-AVX instructions that employ 256-bit operations, including (but not necessarily limited to) rep string instructions that use 256-bit loads and stores for optimized performance, XSAVE* and XRSTOR*, and operations that transition the x87 FPU data registers between x87 and MMX",
	.topic = "floating point",
	.long_desc = "Note that a whole rep string only counts AVX_INST.ALL once",
},
{
	.name = "fp_assist.x87_output",
	.event = "umask=0x2,period=100003,event=0xca",
	.desc = "Number of X87 assists due to output value",
	.topic = "floating point",
	.long_desc = "Number of X87 FP assists due to output values",
},
{
	.name = "fp_assist.x87_input",
	.event = "umask=0x4,period=100003,event=0xca",
	.desc = "Number of X87 assists due to input value",
	.topic = "floating point",
	.long_desc = "Number of X87 FP assists due to input values",
},
{
	.name = "fp_assist.simd_output",
	.event = "umask=0x8,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to Output values",
	.topic = "floating point",
	.long_desc = "Number of SIMD FP assists due to output values",
},
{
	.name = "fp_assist.simd_input",
	.event = "umask=0x10,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to input values",
	.topic = "floating point",
	.long_desc = "Number of SIMD FP assists due to input values",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,cmask=1,period=100003,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
	.long_desc = "Cycles with any input/output SSE* or FP assists",
},
{
	.name = "idq.empty",
	.event = "umask=0x2,period=2000003,event=0x79",
	.desc = "Instruction Decode Queue (IDQ) empty cycles  Spec update: HSD135",
	.topic = "frontend",
	.long_desc = "Counts cycles the IDQ is empty  Spec update: HSD135",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ from MITE path. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Increment each cycle. # of uops delivered to IDQ from DSB path. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_uops",
	.event = "umask=0x10,period=2000003,event=0x79",
	.desc = "Uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ when MS_busy by DSB. Set Cmask = 1 to count cycles. Add Edge=1 to count # of delivery",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_occur",
	.event = "edge=1,umask=0x10,cmask=1,period=2000003,event=0x79",
	.desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts cycles DSB is delivered four uops. Set Cmask = 4",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts cycles DSB is delivered at least one uops. Set Cmask = 1",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ when MS_busy by MITE. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts cycles MITE is delivered four uops. Set Cmask = 4",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts cycles MITE is delivered at least one uop. Set Cmask = 1",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts uops delivered by the Front-end with the assistance of the microcode sequencer.  Microcode assists are used for complex instructions or scenarios that can't be handled by the standard decoder.  Using other instructions, if possible, will usually improve performance",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which the microcode sequencer assisted the Front-end in delivering uops.  Microcode assists are used for complex instructions or scenarios that can't be handled by the standard decoder.  Using other instructions, if possible, will usually improve performance",
},
{
	.name = "idq.ms_switches",
	.event = "edge=1,umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
},
{
	.name = "idq.mite_all_uops",
	.event = "umask=0x3c,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Number of uops delivered to IDQ from any path",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=2000003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
	.topic = "frontend",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Misses. Includes Uncacheable accesses",
	.topic = "frontend",
	.long_desc = "This event counts Instruction Cache (ICACHE) misses",
},
{
	.name = "icache.ifetch_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction-cache miss",
	.topic = "frontend",
},
{
	.name = "icache.ifdata_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction-cache miss",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled  Spec update: HSD135",
	.topic = "frontend",
	.long_desc = "This event count the number of undelivered (unallocated) uops from the Front-end to the Resource Allocation Table (RAT) while the Back-end of the processor is not stalled. The Front-end can allocate up to 4 uops per cycle so this event can increment 0-4 times per cycle depending on the number of unallocated uops. This event is counted on a per-core basis  Spec update: HSD135",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,cmask=4,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled  Spec update: HSD135",
	.topic = "frontend",
	.long_desc = "This event counts the number cycles during which the Front-end allocated exactly zero uops to the Resource Allocation Table (RAT) while the Back-end of the processor is not stalled.  This event is counted on a per-core basis  Spec update: HSD135",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,cmask=3,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled  Spec update: HSD135",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,cmask=2,period=2000003,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end  Spec update: HSD135",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end  Spec update: HSD135",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE  Spec update: HSD135",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "hsx metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "hsx metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "hsx metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "hsx metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "hsx metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "hsx metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "hsx metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "hsx metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "hsx metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Instruction per taken branch",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_taken",
	.metric_name = "IpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Branch instructions per taken branch",
	.topic = "hsx metrics",
	.metric_expr = "br_inst_retired.all_branches / br_inst_retired.near_taken",
	.metric_name = "BpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "hsx metrics",
	.metric_expr = "min( 1 , idq.mite_uops / ( (uops_retired.retire_slots / inst_retired.any) * 16 * ( icache.hit + icache.misses ) / 4.0 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "hsx metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops ) )",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "hsx metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "hsx metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "hsx metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "hsx metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Instructions per Load (lower number means loads are more frequent)",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_loads",
	.metric_name = "IpL",
	.metric_group = "Instruction_Type;L1_Bound",
},
{
	.desc = "Instructions per Store",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_stores",
	.metric_name = "IpS",
	.metric_group = "Instruction_Type;Store_Bound",
},
{
	.desc = "Instructions per Branch",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.all_branches",
	.metric_name = "IpB",
	.metric_group = "Branches;Instruction_Type;Port_5;Port_6",
},
{
	.desc = "Instruction per (near) call",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_call",
	.metric_name = "IpCall",
	.metric_group = "Branches",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "hsx metrics",
	.metric_expr = "( uops_executed.core / 2 / (( cpu@uops_executed.core\\,cmask\\=1@ / 2 ) if #smt_on else cpu@uops_executed.core\\,cmask\\=1@) ) if #smt_on else uops_executed.core / (( cpu@uops_executed.core\\,cmask\\=1@ / 2 ) if #smt_on else cpu@uops_executed.core\\,cmask\\=1@)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Number of Instructions per non-speculative Branch Misprediction (JEClear)",
	.topic = "hsx metrics",
	.metric_expr = "inst_retired.any / br_misp_retired.all_branches",
	.metric_name = "IpMispredict",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "hsx metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads (in core cycles)",
	.topic = "hsx metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_uops_retired.l1_miss + mem_load_uops_retired.hit_lfb )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least one such miss. Per-thread)",
	.topic = "hsx metrics",
	.metric_expr = "l1d_pend_miss.pending / l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "hsx metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration ) / cycles",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "hsx metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration ) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "Page_Walks_Utilization_SMT",
	.metric_group = "TLB_SMT",
},
{
	.desc = "Average data fill bandwidth to the L1 data cache [GB / sec]",
	.topic = "hsx metrics",
	.metric_expr = "64 * l1d.replacement / 1000000000 / duration_time",
	.metric_name = "L1D_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average data fill bandwidth to the L2 cache [GB / sec]",
	.topic = "hsx metrics",
	.metric_expr = "64 * l2_lines_in.all / 1000000000 / duration_time",
	.metric_name = "L2_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "hsx metrics",
	.metric_expr = "64 * longest_lat_cache.miss / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "L1 cache true misses per kilo instruction for retired demand loads",
	.topic = "hsx metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l1_miss / inst_retired.any",
	.metric_name = "L1MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache true misses per kilo instruction for retired demand loads",
	.topic = "hsx metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache misses per kilo instruction for all request types (including speculative)",
	.topic = "hsx metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache hits per kilo instruction for all request types (including speculative)",
	.topic = "hsx metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2HPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L3 cache true misses per kilo instruction for retired demand loads",
	.topic = "hsx metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l3_miss / inst_retired.any",
	.metric_name = "L3MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "Average CPU Utilization",
	.topic = "hsx metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "hsx metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "hsx metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "hsx metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "hsx metrics",
	.metric_expr = "( 64 * ( uncore_imc@cas_count_read@ + uncore_imc@cas_count_write@ ) / 1000000000 ) / duration_time",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average latency of data read request to external memory (in nanoseconds). Accounts for demand loads and L1/L2 prefetches",
	.topic = "hsx metrics",
	.metric_expr = "1000000000 * ( cbox@event\\=0x36\\,umask\\=0x3\\,filter_opc\\=0x182@ / cbox@event\\=0x35\\,umask\\=0x3\\,filter_opc\\=0x182@ ) / ( cbox_0@event\\=0x0@ / duration_time )",
	.metric_name = "DRAM_Read_Latency",
	.metric_group = "Memory_Lat",
},
{
	.desc = "Average number of parallel data read requests to external memory. Accounts for demand loads and L1/L2 prefetches",
	.topic = "hsx metrics",
	.metric_expr = "cbox@event\\=0x36\\,umask\\=0x3\\,filter_opc\\=0x182@ / cbox@event\\=0x36\\,umask\\=0x3\\,filter_opc\\=0x182\\,thresh\\=1@",
	.metric_name = "DRAM_Parallel_Reads",
	.metric_group = "Memory_BW",
},
{
	.desc = "Socket actual clocks when any core is active on that socket",
	.topic = "hsx metrics",
	.metric_expr = "cbox_0@event\\=0x0@",
	.metric_name = "Socket_CLKS",
	.metric_group = "",
},
{
	.desc = "C3 residency percent per core",
	.topic = "hsx metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "hsx metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "hsx metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "hsx metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "hsx metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "hsx metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "hsx metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "misalign_mem_ref.loads",
	.event = "umask=0x1,period=2000003,event=0x5",
	.desc = "Speculative cache line split load uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "Speculative cache-line split load uops dispatched to L1D",
},
{
	.name = "misalign_mem_ref.stores",
	.event = "umask=0x2,period=2000003,event=0x5",
	.desc = "Speculative cache line split STA uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "Speculative cache-line split store-address uops dispatched to L1D",
},
{
	.name = "tx_mem.abort_conflict",
	.event = "umask=0x1,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data conflict on a transactionally accessed address",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_capacity_write",
	.event = "umask=0x2,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data capacity limitation for transactional writes",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_store_to_elided_lock",
	.event = "umask=0x4,period=2000003,event=0x54",
	.desc = "Number of times a HLE transactional region aborted due to a non XRELEASE prefixed instruction writing to an elided lock in the elision buffer",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_not_empty",
	.event = "umask=0x8,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to NoAllocatedElisionBuffer being non-zero",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_mismatch",
	.event = "umask=0x10,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to XRELEASE lock not satisfying the address and value requirements in the elision buffer",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_unsupported_alignment",
	.event = "umask=0x20,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to an unsupported read alignment from the elision buffer",
	.topic = "memory",
},
{
	.name = "tx_mem.hle_elision_buffer_full",
	.event = "umask=0x40,period=2000003,event=0x54",
	.desc = "Number of times HLE lock could not be elided due to ElisionBufferAvailable being zero",
	.topic = "memory",
},
{
	.name = "tx_exec.misc1",
	.event = "umask=0x1,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions that may cause a transactional abort was executed. Since this is the count of execution, it may not always cause a transactional abort",
	.topic = "memory",
},
{
	.name = "tx_exec.misc2",
	.event = "umask=0x2,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions (e.g., vzeroupper) that may cause a transactional abort was executed inside a transactional region",
	.topic = "memory",
},
{
	.name = "tx_exec.misc3",
	.event = "umask=0x4,period=2000003,event=0x5d",
	.desc = "Counts the number of times an instruction execution caused the transactional nest count supported to be exceeded",
	.topic = "memory",
},
{
	.name = "tx_exec.misc4",
	.event = "umask=0x8,period=2000003,event=0x5d",
	.desc = "Counts the number of times a XBEGIN instruction was executed inside an HLE transactional region",
	.topic = "memory",
},
{
	.name = "tx_exec.misc5",
	.event = "umask=0x10,period=2000003,event=0x5d",
	.desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
	.topic = "memory",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts",
	.topic = "memory",
	.long_desc = "This event counts the number of memory ordering machine clears detected. Memory ordering machine clears can result from memory address aliasing or snoops from another hardware thread or core to data inflight in the pipeline.  Machine clears can have a significant performance impact if they are happening frequently",
},
{
	.name = "hle_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution started",
	.topic = "memory",
},
{
	.name = "hle_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution successfully committed",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to uncommon conditions",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to incompatible memory type  Spec update: HSD65",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to none of the previous 4 categories (e.g. interrupts)",
	.topic = "memory",
	.long_desc = "Number of times an HLE execution aborted due to none of the previous 4 categories (e.g. interrupts)",
},
{
	.name = "rtm_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution started",
	.topic = "memory",
},
{
	.name = "rtm_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution successfully committed",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
},
{
	.name = "rtm_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to incompatible memory type  Spec update: HSD65",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Randomly selected loads with latency value being above 4  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Randomly selected loads with latency value being above 8  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Randomly selected loads with latency value being above 16  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x20",
	.desc = "Randomly selected loads with latency value being above 32  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Randomly selected loads with latency value being above 64  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Randomly selected loads with latency value being above 128  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Randomly selected loads with latency value being above 256  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Randomly selected loads with latency value being above 512  Spec update: HSD76, HSD25, HSM26 (Must be precise)",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00001",
	.desc = "Counts demand data reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts demand data reads miss in the L3",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0600400001",
	.desc = "Counts demand data reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts demand data reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.demand_rfo.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00002",
	.desc = "Counts all demand data writes (RFOs) miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) miss in the L3",
},
{
	.name = "offcore_response.demand_rfo.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0600400002",
	.desc = "Counts all demand data writes (RFOs) miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.demand_rfo.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00002",
	.desc = "Counts all demand data writes (RFOs) miss the L3 and the modified data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) miss the L3 and the modified data is transferred from remote cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00004",
	.desc = "Counts all demand code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand code reads miss in the L3",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0600400004",
	.desc = "Counts all demand code reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand code reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00010",
	.desc = "Counts prefetch (that bring data to L2) data reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads miss in the L3",
},
{
	.name = "offcore_response.pf_l2_rfo.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs miss in the L3",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads miss in the L3",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads miss in the L3",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs miss in the L3",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads miss in the L3",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00091",
	.desc = "Counts all demand & prefetch data reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss in the L3",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0600400091",
	.desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063F800091",
	.desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from remote dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss the L3 and the data is returned from remote dram",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00091",
	.desc = "Counts all demand & prefetch data reads miss the L3 and the modified data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss the L3 and the modified data is transferred from remote cache",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00091",
	.desc = "Counts all demand & prefetch data reads miss the L3 and clean or shared data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads miss the L3 and clean or shared data is transferred from remote cache",
},
{
	.name = "offcore_response.all_rfo.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00122",
	.desc = "Counts all demand & prefetch RFOs miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs miss in the L3",
},
{
	.name = "offcore_response.all_rfo.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0600400122",
	.desc = "Counts all demand & prefetch RFOs miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC00244",
	.desc = "Counts all demand & prefetch code reads miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch code reads miss in the L3",
},
{
	.name = "offcore_response.all_code_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0600400244",
	.desc = "Counts all demand & prefetch code reads miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch code reads miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_reads.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss in the L3",
},
{
	.name = "offcore_response.all_reads.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x06004007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the data is returned from local dram",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the data is returned from local dram",
},
{
	.name = "offcore_response.all_reads.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063F8007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the data is returned from remote dram",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the data is returned from remote dram",
},
{
	.name = "offcore_response.all_reads.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the modified data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and the modified data is transferred from remote cache",
},
{
	.name = "offcore_response.all_reads.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC007F7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and clean or shared data is transferred from remote cache",
	.topic = "memory",
	.long_desc = "Counts all data/code/rfo reads (demand & prefetch) miss the L3 and clean or shared data is transferred from remote cache",
},
{
	.name = "offcore_response.all_requests.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBFC08FFF",
	.desc = "Counts all requests miss in the L3",
	.topic = "memory",
	.long_desc = "Counts all requests miss in the L3",
},
{
	.name = "cpl_cycles.ring0",
	.event = "umask=0x1,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when the thread is in ring 0",
	.topic = "other",
	.long_desc = "Unhalted core cycles when the thread is in ring 0",
},
{
	.name = "cpl_cycles.ring0_trans",
	.event = "edge=1,umask=0x1,cmask=1,period=100003,event=0x5c",
	.desc = "Number of intervals between processor halts while thread is in ring 0",
	.topic = "other",
},
{
	.name = "cpl_cycles.ring123",
	.event = "umask=0x2,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when thread is in rings 1, 2, or 3",
	.topic = "other",
	.long_desc = "Unhalted core cycles when the thread is not in ring 0",
},
{
	.name = "lock_cycles.split_lock_uc_lock_duration",
	.event = "umask=0x1,period=2000003,event=0x63",
	.desc = "Cycles when L1 and L2 are locked due to UC or split lock",
	.topic = "other",
	.long_desc = "Cycles in which the L1D and L2 are locked, due to a UC lock or split lock",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution  Spec update: HSD140, HSD143",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, this event counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers. INST_RETIRED.ANY is counted by a designated fixed counter, leaving the programmable counters available for other events. Faulting executions of GETSEC/VM entry/VM Exit/MWait will not count as retired instructions  Spec update: HSD140, HSD143",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "loads blocked by overlapping with store buffer that cannot be forwarded",
	.topic = "pipeline",
	.long_desc = "This event counts loads that followed a store to the same address, where the data could not be forwarded inside the pipeline from the store to the load.  The most common reason why store forwarding would be blocked is when a load's address range overlaps with a preceding smaller uncompleted store. The penalty for blocked store forwarding is that the load must wait for the store to write its value to the cache before it can be issued",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
	.long_desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare on address",
	.topic = "pipeline",
	.long_desc = "Aliasing occurs when a load is issued after a store and their memory addresses are offset by 4K.  This event counts the number of loads that aliased with a preceding store, resulting in an extended address check in the pipeline which can have a performance impact",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x3,cmask=1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for this thread (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of cycles spent waiting for a recovery after an event such as a processor nuke, JEClear, assist, hle/rtm abort etc",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x3,any=1,cmask=1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of uops issued by the Front-end of the pipeline to the Back-end. This event is counted at the allocation stage and will count both retired and non-retired uops",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,cmask=1,period=2000003,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for all threads",
	.topic = "pipeline",
},
{
	.name = "uops_issued.flags_merge",
	.event = "umask=0x10,period=2000003,event=0xe",
	.desc = "Number of flags-merge uops being allocated. Such uops considered perf sensitive; added by GSR u-arch",
	.topic = "pipeline",
	.long_desc = "Number of flags-merge uops allocated. Such uops add delay",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
	.long_desc = "Number of slow LEA or similar uops allocated. Such uop has 3 sources (for example, 2 sources + immediate) regardless of whether it is a result of LEA instruction or not",
},
{
	.name = "uops_issued.single_mul",
	.event = "umask=0x40,period=2000003,event=0xe",
	.desc = "Number of Multiply packed/scalar single precision uops allocated",
	.topic = "pipeline",
	.long_desc = "Number of multiply packed/scalar single precision uops allocated",
},
{
	.name = "arith.divider_uops",
	.event = "umask=0x2,period=2000003,event=0x14",
	.desc = "Any uop executed by the Divider. (This includes all divide uops, sqrt, ...)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Increments at the frequency of XCLK (100 MHz) when not halted",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the thread is unhalted. (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "Non-SW-prefetch load dispatches that hit fill buffer allocated for S/W prefetch",
},
{
	.name = "load_hit_pre.hw_pf",
	.event = "umask=0x2,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for hardware prefetch",
	.topic = "pipeline",
	.long_desc = "Non-SW-prefetch load dispatches that hit fill buffer allocated for H/W prefetch",
},
{
	.name = "move_elimination.int_eliminated",
	.event = "umask=0x1,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
	.long_desc = "Number of integer move elimination candidate uops that were eliminated",
},
{
	.name = "move_elimination.simd_eliminated",
	.event = "umask=0x2,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
	.long_desc = "Number of SIMD move elimination candidate uops that were eliminated",
},
{
	.name = "move_elimination.int_not_eliminated",
	.event = "umask=0x4,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
	.long_desc = "Number of integer move elimination candidate uops that were not eliminated",
},
{
	.name = "move_elimination.simd_not_eliminated",
	.event = "umask=0x8,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
	.long_desc = "Number of SIMD move elimination candidate uops that were not eliminated",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles when the Reservation Station ( RS ) is empty for the thread. The RS is a structure that buffers allocated micro-ops from the Front-end. If there are many cycles when the RS is empty, it may represent an underflow of instructions delivered from the Front-end",
},
{
	.name = "rs_events.empty_end",
	.event = "edge=1,inv=1,umask=0x1,cmask=1,period=200003,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
	.long_desc = "This event counts cycles where the decoder is stalled on an instruction with a length changing prefix (LCP)",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000003,event=0x87",
	.desc = "Stall cycles because IQ is full",
	.topic = "pipeline",
	.long_desc = "Stall cycles due to IQ is full",
},
{
	.name = "br_inst_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x88",
	.desc = "Not taken macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_direct_jump",
	.event = "umask=0x82,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_near_return",
	.event = "umask=0x88,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches with return mnemonic",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x88",
	.desc = "Taken speculative and retired direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x88",
	.desc = "Speculative and retired macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_jmp",
	.event = "umask=0xc2,period=200003,event=0x88",
	.desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x88",
	.desc = "Speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_indirect_near_return",
	.event = "umask=0xc8,period=200003,event=0x88",
	.desc = "Speculative and retired indirect return branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x88",
	.desc = "Speculative and retired direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x88",
	.desc = "Speculative and retired  branches",
	.topic = "pipeline",
	.long_desc = "Counts all near executed branches (not necessarily retired)",
},
{
	.name = "br_misp_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x89",
	.desc = "Not taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_return_near",
	.event = "umask=0x88,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x89",
	.desc = "Mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Counts all near executed branches (not necessarily retired)",
},
{
	.name = "uops_executed_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 0 in this thread",
},
{
	.name = "uops_executed_port.port_0_core",
	.event = "umask=0x1,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 0",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 1 in this thread",
},
{
	.name = "uops_executed_port.port_1_core",
	.event = "umask=0x2,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 1",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 2 in this thread",
},
{
	.name = "uops_executed_port.port_2_core",
	.event = "umask=0x4,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 2",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 3 in this thread",
},
{
	.name = "uops_executed_port.port_3_core",
	.event = "umask=0x8,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 3",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 4 in this thread",
},
{
	.name = "uops_executed_port.port_4_core",
	.event = "umask=0x10,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 4",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 5 in this thread",
},
{
	.name = "uops_executed_port.port_5_core",
	.event = "umask=0x20,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 5",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 6 in this thread",
},
{
	.name = "uops_executed_port.port_6_core",
	.event = "umask=0x40,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are exectuted in port 6",
},
{
	.name = "uops_dispatched_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "Cycles which a uop is dispatched on port 7 in this thread",
},
{
	.name = "uops_executed_port.port_7_core",
	.event = "umask=0x80,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 7",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles  Spec update: HSD135",
	.topic = "pipeline",
	.long_desc = "Cycles allocation is stalled due to resource related reason  Spec update: HSD135",
},
{
	.name = "resource_stalls.rs",
	.event = "umask=0x4,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no eligible RS entry available",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which no instructions were allocated because no Store Buffers (SB) were available",
},
{
	.name = "resource_stalls.rob",
	.event = "umask=0x10,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to re-order buffer full",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l2_pending",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa3",
	.desc = "Cycles with pending L2 cache miss loads  Spec update: HSD78",
	.topic = "pipeline",
	.long_desc = "Cycles with pending L2 miss loads. Set Cmask=2 to count cycle  Spec update: HSD78",
},
{
	.name = "cycle_activity.cycles_ldm_pending",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xa3",
	.desc = "Cycles with pending memory loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending memory loads. Set Cmask=2 to count cycle",
},
{
	.name = "cycle_activity.cycles_no_execute",
	.event = "umask=0x4,cmask=4,period=2000003,event=0xa3",
	.desc = "This event increments by 1 for every cycle where there was no execute for this thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which no instructions were executed in the execution stage of the pipeline",
},
{
	.name = "cycle_activity.stalls_l2_pending",
	.event = "umask=0x5,cmask=5,period=2000003,event=0xa3",
	.desc = "Execution stalls due to L2 cache misses",
	.topic = "pipeline",
	.long_desc = "Number of loads missed L2",
},
{
	.name = "cycle_activity.stalls_ldm_pending",
	.event = "umask=0x6,cmask=6,period=2000003,event=0xa3",
	.desc = "Execution stalls due to memory subsystem",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which no instructions were executed in the execution stage of the pipeline and there were memory instructions pending (waiting for data)",
},
{
	.name = "cycle_activity.cycles_l1d_pending",
	.event = "umask=0x8,cmask=8,period=2000003,event=0xa3",
	.desc = "Cycles with pending L1 cache miss loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending L1 data cache miss loads. Set Cmask=8 to count cycle",
},
{
	.name = "cycle_activity.stalls_l1d_pending",
	.event = "umask=0xc,cmask=12,period=2000003,event=0xa3",
	.desc = "Execution stalls due to L1 data cache misses",
	.topic = "pipeline",
	.long_desc = "Execution stalls due to L1 data cache miss loads. Set Cmask=0CH",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
	.long_desc = "Number of uops delivered by the LSD",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
	.long_desc = "This events counts the cycles where at least one uop was executed. It is counted per thread  Spec update: HSD144, HSD30, HSM31",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
	.long_desc = "This events counts the cycles where at least two uop were executed. It is counted per thread  Spec update: HSD144, HSD30, HSM31",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
	.long_desc = "This events counts the cycles where at least three uop were executed. It is counted per thread  Spec update: HSD144, HSD30, HSM31",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread  Spec update: HSD144, HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
	.long_desc = "Counts total number of uops to be executed per-core each cycle  Spec update: HSD30, HSM31",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core  Spec update: HSD30, HSM31",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter   - architectural event  Spec update: HSD11, HSD140",
	.topic = "pipeline",
	.long_desc = "Number of instructions at retirement  Spec update: HSD11, HSD140",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: HSD140 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: HSD140 (Must be precise)",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000003,event=0xc0",
	.desc = "FP operations retired. X87 FP operations that have no exceptions: Counts also flows that have several X87 or flows that use X87 uops in the exception handling",
	.topic = "pipeline",
	.long_desc = "This is a non-precise version (that is, does not use PEBS) of the event that counts FP operations retired. For X87 FP operations that have no exceptions counting also includes flows that have several X87, or flows that use X87 uops in the exception handling",
},
{
	.name = "other_assists.any_wb_assist",
	.event = "umask=0x40,period=100003,event=0xc1",
	.desc = "Number of times any microcode assist is invoked by HW upon uop writeback",
	.topic = "pipeline",
	.long_desc = "Number of microcode assists invoked by HW upon uop writeback",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Actually retired uops  Supports address when precise (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of micro-ops retired. Use Cmask=1 and invert to count active cycles or stalled cycles  Supports address when precise (Precise event)",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,cmask=10,period=2000003,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,cmask=1,period=2000003,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of retirement slots used each cycle.  There are potentially 4 slots that can be used each cycle - meaning, 4 uops or 4 instructions could retire each cycle (Precise event)",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=2000003,event=0xc3",
	.desc = "Cycles there was a Nuke. Account for both thread-specific and All Thread Nukes",
	.topic = "pipeline",
},
{
	.name = "machine_clears.count",
	.event = "edge=1,umask=0x1,cmask=1,period=100003,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "This event is incremented when self-modifying code (SMC) is detected, which causes a machine clear.  Machine clears can have a significant performance impact if they are happening frequently",
},
{
	.name = "machine_clears.maskmov",
	.event = "umask=0x20,period=100003,event=0xc3",
	.desc = "This event counts the number of executed Intel AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Branch instructions at retirement",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of conditional branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100003,event=0xc4",
	.desc = "Direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call_r3",
	.event = "umask=0x2,period=100003,event=0xc4",
	.desc = "Direct and indirect macro near call instructions retired (captured in ring 3) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100003,event=0xc4",
	.desc = "Return instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of near return instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Not taken branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Counts the number of not taken branch instructions retired",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Number of near taken branches retired (Precise event)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100003,event=0xc4",
	.desc = "Far branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Number of far branches retired",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Mispredicted branch instructions at retirement",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This event counts all mispredicted branch instructions retired. This is a precise event (Must be precise)",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "number of near branch instructions retired that were mispredicted and taken (Precise event)",
	.topic = "pipeline",
	.long_desc = "Number of near branch instructions retired that were taken but mispredicted (Precise event)",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Count cases of saving new LBR",
	.topic = "pipeline",
	.long_desc = "Count cases of saving new LBR records by hardware",
},
{
	.name = "baclears.any",
	.event = "umask=0x1f,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
	.long_desc = "Number of front end re-steers due to BPU misprediction",
},
{
	.name = "unc_c_clockticks",
	.event = "event=0",
	.desc = "Uncore cache clock ticks. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_lookup.any",
	.event = "umask=0x11,event=0x34,filter_state=0x1",
	.desc = "All LLC Misses (code+ data rd + data wr - including demand and prefetch). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_victims.m_state",
	.event = "umask=0x1,event=0x37",
	.desc = "M line evictions from LLC (writebacks to memory). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.data_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x182",
	.desc = "LLC misses - demand and prefetch data reads - excludes LLC prefetches. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.uncacheable",
	.event = "umask=0x3,event=0x35,filter_opc=0x187",
	.desc = "LLC misses - Uncacheable reads (from cpu) . Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x187,filter_nc=1",
	.desc = "MMIO reads. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x18f,filter_nc=1",
	.desc = "MMIO writes. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.rfo_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x190",
	.desc = "LLC prefetch misses for RFO. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.code_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x191",
	.desc = "LLC prefetch misses for code reads. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.data_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x192",
	.desc = "LLC prefetch misses for data reads. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x19e",
	.desc = "LLC misses for PCIe read current. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x1c8",
	.desc = "ItoM write misses (as part of fast string memcpy stores) + PCIe full line writes. Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_non_snoop_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x1c8,filter_tid=0x3e",
	.desc = "PCIe write misses (full cache line). Derived from unc_c_tor_inserts.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_ns_partial_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x180,filter_tid=0x3e",
	.desc = "PCIe writes (partial cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "llc_references.code_llc_prefetch",
	.event = "umask=0x1,event=0x35,filter_opc=0x181",
	.desc = "L2 demand and L2 prefetch code references to LLC. Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_full",
	.event = "umask=0x1,event=0x35,filter_opc=0x18c",
	.desc = "Streaming stores (full cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_partial",
	.event = "umask=0x1,event=0x35,filter_opc=0x18d",
	.desc = "Streaming stores (partial cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_read",
	.event = "umask=0x1,event=0x35,filter_opc=0x19e",
	.desc = "PCIe read current. Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x1c8,filter_tid=0x3e",
	.desc = "PCIe write references (full cache line). Derived from unc_c_tor_inserts.opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_tor_occupancy.llc_data_read",
	.event = "umask=0x3,event=0x36,filter_opc=0x182",
	.desc = "Occupancy counter for LLC data reads (demand and L2 prefetch). Derived from unc_c_tor_occupancy.miss_opcode. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads",
	.event = "umask=0x3,event=0x1",
	.desc = "read requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads_local",
	.event = "umask=0x1,event=0x1",
	.desc = "read requests to local home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads_remote",
	.event = "umask=0x2,event=0x1",
	.desc = "read requests to remote home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes",
	.event = "umask=0xC,event=0x1",
	.desc = "write requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes_local",
	.event = "umask=0x4,event=0x1",
	.desc = "write requests to local home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes_remote",
	.event = "umask=0x8,event=0x1",
	.desc = "write requests to remote home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspcnflct",
	.event = "umask=0x40,event=0x21",
	.desc = "Conflict requests (requests for same address from multiple agents simultaneously). Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rsp_fwd_wb",
	.event = "umask=0x20,event=0x21",
	.desc = "M line forwarded from remote cache along with writeback to memory. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspifwd",
	.event = "umask=0x4,event=0x21",
	.desc = "M line forwarded from remote cache with no writeback to memory. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rsps",
	.event = "umask=0x2,event=0x21",
	.desc = "Shared line response from remote cache. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspsfwd",
	.event = "umask=0x8,event=0x21",
	.desc = "Shared line forwarded from remote cache. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_q_clockticks",
	.event = "event=0x14",
	.desc = "QPI clock ticks. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.perpkg = "1",
},
{
	.name = "qpi_data_bandwidth_tx",
	.event = "umask=0x2,event=0",
	.desc = "Number of data flits transmitted . Derived from unc_q_txl_flits_g0.data. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.unit = "8Bytes",
	.perpkg = "1",
},
{
	.name = "qpi_ctl_bandwidth_tx",
	.event = "umask=0x4,event=0",
	.desc = "Number of non data (control) flits transmitted . Derived from unc_q_txl_flits_g0.non_data. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.unit = "8Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_read",
	.event = "umask=0x3,event=0x4",
	.desc = "read requests to memory controller. Derived from unc_m_cas_count.rd. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_write",
	.event = "umask=0xC,event=0x4",
	.desc = "write requests to memory controller. Derived from unc_m_cas_count.wr. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_clockticks",
	.event = "event=0",
	.desc = "Memory controller clock ticks. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_power_channel_ppd",
	.event = "event=0x85",
	.desc = "Cycles where DRAM ranks are in power down (CKE) mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_channel_ppd / unc_m_clockticks) * 100.",
	.metric_name = "power_channel_ppd %",
},
{
	.name = "unc_m_power_critical_throttle_cycles",
	.event = "event=0x86",
	.desc = "Cycles all ranks are in critical thermal throttle. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_critical_throttle_cycles / unc_m_clockticks) * 100.",
	.metric_name = "power_critical_throttle_cycles %",
},
{
	.name = "unc_m_power_self_refresh",
	.event = "event=0x43",
	.desc = "Cycles Memory is in self refresh power mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_self_refresh / unc_m_clockticks) * 100.",
	.metric_name = "power_self_refresh %",
},
{
	.name = "unc_m_pre_count.page_miss",
	.event = "umask=0x1,event=0x2",
	.desc = "Pre-charges due to page misses. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.rd",
	.event = "umask=0x4,event=0x2",
	.desc = "Pre-charge for reads. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.wr",
	.event = "umask=0x8,event=0x2",
	.desc = "Pre-charge for writes. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_p_clockticks",
	.event = "event=0",
	.desc = "PCU clock ticks. Use to get percentages of PCU cycles events. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
},
{
	.name = "unc_p_power_state_occupancy.cores_c0",
	.event = "event=0x80,occ_sel=1",
	.desc = "This is an occupancy event that tracks the number of cores that are in C0.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c0 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c0 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c3",
	.event = "event=0x80,occ_sel=2",
	.desc = "This is an occupancy event that tracks the number of cores that are in C3.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c3 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c3 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c6",
	.event = "event=0x80,occ_sel=3",
	.desc = "This is an occupancy event that tracks the number of cores that are in C6.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events . Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c6 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c6 %",
},
{
	.name = "unc_p_prochot_external_cycles",
	.event = "event=0xa",
	.desc = "Counts the number of cycles that we are in external PROCHOT mode.  This mode is triggered when a sensor off the die determines that something off-die (like DRAM) is too hot and must throttle to avoid damaging the chip. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_prochot_external_cycles / unc_p_clockticks) * 100.",
	.metric_name = "prochot_external_cycles %",
},
{
	.name = "unc_p_freq_max_limit_thermal_cycles",
	.event = "event=0x4",
	.desc = "Counts the number of cycles when temperature is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_limit_thermal_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_limit_thermal_cycles %",
},
{
	.name = "unc_p_freq_max_os_cycles",
	.event = "event=0x6",
	.desc = "Counts the number of cycles when the OS is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_os_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_os_cycles %",
},
{
	.name = "unc_p_freq_max_power_cycles",
	.event = "event=0x5",
	.desc = "Counts the number of cycles when power is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_power_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_power_cycles %",
},
{
	.name = "unc_p_freq_trans_cycles",
	.event = "event=0x74",
	.desc = "Counts the number of cycles when current is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_trans_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_trans_cycles %",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Misses in all TLB levels that cause a page walk of any page size",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (4K)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to demand load misses that caused 4K page walks in any TLB levels",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to demand load misses that caused 2M/4M page walks in any TLB levels",
},
{
	.name = "dtlb_load_misses.walk_completed_1g",
	.event = "umask=0x8,period=2000003,event=0x8",
	.desc = "Load miss in all TLB levels causes a page walk that completes. (1G)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes of any page size",
	.topic = "virtual memory",
	.long_desc = "Completed page walks in any TLB of any page size due to demand load misses",
},
{
	.name = "dtlb_load_misses.walk_duration",
	.event = "umask=0x10,period=2000003,event=0x8",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles when the  page miss handler (PMH) is servicing page walks caused by DTLB load misses",
},
{
	.name = "dtlb_load_misses.stlb_hit_4k",
	.event = "umask=0x20,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
	.long_desc = "This event counts load operations from a 4K page that miss the first DTLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_load_misses.stlb_hit_2m",
	.event = "umask=0x40,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
	.long_desc = "This event counts load operations from a 2M page that miss the first DTLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x60,period=2000003,event=0x8",
	.desc = "Load operations that miss the first DTLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "Number of cache load STLB hits. No page walk",
},
{
	.name = "dtlb_load_misses.pde_cache_miss",
	.event = "umask=0x80,period=100003,event=0x8",
	.desc = "DTLB demand load misses with low part of linear-to-physical address translation missed",
	.topic = "virtual memory",
	.long_desc = "DTLB demand load misses with low part of linear-to-physical address translation missed",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Miss in all TLB levels causes a page walk of any page size (4K/2M/4M/1G)",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (4K)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to store misses in one or more TLB levels of 4K page structure",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to store misses in one or more TLB levels of 2M/4M page structure",
},
{
	.name = "dtlb_store_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks. (1G)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to store miss in any TLB levels of any page size (4K/2M/4M/1G)",
},
{
	.name = "dtlb_store_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x49",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles when the  page miss handler (PMH) is servicing page walks caused by DTLB store misses",
},
{
	.name = "dtlb_store_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
	.long_desc = "This event counts store operations from a 4K page that miss the first DTLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_store_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
	.long_desc = "This event counts store operations from a 2M page that miss the first DTLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x49",
	.desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
},
{
	.name = "dtlb_store_misses.pde_cache_miss",
	.event = "umask=0x80,period=100003,event=0x49",
	.desc = "DTLB store misses with low part of linear-to-physical address translation missed",
	.topic = "virtual memory",
	.long_desc = "DTLB store misses with low part of linear-to-physical address translation missed",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Cycle count for an Extended Page table walk",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Misses in ITLB that causes a page walk of any page size",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (4K)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to misses in ITLB 4K page entries",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Completed page walks due to misses in ITLB 2M/4M page entries",
},
{
	.name = "itlb_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x85",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (1G)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x85",
	.desc = "Misses in all ITLB levels that cause completed page walks",
	.topic = "virtual memory",
	.long_desc = "Completed page walks in ITLB of any page size",
},
{
	.name = "itlb_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles when the  page miss handler (PMH) is servicing page walks caused by ITLB misses",
},
{
	.name = "itlb_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x85",
	.desc = "Core misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
	.long_desc = "ITLB misses that hit STLB (4K)",
},
{
	.name = "itlb_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x85",
	.desc = "Code misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
	.long_desc = "ITLB misses that hit STLB (2M)",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x85",
	.desc = "Operations that miss the first ITLB level but hit the second and do not cause any page walks",
	.topic = "virtual memory",
	.long_desc = "ITLB misses that hit STLB. No page walk",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100003,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "Counts the number of ITLB flushes, includes 4k/2M/4M pages",
},
{
	.name = "page_walker_loads.dtlb_l1",
	.event = "umask=0x11,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L1+FB",
	.topic = "virtual memory",
	.long_desc = "Number of DTLB page walker loads that hit in the L1+FB",
},
{
	.name = "page_walker_loads.dtlb_l2",
	.event = "umask=0x12,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L2",
	.topic = "virtual memory",
	.long_desc = "Number of DTLB page walker loads that hit in the L2",
},
{
	.name = "page_walker_loads.dtlb_l3",
	.event = "umask=0x14,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L3 + XSNP  Spec update: HSD25",
	.topic = "virtual memory",
	.long_desc = "Number of DTLB page walker loads that hit in the L3  Spec update: HSD25",
},
{
	.name = "page_walker_loads.dtlb_memory",
	.event = "umask=0x18,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in Memory  Spec update: HSD25",
	.topic = "virtual memory",
	.long_desc = "Number of DTLB page walker loads from memory  Spec update: HSD25",
},
{
	.name = "page_walker_loads.itlb_l1",
	.event = "umask=0x21,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L1+FB",
	.topic = "virtual memory",
	.long_desc = "Number of ITLB page walker loads that hit in the L1+FB",
},
{
	.name = "page_walker_loads.itlb_l2",
	.event = "umask=0x22,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L2",
	.topic = "virtual memory",
	.long_desc = "Number of ITLB page walker loads that hit in the L2",
},
{
	.name = "page_walker_loads.itlb_l3",
	.event = "umask=0x24,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L3 + XSNP  Spec update: HSD25",
	.topic = "virtual memory",
	.long_desc = "Number of ITLB page walker loads that hit in the L3  Spec update: HSD25",
},
{
	.name = "page_walker_loads.itlb_memory",
	.event = "umask=0x28,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in Memory  Spec update: HSD25",
	.topic = "virtual memory",
	.long_desc = "Number of ITLB page walker loads from memory  Spec update: HSD25",
},
{
	.name = "page_walker_loads.ept_dtlb_l1",
	.event = "umask=0x41,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the DTLB that hit in the L1 and FB",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_dtlb_l2",
	.event = "umask=0x42,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the DTLB that hit in the L2",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_dtlb_l3",
	.event = "umask=0x44,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the DTLB that hit in the L3",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_dtlb_memory",
	.event = "umask=0x48,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the DTLB that hit in memory",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_itlb_l1",
	.event = "umask=0x81,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the ITLB that hit in the L1 and FB",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_itlb_l2",
	.event = "umask=0x82,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the ITLB that hit in the L2",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_itlb_l3",
	.event = "umask=0x84,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the ITLB that hit in the L2",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.ept_itlb_memory",
	.event = "umask=0x88,period=2000003,event=0xbc",
	.desc = "Counts the number of Extended Page Table walks from the ITLB that hit in memory",
	.topic = "virtual memory",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100003,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100003,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "Count number of STLB flush attempts",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_icelake[] = {
{
	.name = "l2_rqsts.demand_data_rd_miss",
	.event = "umask=0x21,period=200003,event=0x24",
	.desc = "Demand Data Read miss L2, no rejects",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests that miss L2 cache. Only not rejected loads are counted",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x22,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts the RFO (Read-for-Ownership) requests that miss L2 cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x24,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
	.long_desc = "Counts L2 cache misses when fetching instructions",
},
{
	.name = "l2_rqsts.all_demand_miss",
	.event = "umask=0x27,period=200003,event=0x24",
	.desc = "Demand requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand requests that miss L2 cache",
},
{
	.name = "l2_rqsts.swpf_miss",
	.event = "umask=0x28,period=200003,event=0x24",
	.desc = "SW prefetch requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts Software prefetch requests that miss the L2 cache. This event accounts for PREFETCHNTA and PREFETCHT0/1/2 instructions",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0xc1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests initiated by load instructions that hit L2 cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0xc2,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the RFO (Read-for-Ownership) requests that hit L2 cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0xc4,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
	.long_desc = "Counts L2 cache hits when fetching instructions, code reads",
},
{
	.name = "l2_rqsts.swpf_hit",
	.event = "umask=0xc8,period=200003,event=0x24",
	.desc = "SW prefetch requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts Software prefetch requests that hit the L2 cache. This event accounts for PREFETCHNTA and PREFETCHT0/1/2 instructions",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0xe1,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests (including requests from L1D hardware prefetchers). These loads may hit or miss L2 cache. Only non rejected loads are counted",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xe2,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "Counts the total number of RFO (read for ownership) requests to L2 cache. L2 RFO requests include both L1D demand RFO misses as well as L1D RFO prefetches",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0xe4,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "Counts the total number of L2 code requests",
},
{
	.name = "l2_rqsts.all_demand_references",
	.event = "umask=0xe7,period=200003,event=0x24",
	.desc = "Demand requests to L2 cache",
	.topic = "cache",
	.long_desc = "Counts demand requests to L2 cache",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "Number of L1D misses that are outstanding",
	.topic = "cache",
	.long_desc = "Counts number of L1D misses that are outstanding in each cycle, that is each cycle the number of Fill Buffers (FB) outstanding required by Demand Reads. FB either is held by demand loads, or it is held by non-demand loads and gets hit at least once by demand. The valid outstanding interval is defined until the FB deallocation by one of the following ways: from FB allocation, if FB is allocated by demand from the demand Hit FB, if it is allocated by hardware or software prefetch. Note: In the L1D, a Demand Read contains cacheable or noncacheable demand loads, including ones causing cache-line splits and reads due to page walks resulted from any request type",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
	.long_desc = "Counts duration of L1D miss outstanding in cycles",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,event=0x48",
	.desc = "Number of cycles a demand request has waited due to L1D Fill Buffer (FB) unavailablability",
	.topic = "cache",
	.long_desc = "Counts number of cycles a demand request has waited due to L1D Fill Buffer (FB) unavailablability. Demand requests include cacheable/uncacheable demand load, store, lock or SW prefetch accesses",
},
{
	.name = "l1d_pend_miss.fb_full_periods",
	.event = "umask=0x2,period=2000003,cmask=1,edge=1,event=0x48",
	.desc = "Number of phases a demand request has waited due to L1D Fill Buffer (FB) unavailablability",
	.topic = "cache",
	.long_desc = "Counts number of phases a demand request has waited due to L1D Fill Buffer (FB) unavailablability. Demand requests include cacheable/uncacheable demand load, store, lock or SW prefetch accesses",
},
{
	.name = "l1d_pend_miss.l2_stall",
	.event = "umask=0x4,period=2000003,event=0x48",
	.desc = "Number of cycles a demand request has waited due to L1D due to lack of L2 resources",
	.topic = "cache",
	.long_desc = "Counts number of cycles a demand request has waited due to L1D due to lack of L2 resources. Demand requests include cacheable/uncacheable demand load, store, lock or SW prefetch accesses",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "Counts the number of cache lines replaced in L1 data cache",
	.topic = "cache",
	.long_desc = "Counts L1D data line replacements including opportunistic replacements, and replacements that require stall-for-replace or block-for-replace",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles with offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding demand rfo Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding cacheable Core Data Read transactions in the super queue every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts cycles when offcore outstanding cacheable Core Data Read transactions are present in the super queue. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
	.long_desc = "Counts the Demand Data Read requests sent to uncore. Use it in conjunction with OFFCORE_REQUESTS_OUTSTANDING to determine average latency in the uncore",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "Counts the demand RFO (read for ownership) requests including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts the demand and prefetch data reads. All Core Data Reads include cacheable 'Demands' and L2 prefetchers (not L3 prefetchers). Counting also covers reads due to page walks resulted from any request type",
},
{
	.name = "offcore_requests.all_requests",
	.event = "umask=0x80,period=100003,event=0xb0",
	.desc = "Any memory transaction that reached the SQ",
	.topic = "cache",
	.long_desc = "Counts memory transactions reached the super queue including requests initiated by the core, all L3 prefetches, page walks, etc.",
},
{
	.name = "mem_inst_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load instructions that miss the STLB  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions that true miss the STLB  Supports address when precise (Precise event)",
},
{
	.name = "mem_inst_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store instructions that miss the STLB  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired store instructions that true miss the STLB  Supports address when precise (Precise event)",
},
{
	.name = "mem_inst_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load instructions with locked access  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with locked access  Supports address when precise (Precise event)",
},
{
	.name = "mem_inst_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load instructions that split across a cacheline boundary  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions that split across a cacheline boundary  Supports address when precise (Precise event)",
},
{
	.name = "mem_inst_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store instructions that split across a cacheline boundary  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired store instructions that split across a cacheline boundary  Supports address when precise (Precise event)",
},
{
	.name = "mem_inst_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load instructions  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts all retired load instructions. This event accounts for SW prefetch instructions for loads  Supports address when precise (Precise event)",
},
{
	.name = "mem_inst_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store instructions  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts all retired store instructions. This event account for SW prefetch instructions and PREFETCHW instruction for stores  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load instructions with L1 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that hit in the L1 data cache. This event includes all SW prefetches and lock instructions regardless of the data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load instructions with L2 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with L2 cache hits as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l3_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load instructions with L3 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that hit in the L3 cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load instructions missed L1 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that missed in the L1 cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Retired load instructions missed L2 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions missed L2 cache as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l3_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Retired load instructions missed L3 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that missed in the L3 cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.fb_hit",
	.event = "umask=0x40,period=100007,event=0xd1",
	.desc = "Number of completed demand load requests that missed the L1, but hit the FB(fill buffer), because a preceding miss to the same cacheline initiated the line to be brought into L1, but data is not yet ready in L1  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop was load missed in L1 but hit FB (Fill Buffers) due to preceding miss to the same cache line with data not ready  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load instructions whose data sources were L3 hit and cross-core snoop missed in on-pkg core cache  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts the retired load instructions whose data sources were L3 hit and cross-core snoop missed in on-pkg core cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load instructions whose data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions whose data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load instructions whose data sources were HitM responses from shared L3  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions whose data sources were HitM responses from shared L3  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load instructions whose data sources were hits in L3 without snoops required  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions whose data sources were hits in L3 without snoops required  Supports address when precise (Precise event)",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x1f,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "Counts the number of L2 cache lines filling the L2. Counting does not cover rejects",
},
{
	.name = "sq_misc.sq_full",
	.event = "umask=0x4,period=100003,event=0xf4",
	.desc = "Cycles the thread is active and superQ cannot take any more entries",
	.topic = "cache",
	.long_desc = "Counts the cycles for which the thread is active and the superQ cannot take any more entries",
},
{
	.name = "assists.fp",
	.event = "umask=0x2,period=100003,cmask=1,event=0xc1",
	.desc = "Counts all microcode FP assists",
	.topic = "floating point",
	.long_desc = "Counts all microcode Floating Point assists",
},
{
	.name = "fp_arith_inst_retired.scalar_double",
	.event = "umask=0x1,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 1 computation. Applies to SSE* and AVX* scalar double precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 RANGE SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
	.topic = "floating point",
	.long_desc = "Counts number of SSE/AVX computational scalar double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 1 computational operation. Applies to SSE* and AVX* scalar double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
},
{
	.name = "fp_arith_inst_retired.scalar_single",
	.event = "umask=0x2,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 1 computation. Applies to SSE* and AVX* scalar single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 RANGE SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
	.topic = "floating point",
	.long_desc = "Counts number of SSE/AVX computational scalar single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 1 computational operation. Applies to SSE* and AVX* scalar single precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT RSQRT RCP FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
},
{
	.name = "fp_arith_inst_retired.128b_packed_double",
	.event = "umask=0x4,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 2 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT RSQRT14 RCP14 RANGE DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
	.topic = "floating point",
	.long_desc = "Counts number of SSE/AVX computational 128-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 2 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
},
{
	.name = "fp_arith_inst_retired.128b_packed_single",
	.event = "umask=0x8,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 4 computation operations, one for each element.  Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
	.topic = "floating point",
	.long_desc = "Counts number of SSE/AVX computational 128-bit packed single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 4 computation operations, one for each element.  Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT RSQRT RCP DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
},
{
	.name = "fp_arith_inst_retired.256b_packed_double",
	.event = "umask=0x10,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 4 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 RANGE SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
	.topic = "floating point",
	.long_desc = "Counts number of SSE/AVX computational 256-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 4 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
},
{
	.name = "fp_arith_inst_retired.256b_packed_single",
	.event = "umask=0x20,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 8 computation operations, one for each element.  Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP14 RSQRT14 RANGE SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
	.topic = "floating point",
	.long_desc = "Counts number of SSE/AVX computational 256-bit packed single precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 8 computation operations, one for each element.  Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT RSQRT RCP DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
},
{
	.name = "fp_arith_inst_retired.512b_packed_double",
	.event = "umask=0x40,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 512-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 16 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT RSQRT14 RCP14 RANGE FM(N)ADD/SUB. FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
	.topic = "floating point",
	.long_desc = "Counts number of SSE/AVX computational 512-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 8 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT RSQRT14 RCP14 RANGE FM(N)ADD/SUB. FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
},
{
	.name = "fp_arith_inst_retired.512b_packed_single",
	.event = "umask=0x80,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 512-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 8 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT RSQRT14 RCP14 RANGE FM(N)ADD/SUB. FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
	.topic = "floating point",
	.long_desc = "Counts number of SSE/AVX computational 512-bit packed double precision floating-point instructions retired; some instructions will count twice as noted below.  Each count represents 16 computation operations, one for each element.  Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT RSQRT14 RCP14 RANGE FM(N)ADD/SUB. FM(N)ADD/SUB instructions count twice as they perform 2 calculations per element",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.mite_cycles_ok",
	.event = "umask=0x4,period=2000003,cmask=5,event=0x79",
	.desc = "Cycles MITE is delivering optimal number of Uops",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles where optimal number of uops was delivered to the Instruction Decode Queue (IDQ) from the MITE (legacy decode pipeline) path. During these cycles uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.mite_cycles_any",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles uops were delivered to the Instruction Decode Queue (IDQ) from the MITE (legacy decode pipeline) path. During these cycles uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Counts the number of uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
},
{
	.name = "idq.dsb_cycles_ok",
	.event = "umask=0x8,period=2000003,cmask=5,event=0x79",
	.desc = "Cycles DSB is delivering optimal number of Uops",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles where optimal number of uops was delivered to the Instruction Decode Queue (IDQ) from the MITE (legacy decode pipeline) path. During these cycles uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.dsb_cycles_any",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles uops were delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
},
{
	.name = "idq.ms_switches",
	.event = "umask=0x30,period=2000003,cmask=1,edge=1,event=0x79",
	.desc = "Number of switches from DSB or MITE to the MS",
	.topic = "frontend",
	.long_desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to IDQ while MS is busy",
	.topic = "frontend",
	.long_desc = "Counts the total number of uops delivered by the Microcode Sequencer (MS). Any instruction over 4 uops will be delivered by the MS. Some instructions such as transcendentals may additionally generate uops from the MS",
},
{
	.name = "idq.ms_cycles_any",
	.event = "umask=0x30,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to IDQ while MS is busy",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "icache_16b.ifdata_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction cache miss",
	.topic = "frontend",
	.long_desc = "Counts cycles where a code line fetch is stalled due to an L1 instruction cache miss. The legacy decode pipeline works at a 16 Byte granularity",
},
{
	.name = "icache_64b.iftag_hit",
	.event = "umask=0x1,period=200003,event=0x83",
	.desc = "Instruction fetch tag lookups that hit in the instruction cache (L1I). Counts at 64-byte cache-line granularity",
	.topic = "frontend",
	.long_desc = "Counts instruction fetch tag lookups that hit in the instruction cache (L1I). Counts at 64-byte cache-line granularity. Accounts for both cacheable and uncacheable accesses",
},
{
	.name = "icache_64b.iftag_miss",
	.event = "umask=0x2,period=200003,event=0x83",
	.desc = "Instruction fetch tag lookups that miss in the instruction cache (L1I). Counts at 64-byte cache-line granularity",
	.topic = "frontend",
	.long_desc = "Counts instruction fetch tag lookups that miss in the instruction cache (L1I). Counts at 64-byte cache-line granularity. Accounts for both cacheable and uncacheable accesses",
},
{
	.name = "icache_64b.iftag_stall",
	.event = "umask=0x4,period=200003,event=0x83",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction cache tag miss",
	.topic = "frontend",
	.long_desc = "Counts cycles where a code fetch is stalled due to L1 instruction cache tag miss",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered by IDQ when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts the number of uops not delivered to by the Instruction Decode Queue (IDQ) to the back-end of the pipeline when there was no back-end stalls. This event counts for one SMT thread in a given cycle",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=5,event=0x9c",
	.desc = "Cycles when no uops are not delivered by the IDQ when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles when no uops were delivered by the Instruction Decode Queue (IDQ) to the back-end of the pipeline when there was no back-end stalls. This event counts for one SMT thread in a given cycle",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Cycles when optimal number of uops was delivered to the back-end when the back-end is not stalled",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles when the optimal number of uops were delivered by the Instruction Decode Queue (IDQ) to the back-end of the pipeline when there was no back-end stalls. This event counts for one SMT thread in a given cycle",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "DSB-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "Decode Stream Buffer (DSB) is a Uop-cache that holds translations of previously fetched instructions that were decoded by the legacy x86 decode pipeline (MITE). This event counts fetch penalty cycles when a transition occurs from DSB to MITE",
},
{
	.name = "frontend_retired.dsb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x11",
	.desc = "Retired Instructions who experienced DSB miss (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced DSB (Decode stream buffer i.e. the decoded instruction-cache) miss (Precise event)",
},
{
	.name = "frontend_retired.l1i_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x12",
	.desc = "Retired Instructions who experienced Instruction L1 Cache true miss (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions who experienced Instruction L1 Cache true miss (Precise event)",
},
{
	.name = "frontend_retired.l2_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x13",
	.desc = "Retired Instructions who experienced Instruction L2 Cache true miss (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions who experienced Instruction L2 Cache true miss (Precise event)",
},
{
	.name = "frontend_retired.itlb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x14",
	.desc = "Retired Instructions who experienced iTLB true miss (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced iTLB (Instruction TLB) true miss (Precise event)",
},
{
	.name = "frontend_retired.stlb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x15",
	.desc = "Retired Instructions who experienced STLB (2nd level TLB) true miss (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced STLB (2nd level TLB) true miss (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_2",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x500206",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 2 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 2 cycles which was not interrupted by a back-end stall (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_4",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x500406",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 4 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 4 cycles which was not interrupted by a back-end stall (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_8",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x500806",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 8 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 8 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_16",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x501006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 16 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 16 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_32",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x502006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 32 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 32 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_64",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x504006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 64 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 64 cycles which was not interrupted by a back-end stall (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_128",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x508006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 128 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 128 cycles which was not interrupted by a back-end stall (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_256",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x510006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 256 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 256 cycles which was not interrupted by a back-end stall (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_512",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x520006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 512 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 512 cycles which was not interrupted by a back-end stall (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_1",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x100206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 1 bubble-slot for a period of 2 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after the front-end had at least 1 bubble-slot for a period of 2 cycles. A bubble-slot is an empty issue-pipeline slot while there was no RAT stall (Precise event)",
},
{
	.name = "tx_mem.abort_conflict",
	.event = "umask=0x1,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data conflict on a transactionally accessed address",
	.topic = "memory",
	.long_desc = "Counts the number of times a TSX line had a cache conflict",
},
{
	.name = "tx_mem.abort_capacity_write",
	.event = "umask=0x2,period=2000003,event=0x54",
	.desc = "Speculatively counts the number TSX Aborts due to a data capacity limitation for transactional writes",
	.topic = "memory",
	.long_desc = "Speculatively counts the number Transactional Synchronization Extensions (TSX) Aborts due to a data capacity limitation for transactional writes",
},
{
	.name = "tx_mem.abort_hle_store_to_elided_lock",
	.event = "umask=0x4,period=100003,event=0x54",
	.desc = "Number of times a HLE transactional region aborted due to a non XRELEASE prefixed instruction writing to an elided lock in the elision buffer",
	.topic = "memory",
	.long_desc = "Counts the number of times a TSX Abort was triggered due to a non-release/commit store to lock",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_not_empty",
	.event = "umask=0x8,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to NoAllocatedElisionBuffer being non-zero",
	.topic = "memory",
	.long_desc = "Counts the number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_mismatch",
	.event = "umask=0x10,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to XRELEASE lock not satisfying the address and value requirements in the elision buffer",
	.topic = "memory",
	.long_desc = "Counts the number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_unsupported_alignment",
	.event = "umask=0x20,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to an unsupported read alignment from the elision buffer",
	.topic = "memory",
	.long_desc = "Counts the number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
},
{
	.name = "tx_mem.hle_elision_buffer_full",
	.event = "umask=0x40,period=2000003,event=0x54",
	.desc = "Number of times HLE lock could not be elided due to ElisionBufferAvailable being zero",
	.topic = "memory",
	.long_desc = "Counts the number of times we could not allocate Lock Buffer",
},
{
	.name = "tx_exec.misc2",
	.event = "umask=0x2,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions that may cause a transactional abort was executed inside a transactional region",
	.topic = "memory",
	.long_desc = "Counts Unfriendly TSX abort triggered by a vzeroupper instruction",
},
{
	.name = "tx_exec.misc3",
	.event = "umask=0x4,period=2000003,event=0x5d",
	.desc = "Number of times an instruction execution caused the transactional nest count supported to be exceeded",
	.topic = "memory",
	.long_desc = "Counts Unfriendly TSX abort triggered by a nest count that is too deep",
},
{
	.name = "cycle_activity.cycles_l3_miss",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Cycles while L3 cache miss demand load is outstanding",
	.topic = "memory",
},
{
	.name = "cycle_activity.stalls_l3_miss",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Execution stalls while L3 cache miss demand load is outstanding",
	.topic = "memory",
},
{
	.name = "offcore_requests.l3_miss_demand_data_rd",
	.event = "umask=0x10,period=100003,event=0xb0",
	.desc = "Demand Data Read requests who miss L3 cache",
	.topic = "memory",
	.long_desc = "Demand Data Read requests who miss L3 cache",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Number of machine clears due to memory ordering conflicts",
	.topic = "memory",
	.long_desc = "Counts the number of Machine Clears detected dye to memory ordering. Memory Ordering Machine Clears may apply when a memory read may not conform to the memory ordering rules of the x86 architecture",
},
{
	.name = "hle_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution started",
	.topic = "memory",
	.long_desc = "Counts the number of times we entered an HLE region. Does not count nested transactions",
},
{
	.name = "hle_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution successfully committed  Supports address when precise",
	.topic = "memory",
	.long_desc = "Counts the number of times HLE commit succeeded  Supports address when precise",
},
{
	.name = "hle_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to any reasons (multiple categories may count as one)",
	.topic = "memory",
	.long_desc = "Counts the number of times HLE abort was triggered",
},
{
	.name = "hle_retired.aborted_mem",
	.event = "umask=0x8,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Counts the number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
},
{
	.name = "hle_retired.aborted_unfriendly",
	.event = "umask=0x20,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions and certain unfriendly events (such as AD assists etc.)",
	.topic = "memory",
	.long_desc = "Counts the number of times an HLE execution aborted due to HLE-unfriendly instructions and certain unfriendly events (such as AD assists etc.)",
},
{
	.name = "hle_retired.aborted_events",
	.event = "umask=0x80,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to unfriendly events (such as interrupts)",
	.topic = "memory",
	.long_desc = "Counts the number of times an HLE execution aborted due to unfriendly events (such as interrupts)",
},
{
	.name = "rtm_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution started",
	.topic = "memory",
	.long_desc = "Counts the number of times we entered an RTM region. Does not count nested transactions",
},
{
	.name = "rtm_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution successfully committed",
	.topic = "memory",
	.long_desc = "Counts the number of times RTM commit succeeded",
},
{
	.name = "rtm_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted  Supports address when precise",
	.topic = "memory",
	.long_desc = "Counts the number of times RTM abort was triggered  Supports address when precise",
},
{
	.name = "rtm_retired.aborted_mem",
	.event = "umask=0x8,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Counts the number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
},
{
	.name = "rtm_retired.aborted_unfriendly",
	.event = "umask=0x20,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Counts the number of times an RTM execution aborted due to HLE-unfriendly instructions",
},
{
	.name = "rtm_retired.aborted_memtype",
	.event = "umask=0x40,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Counts the number of times an RTM execution aborted due to incompatible memory type",
},
{
	.name = "rtm_retired.aborted_events",
	.event = "umask=0x80,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
	.topic = "memory",
	.long_desc = "Counts the number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 4 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 4 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 8 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 8 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 16 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 16 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 32 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 32 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 64 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 64 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 128 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 128 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 256 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 256 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 512 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 512 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "topdown.slots",
	.event = "umask=0x4,period=10000003,event=0",
	.desc = "Counts the number of available slots for an unhalted logical processor",
	.topic = "other",
	.long_desc = "Counts the number of available slots for an unhalted logical processor. The event increments by machine-width of the narrowest pipeline as employed by the Top-down Microarchitecture Analysis method. The count is distributed among unhalted logical processors (hyper-threads) who share the same physical core. Software can use this event as the denominator for the top-level metrics of the Top-down Microarchitecture Analysis method. This event is counted on a designated fixed counter (Fixed Counter 3) and is an architectural event",
},
{
	.name = "core_power.lvl0_turbo_license",
	.event = "umask=0x7,period=200003,event=0x28",
	.desc = "Core cycles where the core was running in a manner where Turbo may be clipped to the Non-AVX turbo schedule",
	.topic = "other",
	.long_desc = "Counts Core cycles where the core was running with power-delivery for baseline license level 0.  This includes non-AVX codes, SSE, AVX 128-bit, and low-current AVX 256-bit codes",
},
{
	.name = "core_power.lvl1_turbo_license",
	.event = "umask=0x18,period=200003,event=0x28",
	.desc = "Core cycles where the core was running in a manner where Turbo may be clipped to the AVX2 turbo schedule",
	.topic = "other",
	.long_desc = "Counts Core cycles where the core was running with power-delivery for license level 1.  This includes high current AVX 256-bit instructions as well as low current AVX 512-bit instructions",
},
{
	.name = "core_power.lvl2_turbo_license",
	.event = "umask=0x20,period=200003,event=0x28",
	.desc = "Core cycles where the core was running in a manner where Turbo may be clipped to the AVX512 turbo schedule",
	.topic = "other",
	.long_desc = "Core cycles where the core was running with power-delivery for license level 2 (introduced in Skylake Server microarchtecture).  This includes high current AVX 512-bit instructions",
},
{
	.name = "sw_prefetch_access.nta",
	.event = "umask=0x1,period=2000003,event=0x32",
	.desc = "Number of PREFETCHNTA instructions executed",
	.topic = "other",
	.long_desc = "Counts the number of PREFETCHNTA instructions executed",
},
{
	.name = "sw_prefetch_access.t0",
	.event = "umask=0x2,period=2000003,event=0x32",
	.desc = "Number of PREFETCHT0 instructions executed",
	.topic = "other",
	.long_desc = "Counts the number of PREFETCHT0 instructions executed",
},
{
	.name = "sw_prefetch_access.t1_t2",
	.event = "umask=0x4,period=2000003,event=0x32",
	.desc = "Number of PREFETCHT1 or PREFETCHT2 instructions executed",
	.topic = "other",
	.long_desc = "Counts the number of PREFETCHT1 or PREFETCHT2 instructions executed",
},
{
	.name = "sw_prefetch_access.prefetchw",
	.event = "umask=0x8,period=2000003,event=0x32",
	.desc = "Number of PREFETCHW instructions executed",
	.topic = "other",
	.long_desc = "Counts the number of PREFETCHW instructions executed",
},
{
	.name = "topdown.slots_p",
	.event = "umask=0x1,period=10000003,event=0xa4",
	.desc = "Counts the number of available slots for an unhalted logical processor",
	.topic = "other",
	.long_desc = "Counts the number of available slots for an unhalted logical processor. The event increments by machine-width of the narrowest pipeline as employed by the Top-down Microarchitecture Analysis method. The count is distributed among unhalted logical processors (hyper-threads) who share the same physical core",
},
{
	.name = "topdown.backend_bound_slots",
	.event = "umask=0x2,period=10000003,event=0xa4",
	.desc = "Issue slots where no uops were being issued due to lack of back end resources",
	.topic = "other",
},
{
	.name = "assists.any",
	.event = "umask=0x7,period=100003,event=0xc1",
	.desc = "Number of occurrences where a microcode assist is invoked by hardware",
	.topic = "other",
	.long_desc = "Counts the number of occurrences where a microcode assist is invoked by hardware Examples include AD (page Access Dirty), FP and AVX related assists",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. Fixed Counter - architectural event",
	.topic = "pipeline",
	.long_desc = "Counts the number of X86 instructions retired - an Architectural PerfMon event. Counting continues during hardware interrupts, traps, and inside interrupt handlers. Notes: INST_RETIRED.ANY is counted by a designated fixed counter freeing up programmable counters to count other events. INST_RETIRED.ANY_P is counted by a programmable counter",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0",
	.desc = "Precise instruction retired event with a reduced effect of PEBS shadow in IP distribution (Must be precise)",
	.topic = "pipeline",
	.long_desc = "A version of INST_RETIRED that allows for a more unbiased distribution of samples across instructions retired. It utilizes the Precise Distribution of Instructions Retired (PDIR) feature to mitigate some bias in how retired instructions get sampled. Use on Fixed Counter 0 (Must be precise)",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state. This event has a constant ratio with the CPU_CLK_UNHALTED.REF_XCLK event. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. Note: On all current platforms this event stops counting during 'throttling (TM)' states duty off periods the processor is 'halted'.  The counter update is done at a lower clock rate then the core clock the overflow status bit for this counter may appear 'sticky'.  After the counter has overflowed and software clears the overflow status bit and resets the counter to less than MAX. The reset value to the counter is not clocked immediately so the overflow status bit will flip 'high (1)' and generate another PMI (if enabled) after which the reset value gets clocked into the counter. Therefore, software will get the interrupt, read the overflow status bit '1 for bit 34 while the counter value is less than MAX. Software should ignore this case",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Loads blocked by overlapping with store buffer that cannot be forwarded",
	.topic = "pipeline",
	.long_desc = "Counts the number of times the load operation got the true Block-on-Store blocking code preventing store forwarding. This includes cases when: a. preceding store conflicts with the load (incomplete overlap),b. store forwarding is impossible due to u-arch limitations, c. preceding lock RMW operations are not forwarded, d. store has the no-forward bit set (uncacheable/page-split/masked stores), e. all-blocking stores are used (mostly, fences and port I/O), and others. The most common case is a load blocked due to its address range overlapping with a preceding smaller uncompleted store. Note: This event does not take into account cases of out-of-SW-control (for example, SbTailHit), unknown physical STA, and cases of blocking loads on store due to being non-WB memory type or a lock. These cases are covered by other events. See the table of not supported store forwards in the Optimization Guide",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
	.long_desc = "Counts the number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare on address",
	.topic = "pipeline",
	.long_desc = "Counts the number of times a load got blocked due to false dependencies in MOB due to partial compare on address",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for this thread",
	.topic = "pipeline",
	.long_desc = "Counts core cycles when the Resource allocator was stalled due to recovery from an earlier branch misprediction or machine clear event",
},
{
	.name = "int_misc.all_recovery_cycles",
	.event = "umask=0x3,period=2000003,cmask=1,event=0xd",
	.desc = "Cycles the Backend cluster is recovering after a miss-speculation or a Store Buffer or Load Buffer drain stall",
	.topic = "pipeline",
	.long_desc = "Counts cycles the Backend cluster is recovering after a miss-speculation or a Store Buffer or Load Buffer drain stall",
},
{
	.name = "int_misc.clear_resteer_cycles",
	.event = "umask=0x80,period=2000003,event=0xd",
	.desc = "Counts cycles after recovery from a branch misprediction or machine clear till the first uop is issued from the resteered path",
	.topic = "pipeline",
	.long_desc = "Cycles after recovery from a branch misprediction or machine clear till the first uop is issued from the resteered path",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that RAT issues to RS",
	.topic = "pipeline",
	.long_desc = "Counts the number of uops that the Resource Allocation Table (RAT) issues to the Reservation Station (RS)",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when RAT does not issue Uops to RS for the thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which the Resource Allocation Table (RAT) does not issue any Uops to the reservation station (RS) for the current thread",
},
{
	.name = "arith.divider_active",
	.event = "umask=0x9,period=2000003,cmask=1,event=0x14",
	.desc = "Cycles when divide unit is busy executing divide or square root operations",
	.topic = "pipeline",
	.long_desc = "Counts cycles when divide unit is busy executing divide or square root operations. Accounts for integer and floating-point operations",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This is an architectural event that counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling. For this reason, this event may have a changing ratio with regards to wall clock time",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=25003,event=0x3c",
	.desc = "Core crystal clock cycles when the thread is unhalted",
	.topic = "pipeline",
	.long_desc = "Counts core crystal clock cycles when the thread is unhalted",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=25003,event=0x3c",
	.desc = "Core crystal clock cycles when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
	.long_desc = "Counts Core crystal clock cycles when current thread is unhalted and the other thread is halted",
},
{
	.name = "load_hit_prefetch.swpf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Counts the number of demand load dispatches that hit L1D fill buffer (FB) allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "Counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the software prefetch. It can also be incremented by some lock instructions. So it should only be used with profiling so that the locks can be excluded by ASM (Assembly File) inspection of the nearby instructions",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which the reservation station (RS) is empty for this logical processor. This is usually caused when the front-end pipeline runs into stravation periods (e.g. branch mispredictions or i-cache misses)",
},
{
	.name = "rs_events.empty_end",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,edge=1,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty",
	.topic = "pipeline",
	.long_desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to closely sample on front-end latency issues (see the FRONTEND_RETIRED event of designated precise events)",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
	.long_desc = "Counts cycles that the Instruction Length decoder (ILD) stalls occurred due to dynamically changing prefix length of the decoded instruction (by operand size prefix instruction 0x66, address size prefix instruction 0x67 or REX.W for Intel64). Count is proportional to the number of prefixes in a 16B-line. This may result in a three-cycle penalty for each LCP (Length changing prefix) in a 16-byte chunk",
},
{
	.name = "uops_dispatched.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Number of uops executed on port 0",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_dispatched.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Number of uops executed on port 1",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_dispatched.port_2_3",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Number of uops executed on port 2 and 3",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to ports 2 and 3",
},
{
	.name = "uops_dispatched.port_4_9",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Number of uops executed on port 4 and 9",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to ports 5 and 9",
},
{
	.name = "uops_dispatched.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Number of uops executed on port 5",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_dispatched.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Number of uops executed on port 6",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_dispatched.port_7_8",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Number of uops executed on port 7 and 8",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to ports 7 and 8",
},
{
	.name = "resource_stalls.scoreboard",
	.event = "umask=0x2,period=2000003,event=0xa2",
	.desc = "Counts cycles where the pipeline is stalled due to serializing operations",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "Counts allocation stall cycles caused by the store buffer (SB) being full. This counts cycles that the pipeline back-end blocked uop delivery from the front-end",
},
{
	.name = "cycle_activity.cycles_l2_miss",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_total",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "Total execution stalls",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_miss",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_miss",
	.event = "umask=0x8,period=2000003,cmask=8,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_miss",
	.event = "umask=0xc,period=2000003,cmask=12,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_mem_any",
	.event = "umask=0x10,period=2000003,cmask=16,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_mem_any",
	.event = "umask=0x14,period=2000003,cmask=20,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "exe_activity.1_ports_util",
	.event = "umask=0x2,period=2000003,event=0xa6",
	.desc = "Cycles total of 1 uop is executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which a total of 1 uop was executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.2_ports_util",
	.event = "umask=0x4,period=2000003,event=0xa6",
	.desc = "Cycles total of 2 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which a total of 2 uops were executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.bound_on_stores",
	.event = "umask=0x40,period=2000003,cmask=2,event=0xa6",
	.desc = "Cycles where the Store Buffer was full and no loads caused an execution stall",
	.topic = "pipeline",
	.long_desc = "Counts cycles where the Store Buffer was full and no loads caused an execution stall",
},
{
	.name = "exe_activity.exe_bound_0_ports",
	.event = "umask=0x80,period=2000003,event=0xa6",
	.desc = "Cycles where no uops were executed, the Reservation Station was not empty, the Store Buffer was full and there was no outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which no uops were executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
	.long_desc = "Counts the number of uops delivered to the back-end by the LSD(Loop Stream Detector)",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Counts the cycles when at least one uop is delivered by the LSD (Loop-stream detector)",
},
{
	.name = "lsd.cycles_ok",
	.event = "umask=0x1,period=2000003,cmask=5,event=0xa8",
	.desc = "Cycles optimal number of Uops delivered by the LSD, but did not come from the decoder",
	.topic = "pipeline",
	.long_desc = "Counts the cycles when optimal number of uops is delivered by the LSD (Loop-stream detector)",
},
{
	.name = "uops_executed.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Counts the number of uops to be executed per-thread each cycle",
	.topic = "pipeline",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which no uops were dispatched from the Reservation Station (RS) per thread",
},
{
	.name = "uops_executed.cycles_ge_1",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 1 uop was executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_2",
	.event = "umask=0x1,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 2 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_3",
	.event = "umask=0x1,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 3 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_4",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 4 uops were executed per-thread",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core",
	.topic = "pipeline",
	.long_desc = "Counts the number of uops executed from any thread",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Counts cycles when at least 1 micro-op is executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Counts cycles when at least 2 micro-ops are executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Counts cycles when at least 3 micro-ops are executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Counts cycles when at least 4 micro-ops are executed from any thread on physical core",
},
{
	.name = "uops_executed.x87",
	.event = "umask=0x10,period=2000003,event=0xb1",
	.desc = "Counts the number of x87 uops dispatched",
	.topic = "pipeline",
	.long_desc = "Counts the number of x87 uops executed",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter - architectural event",
	.topic = "pipeline",
	.long_desc = "Counts the number of X86 instructions retired - an Architectural PerfMon event. Counting continues during hardware interrupts, traps, and inside interrupt handlers. Notes: INST_RETIRED.ANY is counted by a designated fixed counter freeing up programmable counters to count other events. INST_RETIRED.ANY_P is counted by a programmable counter",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x2,period=2000003,cmask=10,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
	.long_desc = "Counts the number of cycles using always true condition (uops_ret &amp;lt; 16) applied to non PEBS uops retired event",
},
{
	.name = "uops_retired.slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used",
	.topic = "pipeline",
	.long_desc = "Counts the retirement slots used each cycle",
},
{
	.name = "machine_clears.count",
	.event = "umask=0x1,period=100003,cmask=1,edge=1,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
	.long_desc = "Counts the number of machine clears (nukes) of any type",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "Counts self-modifying code (SMC) detected, which causes a machine clear",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "period=400009,event=0xc4",
	.desc = "All branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts all branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.cond_taken",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Taken conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts taken conditional branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts both direct and indirect near call instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts return instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.cond_ntaken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Not taken branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts not taken branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.cond",
	.event = "umask=0x11,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts conditional branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts taken branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Far branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts far branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.indirect",
	.event = "umask=0x80,period=100003,event=0xc4",
	.desc = "All indirect branch instructions retired (excluding RETs. TSX aborts are considered indirect branch) (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts all indirect branch instructions retired (excluding RETs. TSX aborts is considered indirect branch) (Precise event)",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "period=400009,event=0xc5",
	.desc = "All mispredicted branch instructions retired  Supports address when precise (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts all the retired branch instructions that were mispredicted by the processor. A branch misprediction occurs when the processor incorrectly predicts the destination of the branch.  When the misprediction is discovered at execution, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path  Supports address when precise (Precise event)",
},
{
	.name = "br_misp_retired.cond_taken",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "number of branch instructions retired that were mispredicted and taken. Non PEBS  Supports address when precise (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts taken conditional mispredicted branch instructions retired  Supports address when precise (Precise event)",
},
{
	.name = "br_misp_retired.cond",
	.event = "umask=0x11,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired  Supports address when precise (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted conditional branch instructions retired  Supports address when precise (Precise event)",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "Number of near branch instructions retired that were mispredicted and taken  Supports address when precise (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts number of near branch instructions retired that were mispredicted and taken  Supports address when precise (Precise event)",
},
{
	.name = "br_misp_retired.indirect",
	.event = "umask=0x80,period=100003,event=0xc5",
	.desc = "All miss-predicted indirect branch instructions retired (excluding RETs. TSX aborts is considered indirect branch)  Supports address when precise (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts all miss-predicted indirect branch instructions retired (excluding RETs. TSX aborts is considered indirect branch)  Supports address when precise (Precise event)",
},
{
	.name = "misc_retired.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Increments whenever there is an update to the LBR array",
	.topic = "pipeline",
	.long_desc = "Increments when an entry is added to the Last Branch Record (LBR) array (or removed from the array in case of RETURNs in call stack mode). The event requires LBR enable via IA32_DEBUGCTL MSR and branch type selection via MSR_LBR_SELECT",
},
{
	.name = "misc_retired.pause_inst",
	.event = "umask=0x40,period=2000003,event=0xcc",
	.desc = "Number of retired PAUSE instructions",
	.topic = "pipeline",
	.long_desc = "Counts number of retired PAUSE instructions (that do not end up with a VMExit to the VMM; TSX aborted Instructions may be counted)",
},
{
	.name = "baclears.any",
	.event = "umask=0x1,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
	.long_desc = "Counts the number of times the front-end is resteered when it finds a branch instruction in a fetch line. This occurs for the first time a branch instruction is fetched or when the branch is not tracked by the BPU (Branch Prediction Unit) anymore",
},
{
	.name = "cpu_clk_unhalted.distributed",
	.event = "umask=0x2,period=2000003,event=0xec",
	.desc = "Cycle counts are evenly distributed between active threads in the Core",
	.topic = "pipeline",
	.long_desc = "This event distributes cycle counts between active hyperthreads, i.e., those in C0.  A hyperthread becomes inactive when it executes the HLT or MWAIT instructions.  If all other hyperthreads are inactive (or disabled or do not exist), all counts are attributed to this hyperthread. To obtain the full count when the Core is active, sum the counts from each hyperthread",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x8",
	.desc = "Page walks completed due to a demand data load to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Page walks completed due to a demand data load to a 2M/4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 2M/4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x8",
	.desc = "Load miss in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts demand data loads that caused a completed page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels. The page walk can end with or without a fault",
},
{
	.name = "dtlb_load_misses.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x8",
	.desc = "Number of page walks outstanding for a demand load in the PMH each cycle",
	.topic = "virtual memory",
	.long_desc = "Counts the number of page walks outstanding for a demand load in the PMH (Page Miss Handler) each cycle",
},
{
	.name = "dtlb_load_misses.walk_active",
	.event = "umask=0x10,period=100003,cmask=1,event=0x8",
	.desc = "Cycles when at least one PMH is busy with a page walk for a demand load",
	.topic = "virtual memory",
	.long_desc = "Counts cycles when at least one PMH (Page Miss Handler) is busy with a page walk for a demand load",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x20,period=2000003,event=0x8",
	.desc = "Loads that miss the DTLB and hit the STLB",
	.topic = "virtual memory",
	.long_desc = "Counts loads that miss the DTLB (Data TLB) and hit the STLB (Second level TLB)",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Page walks completed due to a demand data store to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x49",
	.desc = "Page walks completed due to a demand data store to a 2M/4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 2M/4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x49",
	.desc = "Store misses in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts demand data stores that caused a completed page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels. The page walk can end with or without a fault",
},
{
	.name = "dtlb_store_misses.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x49",
	.desc = "Number of page walks outstanding for a store in the PMH each cycle",
	.topic = "virtual memory",
	.long_desc = "Counts the number of page walks outstanding for a store in the PMH (Page Miss Handler) each cycle",
},
{
	.name = "dtlb_store_misses.walk_active",
	.event = "umask=0x10,period=100003,cmask=1,event=0x49",
	.desc = "Cycles when at least one PMH is busy with a page walk for a store",
	.topic = "virtual memory",
	.long_desc = "Counts cycles when at least one PMH (Page Miss Handler) is busy with a page walk for a store",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x20,period=100003,event=0x49",
	.desc = "Stores that miss the DTLB and hit the STLB",
	.topic = "virtual memory",
	.long_desc = "Counts stores that miss the DTLB (Data TLB) and hit the STLB (2nd Level TLB)",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (4K)",
	.topic = "virtual memory",
	.long_desc = "Counts completed page walks (4K page size) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB. The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Counts code misses in all ITLB (Instruction TLB) levels that caused a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts completed page walks (2M and 4M page sizes) caused by a code fetch. This implies it missed in the ITLB (Instruction TLB) and further levels of TLB. The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_pending",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Number of page walks outstanding for an outstanding code request in the PMH each cycle",
	.topic = "virtual memory",
	.long_desc = "Counts the number of page walks outstanding for an outstanding code (instruction fetch) request in the PMH (Page Miss Handler) each cycle",
},
{
	.name = "itlb_misses.walk_active",
	.event = "umask=0x10,period=100003,cmask=1,event=0x85",
	.desc = "Cycles when at least one PMH is busy with a page walk for code (instruction fetch) request",
	.topic = "virtual memory",
	.long_desc = "Counts cycles when at least one PMH (Page Miss Handler) is busy with a page walk for a code (instruction fetch) request",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x20,period=100003,event=0x85",
	.desc = "Instruction fetch requests that miss the ITLB and hit the STLB",
	.topic = "virtual memory",
	.long_desc = "Counts instruction fetch requests that miss the ITLB (Instruction TLB) and hit the STLB (Second-level TLB)",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "Counts the number of flushes of the big or small ITLB pages. Counting include both TLB Flush (covering all sets) and TLB Set Clear (set-specific)",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "Counts the number of DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "Counts the number of any STLB flush attempts (such as entire, VPID, PCID, InvPage, CR3 write, etc.)",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_ivybridge[] = {
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0x1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Demand Data Read requests that hit L2 cache",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0x3,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
	.long_desc = "Counts any demand and L1 HW prefetch data load requests to L2",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x4,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "RFO requests that hit L2 cache",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x8,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of store RFO requests that miss the L2 cache",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xc,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 store RFO requests",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0x10,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
	.long_desc = "Number of instruction fetches that hit the L2 cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x20,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
	.long_desc = "Number of instruction fetches that missed the L2 cache",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0x30,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "Counts all L2 code requests",
},
{
	.name = "l2_rqsts.pf_hit",
	.event = "umask=0x40,period=200003,event=0x24",
	.desc = "Requests from the L2 hardware prefetchers that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests that hit L2",
},
{
	.name = "l2_rqsts.pf_miss",
	.event = "umask=0x80,period=200003,event=0x24",
	.desc = "Requests from the L2 hardware prefetchers that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests that missed L2",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xc0,period=200003,event=0x24",
	.desc = "Requests from L2 hardware prefetchers",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests",
},
{
	.name = "l2_store_lock_rqsts.miss",
	.event = "umask=0x1,period=200003,event=0x27",
	.desc = "RFOs that miss cache lines",
	.topic = "cache",
	.long_desc = "RFOs that miss cache lines",
},
{
	.name = "l2_store_lock_rqsts.hit_m",
	.event = "umask=0x8,period=200003,event=0x27",
	.desc = "RFOs that hit cache lines in M state",
	.topic = "cache",
	.long_desc = "RFOs that hit cache lines in M state",
},
{
	.name = "l2_store_lock_rqsts.all",
	.event = "umask=0xf,period=200003,event=0x27",
	.desc = "RFOs that access cache lines in any state",
	.topic = "cache",
	.long_desc = "RFOs that access cache lines in any state",
},
{
	.name = "l2_l1d_wb_rqsts.miss",
	.event = "umask=0x1,period=200003,event=0x28",
	.desc = "Count the number of modified Lines evicted from L1 and missed L2. (Non-rejected WBs from the DCU.)",
	.topic = "cache",
	.long_desc = "Not rejected writebacks that missed LLC",
},
{
	.name = "l2_l1d_wb_rqsts.hit_e",
	.event = "umask=0x4,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in E state",
	.topic = "cache",
	.long_desc = "Not rejected writebacks from L1D to L2 cache lines in E state",
},
{
	.name = "l2_l1d_wb_rqsts.hit_m",
	.event = "umask=0x8,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in M state",
	.topic = "cache",
	.long_desc = "Not rejected writebacks from L1D to L2 cache lines in M state",
},
{
	.name = "l2_l1d_wb_rqsts.all",
	.event = "umask=0xf,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in any state",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed LLC",
	.topic = "cache",
	.long_desc = "This event counts each cache miss condition for references to the last level cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to LLC",
	.topic = "cache",
	.long_desc = "This event counts requests originating from the core that reference a cache line in the last level cache",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss oustandings duration in cycles",
	.topic = "cache",
	.long_desc = "Increments the number of outstanding L1D misses every cycle. Set Cmask = 1 and Edge =1 to count occurrences",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
	.long_desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
	.topic = "cache",
	.long_desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "Counts the number of lines brought into the L1 data cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
	.long_desc = "Offcore outstanding Demand Data Read transactions in SQ to uncore. Set Cmask=1 to count cycles",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,period=2000003,cmask=6,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
	.long_desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Offcore outstanding Demand Code Read transactions in SQ to uncore. Set Cmask=1 to count cycles",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_code_rd",
	.event = "umask=0x2,period=2000003,cmask=1,event=0x60",
	.desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding RFO store transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Offcore outstanding RFO store transactions in SQ to uncore. Set Cmask=1 to count cycles",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Offcore outstanding cacheable data read transactions in SQ to uncore. Set Cmask=1 to count cycles",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
},
{
	.name = "lock_cycles.cache_lock_duration",
	.event = "umask=0x2,period=2000003,event=0x63",
	.desc = "Cycles when L1D is locked",
	.topic = "cache",
	.long_desc = "Cycles in which the L1D is locked",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
	.long_desc = "Demand data read requests sent to uncore",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "Demand code read requests sent to uncore",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "Demand RFO read requests sent to uncore, including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "Data read requests sent to uncore (demand and prefetch)",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Cases when offcore requests buffer cannot take more entries for core",
	.topic = "cache",
	.long_desc = "Cases when offcore requests buffer cannot take more entries for core",
},
{
	.name = "mem_uops_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load uops that miss the STLB. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store uops that miss the STLB. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load uops with locked access. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load uops that split across a cacheline boundary. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load uops. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store uops. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load uops with L1 cache hits as data sources (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load uops with L2 cache hits as data sources (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.llc_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load uops which data sources were data hits in LLC without snoops required (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources following L1 data-cache miss (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Retired load uops with L2 cache misses as data sources (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.llc_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Miss in last-level (L3) cache. Excludes Unknown data-source (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.hit_lfb",
	.event = "umask=0x40,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources were load uops missed L1 but hit FB due to preceding miss to the same cache line with data not ready (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were LLC hit and cross-core snoop missed in on-pkg core cache (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were LLC and cross-core snoop hits in on-pkg core cache (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were HitM responses from shared LLC (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load uops which data sources were hits in LLC without snoops required (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_miss_retired.local_dram",
	.event = "umask=0x1,period=100007,event=0xd3",
	.desc = "Retired load uops which data sources missed LLC but serviced from local dram",
	.topic = "cache",
	.long_desc = "Retired load uops whose data source was local memory (cross-socket snoop not needed or missed)",
},
{
	.name = "l2_trans.demand_data_rd",
	.event = "umask=0x1,period=200003,event=0xf0",
	.desc = "Demand Data Read requests that access L2 cache",
	.topic = "cache",
	.long_desc = "Demand Data Read requests that access L2 cache",
},
{
	.name = "l2_trans.rfo",
	.event = "umask=0x2,period=200003,event=0xf0",
	.desc = "RFO requests that access L2 cache",
	.topic = "cache",
	.long_desc = "RFO requests that access L2 cache",
},
{
	.name = "l2_trans.code_rd",
	.event = "umask=0x4,period=200003,event=0xf0",
	.desc = "L2 cache accesses when fetching instructions",
	.topic = "cache",
	.long_desc = "L2 cache accesses when fetching instructions",
},
{
	.name = "l2_trans.all_pf",
	.event = "umask=0x8,period=200003,event=0xf0",
	.desc = "L2 or LLC HW prefetches that access L2 cache",
	.topic = "cache",
	.long_desc = "Any MLC or LLC HW prefetch accessing L2, including rejects",
},
{
	.name = "l2_trans.l1d_wb",
	.event = "umask=0x10,period=200003,event=0xf0",
	.desc = "L1D writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "L1D writebacks that access L2 cache",
},
{
	.name = "l2_trans.l2_fill",
	.event = "umask=0x20,period=200003,event=0xf0",
	.desc = "L2 fill requests that access L2 cache",
	.topic = "cache",
	.long_desc = "L2 fill requests that access L2 cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "L2 writebacks that access L2 cache",
},
{
	.name = "l2_trans.all_requests",
	.event = "umask=0x80,period=200003,event=0xf0",
	.desc = "Transactions accessing L2 pipe",
	.topic = "cache",
	.long_desc = "Transactions accessing L2 pipe",
},
{
	.name = "l2_lines_in.i",
	.event = "umask=0x1,period=100003,event=0xf1",
	.desc = "L2 cache lines in I state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in I state filling L2",
},
{
	.name = "l2_lines_in.s",
	.event = "umask=0x2,period=100003,event=0xf1",
	.desc = "L2 cache lines in S state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in S state filling L2",
},
{
	.name = "l2_lines_in.e",
	.event = "umask=0x4,period=100003,event=0xf1",
	.desc = "L2 cache lines in E state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in E state filling L2",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x7,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines filling L2",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x1,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by demand",
	.topic = "cache",
	.long_desc = "Clean L2 cache lines evicted by demand",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x2,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by demand",
	.topic = "cache",
	.long_desc = "Dirty L2 cache lines evicted by demand",
},
{
	.name = "l2_lines_out.pf_clean",
	.event = "umask=0x4,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by L2 prefetch",
	.topic = "cache",
	.long_desc = "Clean L2 cache lines evicted by the MLC prefetcher",
},
{
	.name = "l2_lines_out.pf_dirty",
	.event = "umask=0x8,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by L2 prefetch",
	.topic = "cache",
	.long_desc = "Dirty L2 cache lines evicted by the MLC prefetcher",
},
{
	.name = "l2_lines_out.dirty_all",
	.event = "umask=0xa,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines filling the L2",
	.topic = "cache",
	.long_desc = "Dirty L2 cache lines filling the L2",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Split locks in SQ",
	.topic = "cache",
},
{
	.name = "offcore_response.all_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0244",
	.desc = "Counts all demand & prefetch code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_code_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0244",
	.desc = "Counts demand & prefetch code reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0091",
	.desc = "Counts all demand & prefetch data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0122",
	.desc = "Counts all demand & prefetch RFOs that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0122",
	.desc = "Counts demand & prefetch RFOs that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10008",
	.desc = "Counts all writebacks from the core to the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0004",
	.desc = "Counts all demand code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0004",
	.desc = "Counts demand code reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0001",
	.desc = "Counts all demand data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0001",
	.desc = "Counts demand data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0001",
	.desc = "Counts demand data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0001",
	.desc = "Counts demand data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0002",
	.desc = "Counts all demand data writes (RFOs) that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0002",
	.desc = "Counts demand data writes (RFOs) that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0002",
	.desc = "Counts demand data writes (RFOs) that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x18000",
	.desc = "Counts miscellaneous accesses that include port i/o, MMIO and uncacheable memory accesses. It also includes L2 hints sent to LLC to keep a line from being evicted out of the core caches",
	.topic = "cache",
},
{
	.name = "offcore_response.split_lock_uc_lock.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10400",
	.desc = "Counts requests where the address of an atomic lock instruction spans a cache line boundary or the lock instruction is executed on uncacheable address",
	.topic = "cache",
},
{
	.name = "offcore_response.streaming_stores.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10800",
	.desc = "Counts non-temporal stores",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010001",
	.desc = "Counts all demand data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010002",
	.desc = "Counts all demand rfo's",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000105B3",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010122",
	.desc = "Counts all demand & prefetch prefetch RFOs",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000107F7",
	.desc = "Counts all data/code/rfo references (demand & prefetch)",
	.topic = "cache",
},
{
	.name = "fp_comp_ops_exe.x87",
	.event = "umask=0x1,period=2000003,event=0x10",
	.desc = "Number of FP Computational Uops Executed this cycle. The number of FADD, FSUB, FCOM, FMULs, integer MULsand IMULs, FDIVs, FPREMs, FSQRTS, integer DIVs, and IDIVs. This event does not distinguish an FADD used in the middle of a transcendental flow from a s",
	.topic = "floating point",
	.long_desc = "Counts number of X87 uops executed",
},
{
	.name = "fp_comp_ops_exe.sse_packed_double",
	.event = "umask=0x10,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational packed double-precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Number of SSE* or AVX-128 FP Computational packed double-precision uops issued this cycle",
},
{
	.name = "fp_comp_ops_exe.sse_scalar_single",
	.event = "umask=0x20,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational scalar single-precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Number of SSE* or AVX-128 FP Computational scalar single-precision uops issued this cycle",
},
{
	.name = "fp_comp_ops_exe.sse_packed_single",
	.event = "umask=0x40,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational packed single-precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Number of SSE* or AVX-128 FP Computational packed single-precision uops issued this cycle",
},
{
	.name = "fp_comp_ops_exe.sse_scalar_double",
	.event = "umask=0x80,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational scalar double-precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Counts number of SSE* or AVX-128 double precision FP scalar uops executed",
},
{
	.name = "simd_fp_256.packed_single",
	.event = "umask=0x1,period=2000003,event=0x11",
	.desc = "number of GSSE-256 Computational FP single precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Counts 256-bit packed single-precision floating-point instructions",
},
{
	.name = "simd_fp_256.packed_double",
	.event = "umask=0x2,period=2000003,event=0x11",
	.desc = "number of AVX-256 Computational FP double precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Counts 256-bit packed double-precision floating-point instructions",
},
{
	.name = "other_assists.avx_store",
	.event = "umask=0x8,period=100003,event=0xc1",
	.desc = "Number of GSSE memory assist for stores. GSSE microcode assist is being invoked whenever the hardware is unable to properly handle GSSE-256b operations",
	.topic = "floating point",
	.long_desc = "Number of assists associated with 256-bit AVX store operations",
},
{
	.name = "other_assists.avx_to_sse",
	.event = "umask=0x10,period=100003,event=0xc1",
	.desc = "Number of transitions from AVX-256 to legacy SSE when penalty applicable",
	.topic = "floating point",
},
{
	.name = "other_assists.sse_to_avx",
	.event = "umask=0x20,period=100003,event=0xc1",
	.desc = "Number of transitions from SSE to AVX-256 when penalty applicable",
	.topic = "floating point",
},
{
	.name = "fp_assist.x87_output",
	.event = "umask=0x2,period=100003,event=0xca",
	.desc = "Number of X87 assists due to output value",
	.topic = "floating point",
	.long_desc = "Number of X87 FP assists due to output values",
},
{
	.name = "fp_assist.x87_input",
	.event = "umask=0x4,period=100003,event=0xca",
	.desc = "Number of X87 assists due to input value",
	.topic = "floating point",
	.long_desc = "Number of X87 FP assists due to input values",
},
{
	.name = "fp_assist.simd_output",
	.event = "umask=0x8,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to Output values",
	.topic = "floating point",
	.long_desc = "Number of SIMD FP assists due to output values",
},
{
	.name = "fp_assist.simd_input",
	.event = "umask=0x10,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to input values",
	.topic = "floating point",
	.long_desc = "Number of SIMD FP assists due to input values",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,period=100003,cmask=1,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
	.long_desc = "Cycles with any input/output SSE* or FP assists",
},
{
	.name = "idq.empty",
	.event = "umask=0x2,period=2000003,event=0x79",
	.desc = "Instruction Decode Queue (IDQ) empty cycles",
	.topic = "frontend",
	.long_desc = "Counts cycles the IDQ is empty",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ from MITE path. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Increment each cycle. # of uops delivered to IDQ from DSB path. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
},
{
	.name = "idq.ms_dsb_uops",
	.event = "umask=0x10,period=2000003,event=0x79",
	.desc = "Uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ when MS_busy by DSB. Set Cmask = 1 to count cycles. Add Edge=1 to count # of delivery",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
},
{
	.name = "idq.ms_dsb_occur",
	.event = "umask=0x10,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts cycles DSB is delivered four uops. Set Cmask = 4",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts cycles DSB is delivered at least one uops. Set Cmask = 1",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ when MS_busy by MITE. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts cycles MITE is delivered four uops. Set Cmask = 4",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts cycles MITE is delivered at least one uops. Set Cmask = 1",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ from MS by either DSB or MITE. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
},
{
	.name = "idq.ms_switches",
	.event = "umask=0x30,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
	.long_desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
},
{
	.name = "idq.mite_all_uops",
	.event = "umask=0x3c,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Number of uops delivered to IDQ from any path",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=2000003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
	.topic = "frontend",
	.long_desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Instruction cache, streaming buffer and victim cache misses",
	.topic = "frontend",
	.long_desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Misses. Includes UC accesses",
},
{
	.name = "icache.ifetch_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code-fetch stalled due to L1 instruction-cache miss or an iTLB miss",
	.topic = "frontend",
	.long_desc = "Cycles where a code-fetch stalled due to L1 instruction-cache miss or an iTLB miss",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Count issue pipeline slots where no uop was delivered from the front end to the back end when there is no back-end stall",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=4,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=3,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=2,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.count",
	.event = "umask=0x1,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switches",
	.topic = "frontend",
	.long_desc = "Number of DSB to MITE switches",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "Cycles DSB to MITE switches caused delay",
},
{
	.name = "dsb_fill.exceed_dsb_lines",
	.event = "umask=0x8,period=2000003,event=0xac",
	.desc = "Cycles when Decode Stream Buffer (DSB) fill encounter more than 3 Decode Stream Buffer (DSB) lines",
	.topic = "frontend",
	.long_desc = "DSB Fill encountered > 3 DSB lines",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "ivb metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "ivb metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "ivb metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "ivb metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "ivb metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "ivb metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "ivb metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "ivb metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "ivb metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Instruction per taken branch",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_taken",
	.metric_name = "IpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Branch instructions per taken branch",
	.topic = "ivb metrics",
	.metric_expr = "br_inst_retired.all_branches / br_inst_retired.near_taken",
	.metric_name = "BpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "ivb metrics",
	.metric_expr = "min( 1 , uops_issued.any / ( (uops_retired.retire_slots / inst_retired.any) * 32 * ( icache.hit + icache.misses ) / 4 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "ivb metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops ) )",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "ivb metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "ivb metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "ivb metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "ivb metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Instructions per Load (lower number means loads are more frequent)",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_loads",
	.metric_name = "IpL",
	.metric_group = "Instruction_Type;L1_Bound",
},
{
	.desc = "Instructions per Store",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_stores",
	.metric_name = "IpS",
	.metric_group = "Instruction_Type;Store_Bound",
},
{
	.desc = "Instructions per Branch",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.all_branches",
	.metric_name = "IpB",
	.metric_group = "Branches;Instruction_Type;Port_5;Port_6",
},
{
	.desc = "Instruction per (near) call",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_call",
	.metric_name = "IpCall",
	.metric_group = "Branches",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "ivb metrics",
	.metric_expr = "(( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / cycles",
	.metric_name = "FLOPc",
	.metric_group = "FLOPS",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "ivb metrics",
	.metric_expr = "(( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "FLOPc_SMT",
	.metric_group = "FLOPS_SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "ivb metrics",
	.metric_expr = "uops_executed.thread / (( cpu@uops_executed.core\\,cmask\\=1@ / 2 ) if #smt_on else uops_executed.cycles_ge_1_uop_exec)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Number of Instructions per non-speculative Branch Misprediction (JEClear)",
	.topic = "ivb metrics",
	.metric_expr = "inst_retired.any / br_misp_retired.all_branches",
	.metric_name = "IpMispredict",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "ivb metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads (in core cycles)",
	.topic = "ivb metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_uops_retired.l1_miss + mem_load_uops_retired.hit_lfb )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least one such miss. Per-thread)",
	.topic = "ivb metrics",
	.metric_expr = "l1d_pend_miss.pending / l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "ivb metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration ) / cycles",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "ivb metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration ) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "Page_Walks_Utilization_SMT",
	.metric_group = "TLB_SMT",
},
{
	.desc = "Average data fill bandwidth to the L1 data cache [GB / sec]",
	.topic = "ivb metrics",
	.metric_expr = "64 * l1d.replacement / 1000000000 / duration_time",
	.metric_name = "L1D_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average data fill bandwidth to the L2 cache [GB / sec]",
	.topic = "ivb metrics",
	.metric_expr = "64 * l2_lines_in.all / 1000000000 / duration_time",
	.metric_name = "L2_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "ivb metrics",
	.metric_expr = "64 * longest_lat_cache.miss / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "L1 cache true misses per kilo instruction for retired demand loads",
	.topic = "ivb metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l1_miss / inst_retired.any",
	.metric_name = "L1MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache true misses per kilo instruction for retired demand loads",
	.topic = "ivb metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache misses per kilo instruction for all request types (including speculative)",
	.topic = "ivb metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache hits per kilo instruction for all request types (including speculative)",
	.topic = "ivb metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2HPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L3 cache true misses per kilo instruction for retired demand loads",
	.topic = "ivb metrics",
	.metric_expr = "1000 * mem_load_uops_retired.llc_miss / inst_retired.any",
	.metric_name = "L3MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "Average CPU Utilization",
	.topic = "ivb metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "ivb metrics",
	.metric_expr = "( (( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / 1000000000 ) / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "ivb metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "ivb metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "ivb metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "ivb metrics",
	.metric_expr = "64 * ( arb@event\\=0x81\\,umask\\=0x1@ + arb@event\\=0x84\\,umask\\=0x1@ ) / 1000000 / duration_time / 1000",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "C3 residency percent per core",
	.topic = "ivb metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "ivb metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "ivb metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "ivb metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "ivb metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "ivb metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "ivb metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "misalign_mem_ref.loads",
	.event = "umask=0x1,period=2000003,event=0x5",
	.desc = "Speculative cache line split load uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "Speculative cache-line split load uops dispatched to L1D",
},
{
	.name = "misalign_mem_ref.stores",
	.event = "umask=0x2,period=2000003,event=0x5",
	.desc = "Speculative cache line split STA uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "Speculative cache-line split Store-address uops dispatched to L1D",
},
{
	.name = "page_walks.llc_miss",
	.event = "umask=0x1,period=100003,event=0xbe",
	.desc = "Number of any page walk that had a miss in LLC",
	.topic = "memory",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Loads with latency value being above 4 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 4 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Loads with latency value being above 8 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 8 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Loads with latency value being above 16 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 16 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Loads with latency value being above 32 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 32 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Loads with latency value being above 64 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 64 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Loads with latency value being above 128 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 128 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Loads with latency value being above 256 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 256 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Loads with latency value being above 512 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 512 (Must be precise)",
},
{
	.name = "mem_trans_retired.precise_store",
	.event = "umask=0x2,period=2000003,event=0xcd",
	.desc = "Sample stores and collect precise store operation via PEBS record. PMC3 only (Must be precise)",
	.topic = "memory",
},
{
	.name = "offcore_response.all_code_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400244",
	.desc = "Counts all demand & prefetch code reads that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400091",
	.desc = "Counts all demand & prefetch data reads that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_reads.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3004003f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400004",
	.desc = "Counts demand code reads that miss the LLC and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400001",
	.desc = "Counts demand data reads that miss the LLC and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in_socket.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x6004001b3",
	.desc = "Counts LLC replacements",
	.topic = "memory",
},
{
	.name = "cpl_cycles.ring0",
	.event = "umask=0x1,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when the thread is in ring 0",
	.topic = "other",
	.long_desc = "Unhalted core cycles when the thread is in ring 0",
},
{
	.name = "cpl_cycles.ring0_trans",
	.event = "umask=0x1,edge=1,period=100007,cmask=1,event=0x5c",
	.desc = "Number of intervals between processor halts while thread is in ring 0",
	.topic = "other",
	.long_desc = "Number of intervals between processor halts while thread is in ring 0",
},
{
	.name = "cpl_cycles.ring123",
	.event = "umask=0x2,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when thread is in rings 1, 2, or 3",
	.topic = "other",
	.long_desc = "Unhalted core cycles when the thread is not in ring 0",
},
{
	.name = "lock_cycles.split_lock_uc_lock_duration",
	.event = "umask=0x1,period=2000003,event=0x63",
	.desc = "Cycles when L1 and L2 are locked due to UC or split lock",
	.topic = "other",
	.long_desc = "Cycles in which the L1D and L2 are locked, due to a UC lock or split lock",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
	.long_desc = "Core cycles when at least one thread on the physical core is not in halt state",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Cases when loads get true Block-on-Store blocking code preventing store forwarding",
	.topic = "pipeline",
	.long_desc = "Loads blocked by overlapping with store buffer that cannot be forwarded",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "This event counts the number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
	.long_desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare on address",
	.topic = "pipeline",
	.long_desc = "False dependencies in MOB due to partial compare on address",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x3,period=2000003,cmask=1,event=0xd",
	.desc = "Number of cycles waiting for the checkpoints in Resource Allocation Table (RAT) to be recovered after Nuke due to all other cases except JEClear (e.g. whenever a ucode assist is needed like SSE exception, memory disambiguation, etc.)",
	.topic = "pipeline",
},
{
	.name = "int_misc.recovery_stalls_count",
	.event = "umask=0x3,edge=1,period=2000003,cmask=1,event=0xd",
	.desc = "Number of occurences waiting for the checkpoints in Resource Allocation Table (RAT) to be recovered after Nuke due to all other cases except JEClear (e.g. whenever a ucode assist is needed like SSE exception, memory disambiguation, etc.)",
	.topic = "pipeline",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x3,any=1,period=2000003,cmask=1,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "Increments each cycle the # of Uops issued by the RAT to RS. Set Cmask = 1, Inv = 1, Any= 1to count stalled cycles of this core",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
	.long_desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for all threads",
	.topic = "pipeline",
	.long_desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for all threads",
},
{
	.name = "uops_issued.flags_merge",
	.event = "umask=0x10,period=2000003,event=0xe",
	.desc = "Number of flags-merge uops being allocated",
	.topic = "pipeline",
	.long_desc = "Number of flags-merge uops allocated. Such uops adds delay",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
	.long_desc = "Number of slow LEA or similar uops allocated. Such uop has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
},
{
	.name = "uops_issued.single_mul",
	.event = "umask=0x40,period=2000003,event=0xe",
	.desc = "Number of Multiply packed/scalar single precision uops allocated",
	.topic = "pipeline",
	.long_desc = "Number of multiply packed/scalar single precision uops allocated",
},
{
	.name = "arith.fpu_div_active",
	.event = "umask=0x1,period=2000003,event=0x14",
	.desc = "Cycles when divider is busy executing divide operations",
	.topic = "pipeline",
	.long_desc = "Cycles that the divider is active, includes INT and FP. Set 'edge =1, cmask=1' to count the number of divides",
},
{
	.name = "arith.fpu_div",
	.event = "umask=0x4,edge=1,period=100003,cmask=1,event=0x14",
	.desc = "Divide operations executed",
	.topic = "pipeline",
	.long_desc = "Divide operations executed",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
	.long_desc = "Core cycles when at least one thread on the physical core is not in halt state",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Increments at the frequency of XCLK (100 MHz) when not halted",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted. (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the thread is unhalted. (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted. (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "Non-SW-prefetch load dispatches that hit fill buffer allocated for S/W prefetch",
},
{
	.name = "load_hit_pre.hw_pf",
	.event = "umask=0x2,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for hardware prefetch",
	.topic = "pipeline",
	.long_desc = "Non-SW-prefetch load dispatches that hit fill buffer allocated for H/W prefetch",
},
{
	.name = "move_elimination.int_eliminated",
	.event = "umask=0x1,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_eliminated",
	.event = "umask=0x2,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.int_not_eliminated",
	.event = "umask=0x4,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_not_eliminated",
	.event = "umask=0x8,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "Cycles the RS is empty for the thread",
},
{
	.name = "rs_events.empty_end",
	.event = "inv=1,umask=0x1,edge=1,period=200003,cmask=1,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000003,event=0x87",
	.desc = "Stall cycles because IQ is full",
	.topic = "pipeline",
	.long_desc = "Stall cycles due to IQ is full",
},
{
	.name = "br_inst_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x88",
	.desc = "Not taken macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "Not taken macro-conditional branches",
},
{
	.name = "br_inst_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired macro-conditional branches",
},
{
	.name = "br_inst_exec.taken_direct_jump",
	.event = "umask=0x82,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
},
{
	.name = "br_inst_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired indirect branches excluding calls and returns",
},
{
	.name = "br_inst_exec.taken_indirect_near_return",
	.event = "umask=0x88,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired indirect branches with return mnemonic",
},
{
	.name = "br_inst_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x88",
	.desc = "Taken speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect calls",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired indirect calls",
},
{
	.name = "br_inst_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x88",
	.desc = "Speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "Speculative and retired macro-conditional branches",
},
{
	.name = "br_inst_exec.all_direct_jmp",
	.event = "umask=0xc2,period=200003,event=0x88",
	.desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
},
{
	.name = "br_inst_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x88",
	.desc = "Speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "Speculative and retired indirect branches excluding calls and returns",
},
{
	.name = "br_inst_exec.all_indirect_near_return",
	.event = "umask=0xc8,period=200003,event=0x88",
	.desc = "Speculative and retired indirect return branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x88",
	.desc = "Speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "Speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x88",
	.desc = "Speculative and retired  branches",
	.topic = "pipeline",
	.long_desc = "Counts all near executed branches (not necessarily retired)",
},
{
	.name = "br_misp_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x89",
	.desc = "Not taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Not taken speculative and retired mispredicted macro conditional branches",
},
{
	.name = "br_misp_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired mispredicted macro conditional branches",
},
{
	.name = "br_misp_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.taken_return_near",
	.event = "umask=0x88,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
},
{
	.name = "br_misp_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect calls",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired mispredicted indirect calls",
},
{
	.name = "br_misp_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Speculative and retired mispredicted macro conditional branches",
},
{
	.name = "br_misp_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x89",
	.desc = "Mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "Mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Counts all near executed branches (not necessarily retired)",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 0",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 0",
},
{
	.name = "uops_dispatched_port.port_0_core",
	.event = "umask=0x1,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 0",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are dispatched to port 0",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 1",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 1",
},
{
	.name = "uops_dispatched_port.port_1_core",
	.event = "umask=0x2,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 1",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are dispatched to port 1",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0xc,period=2000003,event=0xa1",
	.desc = "Cycles per thread when load or STA uops are dispatched to port 2",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 2",
},
{
	.name = "uops_dispatched_port.port_2_core",
	.event = "umask=0xc,any=1,period=2000003,event=0xa1",
	.desc = "Uops dispatched to port 2, loads and stores per core (speculative and retired)",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x30,period=2000003,event=0xa1",
	.desc = "Cycles per thread when load or STA uops are dispatched to port 3",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 3",
},
{
	.name = "uops_dispatched_port.port_3_core",
	.event = "umask=0x30,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when load or STA uops are dispatched to port 3",
	.topic = "pipeline",
	.long_desc = "Cycles per core when load or STA uops are dispatched to port 3",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 4",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 4",
},
{
	.name = "uops_dispatched_port.port_4_core",
	.event = "umask=0x40,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 4",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are dispatched to port 4",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 5",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 5",
},
{
	.name = "uops_dispatched_port.port_5_core",
	.event = "umask=0x80,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 5",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are dispatched to port 5",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
	.long_desc = "Cycles Allocation is stalled due to Resource Related reason",
},
{
	.name = "resource_stalls.rs",
	.event = "umask=0x4,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no eligible RS entry available",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "Cycles stalled due to no store buffers available (not including draining form sync)",
},
{
	.name = "resource_stalls.rob",
	.event = "umask=0x10,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to re-order buffer full",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l2_pending",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Cycles with pending L2 cache miss loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending L2 miss loads. Set AnyThread to count per core",
},
{
	.name = "cycle_activity.cycles_l2_miss",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Cycles while L2 cache miss load* is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_ldm_pending",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Cycles with pending memory loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending memory loads. Set AnyThread to count per core",
},
{
	.name = "cycle_activity.cycles_mem_any",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_no_execute",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "This event increments by 1 for every cycle where there was no execute for this thread",
	.topic = "pipeline",
	.long_desc = "Total execution stalls",
},
{
	.name = "cycle_activity.stalls_total",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "Total execution stalls",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_pending",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Execution stalls due to L2 cache misses",
	.topic = "pipeline",
	.long_desc = "Number of loads missed L2",
},
{
	.name = "cycle_activity.stalls_l2_miss",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Execution stalls while L2 cache miss load* is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_ldm_pending",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Execution stalls due to memory subsystem",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_mem_any",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_pending",
	.event = "umask=0x8,period=2000003,cmask=8,event=0xa3",
	.desc = "Cycles with pending L1 cache miss loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending L1 cache miss loads. Set AnyThread to count per core",
},
{
	.name = "cycle_activity.cycles_l1d_miss",
	.event = "umask=0x8,period=2000003,cmask=8,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_pending",
	.event = "umask=0xc,period=2000003,cmask=12,event=0xa3",
	.desc = "Execution stalls due to L1 data cache misses",
	.topic = "pipeline",
	.long_desc = "Execution stalls due to L1 data cache miss loads. Set Cmask=0CH",
},
{
	.name = "cycle_activity.stalls_l1d_miss",
	.event = "umask=0xc,period=2000003,cmask=12,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
},
{
	.name = "uops_executed.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Counts the number of uops to be executed per-thread each cycle",
	.topic = "pipeline",
	.long_desc = "Counts total number of uops to be executed per-thread each cycle. Set Cmask = 1, INV =1 to count stall cycles",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 1 uop was executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 2 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 3 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 4 uops were executed per-thread",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core",
	.topic = "pipeline",
	.long_desc = "Counts total number of uops to be executed per-core each cycle",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles with no micro-ops executed from any thread on physical core",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter   - architectural event",
	.topic = "pipeline",
	.long_desc = "Number of instructions at retirement",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution (Must be precise)",
},
{
	.name = "other_assists.any_wb_assist",
	.event = "umask=0x80,period=100003,event=0xc1",
	.desc = "Number of times any microcode assist is invoked by HW upon uop writeback",
	.topic = "pipeline",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Retired uops (Precise event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=10,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used (Precise event)",
	.topic = "pipeline",
},
{
	.name = "machine_clears.count",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "Number of self-modifying-code machine clears detected",
},
{
	.name = "machine_clears.maskmov",
	.event = "umask=0x20,period=100003,event=0xc3",
	.desc = "This event counts the number of executed Intel AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
	.topic = "pipeline",
	.long_desc = "Counts the number of executed AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Branch instructions at retirement",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call_r3",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect macro near call instructions retired (captured in ring 3) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Not taken branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Counts the number of not taken branch instructions retired",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Far branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Number of far branches retired",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Mispredicted branch instructions at retirement",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "number of near branch instructions retired that were mispredicted and taken (Precise event)",
	.topic = "pipeline",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Count cases of saving new LBR",
	.topic = "pipeline",
	.long_desc = "Count cases of saving new LBR records by hardware",
},
{
	.name = "baclears.any",
	.event = "umask=0x1f,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
	.long_desc = "Number of front end re-steers due to BPU misprediction",
},
{
	.name = "unc_cbo_xsnp_response.miss",
	.event = "umask=0x01,event=0x22",
	.desc = "Unit: uncore_cbox A snoop misses in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop misses in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.inval",
	.event = "umask=0x02,event=0x22",
	.desc = "Unit: uncore_cbox A snoop invalidates a non-modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop invalidates a non-modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hit",
	.event = "umask=0x04,event=0x22",
	.desc = "Unit: uncore_cbox A snoop hits a non-modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop hits a non-modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hitm",
	.event = "umask=0x08,event=0x22",
	.desc = "Unit: uncore_cbox A snoop hits a modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop hits a modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.inval_m",
	.event = "umask=0x10,event=0x22",
	.desc = "Unit: uncore_cbox A snoop invalidates a modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop invalidates a modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.external_filter",
	.event = "umask=0x20,event=0x22",
	.desc = "Unit: uncore_cbox Filter on cross-core snoops initiated by this Cbox due to external snoop request",
	.topic = "uncore",
	.long_desc = "Filter on cross-core snoops initiated by this Cbox due to external snoop request",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.xcore_filter",
	.event = "umask=0x40,event=0x22",
	.desc = "Unit: uncore_cbox Filter on cross-core snoops initiated by this Cbox due to processor core memory request",
	.topic = "uncore",
	.long_desc = "Filter on cross-core snoops initiated by this Cbox due to processor core memory request",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.eviction_filter",
	.event = "umask=0x80,event=0x22",
	.desc = "Unit: uncore_cbox Filter on cross-core snoops initiated by this Cbox due to LLC eviction",
	.topic = "uncore",
	.long_desc = "Filter on cross-core snoops initiated by this Cbox due to LLC eviction",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.m",
	.event = "umask=0x01,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.e",
	.event = "umask=0x02,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in E-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in E-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.s",
	.event = "umask=0x04,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in S-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.i",
	.event = "umask=0x08,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_filter",
	.event = "umask=0x10,event=0x34",
	.desc = "Unit: uncore_cbox Filter on processor core initiated cacheable read requests",
	.topic = "uncore",
	.long_desc = "Filter on processor core initiated cacheable read requests",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_filter",
	.event = "umask=0x20,event=0x34",
	.desc = "Unit: uncore_cbox Filter on processor core initiated cacheable write requests",
	.topic = "uncore",
	.long_desc = "Filter on processor core initiated cacheable write requests",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.extsnp_filter",
	.event = "umask=0x40,event=0x34",
	.desc = "Unit: uncore_cbox Filter on external snoop requests",
	.topic = "uncore",
	.long_desc = "Filter on external snoop requests",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_request_filter",
	.event = "umask=0x80,event=0x34",
	.desc = "Unit: uncore_cbox Filter on any IRQ or IPQ initiated requests including uncacheable, non-coherent requests",
	.topic = "uncore",
	.long_desc = "Filter on any IRQ or IPQ initiated requests including uncacheable, non-coherent requests",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_arb_trk_occupancy.all",
	.event = "umask=0x01,event=0x80",
	.desc = "Unit: uncore_arb Counts cycles weighted by the number of requests waiting for data returning from the memory controller. Accounts for coherent and non-coherent requests initiated by IA cores, processor graphic units, or LLC",
	.topic = "uncore",
	.long_desc = "Counts cycles weighted by the number of requests waiting for data returning from the memory controller. Accounts for coherent and non-coherent requests initiated by IA cores, processor graphic units, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.all",
	.event = "umask=0x01,event=0x81",
	.desc = "Unit: uncore_arb Counts the number of coherent and in-coherent requests initiated by IA cores, processor graphic units, or LLC",
	.topic = "uncore",
	.long_desc = "Counts the number of coherent and in-coherent requests initiated by IA cores, processor graphic units, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.writes",
	.event = "umask=0x20,event=0x81",
	.desc = "Unit: uncore_arb Counts the number of allocated write entries, include full, partial, and LLC evictions",
	.topic = "uncore",
	.long_desc = "Counts the number of allocated write entries, include full, partial, and LLC evictions",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.evictions",
	.event = "umask=0x80,event=0x81",
	.desc = "Unit: uncore_arb Counts the number of LLC evictions allocated",
	.topic = "uncore",
	.long_desc = "Counts the number of LLC evictions allocated",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_coh_trk_occupancy.all",
	.event = "umask=0x01,event=0x83",
	.desc = "Unit: uncore_arb Cycles weighted by number of requests pending in Coherency Tracker",
	.topic = "uncore",
	.long_desc = "Cycles weighted by number of requests pending in Coherency Tracker",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_coh_trk_requests.all",
	.event = "umask=0x01,event=0x84",
	.desc = "Unit: uncore_arb Number of requests allocated in Coherency Tracker",
	.topic = "uncore",
	.long_desc = "Number of requests allocated in Coherency Tracker",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_occupancy.cycles_with_any_request",
	.event = "umask=0x01,cmask=1,event=0x80",
	.desc = "Unit: uncore_arb Cycles with at least one request outstanding is waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.topic = "uncore",
	.long_desc = "Cycles with at least one request outstanding is waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_occupancy.cycles_over_half_full",
	.event = "umask=0x01,cmask=10,event=0x80",
	.desc = "Unit: uncore_arb Cycles with at least half of the requests outstanding are waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.topic = "uncore",
	.long_desc = "Cycles with at least half of the requests outstanding are waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_clock.socket",
	.event = "umask=0x01,event=0",
	.desc = "Unit: uncore_arb This 48-bit fixed counter counts the UCLK cycles",
	.topic = "uncore",
	.long_desc = "This 48-bit fixed counter counts the UCLK cycles",
	.pmu = "uncore_arb",
},
{
	.name = "unc_cbo_cache_lookup.es",
	.event = "umask=0x06,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in E-state or S-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in E-state or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x81,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes an page walk of any page size",
	.topic = "virtual memory",
	.long_desc = "Misses in all TLB levels that cause a page walk of any page size from demand loads",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0x82,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes of any page size",
	.topic = "virtual memory",
	.long_desc = "Misses in all TLB levels that caused page walk completed of any size by demand loads",
},
{
	.name = "dtlb_load_misses.walk_duration",
	.event = "umask=0x84,period=2000003,event=0x8",
	.desc = "Demand load cycles page miss handler (PMH) is busy with this walk",
	.topic = "virtual memory",
	.long_desc = "Cycle PMH is busy with a walk due to demand loads",
},
{
	.name = "dtlb_load_misses.large_page_walk_completed",
	.event = "umask=0x88,period=100003,event=0x8",
	.desc = "Page walk for a large page completed for Demand load",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Miss in all TLB levels causes a page walk of any page size (4K/2M/4M/1G)",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks",
	.topic = "virtual memory",
	.long_desc = "Miss in all TLB levels causes a page walk that completes of any page size (4K/2M/4M/1G)",
},
{
	.name = "dtlb_store_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x49",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "Cycles PMH is busy with this walk",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x49",
	.desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Cycle count for an Extended Page table walk.  The Extended Page Directory cache is used by Virtual Machine operating systems while the guest operating systems use the standard TLB caches",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x4,period=100003,event=0x5f",
	.desc = "Load operations that miss the first DTLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts load operations that missed 1st level DTLB but hit the 2nd level",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Misses in all ITLB levels that cause page walks",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Misses in all ITLB levels that cause completed page walks",
	.topic = "virtual memory",
	.long_desc = "Misses in all ITLB levels that cause completed page walks",
},
{
	.name = "itlb_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x85",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "Cycle PMH is busy with a walk",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Operations that miss the first ITLB level but hit the second and do not cause any page walks",
	.topic = "virtual memory",
	.long_desc = "Number of cache load STLB hits. No page walk",
},
{
	.name = "itlb_misses.large_page_walk_completed",
	.event = "umask=0x80,period=100003,event=0x85",
	.desc = "Completed page walks in ITLB due to STLB load misses for large pages",
	.topic = "virtual memory",
	.long_desc = "Completed page walks in ITLB due to STLB load misses for large pages",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "Counts the number of ITLB flushes, includes 4k/2M/4M pages",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "Count number of STLB flush attempts",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_ivytown[] = {
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0x1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Demand Data Read requests that hit L2 cache",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0x3,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
	.long_desc = "Counts any demand and L1 HW prefetch data load requests to L2",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x4,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "RFO requests that hit L2 cache",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x8,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of store RFO requests that miss the L2 cache",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xc,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 store RFO requests",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0x10,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
	.long_desc = "Number of instruction fetches that hit the L2 cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x20,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
	.long_desc = "Number of instruction fetches that missed the L2 cache",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0x30,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "Counts all L2 code requests",
},
{
	.name = "l2_rqsts.pf_hit",
	.event = "umask=0x40,period=200003,event=0x24",
	.desc = "Requests from the L2 hardware prefetchers that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests that hit L2",
},
{
	.name = "l2_rqsts.pf_miss",
	.event = "umask=0x80,period=200003,event=0x24",
	.desc = "Requests from the L2 hardware prefetchers that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests that missed L2",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xc0,period=200003,event=0x24",
	.desc = "Requests from L2 hardware prefetchers",
	.topic = "cache",
	.long_desc = "Counts all L2 HW prefetcher requests",
},
{
	.name = "l2_store_lock_rqsts.miss",
	.event = "umask=0x1,period=200003,event=0x27",
	.desc = "RFOs that miss cache lines",
	.topic = "cache",
	.long_desc = "RFOs that miss cache lines",
},
{
	.name = "l2_store_lock_rqsts.hit_m",
	.event = "umask=0x8,period=200003,event=0x27",
	.desc = "RFOs that hit cache lines in M state",
	.topic = "cache",
	.long_desc = "RFOs that hit cache lines in M state",
},
{
	.name = "l2_store_lock_rqsts.all",
	.event = "umask=0xf,period=200003,event=0x27",
	.desc = "RFOs that access cache lines in any state",
	.topic = "cache",
	.long_desc = "RFOs that access cache lines in any state",
},
{
	.name = "l2_l1d_wb_rqsts.miss",
	.event = "umask=0x1,period=200003,event=0x28",
	.desc = "Count the number of modified Lines evicted from L1 and missed L2. (Non-rejected WBs from the DCU.)",
	.topic = "cache",
	.long_desc = "Not rejected writebacks that missed LLC",
},
{
	.name = "l2_l1d_wb_rqsts.hit_e",
	.event = "umask=0x4,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in E state",
	.topic = "cache",
	.long_desc = "Not rejected writebacks from L1D to L2 cache lines in E state",
},
{
	.name = "l2_l1d_wb_rqsts.hit_m",
	.event = "umask=0x8,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in M state",
	.topic = "cache",
	.long_desc = "Not rejected writebacks from L1D to L2 cache lines in M state",
},
{
	.name = "l2_l1d_wb_rqsts.all",
	.event = "umask=0xf,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in any state",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed LLC",
	.topic = "cache",
	.long_desc = "This event counts each cache miss condition for references to the last level cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to LLC",
	.topic = "cache",
	.long_desc = "This event counts requests originating from the core that reference a cache line in the last level cache",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss oustandings duration in cycles",
	.topic = "cache",
	.long_desc = "Increments the number of outstanding L1D misses every cycle. Set Cmask = 1 and Edge =1 to count occurrences",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
	.long_desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
	.topic = "cache",
	.long_desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "Counts the number of lines brought into the L1 data cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
	.long_desc = "Offcore outstanding Demand Data Read transactions in SQ to uncore. Set Cmask=1 to count cycles",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,period=2000003,cmask=6,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
	.long_desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Offcore outstanding Demand Code Read transactions in SQ to uncore. Set Cmask=1 to count cycles",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_code_rd",
	.event = "umask=0x2,period=2000003,cmask=1,event=0x60",
	.desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding RFO store transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Offcore outstanding RFO store transactions in SQ to uncore. Set Cmask=1 to count cycles",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Offcore outstanding cacheable data read transactions in SQ to uncore. Set Cmask=1 to count cycles",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
},
{
	.name = "lock_cycles.cache_lock_duration",
	.event = "umask=0x2,period=2000003,event=0x63",
	.desc = "Cycles when L1D is locked",
	.topic = "cache",
	.long_desc = "Cycles in which the L1D is locked",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
	.long_desc = "Demand data read requests sent to uncore",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "Demand code read requests sent to uncore",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "Demand RFO read requests sent to uncore, including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "Data read requests sent to uncore (demand and prefetch)",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Cases when offcore requests buffer cannot take more entries for core",
	.topic = "cache",
	.long_desc = "Cases when offcore requests buffer cannot take more entries for core",
},
{
	.name = "mem_uops_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load uops that miss the STLB. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store uops that miss the STLB. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load uops with locked access. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load uops that split across a cacheline boundary. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load uops. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store uops. (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load uops with L1 cache hits as data sources (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load uops with L2 cache hits as data sources (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.llc_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load uops which data sources were data hits in LLC without snoops required (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources following L1 data-cache miss (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Retired load uops with L2 cache misses as data sources (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.llc_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Miss in last-level (L3) cache. Excludes Unknown data-source (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.hit_lfb",
	.event = "umask=0x40,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources were load uops missed L1 but hit FB due to preceding miss to the same cache line with data not ready (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were LLC hit and cross-core snoop missed in on-pkg core cache (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were LLC and cross-core snoop hits in on-pkg core cache (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were HitM responses from shared LLC (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load uops which data sources were hits in LLC without snoops required (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_miss_retired.local_dram",
	.event = "umask=0x3,period=100007,event=0xd3",
	.desc = "Retired load uops whose data source was local DRAM (Snoop not needed, Snoop Miss, or Snoop Hit data not forwarded)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_miss_retired.remote_dram",
	.event = "umask=0xc,period=100007,event=0xd3",
	.desc = "Retired load uops whose data source was remote DRAM (Snoop not needed, Snoop Miss, or Snoop Hit data not forwarded)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_miss_retired.remote_hitm",
	.event = "umask=0x10,period=100007,event=0xd3",
	.desc = "Remote cache HITM",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_miss_retired.remote_fwd",
	.event = "umask=0x20,period=100007,event=0xd3",
	.desc = "Data forwarded from remote cache",
	.topic = "cache",
},
{
	.name = "l2_trans.demand_data_rd",
	.event = "umask=0x1,period=200003,event=0xf0",
	.desc = "Demand Data Read requests that access L2 cache",
	.topic = "cache",
	.long_desc = "Demand Data Read requests that access L2 cache",
},
{
	.name = "l2_trans.rfo",
	.event = "umask=0x2,period=200003,event=0xf0",
	.desc = "RFO requests that access L2 cache",
	.topic = "cache",
	.long_desc = "RFO requests that access L2 cache",
},
{
	.name = "l2_trans.code_rd",
	.event = "umask=0x4,period=200003,event=0xf0",
	.desc = "L2 cache accesses when fetching instructions",
	.topic = "cache",
	.long_desc = "L2 cache accesses when fetching instructions",
},
{
	.name = "l2_trans.all_pf",
	.event = "umask=0x8,period=200003,event=0xf0",
	.desc = "L2 or LLC HW prefetches that access L2 cache",
	.topic = "cache",
	.long_desc = "Any MLC or LLC HW prefetch accessing L2, including rejects",
},
{
	.name = "l2_trans.l1d_wb",
	.event = "umask=0x10,period=200003,event=0xf0",
	.desc = "L1D writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "L1D writebacks that access L2 cache",
},
{
	.name = "l2_trans.l2_fill",
	.event = "umask=0x20,period=200003,event=0xf0",
	.desc = "L2 fill requests that access L2 cache",
	.topic = "cache",
	.long_desc = "L2 fill requests that access L2 cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "L2 writebacks that access L2 cache",
},
{
	.name = "l2_trans.all_requests",
	.event = "umask=0x80,period=200003,event=0xf0",
	.desc = "Transactions accessing L2 pipe",
	.topic = "cache",
	.long_desc = "Transactions accessing L2 pipe",
},
{
	.name = "l2_lines_in.i",
	.event = "umask=0x1,period=100003,event=0xf1",
	.desc = "L2 cache lines in I state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in I state filling L2",
},
{
	.name = "l2_lines_in.s",
	.event = "umask=0x2,period=100003,event=0xf1",
	.desc = "L2 cache lines in S state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in S state filling L2",
},
{
	.name = "l2_lines_in.e",
	.event = "umask=0x4,period=100003,event=0xf1",
	.desc = "L2 cache lines in E state filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines in E state filling L2",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x7,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "L2 cache lines filling L2",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x1,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by demand",
	.topic = "cache",
	.long_desc = "Clean L2 cache lines evicted by demand",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x2,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by demand",
	.topic = "cache",
	.long_desc = "Dirty L2 cache lines evicted by demand",
},
{
	.name = "l2_lines_out.pf_clean",
	.event = "umask=0x4,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by L2 prefetch",
	.topic = "cache",
	.long_desc = "Clean L2 cache lines evicted by the MLC prefetcher",
},
{
	.name = "l2_lines_out.pf_dirty",
	.event = "umask=0x8,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by L2 prefetch",
	.topic = "cache",
	.long_desc = "Dirty L2 cache lines evicted by the MLC prefetcher",
},
{
	.name = "l2_lines_out.dirty_all",
	.event = "umask=0xa,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines filling the L2",
	.topic = "cache",
	.long_desc = "Dirty L2 cache lines filling the L2",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Split locks in SQ",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and sibling core snoop returned a clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0090",
	.desc = "Counts all prefetch data reads that hit the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and sibling core snoop returned a clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC and sibling core snoop returned a clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10008",
	.desc = "Counts all writebacks from the core to the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0004",
	.desc = "Counts all demand code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0001",
	.desc = "Counts all demand data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0001",
	.desc = "Counts demand data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0001",
	.desc = "Counts demand data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0001",
	.desc = "Counts demand data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0001",
	.desc = "Counts demand data reads that hit in the LLC and sibling core snoop returned a clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0002",
	.desc = "Counts demand data writes (RFOs) that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.other.lru_hints",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x803c8000",
	.desc = "Counts L2 hints sent to LLC to keep a line from being evicted out of the core caches",
	.topic = "cache",
},
{
	.name = "offcore_response.other.portio_mmio_uc",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x23ffc08000",
	.desc = "Counts miscellaneous accesses that include port i/o, MMIO and uncacheable memory accesses",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0040",
	.desc = "Counts all prefetch (that bring data to L2) code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0200",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.split_lock_uc_lock.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10400",
	.desc = "Counts requests where the address of an atomic lock instruction spans a cache line boundary or the lock instruction is executed on uncacheable address",
	.topic = "cache",
},
{
	.name = "offcore_response.streaming_stores.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10800",
	.desc = "Counts non-temporal stores",
	.topic = "cache",
},
{
	.name = "fp_comp_ops_exe.x87",
	.event = "umask=0x1,period=2000003,event=0x10",
	.desc = "Number of FP Computational Uops Executed this cycle. The number of FADD, FSUB, FCOM, FMULs, integer MULsand IMULs, FDIVs, FPREMs, FSQRTS, integer DIVs, and IDIVs. This event does not distinguish an FADD used in the middle of a transcendental flow from a s",
	.topic = "floating point",
	.long_desc = "Counts number of X87 uops executed",
},
{
	.name = "fp_comp_ops_exe.sse_packed_double",
	.event = "umask=0x10,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational packed double-precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Number of SSE* or AVX-128 FP Computational packed double-precision uops issued this cycle",
},
{
	.name = "fp_comp_ops_exe.sse_scalar_single",
	.event = "umask=0x20,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational scalar single-precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Number of SSE* or AVX-128 FP Computational scalar single-precision uops issued this cycle",
},
{
	.name = "fp_comp_ops_exe.sse_packed_single",
	.event = "umask=0x40,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational packed single-precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Number of SSE* or AVX-128 FP Computational packed single-precision uops issued this cycle",
},
{
	.name = "fp_comp_ops_exe.sse_scalar_double",
	.event = "umask=0x80,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational scalar double-precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Counts number of SSE* or AVX-128 double precision FP scalar uops executed",
},
{
	.name = "simd_fp_256.packed_single",
	.event = "umask=0x1,period=2000003,event=0x11",
	.desc = "number of GSSE-256 Computational FP single precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Counts 256-bit packed single-precision floating-point instructions",
},
{
	.name = "simd_fp_256.packed_double",
	.event = "umask=0x2,period=2000003,event=0x11",
	.desc = "number of AVX-256 Computational FP double precision uops issued this cycle",
	.topic = "floating point",
	.long_desc = "Counts 256-bit packed double-precision floating-point instructions",
},
{
	.name = "other_assists.avx_store",
	.event = "umask=0x8,period=100003,event=0xc1",
	.desc = "Number of GSSE memory assist for stores. GSSE microcode assist is being invoked whenever the hardware is unable to properly handle GSSE-256b operations",
	.topic = "floating point",
	.long_desc = "Number of assists associated with 256-bit AVX store operations",
},
{
	.name = "other_assists.avx_to_sse",
	.event = "umask=0x10,period=100003,event=0xc1",
	.desc = "Number of transitions from AVX-256 to legacy SSE when penalty applicable",
	.topic = "floating point",
},
{
	.name = "other_assists.sse_to_avx",
	.event = "umask=0x20,period=100003,event=0xc1",
	.desc = "Number of transitions from SSE to AVX-256 when penalty applicable",
	.topic = "floating point",
},
{
	.name = "fp_assist.x87_output",
	.event = "umask=0x2,period=100003,event=0xca",
	.desc = "Number of X87 assists due to output value",
	.topic = "floating point",
	.long_desc = "Number of X87 FP assists due to output values",
},
{
	.name = "fp_assist.x87_input",
	.event = "umask=0x4,period=100003,event=0xca",
	.desc = "Number of X87 assists due to input value",
	.topic = "floating point",
	.long_desc = "Number of X87 FP assists due to input values",
},
{
	.name = "fp_assist.simd_output",
	.event = "umask=0x8,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to Output values",
	.topic = "floating point",
	.long_desc = "Number of SIMD FP assists due to output values",
},
{
	.name = "fp_assist.simd_input",
	.event = "umask=0x10,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to input values",
	.topic = "floating point",
	.long_desc = "Number of SIMD FP assists due to input values",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,period=100003,cmask=1,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
	.long_desc = "Cycles with any input/output SSE* or FP assists",
},
{
	.name = "idq.empty",
	.event = "umask=0x2,period=2000003,event=0x79",
	.desc = "Instruction Decode Queue (IDQ) empty cycles",
	.topic = "frontend",
	.long_desc = "Counts cycles the IDQ is empty",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ from MITE path. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Increment each cycle. # of uops delivered to IDQ from DSB path. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
},
{
	.name = "idq.ms_dsb_uops",
	.event = "umask=0x10,period=2000003,event=0x79",
	.desc = "Uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ when MS_busy by DSB. Set Cmask = 1 to count cycles. Add Edge=1 to count # of delivery",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
},
{
	.name = "idq.ms_dsb_occur",
	.event = "umask=0x10,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts cycles DSB is delivered four uops. Set Cmask = 4",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts cycles DSB is delivered at least one uops. Set Cmask = 1",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ when MS_busy by MITE. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts cycles MITE is delivered four uops. Set Cmask = 4",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts cycles MITE is delivered at least one uops. Set Cmask = 1",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Increment each cycle # of uops delivered to IDQ from MS by either DSB or MITE. Set Cmask = 1 to count cycles",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
},
{
	.name = "idq.ms_switches",
	.event = "umask=0x30,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
	.long_desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
},
{
	.name = "idq.mite_all_uops",
	.event = "umask=0x3c,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Number of uops delivered to IDQ from any path",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=2000003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
	.topic = "frontend",
	.long_desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Instruction cache, streaming buffer and victim cache misses",
	.topic = "frontend",
	.long_desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Misses. Includes UC accesses",
},
{
	.name = "icache.ifetch_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code-fetch stalled due to L1 instruction-cache miss or an iTLB miss",
	.topic = "frontend",
	.long_desc = "Cycles where a code-fetch stalled due to L1 instruction-cache miss or an iTLB miss",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Count issue pipeline slots where no uop was delivered from the front end to the back end when there is no back-end stall",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=4,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=3,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=2,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.count",
	.event = "umask=0x1,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switches",
	.topic = "frontend",
	.long_desc = "Number of DSB to MITE switches",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "Cycles DSB to MITE switches caused delay",
},
{
	.name = "dsb_fill.exceed_dsb_lines",
	.event = "umask=0x8,period=2000003,event=0xac",
	.desc = "Cycles when Decode Stream Buffer (DSB) fill encounter more than 3 Decode Stream Buffer (DSB) lines",
	.topic = "frontend",
	.long_desc = "DSB Fill encountered > 3 DSB lines",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "ivt metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "ivt metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "ivt metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "ivt metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "ivt metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "ivt metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "ivt metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "ivt metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "ivt metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Instruction per taken branch",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_taken",
	.metric_name = "IpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Branch instructions per taken branch",
	.topic = "ivt metrics",
	.metric_expr = "br_inst_retired.all_branches / br_inst_retired.near_taken",
	.metric_name = "BpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "ivt metrics",
	.metric_expr = "min( 1 , uops_issued.any / ( (uops_retired.retire_slots / inst_retired.any) * 32 * ( icache.hit + icache.misses ) / 4 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "ivt metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops ) )",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "ivt metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "ivt metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "ivt metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "ivt metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Instructions per Load (lower number means loads are more frequent)",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_loads",
	.metric_name = "IpL",
	.metric_group = "Instruction_Type;L1_Bound",
},
{
	.desc = "Instructions per Store",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_stores",
	.metric_name = "IpS",
	.metric_group = "Instruction_Type;Store_Bound",
},
{
	.desc = "Instructions per Branch",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.all_branches",
	.metric_name = "IpB",
	.metric_group = "Branches;Instruction_Type;Port_5;Port_6",
},
{
	.desc = "Instruction per (near) call",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_call",
	.metric_name = "IpCall",
	.metric_group = "Branches",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "ivt metrics",
	.metric_expr = "(( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / cycles",
	.metric_name = "FLOPc",
	.metric_group = "FLOPS",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "ivt metrics",
	.metric_expr = "(( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "FLOPc_SMT",
	.metric_group = "FLOPS_SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "ivt metrics",
	.metric_expr = "uops_executed.thread / (( cpu@uops_executed.core\\,cmask\\=1@ / 2 ) if #smt_on else uops_executed.cycles_ge_1_uop_exec)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Number of Instructions per non-speculative Branch Misprediction (JEClear)",
	.topic = "ivt metrics",
	.metric_expr = "inst_retired.any / br_misp_retired.all_branches",
	.metric_name = "IpMispredict",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "ivt metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads (in core cycles)",
	.topic = "ivt metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_uops_retired.l1_miss + mem_load_uops_retired.hit_lfb )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least one such miss. Per-thread)",
	.topic = "ivt metrics",
	.metric_expr = "l1d_pend_miss.pending / l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "ivt metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration ) / cycles",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "ivt metrics",
	.metric_expr = "( itlb_misses.walk_duration + dtlb_load_misses.walk_duration + dtlb_store_misses.walk_duration ) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "Page_Walks_Utilization_SMT",
	.metric_group = "TLB_SMT",
},
{
	.desc = "Average data fill bandwidth to the L1 data cache [GB / sec]",
	.topic = "ivt metrics",
	.metric_expr = "64 * l1d.replacement / 1000000000 / duration_time",
	.metric_name = "L1D_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average data fill bandwidth to the L2 cache [GB / sec]",
	.topic = "ivt metrics",
	.metric_expr = "64 * l2_lines_in.all / 1000000000 / duration_time",
	.metric_name = "L2_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "ivt metrics",
	.metric_expr = "64 * longest_lat_cache.miss / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "L1 cache true misses per kilo instruction for retired demand loads",
	.topic = "ivt metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l1_miss / inst_retired.any",
	.metric_name = "L1MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache true misses per kilo instruction for retired demand loads",
	.topic = "ivt metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache misses per kilo instruction for all request types (including speculative)",
	.topic = "ivt metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache hits per kilo instruction for all request types (including speculative)",
	.topic = "ivt metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2HPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L3 cache true misses per kilo instruction for retired demand loads",
	.topic = "ivt metrics",
	.metric_expr = "1000 * mem_load_uops_retired.llc_miss / inst_retired.any",
	.metric_name = "L3MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "Average CPU Utilization",
	.topic = "ivt metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "ivt metrics",
	.metric_expr = "( (( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / 1000000000 ) / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "ivt metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "ivt metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "ivt metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "ivt metrics",
	.metric_expr = "( 64 * ( uncore_imc@cas_count_read@ + uncore_imc@cas_count_write@ ) / 1000000000 ) / duration_time",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "Socket actual clocks when any core is active on that socket",
	.topic = "ivt metrics",
	.metric_expr = "cbox_0@event\\=0x0@",
	.metric_name = "Socket_CLKS",
	.metric_group = "",
},
{
	.desc = "C3 residency percent per core",
	.topic = "ivt metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "ivt metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "ivt metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "ivt metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "ivt metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "ivt metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "ivt metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "misalign_mem_ref.loads",
	.event = "umask=0x1,period=2000003,event=0x5",
	.desc = "Speculative cache line split load uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "Speculative cache-line split load uops dispatched to L1D",
},
{
	.name = "misalign_mem_ref.stores",
	.event = "umask=0x2,period=2000003,event=0x5",
	.desc = "Speculative cache line split STA uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "Speculative cache-line split Store-address uops dispatched to L1D",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Loads with latency value being above 4 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 4 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Loads with latency value being above 8 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 8 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Loads with latency value being above 16 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 16 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Loads with latency value being above 32 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 32 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Loads with latency value being above 64 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 64 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Loads with latency value being above 128 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 128 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Loads with latency value being above 256 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 256 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Loads with latency value being above 512 (Must be precise)",
	.topic = "memory",
	.long_desc = "Loads with latency value being above 512 (Must be precise)",
},
{
	.name = "mem_trans_retired.precise_store",
	.event = "umask=0x2,period=2000003,event=0xcd",
	.desc = "Sample stores and collect precise store operation via PEBS record. PMC3 only (Must be precise)",
	.topic = "memory",
},
{
	.name = "offcore_response.all_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc00244",
	.desc = "Counts all demand & prefetch code reads that miss the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.all_code_rd.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67f800244",
	.desc = "Counts all demand & prefetch code reads that miss the LLC  and the data returned from remote dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_code_rd.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x87f800244",
	.desc = "Counts all demand & prefetch code reads that miss the LLC  and the data forwarded from remote cache",
	.topic = "memory",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20091",
	.desc = "Counts all demand & prefetch data reads that hits the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.all_reads.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc203f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.all_reads.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x6004003f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that miss the LLC  and the data returned from local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_reads.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x87f8203f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that miss the LLC  and the data forwarded from remote cache",
	.topic = "memory",
},
{
	.name = "offcore_response.all_reads.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x107fc003f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that miss the LLC  the data is found in M state in remote cache and forwarded from there",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20004",
	.desc = "Counts all demand code reads that miss the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x600400004",
	.desc = "Counts all demand code reads that miss the LLC  and the data returned from local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67f800004",
	.desc = "Counts all demand code reads that miss the LLC  and the data returned from remote dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x87f820004",
	.desc = "Counts all demand code reads that miss the LLC  and the data forwarded from remote cache",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x107fc00004",
	.desc = "Counts all demand code reads that miss the LLC  the data is found in M state in remote cache and forwarded from there",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.any_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67fc00001",
	.desc = "Counts demand data reads that miss the LLC  and the data returned from remote & local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20001",
	.desc = "Counts demand data reads that miss in the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x600400001",
	.desc = "Counts demand data reads that miss the LLC  and the data returned from local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67f800001",
	.desc = "Counts demand data reads that miss the LLC  and the data returned from remote dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x87f820001",
	.desc = "Counts demand data reads that miss the LLC  and the data forwarded from remote cache",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x107fc00001",
	.desc = "Counts demand data reads that miss the LLC  the data is found in M state in remote cache and forwarded from there",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x107fc20002",
	.desc = "Counts all demand data writes (RFOs) that miss the LLC and the data is found in M state in remote cache and forwarded from there",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20040",
	.desc = "Counts all prefetch (that bring data to L2) code reads that miss the LLC  and the data returned from remote & local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.any_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67fc00010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss the LLC  and the data returned from remote & local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss in the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x600400010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss the LLC  and the data returned from local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67f800010",
	.desc = "Counts prefetch (that bring data to L2) data reads  that miss the LLC  and the data returned from remote dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x87f820010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss the LLC  and the data forwarded from remote cache",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x107fc00010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss the LLC  the data is found in M state in remote cache and forwarded from there",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20200",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads that miss in the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that miss in the LLC",
	.topic = "memory",
},
{
	.name = "cpl_cycles.ring0",
	.event = "umask=0x1,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when the thread is in ring 0",
	.topic = "other",
	.long_desc = "Unhalted core cycles when the thread is in ring 0",
},
{
	.name = "cpl_cycles.ring0_trans",
	.event = "umask=0x1,edge=1,period=100007,cmask=1,event=0x5c",
	.desc = "Number of intervals between processor halts while thread is in ring 0",
	.topic = "other",
	.long_desc = "Number of intervals between processor halts while thread is in ring 0",
},
{
	.name = "cpl_cycles.ring123",
	.event = "umask=0x2,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when thread is in rings 1, 2, or 3",
	.topic = "other",
	.long_desc = "Unhalted core cycles when the thread is not in ring 0",
},
{
	.name = "lock_cycles.split_lock_uc_lock_duration",
	.event = "umask=0x1,period=2000003,event=0x63",
	.desc = "Cycles when L1 and L2 are locked due to UC or split lock",
	.topic = "other",
	.long_desc = "Cycles in which the L1D and L2 are locked, due to a UC lock or split lock",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
	.long_desc = "Core cycles when at least one thread on the physical core is not in halt state",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Cases when loads get true Block-on-Store blocking code preventing store forwarding",
	.topic = "pipeline",
	.long_desc = "Loads blocked by overlapping with store buffer that cannot be forwarded",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "This event counts the number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
	.long_desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare on address",
	.topic = "pipeline",
	.long_desc = "False dependencies in MOB due to partial compare on address",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x3,period=2000003,cmask=1,event=0xd",
	.desc = "Number of cycles waiting for the checkpoints in Resource Allocation Table (RAT) to be recovered after Nuke due to all other cases except JEClear (e.g. whenever a ucode assist is needed like SSE exception, memory disambiguation, etc.)",
	.topic = "pipeline",
},
{
	.name = "int_misc.recovery_stalls_count",
	.event = "umask=0x3,edge=1,period=2000003,cmask=1,event=0xd",
	.desc = "Number of occurences waiting for the checkpoints in Resource Allocation Table (RAT) to be recovered after Nuke due to all other cases except JEClear (e.g. whenever a ucode assist is needed like SSE exception, memory disambiguation, etc.)",
	.topic = "pipeline",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x3,any=1,period=2000003,cmask=1,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "Increments each cycle the # of Uops issued by the RAT to RS. Set Cmask = 1, Inv = 1, Any= 1to count stalled cycles of this core",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
	.long_desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for all threads",
	.topic = "pipeline",
	.long_desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for all threads",
},
{
	.name = "uops_issued.flags_merge",
	.event = "umask=0x10,period=2000003,event=0xe",
	.desc = "Number of flags-merge uops being allocated",
	.topic = "pipeline",
	.long_desc = "Number of flags-merge uops allocated. Such uops adds delay",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
	.long_desc = "Number of slow LEA or similar uops allocated. Such uop has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
},
{
	.name = "uops_issued.single_mul",
	.event = "umask=0x40,period=2000003,event=0xe",
	.desc = "Number of Multiply packed/scalar single precision uops allocated",
	.topic = "pipeline",
	.long_desc = "Number of multiply packed/scalar single precision uops allocated",
},
{
	.name = "arith.fpu_div_active",
	.event = "umask=0x1,period=2000003,event=0x14",
	.desc = "Cycles when divider is busy executing divide operations",
	.topic = "pipeline",
	.long_desc = "Cycles that the divider is active, includes INT and FP. Set 'edge =1, cmask=1' to count the number of divides",
},
{
	.name = "arith.fpu_div",
	.event = "umask=0x4,edge=1,period=100003,cmask=1,event=0x14",
	.desc = "Divide operations executed",
	.topic = "pipeline",
	.long_desc = "Divide operations executed",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
	.long_desc = "Core cycles when at least one thread on the physical core is not in halt state",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Increments at the frequency of XCLK (100 MHz) when not halted",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted. (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the thread is unhalted. (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted. (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "Non-SW-prefetch load dispatches that hit fill buffer allocated for S/W prefetch",
},
{
	.name = "load_hit_pre.hw_pf",
	.event = "umask=0x2,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for hardware prefetch",
	.topic = "pipeline",
	.long_desc = "Non-SW-prefetch load dispatches that hit fill buffer allocated for H/W prefetch",
},
{
	.name = "move_elimination.int_eliminated",
	.event = "umask=0x1,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_eliminated",
	.event = "umask=0x2,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.int_not_eliminated",
	.event = "umask=0x4,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_not_eliminated",
	.event = "umask=0x8,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "Cycles the RS is empty for the thread",
},
{
	.name = "rs_events.empty_end",
	.event = "inv=1,umask=0x1,edge=1,period=200003,cmask=1,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000003,event=0x87",
	.desc = "Stall cycles because IQ is full",
	.topic = "pipeline",
	.long_desc = "Stall cycles due to IQ is full",
},
{
	.name = "br_inst_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x88",
	.desc = "Not taken macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "Not taken macro-conditional branches",
},
{
	.name = "br_inst_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired macro-conditional branches",
},
{
	.name = "br_inst_exec.taken_direct_jump",
	.event = "umask=0x82,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
},
{
	.name = "br_inst_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired indirect branches excluding calls and returns",
},
{
	.name = "br_inst_exec.taken_indirect_near_return",
	.event = "umask=0x88,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired indirect branches with return mnemonic",
},
{
	.name = "br_inst_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x88",
	.desc = "Taken speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect calls",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired indirect calls",
},
{
	.name = "br_inst_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x88",
	.desc = "Speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "Speculative and retired macro-conditional branches",
},
{
	.name = "br_inst_exec.all_direct_jmp",
	.event = "umask=0xc2,period=200003,event=0x88",
	.desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
},
{
	.name = "br_inst_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x88",
	.desc = "Speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "Speculative and retired indirect branches excluding calls and returns",
},
{
	.name = "br_inst_exec.all_indirect_near_return",
	.event = "umask=0xc8,period=200003,event=0x88",
	.desc = "Speculative and retired indirect return branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x88",
	.desc = "Speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "Speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x88",
	.desc = "Speculative and retired  branches",
	.topic = "pipeline",
	.long_desc = "Counts all near executed branches (not necessarily retired)",
},
{
	.name = "br_misp_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x89",
	.desc = "Not taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Not taken speculative and retired mispredicted macro conditional branches",
},
{
	.name = "br_misp_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired mispredicted macro conditional branches",
},
{
	.name = "br_misp_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.taken_return_near",
	.event = "umask=0x88,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
},
{
	.name = "br_misp_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect calls",
	.topic = "pipeline",
	.long_desc = "Taken speculative and retired mispredicted indirect calls",
},
{
	.name = "br_misp_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Speculative and retired mispredicted macro conditional branches",
},
{
	.name = "br_misp_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x89",
	.desc = "Mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "Mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "Counts all near executed branches (not necessarily retired)",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 0",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 0",
},
{
	.name = "uops_dispatched_port.port_0_core",
	.event = "umask=0x1,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 0",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are dispatched to port 0",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 1",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 1",
},
{
	.name = "uops_dispatched_port.port_1_core",
	.event = "umask=0x2,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 1",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are dispatched to port 1",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0xc,period=2000003,event=0xa1",
	.desc = "Cycles per thread when load or STA uops are dispatched to port 2",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 2",
},
{
	.name = "uops_dispatched_port.port_2_core",
	.event = "umask=0xc,any=1,period=2000003,event=0xa1",
	.desc = "Uops dispatched to port 2, loads and stores per core (speculative and retired)",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x30,period=2000003,event=0xa1",
	.desc = "Cycles per thread when load or STA uops are dispatched to port 3",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 3",
},
{
	.name = "uops_dispatched_port.port_3_core",
	.event = "umask=0x30,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when load or STA uops are dispatched to port 3",
	.topic = "pipeline",
	.long_desc = "Cycles per core when load or STA uops are dispatched to port 3",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 4",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 4",
},
{
	.name = "uops_dispatched_port.port_4_core",
	.event = "umask=0x40,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 4",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are dispatched to port 4",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 5",
	.topic = "pipeline",
	.long_desc = "Cycles which a Uop is dispatched on port 5",
},
{
	.name = "uops_dispatched_port.port_5_core",
	.event = "umask=0x80,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 5",
	.topic = "pipeline",
	.long_desc = "Cycles per core when uops are dispatched to port 5",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
	.long_desc = "Cycles Allocation is stalled due to Resource Related reason",
},
{
	.name = "resource_stalls.rs",
	.event = "umask=0x4,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no eligible RS entry available",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "Cycles stalled due to no store buffers available (not including draining form sync)",
},
{
	.name = "resource_stalls.rob",
	.event = "umask=0x10,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to re-order buffer full",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l2_pending",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Cycles with pending L2 cache miss loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending L2 miss loads. Set AnyThread to count per core",
},
{
	.name = "cycle_activity.cycles_l2_miss",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Cycles while L2 cache miss load* is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_ldm_pending",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Cycles with pending memory loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending memory loads. Set AnyThread to count per core",
},
{
	.name = "cycle_activity.cycles_mem_any",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_no_execute",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "This event increments by 1 for every cycle where there was no execute for this thread",
	.topic = "pipeline",
	.long_desc = "Total execution stalls",
},
{
	.name = "cycle_activity.stalls_total",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "Total execution stalls",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_pending",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Execution stalls due to L2 cache misses",
	.topic = "pipeline",
	.long_desc = "Number of loads missed L2",
},
{
	.name = "cycle_activity.stalls_l2_miss",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Execution stalls while L2 cache miss load* is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_ldm_pending",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Execution stalls due to memory subsystem",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_mem_any",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_pending",
	.event = "umask=0x8,period=2000003,cmask=8,event=0xa3",
	.desc = "Cycles with pending L1 cache miss loads",
	.topic = "pipeline",
	.long_desc = "Cycles with pending L1 cache miss loads. Set AnyThread to count per core",
},
{
	.name = "cycle_activity.cycles_l1d_miss",
	.event = "umask=0x8,period=2000003,cmask=8,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_pending",
	.event = "umask=0xc,period=2000003,cmask=12,event=0xa3",
	.desc = "Execution stalls due to L1 data cache misses",
	.topic = "pipeline",
	.long_desc = "Execution stalls due to L1 data cache miss loads. Set Cmask=0CH",
},
{
	.name = "cycle_activity.stalls_l1d_miss",
	.event = "umask=0xc,period=2000003,cmask=12,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
},
{
	.name = "uops_executed.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Counts the number of uops to be executed per-thread each cycle",
	.topic = "pipeline",
	.long_desc = "Counts total number of uops to be executed per-thread each cycle. Set Cmask = 1, INV =1 to count stall cycles",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 1 uop was executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 2 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 3 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 4 uops were executed per-thread",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core",
	.topic = "pipeline",
	.long_desc = "Counts total number of uops to be executed per-core each cycle",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
	.long_desc = "Cycles with no micro-ops executed from any thread on physical core",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter   - architectural event",
	.topic = "pipeline",
	.long_desc = "Number of instructions at retirement",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution (Must be precise)",
},
{
	.name = "other_assists.any_wb_assist",
	.event = "umask=0x80,period=100003,event=0xc1",
	.desc = "Number of times any microcode assist is invoked by HW upon uop writeback",
	.topic = "pipeline",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Retired uops (Precise event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=10,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used (Precise event)",
	.topic = "pipeline",
},
{
	.name = "machine_clears.count",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "Number of self-modifying-code machine clears detected",
},
{
	.name = "machine_clears.maskmov",
	.event = "umask=0x20,period=100003,event=0xc3",
	.desc = "This event counts the number of executed Intel AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
	.topic = "pipeline",
	.long_desc = "Counts the number of executed AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Branch instructions at retirement",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call_r3",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect macro near call instructions retired (captured in ring 3) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Not taken branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Counts the number of not taken branch instructions retired",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Far branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Number of far branches retired",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Mispredicted branch instructions at retirement",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "number of near branch instructions retired that were mispredicted and taken (Precise event)",
	.topic = "pipeline",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Count cases of saving new LBR",
	.topic = "pipeline",
	.long_desc = "Count cases of saving new LBR records by hardware",
},
{
	.name = "baclears.any",
	.event = "umask=0x1f,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
	.long_desc = "Number of front end re-steers due to BPU misprediction",
},
{
	.name = "unc_c_clockticks",
	.event = "event=0",
	.desc = "Uncore cache clock ticks. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_lookup.any",
	.event = "umask=0x11,event=0x34,filter_state=0x1",
	.desc = "All LLC Misses (code+ data rd + data wr - including demand and prefetch). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_victims.m_state",
	.event = "umask=0x1,event=0x37",
	.desc = "M line evictions from LLC (writebacks to memory). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.data_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x182",
	.desc = "LLC misses - demand and prefetch data reads - excludes LLC prefetches. Derived from unc_c_tor_inserts.miss_opcode.demand. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.uncacheable",
	.event = "umask=0x3,event=0x35,filter_opc=0x187",
	.desc = "LLC misses - Uncacheable reads. Derived from unc_c_tor_inserts.miss_opcode.uncacheable. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.rfo_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x190",
	.desc = "LLC prefetch misses for RFO. Derived from unc_c_tor_inserts.miss_opcode.rfo_prefetch. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.code_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x191",
	.desc = "LLC prefetch misses for code reads. Derived from unc_c_tor_inserts.miss_opcode.code. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.data_llc_prefetch",
	.event = "umask=0x3,event=0x35,filter_opc=0x192",
	.desc = "LLC prefetch misses for data reads. Derived from unc_c_tor_inserts.miss_opcode.data_read. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x19c",
	.desc = "PCIe allocating writes that miss LLC - DDIO misses. Derived from unc_c_tor_inserts.miss_opcode.ddio_miss. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x19e",
	.desc = "LLC misses for PCIe read current. Derived from unc_c_tor_inserts.miss_opcode.pcie_read. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.itom_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x1c8",
	.desc = "LLC misses for ItoM writes (as part of fast string memcpy stores). Derived from unc_c_tor_inserts.miss_opcode.itom_write. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_non_snoop_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x1e4",
	.desc = "LLC misses for PCIe non-snoop reads. Derived from unc_c_tor_inserts.miss_opcode.pcie_read. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_non_snoop_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x1e6",
	.desc = "LLC misses for PCIe non-snoop writes (full line). Derived from unc_c_tor_inserts.miss_opcode.pcie_write. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_full",
	.event = "umask=0x1,event=0x35,filter_opc=0x18c",
	.desc = "Streaming stores (full cache line). Derived from unc_c_tor_inserts.opcode.streaming_full. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_partial",
	.event = "umask=0x1,event=0x35,filter_opc=0x18d",
	.desc = "Streaming stores (partial cache line). Derived from unc_c_tor_inserts.opcode.streaming_partial. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_partial_read",
	.event = "umask=0x1,event=0x35,filter_opc=0x195",
	.desc = "Partial PCIe reads. Derived from unc_c_tor_inserts.opcode.pcie_partial. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x19c",
	.desc = "PCIe allocating writes that hit in LLC (DDIO hits). Derived from unc_c_tor_inserts.opcode.ddio_hit. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_read",
	.event = "umask=0x1,event=0x35,filter_opc=0x19e",
	.desc = "PCIe read current. Derived from unc_c_tor_inserts.opcode.pcie_read_current. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.itom_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x1c8",
	.desc = "ItoM write hits (as part of fast string memcpy stores). Derived from unc_c_tor_inserts.opcode.itom_write_hit. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_ns_read",
	.event = "umask=0x1,event=0x35,filter_opc=0x1e4",
	.desc = "PCIe non-snoop reads. Derived from unc_c_tor_inserts.opcode.pcie_read. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_ns_partial_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x1e5",
	.desc = "PCIe non-snoop writes (partial). Derived from unc_c_tor_inserts.opcode.pcie_partial_write. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_ns_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x1e6",
	.desc = "PCIe non-snoop writes (full line). Derived from unc_c_tor_inserts.opcode.pcie_full_write. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_tor_occupancy.miss_local",
	.event = "umask=0x2A,event=0x36",
	.desc = "Occupancy for all LLC misses that are addressed to local memory. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_c_tor_occupancy.llc_data_read",
	.event = "umask=0x3,event=0x36,filter_opc=0x182",
	.desc = "Occupancy counter for LLC data reads (demand and L2 prefetch). Derived from unc_c_tor_occupancy.miss_opcode.llc_data_read. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_c_tor_occupancy.miss_remote",
	.event = "umask=0x8A,event=0x36",
	.desc = "Occupancy for all LLC misses that are addressed to remote memory. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads",
	.event = "umask=0x3,event=0x1",
	.desc = "Read requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes",
	.event = "umask=0xC,event=0x1",
	.desc = "Write requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rsp_fwd_wb",
	.event = "umask=0x20,event=0x21",
	.desc = "M line forwarded from remote cache along with writeback to memory. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspifwd",
	.event = "umask=0x4,event=0x21",
	.desc = "M line forwarded from remote cache with no writeback to memory. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rsps",
	.event = "umask=0x2,event=0x21",
	.desc = "Shared line response from remote cache. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_h_snoop_resp.rspsfwd",
	.event = "umask=0x8,event=0x21",
	.desc = "Shared line forwarded from remote cache. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_q_clockticks",
	.event = "event=0x14",
	.desc = "QPI clock ticks. Use to get percentages for QPI cycles events. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.perpkg = "1",
},
{
	.name = "unc_q_rxl0p_power_cycles",
	.event = "event=0x10",
	.desc = "Cycles where receiving QPI link is in half-width mode. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.perpkg = "1",
	.metric_expr = "(unc_q_rxl0p_power_cycles / unc_q_clockticks) * 100.",
	.metric_name = "rxl0p_power_cycles %",
},
{
	.name = "unc_q_txl0p_power_cycles",
	.event = "event=0xd",
	.desc = "Cycles where transmitting QPI link is in half-width mode. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.perpkg = "1",
	.metric_expr = "(unc_q_txl0p_power_cycles / unc_q_clockticks) * 100.",
	.metric_name = "txl0p_power_cycles %",
},
{
	.name = "unc_q_txl_flits_g0.data",
	.event = "umask=0x2,event=0",
	.desc = "Number of data flits transmitted . Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.unit = "8Bytes",
	.perpkg = "1",
},
{
	.name = "unc_q_txl_flits_g0.non_data",
	.event = "umask=0x4,event=0",
	.desc = "Number of non data (control) flits transmitted . Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.unit = "8Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_act_count.rd",
	.event = "umask=0x1,umask=0x3,event=0x1",
	.desc = "Memory page activates for reads and writes. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_read",
	.event = "umask=0x3,event=0x4",
	.desc = "Read requests to memory controller. Derived from unc_m_cas_count.rd. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_write",
	.event = "umask=0xC,event=0x4",
	.desc = "Write requests to memory controller. Derived from unc_m_cas_count.wr. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_clockticks",
	.event = "event=0",
	.desc = "Memory controller clock ticks. Use to generate percentages for memory controller CYCLES events. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_power_channel_ppd",
	.event = "event=0x85",
	.desc = "Cycles where DRAM ranks are in power down (CKE) mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_channel_ppd / unc_m_clockticks) * 100.",
	.metric_name = "power_channel_ppd %",
},
{
	.name = "unc_m_power_critical_throttle_cycles",
	.event = "event=0x86",
	.desc = "Cycles all ranks are in critical thermal throttle. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_critical_throttle_cycles / unc_m_clockticks) * 100.",
	.metric_name = "power_critical_throttle_cycles %",
},
{
	.name = "unc_m_power_self_refresh",
	.event = "event=0x43",
	.desc = "Cycles Memory is in self refresh power mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_self_refresh / unc_m_clockticks) * 100.",
	.metric_name = "power_self_refresh %",
},
{
	.name = "unc_m_pre_count.page_miss",
	.event = "umask=0x1,event=0x2",
	.desc = "Memory page conflicts. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_p_clockticks",
	.event = "event=0",
	.desc = "PCU clock ticks. Use to get percentages of PCU cycles events. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
},
{
	.name = "unc_p_freq_band0_cycles",
	.event = "event=0xb",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band0=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band0_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band0_cycles %",
},
{
	.name = "unc_p_freq_band1_cycles",
	.event = "event=0xc",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band1=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band1_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band1_cycles %",
},
{
	.name = "unc_p_freq_band2_cycles",
	.event = "event=0xd",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band2=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band2_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band2_cycles %",
},
{
	.name = "unc_p_freq_band3_cycles",
	.event = "event=0xe",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band3=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band3_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band3_cycles %",
},
{
	.name = "unc_p_freq_band0_transitions",
	.event = "event=0xb,edge=1",
	.desc = "Counts the number of times that the uncore transitioned a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band0=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Derived from unc_p_freq_band0_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band0_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band0_cycles %",
},
{
	.name = "unc_p_freq_band1_transitions",
	.event = "event=0xc,edge=1",
	.desc = "Counts the number of times that the uncore transitioned to a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band1=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Derived from unc_p_freq_band1_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band1_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band1_cycles %",
},
{
	.name = "unc_p_freq_band2_transitions",
	.event = "event=0xd,edge=1",
	.desc = "Counts the number of cycles that the uncore transitioned to a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band2=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Derived from unc_p_freq_band2_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band2_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band2_cycles %",
},
{
	.name = "unc_p_freq_band3_transitions",
	.event = "event=0xe,edge=1",
	.desc = "Counts the number of cycles that the uncore transitioned to a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band3=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Derived from unc_p_freq_band3_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band3_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band3_cycles %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c0",
	.event = "event=0x80,occ_sel=1",
	.desc = "This is an occupancy event that tracks the number of cores that are in C0.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c0 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c0 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c3",
	.event = "event=0x80,occ_sel=2",
	.desc = "This is an occupancy event that tracks the number of cores that are in C3.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c3 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c3 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c6",
	.event = "event=0x80,occ_sel=3",
	.desc = "This is an occupancy event that tracks the number of cores that are in C6.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events . Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c6 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c6 %",
},
{
	.name = "unc_p_prochot_external_cycles",
	.event = "event=0xa",
	.desc = "Counts the number of cycles that we are in external PROCHOT mode.  This mode is triggered when a sensor off the die determines that something off-die (like DRAM) is too hot and must throttle to avoid damaging the chip. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_prochot_external_cycles / unc_p_clockticks) * 100.",
	.metric_name = "prochot_external_cycles %",
},
{
	.name = "unc_p_freq_max_limit_thermal_cycles",
	.event = "event=0x4",
	.desc = "Counts the number of cycles when temperature is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_limit_thermal_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_limit_thermal_cycles %",
},
{
	.name = "unc_p_freq_max_os_cycles",
	.event = "event=0x6",
	.desc = "Counts the number of cycles when the OS is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_os_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_os_cycles %",
},
{
	.name = "unc_p_freq_max_power_cycles",
	.event = "event=0x5",
	.desc = "Counts the number of cycles when power is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_power_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_power_cycles %",
},
{
	.name = "unc_p_freq_max_current_cycles",
	.event = "event=0x7",
	.desc = "Counts the number of cycles when current is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_current_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_current_cycles %",
},
{
	.name = "unc_p_freq_trans_cycles",
	.event = "event=0x60",
	.desc = "Cycles spent changing Frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_trans_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_trans_cycles %",
},
{
	.name = "unc_p_freq_ge_1200mhz_cycles",
	.event = "event=0xb,filter_band0=12",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to 1.2Ghz. Derived from unc_p_freq_band0_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_1200mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_1200mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_2000mhz_cycles",
	.event = "event=0xc,filter_band1=20",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to 2Ghz. Derived from unc_p_freq_band1_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_2000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_2000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_3000mhz_cycles",
	.event = "event=0xd,filter_band2=30",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to 3Ghz. Derived from unc_p_freq_band2_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_3000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_3000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_4000mhz_cycles",
	.event = "event=0xe,filter_band3=40",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to 4Ghz. Derived from unc_p_freq_band3_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_4000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_4000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_1200mhz_transitions",
	.event = "event=0xb,edge=1,filter_band0=12",
	.desc = "Counts the number of times that the uncore transitioned to a frequency greater than or equal to 1.2Ghz. Derived from unc_p_freq_band0_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_1200mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_1200mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_2000mhz_transitions",
	.event = "event=0xc,edge=1,filter_band1=20",
	.desc = "Counts the number of times that the uncore transitioned to a frequency greater than or equal to 2Ghz. Derived from unc_p_freq_band1_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_2000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_2000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_3000mhz_transitions",
	.event = "event=0xd,edge=1,filter_band2=30",
	.desc = "Counts the number of cycles that the uncore transitioned to a frequency greater than or equal to 3Ghz. Derived from unc_p_freq_band2_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_3000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_3000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_4000mhz_transitions",
	.event = "event=0xe,edge=1,filter_band3=40",
	.desc = "Counts the number of cycles that the uncore transitioned to a frequency greater than or equal to 4Ghz. Derived from unc_p_freq_band3_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_4000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_4000mhz_cycles %",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x81,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes an page walk of any page size",
	.topic = "virtual memory",
	.long_desc = "Misses in all TLB levels that cause a page walk of any page size from demand loads",
},
{
	.name = "dtlb_load_misses.demand_ld_walk_completed",
	.event = "umask=0x82,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes of any page size",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0x82,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes of any page size",
	.topic = "virtual memory",
	.long_desc = "Misses in all TLB levels that caused page walk completed of any size by demand loads",
},
{
	.name = "dtlb_load_misses.demand_ld_walk_duration",
	.event = "umask=0x84,period=2000003,event=0x8",
	.desc = "Demand load cycles page miss handler (PMH) is busy with this walk",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_duration",
	.event = "umask=0x84,period=2000003,event=0x8",
	.desc = "Demand load cycles page miss handler (PMH) is busy with this walk",
	.topic = "virtual memory",
	.long_desc = "Cycle PMH is busy with a walk due to demand loads",
},
{
	.name = "dtlb_load_misses.large_page_walk_completed",
	.event = "umask=0x88,period=100003,event=0x8",
	.desc = "Page walk for a large page completed for Demand load",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Miss in all TLB levels causes a page walk of any page size (4K/2M/4M/1G)",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks",
	.topic = "virtual memory",
	.long_desc = "Miss in all TLB levels causes a page walk that completes of any page size (4K/2M/4M/1G)",
},
{
	.name = "dtlb_store_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x49",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "Cycles PMH is busy with this walk",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x49",
	.desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Cycle count for an Extended Page table walk.  The Extended Page Directory cache is used by Virtual Machine operating systems while the guest operating systems use the standard TLB caches",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x4,period=100003,event=0x5f",
	.desc = "Load operations that miss the first DTLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts load operations that missed 1st level DTLB but hit the 2nd level",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Misses in all ITLB levels that cause page walks",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Misses in all ITLB levels that cause completed page walks",
	.topic = "virtual memory",
	.long_desc = "Misses in all ITLB levels that cause completed page walks",
},
{
	.name = "itlb_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x85",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "Cycle PMH is busy with a walk",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Operations that miss the first ITLB level but hit the second and do not cause any page walks",
	.topic = "virtual memory",
	.long_desc = "Number of cache load STLB hits. No page walk",
},
{
	.name = "itlb_misses.large_page_walk_completed",
	.event = "umask=0x80,period=100003,event=0x85",
	.desc = "Completed page walks in ITLB due to STLB load misses for large pages",
	.topic = "virtual memory",
	.long_desc = "Completed page walks in ITLB due to STLB load misses for large pages",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "Counts the number of ITLB flushes, includes 4k/2M/4M pages",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "Count number of STLB flush attempts",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_jaketown[] = {
{
	.name = "mem_uops_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load uops that miss the STLB (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store uops that miss the STLB (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load uops with locked access (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load uops that split across a cacheline boundary (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts line-splitted load uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K) (Precise event)",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts line-splitted store uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K) (Precise event)",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load uops (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts the number of load uops retired (Precise event)",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store uops (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts the number of store uops retired (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load uops with L1 cache hits as data sources (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load uops with L2 cache hits as data sources (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.llc_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load uops which data sources were data hits in LLC without snoops required",
	.topic = "cache",
	.long_desc = "This event counts retired load uops that hit in the last-level (L3) cache without snoops required",
},
{
	.name = "mem_load_uops_retired.llc_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Miss in last-level (L3) cache. Excludes Unknown data-source",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.hit_lfb",
	.event = "umask=0x40,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources were load uops missed L1 but hit FB due to preceding miss to the same cache line with data not ready (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were LLC hit and cross-core snoop missed in on-pkg core cache",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were LLC and cross-core snoop hits in on-pkg core cache",
	.topic = "cache",
	.long_desc = "This event counts retired load uops that hit in the last-level cache (L3) and were found in a non-modified state in a neighboring core's private cache (same package).  Since the last level cache is inclusive, hits to the L3 may require snooping the private L2 caches of any cores on the same socket that have the line.  In this case, a snoop was required, and another L2 had the line in a non-modified state",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were HitM responses from shared LLC",
	.topic = "cache",
	.long_desc = "This event counts retired load uops that hit in the last-level cache (L3) and were found in a non-modified state in a neighboring core's private cache (same package).  Since the last level cache is inclusive, hits to the L3 may require snooping the private L2 caches of any cores on the same socket that have the line.  In this case, a snoop was required, and another L2 had the line in a modified state, so the line had to be invalidated in that L2 cache and transferred to the requesting L2",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load uops which data sources were hits in LLC without snoops required",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_miss_retired.local_dram",
	.event = "umask=0x1,period=100007,event=0xd3",
	.desc = "Data from local DRAM either Snoop not needed or Snoop Miss (RspI)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_miss_retired.remote_dram",
	.event = "umask=0x4,period=100007,event=0xd3",
	.desc = "Data from remote DRAM either Snoop not needed or Snoop Miss (RspI)",
	.topic = "cache",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "This event counts L1D data line replacements.  Replacements occur when a new line is brought into the cache, causing eviction of a line loaded earlier",
},
{
	.name = "l1d.allocated_in_m",
	.event = "umask=0x2,period=2000003,event=0x51",
	.desc = "Allocated L1D data cache lines in M state",
	.topic = "cache",
},
{
	.name = "l1d.eviction",
	.event = "umask=0x4,period=2000003,event=0x51",
	.desc = "L1D data cache lines in M state evicted due to replacement",
	.topic = "cache",
},
{
	.name = "l1d.all_m_replacement",
	.event = "umask=0x8,period=2000003,event=0x51",
	.desc = "Cache lines in M state evicted out of L1D due to Snoop HitM or dirty line replacement",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss oustandings duration in cycles",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
},
{
	.name = "lock_cycles.cache_lock_duration",
	.event = "umask=0x2,period=2000003,event=0x63",
	.desc = "Cycles when L1D is locked",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding RFO store transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Cases when offcore requests buffer cannot take more entries for core",
	.topic = "cache",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0x1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x4,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x8,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0x10,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x20,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
},
{
	.name = "l2_rqsts.pf_hit",
	.event = "umask=0x40,period=200003,event=0x24",
	.desc = "Requests from the L2 hardware prefetchers that hit L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.pf_miss",
	.event = "umask=0x80,period=200003,event=0x24",
	.desc = "Requests from the L2 hardware prefetchers that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_store_lock_rqsts.miss",
	.event = "umask=0x1,period=200003,event=0x27",
	.desc = "RFOs that miss cache lines",
	.topic = "cache",
},
{
	.name = "l2_store_lock_rqsts.hit_e",
	.event = "umask=0x4,period=200003,event=0x27",
	.desc = "RFOs that hit cache lines in E state",
	.topic = "cache",
},
{
	.name = "l2_store_lock_rqsts.hit_m",
	.event = "umask=0x8,period=200003,event=0x27",
	.desc = "RFOs that hit cache lines in M state",
	.topic = "cache",
},
{
	.name = "l2_store_lock_rqsts.all",
	.event = "umask=0xf,period=200003,event=0x27",
	.desc = "RFOs that access cache lines in any state",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.miss",
	.event = "umask=0x1,period=200003,event=0x28",
	.desc = "Count the number of modified Lines evicted from L1 and missed L2. (Non-rejected WBs from the DCU.)",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.hit_s",
	.event = "umask=0x2,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in S state",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.hit_e",
	.event = "umask=0x4,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in E state",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.hit_m",
	.event = "umask=0x8,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in M state",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.all",
	.event = "umask=0xf,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in any state",
	.topic = "cache",
},
{
	.name = "l2_trans.demand_data_rd",
	.event = "umask=0x1,period=200003,event=0xf0",
	.desc = "Demand Data Read requests that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.rfo",
	.event = "umask=0x2,period=200003,event=0xf0",
	.desc = "RFO requests that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.code_rd",
	.event = "umask=0x4,period=200003,event=0xf0",
	.desc = "L2 cache accesses when fetching instructions",
	.topic = "cache",
},
{
	.name = "l2_trans.all_pf",
	.event = "umask=0x8,period=200003,event=0xf0",
	.desc = "L2 or LLC HW prefetches that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.l1d_wb",
	.event = "umask=0x10,period=200003,event=0xf0",
	.desc = "L1D writebacks that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.l2_fill",
	.event = "umask=0x20,period=200003,event=0xf0",
	.desc = "L2 fill requests that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.all_requests",
	.event = "umask=0x80,period=200003,event=0xf0",
	.desc = "Transactions accessing L2 pipe",
	.topic = "cache",
},
{
	.name = "l2_lines_in.i",
	.event = "umask=0x1,period=100003,event=0xf1",
	.desc = "L2 cache lines in I state filling L2",
	.topic = "cache",
},
{
	.name = "l2_lines_in.s",
	.event = "umask=0x2,period=100003,event=0xf1",
	.desc = "L2 cache lines in S state filling L2",
	.topic = "cache",
},
{
	.name = "l2_lines_in.e",
	.event = "umask=0x4,period=100003,event=0xf1",
	.desc = "L2 cache lines in E state filling L2",
	.topic = "cache",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x7,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines brought into the L2 cache.  Lines are filled into the L2 cache when there was an L2 miss",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x1,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by demand",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x2,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by demand",
	.topic = "cache",
},
{
	.name = "l2_lines_out.pf_clean",
	.event = "umask=0x4,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by L2 prefetch",
	.topic = "cache",
},
{
	.name = "l2_lines_out.pf_dirty",
	.event = "umask=0x8,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by L2 prefetch",
	.topic = "cache",
},
{
	.name = "l2_lines_out.dirty_all",
	.event = "umask=0xa,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines filling the L2",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed LLC",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to LLC",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Split locks in SQ",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0x3,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xc,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0x30,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xc0,period=200003,event=0x24",
	.desc = "Requests from L2 hardware prefetchers",
	.topic = "cache",
},
{
	.name = "l1d_blocks.bank_conflict_cycles",
	.event = "umask=0x5,period=100003,cmask=1,event=0xbf",
	.desc = "Cycles when dispatched loads are cancelled due to L1D bank conflicts with other load ports",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_c6",
	.event = "umask=0x1,period=2000003,cmask=6,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and sibling core snoop returned a clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0090",
	.desc = "Counts all prefetch data reads that hit the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and sibling core snoop returned a clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC and sibling core snoop returned a clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10008",
	.desc = "Counts all writebacks from the core to the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0004",
	.desc = "Counts all demand code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0001",
	.desc = "Counts all demand data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0001",
	.desc = "Counts demand data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0001",
	.desc = "Counts demand data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0001",
	.desc = "Counts demand data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0001",
	.desc = "Counts demand data reads that hit in the LLC and sibling core snoop returned a clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.other.lru_hints",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x803c8000",
	.desc = "Counts L2 hints sent to LLC to keep a line from being evicted out of the core caches",
	.topic = "cache",
},
{
	.name = "offcore_response.other.portio_mmio_uc",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x23ffc08000",
	.desc = "Counts miscellaneous accesses that include port i/o, MMIO and uncacheable memory accesses",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0040",
	.desc = "Counts all prefetch (that bring data to L2) code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0200",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.split_lock_uc_lock.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10400",
	.desc = "Counts requests where the address of an atomic lock instruction spans a cache line boundary or the lock instruction is executed on uncacheable address",
	.topic = "cache",
},
{
	.name = "offcore_response.streaming_stores.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10800",
	.desc = "Counts non-temporal stores",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010008",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010001",
	.desc = "Counts all demand data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010002",
	.desc = "Counts all demand rfo's",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010008",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000105B3",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010122",
	.desc = "Counts all demand & prefetch prefetch RFOs",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000107F7",
	.desc = "Counts all data/code/rfo references (demand & prefetch)",
	.topic = "cache",
},
{
	.name = "other_assists.avx_store",
	.event = "umask=0x8,period=100003,event=0xc1",
	.desc = "Number of GSSE memory assist for stores. GSSE microcode assist is being invoked whenever the hardware is unable to properly handle GSSE-256b operations",
	.topic = "floating point",
},
{
	.name = "other_assists.avx_to_sse",
	.event = "umask=0x10,period=100003,event=0xc1",
	.desc = "Number of transitions from AVX-256 to legacy SSE when penalty applicable",
	.topic = "floating point",
},
{
	.name = "other_assists.sse_to_avx",
	.event = "umask=0x20,period=100003,event=0xc1",
	.desc = "Number of transitions from SSE to AVX-256 when penalty applicable",
	.topic = "floating point",
},
{
	.name = "fp_assist.x87_output",
	.event = "umask=0x2,period=100003,event=0xca",
	.desc = "Number of X87 assists due to output value",
	.topic = "floating point",
},
{
	.name = "fp_assist.x87_input",
	.event = "umask=0x4,period=100003,event=0xca",
	.desc = "Number of X87 assists due to input value",
	.topic = "floating point",
},
{
	.name = "fp_assist.simd_output",
	.event = "umask=0x8,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to Output values",
	.topic = "floating point",
},
{
	.name = "fp_assist.simd_input",
	.event = "umask=0x10,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to input values",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.x87",
	.event = "umask=0x1,period=2000003,event=0x10",
	.desc = "Number of FP Computational Uops Executed this cycle. The number of FADD, FSUB, FCOM, FMULs, integer MULsand IMULs, FDIVs, FPREMs, FSQRTS, integer DIVs, and IDIVs. This event does not distinguish an FADD used in the middle of a transcendental flow from a s",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_packed_double",
	.event = "umask=0x10,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational packed double-precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_scalar_single",
	.event = "umask=0x20,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational scalar single-precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_packed_single",
	.event = "umask=0x40,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational packed single-precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_scalar_double",
	.event = "umask=0x80,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational scalar double-precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "simd_fp_256.packed_single",
	.event = "umask=0x1,period=2000003,event=0x11",
	.desc = "Number of GSSE-256 Computational FP single precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "simd_fp_256.packed_double",
	.event = "umask=0x2,period=2000003,event=0x11",
	.desc = "Number of AVX-256 Computational FP double precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,period=100003,cmask=1,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=2000003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
	.topic = "frontend",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Instruction cache, streaming buffer and victim cache misses",
	.topic = "frontend",
	.long_desc = "This event counts the number of instruction cache, streaming buffer and victim cache misses. Counting includes unchacheable accesses",
},
{
	.name = "idq.empty",
	.event = "umask=0x2,period=2000003,event=0x79",
	.desc = "Instruction Decode Queue (IDQ) empty cycles",
	.topic = "frontend",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_uops",
	.event = "umask=0x10,period=2000003,event=0x79",
	.desc = "Uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which the microcode sequencer assisted the front-end in delivering uops.  Microcode assists are used for complex instructions or scenarios that can't be handled by the standard decoder.  Using other instructions, if possible, will usually improve performance.  See the Intel? 64 and IA-32 Architectures Optimization Reference Manual for more information",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled ",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops not delivered to the back-end per cycle, per thread, when the back-end was not stalled.  In the ideal case 4 uops can be delivered each cycle.  The event counts the undelivered uops - so if 3 were delivered in one cycle, the counter would be incremented by 1 for that cycle (4 - 3). If the back-end is stalled, the count for this event is not incremented even when uops were not delivered, because the back-end would not have been able to accept them.  This event is used in determining the front-end bound category of the top-down pipeline slots characterization",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=4,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=3,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.count",
	.event = "umask=0x1,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switches",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "This event counts the cycles attributed to a switch from the Decoded Stream Buffer (DSB), which holds decoded instructions, to the legacy decode pipeline.  It excludes cycles when the back-end cannot  accept new micro-ops.  The penalty for these switches is potentially several cycles of instruction starvation, where no micro-ops are delivered to the back-end",
},
{
	.name = "dsb_fill.other_cancel",
	.event = "umask=0x2,period=2000003,event=0xac",
	.desc = "Cases of cancelling valid DSB fill not because of exceeding way limit",
	.topic = "frontend",
},
{
	.name = "dsb_fill.exceed_dsb_lines",
	.event = "umask=0x8,period=2000003,event=0xac",
	.desc = "Cycles when Decode Stream Buffer (DSB) fill encounter more than 3 Decode Stream Buffer (DSB) lines",
	.topic = "frontend",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_occur",
	.event = "umask=0x10,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=2,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_ge_1_uop_deliv.core",
	.event = "inv=1,umask=0x1,period=2000003,cmask=4,event=0x9c",
	.desc = "Cycles when 1 or more uops were delivered to the by the front end",
	.topic = "frontend",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
},
{
	.name = "dsb_fill.all_cancel",
	.event = "umask=0xa,period=2000003,event=0xac",
	.desc = "Cases of cancelling valid Decode Stream Buffer (DSB) fill not because of exceeding way limit",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "idq.mite_all_uops",
	.event = "umask=0x3c,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
},
{
	.name = "idq.ms_switches",
	.event = "umask=0x30,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "jkt metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "jkt metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "jkt metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "jkt metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "jkt metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "jkt metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "jkt metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "jkt metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "jkt metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "jkt metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "jkt metrics",
	.metric_expr = "min( 1 , uops_issued.any / ( (uops_retired.retire_slots / inst_retired.any) * 32 * ( icache.hit + icache.misses ) / 4 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "jkt metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops ) )",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "jkt metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "jkt metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "jkt metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "jkt metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "jkt metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "jkt metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "jkt metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "jkt metrics",
	.metric_expr = "(( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / cycles",
	.metric_name = "FLOPc",
	.metric_group = "FLOPS",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "jkt metrics",
	.metric_expr = "(( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "FLOPc_SMT",
	.metric_group = "FLOPS_SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "jkt metrics",
	.metric_expr = "uops_dispatched.thread / (( cpu@uops_dispatched.core\\,cmask\\=1@ / 2 ) if #smt_on else cpu@uops_dispatched.core\\,cmask\\=1@)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "jkt metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Average CPU Utilization",
	.topic = "jkt metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "jkt metrics",
	.metric_expr = "( (( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / 1000000000 ) / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "jkt metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "jkt metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "jkt metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "jkt metrics",
	.metric_expr = "( 64 * ( uncore_imc@cas_count_read@ + uncore_imc@cas_count_write@ ) / 1000000000 ) / duration_time",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "Socket actual clocks when any core is active on that socket",
	.topic = "jkt metrics",
	.metric_expr = "cbox_0@event\\=0x0@",
	.metric_name = "Socket_CLKS",
	.metric_group = "",
},
{
	.desc = "C3 residency percent per core",
	.topic = "jkt metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "jkt metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "jkt metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "jkt metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "jkt metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "jkt metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "jkt metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts",
	.topic = "memory",
	.long_desc = "This event counts the number of memory ordering Machine Clears detected. Memory Ordering Machine Clears can result from memory disambiguation, external snoops, or cross SMT-HW-thread snoop (stores) hitting load buffers.  Machine clears can have a significant performance impact if they are happening frequently",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Loads with latency value being above 4  (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Loads with latency value being above 8 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Loads with latency value being above 16 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Loads with latency value being above 32 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Loads with latency value being above 64 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Loads with latency value being above 128 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Loads with latency value being above 256 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Loads with latency value being above 512 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.precise_store",
	.event = "umask=0x2,period=2000003,event=0xcd",
	.desc = "Sample stores and collect precise store operation via PEBS record. PMC3 only. (Precise Event - PEBS) (Must be precise)",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.loads",
	.event = "umask=0x1,period=2000003,event=0x5",
	.desc = "Speculative cache line split load uops dispatched to L1 cache",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.stores",
	.event = "umask=0x2,period=2000003,event=0x5",
	.desc = "Speculative cache line split STA uops dispatched to L1 cache",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20004",
	.desc = "Counts all demand code reads that miss the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x600400004",
	.desc = "Counts all demand code reads that miss the LLC  and the data returned from local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67f800004",
	.desc = "Counts all demand code reads that miss the LLC  and the data returned from remote dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x87f820004",
	.desc = "Counts all demand code reads that miss the LLC  and the data forwarded from remote cache",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x107fc00004",
	.desc = "Counts all demand code reads that miss the LLC  the data is found in M state in remote cache and forwarded from there",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.any_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67fc00001",
	.desc = "Counts demand data reads that miss the LLC  and the data returned from remote & local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20001",
	.desc = "Counts demand data reads that miss in the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x600400001",
	.desc = "Counts demand data reads that miss the LLC  and the data returned from local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67f800001",
	.desc = "Counts demand data reads that miss the LLC  and the data returned from remote dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x87f820001",
	.desc = "Counts demand data reads that miss the LLC  and the data forwarded from remote cache",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x107fc00001",
	.desc = "Counts demand data reads that miss the LLC  the data is found in M state in remote cache and forwarded from there",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20040",
	.desc = "Counts all prefetch (that bring data to L2) code reads that miss the LLC  and the data returned from remote & local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.any_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67fc00010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss the LLC  and the data returned from remote & local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss in the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x600400010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss the LLC  and the data returned from local dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.remote_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x67f800010",
	.desc = "Counts prefetch (that bring data to L2) data reads  that miss the LLC  and the data returned from remote dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x87f820010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss the LLC  and the data forwarded from remote cache",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x107fc00010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss the LLC  the data is found in M state in remote cache and forwarded from there",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20200",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads that miss in the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3fffc20080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "memory",
},
{
	.name = "offcore_response.all_demand_mlc_pref_reads.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x600400077",
	.desc = "Counts all local dram accesses for all demand and L2 prefetches. LLC prefetches are excluded",
	.topic = "memory",
},
{
	.name = "offcore_response.all_demand_mlc_pref_reads.llc_miss.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFFC20077",
	.desc = "This event counts all LLC misses for all demand and L2 prefetches. LLC prefetches are excluded",
	.topic = "memory",
},
{
	.name = "offcore_response.all_demand_mlc_pref_reads.llc_miss.remote_hitm_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x187FC20077",
	.desc = "This event counts all remote cache-to-cache transfers (includes HITM and HIT-Forward) for all demand and L2 prefetches. LLC prefetches are excluded",
	.topic = "memory",
},
{
	.name = "insts_written_to_iq.insts",
	.event = "umask=0x1,period=2000003,event=0x17",
	.desc = "Valid instructions written to IQ per cycle",
	.topic = "other",
},
{
	.name = "cpl_cycles.ring0",
	.event = "umask=0x1,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when the thread is in ring 0",
	.topic = "other",
},
{
	.name = "cpl_cycles.ring0_trans",
	.event = "umask=0x1,edge=1,period=100007,cmask=1,event=0x5c",
	.desc = "Number of intervals between processor halts while thread is in ring 0",
	.topic = "other",
},
{
	.name = "cpl_cycles.ring123",
	.event = "umask=0x2,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when thread is in rings 1, 2, or 3",
	.topic = "other",
},
{
	.name = "hw_pre_req.dl1_miss",
	.event = "umask=0x2,period=2000003,event=0x4e",
	.desc = "Hardware Prefetch requests that miss the L1D cache. This accounts for both L1 streamer and IP-based (IPP) HW prefetchers. A request is being counted each time it access the cache & miss it, including if a block is applicable or if hit the Fill Buffer for ",
	.topic = "other",
},
{
	.name = "lock_cycles.split_lock_uc_lock_duration",
	.event = "umask=0x1,period=2000003,event=0x63",
	.desc = "Cycles when L1 and L2 are locked due to UC or split lock",
	.topic = "other",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, this event counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state. This event has a constant ratio with the CPU_CLK_UNHALTED.REF_XCLK event. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "br_inst_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x88",
	.desc = "Not taken macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_direct_jump",
	.event = "umask=0x82,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_near_return",
	.event = "umask=0x88,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches with return mnemonic",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x88",
	.desc = "Taken speculative and retired direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x88",
	.desc = "Speculative and retired macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_jmp",
	.event = "umask=0xc2,period=200003,event=0x88",
	.desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x88",
	.desc = "Speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_indirect_near_return",
	.event = "umask=0xc8,period=200003,event=0x88",
	.desc = "Speculative and retired indirect return branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x88",
	.desc = "Speculative and retired direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x89",
	.desc = "Not taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_return_near",
	.event = "umask=0x88,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x89",
	.desc = "Mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted direct near calls",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000003,event=0x87",
	.desc = "Stall cycles because IQ is full",
	.topic = "pipeline",
},
{
	.name = "int_misc.rat_stall_cycles",
	.event = "umask=0x40,period=2000003,event=0xd",
	.desc = "Cycles when Resource Allocation Table (RAT) external stall is sent to Instruction Decode Queue (IDQ) for the thread",
	.topic = "pipeline",
},
{
	.name = "partial_rat_stalls.flags_merge_uop",
	.event = "umask=0x20,period=2000003,event=0x59",
	.desc = "Increments the number of flags-merge uops in flight each cycle",
	.topic = "pipeline",
},
{
	.name = "partial_rat_stalls.slow_lea_window",
	.event = "umask=0x40,period=2000003,event=0x59",
	.desc = "Cycles with at least one slow LEA uop being allocated",
	.topic = "pipeline",
	.long_desc = "This event counts the number of cycles with at least one slow LEA uop being allocated. A uop is generally considered as slow LEA if it has three sources (for example, two sources and immediate) regardless of whether it is a result of LEA instruction or not. Examples of the slow LEA uop are or uops with base, index, and offset source operands using base and index reqisters, where base is EBR/RBP/R13, using RIP relative or 16-bit addressing modes. See the Intel? 64 and IA-32 Architectures Optimization Reference Manual for more details about slow LEA instructions",
},
{
	.name = "partial_rat_stalls.mul_single_uop",
	.event = "umask=0x80,period=2000003,event=0x59",
	.desc = "Multiply packed/scalar single precision uops allocated",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.lb",
	.event = "umask=0x2,period=2000003,event=0xa2",
	.desc = "Counts the cycles of stall due to lack of load buffers",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rs",
	.event = "umask=0x4,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no eligible RS entry available",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rob",
	.event = "umask=0x10,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to re-order buffer full",
	.topic = "pipeline",
},
{
	.name = "resource_stalls2.bob_full",
	.event = "umask=0x40,period=2000003,event=0x5b",
	.desc = "Cycles when Allocator is stalled if BOB is full and new branch needs it",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of Uops issued by the front-end of the pipeilne to the back-end",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for all threads",
	.topic = "pipeline",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Count cases of saving new LBR",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "This event is incremented when self-modifying code (SMC) is detected, which causes a machine clear.  Machine clears can have a significant performance impact if they are happening frequently",
},
{
	.name = "machine_clears.maskmov",
	.event = "umask=0x20,period=100003,event=0xc3",
	.desc = "This event counts the number of executed Intel AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
	.topic = "pipeline",
	.long_desc = "Maskmov false fault - counts number of time ucode passes through Maskmov flow due to instruction's mask being 0 while the flow was completed without raising a fault",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter   - architectural event",
	.topic = "pipeline",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Actually retired uops (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of micro-ops retired (Precise event)",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of retirement slots used each cycle.  There are potentially 4 slots that can be used each cycle - meaning, 4 micro-ops or 4 instructions could retire each cycle.  This event is used in determining the 'Retiring' category of the Top-Down pipeline slots characterization (Precise event)",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=10,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Not taken branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Far branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired. (Precise Event - PEBS) (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc5",
	.desc = "Direct and indirect mispredicted near call instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc5",
	.desc = "Mispredicted not taken branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "Mispredicted taken branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired. (Precise Event - PEBS) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Mispredicted macro branch instructions retired. (Precise Event - PEBS) (Must be precise)",
},
{
	.name = "other_assists.itlb_miss_retired",
	.event = "umask=0x2,period=100003,event=0xc1",
	.desc = "Retired instructions experiencing ITLB misses",
	.topic = "pipeline",
},
{
	.name = "arith.fpu_div_active",
	.event = "umask=0x1,period=2000003,event=0x14",
	.desc = "Cycles when divider is busy executing divide operations",
	.topic = "pipeline",
},
{
	.name = "arith.fpu_div",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0x14",
	.desc = "Divide operations executed",
	.topic = "pipeline",
	.long_desc = "This event counts the number of the divide operations executed",
},
{
	.name = "uops_dispatched.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Uops dispatched per thread",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Uops dispatched from any thread",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 0",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 1",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 4",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 5",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_no_dispatch",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "Each cycle there was no dispatch for this thread, increment by 1. Note this is connect to Umask 2. No dispatch can be deduced from the UOPS_EXECUTED event",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_pending",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Each cycle there was a miss-pending demand load this thread, increment by 1. Note this is in DCU and connected to Umask 1. Miss Pending demand load should be deduced by OR-ing increment bits of DCACHE_MISS_PEND.PENDING",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l2_pending",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Each cycle there was a MLC-miss pending demand load this thread (i.e. Non-completed valid SQ entry allocated for demand load and waiting for Uncore), increment by 1. Note this is in MLC and connected to Umask 0",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_pending",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Each cycle there was a miss-pending demand load this thread and no uops dispatched, increment by 1. Note this is in DCU and connected to Umask 1 and 2. Miss Pending demand load should be deduced by OR-ing increment bits of DCACHE_MISS_PEND.PENDING",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_pending",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Each cycle there was a MLC-miss pending demand load and no uops dispatched on this thread (i.e. Non-completed valid SQ entry allocated for demand load and waiting for Uncore), increment by 1. Note this is in MLC and connected to Umask 0 and 2",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for software prefetch",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.hw_pf",
	.event = "umask=0x2,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for hardware prefetch",
	.topic = "pipeline",
},
{
	.name = "ld_blocks.data_unknown",
	.event = "umask=0x1,period=100003,event=0x3",
	.desc = "Loads delayed due to SB blocks, preceding store operations with known addresses but unknown data",
	.topic = "pipeline",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Cases when loads get true Block-on-Store blocking code preventing store forwarding",
	.topic = "pipeline",
	.long_desc = "This event counts loads that followed a store to the same address, where the data could not be forwarded inside the pipeline from the store to the load.  The most common reason why store forwarding would be blocked is when a load's address range overlaps with a preceeding smaller uncompleted store.  See the table of not supported store forwards in the Intel? 64 and IA-32 Architectures Optimization Reference Manual.  The penalty for blocked store forwarding is that the load must wait for the store to complete before it can be issued",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "This event counts the number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
},
{
	.name = "ld_blocks.all_block",
	.event = "umask=0x10,period=100003,event=0x3",
	.desc = "Number of cases where any load ends up with a valid block-code written to the load buffer (including blocks due to Memory Order Buffer (MOB), Data Cache Unit (DCU), TLB, but load has no DCU miss)",
	.topic = "pipeline",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare",
	.topic = "pipeline",
	.long_desc = "Aliasing occurs when a load is issued after a store and their memory addresses are offset by 4K.  This event counts the number of loads that aliased with a preceding store, resulting in an extended address check in the pipeline.  The enhanced address check typically has a performance penalty of 5 cycles",
},
{
	.name = "ld_blocks_partial.all_sta_block",
	.event = "umask=0x8,period=100003,event=0x7",
	.desc = "This event counts the number of times that load operations are temporarily blocked because of older stores, with addresses that are not yet known. A load operation may incur more than one block of this type",
	.topic = "pipeline",
},
{
	.name = "agu_bypass_cancel.count",
	.event = "umask=0x1,period=100003,event=0xb6",
	.desc = "This event counts executed load operations with all the following traits: 1. addressing of the format [base + offset], 2. the offset is between 1 and 2047, 3. the address specified in the base register is in one page and the address [base+offset] is in an",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other is halted",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_0_core",
	.event = "umask=0x1,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 0",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_1_core",
	.event = "umask=0x2,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 1",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_4_core",
	.event = "umask=0x40,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 4",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_5_core",
	.event = "umask=0x80,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 5",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0xc,period=2000003,event=0xa1",
	.desc = "Cycles per thread when load or STA uops are dispatched to port 2",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x30,period=2000003,event=0xa1",
	.desc = "Cycles per thread when load or STA uops are dispatched to port 3",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_2_core",
	.event = "umask=0xc,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when load or STA uops are dispatched to port 2",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_3_core",
	.event = "umask=0x30,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when load or STA uops are dispatched to port 3",
	.topic = "pipeline",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Instructions retired. (Precise Event - PEBS) (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "resource_stalls2.all_prf_control",
	.event = "umask=0xf,period=2000003,event=0x5b",
	.desc = "Resource stalls2 control structures full for physical registers",
	.topic = "pipeline",
},
{
	.name = "resource_stalls2.all_fl_empty",
	.event = "umask=0xc,period=2000003,event=0x5b",
	.desc = "Cycles with either free list is empty",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.mem_rs",
	.event = "umask=0xe,period=2000003,event=0xa2",
	.desc = "Resource stalls due to memory buffers or Reservation Station (RS) being fully utilized",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.ooo_rsrc",
	.event = "umask=0xf0,period=2000003,event=0xa2",
	.desc = "Resource stalls due to Rob being full, FCSW, MXCSR and OTHER",
	.topic = "pipeline",
},
{
	.name = "resource_stalls2.ooo_rsrc",
	.event = "umask=0x4f,period=2000003,event=0x5b",
	.desc = "Resource stalls out of order resources full",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.lb_sb",
	.event = "umask=0xa,period=2000003,event=0xa2",
	.desc = "Resource stalls due to load or store buffers all being in use",
	.topic = "pipeline",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x3,period=2000003,cmask=1,event=0xd",
	.desc = "Number of cycles waiting for the checkpoints in Resource Allocation Table (RAT) to be recovered after Nuke due to all other cases except JEClear (e.g. whenever a ucode assist is needed like SSE exception, memory disambiguation, etc...)",
	.topic = "pipeline",
},
{
	.name = "partial_rat_stalls.flags_merge_uop_cycles",
	.event = "umask=0x20,period=2000003,cmask=1,event=0x59",
	.desc = "Performance sensitive flags-merging uops added by Sandy Bridge u-arch",
	.topic = "pipeline",
	.long_desc = "This event counts the number of cycles spent executing performance-sensitive flags-merging uops. For example, shift CL (merge_arith_flags). For more details, See the Intel? 64 and IA-32 Architectures Optimization Reference Manual",
},
{
	.name = "int_misc.recovery_stalls_count",
	.event = "umask=0x3,edge=1,period=2000003,cmask=1,event=0xd",
	.desc = "Number of occurences waiting for the checkpoints in Resource Allocation Table (RAT) to be recovered after Nuke due to all other cases except JEClear (e.g. whenever a ucode assist is needed like SSE exception, memory disambiguation, etc...)",
	.topic = "pipeline",
},
{
	.name = "baclears.any",
	.event = "umask=0x1f,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x88",
	.desc = "Speculative and retired  branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "uops_retired.core_stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "machine_clears.count",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "rs_events.empty_end",
	.event = "inv=1,umask=0x1,edge=1,period=2000003,cmask=1,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x3,any=1,period=2000003,cmask=1,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "unc_c_clockticks",
	.event = "event=0",
	.desc = "Uncore cache clock ticks. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_lookup.any",
	.event = "umask=0x11,event=0x34,filter_state=0x1",
	.desc = "All LLC Misses (code+ data rd + data wr - including demand and prefetch). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_llc_victims.m_state",
	.event = "umask=0x1,event=0x37",
	.desc = "M line evictions from LLC (writebacks to memory). Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.data_read",
	.event = "umask=0x3,event=0x35,filter_opc=0x182",
	.desc = "LLC misses - demand and prefetch data reads - excludes LLC prefetches. Derived from unc_c_tor_inserts.miss_opcode.demand. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.uncacheable",
	.event = "umask=0x3,event=0x35,filter_opc=0x187",
	.desc = "LLC misses - Uncacheable reads. Derived from unc_c_tor_inserts.miss_opcode.uncacheable. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x19c",
	.desc = "PCIe allocating writes that miss LLC - DDIO misses. Derived from unc_c_tor_inserts.miss_opcode.ddio_miss. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.itom_write",
	.event = "umask=0x3,event=0x35,filter_opc=0x1c8",
	.desc = "LLC misses for ItoM writes (as part of fast string memcpy stores). Derived from unc_c_tor_inserts.miss_opcode.itom_write. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_full",
	.event = "umask=0x1,event=0x35,filter_opc=0x18c",
	.desc = "Streaming stores (full cache line). Derived from unc_c_tor_inserts.opcode.streaming_full. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_partial",
	.event = "umask=0x1,event=0x35,filter_opc=0x18d",
	.desc = "Streaming stores (partial cache line). Derived from unc_c_tor_inserts.opcode.streaming_partial. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_partial_read",
	.event = "umask=0x1,event=0x35,filter_opc=0x195",
	.desc = "Partial PCIe reads. Derived from unc_c_tor_inserts.opcode.pcie_partial. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x19c",
	.desc = "PCIe allocating writes that hit in LLC (DDIO hits). Derived from unc_c_tor_inserts.opcode.ddio_hit. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_read",
	.event = "umask=0x1,event=0x35,filter_opc=0x19e",
	.desc = "PCIe read current. Derived from unc_c_tor_inserts.opcode.pcie_read_current. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.itom_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x1c8",
	.desc = "ItoM write hits (as part of fast string memcpy stores). Derived from unc_c_tor_inserts.opcode.itom_write_hit. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_ns_read",
	.event = "umask=0x1,event=0x35,filter_opc=0x1e4",
	.desc = "PCIe non-snoop reads. Derived from unc_c_tor_inserts.opcode.pcie_read. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_ns_partial_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x1e5",
	.desc = "PCIe non-snoop writes (partial). Derived from unc_c_tor_inserts.opcode.pcie_partial_write. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.pcie_ns_write",
	.event = "umask=0x1,event=0x35,filter_opc=0x1e6",
	.desc = "PCIe non-snoop writes (full line). Derived from unc_c_tor_inserts.opcode.pcie_full_write. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_c_tor_occupancy.miss_all",
	.event = "umask=0xa,event=0x36,filter_opc=0x182",
	.desc = "Occupancy counter for all LLC misses; we divide this by UNC_C_CLOCKTICKS to get average Q depth. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
	.metric_expr = "(unc_c_tor_occupancy.miss_all / unc_c_clockticks) * 100.",
	.metric_name = "tor_occupancy.miss_all %",
},
{
	.name = "unc_c_tor_occupancy.llc_data_read",
	.event = "umask=0x3,event=0x36",
	.desc = "Occupancy counter for LLC data reads (demand and L2 prefetch). Derived from unc_c_tor_occupancy.miss_opcode.llc_data_read. Unit: uncore_cbox ",
	.topic = "uncore cache",
	.pmu = "uncore_cbox",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.reads",
	.event = "umask=0x3,event=0x1",
	.desc = "read requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_h_requests.writes",
	.event = "umask=0xc,event=0x1",
	.desc = "write requests to home agent. Unit: uncore_ha ",
	.topic = "uncore cache",
	.pmu = "uncore_ha",
	.perpkg = "1",
},
{
	.name = "unc_q_clockticks",
	.event = "event=0x14",
	.desc = "QPI clock ticks. Used to get percentages of QPI cycles events. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.perpkg = "1",
},
{
	.name = "unc_q_rxl0p_power_cycles",
	.event = "event=0x10",
	.desc = "Cycles where receiving QPI link is in half-width mode. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.perpkg = "1",
	.metric_expr = "(unc_q_rxl0p_power_cycles / unc_q_clockticks) * 100.",
	.metric_name = "rxl0p_power_cycles %",
},
{
	.name = "unc_q_txl0p_power_cycles",
	.event = "event=0xd",
	.desc = "Cycles where transmitting QPI link is in half-width mode. Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.perpkg = "1",
	.metric_expr = "(unc_q_txl0p_power_cycles / unc_q_clockticks) * 100.",
	.metric_name = "txl0p_power_cycles %",
},
{
	.name = "unc_q_txl_flits_g0.data",
	.event = "umask=0x2,event=0",
	.desc = "Number of data flits transmitted . Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.unit = "8Bytes",
	.perpkg = "1",
},
{
	.name = "unc_q_txl_flits_g0.non_data",
	.event = "umask=0x4,event=0",
	.desc = "Number of non data (control) flits transmitted . Unit: uncore_qpi ",
	.topic = "uncore interconnect",
	.pmu = "uncore_qpi",
	.unit = "8Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_act_count",
	.event = "event=0x1",
	.desc = "Memory page activates. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_read",
	.event = "umask=0x3,event=0x4",
	.desc = "read requests to memory controller. Derived from unc_m_cas_count.rd. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_write",
	.event = "umask=0xc,event=0x4",
	.desc = "write requests to memory controller. Derived from unc_m_cas_count.wr. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_clockticks",
	.event = "event=0",
	.desc = "Memory controller clock ticks. Used to get percentages of memory controller cycles events. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_power_channel_ppd",
	.event = "event=0x85",
	.desc = "Cycles where DRAM ranks are in power down (CKE) mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_channel_ppd / unc_m_clockticks) * 100.",
	.metric_name = "power_channel_ppd %",
},
{
	.name = "unc_m_power_critical_throttle_cycles",
	.event = "event=0x86",
	.desc = "Cycles all ranks are in critical thermal throttle. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_critical_throttle_cycles / unc_m_clockticks) * 100.",
	.metric_name = "power_critical_throttle_cycles %",
},
{
	.name = "unc_m_power_self_refresh",
	.event = "event=0x43",
	.desc = "Cycles Memory is in self refresh power mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_self_refresh / unc_m_clockticks) * 100.",
	.metric_name = "power_self_refresh %",
},
{
	.name = "unc_m_pre_count.page_miss",
	.event = "umask=0x1,event=0x2",
	.desc = "Memory page conflicts. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_rpq_occupancy",
	.event = "event=0x80",
	.desc = "Occupancy counter for memory read queue. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_p_clockticks",
	.event = "event=0",
	.desc = "PCU clock ticks. Use to get percentages of PCU cycles events. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
},
{
	.name = "unc_p_freq_band0_cycles",
	.event = "event=0xb",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band0=XXX with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band0_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band0_cycles %",
},
{
	.name = "unc_p_freq_band1_cycles",
	.event = "event=0xc",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band1=XXX with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band1_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band1_cycles %",
},
{
	.name = "unc_p_freq_band2_cycles",
	.event = "event=0xd",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band2=XXX with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band2_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band2_cycles %",
},
{
	.name = "unc_p_freq_band3_cycles",
	.event = "event=0xe",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band3=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band3_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band3_cycles %",
},
{
	.name = "unc_p_freq_band0_transitions",
	.event = "event=0xb,edge=1",
	.desc = "Counts the number of times that the uncore transitioned a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band0=XXX with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Derived from unc_p_freq_band0_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band0_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band0_cycles %",
},
{
	.name = "unc_p_freq_band1_transitions",
	.event = "event=0xc,edge=1",
	.desc = "Counts the number of times that the uncore transistioned to a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band1=XXX with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Derived from unc_p_freq_band1_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band1_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band1_cycles %",
},
{
	.name = "unc_p_freq_band2_transitions",
	.event = "event=0xd,edge=1",
	.desc = "Counts the number of cycles that the uncore transitioned to a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band2=XXX with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Derived from unc_p_freq_band2_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band2_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band2_cycles %",
},
{
	.name = "unc_p_freq_band3_transitions",
	.event = "event=0xe,edge=1",
	.desc = "Counts the number of cycles that the uncore transitioned to a frequency greater than or equal to the frequency that is configured in the filter.  (filter_band3=XXX, with XXX in 100Mhz units). One can also use inversion (filter_inv=1) to track cycles when we were less than the configured frequency. Derived from unc_p_freq_band3_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_band3_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_band3_cycles %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c0",
	.event = "event=0x80,occ_sel=1",
	.desc = "This is an occupancy event that tracks the number of cores that are in C0.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c0 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c0 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c3",
	.event = "event=0x80,occ_sel=2",
	.desc = "This is an occupancy event that tracks the number of cores that are in C3.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events and occupancy triggering to capture other details. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c3 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c3 %",
},
{
	.name = "unc_p_power_state_occupancy.cores_c6",
	.event = "event=0x80,occ_sel=3",
	.desc = "This is an occupancy event that tracks the number of cores that are in C6.  It can be used by itself to get the average number of cores in C0, with threshholding to generate histograms, or with other PCU events . Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_power_state_occupancy.cores_c6 / unc_p_clockticks) * 100.",
	.metric_name = "power_state_occupancy.cores_c6 %",
},
{
	.name = "unc_p_prochot_external_cycles",
	.event = "event=0xa",
	.desc = "Counts the number of cycles that we are in external PROCHOT mode.  This mode is triggered when a sensor off the die determines that something off-die (like DRAM) is too hot and must throttle to avoid damaging the chip. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_prochot_external_cycles / unc_p_clockticks) * 100.",
	.metric_name = "prochot_external_cycles %",
},
{
	.name = "unc_p_freq_max_limit_thermal_cycles",
	.event = "event=0x4",
	.desc = "Counts the number of cycles when temperature is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_limit_thermal_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_limit_thermal_cycles %",
},
{
	.name = "unc_p_freq_max_os_cycles",
	.event = "event=0x6",
	.desc = "Counts the number of cycles when the OS is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_os_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_os_cycles %",
},
{
	.name = "unc_p_freq_max_power_cycles",
	.event = "event=0x5",
	.desc = "Counts the number of cycles when power is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_power_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_power_cycles %",
},
{
	.name = "unc_p_freq_max_current_cycles",
	.event = "event=0x7",
	.desc = "Counts the number of cycles when current is the upper limit on frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_max_current_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_max_current_cycles %",
},
{
	.name = "unc_p_freq_trans_cycles",
	.event = "event=0",
	.desc = "Cycles spent changing Frequency. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_trans_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_trans_cycles %",
},
{
	.name = "unc_p_freq_ge_1200mhz_cycles",
	.event = "event=0xb,filter_band0=12",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to 1.2Ghz. Derived from unc_p_freq_band0_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_1200mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_1200mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_2000mhz_cycles",
	.event = "event=0xc,filter_band1=20",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to 2Ghz. Derived from unc_p_freq_band1_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_2000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_2000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_3000mhz_cycles",
	.event = "event=0xd,filter_band2=30",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to 3Ghz. Derived from unc_p_freq_band2_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_3000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_3000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_4000mhz_cycles",
	.event = "event=0xe,filter_band3=40",
	.desc = "Counts the number of cycles that the uncore was running at a frequency greater than or equal to 4Ghz. Derived from unc_p_freq_band3_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_4000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_4000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_1200mhz_transitions",
	.event = "event=0xb,edge=1,filter_band0=12",
	.desc = "Counts the number of times that the uncore transitioned to a frequency greater than or equal to 1.2Ghz. Derived from unc_p_freq_band0_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_1200mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_1200mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_2000mhz_transitions",
	.event = "event=0xc,edge=1,filter_band1=20",
	.desc = "Counts the number of times that the uncore transitioned to a frequency greater than or equal to 2Ghz. Derived from unc_p_freq_band1_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_2000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_2000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_3000mhz_transitions",
	.event = "event=0xd,edge=1,filter_band2=30",
	.desc = "Counts the number of cycles that the uncore transitioned to a frequency greater than or equal to 3Ghz. Derived from unc_p_freq_band2_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_3000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_3000mhz_cycles %",
},
{
	.name = "unc_p_freq_ge_4000mhz_transitions",
	.event = "event=0xe,edge=1,filter_band3=40",
	.desc = "Counts the number of cycles that the uncore transitioned to a frequency greater than or equal to 4Ghz. Derived from unc_p_freq_band3_cycles. Unit: uncore_pcu ",
	.topic = "uncore power",
	.pmu = "uncore_pcu",
	.perpkg = "1",
	.metric_expr = "(unc_p_freq_ge_4000mhz_cycles / unc_p_clockticks) * 100.",
	.metric_name = "freq_ge_4000mhz_cycles %",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Cycle count for an Extended Page table walk.  The Extended Page Directory cache is used by Virtual Machine operating systems while the guest operating systems use the standard TLB caches",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Misses in all ITLB levels that cause completed page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x85",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event count cycles when Page Miss Handler (PMH) is servicing page walks caused by ITLB misses",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Operations that miss the first ITLB level but hit the second and do not cause any page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x8",
	.desc = "Load misses at all DTLB levels that cause completed page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles when the  page miss handler (PMH) is servicing page walks caused by DTLB load misses",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x8",
	.desc = "Load operations that miss the first DTLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts load operations that miss the first DTLB level but hit the second and do not cause any page walks. The penalty in this case is approximately 7 cycles",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x49",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x49",
	.desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_knightslanding[] = {
{
	.name = "l2_requests_reject.all",
	.event = "umask=0x0,period=200003,event=0x30",
	.desc = "Counts the number of MEC requests from the L2Q that reference a cache line (cacheable requests) exlcuding SW prefetches filling only to L2 cache and L1 evictions (automatically exlcudes L2HWP, UC, WC) that were rejected - Multiple repeated rejects should be counted multiple times",
	.topic = "cache",
},
{
	.name = "core_reject_l2q.all",
	.event = "umask=0x0,period=200003,event=0x31",
	.desc = "Counts the number of MEC requests that were not accepted into the L2Q because of any L2  queue reject condition. There is no concept of at-ret here. It might include requests due to instructions in the speculative path",
	.topic = "cache",
},
{
	.name = "l2_requests.reference",
	.event = "umask=0x4f,period=200003,event=0x2e",
	.desc = "Counts the total number of L2 cache references",
	.topic = "cache",
},
{
	.name = "l2_requests.miss",
	.event = "umask=0x41,period=200003,event=0x2e",
	.desc = "Counts the number of L2 cache misses",
	.topic = "cache",
},
{
	.name = "fetch_stall.icache_fill_pending_cycles",
	.event = "umask=0x4,period=200003,event=0x86",
	.desc = "Counts the number of core cycles the fetch stalls because of an icache miss. This is a cummulative count of core cycles the fetch stalled for all icache misses",
	.topic = "cache",
	.long_desc = "This event counts the number of core cycles the fetch stalls because of an icache miss. This is a cumulative count of cycles the NIP stalled for all icache misses",
},
{
	.name = "mem_uops_retired.l1_miss_loads",
	.event = "umask=0x1,period=200003,event=0x4",
	.desc = "Counts the number of load micro-ops retired that miss in L1 D cache",
	.topic = "cache",
	.long_desc = "This event counts the number of load micro-ops retired that miss in L1 Data cache. Note that prefetch misses will not be counted",
},
{
	.name = "mem_uops_retired.l2_hit_loads",
	.event = "umask=0x2,period=200003,event=0x4",
	.desc = "Counts the number of load micro-ops retired that hit in the L2  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.l2_miss_loads",
	.event = "umask=0x4,period=100007,event=0x4",
	.desc = "Counts the number of load micro-ops retired that miss in the L2  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.utlb_miss_loads",
	.event = "umask=0x10,period=200003,event=0x4",
	.desc = "Counts the number of load micro-ops retired that caused micro TLB miss",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.hitm",
	.event = "umask=0x20,period=200003,event=0x4",
	.desc = "Counts the loads retired that get the data from the other core in the same tile in M state  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x40,period=200003,event=0x4",
	.desc = "Counts all the load micro-ops retired",
	.topic = "cache",
	.long_desc = "This event counts the number of load micro-ops retired",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x80,period=200003,event=0x4",
	.desc = "Counts all the store micro-ops retired",
	.topic = "cache",
	.long_desc = "This event counts the number of store micro-ops retired",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100007,event=0xb7",
	.desc = "Counts the matrix events specified by MSR_OFFCORE_RESPx",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000070",
	.desc = "Counts any Prefetch requests that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400070",
	.desc = "Counts any Prefetch requests that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400070",
	.desc = "Counts any Prefetch requests that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080070",
	.desc = "Counts any Prefetch requests that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080070",
	.desc = "Counts any Prefetch requests that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010070",
	.desc = "Counts any Prefetch requests that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x40000032f7",
	.desc = "Counts any Read request  that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x10004032f7",
	.desc = "Counts any Read request  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x08004032f7",
	.desc = "Counts any Read request  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x10000832f7",
	.desc = "Counts any Read request  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x08000832f7",
	.desc = "Counts any Read request  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00000132f7",
	.desc = "Counts any Read request  that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000044",
	.desc = "Counts Demand code reads and prefetch code read requests  that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000022",
	.desc = "Counts Demand cacheable data write requests  that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010022",
	.desc = "Counts Demand cacheable data write requests  that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000003091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000403091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800403091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000083091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800083091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000013091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000008000",
	.desc = "Counts any request that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000408000",
	.desc = "Counts any request that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800408000",
	.desc = "Counts any request that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000088000",
	.desc = "Counts any request that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800088000",
	.desc = "Counts any request that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000018000",
	.desc = "Counts any request that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.streaming_stores.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000014800",
	.desc = "Counts all streaming stores (WC and should be programmed on PMC1) that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_streaming_stores.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000014000",
	.desc = "Counts Partial streaming stores (WC and should be programmed on PMC1) that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000002000",
	.desc = "Counts L1 data HW prefetches that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000402000",
	.desc = "Counts L1 data HW prefetches that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800402000",
	.desc = "Counts L1 data HW prefetches that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000082000",
	.desc = "Counts L1 data HW prefetches that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800082000",
	.desc = "Counts L1 data HW prefetches that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000012000",
	.desc = "Counts L1 data HW prefetches that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000001000",
	.desc = "Counts Software Prefetches that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000401000",
	.desc = "Counts Software Prefetches that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800401000",
	.desc = "Counts Software Prefetches that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000081000",
	.desc = "Counts Software Prefetches that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800081000",
	.desc = "Counts Software Prefetches that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000011000",
	.desc = "Counts Software Prefetches that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.full_streaming_stores.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010800",
	.desc = "Counts Full streaming stores (WC and should be programmed on PMC1) that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000400",
	.desc = "Counts Bus locks and split lock requests that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010400",
	.desc = "Counts Bus locks and split lock requests that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000040",
	.desc = "Counts L2 code HW prefetches that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400040",
	.desc = "Counts L2 code HW prefetches that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400040",
	.desc = "Counts L2 code HW prefetches that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080040",
	.desc = "Counts L2 code HW prefetches that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080040",
	.desc = "Counts L2 code HW prefetches that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010040",
	.desc = "Counts L2 code HW prefetches that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.supplier_none",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000020020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that provides no supplier details",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000004",
	.desc = "Counts demand code reads and prefetch code reads that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000002",
	.desc = "Counts Demand cacheable data writes that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400002",
	.desc = "Counts Demand cacheable data writes that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400002",
	.desc = "Counts Demand cacheable data writes that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080002",
	.desc = "Counts Demand cacheable data writes that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080002",
	.desc = "Counts Demand cacheable data writes that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010002",
	.desc = "Counts Demand cacheable data writes that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that are outstanding, per weighted cycle, from the time of the request to when any response is received. The oustanding response should be programmed only on PMC0",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit_far_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000400001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit_far_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800400001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses from a snoop request hit with data forwarded from its Far(not in the same quadrant as the request)-other tile's L2 in E/F state. Valid only for SNC4 cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit_near_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000080001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit_near_tile_e_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0800080001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses from a snoop request hit with data forwarded from its Near-other tile's L2 in E/F state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for any response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000002",
	.desc = "Counts Demand cacheable data writes that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002001000",
	.desc = "Counts Software Prefetches that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002002000",
	.desc = "Counts L1 data HW prefetches that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002008000",
	.desc = "Counts any request that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002003091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00020032f7",
	.desc = "Counts any Read request  that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.l2_hit_this_tile_m",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0002000070",
	.desc = "Counts any Prefetch requests that accounts for responses which hit its own tile's L2 with data in M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000002",
	.desc = "Counts Demand cacheable data writes that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000040",
	.desc = "Counts L2 code HW prefetches that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004001000",
	.desc = "Counts Software Prefetches that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004002000",
	.desc = "Counts L1 data HW prefetches that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004008000",
	.desc = "Counts any request that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004003091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00040032f7",
	.desc = "Counts any Read request  that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.l2_hit_this_tile_e",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0004000070",
	.desc = "Counts any Prefetch requests that accounts for responses which hit its own tile's L2 with data in E state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000002",
	.desc = "Counts Demand cacheable data writes that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008001000",
	.desc = "Counts Software Prefetches that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008002000",
	.desc = "Counts L1 data HW prefetches that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008008000",
	.desc = "Counts any request that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008003091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0008000044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_this_tile_s",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00080032f7",
	.desc = "Counts any Read request  that accounts for responses which hit its own tile's L2 with data in S state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000002",
	.desc = "Counts Demand cacheable data writes that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000040",
	.desc = "Counts L2 code HW prefetches that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010001000",
	.desc = "Counts Software Prefetches that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010002000",
	.desc = "Counts L1 data HW prefetches that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010008000",
	.desc = "Counts any request that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010003091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00100032f7",
	.desc = "Counts any Read request  that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.l2_hit_this_tile_f",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0010000070",
	.desc = "Counts any Prefetch requests that accounts for responses which hit its own tile's L2 with data in F state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180002",
	.desc = "Counts Demand cacheable data writes that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180040",
	.desc = "Counts L2 code HW prefetches that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.uc_code_reads.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180400",
	.desc = "Counts Bus locks and split lock requests that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800181000",
	.desc = "Counts Software Prefetches that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800182000",
	.desc = "Counts L1 data HW prefetches that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800188000",
	.desc = "Counts any request that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800183091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180022",
	.desc = "Counts Demand cacheable data write requests  that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x18001832f7",
	.desc = "Counts any Read request  that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.l2_hit_near_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800180070",
	.desc = "Counts any Prefetch requests that accounts for reponses from snoop request hit with data forwarded from its Near-other tile L2 in E/F/M state",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800400002",
	.desc = "Counts Demand cacheable data writes that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800400004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800400040",
	.desc = "Counts L2 code HW prefetches that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800400080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800400100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.bus_locks.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800400400",
	.desc = "Counts Bus locks and split lock requests that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_software.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800401000",
	.desc = "Counts Software Prefetches that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800402000",
	.desc = "Counts L1 data HW prefetches that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800408000",
	.desc = "Counts any request that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800403091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800400022",
	.desc = "Counts Demand cacheable data write requests  that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800400044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_read.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x18004032f7",
	.desc = "Counts any Read request  that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "offcore_response.any_pf_l2.l2_hit_far_tile",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1800400070",
	.desc = "Counts any Prefetch requests that accounts for reponses from snoop request hit with data forwarded from it Far(not in the same quadrant as the request)-other tile L2 in E/F/M state. Valid only in SNC4 Cluster mode",
	.topic = "cache",
},
{
	.name = "icache.accesses",
	.event = "umask=0x3,period=200003,event=0x80",
	.desc = "Counts all instruction fetches, including uncacheable fetches",
	.topic = "frontend",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=200003,event=0x80",
	.desc = "Counts all instruction fetches that hit the instruction cache",
	.topic = "frontend",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Counts all instruction fetches that miss the instruction cache or produce memory requests. An instruction fetch miss is counted only once and not once for every cycle it is outstanding",
	.topic = "frontend",
},
{
	.name = "ms_decoded.ms_entry",
	.event = "umask=0x1,period=200003,event=0xe7",
	.desc = "Counts the number of times the MSROM starts a flow of uops",
	.topic = "frontend",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=200003,event=0xc3",
	.desc = "Counts the number of times the machine clears due to memory ordering hazards",
	.topic = "memory",
},
{
	.name = "offcore_response.any_pf_l2.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400070",
	.desc = "Counts any Prefetch requests that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_pf_l2.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200070",
	.desc = "Counts any Prefetch requests that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_pf_l2.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000070",
	.desc = "Counts any Prefetch requests that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_pf_l2.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800070",
	.desc = "Counts any Prefetch requests that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_read.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x01004032f7",
	.desc = "Counts any Read request  that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_read.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00802032f7",
	.desc = "Counts any Read request  that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_read.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x01010032f7",
	.desc = "Counts any Read request  that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_read.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x00808032f7",
	.desc = "Counts any Read request  that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_code_rd.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_code_rd.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_code_rd.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_code_rd.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400022",
	.desc = "Counts Demand cacheable data write requests  that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200022",
	.desc = "Counts Demand cacheable data write requests  that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000022",
	.desc = "Counts Demand cacheable data write requests  that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800022",
	.desc = "Counts Demand cacheable data write requests  that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data_rd.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100403091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data_rd.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080203091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data_rd.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101003091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data_rd.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080803091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100408000",
	.desc = "Counts any request that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080208000",
	.desc = "Counts any request that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101008000",
	.desc = "Counts any request that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080808000",
	.desc = "Counts any request that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l1_data_rd.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100402000",
	.desc = "Counts L1 data HW prefetches that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l1_data_rd.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080202000",
	.desc = "Counts L1 data HW prefetches that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l1_data_rd.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101002000",
	.desc = "Counts L1 data HW prefetches that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l1_data_rd.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080802000",
	.desc = "Counts L1 data HW prefetches that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_software.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100401000",
	.desc = "Counts Software Prefetches that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_software.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080201000",
	.desc = "Counts Software Prefetches that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_software.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101001000",
	.desc = "Counts Software Prefetches that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_software.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080801000",
	.desc = "Counts Software Prefetches that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.bus_locks.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400400",
	.desc = "Counts Bus locks and split lock requests that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.bus_locks.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200400",
	.desc = "Counts Bus locks and split lock requests that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.bus_locks.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000400",
	.desc = "Counts Bus locks and split lock requests that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.bus_locks.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800400",
	.desc = "Counts Bus locks and split lock requests that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.uc_code_reads.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.uc_code_reads.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.uc_code_reads.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.uc_code_reads.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_writes.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_writes.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_writes.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_writes.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_reads.non_dram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x2000020080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses from any NON_DRAM system address. This includes MMIO transactions",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_reads.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_reads.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_reads.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_reads.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_code_rd.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400040",
	.desc = "Counts L2 code HW prefetches that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_code_rd.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200040",
	.desc = "Counts L2 code HW prefetches that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_code_rd.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000040",
	.desc = "Counts L2 code HW prefetches that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_code_rd.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800040",
	.desc = "Counts L2 code HW prefetches that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_rfo.non_dram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x2000020020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses from any NON_DRAM system address. This includes MMIO transactions",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_rfo.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_rfo.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_rfo.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_rfo.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400002",
	.desc = "Counts Demand cacheable data writes that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200002",
	.desc = "Counts Demand cacheable data writes that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000002",
	.desc = "Counts Demand cacheable data writes that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800002",
	.desc = "Counts Demand cacheable data writes that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.mcdram_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0100400001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for data responses from MCDRAM Far or Other tile L2 hit far",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.mcdram_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080200001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for data responses from MCDRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.ddr_far",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0101000001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for data responses from DRAM Far",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.ddr_near",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080800001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for data responses from DRAM Local",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600002",
	.desc = "Counts Demand cacheable data writes that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_rfo.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_reads.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_writes.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600100",
	.desc = "Counts Partial writes (UC or WT or WP and should be programmed on PMC1) that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.uc_code_reads.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.bus_locks.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_software.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180601000",
	.desc = "Counts Software Prefetches that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180608000",
	.desc = "Counts any request that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data_rd.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180603091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_code_rd.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_read.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x01806032f7",
	.desc = "Counts any Read request  that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_pf_l2.mcdram",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0180600070",
	.desc = "Counts any Prefetch requests that accounts for responses from MCDRAM (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800001",
	.desc = "Counts demand cacheable data and L1 prefetch data reads that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800002",
	.desc = "Counts Demand cacheable data writes that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800004",
	.desc = "Counts demand code reads and prefetch code reads that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_rfo.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800020",
	.desc = "Counts L2 data RFO prefetches (includes PREFETCHW instruction) that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_code_rd.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800040",
	.desc = "Counts L2 code HW prefetches that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.partial_reads.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800080",
	.desc = "Counts Partial reads (UC or WC and is valid only for Outstanding response type).  that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.uc_code_reads.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800200",
	.desc = "Counts UC code reads (valid only for Outstanding response type)  that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.bus_locks.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800400",
	.desc = "Counts Bus locks and split lock requests that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_software.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181801000",
	.desc = "Counts Software Prefetches that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l1_data_rd.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181802000",
	.desc = "Counts L1 data HW prefetches that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181808000",
	.desc = "Counts any request that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data_rd.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181803091",
	.desc = "Counts Demand cacheable data and L1 prefetch data read requests  that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800022",
	.desc = "Counts Demand cacheable data write requests  that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_code_rd.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0181800044",
	.desc = "Counts Demand code reads and prefetch code read requests  that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "offcore_response.any_read.ddr",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x01818032f7",
	.desc = "Counts any Read request  that accounts for responses from DDR (local and far)",
	.topic = "memory",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=200003,event=0xc4",
	.desc = "Counts the number of branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.jcc",
	.event = "umask=0x7e,period=200003,event=0xc4",
	.desc = "Counts the number of branch instructions retired that were conditional jumps (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.taken_jcc",
	.event = "umask=0xfe,period=200003,event=0xc4",
	.desc = "Counts the number of branch instructions retired that were conditional jumps and predicted taken (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.call",
	.event = "umask=0xf9,period=200003,event=0xc4",
	.desc = "Counts the number of near CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.rel_call",
	.event = "umask=0xfd,period=200003,event=0xc4",
	.desc = "Counts the number of near relative CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.ind_call",
	.event = "umask=0xfb,period=200003,event=0xc4",
	.desc = "Counts the number of near indirect CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.return",
	.event = "umask=0xf7,period=200003,event=0xc4",
	.desc = "Counts the number of near RET branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.non_return_ind",
	.event = "umask=0xeb,period=200003,event=0xc4",
	.desc = "Counts the number of branch instructions retired that were near indirect CALL or near indirect JMP (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0xbf,period=200003,event=0xc4",
	.desc = "Counts the number of far branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.jcc",
	.event = "umask=0x7e,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted branch instructions retired that were conditional jumps (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.taken_jcc",
	.event = "umask=0xfe,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted branch instructions retired that were conditional jumps and predicted taken (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.ind_call",
	.event = "umask=0xfb,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted near indirect CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.return",
	.event = "umask=0xf7,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted near RET branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.non_return_ind",
	.event = "umask=0xeb,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted branch instructions retired that were near indirect CALL or near indirect JMP (Precise event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.ms",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Counts the number of micro-ops retired that are from the complex flows issued by the micro-sequencer (MS)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of micro-ops retired that were supplied from MSROM",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x10,period=2000003,event=0xc2",
	.desc = "Counts the number of micro-ops retired",
	.topic = "pipeline",
	.long_desc = "This event counts the number of micro-ops (uops) retired. The processor decodes complex macro instructions into a sequence of simpler uops. Most instructions are composed of one or two uops. Some instructions are decoded into longer sequences such as repeat instructions, floating point transcendental instructions, and assists",
},
{
	.name = "uops_retired.scalar_simd",
	.event = "umask=0x20,period=200003,event=0xc2",
	.desc = "Counts the number of scalar SSE, AVX, AVX2, AVX-512 micro-ops retired. More specifically, it counts scalar SSE, AVX, AVX2, AVX-512 micro-ops except for loads (memory-to-register mov-type micro ops), division, sqrt",
	.topic = "pipeline",
	.long_desc = "This event counts the number of scalar SSE, AVX, AVX2, AVX-512 micro-ops retired (floating point, integer and store) except for loads (memory-to-register mov-type micro ops), division, sqrt",
},
{
	.name = "uops_retired.packed_simd",
	.event = "umask=0x40,period=200003,event=0xc2",
	.desc = "Counts the number of vector SSE, AVX, AVX2, AVX-512 micro-ops retired. More specifically, it counts packed SSE, AVX, AVX2, AVX-512 micro-ops (both floating point and integer) except for loads (memory-to-register mov-type micro-ops), packed byte and word multiplies",
	.topic = "pipeline",
	.long_desc = "This event counts the number of packed vector SSE, AVX, AVX2, and AVX-512 micro-ops retired (floating point, integer and store) except for loads (memory-to-register mov-type micro-ops), packed byte and word multiplies",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x1,period=200003,event=0xc3",
	.desc = "Counts the number of times that the machine clears due to program modifying data within 1K of a recently fetched code page",
	.topic = "pipeline",
},
{
	.name = "machine_clears.fp_assist",
	.event = "umask=0x4,period=200003,event=0xc3",
	.desc = "Counts the number of floating operations retired that required microcode assists",
	.topic = "pipeline",
	.long_desc = "This event counts the number of times that the pipeline stalled due to FP operations needing assists",
},
{
	.name = "machine_clears.all",
	.event = "umask=0x8,period=200003,event=0xc3",
	.desc = "Counts all nukes",
	.topic = "pipeline",
},
{
	.name = "no_alloc_cycles.rob_full",
	.event = "umask=0x1,period=200003,event=0xca",
	.desc = "Counts the number of core cycles when no micro-ops are allocated and the ROB is full",
	.topic = "pipeline",
},
{
	.name = "no_alloc_cycles.mispredicts",
	.event = "umask=0x4,period=200003,event=0xca",
	.desc = "Counts the number of core cycles when no micro-ops are allocated and the alloc pipe is stalled waiting for a mispredicted branch to retire",
	.topic = "pipeline",
	.long_desc = "This event counts the number of core cycles when no uops are allocated and the alloc pipe is stalled waiting for a mispredicted branch to retire",
},
{
	.name = "no_alloc_cycles.rat_stall",
	.event = "umask=0x20,period=200003,event=0xca",
	.desc = "Counts the number of core cycles when no micro-ops are allocated and a RATstall (caused by reservation station full) is asserted",
	.topic = "pipeline",
},
{
	.name = "no_alloc_cycles.not_delivered",
	.event = "umask=0x90,period=200003,event=0xca",
	.desc = "Counts the number of core cycles when no micro-ops are allocated, the IQ is empty, and no other condition is blocking allocation",
	.topic = "pipeline",
	.long_desc = "This event counts the number of core cycles when no uops are allocated, the instruction queue is empty and the alloc pipe is stalled waiting for instructions to be fetched",
},
{
	.name = "no_alloc_cycles.all",
	.event = "umask=0x7f,period=200003,event=0xca",
	.desc = "Counts the total number of core cycles when no micro-ops are allocated for any reason",
	.topic = "pipeline",
},
{
	.name = "rs_full_stall.mec",
	.event = "umask=0x1,period=200003,event=0xcb",
	.desc = "Counts the number of core cycles when allocation pipeline is stalled and is waiting for a free MEC reservation station entry",
	.topic = "pipeline",
},
{
	.name = "rs_full_stall.all",
	.event = "umask=0x1f,period=200003,event=0xcb",
	.desc = "Counts the total number of core cycles the Alloc pipeline is stalled when any one of the reservation stations is full",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Counts the total number of instructions retired",
	.topic = "pipeline",
},
{
	.name = "cycles_div_busy.all",
	.event = "umask=0x1,period=2000003,event=0xcd",
	.desc = "Cycles the number of core cycles when divider is busy.  Does not imply a stall waiting for the divider",
	.topic = "pipeline",
	.long_desc = "This event counts cycles when the divider is busy. More specifically cycles when the divide unit is unable to accept a new divide uop because it is busy processing a previously dispatched uop. The cycles will be counted irrespective of whether or not another divide uop is waiting to enter the divide unit (from the RS). This event counts integer divides, x87 divides, divss, divsd, sqrtss, sqrtsd event and does not count vector divides",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Fixed Counter: Counts the number of instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions that retire.  For instructions that consist of multiple micro-ops, this event counts exactly once, as the last micro-op of the instruction retires.  The event continues counting while instructions retire, including during interrupt service routines caused by hardware interrupts, faults or traps",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Counts the number of unhalted core clock cycles",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Counts the number of unhalted reference clock cycles",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Fixed Counter: Counts the number of unhalted core clock cycles",
	.topic = "pipeline",
	.long_desc = "This event counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Fixed Counter: Counts the number of unhalted reference clock cycles",
	.topic = "pipeline",
},
{
	.name = "baclears.all",
	.event = "umask=0x1,period=200003,event=0xe6",
	.desc = "Counts the number of times the front end resteers for any branch as a result of another branch handling mechanism in the front end",
	.topic = "pipeline",
},
{
	.name = "baclears.return",
	.event = "umask=0x8,period=200003,event=0xe6",
	.desc = "Counts the number of times the front end resteers for RET branches as a result of another branch handling mechanism in the front end",
	.topic = "pipeline",
},
{
	.name = "baclears.cond",
	.event = "umask=0x10,period=200003,event=0xe6",
	.desc = "Counts the number of times the front end resteers for conditional branches as a result of another branch handling mechanism in the front end",
	.topic = "pipeline",
},
{
	.name = "recycleq.ld_block_st_forward",
	.event = "umask=0x1,period=200003,event=0x3",
	.desc = "Counts the number of occurences a retired load gets blocked because its address partially overlaps with a store  Supports address when precise (Precise event)",
	.topic = "pipeline",
},
{
	.name = "recycleq.ld_block_std_notready",
	.event = "umask=0x2,period=200003,event=0x3",
	.desc = "Counts the number of occurences a retired load gets blocked because its address overlaps with a store whose data is not ready",
	.topic = "pipeline",
},
{
	.name = "recycleq.st_splits",
	.event = "umask=0x4,period=200003,event=0x3",
	.desc = "Counts the number of occurences a retired store that is a cache line split. Each split should be counted only once",
	.topic = "pipeline",
	.long_desc = "This event counts the number of retired store that experienced a cache line boundary split(Precise Event). Note that each spilt should be counted only once",
},
{
	.name = "recycleq.ld_splits",
	.event = "umask=0x8,period=200003,event=0x3",
	.desc = "Counts the number of occurences a retired load that is a cache line split. Each split should be counted only once  Supports address when precise (Precise event)",
	.topic = "pipeline",
},
{
	.name = "recycleq.lock",
	.event = "umask=0x10,period=200003,event=0x3",
	.desc = "Counts all the retired locked loads. It does not include stores because we would double count if we count stores",
	.topic = "pipeline",
},
{
	.name = "recycleq.sta_full",
	.event = "umask=0x20,period=200003,event=0x3",
	.desc = "Counts the store micro-ops retired that were pushed in the rehad queue because the store address buffer is full",
	.topic = "pipeline",
},
{
	.name = "recycleq.any_ld",
	.event = "umask=0x40,period=200003,event=0x3",
	.desc = "Counts any retired load that was pushed into the recycle queue for any reason",
	.topic = "pipeline",
},
{
	.name = "recycleq.any_st",
	.event = "umask=0x80,period=200003,event=0x3",
	.desc = "Counts any retired store that was pushed into the recycle queue for any reason",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.call",
	.event = "umask=0xf9,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted near CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.rel_call",
	.event = "umask=0xfd,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted near relative CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.far_branch",
	.event = "umask=0xbf,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted far branch instructions retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "unc_m_cas_count.rd",
	.event = "umask=0x01,event=0x3",
	.desc = "ddr bandwidth read (CPU traffic only) (MB/sec). Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "6.4e-05MiB",
	.perpkg = "1",
},
{
	.name = "unc_m_cas_count.wr",
	.event = "umask=0x02,event=0x3",
	.desc = "ddr bandwidth write (CPU traffic only) (MB/sec). Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "6.4e-05MiB",
	.perpkg = "1",
},
{
	.name = "unc_e_rpq_inserts",
	.event = "umask=0x01,event=0x1",
	.desc = "mcdram bandwidth read (CPU traffic only) (MB/sec). Unit: uncore_edc_eclk ",
	.topic = "uncore memory",
	.pmu = "uncore_edc_eclk",
	.unit = "6.4e-05MiB",
	.perpkg = "1",
},
{
	.name = "unc_e_wpq_inserts",
	.event = "umask=0x01,event=0x2",
	.desc = "mcdram bandwidth write (CPU traffic only) (MB/sec). Unit: uncore_edc_eclk ",
	.topic = "uncore memory",
	.pmu = "uncore_edc_eclk",
	.unit = "6.4e-05MiB",
	.perpkg = "1",
},
{
	.name = "mem_uops_retired.dtlb_miss_loads",
	.event = "umask=0x8,period=200003,event=0x4",
	.desc = "Counts the number of load micro-ops retired that cause a DTLB miss  Supports address when precise (Precise event)",
	.topic = "virtual memory",
},
{
	.name = "page_walks.d_side_walks",
	.event = "umask=0x1,period=100003,edge=1,event=0x5",
	.desc = "Counts the total D-side page walks that are completed or started. The page walks started in the speculative path will also be counted",
	.topic = "virtual memory",
},
{
	.name = "page_walks.d_side_cycles",
	.event = "umask=0x1,period=200003,event=0x5",
	.desc = "Counts the total number of core cycles for all the D-side page walks. The cycles for page walks started in speculative path will also be included",
	.topic = "virtual memory",
},
{
	.name = "page_walks.i_side_walks",
	.event = "umask=0x2,period=100003,edge=1,event=0x5",
	.desc = "Counts the total I-side page walks that are completed",
	.topic = "virtual memory",
},
{
	.name = "page_walks.i_side_cycles",
	.event = "umask=0x2,period=200003,event=0x5",
	.desc = "Counts the total number of core cycles for all the I-side page walks. The cycles for page walks started in speculative path will also be included",
	.topic = "virtual memory",
	.long_desc = "This event counts every cycle when an I-side (walks due to an instruction fetch) page walk is in progress",
},
{
	.name = "page_walks.walks",
	.event = "umask=0x3,period=100003,edge=1,event=0x5",
	.desc = "Counts the total page walks that are completed (I-side and D-side)",
	.topic = "virtual memory",
},
{
	.name = "page_walks.cycles",
	.event = "umask=0x3,period=200003,event=0x5",
	.desc = "Counts the total number of core cycles for all the page walks. The cycles for page walks started in speculative path will also be included",
	.topic = "virtual memory",
	.long_desc = "This event counts every cycle when a data (D) page walk or instruction (I) page walk is in progress",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_nehalemep[] = {
{
	.name = "cache_lock_cycles.l1d",
	.event = "umask=0x2,period=2000000,event=0x63",
	.desc = "Cycles L1D locked",
	.topic = "cache",
},
{
	.name = "cache_lock_cycles.l1d_l2",
	.event = "umask=0x1,period=2000000,event=0x63",
	.desc = "Cycles L1D and L2 locked",
	.topic = "cache",
},
{
	.name = "l1d.m_evict",
	.event = "umask=0x4,period=2000000,event=0x51",
	.desc = "L1D cache lines replaced in M state",
	.topic = "cache",
},
{
	.name = "l1d.m_repl",
	.event = "umask=0x2,period=2000000,event=0x51",
	.desc = "L1D cache lines allocated in the M state",
	.topic = "cache",
},
{
	.name = "l1d.m_snoop_evict",
	.event = "umask=0x8,period=2000000,event=0x51",
	.desc = "L1D snoop eviction of cache lines in M state",
	.topic = "cache",
},
{
	.name = "l1d.repl",
	.event = "umask=0x1,period=2000000,event=0x51",
	.desc = "L1 data cache lines allocated",
	.topic = "cache",
},
{
	.name = "l1d_all_ref.any",
	.event = "umask=0x1,period=2000000,event=0x43",
	.desc = "All references to the L1 data cache",
	.topic = "cache",
},
{
	.name = "l1d_all_ref.cacheable",
	.event = "umask=0x2,period=2000000,event=0x43",
	.desc = "L1 data cacheable reads and writes",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.e_state",
	.event = "umask=0x4,period=2000000,event=0x40",
	.desc = "L1 data cache read in E state",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.i_state",
	.event = "umask=0x1,period=2000000,event=0x40",
	.desc = "L1 data cache read in I state (misses)",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.m_state",
	.event = "umask=0x8,period=2000000,event=0x40",
	.desc = "L1 data cache read in M state",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.mesi",
	.event = "umask=0xf,period=2000000,event=0x40",
	.desc = "L1 data cache reads",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.s_state",
	.event = "umask=0x2,period=2000000,event=0x40",
	.desc = "L1 data cache read in S state",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock.e_state",
	.event = "umask=0x4,period=2000000,event=0x42",
	.desc = "L1 data cache load locks in E state",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock.hit",
	.event = "umask=0x1,period=2000000,event=0x42",
	.desc = "L1 data cache load lock hits",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock.m_state",
	.event = "umask=0x8,period=2000000,event=0x42",
	.desc = "L1 data cache load locks in M state",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock.s_state",
	.event = "umask=0x2,period=2000000,event=0x42",
	.desc = "L1 data cache load locks in S state",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock_fb_hit",
	.event = "umask=0x1,period=2000000,event=0x53",
	.desc = "L1D load lock accepted in fill buffer",
	.topic = "cache",
},
{
	.name = "l1d_cache_prefetch_lock_fb_hit",
	.event = "umask=0x1,period=2000000,event=0x52",
	.desc = "L1D prefetch load lock accepted in fill buffer",
	.topic = "cache",
},
{
	.name = "l1d_cache_st.e_state",
	.event = "umask=0x4,period=2000000,event=0x41",
	.desc = "L1 data cache stores in E state",
	.topic = "cache",
},
{
	.name = "l1d_cache_st.m_state",
	.event = "umask=0x8,period=2000000,event=0x41",
	.desc = "L1 data cache stores in M state",
	.topic = "cache",
},
{
	.name = "l1d_cache_st.s_state",
	.event = "umask=0x2,period=2000000,event=0x41",
	.desc = "L1 data cache stores in S state",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.miss",
	.event = "umask=0x2,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch misses",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.requests",
	.event = "umask=0x1,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.triggers",
	.event = "umask=0x4,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests triggered",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.e_state",
	.event = "umask=0x4,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in E state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.i_state",
	.event = "umask=0x1,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in I state (misses)",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.m_state",
	.event = "umask=0x8,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in M state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.mesi",
	.event = "umask=0xf,period=100000,event=0x28",
	.desc = "All L1 writebacks to L2",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.s_state",
	.event = "umask=0x2,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.any",
	.event = "umask=0xff,period=200000,event=0x26",
	.desc = "All L2 data requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.e_state",
	.event = "umask=0x4,period=200000,event=0x26",
	.desc = "L2 data demand loads in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.i_state",
	.event = "umask=0x1,period=200000,event=0x26",
	.desc = "L2 data demand loads in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.m_state",
	.event = "umask=0x8,period=200000,event=0x26",
	.desc = "L2 data demand loads in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.mesi",
	.event = "umask=0xf,period=200000,event=0x26",
	.desc = "L2 data demand requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.s_state",
	.event = "umask=0x2,period=200000,event=0x26",
	.desc = "L2 data demand loads in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.e_state",
	.event = "umask=0x40,period=200000,event=0x26",
	.desc = "L2 data prefetches in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.i_state",
	.event = "umask=0x10,period=200000,event=0x26",
	.desc = "L2 data prefetches in the I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.m_state",
	.event = "umask=0x80,period=200000,event=0x26",
	.desc = "L2 data prefetches in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.mesi",
	.event = "umask=0xf0,period=200000,event=0x26",
	.desc = "All L2 data prefetches",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.s_state",
	.event = "umask=0x20,period=200000,event=0x26",
	.desc = "L2 data prefetches in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.any",
	.event = "umask=0x7,period=100000,event=0xf1",
	.desc = "L2 lines alloacated",
	.topic = "cache",
},
{
	.name = "l2_lines_in.e_state",
	.event = "umask=0x4,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the E state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.s_state",
	.event = "umask=0x2,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_out.any",
	.event = "umask=0xf,period=100000,event=0xf2",
	.desc = "L2 lines evicted",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x1,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x2,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_clean",
	.event = "umask=0x4,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_dirty",
	.event = "umask=0x8,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_hit",
	.event = "umask=0x10,period=200000,event=0x24",
	.desc = "L2 instruction fetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_miss",
	.event = "umask=0x20,period=200000,event=0x24",
	.desc = "L2 instruction fetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetches",
	.event = "umask=0x30,period=200000,event=0x24",
	.desc = "L2 instruction fetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_hit",
	.event = "umask=0x1,period=200000,event=0x24",
	.desc = "L2 load hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_miss",
	.event = "umask=0x2,period=200000,event=0x24",
	.desc = "L2 load misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.loads",
	.event = "umask=0x3,period=200000,event=0x24",
	.desc = "L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0xaa,period=200000,event=0x24",
	.desc = "All L2 misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_hit",
	.event = "umask=0x40,period=200000,event=0x24",
	.desc = "L2 prefetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_miss",
	.event = "umask=0x80,period=200000,event=0x24",
	.desc = "L2 prefetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetches",
	.event = "umask=0xc0,period=200000,event=0x24",
	.desc = "All L2 prefetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200000,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x4,period=200000,event=0x24",
	.desc = "L2 RFO hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x8,period=200000,event=0x24",
	.desc = "L2 RFO misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfos",
	.event = "umask=0xc,period=200000,event=0x24",
	.desc = "L2 RFO requests",
	.topic = "cache",
},
{
	.name = "l2_transactions.any",
	.event = "umask=0x80,period=200000,event=0xf0",
	.desc = "All L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.fill",
	.event = "umask=0x20,period=200000,event=0xf0",
	.desc = "L2 fill transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.ifetch",
	.event = "umask=0x4,period=200000,event=0xf0",
	.desc = "L2 instruction fetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.l1d_wb",
	.event = "umask=0x10,period=200000,event=0xf0",
	.desc = "L1D writeback to L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.load",
	.event = "umask=0x1,period=200000,event=0xf0",
	.desc = "L2 Load transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.prefetch",
	.event = "umask=0x8,period=200000,event=0xf0",
	.desc = "L2 prefetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.rfo",
	.event = "umask=0x2,period=200000,event=0xf0",
	.desc = "L2 RFO transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.wb",
	.event = "umask=0x40,period=200000,event=0xf0",
	.desc = "L2 writeback to LLC transactions",
	.topic = "cache",
},
{
	.name = "l2_write.lock.e_state",
	.event = "umask=0x40,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in E state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.hit",
	.event = "umask=0xe0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.lock.i_state",
	.event = "umask=0x10,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.lock.m_state",
	.event = "umask=0x80,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.mesi",
	.event = "umask=0xf0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.lock.s_state",
	.event = "umask=0x20,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in S state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.hit",
	.event = "umask=0xe,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.i_state",
	.event = "umask=0x1,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.m_state",
	.event = "umask=0x8,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.mesi",
	.event = "umask=0xf,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.s_state",
	.event = "umask=0x2,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in S state",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100000,event=0x2e",
	.desc = "Longest latency cache miss",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=200000,event=0x2e",
	.desc = "Longest latency cache reference",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.loads",
	.event = "umask=0x1,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a load (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.stores",
	.event = "umask=0x2,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a store (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.hit_lfb",
	.event = "umask=0x40,period=200000,event=0xcb",
	.desc = "Retired loads that miss L1D and hit an previously allocated LFB (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l1d_hit",
	.event = "umask=0x1,period=2000000,event=0xcb",
	.desc = "Retired loads that hit the L1 data cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x2,period=200000,event=0xcb",
	.desc = "Retired loads that hit the L2 cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_miss",
	.event = "umask=0x10,period=10000,event=0xcb",
	.desc = "Retired loads that miss the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_unshared_hit",
	.event = "umask=0x4,period=40000,event=0xcb",
	.desc = "Retired loads that hit valid versions in the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.other_core_l2_hit_hitm",
	.event = "umask=0x8,period=40000,event=0xcb",
	.desc = "Retired loads that hit sibling core's L2 in modified or unmodified states (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.local_dram",
	.event = "umask=0x20,period=10000,event=0xf",
	.desc = "Load instructions retired with a data source of local DRAM or locally homed remote hitm (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.other_core_l2_hitm",
	.event = "umask=0x2,period=40000,event=0xf",
	.desc = "Load instructions retired that HIT modified data in sibling core (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.remote_cache_local_home_hit",
	.event = "umask=0x8,period=20000,event=0xf",
	.desc = "Load instructions retired remote cache HIT data source (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.remote_dram",
	.event = "umask=0x10,period=10000,event=0xf",
	.desc = "Load instructions retired remote DRAM and remote home-remote cache HITM (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.uncacheable",
	.event = "umask=0x80,period=4000,event=0xf",
	.desc = "Load instructions retired IO (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_requests.l1d_writeback",
	.event = "umask=0x40,period=100000,event=0xb0",
	.desc = "Offcore L1 data cache writebacks",
	.topic = "cache",
},
{
	.name = "offcore_requests_sq_full",
	.event = "umask=0x1,period=100000,event=0xb2",
	.desc = "Offcore requests blocked due to Super Queue full",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=2000000,event=0xf4",
	.desc = "Super Queue lock splits across a cache line",
	.topic = "cache",
},
{
	.name = "store_blocks.at_ret",
	.event = "umask=0x4,period=200000,event=0x6",
	.desc = "Loads delayed with at-Retirement block code",
	.topic = "cache",
},
{
	.name = "store_blocks.l1d_block",
	.event = "umask=0x8,period=200000,event=0x6",
	.desc = "Cacheable loads delayed with L1D block code",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_0",
	.event = "umask=0x10,period=2000000,event=0xb,ldlat=0x0",
	.desc = "Memory instructions retired above 0 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_1024",
	.event = "umask=0x10,period=100,event=0xb,ldlat=0x400",
	.desc = "Memory instructions retired above 1024 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_128",
	.event = "umask=0x10,period=1000,event=0xb,ldlat=0x80",
	.desc = "Memory instructions retired above 128 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16",
	.event = "umask=0x10,period=10000,event=0xb,ldlat=0x10",
	.desc = "Memory instructions retired above 16 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16384",
	.event = "umask=0x10,period=5,event=0xb,ldlat=0x4000",
	.desc = "Memory instructions retired above 16384 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_2048",
	.event = "umask=0x10,period=50,event=0xb,ldlat=0x800",
	.desc = "Memory instructions retired above 2048 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_256",
	.event = "umask=0x10,period=500,event=0xb,ldlat=0x100",
	.desc = "Memory instructions retired above 256 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32",
	.event = "umask=0x10,period=5000,event=0xb,ldlat=0x20",
	.desc = "Memory instructions retired above 32 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32768",
	.event = "umask=0x10,period=3,event=0xb,ldlat=0x8000",
	.desc = "Memory instructions retired above 32768 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4",
	.event = "umask=0x10,period=50000,event=0xb,ldlat=0x4",
	.desc = "Memory instructions retired above 4 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4096",
	.event = "umask=0x10,period=20,event=0xb,ldlat=0x1000",
	.desc = "Memory instructions retired above 4096 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_512",
	.event = "umask=0x10,period=200,event=0xb,ldlat=0x200",
	.desc = "Memory instructions retired above 512 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_64",
	.event = "umask=0x10,period=2000,event=0xb,ldlat=0x40",
	.desc = "Memory instructions retired above 64 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8",
	.event = "umask=0x10,period=20000,event=0xb,ldlat=0x8",
	.desc = "Memory instructions retired above 8 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8192",
	.event = "umask=0x10,period=10,event=0xb,ldlat=0x2000",
	.desc = "Memory instructions retired above 8192 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F11",
	.desc = "Offcore data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF11",
	.desc = "All offcore data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8011",
	.desc = "Offcore data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x111",
	.desc = "Offcore data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x211",
	.desc = "Offcore data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x411",
	.desc = "Offcore data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x711",
	.desc = "Offcore data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4711",
	.desc = "Offcore data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1811",
	.desc = "Offcore data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3811",
	.desc = "Offcore data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1011",
	.desc = "Offcore data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x811",
	.desc = "Offcore data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F44",
	.desc = "Offcore code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF44",
	.desc = "All offcore code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8044",
	.desc = "Offcore code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x144",
	.desc = "Offcore code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x244",
	.desc = "Offcore code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x444",
	.desc = "Offcore code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x744",
	.desc = "Offcore code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4744",
	.desc = "Offcore code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1844",
	.desc = "Offcore code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3844",
	.desc = "Offcore code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1044",
	.desc = "Offcore code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x844",
	.desc = "Offcore code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7FFF",
	.desc = "Offcore requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFFFF",
	.desc = "All offcore requests",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x80FF",
	.desc = "Offcore requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1FF",
	.desc = "Offcore requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2FF",
	.desc = "Offcore requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4FF",
	.desc = "Offcore requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7FF",
	.desc = "Offcore requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x47FF",
	.desc = "Offcore requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x18FF",
	.desc = "Offcore requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x38FF",
	.desc = "Offcore requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x10FF",
	.desc = "Offcore requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8FF",
	.desc = "Offcore requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F22",
	.desc = "Offcore RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF22",
	.desc = "All offcore RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8022",
	.desc = "Offcore RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x122",
	.desc = "Offcore RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x222",
	.desc = "Offcore RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x422",
	.desc = "Offcore RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x722",
	.desc = "Offcore RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4722",
	.desc = "Offcore RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1822",
	.desc = "Offcore RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3822",
	.desc = "Offcore RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1022",
	.desc = "Offcore RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x822",
	.desc = "Offcore RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F08",
	.desc = "Offcore writebacks to any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF08",
	.desc = "All offcore writebacks",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8008",
	.desc = "Offcore writebacks to the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x108",
	.desc = "Offcore writebacks to the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x408",
	.desc = "Offcore writebacks to the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x708",
	.desc = "Offcore writebacks to the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4708",
	.desc = "Offcore writebacks to the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1808",
	.desc = "Offcore writebacks to a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3808",
	.desc = "Offcore writebacks to a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1008",
	.desc = "Offcore writebacks that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x808",
	.desc = "Offcore writebacks that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F77",
	.desc = "Offcore code or data read requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF77",
	.desc = "All offcore code or data read requests",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8077",
	.desc = "Offcore code or data read requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x177",
	.desc = "Offcore code or data read requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x277",
	.desc = "Offcore code or data read requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x477",
	.desc = "Offcore code or data read requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x777",
	.desc = "Offcore code or data read requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4777",
	.desc = "Offcore code or data read requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1877",
	.desc = "Offcore code or data read requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3877",
	.desc = "Offcore code or data read requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1077",
	.desc = "Offcore code or data read requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x877",
	.desc = "Offcore code or data read requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F33",
	.desc = "Offcore request = all data, response = any cache_dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF33",
	.desc = "Offcore request = all data, response = any location",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8033",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x133",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x233",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x433",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x733",
	.desc = "Offcore request = all data, response = local cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4733",
	.desc = "Offcore request = all data, response = local cache or dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1833",
	.desc = "Offcore request = all data, response = remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3833",
	.desc = "Offcore request = all data, response = remote cache or dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1033",
	.desc = "Offcore data reads, RFO's and prefetches that HIT in a remote cache ",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x833",
	.desc = "Offcore data reads, RFO's and prefetches that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F03",
	.desc = "Offcore demand data requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF03",
	.desc = "All offcore demand data requests",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8003",
	.desc = "Offcore demand data requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x103",
	.desc = "Offcore demand data requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x203",
	.desc = "Offcore demand data requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x403",
	.desc = "Offcore demand data requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x703",
	.desc = "Offcore demand data requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4703",
	.desc = "Offcore demand data requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1803",
	.desc = "Offcore demand data requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3803",
	.desc = "Offcore demand data requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1003",
	.desc = "Offcore demand data requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x803",
	.desc = "Offcore demand data requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F01",
	.desc = "Offcore demand data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF01",
	.desc = "All offcore demand data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8001",
	.desc = "Offcore demand data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x101",
	.desc = "Offcore demand data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x201",
	.desc = "Offcore demand data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x401",
	.desc = "Offcore demand data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x701",
	.desc = "Offcore demand data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4701",
	.desc = "Offcore demand data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1801",
	.desc = "Offcore demand data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3801",
	.desc = "Offcore demand data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1001",
	.desc = "Offcore demand data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x801",
	.desc = "Offcore demand data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F04",
	.desc = "Offcore demand code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF04",
	.desc = "All offcore demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8004",
	.desc = "Offcore demand code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x104",
	.desc = "Offcore demand code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x204",
	.desc = "Offcore demand code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x404",
	.desc = "Offcore demand code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x704",
	.desc = "Offcore demand code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4704",
	.desc = "Offcore demand code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1804",
	.desc = "Offcore demand code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3804",
	.desc = "Offcore demand code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1004",
	.desc = "Offcore demand code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x804",
	.desc = "Offcore demand code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F02",
	.desc = "Offcore demand RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF02",
	.desc = "All offcore demand RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8002",
	.desc = "Offcore demand RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x102",
	.desc = "Offcore demand RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x202",
	.desc = "Offcore demand RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x402",
	.desc = "Offcore demand RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x702",
	.desc = "Offcore demand RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4702",
	.desc = "Offcore demand RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1802",
	.desc = "Offcore demand RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3802",
	.desc = "Offcore demand RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1002",
	.desc = "Offcore demand RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x802",
	.desc = "Offcore demand RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F80",
	.desc = "Offcore other requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF80",
	.desc = "All offcore other requests",
	.topic = "cache",
},
{
	.name = "offcore_response.other.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8080",
	.desc = "Offcore other requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x180",
	.desc = "Offcore other requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x280",
	.desc = "Offcore other requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x480",
	.desc = "Offcore other requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x780",
	.desc = "Offcore other requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4780",
	.desc = "Offcore other requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1880",
	.desc = "Offcore other requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3880",
	.desc = "Offcore other requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1080",
	.desc = "Offcore other requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x880",
	.desc = "Offcore other requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F30",
	.desc = "Offcore prefetch data requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF30",
	.desc = "All offcore prefetch data requests",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8030",
	.desc = "Offcore prefetch data requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x130",
	.desc = "Offcore prefetch data requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x230",
	.desc = "Offcore prefetch data requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x430",
	.desc = "Offcore prefetch data requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x730",
	.desc = "Offcore prefetch data requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4730",
	.desc = "Offcore prefetch data requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1830",
	.desc = "Offcore prefetch data requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3830",
	.desc = "Offcore prefetch data requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1030",
	.desc = "Offcore prefetch data requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x830",
	.desc = "Offcore prefetch data requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F10",
	.desc = "Offcore prefetch data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF10",
	.desc = "All offcore prefetch data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8010",
	.desc = "Offcore prefetch data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x110",
	.desc = "Offcore prefetch data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x210",
	.desc = "Offcore prefetch data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x410",
	.desc = "Offcore prefetch data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x710",
	.desc = "Offcore prefetch data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4710",
	.desc = "Offcore prefetch data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1810",
	.desc = "Offcore prefetch data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3810",
	.desc = "Offcore prefetch data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1010",
	.desc = "Offcore prefetch data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x810",
	.desc = "Offcore prefetch data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F40",
	.desc = "Offcore prefetch code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF40",
	.desc = "All offcore prefetch code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8040",
	.desc = "Offcore prefetch code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x140",
	.desc = "Offcore prefetch code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x240",
	.desc = "Offcore prefetch code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x440",
	.desc = "Offcore prefetch code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x740",
	.desc = "Offcore prefetch code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4740",
	.desc = "Offcore prefetch code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1840",
	.desc = "Offcore prefetch code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3840",
	.desc = "Offcore prefetch code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1040",
	.desc = "Offcore prefetch code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x840",
	.desc = "Offcore prefetch code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F20",
	.desc = "Offcore prefetch RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF20",
	.desc = "All offcore prefetch RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8020",
	.desc = "Offcore prefetch RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x120",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x220",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x420",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x720",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4720",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1820",
	.desc = "Offcore prefetch RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3820",
	.desc = "Offcore prefetch RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1020",
	.desc = "Offcore prefetch RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x820",
	.desc = "Offcore prefetch RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F70",
	.desc = "Offcore prefetch requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF70",
	.desc = "All offcore prefetch requests",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8070",
	.desc = "Offcore prefetch requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x170",
	.desc = "Offcore prefetch requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x270",
	.desc = "Offcore prefetch requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x470",
	.desc = "Offcore prefetch requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x770",
	.desc = "Offcore prefetch requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4770",
	.desc = "Offcore prefetch requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1870",
	.desc = "Offcore prefetch requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3870",
	.desc = "Offcore prefetch requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1070",
	.desc = "Offcore prefetch requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x870",
	.desc = "Offcore prefetch requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "fp_assist.all",
	.event = "umask=0x1,period=20000,event=0xf7",
	.desc = "X87 Floating point assists (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.input",
	.event = "umask=0x4,period=20000,event=0xf7",
	.desc = "X87 Floating poiint assists for invalid input value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.output",
	.event = "umask=0x2,period=20000,event=0xf7",
	.desc = "X87 Floating point assists for invalid output value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.mmx",
	.event = "umask=0x2,period=2000000,event=0x10",
	.desc = "MMX Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_double_precision",
	.event = "umask=0x80,period=2000000,event=0x10",
	.desc = "SSE* FP double precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp",
	.event = "umask=0x4,period=2000000,event=0x10",
	.desc = "SSE and SSE2 FP Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_packed",
	.event = "umask=0x10,period=2000000,event=0x10",
	.desc = "SSE FP packed Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_scalar",
	.event = "umask=0x20,period=2000000,event=0x10",
	.desc = "SSE FP scalar Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_single_precision",
	.event = "umask=0x40,period=2000000,event=0x10",
	.desc = "SSE* FP single precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse2_integer",
	.event = "umask=0x8,period=2000000,event=0x10",
	.desc = "SSE2 integer Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.x87",
	.event = "umask=0x1,period=2000000,event=0x10",
	.desc = "Computational floating-point operations executed",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.any",
	.event = "umask=0x3,period=2000000,event=0xcc",
	.desc = "All Floating Point to and from MMX transitions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_fp",
	.event = "umask=0x1,period=2000000,event=0xcc",
	.desc = "Transitions from MMX to Floating Point instructions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_mmx",
	.event = "umask=0x2,period=2000000,event=0xcc",
	.desc = "Transitions from Floating Point to MMX instructions",
	.topic = "floating point",
},
{
	.name = "simd_int_128.pack",
	.event = "umask=0x4,period=200000,event=0x12",
	.desc = "128 bit SIMD integer pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_arith",
	.event = "umask=0x20,period=200000,event=0x12",
	.desc = "128 bit SIMD integer arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_logical",
	.event = "umask=0x10,period=200000,event=0x12",
	.desc = "128 bit SIMD integer logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_mpy",
	.event = "umask=0x1,period=200000,event=0x12",
	.desc = "128 bit SIMD integer multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_shift",
	.event = "umask=0x2,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.shuffle_move",
	.event = "umask=0x40,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.unpack",
	.event = "umask=0x8,period=200000,event=0x12",
	.desc = "128 bit SIMD integer unpack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.pack",
	.event = "umask=0x4,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_arith",
	.event = "umask=0x20,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_logical",
	.event = "umask=0x10,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_mpy",
	.event = "umask=0x1,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit packed multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_shift",
	.event = "umask=0x2,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.shuffle_move",
	.event = "umask=0x40,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.unpack",
	.event = "umask=0x8,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit unpack operations",
	.topic = "floating point",
},
{
	.name = "macro_insts.decoded",
	.event = "umask=0x1,period=2000000,event=0xd0",
	.desc = "Instructions decoded",
	.topic = "frontend",
},
{
	.name = "macro_insts.fusions_decoded",
	.event = "umask=0x1,period=2000000,event=0xa6",
	.desc = "Macro-fused instructions decoded",
	.topic = "frontend",
},
{
	.name = "two_uop_insts_decoded",
	.event = "umask=0x1,period=2000000,event=0x19",
	.desc = "Two Uop instructions decoded",
	.topic = "frontend",
},
{
	.name = "offcore_response.any_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6011",
	.desc = "Offcore data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF811",
	.desc = "Offcore data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4011",
	.desc = "Offcore data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2011",
	.desc = "Offcore data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6044",
	.desc = "Offcore code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF844",
	.desc = "Offcore code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4044",
	.desc = "Offcore code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2044",
	.desc = "Offcore code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x60FF",
	.desc = "Offcore requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF8FF",
	.desc = "Offcore requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x40FF",
	.desc = "Offcore requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x20FF",
	.desc = "Offcore requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6022",
	.desc = "Offcore RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF822",
	.desc = "Offcore RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4022",
	.desc = "Offcore RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2022",
	.desc = "Offcore RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6008",
	.desc = "Offcore writebacks to any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF808",
	.desc = "Offcore writebacks that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4008",
	.desc = "Offcore writebacks to the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2008",
	.desc = "Offcore writebacks to a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6077",
	.desc = "Offcore code or data read requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF877",
	.desc = "Offcore code or data read requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4077",
	.desc = "Offcore code or data read requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2077",
	.desc = "Offcore code or data read requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6033",
	.desc = "Offcore request = all data, response = any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF833",
	.desc = "Offcore request = all data, response = any LLC miss",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4033",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2033",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6003",
	.desc = "Offcore demand data requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF803",
	.desc = "Offcore demand data requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4003",
	.desc = "Offcore demand data requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2003",
	.desc = "Offcore demand data requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6001",
	.desc = "Offcore demand data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF801",
	.desc = "Offcore demand data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4001",
	.desc = "Offcore demand data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2001",
	.desc = "Offcore demand data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6004",
	.desc = "Offcore demand code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF804",
	.desc = "Offcore demand code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4004",
	.desc = "Offcore demand code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2004",
	.desc = "Offcore demand code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6002",
	.desc = "Offcore demand RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF802",
	.desc = "Offcore demand RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4002",
	.desc = "Offcore demand RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2002",
	.desc = "Offcore demand RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6080",
	.desc = "Offcore other requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF880",
	.desc = "Offcore other requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.other.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2080",
	.desc = "Offcore other requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6030",
	.desc = "Offcore prefetch data requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF830",
	.desc = "Offcore prefetch data requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4030",
	.desc = "Offcore prefetch data requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2030",
	.desc = "Offcore prefetch data requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6010",
	.desc = "Offcore prefetch data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF810",
	.desc = "Offcore prefetch data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4010",
	.desc = "Offcore prefetch data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2010",
	.desc = "Offcore prefetch data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6040",
	.desc = "Offcore prefetch code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF840",
	.desc = "Offcore prefetch code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4040",
	.desc = "Offcore prefetch code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2040",
	.desc = "Offcore prefetch code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6020",
	.desc = "Offcore prefetch RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF820",
	.desc = "Offcore prefetch RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4020",
	.desc = "Offcore prefetch RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2020",
	.desc = "Offcore prefetch RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6070",
	.desc = "Offcore prefetch requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF870",
	.desc = "Offcore prefetch requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4070",
	.desc = "Offcore prefetch requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2070",
	.desc = "Offcore prefetch requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "bpu_clears.early",
	.event = "umask=0x1,period=2000000,event=0xe8",
	.desc = "Early Branch Prediciton Unit clears",
	.topic = "other",
},
{
	.name = "bpu_clears.late",
	.event = "umask=0x2,period=2000000,event=0xe8",
	.desc = "Late Branch Prediction Unit clears",
	.topic = "other",
},
{
	.name = "bpu_missed_call_ret",
	.event = "umask=0x1,period=2000000,event=0xe5",
	.desc = "Branch prediction unit missed call or return",
	.topic = "other",
},
{
	.name = "es_reg_renames",
	.event = "umask=0x1,period=2000000,event=0xd5",
	.desc = "ES segment renames",
	.topic = "other",
},
{
	.name = "io_transactions",
	.event = "umask=0x1,period=2000000,event=0x6c",
	.desc = "I/O transactions",
	.topic = "other",
},
{
	.name = "l1i.cycles_stalled",
	.event = "umask=0x4,period=2000000,event=0x80",
	.desc = "L1I instruction fetch stall cycles",
	.topic = "other",
},
{
	.name = "l1i.hits",
	.event = "umask=0x1,period=2000000,event=0x80",
	.desc = "L1I instruction fetch hits",
	.topic = "other",
},
{
	.name = "l1i.misses",
	.event = "umask=0x2,period=2000000,event=0x80",
	.desc = "L1I instruction fetch misses",
	.topic = "other",
},
{
	.name = "l1i.reads",
	.event = "umask=0x3,period=2000000,event=0x80",
	.desc = "L1I Instruction fetches",
	.topic = "other",
},
{
	.name = "large_itlb.hit",
	.event = "umask=0x1,period=200000,event=0x82",
	.desc = "Large ITLB hit",
	.topic = "other",
},
{
	.name = "load_dispatch.any",
	.event = "umask=0x7,period=2000000,event=0x13",
	.desc = "All loads dispatched",
	.topic = "other",
},
{
	.name = "load_dispatch.mob",
	.event = "umask=0x4,period=2000000,event=0x13",
	.desc = "Loads dispatched from the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs",
	.event = "umask=0x1,period=2000000,event=0x13",
	.desc = "Loads dispatched that bypass the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs_delayed",
	.event = "umask=0x2,period=2000000,event=0x13",
	.desc = "Loads dispatched from stage 305",
	.topic = "other",
},
{
	.name = "partial_address_alias",
	.event = "umask=0x1,period=200000,event=0x7",
	.desc = "False dependencies due to partial address aliasing",
	.topic = "other",
},
{
	.name = "rat_stalls.any",
	.event = "umask=0xf,period=2000000,event=0xd2",
	.desc = "All RAT stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.flags",
	.event = "umask=0x1,period=2000000,event=0xd2",
	.desc = "Flag stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.registers",
	.event = "umask=0x2,period=2000000,event=0xd2",
	.desc = "Partial register stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.rob_read_port",
	.event = "umask=0x4,period=2000000,event=0xd2",
	.desc = "ROB read port stalls cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.scoreboard",
	.event = "umask=0x8,period=2000000,event=0xd2",
	.desc = "Scoreboard stall cycles",
	.topic = "other",
},
{
	.name = "sb_drain.any",
	.event = "umask=0x7,period=200000,event=0x4",
	.desc = "All Store buffer stall cycles",
	.topic = "other",
},
{
	.name = "seg_rename_stalls",
	.event = "umask=0x1,period=2000000,event=0xd4",
	.desc = "Segment rename stall cycles",
	.topic = "other",
},
{
	.name = "snoop_response.hit",
	.event = "umask=0x1,period=100000,event=0xb8",
	.desc = "Thread responded HIT to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hite",
	.event = "umask=0x2,period=100000,event=0xb8",
	.desc = "Thread responded HITE to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hitm",
	.event = "umask=0x4,period=100000,event=0xb8",
	.desc = "Thread responded HITM to snoop",
	.topic = "other",
},
{
	.name = "sq_full_stall_cycles",
	.event = "umask=0x1,period=2000000,event=0xf6",
	.desc = "Super Queue full stall cycles",
	.topic = "other",
},
{
	.name = "arith.cycles_div_busy",
	.event = "umask=0x1,period=2000000,event=0x14",
	.desc = "Cycles the divider is busy",
	.topic = "pipeline",
},
{
	.name = "arith.div",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,edge=1,event=0x14",
	.desc = "Divide Operations executed",
	.topic = "pipeline",
},
{
	.name = "arith.mul",
	.event = "umask=0x2,period=2000000,event=0x14",
	.desc = "Multiply operations executed",
	.topic = "pipeline",
},
{
	.name = "baclear.bad_target",
	.event = "umask=0x2,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted with bad target address",
	.topic = "pipeline",
},
{
	.name = "baclear.clear",
	.event = "umask=0x1,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted, regardless of cause ",
	.topic = "pipeline",
},
{
	.name = "baclear_force_iq",
	.event = "umask=0x1,period=2000000,event=0xa7",
	.desc = "Instruction queue forced BACLEAR",
	.topic = "pipeline",
},
{
	.name = "br_inst_decoded",
	.event = "umask=0x1,period=2000000,event=0xe0",
	.desc = "Branch instructions decoded",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.any",
	.event = "umask=0x7f,period=200000,event=0x88",
	.desc = "Branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.cond",
	.event = "umask=0x1,period=200000,event=0x88",
	.desc = "Conditional branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct",
	.event = "umask=0x2,period=200000,event=0x88",
	.desc = "Unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct_near_call",
	.event = "umask=0x10,period=20000,event=0x88",
	.desc = "Unconditional call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_near_call",
	.event = "umask=0x20,period=20000,event=0x88",
	.desc = "Indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_non_call",
	.event = "umask=0x4,period=20000,event=0x88",
	.desc = "Indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.near_calls",
	.event = "umask=0x30,period=20000,event=0x88",
	.desc = "Call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.non_calls",
	.event = "umask=0x7,period=200000,event=0x88",
	.desc = "All non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.return_near",
	.event = "umask=0x8,period=20000,event=0x88",
	.desc = "Indirect return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken",
	.event = "umask=0x40,period=200000,event=0x88",
	.desc = "Taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x4,period=200000,event=0xc4",
	.desc = "Retired branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=200000,event=0xc4",
	.desc = "Retired conditional branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=20000,event=0xc4",
	.desc = "Retired near call instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.any",
	.event = "umask=0x7f,period=20000,event=0x89",
	.desc = "Mispredicted branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.cond",
	.event = "umask=0x1,period=20000,event=0x89",
	.desc = "Mispredicted conditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct",
	.event = "umask=0x2,period=20000,event=0x89",
	.desc = "Mispredicted unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct_near_call",
	.event = "umask=0x10,period=2000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_near_call",
	.event = "umask=0x20,period=2000,event=0x89",
	.desc = "Mispredicted indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_non_call",
	.event = "umask=0x4,period=2000,event=0x89",
	.desc = "Mispredicted indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.near_calls",
	.event = "umask=0x30,period=2000,event=0x89",
	.desc = "Mispredicted call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.non_calls",
	.event = "umask=0x7,period=20000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.return_near",
	.event = "umask=0x8,period=2000,event=0x89",
	.desc = "Mispredicted return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken",
	.event = "umask=0x40,period=20000,event=0x89",
	.desc = "Mispredicted taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=2000,event=0xc5",
	.desc = "Mispredicted near retired calls (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Reference cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_p",
	.event = "umask=0x1,period=100000,event=0x3c",
	.desc = "Reference base clock (133 Mhz) cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000000,event=0x3c",
	.desc = "Cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.total_cycles",
	.event = "inv=1,umask=0x0,period=2000000,cmask=2,event=0x3c",
	.desc = "Total CPU cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.any",
	.event = "umask=0xf,period=2000000,event=0x87",
	.desc = "Any Instruction Length Decoder stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000000,event=0x87",
	.desc = "Instruction Queue full stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000000,event=0x87",
	.desc = "Length Change Prefix stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.mru",
	.event = "umask=0x2,period=2000000,event=0x87",
	.desc = "Stall cycles due to BPU MRU bypass",
	.topic = "pipeline",
},
{
	.name = "ild_stall.regen",
	.event = "umask=0x8,period=2000000,event=0x87",
	.desc = "Regen stall cycles",
	.topic = "pipeline",
},
{
	.name = "inst_decoded.dec0",
	.event = "umask=0x1,period=2000000,event=0x18",
	.desc = "Instructions that must be decoded by decoder 0",
	.topic = "pipeline",
},
{
	.name = "inst_queue_write_cycles",
	.event = "umask=0x1,period=2000000,event=0x1e",
	.desc = "Cycles instructions are written to the instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_queue_writes",
	.event = "umask=0x1,period=2000000,event=0x17",
	.desc = "Instructions written to instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (Programmable counter and Precise Event) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.mmx",
	.event = "umask=0x4,period=2000000,event=0xc0",
	.desc = "Retired MMX instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000000,event=0xc0",
	.desc = "Retired floating-point operations (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre",
	.event = "umask=0x1,period=200000,event=0x4c",
	.desc = "Load operations conflicting with software prefetches",
	.topic = "pipeline",
},
{
	.name = "lsd.active",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles when uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.inactive",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles no uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd_overflow",
	.event = "umask=0x1,period=2000000,event=0x20",
	.desc = "Loops that can't stream from the instruction queue",
	.topic = "pipeline",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=20000,event=0xc3",
	.desc = "Cycles machine clear asserted",
	.topic = "pipeline",
},
{
	.name = "machine_clears.mem_order",
	.event = "umask=0x2,period=20000,event=0xc3",
	.desc = "Execution pipeline restart due to Memory ordering conflicts",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=20000,event=0xc3",
	.desc = "Self-Modifying Code detected",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000000,event=0xa2",
	.desc = "Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.fpcw",
	.event = "umask=0x20,period=2000000,event=0xa2",
	.desc = "FPU control word write stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.load",
	.event = "umask=0x2,period=2000000,event=0xa2",
	.desc = "Load buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.mxcsr",
	.event = "umask=0x40,period=2000000,event=0xa2",
	.desc = "MXCSR rename stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.other",
	.event = "umask=0x80,period=2000000,event=0xa2",
	.desc = "Other Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rob_full",
	.event = "umask=0x10,period=2000000,event=0xa2",
	.desc = "ROB full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rs_full",
	.event = "umask=0x4,period=2000000,event=0xa2",
	.desc = "Reservation Station full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.store",
	.event = "umask=0x8,period=2000000,event=0xa2",
	.desc = "Store buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_double",
	.event = "umask=0x4,period=200000,event=0xc7",
	.desc = "SIMD Packed-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_single",
	.event = "umask=0x1,period=200000,event=0xc7",
	.desc = "SIMD Packed-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_double",
	.event = "umask=0x8,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_single",
	.event = "umask=0x2,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.vector_integer",
	.event = "umask=0x10,period=200000,event=0xc7",
	.desc = "SIMD Vector Integer Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uop_unfusion",
	.event = "umask=0x1,period=2000000,event=0xdb",
	.desc = "Uop unfusions due to FP exceptions",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_folding",
	.event = "umask=0x4,period=2000000,event=0xd1",
	.desc = "Stack pointer instructions decoded",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_sync",
	.event = "umask=0x8,period=2000000,event=0xd1",
	.desc = "Stack pointer sync operations",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.ms_cycles_active",
	.event = "umask=0x2,period=2000000,cmask=1,event=0xd1",
	.desc = "Uops decoded by Microcode Sequencer",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xd1",
	.desc = "Cycles no Uops are decoded",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles",
	.event = "umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles_no_port5",
	.event = "umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count",
	.event = "inv=1,umask=0x3f,any=1,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count_no_port5",
	.event = "inv=1,umask=0x1f,any=1,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles",
	.event = "inv=1,umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles_no_port5",
	.event = "inv=1,umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port0",
	.event = "umask=0x1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015",
	.event = "umask=0x40,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015_stall_cycles",
	.event = "inv=1,umask=0x40,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port1",
	.event = "umask=0x2,period=2000000,event=0xb1",
	.desc = "Uops executed on port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port2_core",
	.event = "umask=0x4,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 2 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port234_core",
	.event = "umask=0x80,any=1,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 2, 3 or 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port3_core",
	.event = "umask=0x8,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 3 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port4_core",
	.event = "umask=0x10,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port5",
	.event = "umask=0x20,period=2000000,event=0xb1",
	.desc = "Uops executed on port 5",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000000,event=0xe",
	.desc = "Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued on any thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.cycles_all_threads",
	.event = "umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles Uops were issued on either thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.fused",
	.event = "umask=0x2,period=2000000,event=0xe",
	.desc = "Fused Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued",
	.topic = "pipeline",
},
{
	.name = "uops_retired.active_cycles",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are being retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.any",
	.event = "umask=0x1,period=2000000,event=0xc2",
	.desc = "Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.macro_fused",
	.event = "umask=0x4,period=2000000,event=0xc2",
	.desc = "Macro-fused Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000000,event=0xc2",
	.desc = "Retirement slots used (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are not retiring (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc2",
	.desc = "Total cycles using precise uop retired event (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles_ps",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "dtlb_load_misses.any",
	.event = "umask=0x1,period=200000,event=0x8",
	.desc = "DTLB load misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.pde_miss",
	.event = "umask=0x20,period=200000,event=0x8",
	.desc = "DTLB load miss caused by low part of address",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x10,period=2000000,event=0x8",
	.desc = "DTLB second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x8",
	.desc = "DTLB load miss page walks complete",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x49",
	.desc = "DTLB misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.stlb_hit",
	.event = "umask=0x10,period=200000,event=0x49",
	.desc = "DTLB first level misses but second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x49",
	.desc = "DTLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_flush",
	.event = "umask=0x1,period=2000000,event=0xae",
	.desc = "ITLB flushes",
	.topic = "virtual memory",
},
{
	.name = "itlb_miss_retired",
	.event = "umask=0x20,period=200000,event=0xc8",
	.desc = "Retired instructions that missed the ITLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x85",
	.desc = "ITLB miss",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x85",
	.desc = "ITLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "mem_load_retired.dtlb_miss",
	.event = "umask=0x80,period=200000,event=0xcb",
	.desc = "Retired loads that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "mem_store_retired.dtlb_miss",
	.event = "umask=0x1,period=200000,event=0xc",
	.desc = "Retired stores that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_nehalemex[] = {
{
	.name = "cache_lock_cycles.l1d",
	.event = "umask=0x2,period=2000000,event=0x63",
	.desc = "Cycles L1D locked",
	.topic = "cache",
},
{
	.name = "cache_lock_cycles.l1d_l2",
	.event = "umask=0x1,period=2000000,event=0x63",
	.desc = "Cycles L1D and L2 locked",
	.topic = "cache",
},
{
	.name = "l1d.m_evict",
	.event = "umask=0x4,period=2000000,event=0x51",
	.desc = "L1D cache lines replaced in M state",
	.topic = "cache",
},
{
	.name = "l1d.m_repl",
	.event = "umask=0x2,period=2000000,event=0x51",
	.desc = "L1D cache lines allocated in the M state",
	.topic = "cache",
},
{
	.name = "l1d.m_snoop_evict",
	.event = "umask=0x8,period=2000000,event=0x51",
	.desc = "L1D snoop eviction of cache lines in M state",
	.topic = "cache",
},
{
	.name = "l1d.repl",
	.event = "umask=0x1,period=2000000,event=0x51",
	.desc = "L1 data cache lines allocated",
	.topic = "cache",
},
{
	.name = "l1d_all_ref.any",
	.event = "umask=0x1,period=2000000,event=0x43",
	.desc = "All references to the L1 data cache",
	.topic = "cache",
},
{
	.name = "l1d_all_ref.cacheable",
	.event = "umask=0x2,period=2000000,event=0x43",
	.desc = "L1 data cacheable reads and writes",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.e_state",
	.event = "umask=0x4,period=2000000,event=0x40",
	.desc = "L1 data cache read in E state",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.i_state",
	.event = "umask=0x1,period=2000000,event=0x40",
	.desc = "L1 data cache read in I state (misses)",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.m_state",
	.event = "umask=0x8,period=2000000,event=0x40",
	.desc = "L1 data cache read in M state",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.mesi",
	.event = "umask=0xf,period=2000000,event=0x40",
	.desc = "L1 data cache reads",
	.topic = "cache",
},
{
	.name = "l1d_cache_ld.s_state",
	.event = "umask=0x2,period=2000000,event=0x40",
	.desc = "L1 data cache read in S state",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock.e_state",
	.event = "umask=0x4,period=2000000,event=0x42",
	.desc = "L1 data cache load locks in E state",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock.hit",
	.event = "umask=0x1,period=2000000,event=0x42",
	.desc = "L1 data cache load lock hits",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock.m_state",
	.event = "umask=0x8,period=2000000,event=0x42",
	.desc = "L1 data cache load locks in M state",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock.s_state",
	.event = "umask=0x2,period=2000000,event=0x42",
	.desc = "L1 data cache load locks in S state",
	.topic = "cache",
},
{
	.name = "l1d_cache_lock_fb_hit",
	.event = "umask=0x1,period=2000000,event=0x53",
	.desc = "L1D load lock accepted in fill buffer",
	.topic = "cache",
},
{
	.name = "l1d_cache_prefetch_lock_fb_hit",
	.event = "umask=0x1,period=2000000,event=0x52",
	.desc = "L1D prefetch load lock accepted in fill buffer",
	.topic = "cache",
},
{
	.name = "l1d_cache_st.e_state",
	.event = "umask=0x4,period=2000000,event=0x41",
	.desc = "L1 data cache stores in E state",
	.topic = "cache",
},
{
	.name = "l1d_cache_st.m_state",
	.event = "umask=0x8,period=2000000,event=0x41",
	.desc = "L1 data cache stores in M state",
	.topic = "cache",
},
{
	.name = "l1d_cache_st.s_state",
	.event = "umask=0x2,period=2000000,event=0x41",
	.desc = "L1 data cache stores in S state",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.miss",
	.event = "umask=0x2,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch misses",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.requests",
	.event = "umask=0x1,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.triggers",
	.event = "umask=0x4,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests triggered",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.e_state",
	.event = "umask=0x4,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in E state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.i_state",
	.event = "umask=0x1,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in I state (misses)",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.m_state",
	.event = "umask=0x8,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in M state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.mesi",
	.event = "umask=0xf,period=100000,event=0x28",
	.desc = "All L1 writebacks to L2",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.s_state",
	.event = "umask=0x2,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.any",
	.event = "umask=0xff,period=200000,event=0x26",
	.desc = "All L2 data requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.e_state",
	.event = "umask=0x4,period=200000,event=0x26",
	.desc = "L2 data demand loads in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.i_state",
	.event = "umask=0x1,period=200000,event=0x26",
	.desc = "L2 data demand loads in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.m_state",
	.event = "umask=0x8,period=200000,event=0x26",
	.desc = "L2 data demand loads in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.mesi",
	.event = "umask=0xf,period=200000,event=0x26",
	.desc = "L2 data demand requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.s_state",
	.event = "umask=0x2,period=200000,event=0x26",
	.desc = "L2 data demand loads in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.e_state",
	.event = "umask=0x40,period=200000,event=0x26",
	.desc = "L2 data prefetches in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.i_state",
	.event = "umask=0x10,period=200000,event=0x26",
	.desc = "L2 data prefetches in the I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.m_state",
	.event = "umask=0x80,period=200000,event=0x26",
	.desc = "L2 data prefetches in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.mesi",
	.event = "umask=0xf0,period=200000,event=0x26",
	.desc = "All L2 data prefetches",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.s_state",
	.event = "umask=0x20,period=200000,event=0x26",
	.desc = "L2 data prefetches in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.any",
	.event = "umask=0x7,period=100000,event=0xf1",
	.desc = "L2 lines alloacated",
	.topic = "cache",
},
{
	.name = "l2_lines_in.e_state",
	.event = "umask=0x4,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the E state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.s_state",
	.event = "umask=0x2,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_out.any",
	.event = "umask=0xf,period=100000,event=0xf2",
	.desc = "L2 lines evicted",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x1,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x2,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_clean",
	.event = "umask=0x4,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_dirty",
	.event = "umask=0x8,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_hit",
	.event = "umask=0x10,period=200000,event=0x24",
	.desc = "L2 instruction fetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_miss",
	.event = "umask=0x20,period=200000,event=0x24",
	.desc = "L2 instruction fetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetches",
	.event = "umask=0x30,period=200000,event=0x24",
	.desc = "L2 instruction fetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_hit",
	.event = "umask=0x1,period=200000,event=0x24",
	.desc = "L2 load hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_miss",
	.event = "umask=0x2,period=200000,event=0x24",
	.desc = "L2 load misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.loads",
	.event = "umask=0x3,period=200000,event=0x24",
	.desc = "L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0xaa,period=200000,event=0x24",
	.desc = "All L2 misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_hit",
	.event = "umask=0x40,period=200000,event=0x24",
	.desc = "L2 prefetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_miss",
	.event = "umask=0x80,period=200000,event=0x24",
	.desc = "L2 prefetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetches",
	.event = "umask=0xc0,period=200000,event=0x24",
	.desc = "All L2 prefetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200000,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x4,period=200000,event=0x24",
	.desc = "L2 RFO hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x8,period=200000,event=0x24",
	.desc = "L2 RFO misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfos",
	.event = "umask=0xc,period=200000,event=0x24",
	.desc = "L2 RFO requests",
	.topic = "cache",
},
{
	.name = "l2_transactions.any",
	.event = "umask=0x80,period=200000,event=0xf0",
	.desc = "All L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.fill",
	.event = "umask=0x20,period=200000,event=0xf0",
	.desc = "L2 fill transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.ifetch",
	.event = "umask=0x4,period=200000,event=0xf0",
	.desc = "L2 instruction fetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.l1d_wb",
	.event = "umask=0x10,period=200000,event=0xf0",
	.desc = "L1D writeback to L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.load",
	.event = "umask=0x1,period=200000,event=0xf0",
	.desc = "L2 Load transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.prefetch",
	.event = "umask=0x8,period=200000,event=0xf0",
	.desc = "L2 prefetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.rfo",
	.event = "umask=0x2,period=200000,event=0xf0",
	.desc = "L2 RFO transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.wb",
	.event = "umask=0x40,period=200000,event=0xf0",
	.desc = "L2 writeback to LLC transactions",
	.topic = "cache",
},
{
	.name = "l2_write.lock.e_state",
	.event = "umask=0x40,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in E state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.hit",
	.event = "umask=0xe0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.lock.i_state",
	.event = "umask=0x10,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.lock.m_state",
	.event = "umask=0x80,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.mesi",
	.event = "umask=0xf0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.lock.s_state",
	.event = "umask=0x20,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in S state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.hit",
	.event = "umask=0xe,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.i_state",
	.event = "umask=0x1,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.m_state",
	.event = "umask=0x8,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.mesi",
	.event = "umask=0xf,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.s_state",
	.event = "umask=0x2,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in S state",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100000,event=0x2e",
	.desc = "Longest latency cache miss",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=200000,event=0x2e",
	.desc = "Longest latency cache reference",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.loads",
	.event = "umask=0x1,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a load (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.stores",
	.event = "umask=0x2,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a store (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.hit_lfb",
	.event = "umask=0x40,period=200000,event=0xcb",
	.desc = "Retired loads that miss L1D and hit an previously allocated LFB (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l1d_hit",
	.event = "umask=0x1,period=2000000,event=0xcb",
	.desc = "Retired loads that hit the L1 data cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x2,period=200000,event=0xcb",
	.desc = "Retired loads that hit the L2 cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_miss",
	.event = "umask=0x10,period=10000,event=0xcb",
	.desc = "Retired loads that miss the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_unshared_hit",
	.event = "umask=0x4,period=40000,event=0xcb",
	.desc = "Retired loads that hit valid versions in the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.other_core_l2_hit_hitm",
	.event = "umask=0x8,period=40000,event=0xcb",
	.desc = "Retired loads that hit sibling core's L2 in modified or unmodified states (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_requests.l1d_writeback",
	.event = "umask=0x40,period=100000,event=0xb0",
	.desc = "Offcore L1 data cache writebacks",
	.topic = "cache",
},
{
	.name = "offcore_requests_sq_full",
	.event = "umask=0x1,period=100000,event=0xb2",
	.desc = "Offcore requests blocked due to Super Queue full",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=2000000,event=0xf4",
	.desc = "Super Queue lock splits across a cache line",
	.topic = "cache",
},
{
	.name = "store_blocks.at_ret",
	.event = "umask=0x4,period=200000,event=0x6",
	.desc = "Loads delayed with at-Retirement block code",
	.topic = "cache",
},
{
	.name = "store_blocks.l1d_block",
	.event = "umask=0x8,period=200000,event=0x6",
	.desc = "Cacheable loads delayed with L1D block code",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_0",
	.event = "umask=0x10,period=2000000,event=0xb,ldlat=0x0",
	.desc = "Memory instructions retired above 0 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_1024",
	.event = "umask=0x10,period=100,event=0xb,ldlat=0x400",
	.desc = "Memory instructions retired above 1024 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_128",
	.event = "umask=0x10,period=1000,event=0xb,ldlat=0x80",
	.desc = "Memory instructions retired above 128 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16",
	.event = "umask=0x10,period=10000,event=0xb,ldlat=0x10",
	.desc = "Memory instructions retired above 16 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16384",
	.event = "umask=0x10,period=5,event=0xb,ldlat=0x4000",
	.desc = "Memory instructions retired above 16384 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_2048",
	.event = "umask=0x10,period=50,event=0xb,ldlat=0x800",
	.desc = "Memory instructions retired above 2048 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_256",
	.event = "umask=0x10,period=500,event=0xb,ldlat=0x100",
	.desc = "Memory instructions retired above 256 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32",
	.event = "umask=0x10,period=5000,event=0xb,ldlat=0x20",
	.desc = "Memory instructions retired above 32 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32768",
	.event = "umask=0x10,period=3,event=0xb,ldlat=0x8000",
	.desc = "Memory instructions retired above 32768 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4",
	.event = "umask=0x10,period=50000,event=0xb,ldlat=0x4",
	.desc = "Memory instructions retired above 4 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4096",
	.event = "umask=0x10,period=20,event=0xb,ldlat=0x1000",
	.desc = "Memory instructions retired above 4096 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_512",
	.event = "umask=0x10,period=200,event=0xb,ldlat=0x200",
	.desc = "Memory instructions retired above 512 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_64",
	.event = "umask=0x10,period=2000,event=0xb,ldlat=0x40",
	.desc = "Memory instructions retired above 64 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8",
	.event = "umask=0x10,period=20000,event=0xb,ldlat=0x8",
	.desc = "Memory instructions retired above 8 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8192",
	.event = "umask=0x10,period=10,event=0xb,ldlat=0x2000",
	.desc = "Memory instructions retired above 8192 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F11",
	.desc = "Offcore data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF11",
	.desc = "All offcore data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8011",
	.desc = "Offcore data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x111",
	.desc = "Offcore data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x211",
	.desc = "Offcore data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x411",
	.desc = "Offcore data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x711",
	.desc = "Offcore data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4711",
	.desc = "Offcore data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1811",
	.desc = "Offcore data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3811",
	.desc = "Offcore data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1011",
	.desc = "Offcore data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x811",
	.desc = "Offcore data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F44",
	.desc = "Offcore code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF44",
	.desc = "All offcore code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8044",
	.desc = "Offcore code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x144",
	.desc = "Offcore code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x244",
	.desc = "Offcore code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x444",
	.desc = "Offcore code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x744",
	.desc = "Offcore code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4744",
	.desc = "Offcore code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1844",
	.desc = "Offcore code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3844",
	.desc = "Offcore code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1044",
	.desc = "Offcore code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x844",
	.desc = "Offcore code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7FFF",
	.desc = "Offcore requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFFFF",
	.desc = "All offcore requests",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x80FF",
	.desc = "Offcore requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1FF",
	.desc = "Offcore requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2FF",
	.desc = "Offcore requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4FF",
	.desc = "Offcore requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7FF",
	.desc = "Offcore requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x47FF",
	.desc = "Offcore requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x18FF",
	.desc = "Offcore requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x38FF",
	.desc = "Offcore requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x10FF",
	.desc = "Offcore requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8FF",
	.desc = "Offcore requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F22",
	.desc = "Offcore RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF22",
	.desc = "All offcore RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8022",
	.desc = "Offcore RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x122",
	.desc = "Offcore RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x222",
	.desc = "Offcore RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x422",
	.desc = "Offcore RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x722",
	.desc = "Offcore RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4722",
	.desc = "Offcore RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1822",
	.desc = "Offcore RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3822",
	.desc = "Offcore RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1022",
	.desc = "Offcore RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x822",
	.desc = "Offcore RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F08",
	.desc = "Offcore writebacks to any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF08",
	.desc = "All offcore writebacks",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8008",
	.desc = "Offcore writebacks to the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x108",
	.desc = "Offcore writebacks to the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x408",
	.desc = "Offcore writebacks to the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x708",
	.desc = "Offcore writebacks to the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4708",
	.desc = "Offcore writebacks to the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1808",
	.desc = "Offcore writebacks to a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3808",
	.desc = "Offcore writebacks to a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1008",
	.desc = "Offcore writebacks that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x808",
	.desc = "Offcore writebacks that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F77",
	.desc = "Offcore code or data read requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF77",
	.desc = "All offcore code or data read requests",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8077",
	.desc = "Offcore code or data read requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x177",
	.desc = "Offcore code or data read requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x277",
	.desc = "Offcore code or data read requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x477",
	.desc = "Offcore code or data read requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x777",
	.desc = "Offcore code or data read requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4777",
	.desc = "Offcore code or data read requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1877",
	.desc = "Offcore code or data read requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3877",
	.desc = "Offcore code or data read requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1077",
	.desc = "Offcore code or data read requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x877",
	.desc = "Offcore code or data read requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F33",
	.desc = "Offcore request = all data, response = any cache_dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF33",
	.desc = "Offcore request = all data, response = any location",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8033",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x133",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x233",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x433",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x733",
	.desc = "Offcore request = all data, response = local cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4733",
	.desc = "Offcore request = all data, response = local cache or dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1833",
	.desc = "Offcore request = all data, response = remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3833",
	.desc = "Offcore request = all data, response = remote cache or dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1033",
	.desc = "Offcore data reads, RFO's and prefetches that HIT in a remote cache ",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x833",
	.desc = "Offcore data reads, RFO's and prefetches that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F03",
	.desc = "Offcore demand data requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF03",
	.desc = "All offcore demand data requests",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8003",
	.desc = "Offcore demand data requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x103",
	.desc = "Offcore demand data requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x203",
	.desc = "Offcore demand data requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x403",
	.desc = "Offcore demand data requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x703",
	.desc = "Offcore demand data requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4703",
	.desc = "Offcore demand data requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1803",
	.desc = "Offcore demand data requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3803",
	.desc = "Offcore demand data requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1003",
	.desc = "Offcore demand data requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x803",
	.desc = "Offcore demand data requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F01",
	.desc = "Offcore demand data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF01",
	.desc = "All offcore demand data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8001",
	.desc = "Offcore demand data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x101",
	.desc = "Offcore demand data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x201",
	.desc = "Offcore demand data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x401",
	.desc = "Offcore demand data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x701",
	.desc = "Offcore demand data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4701",
	.desc = "Offcore demand data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1801",
	.desc = "Offcore demand data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3801",
	.desc = "Offcore demand data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1001",
	.desc = "Offcore demand data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x801",
	.desc = "Offcore demand data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F04",
	.desc = "Offcore demand code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF04",
	.desc = "All offcore demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8004",
	.desc = "Offcore demand code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x104",
	.desc = "Offcore demand code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x204",
	.desc = "Offcore demand code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x404",
	.desc = "Offcore demand code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x704",
	.desc = "Offcore demand code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4704",
	.desc = "Offcore demand code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1804",
	.desc = "Offcore demand code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3804",
	.desc = "Offcore demand code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1004",
	.desc = "Offcore demand code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x804",
	.desc = "Offcore demand code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F02",
	.desc = "Offcore demand RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF02",
	.desc = "All offcore demand RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8002",
	.desc = "Offcore demand RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x102",
	.desc = "Offcore demand RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x202",
	.desc = "Offcore demand RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x402",
	.desc = "Offcore demand RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x702",
	.desc = "Offcore demand RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4702",
	.desc = "Offcore demand RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1802",
	.desc = "Offcore demand RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3802",
	.desc = "Offcore demand RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1002",
	.desc = "Offcore demand RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x802",
	.desc = "Offcore demand RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F80",
	.desc = "Offcore other requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF80",
	.desc = "All offcore other requests",
	.topic = "cache",
},
{
	.name = "offcore_response.other.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8080",
	.desc = "Offcore other requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x180",
	.desc = "Offcore other requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x280",
	.desc = "Offcore other requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x480",
	.desc = "Offcore other requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x780",
	.desc = "Offcore other requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4780",
	.desc = "Offcore other requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1880",
	.desc = "Offcore other requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3880",
	.desc = "Offcore other requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1080",
	.desc = "Offcore other requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x880",
	.desc = "Offcore other requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F30",
	.desc = "Offcore prefetch data requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF30",
	.desc = "All offcore prefetch data requests",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8030",
	.desc = "Offcore prefetch data requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x130",
	.desc = "Offcore prefetch data requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x230",
	.desc = "Offcore prefetch data requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x430",
	.desc = "Offcore prefetch data requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x730",
	.desc = "Offcore prefetch data requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4730",
	.desc = "Offcore prefetch data requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1830",
	.desc = "Offcore prefetch data requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3830",
	.desc = "Offcore prefetch data requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1030",
	.desc = "Offcore prefetch data requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x830",
	.desc = "Offcore prefetch data requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F10",
	.desc = "Offcore prefetch data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF10",
	.desc = "All offcore prefetch data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8010",
	.desc = "Offcore prefetch data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x110",
	.desc = "Offcore prefetch data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x210",
	.desc = "Offcore prefetch data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x410",
	.desc = "Offcore prefetch data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x710",
	.desc = "Offcore prefetch data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4710",
	.desc = "Offcore prefetch data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1810",
	.desc = "Offcore prefetch data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3810",
	.desc = "Offcore prefetch data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1010",
	.desc = "Offcore prefetch data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x810",
	.desc = "Offcore prefetch data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F40",
	.desc = "Offcore prefetch code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF40",
	.desc = "All offcore prefetch code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8040",
	.desc = "Offcore prefetch code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x140",
	.desc = "Offcore prefetch code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x240",
	.desc = "Offcore prefetch code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x440",
	.desc = "Offcore prefetch code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x740",
	.desc = "Offcore prefetch code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4740",
	.desc = "Offcore prefetch code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1840",
	.desc = "Offcore prefetch code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3840",
	.desc = "Offcore prefetch code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1040",
	.desc = "Offcore prefetch code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x840",
	.desc = "Offcore prefetch code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F20",
	.desc = "Offcore prefetch RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF20",
	.desc = "All offcore prefetch RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8020",
	.desc = "Offcore prefetch RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x120",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x220",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x420",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x720",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4720",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1820",
	.desc = "Offcore prefetch RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3820",
	.desc = "Offcore prefetch RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1020",
	.desc = "Offcore prefetch RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x820",
	.desc = "Offcore prefetch RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F70",
	.desc = "Offcore prefetch requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF70",
	.desc = "All offcore prefetch requests",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8070",
	.desc = "Offcore prefetch requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x170",
	.desc = "Offcore prefetch requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x270",
	.desc = "Offcore prefetch requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x470",
	.desc = "Offcore prefetch requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x770",
	.desc = "Offcore prefetch requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4770",
	.desc = "Offcore prefetch requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1870",
	.desc = "Offcore prefetch requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3870",
	.desc = "Offcore prefetch requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1070",
	.desc = "Offcore prefetch requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x870",
	.desc = "Offcore prefetch requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "fp_assist.all",
	.event = "umask=0x1,period=20000,event=0xf7",
	.desc = "X87 Floating point assists (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.input",
	.event = "umask=0x4,period=20000,event=0xf7",
	.desc = "X87 Floating poiint assists for invalid input value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.output",
	.event = "umask=0x2,period=20000,event=0xf7",
	.desc = "X87 Floating point assists for invalid output value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.mmx",
	.event = "umask=0x2,period=2000000,event=0x10",
	.desc = "MMX Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_double_precision",
	.event = "umask=0x80,period=2000000,event=0x10",
	.desc = "SSE* FP double precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp",
	.event = "umask=0x4,period=2000000,event=0x10",
	.desc = "SSE and SSE2 FP Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_packed",
	.event = "umask=0x10,period=2000000,event=0x10",
	.desc = "SSE FP packed Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_scalar",
	.event = "umask=0x20,period=2000000,event=0x10",
	.desc = "SSE FP scalar Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_single_precision",
	.event = "umask=0x40,period=2000000,event=0x10",
	.desc = "SSE* FP single precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse2_integer",
	.event = "umask=0x8,period=2000000,event=0x10",
	.desc = "SSE2 integer Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.x87",
	.event = "umask=0x1,period=2000000,event=0x10",
	.desc = "Computational floating-point operations executed",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.any",
	.event = "umask=0x3,period=2000000,event=0xcc",
	.desc = "All Floating Point to and from MMX transitions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_fp",
	.event = "umask=0x1,period=2000000,event=0xcc",
	.desc = "Transitions from MMX to Floating Point instructions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_mmx",
	.event = "umask=0x2,period=2000000,event=0xcc",
	.desc = "Transitions from Floating Point to MMX instructions",
	.topic = "floating point",
},
{
	.name = "simd_int_128.pack",
	.event = "umask=0x4,period=200000,event=0x12",
	.desc = "128 bit SIMD integer pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_arith",
	.event = "umask=0x20,period=200000,event=0x12",
	.desc = "128 bit SIMD integer arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_logical",
	.event = "umask=0x10,period=200000,event=0x12",
	.desc = "128 bit SIMD integer logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_mpy",
	.event = "umask=0x1,period=200000,event=0x12",
	.desc = "128 bit SIMD integer multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_shift",
	.event = "umask=0x2,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.shuffle_move",
	.event = "umask=0x40,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.unpack",
	.event = "umask=0x8,period=200000,event=0x12",
	.desc = "128 bit SIMD integer unpack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.pack",
	.event = "umask=0x4,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_arith",
	.event = "umask=0x20,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_logical",
	.event = "umask=0x10,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_mpy",
	.event = "umask=0x1,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit packed multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_shift",
	.event = "umask=0x2,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.shuffle_move",
	.event = "umask=0x40,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.unpack",
	.event = "umask=0x8,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit unpack operations",
	.topic = "floating point",
},
{
	.name = "macro_insts.decoded",
	.event = "umask=0x1,period=2000000,event=0xd0",
	.desc = "Instructions decoded",
	.topic = "frontend",
},
{
	.name = "macro_insts.fusions_decoded",
	.event = "umask=0x1,period=2000000,event=0xa6",
	.desc = "Macro-fused instructions decoded",
	.topic = "frontend",
},
{
	.name = "two_uop_insts_decoded",
	.event = "umask=0x1,period=2000000,event=0x19",
	.desc = "Two Uop instructions decoded",
	.topic = "frontend",
},
{
	.name = "offcore_response.any_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6011",
	.desc = "Offcore data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF811",
	.desc = "Offcore data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4011",
	.desc = "Offcore data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2011",
	.desc = "Offcore data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6044",
	.desc = "Offcore code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF844",
	.desc = "Offcore code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4044",
	.desc = "Offcore code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2044",
	.desc = "Offcore code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x60FF",
	.desc = "Offcore requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF8FF",
	.desc = "Offcore requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x40FF",
	.desc = "Offcore requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x20FF",
	.desc = "Offcore requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6022",
	.desc = "Offcore RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF822",
	.desc = "Offcore RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4022",
	.desc = "Offcore RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2022",
	.desc = "Offcore RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6008",
	.desc = "Offcore writebacks to any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF808",
	.desc = "Offcore writebacks that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4008",
	.desc = "Offcore writebacks to the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2008",
	.desc = "Offcore writebacks to a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6077",
	.desc = "Offcore code or data read requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF877",
	.desc = "Offcore code or data read requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4077",
	.desc = "Offcore code or data read requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2077",
	.desc = "Offcore code or data read requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6033",
	.desc = "Offcore request = all data, response = any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF833",
	.desc = "Offcore request = all data, response = any LLC miss",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4033",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2033",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6003",
	.desc = "Offcore demand data requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF803",
	.desc = "Offcore demand data requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4003",
	.desc = "Offcore demand data requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2003",
	.desc = "Offcore demand data requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6001",
	.desc = "Offcore demand data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF801",
	.desc = "Offcore demand data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4001",
	.desc = "Offcore demand data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2001",
	.desc = "Offcore demand data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6004",
	.desc = "Offcore demand code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF804",
	.desc = "Offcore demand code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4004",
	.desc = "Offcore demand code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2004",
	.desc = "Offcore demand code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6002",
	.desc = "Offcore demand RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF802",
	.desc = "Offcore demand RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4002",
	.desc = "Offcore demand RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2002",
	.desc = "Offcore demand RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6080",
	.desc = "Offcore other requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF880",
	.desc = "Offcore other requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.other.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2080",
	.desc = "Offcore other requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6030",
	.desc = "Offcore prefetch data requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF830",
	.desc = "Offcore prefetch data requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4030",
	.desc = "Offcore prefetch data requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2030",
	.desc = "Offcore prefetch data requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6010",
	.desc = "Offcore prefetch data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF810",
	.desc = "Offcore prefetch data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4010",
	.desc = "Offcore prefetch data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2010",
	.desc = "Offcore prefetch data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6040",
	.desc = "Offcore prefetch code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF840",
	.desc = "Offcore prefetch code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4040",
	.desc = "Offcore prefetch code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2040",
	.desc = "Offcore prefetch code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6020",
	.desc = "Offcore prefetch RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF820",
	.desc = "Offcore prefetch RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4020",
	.desc = "Offcore prefetch RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2020",
	.desc = "Offcore prefetch RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6070",
	.desc = "Offcore prefetch requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF870",
	.desc = "Offcore prefetch requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4070",
	.desc = "Offcore prefetch requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2070",
	.desc = "Offcore prefetch requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "bpu_clears.early",
	.event = "umask=0x1,period=2000000,event=0xe8",
	.desc = "Early Branch Prediciton Unit clears",
	.topic = "other",
},
{
	.name = "bpu_clears.late",
	.event = "umask=0x2,period=2000000,event=0xe8",
	.desc = "Late Branch Prediction Unit clears",
	.topic = "other",
},
{
	.name = "bpu_missed_call_ret",
	.event = "umask=0x1,period=2000000,event=0xe5",
	.desc = "Branch prediction unit missed call or return",
	.topic = "other",
},
{
	.name = "es_reg_renames",
	.event = "umask=0x1,period=2000000,event=0xd5",
	.desc = "ES segment renames",
	.topic = "other",
},
{
	.name = "io_transactions",
	.event = "umask=0x1,period=2000000,event=0x6c",
	.desc = "I/O transactions",
	.topic = "other",
},
{
	.name = "l1i.cycles_stalled",
	.event = "umask=0x4,period=2000000,event=0x80",
	.desc = "L1I instruction fetch stall cycles",
	.topic = "other",
},
{
	.name = "l1i.hits",
	.event = "umask=0x1,period=2000000,event=0x80",
	.desc = "L1I instruction fetch hits",
	.topic = "other",
},
{
	.name = "l1i.misses",
	.event = "umask=0x2,period=2000000,event=0x80",
	.desc = "L1I instruction fetch misses",
	.topic = "other",
},
{
	.name = "l1i.reads",
	.event = "umask=0x3,period=2000000,event=0x80",
	.desc = "L1I Instruction fetches",
	.topic = "other",
},
{
	.name = "large_itlb.hit",
	.event = "umask=0x1,period=200000,event=0x82",
	.desc = "Large ITLB hit",
	.topic = "other",
},
{
	.name = "load_dispatch.any",
	.event = "umask=0x7,period=2000000,event=0x13",
	.desc = "All loads dispatched",
	.topic = "other",
},
{
	.name = "load_dispatch.mob",
	.event = "umask=0x4,period=2000000,event=0x13",
	.desc = "Loads dispatched from the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs",
	.event = "umask=0x1,period=2000000,event=0x13",
	.desc = "Loads dispatched that bypass the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs_delayed",
	.event = "umask=0x2,period=2000000,event=0x13",
	.desc = "Loads dispatched from stage 305",
	.topic = "other",
},
{
	.name = "partial_address_alias",
	.event = "umask=0x1,period=200000,event=0x7",
	.desc = "False dependencies due to partial address aliasing",
	.topic = "other",
},
{
	.name = "rat_stalls.any",
	.event = "umask=0xf,period=2000000,event=0xd2",
	.desc = "All RAT stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.flags",
	.event = "umask=0x1,period=2000000,event=0xd2",
	.desc = "Flag stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.registers",
	.event = "umask=0x2,period=2000000,event=0xd2",
	.desc = "Partial register stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.rob_read_port",
	.event = "umask=0x4,period=2000000,event=0xd2",
	.desc = "ROB read port stalls cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.scoreboard",
	.event = "umask=0x8,period=2000000,event=0xd2",
	.desc = "Scoreboard stall cycles",
	.topic = "other",
},
{
	.name = "sb_drain.any",
	.event = "umask=0x7,period=200000,event=0x4",
	.desc = "All Store buffer stall cycles",
	.topic = "other",
},
{
	.name = "seg_rename_stalls",
	.event = "umask=0x1,period=2000000,event=0xd4",
	.desc = "Segment rename stall cycles",
	.topic = "other",
},
{
	.name = "snoop_response.hit",
	.event = "umask=0x1,period=100000,event=0xb8",
	.desc = "Thread responded HIT to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hite",
	.event = "umask=0x2,period=100000,event=0xb8",
	.desc = "Thread responded HITE to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hitm",
	.event = "umask=0x4,period=100000,event=0xb8",
	.desc = "Thread responded HITM to snoop",
	.topic = "other",
},
{
	.name = "sq_full_stall_cycles",
	.event = "umask=0x1,period=2000000,event=0xf6",
	.desc = "Super Queue full stall cycles",
	.topic = "other",
},
{
	.name = "arith.cycles_div_busy",
	.event = "umask=0x1,period=2000000,event=0x14",
	.desc = "Cycles the divider is busy",
	.topic = "pipeline",
},
{
	.name = "arith.div",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,edge=1,event=0x14",
	.desc = "Divide Operations executed",
	.topic = "pipeline",
},
{
	.name = "arith.mul",
	.event = "umask=0x2,period=2000000,event=0x14",
	.desc = "Multiply operations executed",
	.topic = "pipeline",
},
{
	.name = "baclear.bad_target",
	.event = "umask=0x2,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted with bad target address",
	.topic = "pipeline",
},
{
	.name = "baclear.clear",
	.event = "umask=0x1,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted, regardless of cause ",
	.topic = "pipeline",
},
{
	.name = "baclear_force_iq",
	.event = "umask=0x1,period=2000000,event=0xa7",
	.desc = "Instruction queue forced BACLEAR",
	.topic = "pipeline",
},
{
	.name = "br_inst_decoded",
	.event = "umask=0x1,period=2000000,event=0xe0",
	.desc = "Branch instructions decoded",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.any",
	.event = "umask=0x7f,period=200000,event=0x88",
	.desc = "Branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.cond",
	.event = "umask=0x1,period=200000,event=0x88",
	.desc = "Conditional branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct",
	.event = "umask=0x2,period=200000,event=0x88",
	.desc = "Unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct_near_call",
	.event = "umask=0x10,period=20000,event=0x88",
	.desc = "Unconditional call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_near_call",
	.event = "umask=0x20,period=20000,event=0x88",
	.desc = "Indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_non_call",
	.event = "umask=0x4,period=20000,event=0x88",
	.desc = "Indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.near_calls",
	.event = "umask=0x30,period=20000,event=0x88",
	.desc = "Call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.non_calls",
	.event = "umask=0x7,period=200000,event=0x88",
	.desc = "All non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.return_near",
	.event = "umask=0x8,period=20000,event=0x88",
	.desc = "Indirect return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken",
	.event = "umask=0x40,period=200000,event=0x88",
	.desc = "Taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x4,period=200000,event=0xc4",
	.desc = "Retired branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=200000,event=0xc4",
	.desc = "Retired conditional branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=20000,event=0xc4",
	.desc = "Retired near call instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.any",
	.event = "umask=0x7f,period=20000,event=0x89",
	.desc = "Mispredicted branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.cond",
	.event = "umask=0x1,period=20000,event=0x89",
	.desc = "Mispredicted conditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct",
	.event = "umask=0x2,period=20000,event=0x89",
	.desc = "Mispredicted unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct_near_call",
	.event = "umask=0x10,period=2000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_near_call",
	.event = "umask=0x20,period=2000,event=0x89",
	.desc = "Mispredicted indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_non_call",
	.event = "umask=0x4,period=2000,event=0x89",
	.desc = "Mispredicted indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.near_calls",
	.event = "umask=0x30,period=2000,event=0x89",
	.desc = "Mispredicted call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.non_calls",
	.event = "umask=0x7,period=20000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.return_near",
	.event = "umask=0x8,period=2000,event=0x89",
	.desc = "Mispredicted return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken",
	.event = "umask=0x40,period=20000,event=0x89",
	.desc = "Mispredicted taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=2000,event=0xc5",
	.desc = "Mispredicted near retired calls (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Reference cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_p",
	.event = "umask=0x1,period=100000,event=0x3c",
	.desc = "Reference base clock (133 Mhz) cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000000,event=0x3c",
	.desc = "Cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.total_cycles",
	.event = "inv=1,umask=0x0,period=2000000,cmask=2,event=0x3c",
	.desc = "Total CPU cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.any",
	.event = "umask=0xf,period=2000000,event=0x87",
	.desc = "Any Instruction Length Decoder stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000000,event=0x87",
	.desc = "Instruction Queue full stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000000,event=0x87",
	.desc = "Length Change Prefix stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.mru",
	.event = "umask=0x2,period=2000000,event=0x87",
	.desc = "Stall cycles due to BPU MRU bypass",
	.topic = "pipeline",
},
{
	.name = "ild_stall.regen",
	.event = "umask=0x8,period=2000000,event=0x87",
	.desc = "Regen stall cycles",
	.topic = "pipeline",
},
{
	.name = "inst_decoded.dec0",
	.event = "umask=0x1,period=2000000,event=0x18",
	.desc = "Instructions that must be decoded by decoder 0",
	.topic = "pipeline",
},
{
	.name = "inst_queue_write_cycles",
	.event = "umask=0x1,period=2000000,event=0x1e",
	.desc = "Cycles instructions are written to the instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_queue_writes",
	.event = "umask=0x1,period=2000000,event=0x17",
	.desc = "Instructions written to instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (Programmable counter and Precise Event) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.mmx",
	.event = "umask=0x4,period=2000000,event=0xc0",
	.desc = "Retired MMX instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000000,event=0xc0",
	.desc = "Retired floating-point operations (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre",
	.event = "umask=0x1,period=200000,event=0x4c",
	.desc = "Load operations conflicting with software prefetches",
	.topic = "pipeline",
},
{
	.name = "lsd.active",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles when uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.inactive",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles no uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd_overflow",
	.event = "umask=0x1,period=2000000,event=0x20",
	.desc = "Loops that can't stream from the instruction queue",
	.topic = "pipeline",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=20000,event=0xc3",
	.desc = "Cycles machine clear asserted",
	.topic = "pipeline",
},
{
	.name = "machine_clears.mem_order",
	.event = "umask=0x2,period=20000,event=0xc3",
	.desc = "Execution pipeline restart due to Memory ordering conflicts",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=20000,event=0xc3",
	.desc = "Self-Modifying Code detected",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000000,event=0xa2",
	.desc = "Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.fpcw",
	.event = "umask=0x20,period=2000000,event=0xa2",
	.desc = "FPU control word write stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.load",
	.event = "umask=0x2,period=2000000,event=0xa2",
	.desc = "Load buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.mxcsr",
	.event = "umask=0x40,period=2000000,event=0xa2",
	.desc = "MXCSR rename stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.other",
	.event = "umask=0x80,period=2000000,event=0xa2",
	.desc = "Other Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rob_full",
	.event = "umask=0x10,period=2000000,event=0xa2",
	.desc = "ROB full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rs_full",
	.event = "umask=0x4,period=2000000,event=0xa2",
	.desc = "Reservation Station full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.store",
	.event = "umask=0x8,period=2000000,event=0xa2",
	.desc = "Store buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_double",
	.event = "umask=0x4,period=200000,event=0xc7",
	.desc = "SIMD Packed-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_single",
	.event = "umask=0x1,period=200000,event=0xc7",
	.desc = "SIMD Packed-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_double",
	.event = "umask=0x8,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_single",
	.event = "umask=0x2,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.vector_integer",
	.event = "umask=0x10,period=200000,event=0xc7",
	.desc = "SIMD Vector Integer Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uop_unfusion",
	.event = "umask=0x1,period=2000000,event=0xdb",
	.desc = "Uop unfusions due to FP exceptions",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_folding",
	.event = "umask=0x4,period=2000000,event=0xd1",
	.desc = "Stack pointer instructions decoded",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_sync",
	.event = "umask=0x8,period=2000000,event=0xd1",
	.desc = "Stack pointer sync operations",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.ms_cycles_active",
	.event = "umask=0x2,period=2000000,cmask=1,event=0xd1",
	.desc = "Uops decoded by Microcode Sequencer",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xd1",
	.desc = "Cycles no Uops are decoded",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles",
	.event = "umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles_no_port5",
	.event = "umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count",
	.event = "inv=1,umask=0x3f,any=1,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count_no_port5",
	.event = "inv=1,umask=0x1f,any=1,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles",
	.event = "inv=1,umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles_no_port5",
	.event = "inv=1,umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port0",
	.event = "umask=0x1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015",
	.event = "umask=0x40,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015_stall_cycles",
	.event = "inv=1,umask=0x40,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port1",
	.event = "umask=0x2,period=2000000,event=0xb1",
	.desc = "Uops executed on port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port2_core",
	.event = "umask=0x4,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 2 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port234_core",
	.event = "umask=0x80,any=1,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 2, 3 or 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port3_core",
	.event = "umask=0x8,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 3 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port4_core",
	.event = "umask=0x10,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port5",
	.event = "umask=0x20,period=2000000,event=0xb1",
	.desc = "Uops executed on port 5",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000000,event=0xe",
	.desc = "Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued on any thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.cycles_all_threads",
	.event = "umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles Uops were issued on either thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.fused",
	.event = "umask=0x2,period=2000000,event=0xe",
	.desc = "Fused Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued",
	.topic = "pipeline",
},
{
	.name = "uops_retired.active_cycles",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are being retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.any",
	.event = "umask=0x1,period=2000000,event=0xc2",
	.desc = "Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.macro_fused",
	.event = "umask=0x4,period=2000000,event=0xc2",
	.desc = "Macro-fused Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000000,event=0xc2",
	.desc = "Retirement slots used (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are not retiring (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc2",
	.desc = "Total cycles using precise uop retired event (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles_ps",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "dtlb_load_misses.any",
	.event = "umask=0x1,period=200000,event=0x8",
	.desc = "DTLB load misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.pde_miss",
	.event = "umask=0x20,period=200000,event=0x8",
	.desc = "DTLB load miss caused by low part of address",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x10,period=2000000,event=0x8",
	.desc = "DTLB second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x8",
	.desc = "DTLB load miss page walks complete",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x49",
	.desc = "DTLB misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.stlb_hit",
	.event = "umask=0x10,period=200000,event=0x49",
	.desc = "DTLB first level misses but second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x49",
	.desc = "DTLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_flush",
	.event = "umask=0x1,period=2000000,event=0xae",
	.desc = "ITLB flushes",
	.topic = "virtual memory",
},
{
	.name = "itlb_miss_retired",
	.event = "umask=0x20,period=200000,event=0xc8",
	.desc = "Retired instructions that missed the ITLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x85",
	.desc = "ITLB miss",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x85",
	.desc = "ITLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "mem_load_retired.dtlb_miss",
	.event = "umask=0x80,period=200000,event=0xcb",
	.desc = "Retired loads that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "mem_store_retired.dtlb_miss",
	.event = "umask=0x1,period=200000,event=0xc",
	.desc = "Retired stores that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_sandybridge[] = {
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0x1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0x3,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x4,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x8,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xc,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0x10,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x20,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0x30,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.pf_hit",
	.event = "umask=0x40,period=200003,event=0x24",
	.desc = "Requests from the L2 hardware prefetchers that hit L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.pf_miss",
	.event = "umask=0x80,period=200003,event=0x24",
	.desc = "Requests from the L2 hardware prefetchers that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xc0,period=200003,event=0x24",
	.desc = "Requests from L2 hardware prefetchers",
	.topic = "cache",
},
{
	.name = "l2_store_lock_rqsts.miss",
	.event = "umask=0x1,period=200003,event=0x27",
	.desc = "RFOs that miss cache lines",
	.topic = "cache",
},
{
	.name = "l2_store_lock_rqsts.hit_e",
	.event = "umask=0x4,period=200003,event=0x27",
	.desc = "RFOs that hit cache lines in E state",
	.topic = "cache",
},
{
	.name = "l2_store_lock_rqsts.hit_m",
	.event = "umask=0x8,period=200003,event=0x27",
	.desc = "RFOs that hit cache lines in M state",
	.topic = "cache",
},
{
	.name = "l2_store_lock_rqsts.all",
	.event = "umask=0xf,period=200003,event=0x27",
	.desc = "RFOs that access cache lines in any state",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.miss",
	.event = "umask=0x1,period=200003,event=0x28",
	.desc = "Count the number of modified Lines evicted from L1 and missed L2. (Non-rejected WBs from the DCU.)",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.hit_s",
	.event = "umask=0x2,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in S state",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.hit_e",
	.event = "umask=0x4,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in E state",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.hit_m",
	.event = "umask=0x8,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in M state",
	.topic = "cache",
},
{
	.name = "l2_l1d_wb_rqsts.all",
	.event = "umask=0xf,period=200003,event=0x28",
	.desc = "Not rejected writebacks from L1D to L2 cache lines in any state",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed LLC",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to LLC",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss oustandings duration in cycles",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
	.topic = "cache",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "This event counts L1D data line replacements.  Replacements occur when a new line is brought into the cache, causing eviction of a line loaded earlier",
},
{
	.name = "l1d.allocated_in_m",
	.event = "umask=0x2,period=2000003,event=0x51",
	.desc = "Allocated L1D data cache lines in M state",
	.topic = "cache",
},
{
	.name = "l1d.eviction",
	.event = "umask=0x4,period=2000003,event=0x51",
	.desc = "L1D data cache lines in M state evicted due to replacement",
	.topic = "cache",
},
{
	.name = "l1d.all_m_replacement",
	.event = "umask=0x8,period=2000003,event=0x51",
	.desc = "Cache lines in M state evicted out of L1D due to Snoop HitM or dirty line replacement",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_c6",
	.event = "umask=0x1,period=2000003,cmask=6,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding RFO store transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
},
{
	.name = "lock_cycles.cache_lock_duration",
	.event = "umask=0x2,period=2000003,event=0x63",
	.desc = "Cycles when L1D is locked",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Cases when offcore requests buffer cannot take more entries for core",
	.topic = "cache",
},
{
	.name = "l1d_blocks.bank_conflict_cycles",
	.event = "umask=0x5,period=100003,cmask=1,event=0xbf",
	.desc = "Cycles when dispatched loads are cancelled due to L1D bank conflicts with other load ports",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load uops that miss the STLB. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store uops that miss the STLB. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load uops with locked access. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load uops that split across a cacheline boundary. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts line-splitted load uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K). (Precise Event - PEBS) (Precise event)",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts line-splitted store uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K). (Precise Event - PEBS) (Precise event)",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load uops. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts the number of load uops retired (Precise Event) (Precise event)",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store uops. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts the number of store uops retired. (Precise Event - PEBS) (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load uops with L1 cache hits as data sources. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load uops with L2 cache hits as data sources. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.llc_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load uops which data sources were data hits in LLC without snoops required. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops that hit in the last-level (L3) cache without snoops required. (Precise Event - PEBS) (Precise event)",
},
{
	.name = "mem_load_uops_retired.hit_lfb",
	.event = "umask=0x40,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources were load uops missed L1 but hit FB due to preceding miss to the same cache line with data not ready. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were LLC hit and cross-core snoop missed in on-pkg core cache. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were LLC and cross-core snoop hits in on-pkg core cache. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops that hit in the last-level cache (L3) and were found in a non-modified state in a neighboring core's private cache (same package).  Since the last level cache is inclusive, hits to the L3 may require snooping the private L2 caches of any cores on the same socket that have the line.  In this case, a snoop was required, and another L2 had the line in a non-modified state. (Precise Event - PEBS) (Precise event)",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were HitM responses from shared LLC. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops that hit in the last-level cache (L3) and were found in a non-modified state in a neighboring core's private cache (same package).  Since the last level cache is inclusive, hits to the L3 may require snooping the private L2 caches of any cores on the same socket that have the line.  In this case, a snoop was required, and another L2 had the line in a modified state, so the line had to be invalidated in that L2 cache and transferred to the requesting L2. (Precise Event - PEBS) (Precise event)",
},
{
	.name = "mem_load_uops_llc_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load uops which data sources were hits in LLC without snoops required. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_misc_retired.llc_miss",
	.event = "umask=0x2,period=100007,event=0xd4",
	.desc = "Retired load uops with unknown information as data source in cache serviced the load. (Precise Event - PEBS) (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts retired demand loads that missed the  last-level (L3) cache. This means that the load is usually satisfied from memory in a client system or possibly from the remote socket in a server. Demand loads are non speculative load uops. (Precise Event - PEBS) (Precise event)",
},
{
	.name = "l2_trans.demand_data_rd",
	.event = "umask=0x1,period=200003,event=0xf0",
	.desc = "Demand Data Read requests that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.rfo",
	.event = "umask=0x2,period=200003,event=0xf0",
	.desc = "RFO requests that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.code_rd",
	.event = "umask=0x4,period=200003,event=0xf0",
	.desc = "L2 cache accesses when fetching instructions",
	.topic = "cache",
},
{
	.name = "l2_trans.all_pf",
	.event = "umask=0x8,period=200003,event=0xf0",
	.desc = "L2 or LLC HW prefetches that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.l1d_wb",
	.event = "umask=0x10,period=200003,event=0xf0",
	.desc = "L1D writebacks that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.l2_fill",
	.event = "umask=0x20,period=200003,event=0xf0",
	.desc = "L2 fill requests that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
},
{
	.name = "l2_trans.all_requests",
	.event = "umask=0x80,period=200003,event=0xf0",
	.desc = "Transactions accessing L2 pipe",
	.topic = "cache",
},
{
	.name = "l2_lines_in.i",
	.event = "umask=0x1,period=100003,event=0xf1",
	.desc = "L2 cache lines in I state filling L2",
	.topic = "cache",
},
{
	.name = "l2_lines_in.s",
	.event = "umask=0x2,period=100003,event=0xf1",
	.desc = "L2 cache lines in S state filling L2",
	.topic = "cache",
},
{
	.name = "l2_lines_in.e",
	.event = "umask=0x4,period=100003,event=0xf1",
	.desc = "L2 cache lines in E state filling L2",
	.topic = "cache",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x7,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines brought into the L2 cache.  Lines are filled into the L2 cache when there was an L2 miss",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x1,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by demand",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x2,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by demand",
	.topic = "cache",
},
{
	.name = "l2_lines_out.pf_clean",
	.event = "umask=0x4,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by L2 prefetch",
	.topic = "cache",
},
{
	.name = "l2_lines_out.pf_dirty",
	.event = "umask=0x8,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines evicted by L2 prefetch",
	.topic = "cache",
},
{
	.name = "l2_lines_out.dirty_all",
	.event = "umask=0xa,period=100003,event=0xf2",
	.desc = "Dirty L2 cache lines filling the L2",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Split locks in SQ",
	.topic = "cache",
},
{
	.name = "offcore_response.all_code_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0244",
	.desc = "Counts demand & prefetch code reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_code_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0244",
	.desc = "Counts demand & prefetch code reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_code_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0244",
	.desc = "Counts demand & prefetch code reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0091",
	.desc = "Counts all demand & prefetch data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0091",
	.desc = "Counts demand & prefetch data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0240",
	.desc = "Counts all prefetch code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_code_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0240",
	.desc = "Counts prefetch code reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_code_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0240",
	.desc = "Counts prefetch code reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_code_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0240",
	.desc = "Counts prefetch code reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_code_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0240",
	.desc = "Counts prefetch code reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0090",
	.desc = "Counts all prefetch data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0090",
	.desc = "Counts prefetch data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0120",
	.desc = "Counts all prefetch RFOs that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_rfo.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0120",
	.desc = "Counts prefetch RFOs that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0120",
	.desc = "Counts prefetch RFOs that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_rfo.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0120",
	.desc = "Counts prefetch RFOs that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_pf_rfo.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0120",
	.desc = "Counts prefetch RFOs that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c03f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c03f7",
	.desc = "Counts data/code/rfo reads (demand & prefetch) that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c03f7",
	.desc = "Counts data/code/rfo reads (demand & prefetch) that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c03f7",
	.desc = "Counts data/code/rfo reads (demand & prefetch) that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c03f7",
	.desc = "Counts data/code/rfo reads (demand & prefetch) that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0122",
	.desc = "Counts all demand & prefetch RFOs that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0122",
	.desc = "Counts demand & prefetch RFOs that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0122",
	.desc = "Counts demand & prefetch RFOs that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0122",
	.desc = "Counts demand & prefetch RFOs that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0122",
	.desc = "Counts demand & prefetch RFOs that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10008",
	.desc = "COREWB & ANY_RESPONSE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0004",
	.desc = "Counts all demand code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0004",
	.desc = "Counts demand code reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0004",
	.desc = "Counts demand code reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0004",
	.desc = "Counts demand code reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0004",
	.desc = "Counts demand code reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0001",
	.desc = "Counts all demand data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0001",
	.desc = "Counts demand data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0001",
	.desc = "Counts demand data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0001",
	.desc = "Counts demand data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0001",
	.desc = "Counts demand data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0002",
	.desc = "Counts all demand data writes (RFOs) that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0002",
	.desc = "Counts demand data writes (RFOs) that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0002",
	.desc = "Counts demand data writes (RFOs) that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0002",
	.desc = "Counts demand data writes (RFOs) that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0002",
	.desc = "Counts demand data writes (RFOs) that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x18000",
	.desc = "Counts miscellaneous accesses that include port i/o, MMIO and uncacheable memory accesses. It also includes L2 hints sent to LLC to keep a line from being evicted out of the core caches",
	.topic = "cache",
},
{
	.name = "offcore_response.other.lru_hints",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x803c8000",
	.desc = "Counts L2 hints sent to LLC to keep a line from being evicted out of the core caches",
	.topic = "cache",
},
{
	.name = "offcore_response.other.portio_mmio_uc",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2380408000",
	.desc = "Counts miscellaneous accesses that include port i/o, MMIO and uncacheable memory accesses",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0040",
	.desc = "Counts all prefetch (that bring data to L2) code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0040",
	.desc = "Counts prefetch (that bring data to L2) code reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0040",
	.desc = "Counts prefetch (that bring data to L2) code reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0040",
	.desc = "Counts prefetch (that bring data to L2) code reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0040",
	.desc = "Counts prefetch (that bring data to L2) code reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0010",
	.desc = "Counts all prefetch (that bring data to L2) data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0010",
	.desc = "Counts prefetch (that bring data to L2) data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0020",
	.desc = "Counts prefetch (that bring data to L2) RFOs that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0020",
	.desc = "Counts prefetch (that bring data to L2) RFOs that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0020",
	.desc = "Counts prefetch (that bring data to L2) RFOs that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0020",
	.desc = "Counts prefetch (that bring data to L2) RFOs that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0200",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0080",
	.desc = "Counts prefetch (that bring data to LLC only) data reads that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_hit.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3f803c0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs that hit in the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x4003c0100",
	.desc = "Counts prefetch (that bring data to LLC only) RFOs that hit in the LLC and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003c0100",
	.desc = "Counts prefetch (that bring data to LLC only) RFOs that hit in the LLC and the snoop to one of the sibling cores hits the line in M state and the line is forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1003c0100",
	.desc = "Counts prefetch (that bring data to LLC only) RFOs that hit in the LLC and sibling core snoops are not needed as either the core-valid bit is not set or the shared line is present in multiple cores",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2003c0100",
	.desc = "Counts prefetch (that bring data to LLC only) RFOs that hit in the LLC and the snoops sent to sibling cores return clean response",
	.topic = "cache",
},
{
	.name = "offcore_response.split_lock_uc_lock.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10400",
	.desc = "Counts requests where the address of an atomic lock instruction spans a cache line boundary or the lock instruction is executed on uncacheable address",
	.topic = "cache",
},
{
	.name = "offcore_response.streaming_stores.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10800",
	.desc = "Counts non-temporal stores",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010001",
	.desc = "Counts all demand data reads ",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010002",
	.desc = "Counts all demand rfo's ",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.all_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000105B3",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.all_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00010122",
	.desc = "Counts all demand & prefetch prefetch RFOs ",
	.topic = "cache",
},
{
	.name = "offcore_response.all_reads.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000107F7",
	.desc = "Counts all data/code/rfo references (demand & prefetch) ",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10433",
	.desc = "REQUEST = DATA_INTO_CORE and RESPONSE = ANY_RESPONSE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_m.hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040002",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = LLC_HIT_M and SNOOP = HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10040",
	.desc = "REQUEST = PF_RFO and RESPONSE = ANY_RESPONSE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10080",
	.desc = "REQUEST = PF_LLC_DATA_RD and RESPONSE = ANY_RESPONSE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l_ifetch.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10200",
	.desc = "REQUEST = PF_LLC_IFETCH and RESPONSE = ANY_RESPONSE",
	.topic = "cache",
},
{
	.name = "fp_comp_ops_exe.x87",
	.event = "umask=0x1,period=2000003,event=0x10",
	.desc = "Number of FP Computational Uops Executed this cycle. The number of FADD, FSUB, FCOM, FMULs, integer MULsand IMULs, FDIVs, FPREMs, FSQRTS, integer DIVs, and IDIVs. This event does not distinguish an FADD used in the middle of a transcendental flow from a s",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_packed_double",
	.event = "umask=0x10,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational packed double-precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_scalar_single",
	.event = "umask=0x20,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational scalar single-precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_packed_single",
	.event = "umask=0x40,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational packed single-precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_scalar_double",
	.event = "umask=0x80,period=2000003,event=0x10",
	.desc = "Number of SSE* or AVX-128 FP Computational scalar double-precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "simd_fp_256.packed_single",
	.event = "umask=0x1,period=2000003,event=0x11",
	.desc = "Number of GSSE-256 Computational FP single precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "simd_fp_256.packed_double",
	.event = "umask=0x2,period=2000003,event=0x11",
	.desc = "Number of AVX-256 Computational FP double precision uops issued this cycle",
	.topic = "floating point",
},
{
	.name = "other_assists.avx_store",
	.event = "umask=0x8,period=100003,event=0xc1",
	.desc = "Number of GSSE memory assist for stores. GSSE microcode assist is being invoked whenever the hardware is unable to properly handle GSSE-256b operations",
	.topic = "floating point",
},
{
	.name = "other_assists.avx_to_sse",
	.event = "umask=0x10,period=100003,event=0xc1",
	.desc = "Number of transitions from AVX-256 to legacy SSE when penalty applicable",
	.topic = "floating point",
},
{
	.name = "other_assists.sse_to_avx",
	.event = "umask=0x20,period=100003,event=0xc1",
	.desc = "Number of transitions from SSE to AVX-256 when penalty applicable",
	.topic = "floating point",
},
{
	.name = "fp_assist.x87_output",
	.event = "umask=0x2,period=100003,event=0xca",
	.desc = "Number of X87 assists due to output value",
	.topic = "floating point",
},
{
	.name = "fp_assist.x87_input",
	.event = "umask=0x4,period=100003,event=0xca",
	.desc = "Number of X87 assists due to input value",
	.topic = "floating point",
},
{
	.name = "fp_assist.simd_output",
	.event = "umask=0x8,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to Output values",
	.topic = "floating point",
},
{
	.name = "fp_assist.simd_input",
	.event = "umask=0x10,period=100003,event=0xca",
	.desc = "Number of SIMD FP assists due to input values",
	.topic = "floating point",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,period=100003,cmask=1,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
},
{
	.name = "idq.empty",
	.event = "umask=0x2,period=2000003,event=0x79",
	.desc = "Instruction Decode Queue (IDQ) empty cycles",
	.topic = "frontend",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_uops",
	.event = "umask=0x10,period=2000003,event=0x79",
	.desc = "Uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.ms_dsb_occur",
	.event = "umask=0x10,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which the microcode sequencer assisted the front-end in delivering uops.  Microcode assists are used for complex instructions or scenarios that can't be handled by the standard decoder.  Using other instructions, if possible, will usually improve performance.  See the Intel\u00ae 64 and IA-32 Architectures Optimization Reference Manual for more information",
},
{
	.name = "idq.ms_switches",
	.event = "umask=0x30,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
},
{
	.name = "idq.mite_all_uops",
	.event = "umask=0x3c,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=2000003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
	.topic = "frontend",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Instruction cache, streaming buffer and victim cache misses",
	.topic = "frontend",
	.long_desc = "This event counts the number of instruction cache, streaming buffer and victim cache misses. Counting includes unchacheable accesses",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled ",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops not delivered to the back-end per cycle, per thread, when the back-end was not stalled.  In the ideal case 4 uops can be delivered each cycle.  The event counts the undelivered uops - so if 3 were delivered in one cycle, the counter would be incremented by 1 for that cycle (4 - 3). If the back-end is stalled, the count for this event is not incremented even when uops were not delivered, because the back-end would not have been able to accept them.  This event is used in determining the front-end bound category of the top-down pipeline slots characterization",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=4,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=3,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=2,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_ge_1_uop_deliv.core",
	.event = "inv=1,umask=0x1,period=2000003,cmask=4,event=0x9c",
	.desc = "Cycles when 1 or more uops were delivered to the by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.count",
	.event = "umask=0x1,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switches",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "This event counts the cycles attributed to a switch from the Decoded Stream Buffer (DSB), which holds decoded instructions, to the legacy decode pipeline.  It excludes cycles when the back-end cannot  accept new micro-ops.  The penalty for these switches is potentially several cycles of instruction starvation, where no micro-ops are delivered to the back-end",
},
{
	.name = "dsb_fill.other_cancel",
	.event = "umask=0x2,period=2000003,event=0xac",
	.desc = "Cases of cancelling valid DSB fill not because of exceeding way limit",
	.topic = "frontend",
},
{
	.name = "dsb_fill.exceed_dsb_lines",
	.event = "umask=0x8,period=2000003,event=0xac",
	.desc = "Cycles when Decode Stream Buffer (DSB) fill encounter more than 3 Decode Stream Buffer (DSB) lines",
	.topic = "frontend",
},
{
	.name = "dsb_fill.all_cancel",
	.event = "umask=0xa,period=2000003,event=0xac",
	.desc = "Cases of cancelling valid Decode Stream Buffer (DSB) fill not because of exceeding way limit",
	.topic = "frontend",
},
{
	.name = "misalign_mem_ref.loads",
	.event = "umask=0x1,period=2000003,event=0x5",
	.desc = "Speculative cache line split load uops dispatched to L1 cache",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.stores",
	.event = "umask=0x2,period=2000003,event=0x5",
	.desc = "Speculative cache line split STA uops dispatched to L1 cache",
	.topic = "memory",
},
{
	.name = "page_walks.llc_miss",
	.event = "umask=0x1,period=100003,event=0xbe",
	.desc = "Number of any page walk that had a miss in LLC. Does not necessary cause a SUSPEND",
	.topic = "memory",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts",
	.topic = "memory",
	.long_desc = "This event counts the number of memory ordering Machine Clears detected. Memory Ordering Machine Clears can result from memory disambiguation, external snoops, or cross SMT-HW-thread snoop (stores) hitting load buffers.  Machine clears can have a significant performance impact if they are happening frequently",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Loads with latency value being above 4  (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Loads with latency value being above 8 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Loads with latency value being above 16 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Loads with latency value being above 32 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Loads with latency value being above 64 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Loads with latency value being above 128 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Loads with latency value being above 256 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Loads with latency value being above 512 (Must be precise)",
	.topic = "memory",
},
{
	.name = "mem_trans_retired.precise_store",
	.event = "umask=0x2,period=2000003,event=0xcd",
	.desc = "Sample stores and collect precise store operation via PEBS record. PMC3 only. (Precise Event - PEBS) (Must be precise)",
	.topic = "memory",
},
{
	.name = "offcore_response.all_code_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400244",
	.desc = "Counts all demand & prefetch code reads that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_data_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400091",
	.desc = "Counts all demand & prefetch data reads that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_pf_code_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400240",
	.desc = "Counts all prefetch code reads that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_pf_data_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400090",
	.desc = "Counts all prefetch data reads that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_pf_rfo.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400120",
	.desc = "Counts all prefetch RFOs that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_reads.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3004003f7",
	.desc = "Counts all data/code/rfo reads (demand & prefetch) that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.all_rfo.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400122",
	.desc = "Counts all demand & prefetch RFOs that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_code_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400004",
	.desc = "Counts demand code reads that miss the LLC and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400001",
	.desc = "Counts demand data reads that miss the LLC and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400002",
	.desc = "Counts demand data writes (RFOs) that miss the LLC and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_code_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400040",
	.desc = "Counts all prefetch (that bring data to L2) code reads that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_data_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400010",
	.desc = "Counts prefetch (that bring data to L2) data reads that miss the LLC and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l2_rfo.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_llc_code_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400200",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_llc_data_rd.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_llc_rfo.llc_miss.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x300400100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs that miss the LLC  and the data returned from dram",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in_socket.llc_miss.local_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x6004001b3",
	.desc = "Counts LLC replacements",
	.topic = "memory",
	.long_desc = "This event counts all data requests (demand/prefetch data reads and demand data writes (RFOs) that miss the LLC  where the data is returned from local DRAM",
},
{
	.name = "offcore_response.any_request.llc_miss_local.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1f80408fff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = LLC_MISS_LOCAL and SNOOP = DRAM",
	.topic = "memory",
	.long_desc = "This event counts any requests that miss the LLC where the data was returned from local DRAM",
},
{
	.name = "offcore_response.data_in_socket.llc_miss_local.any_llc_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x17004001b3",
	.desc = "REQUEST = DATA_IN_SOCKET and RESPONSE = LLC_MISS_LOCAL and SNOOP = ANY_LLC_HIT",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.llc_miss_local.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1f80400004",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = LLC_MISS_LOCAL and SNOOP = DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.llc_miss_local.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1f80400010",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = LLC_MISS_LOCAL and SNOOP = DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.llc_miss_local.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1f80400040",
	.desc = "REQUEST = PF_RFO and RESPONSE = LLC_MISS_LOCAL and SNOOP = DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l_data_rd.llc_miss_local.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1f80400080",
	.desc = "REQUEST = PF_LLC_DATA_RD and RESPONSE = LLC_MISS_LOCAL and SNOOP = DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_l_ifetch.llc_miss_local.dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1f80400200",
	.desc = "REQUEST = PF_LLC_IFETCH and RESPONSE = LLC_MISS_LOCAL and SNOOP = DRAM",
	.topic = "memory",
},
{
	.name = "insts_written_to_iq.insts",
	.event = "umask=0x1,period=2000003,event=0x17",
	.desc = "Valid instructions written to IQ per cycle",
	.topic = "other",
},
{
	.name = "hw_pre_req.dl1_miss",
	.event = "umask=0x2,period=2000003,event=0x4e",
	.desc = "Hardware Prefetch requests that miss the L1D cache. This accounts for both L1 streamer and IP-based (IPP) HW prefetchers. A request is being counted each time it access the cache & miss it, including if a block is applicable or if hit the Fill Buffer for ",
	.topic = "other",
},
{
	.name = "cpl_cycles.ring0",
	.event = "umask=0x1,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when the thread is in ring 0",
	.topic = "other",
},
{
	.name = "cpl_cycles.ring0_trans",
	.event = "umask=0x1,edge=1,period=100007,cmask=1,event=0x5c",
	.desc = "Number of intervals between processor halts while thread is in ring 0",
	.topic = "other",
},
{
	.name = "cpl_cycles.ring123",
	.event = "umask=0x2,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when thread is in rings 1, 2, or 3",
	.topic = "other",
},
{
	.name = "lock_cycles.split_lock_uc_lock_duration",
	.event = "umask=0x1,period=2000003,event=0x63",
	.desc = "Cycles when L1 and L2 are locked due to UC or split lock",
	.topic = "other",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state. This event has a constant ratio with the CPU_CLK_UNHALTED.REF_XCLK event. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, this event counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "ld_blocks.data_unknown",
	.event = "umask=0x1,period=100003,event=0x3",
	.desc = "Loads delayed due to SB blocks, preceding store operations with known addresses but unknown data",
	.topic = "pipeline",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Cases when loads get true Block-on-Store blocking code preventing store forwarding",
	.topic = "pipeline",
	.long_desc = "This event counts loads that followed a store to the same address, where the data could not be forwarded inside the pipeline from the store to the load.  The most common reason why store forwarding would be blocked is when a load's address range overlaps with a preceeding smaller uncompleted store.  See the table of not supported store forwards in the Intel\u00ae 64 and IA-32 Architectures Optimization Reference Manual.  The penalty for blocked store forwarding is that the load must wait for the store to complete before it can be issued",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "This event counts the number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
},
{
	.name = "ld_blocks.all_block",
	.event = "umask=0x10,period=100003,event=0x3",
	.desc = "Number of cases where any load ends up with a valid block-code written to the load buffer (including blocks due to Memory Order Buffer (MOB), Data Cache Unit (DCU), TLB, but load has no DCU miss)",
	.topic = "pipeline",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare",
	.topic = "pipeline",
	.long_desc = "Aliasing occurs when a load is issued after a store and their memory addresses are offset by 4K.  This event counts the number of loads that aliased with a preceding store, resulting in an extended address check in the pipeline.  The enhanced address check typically has a performance penalty of 5 cycles",
},
{
	.name = "ld_blocks_partial.all_sta_block",
	.event = "umask=0x8,period=100003,event=0x7",
	.desc = "This event counts the number of times that load operations are temporarily blocked because of older stores, with addresses that are not yet known. A load operation may incur more than one block of this type",
	.topic = "pipeline",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x3,period=2000003,cmask=1,event=0xd",
	.desc = "Number of cycles waiting for the checkpoints in Resource Allocation Table (RAT) to be recovered after Nuke due to all other cases except JEClear (e.g. whenever a ucode assist is needed like SSE exception, memory disambiguation, etc...)",
	.topic = "pipeline",
},
{
	.name = "int_misc.recovery_stalls_count",
	.event = "umask=0x3,edge=1,period=2000003,cmask=1,event=0xd",
	.desc = "Number of occurences waiting for the checkpoints in Resource Allocation Table (RAT) to be recovered after Nuke due to all other cases except JEClear (e.g. whenever a ucode assist is needed like SSE exception, memory disambiguation, etc...)",
	.topic = "pipeline",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x3,any=1,period=2000003,cmask=1,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "int_misc.rat_stall_cycles",
	.event = "umask=0x40,period=2000003,event=0xd",
	.desc = "Cycles when Resource Allocation Table (RAT) external stall is sent to Instruction Decode Queue (IDQ) for the thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of Uops issued by the front-end of the pipeilne to the back-end",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for all threads",
	.topic = "pipeline",
},
{
	.name = "arith.fpu_div_active",
	.event = "umask=0x1,period=2000003,event=0x14",
	.desc = "Cycles when divider is busy executing divide operations",
	.topic = "pipeline",
},
{
	.name = "arith.fpu_div",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0x14",
	.desc = "Divide operations executed",
	.topic = "pipeline",
	.long_desc = "This event counts the number of the divide operations executed",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for software prefetch",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.hw_pf",
	.event = "umask=0x2,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for hardware prefetch",
	.topic = "pipeline",
},
{
	.name = "partial_rat_stalls.flags_merge_uop",
	.event = "umask=0x20,period=2000003,event=0x59",
	.desc = "Increments the number of flags-merge uops in flight each cycle",
	.topic = "pipeline",
},
{
	.name = "partial_rat_stalls.flags_merge_uop_cycles",
	.event = "umask=0x20,period=2000003,cmask=1,event=0x59",
	.desc = "Performance sensitive flags-merging uops added by Sandy Bridge u-arch",
	.topic = "pipeline",
	.long_desc = "This event counts the number of cycles spent executing performance-sensitive flags-merging uops. For example, shift CL (merge_arith_flags). For more details, See the Intel\u00ae 64 and IA-32 Architectures Optimization Reference Manual",
},
{
	.name = "partial_rat_stalls.slow_lea_window",
	.event = "umask=0x40,period=2000003,event=0x59",
	.desc = "Cycles with at least one slow LEA uop being allocated",
	.topic = "pipeline",
	.long_desc = "This event counts the number of cycles with at least one slow LEA uop being allocated. A uop is generally considered as slow LEA if it has three sources (for example, two sources and immediate) regardless of whether it is a result of LEA instruction or not. Examples of the slow LEA uop are or uops with base, index, and offset source operands using base and index reqisters, where base is EBR/RBP/R13, using RIP relative or 16-bit addressing modes. See the Intel\u00ae 64 and IA-32 Architectures Optimization Reference Manual for more details about slow LEA instructions",
},
{
	.name = "partial_rat_stalls.mul_single_uop",
	.event = "umask=0x80,period=2000003,event=0x59",
	.desc = "Multiply packed/scalar single precision uops allocated",
	.topic = "pipeline",
},
{
	.name = "resource_stalls2.all_fl_empty",
	.event = "umask=0xc,period=2000003,event=0x5b",
	.desc = "Cycles with either free list is empty",
	.topic = "pipeline",
},
{
	.name = "resource_stalls2.all_prf_control",
	.event = "umask=0xf,period=2000003,event=0x5b",
	.desc = "Resource stalls2 control structures full for physical registers",
	.topic = "pipeline",
},
{
	.name = "resource_stalls2.bob_full",
	.event = "umask=0x40,period=2000003,event=0x5b",
	.desc = "Cycles when Allocator is stalled if BOB is full and new branch needs it",
	.topic = "pipeline",
},
{
	.name = "resource_stalls2.ooo_rsrc",
	.event = "umask=0x4f,period=2000003,event=0x5b",
	.desc = "Resource stalls out of order resources full",
	.topic = "pipeline",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
},
{
	.name = "rs_events.empty_end",
	.event = "inv=1,umask=0x1,edge=1,period=2000003,cmask=1,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000003,event=0x87",
	.desc = "Stall cycles because IQ is full",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x88",
	.desc = "Not taken macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_direct_jump",
	.event = "umask=0x82,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_near_return",
	.event = "umask=0x88,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches with return mnemonic",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x88",
	.desc = "Taken speculative and retired direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x88",
	.desc = "Speculative and retired macro-conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_jmp",
	.event = "umask=0xc2,period=200003,event=0x88",
	.desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x88",
	.desc = "Speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_indirect_near_return",
	.event = "umask=0xc8,period=200003,event=0x88",
	.desc = "Speculative and retired indirect return branches",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x88",
	.desc = "Speculative and retired direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x88",
	.desc = "Speculative and retired  branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x89",
	.desc = "Not taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_return_near",
	.event = "umask=0x88,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x89",
	.desc = "Mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted direct near calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 0",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_0_core",
	.event = "umask=0x1,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 0",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 1",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_1_core",
	.event = "umask=0x2,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 1",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0xc,period=2000003,event=0xa1",
	.desc = "Cycles per thread when load or STA uops are dispatched to port 2",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_2_core",
	.event = "umask=0xc,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when load or STA uops are dispatched to port 2",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x30,period=2000003,event=0xa1",
	.desc = "Cycles per thread when load or STA uops are dispatched to port 3",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_3_core",
	.event = "umask=0x30,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when load or STA uops are dispatched to port 3",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 4",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_4_core",
	.event = "umask=0x40,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 4",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are dispatched to port 5",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched_port.port_5_core",
	.event = "umask=0x80,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 5",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.lb",
	.event = "umask=0x2,period=2000003,event=0xa2",
	.desc = "Counts the cycles of stall due to lack of load buffers",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rs",
	.event = "umask=0x4,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no eligible RS entry available",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.lb_sb",
	.event = "umask=0xa,period=2000003,event=0xa2",
	.desc = "Resource stalls due to load or store buffers all being in use",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.mem_rs",
	.event = "umask=0xe,period=2000003,event=0xa2",
	.desc = "Resource stalls due to memory buffers or Reservation Station (RS) being fully utilized",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rob",
	.event = "umask=0x10,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to re-order buffer full",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.ooo_rsrc",
	.event = "umask=0xf0,period=2000003,event=0xa2",
	.desc = "Resource stalls due to Rob being full, FCSW, MXCSR and OTHER",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l2_pending",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Each cycle there was a MLC-miss pending demand load this thread (i.e. Non-completed valid SQ entry allocated for demand load and waiting for Uncore), increment by 1. Note this is in MLC and connected to Umask 0",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_pending",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Each cycle there was a miss-pending demand load this thread, increment by 1. Note this is in DCU and connected to Umask 1. Miss Pending demand load should be deduced by OR-ing increment bits of DCACHE_MISS_PEND.PENDING",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_no_dispatch",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "Each cycle there was no dispatch for this thread, increment by 1. Note this is connect to Umask 2. No dispatch can be deduced from the UOPS_EXECUTED event",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_pending",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Each cycle there was a MLC-miss pending demand load and no uops dispatched on this thread (i.e. Non-completed valid SQ entry allocated for demand load and waiting for Uncore), increment by 1. Note this is in MLC and connected to Umask 0 and 2",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_pending",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Each cycle there was a miss-pending demand load this thread and no uops dispatched, increment by 1. Note this is in DCU and connected to Umask 1 and 2. Miss Pending demand load should be deduced by OR-ing increment bits of DCACHE_MISS_PEND.PENDING",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Uops dispatched per thread",
	.topic = "pipeline",
},
{
	.name = "uops_dispatched.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Uops dispatched from any thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "agu_bypass_cancel.count",
	.event = "umask=0x1,period=100003,event=0xb6",
	.desc = "This event counts executed load operations with all the following traits: 1. addressing of the format [base + offset], 2. the offset is between 1 and 2047, 3. the address specified in the base register is in one page and the address [base+offset] is in an",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter   - architectural event",
	.topic = "pipeline",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Instructions retired. (Precise Event - PEBS) (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "other_assists.itlb_miss_retired",
	.event = "umask=0x2,period=100003,event=0xc1",
	.desc = "Retired instructions experiencing ITLB misses",
	.topic = "pipeline",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Actually retired uops. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of micro-ops retired. (Precise Event) (Precise event)",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=10,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.core_stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of retirement slots used each cycle.  There are potentially 4 slots that can be used each cycle - meaning, 4 micro-ops or 4 instructions could retire each cycle.  This event is used in determining the 'Retiring' category of the Top-Down pipeline slots characterization. (Precise Event - PEBS) (Precise event)",
},
{
	.name = "machine_clears.count",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "This event is incremented when self-modifying code (SMC) is detected, which causes a machine clear.  Machine clears can have a significant performance impact if they are happening frequently",
},
{
	.name = "machine_clears.maskmov",
	.event = "umask=0x20,period=100003,event=0xc3",
	.desc = "This event counts the number of executed Intel AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
	.topic = "pipeline",
	.long_desc = "Maskmov false fault - counts number of time ucode passes through Maskmov flow due to instruction's mask being 0 while the flow was completed without raising a fault",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call_r3",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect macro near call instructions retired (captured in ring 3). (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired. (Precise Event - PEBS) (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Not taken branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Far branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc5",
	.desc = "Direct and indirect mispredicted near call instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired. (Precise Event - PEBS) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Mispredicted macro branch instructions retired. (Precise Event - PEBS) (Must be precise)",
},
{
	.name = "br_misp_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc5",
	.desc = "Mispredicted not taken branch instructions retired.(Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "Mispredicted taken branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Count cases of saving new LBR",
	.topic = "pipeline",
},
{
	.name = "baclears.any",
	.event = "umask=0x1f,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "snb metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "snb metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "snb metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "snb metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "snb metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "snb metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "snb metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "snb metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "snb metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "snb metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "snb metrics",
	.metric_expr = "min( 1 , uops_issued.any / ( (uops_retired.retire_slots / inst_retired.any) * 32 * ( icache.hit + icache.misses ) / 4 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "snb metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops ) )",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "snb metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "snb metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "snb metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "snb metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "snb metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "snb metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "snb metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "snb metrics",
	.metric_expr = "(( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / cycles",
	.metric_name = "FLOPc",
	.metric_group = "FLOPS",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "snb metrics",
	.metric_expr = "(( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "FLOPc_SMT",
	.metric_group = "FLOPS_SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "snb metrics",
	.metric_expr = "uops_dispatched.thread / (( cpu@uops_dispatched.core\\,cmask\\=1@ / 2 ) if #smt_on else cpu@uops_dispatched.core\\,cmask\\=1@)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "snb metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Average CPU Utilization",
	.topic = "snb metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "snb metrics",
	.metric_expr = "( (( 1 * ( fp_comp_ops_exe.sse_scalar_single + fp_comp_ops_exe.sse_scalar_double ) + 2 * fp_comp_ops_exe.sse_packed_double + 4 * ( fp_comp_ops_exe.sse_packed_single + simd_fp_256.packed_double ) + 8 * simd_fp_256.packed_single )) / 1000000000 ) / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "snb metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "snb metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "snb metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "snb metrics",
	.metric_expr = "64 * ( arb@event\\=0x81\\,umask\\=0x1@ + arb@event\\=0x84\\,umask\\=0x1@ ) / 1000000 / duration_time / 1000",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "C3 residency percent per core",
	.topic = "snb metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "snb metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "snb metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "snb metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "snb metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "snb metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "snb metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "unc_cbo_xsnp_response.miss",
	.event = "umask=0x01,event=0x22",
	.desc = "Unit: uncore_cbox A snoop misses in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop misses in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.inval",
	.event = "umask=0x02,event=0x22",
	.desc = "Unit: uncore_cbox A snoop invalidates a non-modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop invalidates a non-modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hit",
	.event = "umask=0x04,event=0x22",
	.desc = "Unit: uncore_cbox A snoop hits a non-modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop hits a non-modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hitm",
	.event = "umask=0x08,event=0x22",
	.desc = "Unit: uncore_cbox A snoop hits a modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop hits a modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.inval_m",
	.event = "umask=0x10,event=0x22",
	.desc = "Unit: uncore_cbox A snoop invalidates a modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A snoop invalidates a modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.external_filter",
	.event = "umask=0x20,event=0x22",
	.desc = "Unit: uncore_cbox Filter on cross-core snoops initiated by this Cbox due to external snoop request",
	.topic = "uncore",
	.long_desc = "Filter on cross-core snoops initiated by this Cbox due to external snoop request",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.xcore_filter",
	.event = "umask=0x40,event=0x22",
	.desc = "Unit: uncore_cbox Filter on cross-core snoops initiated by this Cbox due to processor core memory request",
	.topic = "uncore",
	.long_desc = "Filter on cross-core snoops initiated by this Cbox due to processor core memory request",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.eviction_filter",
	.event = "umask=0x80,event=0x22",
	.desc = "Unit: uncore_cbox Filter on cross-core snoops initiated by this Cbox due to LLC eviction",
	.topic = "uncore",
	.long_desc = "Filter on cross-core snoops initiated by this Cbox due to LLC eviction",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.m",
	.event = "umask=0x01,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.e",
	.event = "umask=0x02,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in E-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in E-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.s",
	.event = "umask=0x04,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in S-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.i",
	.event = "umask=0x08,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_filter",
	.event = "umask=0x10,event=0x34",
	.desc = "Unit: uncore_cbox Filter on processor core initiated cacheable read requests",
	.topic = "uncore",
	.long_desc = "Filter on processor core initiated cacheable read requests",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_filter",
	.event = "umask=0x20,event=0x34",
	.desc = "Unit: uncore_cbox Filter on processor core initiated cacheable write requests",
	.topic = "uncore",
	.long_desc = "Filter on processor core initiated cacheable write requests",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.extsnp_filter",
	.event = "umask=0x40,event=0x34",
	.desc = "Unit: uncore_cbox Filter on external snoop requests",
	.topic = "uncore",
	.long_desc = "Filter on external snoop requests",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_request_filter",
	.event = "umask=0x80,event=0x34",
	.desc = "Unit: uncore_cbox Filter on any IRQ or IPQ initiated requests including uncacheable, non-coherent requests",
	.topic = "uncore",
	.long_desc = "Filter on any IRQ or IPQ initiated requests including uncacheable, non-coherent requests",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_arb_trk_occupancy.all",
	.event = "umask=0x01,event=0x80",
	.desc = "Unit: uncore_arb Counts cycles weighted by the number of requests waiting for data returning from the memory controller. Accounts for coherent and non-coherent requests initiated by IA cores, processor graphic units, or LLC",
	.topic = "uncore",
	.long_desc = "Counts cycles weighted by the number of requests waiting for data returning from the memory controller. Accounts for coherent and non-coherent requests initiated by IA cores, processor graphic units, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.all",
	.event = "umask=0x01,event=0x81",
	.desc = "Unit: uncore_arb Counts the number of coherent and in-coherent requests initiated by IA cores, processor graphic units, or LLC",
	.topic = "uncore",
	.long_desc = "Counts the number of coherent and in-coherent requests initiated by IA cores, processor graphic units, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.writes",
	.event = "umask=0x20,event=0x81",
	.desc = "Unit: uncore_arb Counts the number of allocated write entries, include full, partial, and LLC evictions",
	.topic = "uncore",
	.long_desc = "Counts the number of allocated write entries, include full, partial, and LLC evictions",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.evictions",
	.event = "umask=0x80,event=0x81",
	.desc = "Unit: uncore_arb Counts the number of LLC evictions allocated",
	.topic = "uncore",
	.long_desc = "Counts the number of LLC evictions allocated",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_coh_trk_occupancy.all",
	.event = "umask=0x01,event=0x83",
	.desc = "Unit: uncore_arb Cycles weighted by number of requests pending in Coherency Tracker",
	.topic = "uncore",
	.long_desc = "Cycles weighted by number of requests pending in Coherency Tracker",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_coh_trk_requests.all",
	.event = "umask=0x01,event=0x84",
	.desc = "Unit: uncore_arb Number of requests allocated in Coherency Tracker",
	.topic = "uncore",
	.long_desc = "Number of requests allocated in Coherency Tracker",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_occupancy.cycles_with_any_request",
	.event = "umask=0x01,cmask=1,event=0x80",
	.desc = "Unit: uncore_arb Cycles with at least one request outstanding is waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.topic = "uncore",
	.long_desc = "Cycles with at least one request outstanding is waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_occupancy.cycles_over_half_full",
	.event = "umask=0x01,cmask=10,event=0x80",
	.desc = "Unit: uncore_arb Cycles with at least half of the requests outstanding are waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.topic = "uncore",
	.long_desc = "Cycles with at least half of the requests outstanding are waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_clock.socket",
	.event = "umask=0x01,event=0",
	.desc = "Unit: uncore_arb This 48-bit fixed counter counts the UCLK cycles",
	.topic = "uncore",
	.long_desc = "This 48-bit fixed counter counts the UCLK cycles",
	.pmu = "uncore_arb",
},
{
	.name = "unc_cbo_cache_lookup.es",
	.event = "umask=0x06,event=0x34",
	.desc = "Unit: uncore_cbox LLC lookup request that access cache and found line in E-state or S-state",
	.topic = "uncore",
	.long_desc = "LLC lookup request that access cache and found line in E-state or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x8",
	.desc = "Load misses at all DTLB levels that cause completed page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles when the  page miss handler (PMH) is servicing page walks caused by DTLB load misses",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x8",
	.desc = "Load operations that miss the first DTLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
	.long_desc = "This event counts load operations that miss the first DTLB level but hit the second and do not cause any page walks. The penalty in this case is approximately 7 cycles",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x49",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x49",
	.desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Cycle count for an Extended Page table walk.  The Extended Page Directory cache is used by Virtual Machine operating systems while the guest operating systems use the standard TLB caches",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Misses in all ITLB levels that cause completed page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_duration",
	.event = "umask=0x4,period=2000003,event=0x85",
	.desc = "Cycles when PMH is busy with page walks",
	.topic = "virtual memory",
	.long_desc = "This event count cycles when Page Miss Handler (PMH) is servicing page walks caused by ITLB misses",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Operations that miss the first ITLB level but hit the second and do not cause any page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_silvermont[] = {
{
	.name = "l2_reject_xq.all",
	.event = "umask=0x0,period=200003,event=0x30",
	.desc = "Counts the number of request from the L2 that were not accepted into the XQ",
	.topic = "cache",
	.long_desc = "This event counts the number of demand and prefetch transactions that the L2 XQ rejects due to a full or near full condition which likely indicates back pressure from the IDI link. The XQ may reject transactions from the L2Q (non-cacheable requests), BBS (L2 misses) and WOB (L2 write-back victims)",
},
{
	.name = "core_reject_l2q.all",
	.event = "umask=0x0,period=200003,event=0x31",
	.desc = "Counts the number of request that were not accepted into the L2Q because the L2Q is FULL",
	.topic = "cache",
	.long_desc = "Counts the number of (demand and L1 prefetchers) core requests rejected by the L2Q due to a full or nearly full w condition which likely indicates back pressure from L2Q.  It also counts requests that would have gone directly to the XQ, but are rejected due to a full or nearly full condition, indicating back pressure from the IDI link.  The L2Q may also reject transactions  from a core to insure fairness between cores, or to delay a core?s dirty eviction when the address conflicts incoming external snoops.  (Note that L2 prefetcher requests that are dropped are not counted by this event.)",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=200003,event=0x2e",
	.desc = "L2 cache requests from this core",
	.topic = "cache",
	.long_desc = "This event counts requests originating from the core that references a cache line in the L2 cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=200003,event=0x2e",
	.desc = "L2 cache request misses",
	.topic = "cache",
	.long_desc = "This event counts the total number of L2 cache references and the number of L2 cache misses respectively",
},
{
	.name = "fetch_stall.icache_fill_pending_cycles",
	.event = "umask=0x4,period=200003,event=0x86",
	.desc = "Cycles code-fetch stalled due to an outstanding ICache miss",
	.topic = "cache",
	.long_desc = "Counts cycles that fetch is stalled due to an outstanding ICache miss. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes due to an ICache miss.  Note: this event is not the same as the total number of cycles spent retrieving instruction cache lines from the memory hierarchy.\r\nCounts cycles that fetch is stalled due to any reason. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes.  This will include cycles due to an ITLB miss, ICache miss and other events",
},
{
	.name = "rehabq.ld_block_st_forward",
	.event = "umask=0x1,period=200003,event=0x3",
	.desc = "Loads blocked due to store forward restriction (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts the number of retired loads that were prohibited from receiving forwarded data from the store because of address mismatch (Precise event)",
},
{
	.name = "rehabq.ld_block_std_notready",
	.event = "umask=0x2,period=200003,event=0x3",
	.desc = "Loads blocked due to store data not ready",
	.topic = "cache",
	.long_desc = "This event counts the cases where a forward was technically possible, but did not occur because the store data was not available at the right time",
},
{
	.name = "rehabq.st_splits",
	.event = "umask=0x4,period=200003,event=0x3",
	.desc = "Store uops that split cache line boundary",
	.topic = "cache",
	.long_desc = "This event counts the number of retire stores that experienced cache line boundary splits",
},
{
	.name = "rehabq.ld_splits",
	.event = "umask=0x8,period=200003,event=0x3",
	.desc = "Load uops that split cache line boundary (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts the number of retire loads that experienced cache line boundary splits (Precise event)",
},
{
	.name = "rehabq.lock",
	.event = "umask=0x10,period=200003,event=0x3",
	.desc = "Uops with lock semantics",
	.topic = "cache",
	.long_desc = "This event counts the number of retired memory operations with lock semantics. These are either implicit locked instructions such as the XCHG instruction or instructions with an explicit LOCK prefix (0xF0)",
},
{
	.name = "rehabq.sta_full",
	.event = "umask=0x20,period=200003,event=0x3",
	.desc = "Store address buffer full",
	.topic = "cache",
	.long_desc = "This event counts the number of retired stores that are delayed because there is not a store address buffer available",
},
{
	.name = "rehabq.any_ld",
	.event = "umask=0x40,period=200003,event=0x3",
	.desc = "Any reissued load uops",
	.topic = "cache",
	.long_desc = "This event counts the number of load uops reissued from Rehabq",
},
{
	.name = "rehabq.any_st",
	.event = "umask=0x80,period=200003,event=0x3",
	.desc = "Any reissued store uops",
	.topic = "cache",
	.long_desc = "This event counts the number of store uops reissued from Rehabq",
},
{
	.name = "mem_uops_retired.l1_miss_loads",
	.event = "umask=0x1,period=200003,event=0x4",
	.desc = "Loads missed L1",
	.topic = "cache",
	.long_desc = "This event counts the number of load ops retired that miss in L1 Data cache. Note that prefetch misses will not be counted",
},
{
	.name = "mem_uops_retired.l2_hit_loads",
	.event = "umask=0x2,period=200003,event=0x4",
	.desc = "Loads hit L2 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts the number of load ops retired that hit in the L2 (Precise event)",
},
{
	.name = "mem_uops_retired.l2_miss_loads",
	.event = "umask=0x4,period=100007,event=0x4",
	.desc = "Loads missed L2 (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts the number of load ops retired that miss in the L2 (Precise event)",
},
{
	.name = "mem_uops_retired.utlb_miss",
	.event = "umask=0x10,period=200003,event=0x4",
	.desc = "Loads missed UTLB",
	.topic = "cache",
	.long_desc = "This event counts the number of load ops retired that had UTLB miss",
},
{
	.name = "mem_uops_retired.hitm",
	.event = "umask=0x20,period=200003,event=0x4",
	.desc = "Cross core or cross module hitm (Precise event)",
	.topic = "cache",
	.long_desc = "This event counts the number of load ops retired that got data from the other core or from the other module (Precise event)",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x40,period=200003,event=0x4",
	.desc = "All Loads",
	.topic = "cache",
	.long_desc = "This event counts the number of load ops retired",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x80,period=200003,event=0x4",
	.desc = "All Stores",
	.topic = "cache",
	.long_desc = "This event counts the number of store ops retired",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100007,event=0xb7",
	.desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "offcore_response.any_code_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000044",
	.desc = "Counts any code reads (demand & prefetch) that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000044",
	.desc = "Counts any code reads (demand & prefetch) that hit in the other module where modified copies were found in other core's L1 cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000044",
	.desc = "Counts any code reads (demand & prefetch) that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000044",
	.desc = "Counts any code reads (demand & prefetch) that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.any_code_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010044",
	.desc = "Counts any code reads (demand & prefetch) that have any response type",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000022",
	.desc = "Counts any rfo reads (demand & prefetch) that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000022",
	.desc = "Counts any rfo reads (demand & prefetch) that hit in the other module where modified copies were found in other core's L1 cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000022",
	.desc = "Counts any rfo reads (demand & prefetch) that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000022",
	.desc = "Counts any rfo reads (demand & prefetch) that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010022",
	.desc = "Counts any rfo reads (demand & prefetch) that have any response type",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680003091",
	.desc = "Counts any data read (demand & prefetch) that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000003091",
	.desc = "Counts any data read (demand & prefetch) that hit in the other module where modified copies were found in other core's L1 cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400003091",
	.desc = "Counts any data read (demand & prefetch) that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200003091",
	.desc = "Counts any data read (demand & prefetch) that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000013091",
	.desc = "Counts any data read (demand & prefetch) that have any response type",
	.topic = "cache",
},
{
	.name = "offcore_response.streaming_stores.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680004800",
	.desc = "Counts streaming store that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000008008",
	.desc = "Counts any request that hit in the other module where modified copies were found in other core's L1 cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400008008",
	.desc = "Counts any request that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200008008",
	.desc = "Counts any request that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000018008",
	.desc = "Counts any request that have any response type",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680002000",
	.desc = "Counts DCU hardware prefetcher data read that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000002000",
	.desc = "Counts DCU hardware prefetcher data read that hit in the other module where modified copies were found in other core's L1 cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400002000",
	.desc = "Counts DCU hardware prefetcher data read that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200002000",
	.desc = "Counts DCU hardware prefetcher data read that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l1_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000012000",
	.desc = "Counts DCU hardware prefetcher data read that have any response type",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_writes.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000100",
	.desc = "Countsof demand RFO requests to write to partial cache lines that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.partial_reads.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000080",
	.desc = "Counts demand reads of partial cache lines (including UC and WC) that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000040",
	.desc = "Counts code reads generated by L2 prefetchers that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000040",
	.desc = "Counts code reads generated by L2 prefetchers that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_code_rd.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000040",
	.desc = "Counts code reads generated by L2 prefetchers that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000020",
	.desc = "Counts RFO requests generated by L2 prefetchers that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000020",
	.desc = "Counts RFO requests generated by L2 prefetchers that hit in the other module where modified copies were found in other core's L1 cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000020",
	.desc = "Counts RFO requests generated by L2 prefetchers that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_rfo.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000020",
	.desc = "Counts RFO requests generated by L2 prefetchers that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000010",
	.desc = "Counts data cacheline reads generated by L2 prefetchers that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000010",
	.desc = "Counts data cacheline reads generated by L2 prefetchers that hit in the other module where modified copies were found in other core's L1 cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000010",
	.desc = "Counts data cacheline reads generated by L2 prefetchers that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_l2_data_rd.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000010",
	.desc = "Counts data cacheline reads generated by L2 prefetchers that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000008",
	.desc = "Counts writeback (modified to exclusive) that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.l2_miss.no_snoop_needed",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0080000008",
	.desc = "Counts writeback (modified to exclusive) that miss L2 with no details on snoop-related information",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000004",
	.desc = "Counts demand and DCU prefetch instruction cacheline that are are outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000004",
	.desc = "Counts demand and DCU prefetch instruction cacheline that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000004",
	.desc = "Counts demand and DCU prefetch instruction cacheline that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000004",
	.desc = "Counts demand and DCU prefetch instruction cacheline that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_code_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010004",
	.desc = "Counts demand and DCU prefetch instruction cacheline that have any response type",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000002",
	.desc = "Counts demand and DCU prefetch RFOs that are are outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000002",
	.desc = "Counts demand and DCU prefetch RFOs that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000002",
	.desc = "Counts demand and DCU prefetch RFOs that hit in the other module where modified copies were found in other core's L1 cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000002",
	.desc = "Counts demand and DCU prefetch RFOs that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000002",
	.desc = "Counts demand and DCU prefetch RFOs that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.outstanding",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x4000000001",
	.desc = "Counts demand and DCU prefetch data read that are are outstanding, per cycle, from the time of the L2 miss to when any response is received",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.any",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1680000001",
	.desc = "Counts demand and DCU prefetch data read that miss L2",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.hitm_other_core",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x1000000001",
	.desc = "Counts demand and DCU prefetch data read that hit in the other module where modified copies were found in other core's L1 cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0400000001",
	.desc = "Counts demand and DCU prefetch data read that miss L2 and the snoops to sibling cores hit in either E/S state and the line is not forwarded",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.l2_miss.snoop_miss",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0200000001",
	.desc = "Counts demand and DCU prefetch data read that miss L2 with a snoop miss response",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_response",
	.event = "umask=0x1,period=100007,event=0xb7,offcore_rsp=0x0000010001",
	.desc = "Counts demand and DCU prefetch data read that have any response type",
	.topic = "cache",
},
{
	.name = "icache.accesses",
	.event = "umask=0x3,period=200003,event=0x80",
	.desc = "Instruction fetches",
	.topic = "frontend",
	.long_desc = "This event counts all instruction fetches, not including most uncacheable\r\nfetches",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=200003,event=0x80",
	.desc = "Instruction fetches from Icache",
	.topic = "frontend",
	.long_desc = "This event counts all instruction fetches from the instruction cache",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Icache miss",
	.topic = "frontend",
	.long_desc = "This event counts all instruction fetches that miss the Instruction cache or produce memory requests. This includes uncacheable fetches. An instruction fetch miss is counted only once and not once for every cycle it is outstanding",
},
{
	.name = "ms_decoded.ms_entry",
	.event = "umask=0x1,period=200003,event=0xe7",
	.desc = "Counts the number of times entered into a ucode flow in the FEC.  Includes inserted flows due to front-end detected faults or assists.  Speculative count",
	.topic = "frontend",
	.long_desc = "Counts the number of times the MSROM starts a flow of UOPS. It does not count every time a UOP is read from the microcode ROM.  The most common case that this counts is when a micro-coded instruction is encountered by the front end of the machine.  Other cases include when an instruction encounters a fault, trap, or microcode assist of any sort.  The event will count MSROM startups for UOPS that are speculative, and subsequently cleared by branch mispredict or machine clear.  Background: UOPS are produced by two mechanisms.  Either they are generated by hardware that decodes instructions into UOPS, or they are delivered by a ROM (called the MSROM) that holds UOPS associated with a specific instruction.  MSROM UOPS might also be delivered in response to some condition such as a fault or other exceptional condition.  This event is an excellent mechanism for detecting instructions that require the use of MSROM instructions",
},
{
	.name = "decode_restriction.predecode_wrong",
	.event = "umask=0x1,period=200003,event=0xe9",
	.desc = "Counts the number of times a decode restriction reduced the decode throughput due to wrong instruction length prediction",
	.topic = "frontend",
	.long_desc = "Counts the number of times a decode restriction reduced the decode throughput due to wrong instruction length prediction",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=200003,event=0xc3",
	.desc = "Stalls due to Memory ordering",
	.topic = "memory",
	.long_desc = "This event counts the number of times that pipeline was cleared due to memory ordering issues",
},
{
	.name = "fetch_stall.itlb_fill_pending_cycles",
	.event = "umask=0x2,period=200003,event=0x86",
	.desc = "Cycles code-fetch stalled due to an outstanding ITLB miss",
	.topic = "other",
	.long_desc = "Counts cycles that fetch is stalled due to an outstanding ITLB miss. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes due to an ITLB miss.  Note: this event is not the same as page walk cycles to retrieve an instruction translation",
},
{
	.name = "fetch_stall.all",
	.event = "umask=0x3f,period=200003,event=0x86",
	.desc = "Cycles code-fetch stalled due to any reason",
	.topic = "other",
	.long_desc = "Counts cycles that fetch is stalled due to any reason. That is, the decoder queue is able to accept bytes, but the fetch unit is unable to provide bytes.  This will include cycles due to an ITLB miss, ICache miss and other events",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=200003,event=0xc4",
	.desc = "Counts the number of branch instructions retired.. (Precise event)",
	.topic = "pipeline",
	.long_desc = "ALL_BRANCHES counts the number of any branch instructions retired.  Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Precise event)",
},
{
	.name = "br_inst_retired.jcc",
	.event = "umask=0x7e,period=200003,event=0xc4",
	.desc = "Counts the number of JCC branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "JCC counts the number of conditional branch (JCC) instructions retired. Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Precise event)",
},
{
	.name = "br_inst_retired.taken_jcc",
	.event = "umask=0xfe,period=200003,event=0xc4",
	.desc = "Counts the number of taken JCC branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "TAKEN_JCC counts the number of taken conditional branch (JCC) instructions retired. Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Precise event)",
},
{
	.name = "br_inst_retired.call",
	.event = "umask=0xf9,period=200003,event=0xc4",
	.desc = "Counts the number of near CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "CALL counts the number of near CALL branch instructions retired.  Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Precise event)",
},
{
	.name = "br_inst_retired.rel_call",
	.event = "umask=0xfd,period=200003,event=0xc4",
	.desc = "Counts the number of near relative CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "REL_CALL counts the number of near relative CALL branch instructions retired.  Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Precise event)",
},
{
	.name = "br_inst_retired.ind_call",
	.event = "umask=0xfb,period=200003,event=0xc4",
	.desc = "Counts the number of near indirect CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "IND_CALL counts the number of near indirect CALL branch instructions retired.  Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Precise event)",
},
{
	.name = "br_inst_retired.return",
	.event = "umask=0xf7,period=200003,event=0xc4",
	.desc = "Counts the number of near RET branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "RETURN counts the number of near RET branch instructions retired.  Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Precise event)",
},
{
	.name = "br_inst_retired.non_return_ind",
	.event = "umask=0xeb,period=200003,event=0xc4",
	.desc = "Counts the number of near indirect JMP and near indirect CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "NON_RETURN_IND counts the number of near indirect JMP and near indirect CALL branch instructions retired.  Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Precise event)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0xbf,period=200003,event=0xc4",
	.desc = "Counts the number of far branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "FAR counts the number of far branch instructions retired.  Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Precise event)",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "ALL_BRANCHES counts the number of any mispredicted branch instructions retired. This umask is an architecturally defined event. This event counts the number of retired branch instructions that were mispredicted by the processor, categorized by type. A branch misprediction occurs when the processor predicts that the branch would be taken, but it is not, or vice-versa.  When the misprediction is discovered, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path (Precise event)",
},
{
	.name = "br_misp_retired.jcc",
	.event = "umask=0x7e,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted JCC branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "JCC counts the number of mispredicted conditional branches (JCC) instructions retired.  This event counts the number of retired branch instructions that were mispredicted by the processor, categorized by type. A branch misprediction occurs when the processor predicts that the branch would be taken, but it is not, or vice-versa.  When the misprediction is discovered, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path (Precise event)",
},
{
	.name = "br_misp_retired.taken_jcc",
	.event = "umask=0xfe,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted taken JCC branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "TAKEN_JCC counts the number of mispredicted taken conditional branch (JCC) instructions retired.  This event counts the number of retired branch instructions that were mispredicted by the processor, categorized by type. A branch misprediction occurs when the processor predicts that the branch would be taken, but it is not, or vice-versa.  When the misprediction is discovered, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path (Precise event)",
},
{
	.name = "br_misp_retired.ind_call",
	.event = "umask=0xfb,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted near indirect CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "IND_CALL counts the number of mispredicted near indirect CALL branch instructions retired.  This event counts the number of retired branch instructions that were mispredicted by the processor, categorized by type. A branch misprediction occurs when the processor predicts that the branch would be taken, but it is not, or vice-versa.  When the misprediction is discovered, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path (Precise event)",
},
{
	.name = "br_misp_retired.return",
	.event = "umask=0xf7,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted near RET branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "RETURN counts the number of mispredicted near RET branch instructions retired.  This event counts the number of retired branch instructions that were mispredicted by the processor, categorized by type. A branch misprediction occurs when the processor predicts that the branch would be taken, but it is not, or vice-versa.  When the misprediction is discovered, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path (Precise event)",
},
{
	.name = "br_misp_retired.non_return_ind",
	.event = "umask=0xeb,period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted near indirect JMP and near indirect CALL branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "NON_RETURN_IND counts the number of mispredicted near indirect JMP and near indirect CALL branch instructions retired.  This event counts the number of retired branch instructions that were mispredicted by the processor, categorized by type. A branch misprediction occurs when the processor predicts that the branch would be taken, but it is not, or vice-versa.  When the misprediction is discovered, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path (Precise event)",
},
{
	.name = "uops_retired.ms",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "MSROM micro-ops retired",
	.topic = "pipeline",
	.long_desc = "This event counts the number of micro-ops retired that were supplied from MSROM",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x10,period=2000003,event=0xc2",
	.desc = "Micro-ops retired",
	.topic = "pipeline",
	.long_desc = "This event counts the number of micro-ops retired. The processor decodes complex macro instructions into a sequence of simpler micro-ops. Most instructions are composed of one or two micro-ops. Some instructions are decoded into longer sequences such as repeat instructions, floating point transcendental instructions, and assists. In some cases micro-op sequences are fused or whole instructions are fused into one micro-op. See other UOPS_RETIRED events for differentiating retired fused and non-fused micro-ops",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x1,period=200003,event=0xc3",
	.desc = "Self-Modifying Code detected",
	.topic = "pipeline",
	.long_desc = "This event counts the number of times that a program writes to a code section. Self-modifying code causes a severe penalty in all Intel? architecture processors",
},
{
	.name = "machine_clears.fp_assist",
	.event = "umask=0x4,period=200003,event=0xc3",
	.desc = "Stalls due to FP assists",
	.topic = "pipeline",
	.long_desc = "This event counts the number of times that pipeline stalled due to FP operations needing assists",
},
{
	.name = "machine_clears.all",
	.event = "umask=0x8,period=200003,event=0xc3",
	.desc = "Counts all machine clears",
	.topic = "pipeline",
	.long_desc = "Machine clears happen when something happens in the machine that causes the hardware to need to take special care to get the right answer. When such a condition is signaled on an instruction, the front end of the machine is notified that it must restart, so no more instructions will be decoded from the current path.  All instructions \"older\" than this one will be allowed to finish.  This instruction and all \"younger\" instructions must be cleared, since they must not be allowed to complete.  Essentially, the hardware waits until the problematic instruction is the oldest instruction in the machine.  This means all older instructions are retired, and all pending stores (from older instructions) are completed.  Then the new path of instructions from the front end are allowed to start into the machine.  There are many conditions that might cause a machine clear (including the receipt of an interrupt, or a trap or a fault).  All those conditions (including but not limited to MACHINE_CLEARS.MEMORY_ORDERING, MACHINE_CLEARS.SMC, and MACHINE_CLEARS.FP_ASSIST) are captured in the ANY event. In addition, some conditions can be specifically counted (i.e. SMC, MEMORY_ORDERING, FP_ASSIST).  However, the sum of SMC, MEMORY_ORDERING, and FP_ASSIST machine clears will not necessarily equal the number of ANY",
},
{
	.name = "no_alloc_cycles.rob_full",
	.event = "umask=0x1,period=200003,event=0xca",
	.desc = "Counts the number of cycles when no uops are allocated and the ROB is full (less than 2 entries available)",
	.topic = "pipeline",
	.long_desc = "Counts the number of cycles when no uops are allocated and the ROB is full (less than 2 entries available)",
},
{
	.name = "no_alloc_cycles.mispredicts",
	.event = "umask=0x4,period=200003,event=0xca",
	.desc = "Counts the number of cycles when no uops are allocated and the alloc pipe is stalled waiting for a mispredicted jump to retire.  After the misprediction is detected, the front end will start immediately but the allocate pipe stalls until the mispredicted",
	.topic = "pipeline",
	.long_desc = "Counts the number of cycles when no uops are allocated and the alloc pipe is stalled waiting for a mispredicted jump to retire.  After the misprediction is detected, the front end will start immediately but the allocate pipe stalls until the mispredicted",
},
{
	.name = "no_alloc_cycles.rat_stall",
	.event = "umask=0x20,period=200003,event=0xca",
	.desc = "Counts the number of cycles when no uops are allocated and a RATstall is asserted",
	.topic = "pipeline",
},
{
	.name = "no_alloc_cycles.not_delivered",
	.event = "umask=0x50,period=200003,event=0xca",
	.desc = "Counts the number of cycles when no uops are allocated, the IQ is empty, and no other condition is blocking allocation",
	.topic = "pipeline",
	.long_desc = "The NO_ALLOC_CYCLES.NOT_DELIVERED event is used to measure front-end inefficiencies, i.e. when front-end of the machine is not delivering micro-ops to the back-end and the back-end is not stalled. This event can be used to identify if the machine is truly front-end bound.  When this event occurs, it is an indication that the front-end of the machine is operating at less than its theoretical peak performance.  Background: We can think of the processor pipeline as being divided into 2 broader parts: Front-end and Back-end. Front-end is responsible for fetching the instruction, decoding into micro-ops (uops) in machine understandable format and putting them into a micro-op queue to be consumed by back end. The back-end then takes these micro-ops, allocates the required resources.  When all resources are ready, micro-ops are executed. If the back-end is not ready to accept micro-ops from the front-end, then we do not want to count these as front-end bottlenecks.  However, whenever we have bottlenecks in the back-end, we will have allocation unit stalls and eventually forcing the front-end to wait until the back-end is ready to receive more UOPS. This event counts the cycles only when back-end is requesting more uops and front-end is not able to provide them. Some examples of conditions that cause front-end efficiencies are: Icache misses, ITLB misses, and decoder restrictions that limit the the front-end bandwidth",
},
{
	.name = "no_alloc_cycles.all",
	.event = "umask=0x3f,period=200003,event=0xca",
	.desc = "Counts the number of cycles when no uops are allocated for any reason",
	.topic = "pipeline",
	.long_desc = "The NO_ALLOC_CYCLES.ALL event counts the number of cycles when the front-end does not provide any instructions to be allocated for any reason. This event indicates the cycles where an allocation stalls occurs, and no UOPS are allocated in that cycle",
},
{
	.name = "rs_full_stall.mec",
	.event = "umask=0x1,period=200003,event=0xcb",
	.desc = "Counts the number of cycles and allocation pipeline is stalled and is waiting for a free MEC reservation station entry.  The cycles should be appropriately counted in case of the cracked ops e.g. In case of a cracked load-op, the load portion is sent to M",
	.topic = "pipeline",
	.long_desc = "Counts the number of cycles and allocation pipeline is stalled and is waiting for a free MEC reservation station entry.  The cycles should be appropriately counted in case of the cracked ops e.g. In case of a cracked load-op, the load portion is sent to M",
},
{
	.name = "rs_full_stall.all",
	.event = "umask=0x1f,period=200003,event=0xcb",
	.desc = "Counts the number of cycles the Alloc pipeline is stalled when any one of the RSs (IEC, FPC and MEC) is full. This event is a superset of all the individual RS stall event counts",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions that retire execution. For instructions that consist of multiple micro-ops, this event counts the retirement of the last micro-op of the instruction. The counter continues counting during hardware interrupts, traps, and inside interrupt handlers",
},
{
	.name = "cycles_div_busy.all",
	.event = "umask=0x1,period=2000003,event=0xcd",
	.desc = "Cycles the divider is busy.  Does not imply a stall waiting for the divider",
	.topic = "pipeline",
	.long_desc = "Cycles the divider is busy.This event counts the cycles when the divide unit is unable to accept a new divide UOP because it is busy processing a previously dispatched UOP. The cycles will be counted irrespective of whether or not another divide UOP is waiting to enter the divide unit (from the RS). This event might count cycles while a divide is in progress even if the RS is empty.  The divide instruction is one of the longest latency instructions in the machine.  Hence, it has a special event associated with it to help determine if divides are delaying the retirement of instructions",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Fixed Counter: Counts the number of instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions that retire.  For instructions that consist of multiple micro-ops, this event counts exactly once, as the last micro-op of the instruction retires.  The event continues counting while instructions retire, including during interrupt service routines caused by hardware interrupts, faults or traps.  Background: Modern microprocessors employ extensive pipelining and speculative techniques.  Since sometimes an instruction is started but never completed, the notion of \"retirement\" is introduced.  A retired instruction is one that commits its states. Or stated differently, an instruction might be abandoned at some point. No instruction is truly finished until it retires.  This counter measures the number of completed instructions.  The fixed event is INST_RETIRED.ANY and the programmable event is INST_RETIRED.ANY_P",
},
{
	.name = "cpu_clk_unhalted.core",
	.event = "event=0x3c,period=2000003",
	.desc = "Fixed Counter: Counts the number of unhalted core clock cycles",
	.topic = "pipeline",
	.long_desc = "Counts the number of core cycles while the core is not in a halt state. The core enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios.  The core frequency may change from time to time. For this reason this event may have a changing ratio with regards to time. In systems with a constant core frequency, this event can give you a measurement of the elapsed time while the core was not in halt state by dividing the event count by the core frequency. This event is architecturally defined and is a designated fixed counter.  CPU_CLK_UNHALTED.CORE and CPU_CLK_UNHALTED.CORE_P use the core frequency which may change from time to time.  CPU_CLK_UNHALTE.REF_TSC and CPU_CLK_UNHALTED.REF are not affected by core frequency changes but counts as if the core is running at the maximum frequency all the time.  The fixed events are CPU_CLK_UNHALTED.CORE and CPU_CLK_UNHALTED.REF_TSC and the programmable events are CPU_CLK_UNHALTED.CORE_P and CPU_CLK_UNHALTED.REF",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Fixed Counter: Counts the number of unhalted reference clock cycles",
	.topic = "pipeline",
	.long_desc = "Counts the number of reference cycles while the core is not in a halt state. The core enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios.  The core frequency may change from time. This event is not affected by core frequency changes but counts as if the core is running at the maximum frequency all the time.  Divide this event count by core frequency to determine the elapsed time while the core was not in halt state.  Divide this event count by core frequency to determine the elapsed time while the core was not in halt state.  This event is architecturally defined and is a designated fixed counter.  CPU_CLK_UNHALTED.CORE and CPU_CLK_UNHALTED.CORE_P use the core frequency which may change from time to time.  CPU_CLK_UNHALTE.REF_TSC and CPU_CLK_UNHALTED.REF are not affected by core frequency changes but counts as if the core is running at the maximum frequency all the time.  The fixed events are CPU_CLK_UNHALTED.CORE and CPU_CLK_UNHALTED.REF_TSC and the programmable events are CPU_CLK_UNHALTED.CORE_P and CPU_CLK_UNHALTED.REF",
},
{
	.name = "cpu_clk_unhalted.core_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Core cycles when core is not halted",
	.topic = "pipeline",
	.long_desc = "This event counts the number of core cycles while the core is not in a halt state. The core enters the halt state when it is running the HLT instruction. In mobile systems the core frequency may change from time to time. For this reason this event may have a changing ratio with regards to time",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Reference cycles when core is not halted",
	.topic = "pipeline",
	.long_desc = "This event counts the number of reference cycles that the core is not in a halt state. The core enters the halt state when it is running the HLT instruction. In mobile systems the core frequency may change from time. This event is not affected by core frequency changes but counts as if the core is running at the maximum frequency all the time",
},
{
	.name = "baclears.all",
	.event = "umask=0x1,period=200003,event=0xe6",
	.desc = "Counts the number of baclears",
	.topic = "pipeline",
	.long_desc = "The BACLEARS event counts the number of times the front end is resteered, mainly when the Branch Prediction Unit cannot provide a correct prediction and this is corrected by the Branch Address Calculator at the front end.  The BACLEARS.ANY event counts the number of baclears for any type of branch",
},
{
	.name = "baclears.return",
	.event = "umask=0x8,period=200003,event=0xe6",
	.desc = "Counts the number of RETURN baclears",
	.topic = "pipeline",
	.long_desc = "The BACLEARS event counts the number of times the front end is resteered, mainly when the Branch Prediction Unit cannot provide a correct prediction and this is corrected by the Branch Address Calculator at the front end.  The BACLEARS.RETURN event counts the number of RETURN baclears",
},
{
	.name = "baclears.cond",
	.event = "umask=0x10,period=200003,event=0xe6",
	.desc = "Counts the number of JCC baclears",
	.topic = "pipeline",
	.long_desc = "The BACLEARS event counts the number of times the front end is resteered, mainly when the Branch Prediction Unit cannot provide a correct prediction and this is corrected by the Branch Address Calculator at the front end.  The BACLEARS.COND event counts the number of JCC (Jump on Condtional Code) baclears",
},
{
	.name = "br_inst_retired.all_taken_branches",
	.event = "umask=0x80,period=200003,event=0xc4",
	.desc = "Counts the number of taken branch instructions retired (Must be precise)",
	.topic = "pipeline",
	.long_desc = "ALL_TAKEN_BRANCHES counts the number of all taken branch instructions retired.  Branch prediction predicts the branch target and enables the processor to begin executing instructions long before the branch true execution path is known. All branches utilize the branch prediction unit (BPU) for prediction. This unit predicts the target address not only based on the EIP of the branch but also based on the execution path through which execution reached this EIP. The BPU can efficiently predict the following branch types: conditional branches, direct calls and jumps, indirect calls and jumps, returns (Must be precise)",
},
{
	.name = "mem_uops_retired.dtlb_miss_loads",
	.event = "umask=0x8,period=200003,event=0x4",
	.desc = "Loads missed DTLB (Precise event)",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of load ops retired that had DTLB miss (Precise event)",
},
{
	.name = "page_walks.d_side_walks",
	.event = "umask=0x1,period=100003,edge=1,event=0x5",
	.desc = "D-side page-walks",
	.topic = "virtual memory",
	.long_desc = "This event counts when a data (D) page walk is completed or started.  Since a page walk implies a TLB miss, the number of TLB misses can be counted by counting the number of pagewalks",
},
{
	.name = "page_walks.d_side_cycles",
	.event = "umask=0x1,period=200003,event=0x5",
	.desc = "Duration of D-side page-walks in core cycles",
	.topic = "virtual memory",
	.long_desc = "This event counts every cycle when a D-side (walks due to a load) page walk is in progress. Page walk duration divided by number of page walks is the average duration of page-walks",
},
{
	.name = "page_walks.i_side_walks",
	.event = "umask=0x2,period=100003,edge=1,event=0x5",
	.desc = "I-side page-walks",
	.topic = "virtual memory",
	.long_desc = "This event counts when an instruction (I) page walk is completed or started.  Since a page walk implies a TLB miss, the number of TLB misses can be counted by counting the number of pagewalks",
},
{
	.name = "page_walks.i_side_cycles",
	.event = "umask=0x2,period=200003,event=0x5",
	.desc = "Duration of I-side page-walks in core cycles",
	.topic = "virtual memory",
	.long_desc = "This event counts every cycle when a I-side (walks due to an instruction fetch) page walk is in progress. Page walk duration divided by number of page walks is the average duration of page-walks",
},
{
	.name = "page_walks.walks",
	.event = "umask=0x3,period=100003,edge=1,event=0x5",
	.desc = "Total page walks that are completed (I-side and D-side)",
	.topic = "virtual memory",
	.long_desc = "This event counts when a data (D) page walk or an instruction (I) page walk is completed or started.  Since a page walk implies a TLB miss, the number of TLB misses can be counted by counting the number of pagewalks",
},
{
	.name = "page_walks.cycles",
	.event = "umask=0x3,period=200003,event=0x5",
	.desc = "Total cycles for all the page walks. (I-side and D-side)",
	.topic = "virtual memory",
	.long_desc = "This event counts every cycle when a data (D) page walk or instruction (I) page walk is in progress.  Since a pagewalk implies a TLB miss, the approximate cost of a TLB miss can be determined from this event",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_skylake[] = {
{
	.name = "l2_rqsts.demand_data_rd_miss",
	.event = "umask=0x21,period=200003,event=0x24",
	.desc = "Demand Data Read miss L2, no rejects",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests that miss L2 cache. Only not rejected loads are counted",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x22,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts the RFO (Read-for-Ownership) requests that miss L2 cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x24,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
	.long_desc = "Counts L2 cache misses when fetching instructions",
},
{
	.name = "l2_rqsts.all_demand_miss",
	.event = "umask=0x27,period=200003,event=0x24",
	.desc = "Demand requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Demand requests that miss L2 cache",
},
{
	.name = "l2_rqsts.pf_miss",
	.event = "umask=0x38,period=200003,event=0x24",
	.desc = "Requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that miss L2 cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0x3f,period=200003,event=0x24",
	.desc = "All requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "All requests that miss L2 cache",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0xc1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests, initiated by load instructions, that hit L2 cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0xc2,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the RFO (Read-for-Ownership) requests that hit L2 cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0xc4,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
	.long_desc = "Counts L2 cache hits when fetching instructions, code reads",
},
{
	.name = "l2_rqsts.pf_hit",
	.event = "umask=0xd8,period=200003,event=0x24",
	.desc = "Requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that hit L2 cache",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0xe1,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests (including requests from L1D hardware prefetchers). These loads may hit or miss L2 cache. Only non rejected loads are counted",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xe2,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "Counts the total number of RFO (read for ownership) requests to L2 cache. L2 RFO requests include both L1D demand RFO misses as well as L1D RFO prefetches",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0xe4,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "Counts the total number of L2 code requests",
},
{
	.name = "l2_rqsts.all_demand_references",
	.event = "umask=0xe7,period=200003,event=0x24",
	.desc = "Demand requests to L2 cache",
	.topic = "cache",
	.long_desc = "Demand requests to L2 cache",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xf8,period=200003,event=0x24",
	.desc = "Requests from the L1/L2/L3 hardware prefetchers or Load software prefetches",
	.topic = "cache",
	.long_desc = "Counts the total number of requests from the L2 hardware prefetchers",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200003,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
	.long_desc = "All L2 requests",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed L3  Spec update: SKL057",
	.topic = "cache",
	.long_desc = "Counts core-originated cacheable requests that miss the L3 cache (Longest Latency cache). Requests include data and code reads, Reads-for-Ownership (RFOs), speculative accesses and hardware prefetches from L1 and L2. It does not include all misses to the L3  Spec update: SKL057",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to L3  Spec update: SKL057",
	.topic = "cache",
	.long_desc = "Counts core-originated cacheable requests to the  L3 cache (Longest Latency cache). Requests include data and code reads, Reads-for-Ownership (RFOs), speculative accesses and hardware prefetches from L1 and L2.  It does not include all accesses to the L3  Spec update: SKL057",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss outstandings duration in cycles",
	.topic = "cache",
	.long_desc = "Counts duration of L1D miss outstanding, that is each cycle number of Fill Buffers (FB) outstanding required by Demand Reads. FB either is held by demand loads, or it is held by non-demand loads and gets hit at least once by demand. The valid outstanding interval is defined until the FB deallocation by one of the following ways: from FB allocation, if FB is allocated by demand from the demand Hit FB, if it is allocated by hardware or software prefetch.Note: In the L1D, a Demand Read contains cacheable or noncacheable demand loads, including ones causing cache-line splits and reads due to page walks resulted from any request type",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
	.long_desc = "Counts duration of L1D miss outstanding in cycles",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,event=0x48",
	.desc = "Number of times a request needed a FB entry but there was no entry available for it. That is the FB unavailability was dominant reason for blocking the request. A request includes cacheable/uncacheable demands that is load, store or SW prefetch",
	.topic = "cache",
	.long_desc = "Number of times a request needed a FB (Fill Buffer) entry but there was no entry available for it. A request includes cacheable/uncacheable demands that are load, store or SW prefetch instructions",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "Counts L1D data line replacements including opportunistic replacements, and replacements that require stall-for-replace or block-for-replace",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding Demand Data Read transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor. See the corresponding Umask under OFFCORE_REQUESTS.Note: A prefetch promoted to Demand is counted from the promotion point",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts cycles when offcore outstanding Demand Data Read transactions are present in the super queue (SQ). A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation)",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,period=2000003,cmask=6,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding Code Reads transactions in the SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding Code Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_code_rd",
	.event = "umask=0x2,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles with offcore outstanding Code Reads transactions in the SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding Code Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding RFO (store) transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles with offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding demand rfo Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding cacheable Core Data Read transactions in the super queue every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts cycles when offcore outstanding cacheable Core Data Read transactions are present in the super queue. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
	.long_desc = "Counts the Demand Data Read requests sent to uncore. Use it in conjunction with OFFCORE_REQUESTS_OUTSTANDING to determine average latency in the uncore",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "Counts both cacheable and non-cacheable code read requests",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "Counts the demand RFO (read for ownership) requests including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts the demand and prefetch data reads. All Core Data Reads include cacheable 'Demands' and L2 prefetchers (not L3 prefetchers). Counting also covers reads due to page walks resulted from any request type",
},
{
	.name = "offcore_requests.all_requests",
	.event = "umask=0x80,period=100003,event=0xb0",
	.desc = "Any memory transaction that reached the SQ",
	.topic = "cache",
	.long_desc = "Counts memory transactions reached the super queue including requests initiated by the core, all L3 prefetches, page walks, etc.",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Offcore requests buffer cannot take more entries for this thread core",
	.topic = "cache",
	.long_desc = "Counts the number of cases when the offcore requests buffer cannot take more entries for the core. This can happen when the superqueue does not contain eligible entries, or when L1D writeback pending FIFO requests is full.Note: Writeback pending FIFO has six entries",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "mem_inst_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load instructions that miss the STLB. (Precise Event)  Supports address when precise",
	.topic = "cache",
	.long_desc = "Retired load instructions that miss the STLB  Supports address when precise",
},
{
	.name = "mem_inst_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store instructions that miss the STLB. (Precise Event)  Supports address when precise",
	.topic = "cache",
	.long_desc = "Retired store instructions that miss the STLB  Supports address when precise",
},
{
	.name = "mem_inst_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load instructions with locked access. (Precise Event)  Supports address when precise",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load instructions that split across a cacheline boundary. (Precise Event)  Supports address when precise",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store instructions that split across a cacheline boundary. (Precise Event)  Supports address when precise",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load instructions. (Precise Event)  Supports address when precise",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store instructions. (Precise Event)  Supports address when precise",
	.topic = "cache",
	.long_desc = "All retired store instructions  Supports address when precise",
},
{
	.name = "mem_load_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load instructions with L1 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that hit in the L1 data cache. This event includes all SW prefetches and lock instructions regardless of the data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load instructions with L2 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions with L2 cache hits as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l3_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load instructions with L3 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions with L3 cache hits as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load instructions missed L1 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that missed in the L1 cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Retired load instructions missed L2 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions missed L2 cache as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l3_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Retired load instructions missed L3 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions missed L3 cache as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.fb_hit",
	.event = "umask=0x40,period=100007,event=0xd1",
	.desc = "Retired load instructions which data sources were load missed L1 but hit FB due to preceding miss to the same cache line with data not ready  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop was load missed in L1 but hit FB (Fill Buffers) due to preceding miss to the same cache line with data not ready  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load instructions which data sources were L3 hit and cross-core snoop missed in on-pkg core cache  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load instructions which data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load instructions which data sources were HitM responses from shared L3  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources were HitM responses from shared L3  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load instructions which data sources were hits in L3 without snoops required  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources were hits in L3 without snoops required  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_misc_retired.uc",
	.event = "umask=0x4,period=100007,event=0xd4",
	.desc = "Retired instructions with at least 1 uncacheable load or lock  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "Counts L2 writebacks that access L2 cache",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x1f,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "Counts the number of L2 cache lines filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_out.silent",
	.event = "umask=0x1,period=200003,event=0xf2",
	.desc = "Counts the number of lines that are silently dropped by L2 cache when triggered by an L2 cache fill. These lines are typically in Shared or Exclusive state. A non-threaded event",
	.topic = "cache",
},
{
	.name = "l2_lines_out.non_silent",
	.event = "umask=0x2,period=200003,event=0xf2",
	.desc = "Counts the number of lines that are evicted by L2 cache when triggered by an L2 cache fill. Those lines are in Modified state. Modified lines are written back to L3",
	.topic = "cache",
},
{
	.name = "l2_lines_out.useless_pref",
	.event = "umask=0x4,period=200003,event=0xf2",
	.desc = "This event is deprecated. Refer to new event L2_LINES_OUT.USELESS_HWPF",
	.topic = "cache",
	.long_desc = "This event is deprecated. Refer to new event L2_LINES_OUT.USELESS_HWPF",
},
{
	.name = "l2_lines_out.useless_hwpf",
	.event = "umask=0x4,period=200003,event=0xf2",
	.desc = "Counts the number of lines that have been hardware prefetched but not used and now evicted by L2 cache",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Number of cache line split locks sent to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of cache line split locks sent to the uncore",
},
{
	.name = "offcore_response.other.l4_hit_local_l4.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0408000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l4_hit_local_l4.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000408000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l4_hit_local_l4.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400408000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l4_hit_local_l4.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200408000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l4_hit_local_l4.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100408000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l4_hit_local_l4.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080408000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l4_hit_local_l4.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040408000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC01C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10001C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04001C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02001C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01001C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00801C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00401C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0108000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_s.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000108000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_s.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400108000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200108000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_s.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100108000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080108000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_s.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040108000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0088000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_e.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000088000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_e.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400088000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200088000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_e.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100088000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080088000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_e.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040088000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0048000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_m.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000048000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_m.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400048000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200048000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_m.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100048000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080048000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_m.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040048000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000018000",
	.desc = "Counts any other requests have any response type",
	.topic = "cache",
	.long_desc = "Counts any other requests have any response type",
},
{
	.name = "offcore_response.demand_code_rd.l4_hit_local_l4.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l4_hit_local_l4.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l4_hit_local_l4.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l4_hit_local_l4.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l4_hit_local_l4.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l4_hit_local_l4.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l4_hit_local_l4.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC01C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10001C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04001C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02001C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01001C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00801C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00401C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0100004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_s.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_s.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_s.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_s.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040100004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0080004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_e.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_e.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_e.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_e.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040080004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0040004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_m.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_m.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_m.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_m.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040040004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0020004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040020004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that have any response type",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that have any response type",
},
{
	.name = "offcore_response.demand_rfo.l4_hit_local_l4.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l4_hit_local_l4.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l4_hit_local_l4.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l4_hit_local_l4.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l4_hit_local_l4.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l4_hit_local_l4.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l4_hit_local_l4.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC01C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10001C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04001C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02001C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01001C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00801C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00401C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_s.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_s.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_s.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_s.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_e.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_e.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_e.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_e.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_m.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_m.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_m.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_m.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.supplier_none.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010002",
	.desc = "Counts all demand data writes (RFOs) have any response type",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) have any response type",
},
{
	.name = "offcore_response.demand_data_rd.l4_hit_local_l4.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0400001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l4_hit_local_l4.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000400001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l4_hit_local_l4.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400400001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l4_hit_local_l4.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200400001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l4_hit_local_l4.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100400001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l4_hit_local_l4.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080400001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l4_hit_local_l4.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040400001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC01C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10001C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04001C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02001C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01001C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00801C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00401C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_s.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0100001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_s.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000100001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_s.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400100001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_s.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200100001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_s.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100100001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_s.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080100001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_s.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040100001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_e.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0080001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_e.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000080001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_e.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400080001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_e.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200080001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_e.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100080001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_e.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080080001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_e.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040080001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_m.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0040001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_m.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000040001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_m.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400040001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_m.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200040001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_m.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100040001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_m.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080040001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_m.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040040001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC0020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0040020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010001",
	.desc = "Counts demand data reads have any response type",
	.topic = "cache",
	.long_desc = "Counts demand data reads have any response type",
},
{
	.name = "fp_arith_inst_retired.scalar_double",
	.event = "umask=0x1,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar double precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.scalar_single",
	.event = "umask=0x2,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar single precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_double",
	.event = "umask=0x4,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed double precision floating-point instructions retired.  Each count represents 2 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_single",
	.event = "umask=0x8,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed single precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_double",
	.event = "umask=0x10,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed double precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_single",
	.event = "umask=0x20,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed single precision floating-point instructions retired.  Each count represents 8 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,period=100003,cmask=1,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
	.long_desc = "Counts cycles with any input and output SSE or x87 FP assist. If an input and output assist are detected on the same cycle the event increments by 1",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may 'bypass' the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Counts the number of uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles 4 uops were delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Count includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles uops were delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Count includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts the number of uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles 4 uops were delivered to the Instruction Decode Queue (IDQ) from the MITE (legacy decode pipeline) path. Counting includes uops that may 'bypass' the IDQ. During these cycles uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles uops were delivered to the Instruction Decode Queue (IDQ) from the MITE (legacy decode pipeline) path. Counting includes uops that may 'bypass' the IDQ. During these cycles uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may 'bypass' the IDQ. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "idq.ms_switches",
	.event = "umask=0x30,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
	.long_desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts the total number of uops delivered by the Microcode Sequencer (MS). Any instruction over 4 uops will be delivered by the MS. Some instructions such as transcendentals may additionally generate uops from the MS",
},
{
	.name = "icache_16b.ifdata_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction cache miss",
	.topic = "frontend",
	.long_desc = "Cycles where a code line fetch is stalled due to an L1 instruction cache miss. The legacy decode pipeline works at a 16 Byte granularity",
},
{
	.name = "icache_64b.iftag_hit",
	.event = "umask=0x1,period=200003,event=0x83",
	.desc = "Instruction fetch tag lookups that hit in the instruction cache (L1I). Counts at 64-byte cache-line granularity",
	.topic = "frontend",
},
{
	.name = "icache_64b.iftag_miss",
	.event = "umask=0x2,period=200003,event=0x83",
	.desc = "Instruction fetch tag lookups that miss in the instruction cache (L1I). Counts at 64-byte cache-line granularity",
	.topic = "frontend",
},
{
	.name = "icache_64b.iftag_stall",
	.event = "umask=0x4,period=200003,event=0x83",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction cache tag miss",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts the number of uops not delivered to Resource Allocation Table (RAT) per thread adding \u201c4 \u2013 x\u201d when Resource Allocation Table (RAT) is not stalled and Instruction Decode Queue (IDQ) delivers x uops to Resource Allocation Table (RAT) (where x belongs to {0,1,2,3}). Counting does not cover cases when: a. IDQ-Resource Allocation Table (RAT) pipe serves the other thread. b. Resource Allocation Table (RAT) is stalled for the thread (including uop drops and clear BE conditions).  c. Instruction Decode Queue (IDQ) delivers four uops",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=4,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts, on the per-thread basis, cycles when no uops are delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core =4",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=3,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts, on the per-thread basis, cycles when less than 1 uop is delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core >= 3",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=2,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
	.long_desc = "Cycles with less than 2 uops delivered by the front-end",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
	.long_desc = "Cycles with less than 3 uops delivered by the front-end",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "Counts Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles. These cycles do not include uops routed through because of the switch itself, for example, when Instruction Decode Queue (IDQ) pre-allocation is unavailable, or Instruction Decode Queue (IDQ) is full. SBD-to-MITE switch true penalty cycles happen after the merge mux (MM) receives Decode Stream Buffer (DSB) Sync-indication until receiving the first MITE uop. MM is placed before Instruction Decode Queue (IDQ) to merge uops being fed from the MITE and Decode Stream Buffer (DSB) paths. Decode Stream Buffer (DSB) inserts the Sync-indication whenever a Decode Stream Buffer (DSB)-to-MITE switch occurs.Penalty: A Decode Stream Buffer (DSB) hit followed by a Decode Stream Buffer (DSB) miss can cost up to six cycles in which no uops are delivered to the IDQ. Most often, such switches from the Decode Stream Buffer (DSB) to the legacy pipeline cost 0\u20132 cycles",
},
{
	.name = "frontend_retired.dsb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x11",
	.desc = "Retired Instructions who experienced decode stream buffer (DSB - the decoded instruction-cache) miss. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced DSB (Decode stream buffer i.e. the decoded instruction-cache) miss (Precise event)",
},
{
	.name = "frontend_retired.l1i_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x12",
	.desc = "Retired Instructions who experienced Instruction L1 Cache true miss. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.l2_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x13",
	.desc = "Retired Instructions who experienced Instruction L2 Cache true miss. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.itlb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x14",
	.desc = "Retired Instructions who experienced iTLB true miss. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced iTLB (Instruction TLB) true miss (Precise event)",
},
{
	.name = "frontend_retired.stlb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x15",
	.desc = "Retired Instructions who experienced STLB (2nd level TLB) true miss. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced STLB (2nd level TLB) true miss (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_2",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x400206",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 2 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_2",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x200206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 2 bubble-slots for a period of 2 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_4",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x400406",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 4 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_8",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x400806",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 8 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 8 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_16",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x401006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 16 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 16 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_32",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x402006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 32 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end  after a front-end stall of at least 32 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_64",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x404006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 64 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_128",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x408006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 128 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_256",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x410006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 256 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_512",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x420006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 512 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_1",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x100206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 1 bubble-slot for a period of 2 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after the front-end had at least 1 bubble-slot for a period of 2 cycles. A bubble-slot is an empty issue-pipeline slot while there was no RAT stall (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_3",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x300206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 3 bubble-slots for a period of 2 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "tx_mem.abort_conflict",
	.event = "umask=0x1,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data conflict on a transactionally accessed address",
	.topic = "memory",
	.long_desc = "Number of times a TSX line had a cache conflict",
},
{
	.name = "tx_mem.abort_capacity",
	.event = "umask=0x2,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data capacity limitation for transactional reads or writes",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_store_to_elided_lock",
	.event = "umask=0x4,period=2000003,event=0x54",
	.desc = "Number of times a HLE transactional region aborted due to a non XRELEASE prefixed instruction writing to an elided lock in the elision buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to a non-release/commit store to lock",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_not_empty",
	.event = "umask=0x8,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to NoAllocatedElisionBuffer being non-zero",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_mismatch",
	.event = "umask=0x10,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to XRELEASE lock not satisfying the address and value requirements in the elision buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_unsupported_alignment",
	.event = "umask=0x20,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to an unsupported read alignment from the elision buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
},
{
	.name = "tx_mem.hle_elision_buffer_full",
	.event = "umask=0x40,period=2000003,event=0x54",
	.desc = "Number of times HLE lock could not be elided due to ElisionBufferAvailable being zero",
	.topic = "memory",
	.long_desc = "Number of times we could not allocate Lock Buffer",
},
{
	.name = "tx_exec.misc1",
	.event = "umask=0x1,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions that may cause a transactional abort was executed. Since this is the count of execution, it may not always cause a transactional abort",
	.topic = "memory",
},
{
	.name = "tx_exec.misc2",
	.event = "umask=0x2,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions (e.g., vzeroupper) that may cause a transactional abort was executed inside a transactional region",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by a vzeroupper instruction",
},
{
	.name = "tx_exec.misc3",
	.event = "umask=0x4,period=2000003,event=0x5d",
	.desc = "Counts the number of times an instruction execution caused the transactional nest count supported to be exceeded",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by a nest count that is too deep",
},
{
	.name = "tx_exec.misc4",
	.event = "umask=0x8,period=2000003,event=0x5d",
	.desc = "Counts the number of times a XBEGIN instruction was executed inside an HLE transactional region",
	.topic = "memory",
	.long_desc = "RTM region detected inside HLE",
},
{
	.name = "tx_exec.misc5",
	.event = "umask=0x10,period=2000003,event=0x5d",
	.desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
	.topic = "memory",
	.long_desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
},
{
	.name = "offcore_requests_outstanding.l3_miss_demand_data_rd",
	.event = "umask=0x10,period=2000003,event=0x60",
	.desc = "Counts number of Offcore outstanding Demand Data Read requests that miss L3 cache in the superQ every cycle",
	.topic = "memory",
},
{
	.name = "offcore_requests_outstanding.cycles_with_l3_miss_demand_data_rd",
	.event = "umask=0x10,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles with at least 1 Demand Data Read requests who miss L3 cache in the superQ",
	.topic = "memory",
},
{
	.name = "offcore_requests_outstanding.l3_miss_demand_data_rd_ge_6",
	.event = "umask=0x10,period=2000003,cmask=6,event=0x60",
	.desc = "Cycles with at least 6 Demand Data Read requests that miss L3 cache in the superQ",
	.topic = "memory",
},
{
	.name = "cycle_activity.cycles_l3_miss",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Cycles while L3 cache miss demand load is outstanding",
	.topic = "memory",
},
{
	.name = "cycle_activity.stalls_l3_miss",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Execution stalls while L3 cache miss demand load is outstanding",
	.topic = "memory",
},
{
	.name = "offcore_requests.l3_miss_demand_data_rd",
	.event = "umask=0x10,period=100003,event=0xb0",
	.desc = "Demand Data Read requests who miss L3 cache",
	.topic = "memory",
	.long_desc = "Demand Data Read requests who miss L3 cache",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts  Spec update: SKL089",
	.topic = "memory",
	.long_desc = "Counts the number of memory ordering Machine Clears detected. Memory Ordering Machine Clears can result from one of the following:a. memory disambiguation,b. external snoop, orc. cross SMT-HW-thread snoop (stores) hitting load buffer  Spec update: SKL089",
},
{
	.name = "hle_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution started",
	.topic = "memory",
	.long_desc = "Number of times we entered an HLE region. Does not count nested transactions",
},
{
	.name = "hle_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution successfully committed",
	.topic = "memory",
	.long_desc = "Number of times HLE commit succeeded",
},
{
	.name = "hle_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times HLE abort was triggered. (PEBS) (Precise event)",
},
{
	.name = "hle_retired.aborted_mem",
	.event = "umask=0x8,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_timer",
	.event = "umask=0x10,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to hardware timer expiration",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_unfriendly",
	.event = "umask=0x20,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions and certain unfriendly events (such as AD assists etc.)",
	.topic = "memory",
	.long_desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions and certain unfriendly events (such as AD assists etc.)",
},
{
	.name = "hle_retired.aborted_memtype",
	.event = "umask=0x40,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times an HLE execution aborted due to incompatible memory type",
},
{
	.name = "hle_retired.aborted_events",
	.event = "umask=0x80,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to unfriendly events (such as interrupts)",
	.topic = "memory",
},
{
	.name = "rtm_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution started",
	.topic = "memory",
	.long_desc = "Number of times we entered an RTM region. Does not count nested transactions",
},
{
	.name = "rtm_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution successfully committed",
	.topic = "memory",
	.long_desc = "Number of times RTM commit succeeded",
},
{
	.name = "rtm_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times RTM abort was triggered. (PEBS) (Precise event)",
},
{
	.name = "rtm_retired.aborted_mem",
	.event = "umask=0x8,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
},
{
	.name = "rtm_retired.aborted_timer",
	.event = "umask=0x10,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to uncommon conditions",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_unfriendly",
	.event = "umask=0x20,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
},
{
	.name = "rtm_retired.aborted_memtype",
	.event = "umask=0x40,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to incompatible memory type",
},
{
	.name = "rtm_retired.aborted_events",
	.event = "umask=0x80,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 4 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 4 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 8 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 8 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 16 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 16 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 32 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 32 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 64 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 64 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 128 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 128 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 256 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 256 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 512 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 512 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "offcore_response.other.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFC408000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x203C408000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C408000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C408000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C408000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C408000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC408000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x007C408000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC4008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0044008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l4_hit_local_l4.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000408000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20001C8000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_s.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000108000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_e.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000088000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit_m.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000048000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000028000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFC400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x203C400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x007C400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC4000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0044000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l4_hit_local_l4.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000400004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20001C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_s.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000100004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_e.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000080004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit_m.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000040004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFC400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x203C400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x007C400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC4000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0044000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l4_hit_local_l4.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000400002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20001C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_s.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000100002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_e.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000080002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit_m.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000040002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FFC400001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x203C400001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103C400001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C400001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C400001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C400001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC400001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x007C400001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FC4000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.spl_hit",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0044000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l4_hit_local_l4.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000400001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20001C0001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_s.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000100001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_e.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000080001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit_m.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000040001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "sw_prefetch_access.nta",
	.event = "umask=0x1,period=2000003,event=0x32",
	.desc = "Number of PREFETCHNTA instructions executed",
	.topic = "other",
},
{
	.name = "sw_prefetch_access.t0",
	.event = "umask=0x2,period=2000003,event=0x32",
	.desc = "Number of PREFETCHT0 instructions executed",
	.topic = "other",
},
{
	.name = "sw_prefetch_access.t1_t2",
	.event = "umask=0x4,period=2000003,event=0x32",
	.desc = "Number of PREFETCHT1 or PREFETCHT2 instructions executed",
	.topic = "other",
},
{
	.name = "sw_prefetch_access.prefetchw",
	.event = "umask=0x8,period=2000003,event=0x32",
	.desc = "Number of PREFETCHW instructions executed",
	.topic = "other",
},
{
	.name = "hw_interrupts.received",
	.event = "umask=0x1,period=203,event=0xcb",
	.desc = "Number of hardware interrupts received by the processor",
	.topic = "other",
	.long_desc = "Counts the number of hardware interruptions received by the processor",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, Counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers. Notes: INST_RETIRED.ANY is counted by a designated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. INST_RETIRED.ANY_P is counted by a programmable counter and it is an architectural performance event. Counting: Faulting executions of GETSEC/VM entry/VM Exit/MWait will not count as retired instructions",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state. This event has a constant ratio with the CPU_CLK_UNHALTED.REF_XCLK event. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. Note: On all current platforms this event stops counting during 'throttling (TM)' states duty off periods the processor is 'halted'.  The counter update is done at a lower clock rate then the core clock the overflow status bit for this counter may appear 'sticky'.  After the counter has overflowed and software clears the overflow status bit and resets the counter to less than MAX. The reset value to the counter is not clocked immediately so the overflow status bit will flip 'high (1)' and generate another PMI (if enabled) after which the reset value gets clocked into the counter. Therefore, software will get the interrupt, read the overflow status bit '1 for bit 34 while the counter value is less than MAX. Software should ignore this case",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Loads blocked by overlapping with store buffer that cannot be forwarded ",
	.topic = "pipeline",
	.long_desc = "Counts how many times the load operation got the true Block-on-Store blocking code preventing store forwarding. This includes cases when:a. preceding store conflicts with the load (incomplete overlap),b. store forwarding is impossible due to u-arch limitations,c. preceding lock RMW operations are not forwarded,d. store has the no-forward bit set (uncacheable/page-split/masked stores),e. all-blocking stores are used (mostly, fences and port I/O), and others.The most common case is a load blocked due to its address range overlapping with a preceding smaller uncompleted store. Note: This event does not take into account cases of out-of-SW-control (for example, SbTailHit), unknown physical STA, and cases of blocking loads on store due to being non-WB memory type or a lock. These cases are covered by other events. See the table of not supported store forwards in the Optimization Guide",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
	.long_desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare on address",
	.topic = "pipeline",
	.long_desc = "Counts false dependencies in MOB when the partial comparison upon loose net check and dependency was resolved by the Enhanced Loose net mechanism. This may not result in high performance penalties. Loose net checks can fail when loads and stores are 4k aliased",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for this thread (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "Core cycles the Resource allocator was stalled due to recovery from an earlier branch misprediction or machine clear event",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "int_misc.clear_resteer_cycles",
	.event = "umask=0x80,period=2000003,event=0xd",
	.desc = "Cycles the issue-stage is waiting for front-end to fetch from resteered path following branch misprediction or machine clear events",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "Counts the number of uops that the Resource Allocation Table (RAT) issues to the Reservation Station (RS)",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which the Resource Allocation Table (RAT) does not issue any Uops to the reservation station (RS) for the current thread",
},
{
	.name = "uops_issued.vector_width_mismatch",
	.event = "umask=0x2,period=2000003,event=0xe",
	.desc = "Uops inserted at issue-stage in order to preserve upper bits of vector registers",
	.topic = "pipeline",
	.long_desc = "Counts the number of Blend Uops issued by the Resource Allocation Table (RAT) to the reservation station (RS) in order to preserve upper bits of vector registers. Starting with the Skylake microarchitecture, these Blend uops are needed since every Intel SSE instruction executed in Dirty Upper State needs to preserve bits 128-255 of the destination register. For more information, refer to \u201cMixing Intel AVX and Intel SSE Code\u201d section of the Optimization Guide",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
},
{
	.name = "arith.divider_active",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x14",
	.desc = "Cycles when divide unit is busy executing divide or square root operations. Accounts for integer and floating-point operations",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This is an architectural event that counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling. For this reason, this event may have a changing ratio with regards to wall clock time",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ring0_trans",
	.event = "umask=0x0,edge=1,period=100007,cmask=1,event=0x3c",
	.desc = "Counts when there is a transition from ring 1, 2 or 3 to ring 0",
	.topic = "pipeline",
	.long_desc = "Counts when the Current Privilege Level (CPL) transitions from ring 1, 2 or 3 to ring 0 (Kernel)",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when the thread is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when at least one thread on the physical core is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when the thread is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when at least one thread on the physical core is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Core crystal clock cycles when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Demand load dispatches that hit L1D fill buffer (FB) allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "Counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the software prefetch. It can also be incremented by some lock instructions. So it should only be used with profiling so that the locks can be excluded by ASM (Assembly File) inspection of the nearby instructions",
},
{
	.name = "partial_rat_stalls.scoreboard",
	.event = "umask=0x1,period=2000003,event=0x59",
	.desc = "Cycles where the pipeline is stalled due to serializing operations",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which the microcode scoreboard stalls happen",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which the reservation station (RS) is empty for the thread.; Note: In ST-mode, not active thread should drive 0. This is usually caused by severely costly branch mispredictions, or allocator/FE issues",
},
{
	.name = "rs_events.empty_end",
	.event = "inv=1,umask=0x1,edge=1,period=2000003,cmask=1,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
	.long_desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate front-end Latency Bound issues",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
	.long_desc = "Counts cycles that the Instruction Length decoder (ILD) stalls occurred due to dynamically changing prefix length of the decoded instruction (by operand size prefix instruction 0x66, address size prefix instruction 0x67 or REX.W for Intel64). Count is proportional to the number of prefixes in a 16B-line. This may result in a three-cycle penalty for each LCP (Length changing prefix) in a 16-byte chunk",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 2",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 3",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 4",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_dispatched_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_dispatched_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 7",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
	.long_desc = "Counts resource-related stall cycles",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "Counts allocation stall cycles caused by the store buffer (SB) being full. This counts cycles that the pipeline back-end blocked uop delivery from the front-end",
},
{
	.name = "cycle_activity.cycles_l2_miss",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_total",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "Total execution stalls",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_miss",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_miss",
	.event = "umask=0x8,period=2000003,cmask=8,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_miss",
	.event = "umask=0xc,period=2000003,cmask=12,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_mem_any",
	.event = "umask=0x10,period=2000003,cmask=16,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_mem_any",
	.event = "umask=0x14,period=2000003,cmask=20,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "exe_activity.exe_bound_0_ports",
	.event = "umask=0x1,period=2000003,event=0xa6",
	.desc = "Cycles where no uops were executed, the Reservation Station was not empty, the Store Buffer was full and there was no outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which no uops were executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.1_ports_util",
	.event = "umask=0x2,period=2000003,event=0xa6",
	.desc = "Cycles total of 1 uop is executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which a total of 1 uop was executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.2_ports_util",
	.event = "umask=0x4,period=2000003,event=0xa6",
	.desc = "Cycles total of 2 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which a total of 2 uops were executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.3_ports_util",
	.event = "umask=0x8,period=2000003,event=0xa6",
	.desc = "Cycles total of 3 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Cycles total of 3 uops are executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.4_ports_util",
	.event = "umask=0x10,period=2000003,event=0xa6",
	.desc = "Cycles total of 4 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Cycles total of 4 uops are executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.bound_on_stores",
	.event = "umask=0x40,period=2000003,event=0xa6",
	.desc = "Cycles where the Store Buffer was full and no outstanding load",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
	.long_desc = "Number of uops delivered to the back-end by the LSD(Loop Stream Detector)",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Counts the cycles when at least one uop is delivered by the LSD (Loop-stream detector)",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Counts the cycles when 4 uops are delivered by the LSD (Loop-stream detector)",
},
{
	.name = "uops_executed.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Counts the number of uops to be executed per-thread each cycle",
	.topic = "pipeline",
	.long_desc = "Number of uops to be executed per-thread each cycle",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which no uops were dispatched from the Reservation Station (RS) per thread",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 1 uop was executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 2 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 3 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 4 uops were executed per-thread",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core",
	.topic = "pipeline",
	.long_desc = "Number of uops executed from any thread",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.x87",
	.event = "umask=0x10,period=2000003,event=0xb1",
	.desc = "Counts the number of x87 uops dispatched",
	.topic = "pipeline",
	.long_desc = "Counts the number of x87 uops executed",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter - architectural event  Spec update: SKL091, SKL044",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions (EOMs) retired. Counting covers macro-fused instructions individually (that is, increments by two)  Spec update: SKL091, SKL044",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: SKL091, SKL044 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "A version of INST_RETIRED that allows for a more unbiased distribution of samples across instructions retired. It utilizes the Precise Distribution of Instructions Retired (PDIR) feature to mitigate some bias in how retired instructions get sampled  Spec update: SKL091, SKL044 (Must be precise)",
},
{
	.name = "inst_retired.total_cycles_ps",
	.event = "inv=1,umask=0x1,period=2000003,cmask=10,event=0xc0",
	.desc = "Number of cycles using always true condition applied to  PEBS instructions retired event  Spec update: SKL091, SKL044 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Number of cycles using an always true condition applied to  PEBS instructions retired event. (inst_ret< 16)  Spec update: SKL091, SKL044 (Must be precise)",
},
{
	.name = "other_assists.any",
	.event = "umask=0x3f,period=100003,event=0xc1",
	.desc = "Number of times a microcode assist is invoked by HW other than FP-assist. Examples include AD (page Access Dirty) and AVX* related assists",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used",
	.topic = "pipeline",
	.long_desc = "Counts the retirement slots used",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x2,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
	.long_desc = "This event counts cycles without actually retired uops",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x2,period=2000003,cmask=10,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
	.long_desc = "Number of cycles using always true condition (uops_ret < 16) applied to non PEBS uops retired event",
},
{
	.name = "machine_clears.count",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
	.long_desc = "Number of machine clears (nukes) of any type",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "Counts self-modifying code (SMC) detected, which causes a machine clear",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired  Spec update: SKL091",
	.topic = "pipeline",
	.long_desc = "Counts all (macro) branch instructions retired  Spec update: SKL091",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts conditional branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts both direct and indirect near call instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired  Spec update: SKL091 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_INST_RETIRED.ALL_BRANCHES that counts all (macro) branch instructions retired  Spec update: SKL091 (Must be precise)",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts return instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Counts all not taken macro branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts not taken branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts taken branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Counts the number of far branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts far branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Counts all the retired branch instructions that were mispredicted by the processor. A branch misprediction occurs when the processor incorrectly predicts the destination of the branch.  When the misprediction is discovered at execution, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts mispredicted conditional branch instructions retired (Precise event)",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=400009,event=0xc5",
	.desc = "Mispredicted direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken retired mispredicted direct and indirect near calls, including both register and memory indirect (Precise event)",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_MISP_RETIRED.ALL_BRANCHES that counts all mispredicted macro branch instructions retired (Must be precise)",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "Number of near branch instructions retired that were mispredicted and taken (Precise event)",
	.topic = "pipeline",
	.long_desc = "Number of near branch instructions retired that were mispredicted and taken (Precise event)",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Increments whenever there is an update to the LBR array",
	.topic = "pipeline",
	.long_desc = "Increments when an entry is added to the Last Branch Record (LBR) array (or removed from the array in case of RETURNs in call stack mode). The event requires LBR enable via IA32_DEBUGCTL MSR and branch type selection via MSR_LBR_SELECT",
},
{
	.name = "rob_misc_events.pause_inst",
	.event = "umask=0x40,period=2000003,event=0xcc",
	.desc = "Number of retired PAUSE instructions (that do not end up with a VMExit to the VMM; TSX aborted Instructions may be counted). This event is not supported on first SKL and KBL products",
	.topic = "pipeline",
},
{
	.name = "baclears.any",
	.event = "umask=0x1,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
	.long_desc = "Counts the number of times the front-end is resteered when it finds a branch instruction in a fetch line. This occurs for the first time a branch instruction is fetched or when the branch is not tracked by the BPU (Branch Prediction Unit) anymore",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "skl metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "skl metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "skl metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "skl metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "skl metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "skl metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "skl metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "skl metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "skl metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Instruction per taken branch",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_taken",
	.metric_name = "IpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Branch instructions per taken branch",
	.topic = "skl metrics",
	.metric_expr = "br_inst_retired.all_branches / br_inst_retired.near_taken",
	.metric_name = "BpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "skl metrics",
	.metric_expr = "min( 1 , uops_issued.any / ( (uops_retired.retire_slots / inst_retired.any) * 64 * ( icache_64b.iftag_hit + icache_64b.iftag_miss ) / 4.1 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "skl metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + idq.mite_uops + idq.ms_uops ))",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "skl metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "skl metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "skl metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "skl metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Instructions per Load (lower number means loads are more frequent)",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any / mem_inst_retired.all_loads",
	.metric_name = "IpL",
	.metric_group = "Instruction_Type;L1_Bound",
},
{
	.desc = "Instructions per Store",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any / mem_inst_retired.all_stores",
	.metric_name = "IpS",
	.metric_group = "Instruction_Type;Store_Bound",
},
{
	.desc = "Instructions per Branch",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.all_branches",
	.metric_name = "IpB",
	.metric_group = "Branches;Instruction_Type;Port_5;Port_6",
},
{
	.desc = "Instruction per (near) call",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_call",
	.metric_name = "IpCall",
	.metric_group = "Branches",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "skl metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * fp_arith_inst_retired.256b_packed_single )) / cycles",
	.metric_name = "FLOPc",
	.metric_group = "FLOPS",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "skl metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * fp_arith_inst_retired.256b_packed_single )) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "FLOPc_SMT",
	.metric_group = "FLOPS_SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "skl metrics",
	.metric_expr = "uops_executed.thread / (( uops_executed.core_cycles_ge_1 / 2 ) if #smt_on else uops_executed.core_cycles_ge_1)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "skl metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) * (( int_misc.clear_resteer_cycles + 9 * baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) ) * (4 * cycles) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "skl metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) * (( int_misc.clear_resteer_cycles + 9 * baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) ) * (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost_SMT",
	.metric_group = "Branch_Mispredicts_SMT",
},
{
	.desc = "Number of Instructions per non-speculative Branch Misprediction (JEClear)",
	.topic = "skl metrics",
	.metric_expr = "inst_retired.any / br_misp_retired.all_branches",
	.metric_name = "IpMispredict",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "skl metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads (in core cycles)",
	.topic = "skl metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_retired.l1_miss + mem_load_retired.fb_hit )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least one such miss. Per-thread)",
	.topic = "skl metrics",
	.metric_expr = "l1d_pend_miss.pending / l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "skl metrics",
	.metric_expr = "( itlb_misses.walk_pending + dtlb_load_misses.walk_pending + dtlb_store_misses.walk_pending + ept.walk_pending ) / ( 2 * cycles )",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "skl metrics",
	.metric_expr = "( itlb_misses.walk_pending + dtlb_load_misses.walk_pending + dtlb_store_misses.walk_pending + ept.walk_pending ) / ( 2 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )) )",
	.metric_name = "Page_Walks_Utilization_SMT",
	.metric_group = "TLB_SMT",
},
{
	.desc = "Average data fill bandwidth to the L1 data cache [GB / sec]",
	.topic = "skl metrics",
	.metric_expr = "64 * l1d.replacement / 1000000000 / duration_time",
	.metric_name = "L1D_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average data fill bandwidth to the L2 cache [GB / sec]",
	.topic = "skl metrics",
	.metric_expr = "64 * l2_lines_in.all / 1000000000 / duration_time",
	.metric_name = "L2_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "skl metrics",
	.metric_expr = "64 * longest_lat_cache.miss / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "skl metrics",
	.metric_expr = "64 * offcore_requests.all_requests / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Access_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "L1 cache true misses per kilo instruction for retired demand loads",
	.topic = "skl metrics",
	.metric_expr = "1000 * mem_load_retired.l1_miss / inst_retired.any",
	.metric_name = "L1MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache true misses per kilo instruction for retired demand loads",
	.topic = "skl metrics",
	.metric_expr = "1000 * mem_load_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache misses per kilo instruction for all request types (including speculative)",
	.topic = "skl metrics",
	.metric_expr = "1000 * l2_rqsts.miss / inst_retired.any",
	.metric_name = "L2MPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache hits per kilo instruction for all request types (including speculative)",
	.topic = "skl metrics",
	.metric_expr = "1000 * ( l2_rqsts.references - l2_rqsts.miss ) / inst_retired.any",
	.metric_name = "L2HPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L3 cache true misses per kilo instruction for retired demand loads",
	.topic = "skl metrics",
	.metric_expr = "1000 * mem_load_retired.l3_miss / inst_retired.any",
	.metric_name = "L3MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "Average CPU Utilization",
	.topic = "skl metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "skl metrics",
	.metric_expr = "( (( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * fp_arith_inst_retired.256b_packed_single )) / 1000000000 ) / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "skl metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "skl metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "skl metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "skl metrics",
	.metric_expr = "64 * ( arb@event\\=0x81\\,umask\\=0x1@ + arb@event\\=0x84\\,umask\\=0x1@ ) / 1000000 / duration_time / 1000",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average number of parallel data read requests to external memory. Accounts for demand loads and L1/L2 prefetches",
	.topic = "skl metrics",
	.metric_expr = "arb@event\\=0x80\\,umask\\=0x2@ / arb@event\\=0x80\\,umask\\=0x2\\,thresh\\=1@",
	.metric_name = "DRAM_Parallel_Reads",
	.metric_group = "Memory_BW",
},
{
	.desc = "C3 residency percent per core",
	.topic = "skl metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "skl metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "skl metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "skl metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "skl metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "skl metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "skl metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "unc_cbo_xsnp_response.miss_xcore",
	.event = "umask=0x41,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop initiated by this Cbox due to processor core memory request which misses in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop initiated by this Cbox due to processor core memory request which misses in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.miss_eviction",
	.event = "umask=0x81,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop resulted from L3 Eviction which misses in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop resulted from L3 Eviction which misses in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hit_xcore",
	.event = "umask=0x44,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop initiated by this Cbox due to processor core memory request which hits a non-modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop initiated by this Cbox due to processor core memory request which hits a non-modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hitm_xcore",
	.event = "umask=0x48,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop initiated by this Cbox due to processor core memory request which hits a modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop initiated by this Cbox due to processor core memory request which hits a modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_m",
	.event = "umask=0x21,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_m",
	.event = "umask=0x81,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_i",
	.event = "umask=0x18,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_i",
	.event = "umask=0x88,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_mesi",
	.event = "umask=0x1f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in any MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in any MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_mesi",
	.event = "umask=0x2f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_mesi",
	.event = "umask=0x8f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_es",
	.event = "umask=0x86,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_es",
	.event = "umask=0x16,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_es",
	.event = "umask=0x26,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_arb_trk_occupancy.all",
	.event = "umask=0x01,event=0x80",
	.desc = "Unit: uncore_arb Each cycle count number of all Core outgoing valid entries. Such entry is defined as valid from its allocation till first of IDI0 or DRS0 messages is sent out. Accounts for Coherent and non-coherent traffic",
	.topic = "uncore",
	.long_desc = "Each cycle count number of all Core outgoing valid entries. Such entry is defined as valid from its allocation till first of IDI0 or DRS0 messages is sent out. Accounts for Coherent and non-coherent traffic",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.all",
	.event = "umask=0x01,event=0x81",
	.desc = "Unit: uncore_arb Total number of Core outgoing entries allocated. Accounts for Coherent and non-coherent traffic",
	.topic = "uncore",
	.long_desc = "Total number of Core outgoing entries allocated. Accounts for Coherent and non-coherent traffic",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.drd_direct",
	.event = "umask=0x02,event=0x81",
	.desc = "Unit: uncore_arb Number of Core coherent Data Read entries allocated in DirectData mode",
	.topic = "uncore",
	.long_desc = "Number of Core coherent Data Read entries allocated in DirectData mode",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.writes",
	.event = "umask=0x20,event=0x81",
	.desc = "Unit: uncore_arb Number of Writes allocated - any write transactions: full/partials writes and evictions",
	.topic = "uncore",
	.long_desc = "Number of Writes allocated - any write transactions: full/partials writes and evictions",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_coh_trk_requests.all",
	.event = "umask=0x01,event=0x84",
	.desc = "Unit: uncore_arb Number of entries allocated. Account for Any type: e.g. Snoop, Core aperture, etc",
	.topic = "uncore",
	.long_desc = "Number of entries allocated. Account for Any type: e.g. Snoop, Core aperture, etc",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_occupancy.cycles_with_any_request",
	.event = "umask=0x01,cmask=1,event=0x80",
	.desc = "Unit: uncore_arb Cycles with at least one request outstanding is waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC.;",
	.topic = "uncore",
	.long_desc = "Cycles with at least one request outstanding is waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_clock.socket",
	.event = "umask=0x01,event=0",
	.desc = "Unit: uncore_ncu This 48-bit fixed counter counts the UCLK cycles",
	.topic = "uncore",
	.long_desc = "This 48-bit fixed counter counts the UCLK cycles",
	.pmu = "uncore_ncu",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts demand data loads that caused a page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels, but the walk need not have completed",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x8",
	.desc = "Page walk completed due to a demand data load to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Page walk completed due to a demand data load to a 2M/4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 2M/4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_1g",
	.event = "umask=0x8,period=2000003,event=0x8",
	.desc = "Page walk completed due to a demand data load to a 1G page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x8",
	.desc = "Load miss in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts demand data loads that caused a completed page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels. The page walk can end with or without a fault",
},
{
	.name = "dtlb_load_misses.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x8",
	.desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a load. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a load. EPT page walk duration are excluded in Skylake microarchitecture",
},
{
	.name = "dtlb_load_misses.walk_active",
	.event = "umask=0x10,period=100003,cmask=1,event=0x8",
	.desc = "Cycles when at least one PMH is busy with a page walk for a load. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts cycles when at least one PMH (Page Miss Handler) is busy with a page walk for a load",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x20,period=2000003,event=0x8",
	.desc = "Loads that miss the DTLB and hit the STLB",
	.topic = "virtual memory",
	.long_desc = "Counts loads that miss the DTLB (Data TLB) and hit the STLB (Second level TLB)",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts demand data stores that caused a page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels, but the walk need not have completed",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 2M/4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 2M/4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 1G page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 1G pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x49",
	.desc = "Store misses in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts demand data stores that caused a completed page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels. The page walk can end with or without a fault",
},
{
	.name = "dtlb_store_misses.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x49",
	.desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a store. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a store. EPT page walk duration are excluded in Skylake microarchitecture",
},
{
	.name = "dtlb_store_misses.walk_active",
	.event = "umask=0x10,period=100003,cmask=1,event=0x49",
	.desc = "Cycles when at least one PMH is busy with a page walk for a store. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts cycles when at least one PMH (Page Miss Handler) is busy with a page walk for a store",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x20,period=100003,event=0x49",
	.desc = "Stores that miss the DTLB and hit the STLB",
	.topic = "virtual memory",
	.long_desc = "Stores that miss the DTLB (Data TLB) and hit the STLB (2nd Level TLB)",
},
{
	.name = "ept.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Counts 1 per cycle for each PMH that is busy with a EPT (Extended Page Table) walk for any request type",
	.topic = "virtual memory",
	.long_desc = "Counts cycles for each PMH (Page Miss Handler) that is busy with an EPT (Extended Page Table) walk for any request type",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts page walks of any page size (4K/2M/4M/1G) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB, but the walk need not have completed",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (4K)",
	.topic = "virtual memory",
	.long_desc = "Counts completed page walks (4K page size) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB. The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Counts code misses in all ITLB levels that caused a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (1G)",
	.topic = "virtual memory",
	.long_desc = "Counts store misses in all DTLB levels that cause a completed page walk (1G page size). The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts completed page walks (2M and 4M page sizes) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB. The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_pending",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Counts 1 per cycle for each PMH that is busy with a page walk for an instruction fetch request. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts 1 per cycle for each PMH (Page Miss Handler) that is busy with a page walk for an instruction fetch request. EPT page walk duration are excluded in Skylake michroarchitecture",
},
{
	.name = "itlb_misses.walk_active",
	.event = "umask=0x10,period=100003,cmask=1,event=0x85",
	.desc = "Cycles when at least one PMH is busy with a page walk for code (instruction fetch) request. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Cycles when at least one PMH is busy with a page walk for code (instruction fetch) request. EPT page walk duration are excluded in Skylake microarchitecture",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x20,period=100003,event=0x85",
	.desc = "Instruction fetch requests that miss the ITLB and hit the STLB",
	.topic = "virtual memory",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "Counts the number of flushes of the big or small ITLB pages. Counting include both TLB Flush (covering all sets) and TLB Set Clear (set-specific)",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "Counts the number of DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "Counts the number of any STLB flush attempts (such as entire, VPID, PCID, InvPage, CR3 write, etc.)",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_skylakex[] = {
{
	.name = "l2_rqsts.demand_data_rd_miss",
	.event = "umask=0x21,period=200003,event=0x24",
	.desc = "Demand Data Read miss L2, no rejects",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests that miss L2 cache. Only not rejected loads are counted",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x22,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts the RFO (Read-for-Ownership) requests that miss L2 cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x24,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
	.long_desc = "Counts L2 cache misses when fetching instructions",
},
{
	.name = "l2_rqsts.all_demand_miss",
	.event = "umask=0x27,period=200003,event=0x24",
	.desc = "Demand requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "Demand requests that miss L2 cache",
},
{
	.name = "l2_rqsts.pf_miss",
	.event = "umask=0x38,period=200003,event=0x24",
	.desc = "Requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that miss L2 cache",
	.topic = "cache",
	.long_desc = "Counts requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that miss L2 cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0x3f,period=200003,event=0x24",
	.desc = "All requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "All requests that miss L2 cache",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0xc1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests, initiated by load instructions, that hit L2 cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0xc2,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the RFO (Read-for-Ownership) requests that hit L2 cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0xc4,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
	.long_desc = "Counts L2 cache hits when fetching instructions, code reads",
},
{
	.name = "l2_rqsts.pf_hit",
	.event = "umask=0xd8,period=200003,event=0x24",
	.desc = "Requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts requests from the L1/L2/L3 hardware prefetchers or Load software prefetches that hit L2 cache",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0xe1,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests (including requests from L1D hardware prefetchers). These loads may hit or miss L2 cache. Only non rejected loads are counted",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xe2,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "Counts the total number of RFO (read for ownership) requests to L2 cache. L2 RFO requests include both L1D demand RFO misses as well as L1D RFO prefetches",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0xe4,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "Counts the total number of L2 code requests",
},
{
	.name = "l2_rqsts.all_demand_references",
	.event = "umask=0xe7,period=200003,event=0x24",
	.desc = "Demand requests to L2 cache",
	.topic = "cache",
	.long_desc = "Demand requests to L2 cache",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xf8,period=200003,event=0x24",
	.desc = "Requests from the L1/L2/L3 hardware prefetchers or Load software prefetches",
	.topic = "cache",
	.long_desc = "Counts the total number of requests from the L2 hardware prefetchers",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200003,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
	.long_desc = "All L2 requests",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed L3  Spec update: SKL057",
	.topic = "cache",
	.long_desc = "Counts core-originated cacheable requests that miss the L3 cache (Longest Latency cache). Requests include data and code reads, Reads-for-Ownership (RFOs), speculative accesses and hardware prefetches from L1 and L2. It does not include all misses to the L3  Spec update: SKL057",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to L3  Spec update: SKL057",
	.topic = "cache",
	.long_desc = "Counts core-originated cacheable requests to the  L3 cache (Longest Latency cache). Requests include data and code reads, Reads-for-Ownership (RFOs), speculative accesses and hardware prefetches from L1 and L2.  It does not include all accesses to the L3  Spec update: SKL057",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
	.long_desc = "Counts duration of L1D miss outstanding in cycles",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss outstandings duration in cycles",
	.topic = "cache",
	.long_desc = "Counts duration of L1D miss outstanding, that is each cycle number of Fill Buffers (FB) outstanding required by Demand Reads. FB either is held by demand loads, or it is held by non-demand loads and gets hit at least once by demand. The valid outstanding interval is defined until the FB deallocation by one of the following ways: from FB allocation, if FB is allocated by demand from the demand Hit FB, if it is allocated by hardware or software prefetch.Note: In the L1D, a Demand Read contains cacheable or noncacheable demand loads, including ones causing cache-line splits and reads due to page walks resulted from any request type",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,cmask=1,period=2000003,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,event=0x48",
	.desc = "Number of times a request needed a FB entry but there was no entry available for it. That is the FB unavailability was dominant reason for blocking the request. A request includes cacheable/uncacheable demands that is load, store or SW prefetch",
	.topic = "cache",
	.long_desc = "Number of times a request needed a FB (Fill Buffer) entry but there was no entry available for it. A request includes cacheable/uncacheable demands that are load, store or SW prefetch instructions",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "Counts L1D data line replacements including opportunistic replacements, and replacements that require stall-for-replace or block-for-replace",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts cycles when offcore outstanding Demand Data Read transactions are present in the super queue (SQ). A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation)",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding Demand Data Read transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor. See the corresponding Umask under OFFCORE_REQUESTS.Note: A prefetch promoted to Demand is counted from the promotion point",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,cmask=6,period=2000003,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding Code Reads transactions in the SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding Code Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_code_rd",
	.event = "umask=0x2,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles with offcore outstanding Code Reads transactions in the SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding Code Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding RFO (store) transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles with offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding demand rfo Reads transactions in the super queue every cycle. The 'Offcore outstanding' state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts cycles when offcore outstanding cacheable Core Data Read transactions are present in the super queue. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of offcore outstanding cacheable Core Data Read transactions in the super queue every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
	.long_desc = "Counts the Demand Data Read requests sent to uncore. Use it in conjunction with OFFCORE_REQUESTS_OUTSTANDING to determine average latency in the uncore",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "Counts both cacheable and non-cacheable code read requests",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "Counts the demand RFO (read for ownership) requests including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts the demand and prefetch data reads. All Core Data Reads include cacheable 'Demands' and L2 prefetchers (not L3 prefetchers). Counting also covers reads due to page walks resulted from any request type",
},
{
	.name = "offcore_requests.all_requests",
	.event = "umask=0x80,period=100003,event=0xb0",
	.desc = "Any memory transaction that reached the SQ",
	.topic = "cache",
	.long_desc = "Counts memory transactions reached the super queue including requests initiated by the core, all L3 prefetches, page walks, etc.",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Offcore requests buffer cannot take more entries for this thread core",
	.topic = "cache",
	.long_desc = "Counts the number of cases when the offcore requests buffer cannot take more entries for the core. This can happen when the superqueue does not contain eligible entries, or when L1D writeback pending FIFO requests is full.Note: Writeback pending FIFO has six entries",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
	.topic = "cache",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "mem_inst_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load instructions that miss the STLB. (Precise Event)  Supports address when precise",
	.topic = "cache",
	.long_desc = "Retired load instructions that miss the STLB  Supports address when precise",
},
{
	.name = "mem_inst_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store instructions that miss the STLB. (Precise Event)  Supports address when precise",
	.topic = "cache",
	.long_desc = "Retired store instructions that miss the STLB  Supports address when precise",
},
{
	.name = "mem_inst_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load instructions with locked access. (Precise Event)  Supports address when precise",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load instructions that split across a cacheline boundary. (Precise Event)  Supports address when precise",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store instructions that split across a cacheline boundary. (Precise Event)  Supports address when precise",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load instructions. (Precise Event)  Supports address when precise",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "All retired store instructions. (Precise Event)  Supports address when precise",
	.topic = "cache",
	.long_desc = "All retired store instructions  Supports address when precise",
},
{
	.name = "mem_load_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load instructions with L1 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that hit in the L1 data cache. This event includes all SW prefetches and lock instructions regardless of the data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load instructions with L2 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions with L2 cache hits as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l3_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Retired load instructions with L3 cache hits as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions with L3 cache hits as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load instructions missed L1 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop that missed in the L1 cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Retired load instructions missed L2 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions missed L2 cache as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.l3_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Retired load instructions missed L3 cache as data sources  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions missed L3 cache as data sources  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_retired.fb_hit",
	.event = "umask=0x40,period=100007,event=0xd1",
	.desc = "Retired load instructions which data sources were load missed L1 but hit FB due to preceding miss to the same cache line with data not ready  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts retired load instructions with at least one uop was load missed in L1 but hit FB (Fill Buffers) due to preceding miss to the same cache line with data not ready  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load instructions which data sources were L3 hit and cross-core snoop missed in on-pkg core cache  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load instructions which data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources were L3 and cross-core snoop hits in on-pkg core cache  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load instructions which data sources were HitM responses from shared L3  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources were HitM responses from shared L3  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load instructions which data sources were hits in L3 without snoops required  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Retired load instructions which data sources were hits in L3 without snoops required  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_l3_miss_retired.local_dram",
	.event = "umask=0x1,period=100007,event=0xd3",
	.desc = "Retired load instructions which data sources missed L3 but serviced from local dram  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_l3_miss_retired.remote_dram",
	.event = "umask=0x2,period=100007,event=0xd3",
	.desc = "Retired load instructions which data sources missed L3 but serviced from remote dram  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_l3_miss_retired.remote_hitm",
	.event = "umask=0x4,period=100007,event=0xd3",
	.desc = "Retired load instructions whose data sources was remote HITM  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_l3_miss_retired.remote_fwd",
	.event = "umask=0x8,period=100007,event=0xd3",
	.desc = "Retired load instructions whose data sources was forwarded from a remote cache  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_misc_retired.uc",
	.event = "umask=0x4,period=100007,event=0xd4",
	.desc = "Retired instructions with at least 1 uncacheable load or lock  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "Counts L2 writebacks that access L2 cache",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x1f,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "Counts the number of L2 cache lines filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_out.silent",
	.event = "umask=0x1,period=200003,event=0xf2",
	.desc = "Counts the number of lines that are silently dropped by L2 cache when triggered by an L2 cache fill. These lines are typically in Shared state. A non-threaded event",
	.topic = "cache",
	.long_desc = "Counts the number of lines that are silently dropped by L2 cache when triggered by an L2 cache fill. These lines are typically in Shared or Exclusive state. A non-threaded event",
},
{
	.name = "l2_lines_out.non_silent",
	.event = "umask=0x2,period=200003,event=0xf2",
	.desc = "Counts the number of lines that are evicted by L2 cache when triggered by an L2 cache fill. Those lines can be either in modified state or clean state. Modified lines may either be written back to L3 or directly written to memory and not allocated in L3.  Clean lines may either be allocated in L3 or dropped",
	.topic = "cache",
	.long_desc = "Counts the number of lines that are evicted by L2 cache when triggered by an L2 cache fill. Those lines are in Modified state. Modified lines are written back to L3",
},
{
	.name = "l2_lines_out.useless_pref",
	.event = "umask=0x4,period=200003,event=0xf2",
	.desc = "This event is deprecated. Refer to new event L2_LINES_OUT.USELESS_HWPF",
	.topic = "cache",
	.long_desc = "This event is deprecated. Refer to new event L2_LINES_OUT.USELESS_HWPF",
},
{
	.name = "l2_lines_out.useless_hwpf",
	.event = "umask=0x4,period=200003,event=0xf2",
	.desc = "Counts the number of lines that have been hardware prefetched but not used and now evicted by L2 cache",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Number of cache line split locks sent to uncore",
	.topic = "cache",
	.long_desc = "Counts the number of cache line split locks sent to the uncore",
},
{
	.name = "offcore_response.demand_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010001",
	.desc = "Counts demand data reads have any response type",
	.topic = "cache",
	.long_desc = "Counts demand data reads have any response type",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts demand data reads TBD TBD",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts demand data reads TBD TBD",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts demand data reads TBD TBD",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts demand data reads TBD TBD",
},
{
	.name = "offcore_response.demand_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010002",
	.desc = "Counts all demand data writes (RFOs) have any response type",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) have any response type",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) TBD TBD",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) TBD TBD",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) TBD TBD",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) TBD TBD",
},
{
	.name = "offcore_response.demand_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that have any response type",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that have any response type",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010010",
	.desc = "Counts prefetch (that bring data to L2) data reads have any response type",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads have any response type",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs have any response type",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs have any response type",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads have any response type",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads have any response type",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs have any response type",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs have any response type",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests have any response type",
	.topic = "cache",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests have any response type",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "cache",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "cache",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "cache",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "cache",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010490",
	.desc = "TBD have any response type",
	.topic = "cache",
	.long_desc = "TBD have any response type",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0490",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0490",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0490",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0490",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010120",
	.desc = "TBD have any response type",
	.topic = "cache",
	.long_desc = "TBD have any response type",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0120",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0120",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0120",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0120",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010491",
	.desc = "TBD have any response type",
	.topic = "cache",
	.long_desc = "TBD have any response type",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0491",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0491",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0491",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0491",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010122",
	.desc = "TBD have any response type",
	.topic = "cache",
	.long_desc = "TBD have any response type",
},
{
	.name = "offcore_response.all_rfo.l3_hit.no_snoop_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0122",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_rfo.l3_hit.hit_other_core_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0122",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_rfo.l3_hit.hitm_other_core",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0122",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0122",
	.desc = "TBD TBD TBD",
	.topic = "cache",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
	.topic = "cache",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
	.topic = "cache",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "TBD",
	.topic = "cache",
	.long_desc = "TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "TBD",
	.topic = "cache",
	.long_desc = "TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "TBD",
	.topic = "cache",
	.long_desc = "TBD",
},
{
	.name = "offcore_response.all_rfo.l3_hit.snoop_hit_with_fwd",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "TBD",
	.topic = "cache",
	.long_desc = "TBD",
},
{
	.name = "fp_arith_inst_retired.scalar_double",
	.event = "umask=0x1,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar double precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.scalar_single",
	.event = "umask=0x2,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar single precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_double",
	.event = "umask=0x4,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed double precision floating-point instructions retired.  Each count represents 2 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_single",
	.event = "umask=0x8,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed single precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_double",
	.event = "umask=0x10,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed double precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_single",
	.event = "umask=0x20,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed single precision floating-point instructions retired.  Each count represents 8 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.512b_packed_double",
	.event = "umask=0x40,period=2000003,event=0xc7",
	.desc = "Number of Packed Double-Precision FP arithmetic instructions (Use operation multiplier of 8)",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.512b_packed_single",
	.event = "umask=0x80,period=2000003,event=0xc7",
	.desc = "Number of Packed Single-Precision FP arithmetic instructions (Use operation multiplier of 16)",
	.topic = "floating point",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,cmask=1,period=100003,event=0xca",
	.desc = "Cycles with any input/output SSE or FP assist",
	.topic = "floating point",
	.long_desc = "Counts cycles with any input and output SSE or x87 FP assist. If an input and output assist are detected on the same cycle the event increments by 1",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "Counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may 'bypass' the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "Counts the number of uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles uops were delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Count includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles 4 uops were delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Count includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts the number of uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may 'bypass' the IDQ",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles uops were delivered to the Instruction Decode Queue (IDQ) from the MITE (legacy decode pipeline) path. Counting includes uops that may 'bypass' the IDQ. During these cycles uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,cmask=4,period=2000003,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "Counts the number of cycles 4 uops were delivered to the Instruction Decode Queue (IDQ) from the MITE (legacy decode pipeline) path. Counting includes uops that may 'bypass' the IDQ. During these cycles uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may 'bypass' the IDQ. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "Counts the total number of uops delivered by the Microcode Sequencer (MS). Any instruction over 4 uops will be delivered by the MS. Some instructions such as transcendentals may additionally generate uops from the MS",
},
{
	.name = "idq.ms_switches",
	.event = "edge=1,umask=0x30,cmask=1,period=2000003,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
	.long_desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
},
{
	.name = "icache_16b.ifdata_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction cache miss",
	.topic = "frontend",
	.long_desc = "Cycles where a code line fetch is stalled due to an L1 instruction cache miss. The legacy decode pipeline works at a 16 Byte granularity",
},
{
	.name = "icache_64b.iftag_hit",
	.event = "umask=0x1,period=200003,event=0x83",
	.desc = "Instruction fetch tag lookups that hit in the instruction cache (L1I). Counts at 64-byte cache-line granularity",
	.topic = "frontend",
},
{
	.name = "icache_64b.iftag_miss",
	.event = "umask=0x2,period=200003,event=0x83",
	.desc = "Instruction fetch tag lookups that miss in the instruction cache (L1I). Counts at 64-byte cache-line granularity",
	.topic = "frontend",
},
{
	.name = "icache_64b.iftag_stall",
	.event = "umask=0x4,period=200003,event=0x83",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction cache tag miss",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
	.long_desc = "Cycles with less than 3 uops delivered by the front-end",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,cmask=2,period=2000003,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
	.long_desc = "Cycles with less than 2 uops delivered by the front-end",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,cmask=3,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts, on the per-thread basis, cycles when less than 1 uop is delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core >= 3",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,cmask=4,period=2000003,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts, on the per-thread basis, cycles when no uops are delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core =4",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "Counts the number of uops not delivered to Resource Allocation Table (RAT) per thread adding \u201c4 \u2013 x\u201d when Resource Allocation Table (RAT) is not stalled and Instruction Decode Queue (IDQ) delivers x uops to Resource Allocation Table (RAT) (where x belongs to {0,1,2,3}). Counting does not cover cases when: a. IDQ-Resource Allocation Table (RAT) pipe serves the other thread. b. Resource Allocation Table (RAT) is stalled for the thread (including uop drops and clear BE conditions).  c. Instruction Decode Queue (IDQ) delivers four uops",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "Counts Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles. These cycles do not include uops routed through because of the switch itself, for example, when Instruction Decode Queue (IDQ) pre-allocation is unavailable, or Instruction Decode Queue (IDQ) is full. SBD-to-MITE switch true penalty cycles happen after the merge mux (MM) receives Decode Stream Buffer (DSB) Sync-indication until receiving the first MITE uop. MM is placed before Instruction Decode Queue (IDQ) to merge uops being fed from the MITE and Decode Stream Buffer (DSB) paths. Decode Stream Buffer (DSB) inserts the Sync-indication whenever a Decode Stream Buffer (DSB)-to-MITE switch occurs.Penalty: A Decode Stream Buffer (DSB) hit followed by a Decode Stream Buffer (DSB) miss can cost up to six cycles in which no uops are delivered to the IDQ. Most often, such switches from the Decode Stream Buffer (DSB) to the legacy pipeline cost 0\u20132 cycles",
},
{
	.name = "frontend_retired.latency_ge_4",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x400406",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 4 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_2",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x200206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 2 bubble-slots for a period of 2 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_2",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x400206",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 2 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.stlb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x15",
	.desc = "Retired Instructions who experienced STLB (2nd level TLB) true miss. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced STLB (2nd level TLB) true miss (Precise event)",
},
{
	.name = "frontend_retired.itlb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x14",
	.desc = "Retired Instructions who experienced iTLB true miss. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced iTLB (Instruction TLB) true miss (Precise event)",
},
{
	.name = "frontend_retired.l2_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x13",
	.desc = "Retired Instructions who experienced Instruction L2 Cache true miss. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.l1i_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x12",
	.desc = "Retired Instructions who experienced Instruction L1 Cache true miss. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.dsb_miss",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x11",
	.desc = "Retired Instructions who experienced decode stream buffer (DSB - the decoded instruction-cache) miss. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired Instructions that experienced DSB (Decode stream buffer i.e. the decoded instruction-cache) miss (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_3",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x300206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 3 bubble-slots for a period of 2 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_2_bubbles_ge_1",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x100206",
	.desc = "Retired instructions that are fetched after an interval where the front-end had at least 1 bubble-slot for a period of 2 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after the front-end had at least 1 bubble-slot for a period of 2 cycles. A bubble-slot is an empty issue-pipeline slot while there was no RAT stall (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_512",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x420006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 512 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_256",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x410006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 256 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_128",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x408006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 128 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_64",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x404006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 64 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
},
{
	.name = "frontend_retired.latency_ge_32",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x402006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 32 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end  after a front-end stall of at least 32 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_16",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x401006",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 16 cycles which was not interrupted by a back-end stall. Precise Event (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 16 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "frontend_retired.latency_ge_8",
	.event = "umask=0x1,period=100007,event=0xc6,frontend=0x400806",
	.desc = "Retired instructions that are fetched after an interval where the front-end delivered no uops for a period of 8 cycles which was not interrupted by a back-end stall (Precise event)",
	.topic = "frontend",
	.long_desc = "Counts retired instructions that are delivered to the back-end after a front-end stall of at least 8 cycles. During this period the front-end delivered no uops (Precise event)",
},
{
	.name = "tx_mem.abort_conflict",
	.event = "umask=0x1,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data conflict on a transactionally accessed address",
	.topic = "memory",
	.long_desc = "Number of times a TSX line had a cache conflict",
},
{
	.name = "tx_mem.abort_capacity",
	.event = "umask=0x2,period=2000003,event=0x54",
	.desc = "Number of times a transactional abort was signaled due to a data capacity limitation for transactional reads or writes",
	.topic = "memory",
},
{
	.name = "tx_mem.abort_hle_store_to_elided_lock",
	.event = "umask=0x4,period=2000003,event=0x54",
	.desc = "Number of times a HLE transactional region aborted due to a non XRELEASE prefixed instruction writing to an elided lock in the elision buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to a non-release/commit store to lock",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_not_empty",
	.event = "umask=0x8,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to NoAllocatedElisionBuffer being non-zero",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_mismatch",
	.event = "umask=0x10,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to XRELEASE lock not satisfying the address and value requirements in the elision buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_unsupported_alignment",
	.event = "umask=0x20,period=2000003,event=0x54",
	.desc = "Number of times an HLE transactional execution aborted due to an unsupported read alignment from the elision buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
},
{
	.name = "tx_mem.hle_elision_buffer_full",
	.event = "umask=0x40,period=2000003,event=0x54",
	.desc = "Number of times HLE lock could not be elided due to ElisionBufferAvailable being zero",
	.topic = "memory",
	.long_desc = "Number of times we could not allocate Lock Buffer",
},
{
	.name = "tx_exec.misc1",
	.event = "umask=0x1,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions that may cause a transactional abort was executed. Since this is the count of execution, it may not always cause a transactional abort",
	.topic = "memory",
},
{
	.name = "tx_exec.misc2",
	.event = "umask=0x2,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions (e.g., vzeroupper) that may cause a transactional abort was executed inside a transactional region",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by a vzeroupper instruction",
},
{
	.name = "tx_exec.misc3",
	.event = "umask=0x4,period=2000003,event=0x5d",
	.desc = "Counts the number of times an instruction execution caused the transactional nest count supported to be exceeded",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by a nest count that is too deep",
},
{
	.name = "tx_exec.misc4",
	.event = "umask=0x8,period=2000003,event=0x5d",
	.desc = "Counts the number of times a XBEGIN instruction was executed inside an HLE transactional region",
	.topic = "memory",
	.long_desc = "RTM region detected inside HLE",
},
{
	.name = "tx_exec.misc5",
	.event = "umask=0x10,period=2000003,event=0x5d",
	.desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
	.topic = "memory",
	.long_desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
},
{
	.name = "offcore_requests_outstanding.l3_miss_demand_data_rd",
	.event = "umask=0x10,period=2000003,event=0x60",
	.desc = "Counts number of Offcore outstanding Demand Data Read requests that miss L3 cache in the superQ every cycle",
	.topic = "memory",
},
{
	.name = "offcore_requests_outstanding.l3_miss_demand_data_rd_ge_6",
	.event = "umask=0x10,cmask=6,period=2000003,event=0x60",
	.desc = "Cycles with at least 6 Demand Data Read requests that miss L3 cache in the superQ",
	.topic = "memory",
},
{
	.name = "offcore_requests_outstanding.cycles_with_l3_miss_demand_data_rd",
	.event = "umask=0x10,cmask=1,period=2000003,event=0x60",
	.desc = "Cycles with at least 1 Demand Data Read requests who miss L3 cache in the superQ",
	.topic = "memory",
},
{
	.name = "cycle_activity.cycles_l3_miss",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xa3",
	.desc = "Cycles while L3 cache miss demand load is outstanding",
	.topic = "memory",
},
{
	.name = "cycle_activity.stalls_l3_miss",
	.event = "umask=0x6,cmask=6,period=2000003,event=0xa3",
	.desc = "Execution stalls while L3 cache miss demand load is outstanding",
	.topic = "memory",
},
{
	.name = "offcore_requests.l3_miss_demand_data_rd",
	.event = "umask=0x10,period=100003,event=0xb0",
	.desc = "Demand Data Read requests who miss L3 cache",
	.topic = "memory",
	.long_desc = "Demand Data Read requests who miss L3 cache",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts  Spec update: SKL089",
	.topic = "memory",
	.long_desc = "Counts the number of memory ordering Machine Clears detected. Memory Ordering Machine Clears can result from one of the following:a. memory disambiguation,b. external snoop, orc. cross SMT-HW-thread snoop (stores) hitting load buffer  Spec update: SKL089",
},
{
	.name = "hle_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution started",
	.topic = "memory",
	.long_desc = "Number of times we entered an HLE region. Does not count nested transactions",
},
{
	.name = "hle_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution successfully committed",
	.topic = "memory",
	.long_desc = "Number of times HLE commit succeeded",
},
{
	.name = "hle_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times HLE abort was triggered. (PEBS) (Precise event)",
},
{
	.name = "hle_retired.aborted_mem",
	.event = "umask=0x8,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_timer",
	.event = "umask=0x10,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to hardware timer expiration",
	.topic = "memory",
},
{
	.name = "hle_retired.aborted_unfriendly",
	.event = "umask=0x20,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions and certain unfriendly events (such as AD assists etc.)",
	.topic = "memory",
	.long_desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions and certain unfriendly events (such as AD assists etc.)",
},
{
	.name = "hle_retired.aborted_memtype",
	.event = "umask=0x40,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times an HLE execution aborted due to incompatible memory type",
},
{
	.name = "hle_retired.aborted_events",
	.event = "umask=0x80,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to unfriendly events (such as interrupts)",
	.topic = "memory",
},
{
	.name = "rtm_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution started",
	.topic = "memory",
	.long_desc = "Number of times we entered an RTM region. Does not count nested transactions",
},
{
	.name = "rtm_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution successfully committed",
	.topic = "memory",
	.long_desc = "Number of times RTM commit succeeded",
},
{
	.name = "rtm_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to any reasons (multiple categories may count as one) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times RTM abort was triggered. (PEBS) (Precise event)",
},
{
	.name = "rtm_retired.aborted_mem",
	.event = "umask=0x8,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
},
{
	.name = "rtm_retired.aborted_timer",
	.event = "umask=0x10,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to uncommon conditions",
	.topic = "memory",
},
{
	.name = "rtm_retired.aborted_unfriendly",
	.event = "umask=0x20,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
},
{
	.name = "rtm_retired.aborted_memtype",
	.event = "umask=0x40,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to incompatible memory type",
},
{
	.name = "rtm_retired.aborted_events",
	.event = "umask=0x80,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
	.topic = "memory",
	.long_desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 512 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 512 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 256 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 256 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 128 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 128 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 64 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 64 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 32 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 32 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 16 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 16 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 8 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 8 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 4 cycles (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads when the latency from first dispatch to completion is greater than 4 cycles.  Reported latency may be longer than just the memory latency (Must be precise)",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000001",
	.desc = "Counts demand data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Counts demand data reads TBD TBD",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Counts demand data reads TBD",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Counts demand data reads TBD",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Counts demand data reads TBD",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Counts demand data reads TBD",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000001",
	.desc = "Counts demand data reads TBD",
	.topic = "memory",
	.long_desc = "Counts demand data reads TBD",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000002",
	.desc = "Counts all demand data writes (RFOs) TBD TBD",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) TBD TBD",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) TBD",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) TBD",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) TBD",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) TBD",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000002",
	.desc = "Counts all demand data writes (RFOs) TBD",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs) TBD",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD TBD",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000004",
	.desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
	.topic = "memory",
	.long_desc = "Counts demand instruction fetches and L1 instruction cache prefetches that TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000010",
	.desc = "Counts prefetch (that bring data to L2) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
	.topic = "memory",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
},
{
	.name = "offcore_response.pf_l1d_and_sw.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000400",
	.desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
	.topic = "memory",
	.long_desc = "Counts L1 data cache hardware prefetch requests and software prefetch requests TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000490",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000490",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000120",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000120",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000491",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_data_rd.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000491",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_rfo.l3_miss.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3FBC000122",
	.desc = "TBD TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD TBD",
},
{
	.name = "offcore_response.all_rfo.l3_miss.remote_hit_forward",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x083FC00122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_rfo.l3_miss.remote_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x103FC00122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_rfo.l3_miss.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063FC00122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_rfo.l3_miss_remote_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x063B800122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "offcore_response.all_rfo.l3_miss_local_dram.snoop_miss_or_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0604000122",
	.desc = "TBD TBD",
	.topic = "memory",
	.long_desc = "TBD TBD",
},
{
	.name = "core_power.lvl0_turbo_license",
	.event = "umask=0x7,period=200003,event=0x28",
	.desc = "Core cycles where the core was running in a manner where Turbo may be clipped to the Non-AVX turbo schedule",
	.topic = "other",
	.long_desc = "Core cycles where the core was running with power-delivery for baseline license level 0.  This includes non-AVX codes, SSE, AVX 128-bit, and low-current AVX 256-bit codes",
},
{
	.name = "core_power.lvl1_turbo_license",
	.event = "umask=0x18,period=200003,event=0x28",
	.desc = "Core cycles where the core was running in a manner where Turbo may be clipped to the AVX2 turbo schedule",
	.topic = "other",
	.long_desc = "Core cycles where the core was running with power-delivery for license level 1.  This includes high current AVX 256-bit instructions as well as low current AVX 512-bit instructions",
},
{
	.name = "core_power.lvl2_turbo_license",
	.event = "umask=0x20,period=200003,event=0x28",
	.desc = "Core cycles where the core was running in a manner where Turbo may be clipped to the AVX512 turbo schedule",
	.topic = "other",
	.long_desc = "Core cycles where the core was running with power-delivery for license level 2 (introduced in Skylake Server michroarchtecture).  This includes high current AVX 512-bit instructions",
},
{
	.name = "core_power.throttle",
	.event = "umask=0x40,period=200003,event=0x28",
	.desc = "Core cycles the core was throttled due to a pending power level request",
	.topic = "other",
	.long_desc = "Core cycles the out-of-order engine was throttled due to a pending power level request",
},
{
	.name = "sw_prefetch_access.nta",
	.event = "umask=0x1,period=2000003,event=0x32",
	.desc = "Number of PREFETCHNTA instructions executed",
	.topic = "other",
},
{
	.name = "sw_prefetch_access.t0",
	.event = "umask=0x2,period=2000003,event=0x32",
	.desc = "Number of PREFETCHT0 instructions executed",
	.topic = "other",
},
{
	.name = "sw_prefetch_access.t1_t2",
	.event = "umask=0x4,period=2000003,event=0x32",
	.desc = "Number of PREFETCHT1 or PREFETCHT2 instructions executed",
	.topic = "other",
},
{
	.name = "sw_prefetch_access.prefetchw",
	.event = "umask=0x8,period=2000003,event=0x32",
	.desc = "Number of PREFETCHW instructions executed",
	.topic = "other",
},
{
	.name = "hw_interrupts.received",
	.event = "umask=0x1,period=203,event=0xcb",
	.desc = "Number of hardware interrupts received by the processor",
	.topic = "other",
	.long_desc = "Counts the number of hardware interruptions received by the processor",
},
{
	.name = "core_snoop_response.rsp_ihiti",
	.event = "umask=0x1,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_ihitfse",
	.event = "umask=0x2,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_shitfse",
	.event = "umask=0x4,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_sfwdm",
	.event = "umask=0x8,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_ifwdm",
	.event = "umask=0x10,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_ifwdfe",
	.event = "umask=0x20,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "core_snoop_response.rsp_sfwdfe",
	.event = "umask=0x40,period=2000003,event=0xef",
	.desc = "(null)",
	.topic = "other",
},
{
	.name = "idi_misc.wb_upgrade",
	.event = "umask=0x2,period=100003,event=0xfe",
	.desc = "Counts number of cache lines that are allocated and written back to L3 with the intention that they are more likely to be reused shortly",
	.topic = "other",
	.long_desc = "Counts number of cache lines that are allocated and written back to L3 with the intention that they are more likely to be reused shortly",
},
{
	.name = "idi_misc.wb_downgrade",
	.event = "umask=0x4,period=100003,event=0xfe",
	.desc = "Counts number of cache lines that are dropped and not written back to L3 as they are deemed to be less likely to be reused shortly",
	.topic = "other",
	.long_desc = "Counts number of cache lines that are dropped and not written back to L3 as they are deemed to be less likely to be reused shortly",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, Counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers. Notes: INST_RETIRED.ANY is counted by a designated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. INST_RETIRED.ANY_P is counted by a programmable counter and it is an architectural performance event. Counting: Faulting executions of GETSEC/VM entry/VM Exit/MWait will not count as retired instructions",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "Counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state. This event has a constant ratio with the CPU_CLK_UNHALTED.REF_XCLK event. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. Note: On all current platforms this event stops counting during 'throttling (TM)' states duty off periods the processor is 'halted'.  The counter update is done at a lower clock rate then the core clock the overflow status bit for this counter may appear 'sticky'.  After the counter has overflowed and software clears the overflow status bit and resets the counter to less than MAX. The reset value to the counter is not clocked immediately so the overflow status bit will flip 'high (1)' and generate another PMI (if enabled) after which the reset value gets clocked into the counter. Therefore, software will get the interrupt, read the overflow status bit '1 for bit 34 while the counter value is less than MAX. Software should ignore this case",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Loads blocked by overlapping with store buffer that cannot be forwarded ",
	.topic = "pipeline",
	.long_desc = "Counts how many times the load operation got the true Block-on-Store blocking code preventing store forwarding. This includes cases when:a. preceding store conflicts with the load (incomplete overlap),b. store forwarding is impossible due to u-arch limitations,c. preceding lock RMW operations are not forwarded,d. store has the no-forward bit set (uncacheable/page-split/masked stores),e. all-blocking stores are used (mostly, fences and port I/O), and others.The most common case is a load blocked due to its address range overlapping with a preceding smaller uncompleted store. Note: This event does not take into account cases of out-of-SW-control (for example, SbTailHit), unknown physical STA, and cases of blocking loads on store due to being non-WB memory type or a lock. These cases are covered by other events. See the table of not supported store forwards in the Optimization Guide",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
	.long_desc = "The number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare on address",
	.topic = "pipeline",
	.long_desc = "Counts false dependencies in MOB when the partial comparison upon loose net check and dependency was resolved by the Enhanced Loose net mechanism. This may not result in high performance penalties. Loose net checks can fail when loads and stores are 4k aliased",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for this thread (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "Core cycles the Resource allocator was stalled due to recovery from an earlier branch misprediction or machine clear event",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "int_misc.clear_resteer_cycles",
	.event = "umask=0x80,period=2000003,event=0xd",
	.desc = "Cycles the issue-stage is waiting for front-end to fetch from resteered path following branch misprediction or machine clear events",
	.topic = "pipeline",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which the Resource Allocation Table (RAT) does not issue any Uops to the reservation station (RS) for the current thread",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "Counts the number of uops that the Resource Allocation Table (RAT) issues to the Reservation Station (RS)",
},
{
	.name = "uops_issued.vector_width_mismatch",
	.event = "umask=0x2,period=2000003,event=0xe",
	.desc = "Uops inserted at issue-stage in order to preserve upper bits of vector registers",
	.topic = "pipeline",
	.long_desc = "Counts the number of Blend Uops issued by the Resource Allocation Table (RAT) to the reservation station (RS) in order to preserve upper bits of vector registers. Starting with the Skylake microarchitecture, these Blend uops are needed since every Intel SSE instruction executed in Dirty Upper State needs to preserve bits 128-255 of the destination register. For more information, refer to \u201cMixing Intel AVX and Intel SSE Code\u201d section of the Optimization Guide",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
},
{
	.name = "arith.divider_active",
	.event = "umask=0x1,cmask=1,period=2000003,event=0x14",
	.desc = "Cycles when divide unit is busy executing divide or square root operations. Accounts for integer and floating-point operations",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This is an architectural event that counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling. For this reason, this event may have a changing ratio with regards to wall clock time",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ring0_trans",
	.event = "edge=1,umask=0x0,cmask=1,period=100007,event=0x3c",
	.desc = "Counts when there is a transition from ring 1, 2 or 3 to ring 0",
	.topic = "pipeline",
	.long_desc = "Counts when the Current Privilege Level (CPL) transitions from ring 1, 2 or 3 to ring 0 (Kernel)",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when the thread is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when at least one thread on the physical core is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when at least one thread on the physical core is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when the thread is unhalted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Core crystal clock cycles when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2503,event=0x3c",
	.desc = "Core crystal clock cycles when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Demand load dispatches that hit L1D fill buffer (FB) allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "Counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the software prefetch. It can also be incremented by some lock instructions. So it should only be used with profiling so that the locks can be excluded by ASM (Assembly File) inspection of the nearby instructions",
},
{
	.name = "partial_rat_stalls.scoreboard",
	.event = "umask=0x1,period=2000003,event=0x59",
	.desc = "Cycles where the pipeline is stalled due to serializing operations",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which the microcode scoreboard stalls happen",
},
{
	.name = "rs_events.empty_end",
	.event = "edge=1,inv=1,umask=0x1,cmask=1,period=2000003,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
	.long_desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate front-end Latency Bound issues",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which the reservation station (RS) is empty for the thread.; Note: In ST-mode, not active thread should drive 0. This is usually caused by severely costly branch mispredictions, or allocator/FE issues",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
	.long_desc = "Counts cycles that the Instruction Length decoder (ILD) stalls occurred due to dynamically changing prefix length of the decoded instruction (by operand size prefix instruction 0x66, address size prefix instruction 0x67 or REX.W for Intel64). Count is proportional to the number of prefixes in a 16B-line. This may result in a three-cycle penalty for each LCP (Length changing prefix) in a 16-byte chunk",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 2",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 3",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 4",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_dispatched_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_dispatched_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "Counts, on the per-thread basis, cycles during which at least one uop is dispatched from the Reservation Station (RS) to port 7",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
	.long_desc = "Counts resource-related stall cycles",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "Counts allocation stall cycles caused by the store buffer (SB) being full. This counts cycles that the pipeline back-end blocked uop delivery from the front-end",
},
{
	.name = "cycle_activity.cycles_l2_miss",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_total",
	.event = "umask=0x4,cmask=4,period=2000003,event=0xa3",
	.desc = "Total execution stalls",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_miss",
	.event = "umask=0x5,cmask=5,period=2000003,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_miss",
	.event = "umask=0x8,cmask=8,period=2000003,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_miss",
	.event = "umask=0xc,cmask=12,period=2000003,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_mem_any",
	.event = "umask=0x10,cmask=16,period=2000003,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_mem_any",
	.event = "umask=0x14,cmask=20,period=2000003,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "exe_activity.exe_bound_0_ports",
	.event = "umask=0x1,period=2000003,event=0xa6",
	.desc = "Cycles where no uops were executed, the Reservation Station was not empty, the Store Buffer was full and there was no outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which no uops were executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.1_ports_util",
	.event = "umask=0x2,period=2000003,event=0xa6",
	.desc = "Cycles total of 1 uop is executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which a total of 1 uop was executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.2_ports_util",
	.event = "umask=0x4,period=2000003,event=0xa6",
	.desc = "Cycles total of 2 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which a total of 2 uops were executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.3_ports_util",
	.event = "umask=0x8,period=2000003,event=0xa6",
	.desc = "Cycles total of 3 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Cycles total of 3 uops are executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.4_ports_util",
	.event = "umask=0x10,period=2000003,event=0xa6",
	.desc = "Cycles total of 4 uops are executed on all ports and Reservation Station was not empty",
	.topic = "pipeline",
	.long_desc = "Cycles total of 4 uops are executed on all ports and Reservation Station (RS) was not empty",
},
{
	.name = "exe_activity.bound_on_stores",
	.event = "umask=0x40,period=2000003,event=0xa6",
	.desc = "Cycles where the Store Buffer was full and no outstanding load",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
	.long_desc = "Number of uops delivered to the back-end by the LSD(Loop Stream Detector)",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Counts the cycles when 4 uops are delivered by the LSD (Loop-stream detector)",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
	.long_desc = "Counts the cycles when at least one uop is delivered by the LSD (Loop-stream detector)",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 4 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 3 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 2 uops were executed per-thread",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread",
	.topic = "pipeline",
	.long_desc = "Cycles where at least 1 uop was executed per-thread",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,cmask=1,period=2000003,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread",
	.topic = "pipeline",
	.long_desc = "Counts cycles during which no uops were dispatched from the Reservation Station (RS) per thread",
},
{
	.name = "uops_executed.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Counts the number of uops to be executed per-thread each cycle",
	.topic = "pipeline",
	.long_desc = "Number of uops to be executed per-thread each cycle",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core",
	.topic = "pipeline",
	.long_desc = "Number of uops executed from any thread",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,cmask=4,period=2000003,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,cmask=3,period=2000003,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,cmask=2,period=2000003,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,cmask=1,period=2000003,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.x87",
	.event = "umask=0x10,period=2000003,event=0xb1",
	.desc = "Counts the number of x87 uops dispatched",
	.topic = "pipeline",
	.long_desc = "Counts the number of x87 uops executed",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter - architectural event  Spec update: SKL091, SKL044",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions (EOMs) retired. Counting covers macro-fused instructions individually (that is, increments by two)  Spec update: SKL091, SKL044",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: SKL091, SKL044 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "A version of INST_RETIRED that allows for a more unbiased distribution of samples across instructions retired. It utilizes the Precise Distribution of Instructions Retired (PDIR) feature to mitigate some bias in how retired instructions get sampled  Spec update: SKL091, SKL044 (Must be precise)",
},
{
	.name = "inst_retired.total_cycles_ps",
	.event = "inv=1,umask=0x1,cmask=10,period=2000003,event=0xc0",
	.desc = "Number of cycles using always true condition applied to  PEBS instructions retired event  Spec update: SKL091, SKL044 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "Number of cycles using an always true condition applied to  PEBS instructions retired event. (inst_ret< 16)  Spec update: SKL091, SKL044 (Must be precise)",
},
{
	.name = "other_assists.any",
	.event = "umask=0x3f,period=100003,event=0xc1",
	.desc = "Number of times a microcode assist is invoked by HW other than FP-assist. Examples include AD (page Access Dirty) and AVX* related assists",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x2,cmask=10,period=2000003,event=0xc2",
	.desc = "Cycles with less than 10 actually retired uops",
	.topic = "pipeline",
	.long_desc = "Number of cycles using always true condition (uops_ret < 16) applied to non PEBS uops retired event",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x2,cmask=1,period=2000003,event=0xc2",
	.desc = "Cycles without actually retired uops",
	.topic = "pipeline",
	.long_desc = "This event counts cycles without actually retired uops",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used",
	.topic = "pipeline",
	.long_desc = "Counts the retirement slots used",
},
{
	.name = "machine_clears.count",
	.event = "edge=1,umask=0x1,cmask=1,period=100003,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
	.long_desc = "Number of machine clears (nukes) of any type",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "Counts self-modifying code (SMC) detected, which causes a machine clear",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired  Spec update: SKL091",
	.topic = "pipeline",
	.long_desc = "Counts all (macro) branch instructions retired  Spec update: SKL091",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts conditional branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts both direct and indirect near call instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired  Spec update: SKL091 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_INST_RETIRED.ALL_BRANCHES that counts all (macro) branch instructions retired  Spec update: SKL091 (Must be precise)",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts return instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Counts all not taken macro branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts not taken branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts taken branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Counts the number of far branch instructions retired  Spec update: SKL091 (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts far branch instructions retired  Spec update: SKL091 (Precise event)",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "Counts all the retired branch instructions that were mispredicted by the processor. A branch misprediction occurs when the processor incorrectly predicts the destination of the branch.  When the misprediction is discovered at execution, all the instructions executed in the wrong (speculative) path must be discarded, and the processor must start fetching from the correct path",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts mispredicted conditional branch instructions retired (Precise event)",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=400009,event=0xc5",
	.desc = "Mispredicted direct and indirect near call instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken retired mispredicted direct and indirect near calls, including both register and memory indirect (Precise event)",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_MISP_RETIRED.ALL_BRANCHES that counts all mispredicted macro branch instructions retired (Must be precise)",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "Number of near branch instructions retired that were mispredicted and taken (Precise event)",
	.topic = "pipeline",
	.long_desc = "Number of near branch instructions retired that were mispredicted and taken (Precise event)",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Increments whenever there is an update to the LBR array",
	.topic = "pipeline",
	.long_desc = "Increments when an entry is added to the Last Branch Record (LBR) array (or removed from the array in case of RETURNs in call stack mode). The event requires LBR enable via IA32_DEBUGCTL MSR and branch type selection via MSR_LBR_SELECT",
},
{
	.name = "rob_misc_events.pause_inst",
	.event = "umask=0x40,period=2000003,event=0xcc",
	.desc = "Number of retired PAUSE instructions (that do not end up with a VMExit to the VMM; TSX aborted Instructions may be counted). This event is not supported on first SKL and KBL products",
	.topic = "pipeline",
},
{
	.name = "baclears.any",
	.event = "umask=0x1,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
	.long_desc = "Counts the number of times the front-end is resteered when it finds a branch instruction in a fetch line. This occurs for the first time a branch instruction is fetched or when the branch is not tracked by the BPU (Branch Prediction Unit) anymore",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "skx metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "skx metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "skx metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "skx metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "skx metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "skx metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "skx metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "skx metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "skx metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Instruction per taken branch",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_taken",
	.metric_name = "IpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Branch instructions per taken branch",
	.topic = "skx metrics",
	.metric_expr = "br_inst_retired.all_branches / br_inst_retired.near_taken",
	.metric_name = "BpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "skx metrics",
	.metric_expr = "min( 1 , uops_issued.any / ( (uops_retired.retire_slots / inst_retired.any) * 64 * ( icache_64b.iftag_hit + icache_64b.iftag_miss ) / 4.1 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "skx metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + idq.mite_uops + idq.ms_uops ))",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "skx metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "skx metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "skx metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "skx metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Instructions per Load (lower number means loads are more frequent)",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any / mem_inst_retired.all_loads",
	.metric_name = "IpL",
	.metric_group = "Instruction_Type;L1_Bound",
},
{
	.desc = "Instructions per Store",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any / mem_inst_retired.all_stores",
	.metric_name = "IpS",
	.metric_group = "Instruction_Type;Store_Bound",
},
{
	.desc = "Instructions per Branch",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.all_branches",
	.metric_name = "IpB",
	.metric_group = "Branches;Instruction_Type;Port_5;Port_6",
},
{
	.desc = "Instruction per (near) call",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_call",
	.metric_name = "IpCall",
	.metric_group = "Branches",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "skx metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * ( fp_arith_inst_retired.256b_packed_single + fp_arith_inst_retired.512b_packed_double ) + 16 * fp_arith_inst_retired.512b_packed_single )) / cycles",
	.metric_name = "FLOPc",
	.metric_group = "FLOPS",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "skx metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * ( fp_arith_inst_retired.256b_packed_single + fp_arith_inst_retired.512b_packed_double ) + 16 * fp_arith_inst_retired.512b_packed_single )) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "FLOPc_SMT",
	.metric_group = "FLOPS_SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "skx metrics",
	.metric_expr = "uops_executed.thread / (( uops_executed.core_cycles_ge_1 / 2 ) if #smt_on else uops_executed.core_cycles_ge_1)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "skx metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) * (( int_misc.clear_resteer_cycles + 9 * baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) ) * (4 * cycles) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "skx metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) * (( int_misc.clear_resteer_cycles + 9 * baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) ) * (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost_SMT",
	.metric_group = "Branch_Mispredicts_SMT",
},
{
	.desc = "Number of Instructions per non-speculative Branch Misprediction (JEClear)",
	.topic = "skx metrics",
	.metric_expr = "inst_retired.any / br_misp_retired.all_branches",
	.metric_name = "IpMispredict",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "skx metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads (in core cycles)",
	.topic = "skx metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_retired.l1_miss + mem_load_retired.fb_hit )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least one such miss. Per-thread)",
	.topic = "skx metrics",
	.metric_expr = "l1d_pend_miss.pending / l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "skx metrics",
	.metric_expr = "( itlb_misses.walk_pending + dtlb_load_misses.walk_pending + dtlb_store_misses.walk_pending + ept.walk_pending ) / ( 2 * cycles )",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "skx metrics",
	.metric_expr = "( itlb_misses.walk_pending + dtlb_load_misses.walk_pending + dtlb_store_misses.walk_pending + ept.walk_pending ) / ( 2 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )) )",
	.metric_name = "Page_Walks_Utilization_SMT",
	.metric_group = "TLB_SMT",
},
{
	.desc = "Average data fill bandwidth to the L1 data cache [GB / sec]",
	.topic = "skx metrics",
	.metric_expr = "64 * l1d.replacement / 1000000000 / duration_time",
	.metric_name = "L1D_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average data fill bandwidth to the L2 cache [GB / sec]",
	.topic = "skx metrics",
	.metric_expr = "64 * l2_lines_in.all / 1000000000 / duration_time",
	.metric_name = "L2_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "skx metrics",
	.metric_expr = "64 * longest_lat_cache.miss / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "skx metrics",
	.metric_expr = "64 * offcore_requests.all_requests / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Access_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "L1 cache true misses per kilo instruction for retired demand loads",
	.topic = "skx metrics",
	.metric_expr = "1000 * mem_load_retired.l1_miss / inst_retired.any",
	.metric_name = "L1MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache true misses per kilo instruction for retired demand loads",
	.topic = "skx metrics",
	.metric_expr = "1000 * mem_load_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache misses per kilo instruction for all request types (including speculative)",
	.topic = "skx metrics",
	.metric_expr = "1000 * l2_rqsts.miss / inst_retired.any",
	.metric_name = "L2MPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache hits per kilo instruction for all request types (including speculative)",
	.topic = "skx metrics",
	.metric_expr = "1000 * ( l2_rqsts.references - l2_rqsts.miss ) / inst_retired.any",
	.metric_name = "L2HPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L3 cache true misses per kilo instruction for retired demand loads",
	.topic = "skx metrics",
	.metric_expr = "1000 * mem_load_retired.l3_miss / inst_retired.any",
	.metric_name = "L3MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "Average CPU Utilization",
	.topic = "skx metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "skx metrics",
	.metric_expr = "( (( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * ( fp_arith_inst_retired.256b_packed_single + fp_arith_inst_retired.512b_packed_double ) + 16 * fp_arith_inst_retired.512b_packed_single )) / 1000000000 ) / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "skx metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "skx metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "skx metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "skx metrics",
	.metric_expr = "( 64 * ( uncore_imc@cas_count_read@ + uncore_imc@cas_count_write@ ) / 1000000000 ) / duration_time",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average latency of data read request to external memory (in nanoseconds). Accounts for demand loads and L1/L2 prefetches",
	.topic = "skx metrics",
	.metric_expr = "1000000000 * ( cha@event\\=0x36\\\\\\,umask\\=0x21\\\\\\,config\\=0x40433@ / cha@event\\=0x35\\\\\\,umask\\=0x21\\\\\\,config\\=0x40433@ ) / ( cha_0@event\\=0x0@ / duration_time )",
	.metric_name = "DRAM_Read_Latency",
	.metric_group = "Memory_Lat",
},
{
	.desc = "Average number of parallel data read requests to external memory. Accounts for demand loads and L1/L2 prefetches",
	.topic = "skx metrics",
	.metric_expr = "cha@event\\=0x36\\\\\\,umask\\=0x21\\\\\\,config\\=0x40433@ / cha@event\\=0x36\\\\\\,umask\\=0x21\\\\\\,thresh\\=1\\\\\\,config\\=0x40433@",
	.metric_name = "DRAM_Parallel_Reads",
	.metric_group = "Memory_BW",
},
{
	.desc = "Socket actual clocks when any core is active on that socket",
	.topic = "skx metrics",
	.metric_expr = "cha_0@event\\=0x0@",
	.metric_name = "Socket_CLKS",
	.metric_group = "",
},
{
	.desc = "C3 residency percent per core",
	.topic = "skx metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "skx metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "skx metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "skx metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "skx metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "skx metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "skx metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "llc_misses.mem_read",
	.event = "umask=0x3,event=0x4",
	.desc = "read requests to memory controller. Derived from unc_m_cas_count.rd. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_write",
	.event = "umask=0xC,event=0x4",
	.desc = "write requests to memory controller. Derived from unc_m_cas_count.wr. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_clockticks",
	.event = "event=0",
	.desc = "Memory controller clock ticks. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_power_channel_ppd",
	.event = "event=0x85",
	.desc = "Cycles where DRAM ranks are in power down (CKE) mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_channel_ppd / unc_m_clockticks) * 100.",
	.metric_name = "power_channel_ppd %",
},
{
	.name = "unc_m_power_self_refresh",
	.event = "event=0x43",
	.desc = "Cycles Memory is in self refresh power mode. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
	.metric_expr = "(unc_m_power_self_refresh / unc_m_clockticks) * 100.",
	.metric_name = "power_self_refresh %",
},
{
	.name = "unc_m_pre_count.page_miss",
	.event = "umask=0x1,event=0x2",
	.desc = "Pre-charges due to page misses. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.rd",
	.event = "umask=0x4,event=0x2",
	.desc = "Pre-charge for reads. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.wr",
	.event = "umask=0x8,event=0x2",
	.desc = "Pre-charge for writes. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_act_count.wr",
	.event = "umask=0x2,event=0x1",
	.desc = "DRAM Page Activate commands sent due to a write request. Unit: uncore_imc ",
	.topic = "uncore memory",
	.long_desc = "Counts DRAM Page Activate commands sent on this channel due to a write request to the iMC (Memory Controller).  Activate commands are issued to open up a page on the DRAM devices so that it can be read or written to with a CAS (Column Access Select) command",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_cas_count.all",
	.event = "umask=0xF,event=0x4",
	.desc = "All DRAM CAS Commands issued. Unit: uncore_imc ",
	.topic = "uncore memory",
	.long_desc = "Counts all CAS (Column Address Select) commands issued to DRAM per memory channel.  CAS commands are issued to specify the address to read or write on DRAM, so this event increments for every read and write. This event counts whether AutoPrecharge (which closes the DRAM Page automatically after a read/write) is enabled or not",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_read",
	.event = "umask=0x3,event=0x4",
	.desc = "read requests to memory controller. Derived from unc_m_cas_count.rd. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_cas_count.rd_reg",
	.event = "umask=0x1,event=0x4",
	.desc = "All DRAM Read CAS Commands issued (does not include underfills) . Unit: uncore_imc ",
	.topic = "uncore memory",
	.long_desc = "Counts CAS (Column Access Select) regular read commands issued to DRAM on a per channel basis.  CAS commands are issued to specify the address to read or write on DRAM, and this event increments for every regular read.  This event only counts regular reads and does not includes underfill reads due to partial write requests.  This event counts whether AutoPrecharge (which closes the DRAM Page automatically after a read/write)  is enabled or not",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_cas_count.rd_underfill",
	.event = "umask=0x2,event=0x4",
	.desc = "DRAM Underfill Read CAS Commands issued. Unit: uncore_imc ",
	.topic = "uncore memory",
	.long_desc = "Counts CAS (Column Access Select) underfill read commands issued to DRAM due to a partial write, on a per channel basis.  CAS commands are issued to specify the address to read or write on DRAM, and this command counts underfill reads.  Partial writes must be completed by first reading in the underfill from DRAM and then merging in the partial write data before writing the full line back to DRAM. This event will generally count about the same as the number of partial writes, but may be slightly less because of partials hitting in the WPQ (due to a previous write request)",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_write",
	.event = "umask=0xC,event=0x4",
	.desc = "write requests to memory controller. Derived from unc_m_cas_count.wr. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_rpq_inserts",
	.event = "event=0x10",
	.desc = "Read Pending Queue Allocations. Unit: uncore_imc ",
	.topic = "uncore memory",
	.long_desc = "Counts the number of read requests allocated into the Read Pending Queue (RPQ).  This queue is used to schedule reads out to the memory controller and to track the requests.  Requests allocate into the RPQ soon after they enter the memory controller, and need credits for an entry in this buffer before being sent from the CHA to the iMC.  The requests deallocate after the read CAS command has been issued to DRAM.  This event counts both Isochronous and non-Isochronous requests which were issued to the RPQ",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_rpq_occupancy",
	.event = "event=0x80",
	.desc = "Read Pending Queue Occupancy. Unit: uncore_imc ",
	.topic = "uncore memory",
	.long_desc = "Counts the number of entries in the Read Pending Queue (RPQ) at each cycle.  This can then be used to calculate both the average occupancy of the queue (in conjunction with the number of cycles not empty) and the average latency in the queue (in conjunction with the number of allocations).  The RPQ is used to schedule reads out to the memory controller and to track the requests.  Requests allocate into the RPQ soon after they enter the memory controller, and need credits for an entry in this buffer before being sent from the CHA to the iMC. They deallocate from the RPQ after the CAS command has been issued to memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_wpq_inserts",
	.event = "event=0x20",
	.desc = "Write Pending Queue Allocations. Unit: uncore_imc ",
	.topic = "uncore memory",
	.long_desc = "Counts the number of writes requests allocated into the Write Pending Queue (WPQ).  The WPQ is used to schedule writes out to the memory controller and to track the requests.  Requests allocate into the WPQ soon after they enter the memory controller, and need credits for an entry in this buffer before being sent from the CHA to the iMC (Memory Controller).  The write requests deallocate after being issued to DRAM.  Write requests themselves are able to complete (from the perspective of the rest of the system) as soon they have 'posted' to the iMC",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_wpq_occupancy",
	.event = "event=0x81",
	.desc = "Write Pending Queue Occupancy. Unit: uncore_imc ",
	.topic = "uncore memory",
	.long_desc = "Counts the number of entries in the Write Pending Queue (WPQ) at each cycle.  This can then be used to calculate both the average queue occupancy (in conjunction with the number of cycles not empty) and the average latency (in conjunction with the number of allocations).  The WPQ is used to schedule writes out to the memory controller and to track the requests",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_cha_clockticks",
	.event = "event=0",
	.desc = "Uncore cache clock ticks. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_misses.uncacheable",
	.event = "umask=0x21,event=0x35,config1=0x40e33",
	.desc = "LLC misses - Uncacheable reads (from cpu) . Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_read",
	.event = "umask=0x21,event=0x35,config1=0x40040e33",
	.desc = "MMIO reads. Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_write",
	.event = "umask=0x21,event=0x35,config1=0x40041e33",
	.desc = "MMIO writes. Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_full",
	.event = "umask=0x21,event=0x35,config1=0x41833",
	.desc = "Streaming stores (full cache line). Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_partial",
	.event = "umask=0x21,event=0x35,config1=0x41a33",
	.desc = "Streaming stores (partial cache line). Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.reads",
	.event = "umask=0x03,event=0x50",
	.desc = "read requests from home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.reads_local",
	.event = "umask=0x01,event=0x50",
	.desc = "read requests from local home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.reads_remote",
	.event = "umask=0x02,event=0x50",
	.desc = "read requests from remote home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.writes",
	.event = "umask=0x0C,event=0x50",
	.desc = "write requests from home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.writes_local",
	.event = "umask=0x04,event=0x50",
	.desc = "write requests from local home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.writes_remote",
	.event = "umask=0x08,event=0x50",
	.desc = "write requests from remote home agent. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "upi_data_bandwidth_tx",
	.event = "umask=0x0F,event=0x2",
	.desc = "UPI interconnect send bandwidth for payload. Derived from unc_upi_txl_flits.all_data. Unit: uncore_upi ",
	.topic = "uncore other",
	.pmu = "uncore_upi",
	.unit = "7.11E-06Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_read",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x04,event=0x83,ch_mask=0x1f",
	.desc = "PCI Express bandwidth reading at IIO. Derived from unc_iio_data_req_of_cpu.mem_read.part0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_read.part0 + unc_iio_data_req_of_cpu.mem_read.part1 + unc_iio_data_req_of_cpu.mem_read.part2 + unc_iio_data_req_of_cpu.mem_read.part3",
	.metric_name = "LLC_MISSES.PCIE_READ",
},
{
	.name = "llc_misses.pcie_write",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x01,event=0x83,ch_mask=0x1f",
	.desc = "PCI Express bandwidth writing at IIO. Derived from unc_iio_data_req_of_cpu.mem_write.part0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_write.part0 +unc_iio_data_req_of_cpu.mem_write.part1 +unc_iio_data_req_of_cpu.mem_write.part2 +unc_iio_data_req_of_cpu.mem_write.part3",
	.metric_name = "LLC_MISSES.PCIE_WRITE",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_write.part0 +unc_iio_data_req_of_cpu.mem_write.part1 +unc_iio_data_req_of_cpu.mem_write.part2 +unc_iio_data_req_of_cpu.mem_write.part3",
	.metric_name = "LLC_MISSES.PCIE_WRITE",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 1. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 2. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 3. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_read.part0 + unc_iio_data_req_of_cpu.mem_read.part1 + unc_iio_data_req_of_cpu.mem_read.part2 + unc_iio_data_req_of_cpu.mem_read.part3",
	.metric_name = "LLC_MISSES.PCIE_READ",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 1. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 2. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 3. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_cha_core_snp.core_gtone",
	.event = "umask=0x42,event=0x33",
	.desc = "Core Cross Snoops Issued; Multiple Core Requests. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts the number of transactions that trigger a configurable number of cross snoops.  Cores are snooped if the transaction looks up the cache and determines that it is necessary based on the operation type and what CoreValid bits are set.  For example, if 2 CV bits are set on a data read, the cores must have the data in S state so it is not necessary to snoop them.  However, if only 1 CV bit is set the core my have modified the data.  If the transaction was an RFO, it would need to invalidate the lines.  This event can be filtered based on who triggered the initial snoop(s)",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_core_snp.evict_gtone",
	.event = "umask=0x82,event=0x33",
	.desc = "Core Cross Snoops Issued; Multiple Eviction. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts the number of transactions that trigger a configurable number of cross snoops.  Cores are snooped if the transaction looks up the cache and determines that it is necessary based on the operation type and what CoreValid bits are set.  For example, if 2 CV bits are set on a data read, the cores must have the data in S state so it is not necessary to snoop them.  However, if only 1 CV bit is set the core my have modified the data.  If the transaction was an RFO, it would need to invalidate the lines.  This event can be filtered based on who triggered the initial snoop(s)",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_dir_lookup.no_snp",
	.event = "umask=0x02,event=0x53",
	.desc = "Multi-socket cacheline Directory state lookups; Snoop Not Needed. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts transactions that looked into the multi-socket cacheline Directory state, and therefore did not send a snoop because the Directory indicated it was not needed",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_dir_lookup.snp",
	.event = "umask=0x01,event=0x53",
	.desc = "Multi-socket cacheline Directory state lookups; Snoop Needed. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts  transactions that looked into the multi-socket cacheline Directory state, and sent one or more snoops, because the Directory indicated it was needed",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_dir_update.ha",
	.event = "umask=0x01,event=0x54",
	.desc = "Multi-socket cacheline Directory state updates; Directory Updated memory write from the HA pipe. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts only multi-socket cacheline Directory state updates memory writes issued from the HA pipe. This does not include memory write requests which are for I (Invalid) or E (Exclusive) cachelines",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_dir_update.tor",
	.event = "umask=0x02,event=0x54",
	.desc = "Multi-socket cacheline Directory state updates; Directory Updated memory write from TOR pipe. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts only multi-socket cacheline Directory state updates due to memory writes issued from the TOR pipe which are the result of remote transaction hitting the SF/LLC and returning data Core2Core. This does not include memory write requests which are for I (Invalid) or E (Exclusive) cachelines",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_hitme_hit.ex_rds",
	.event = "umask=0x01,event=0x5f",
	.desc = "Read request from a remote socket which hit in the HitMe Cache to a line In the E state. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts read requests from a remote socket which hit in the HitME cache (used to cache the multi-socket Directory state) to a line in the E(Exclusive) state.  This includes the following read opcodes (RdCode, RdData, RdDataMigratory, RdCur, RdInv*, Inv*)",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_imc_reads_count.normal",
	.event = "umask=0x01,event=0x59",
	.desc = "Normal priority reads issued to the memory controller from the CHA. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts when a normal (Non-Isochronous) read is issued to any of the memory controller channels from the CHA",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_imc_writes_count.full",
	.event = "umask=0x01,event=0x5b",
	.desc = "CHA to iMC Full Line Writes Issued; Full Line Non-ISOCH. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts when a normal (Non-Isochronous) full line write is issued from the CHA to the any of the memory controller channels",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_misc.rfo_hit_s",
	.event = "umask=0x08,event=0x39",
	.desc = "Number of times that an RFO hit in S state. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts when a RFO (the Read for Ownership issued before a  write) request hit a cacheline in the S (Shared) state",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.invitoe_local",
	.event = "umask=0x10,event=0x50",
	.desc = "Local requests for exclusive ownership of a cache line  without receiving data. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts the total number of requests coming from a unit on this socket for exclusive ownership of a cache line without receiving data (INVITOE) to the CHA",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_requests.invitoe_remote",
	.event = "umask=0x20,event=0x50",
	.desc = "Local requests for exclusive ownership of a cache line without receiving data. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts the total number of requests coming from a remote socket for exclusive ownership of a cache line without receiving data (INVITOE) to the CHA",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_snoop_resp.rspcnflcts",
	.event = "umask=0x40,event=0x5c",
	.desc = "RspCnflct* Snoop Responses Received. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts when a a transaction with the opcode type RspCnflct* Snoop Response was received. This is returned when a snoop finds an existing outstanding transaction in a remote caching agent. This triggers conflict resolution hardware. This covers both the opcode RspCnflct and RspCnflctWbI",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_snoop_resp.rspi",
	.event = "umask=0x01,event=0x5c",
	.desc = "RspI Snoop Responses Received. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts when a transaction with the opcode type RspI Snoop Response was received which indicates the remote cache does not have the data, or when the remote cache silently evicts data (such as when an RFO: the Read for Ownership issued before a write hits non-modified data)",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_snoop_resp.rspifwd",
	.event = "umask=0x04,event=0x5c",
	.desc = "RspIFwd Snoop Responses Received. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts when a a transaction with the opcode type RspIFwd Snoop Response was received which indicates a remote caching agent forwarded the data and the requesting agent is able to acquire the data in E (Exclusive) or M (modified) states.  This is commonly returned with RFO (the Read for Ownership issued before a write) transactions.  The snoop could have either been to a cacheline in the M,E,F (Modified, Exclusive or Forward)  states",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_snoop_resp.rspsfwd",
	.event = "umask=0x08,event=0x5c",
	.desc = "RspSFwd Snoop Responses Received. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts when a a transaction with the opcode type RspSFwd Snoop Response was received which indicates a remote caching agent forwarded the data but held on to its current copy.  This is common for data and code reads that hit in a remote socket in E (Exclusive) or F (Forward) state",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_snoop_resp.rsp_fwd_wb",
	.event = "umask=0x20,event=0x5c",
	.desc = "Rsp*Fwd*WB Snoop Responses Received. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts when a transaction with the opcode type Rsp*Fwd*WB Snoop Response was received which indicates the data was written back to it's home socket, and the cacheline was forwarded to the requestor socket.  This snoop response is only used in >= 4 socket systems.  It is used when a snoop HITM's in a remote caching agent and it directly forwards data to a requestor, and simultaneously returns data to it's home socket to be written back to memory",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_snoop_resp.rsp_wbwb",
	.event = "umask=0x10,event=0x5c",
	.desc = "Rsp*WB Snoop Responses Received. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "Counts when a transaction with the opcode type Rsp*WB Snoop Response was received which indicates which indicates the data was written back to it's home.  This is returned when a non-RFO request hits a cacheline in the Modified state. The Cache can either downgrade the cacheline to a S (Shared) or I (Invalid) state depending on how the system has been configured.  This response will also be sent when a cache requests E (Exclusive) ownership of a cache line without receiving data, because the cache must acquire ownership",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_iio_clockticks",
	.event = "event=0x1",
	.desc = "Clockticks of the IIO Traffic Controller. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts clockticks of the 1GHz trafiic controller clock in the IIO unit",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_by_cpu.mem_read.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x04,event=0xc0",
	.desc = "Read request for 4 bytes made by the CPU to IIO Part0. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for 4 bytes of data made by a unit on the main die (generally a core) to the MMIO space of a card on IIO Part0. In the general case, Part0 refers to a standard PCIe card of any size (x16,x8,x4) that is plugged directly into one of the PCIe slots. Part0 could also refer to any device plugged into the first slot of a PCIe riser card or to a device attached to the IIO unit which starts its use of the bus using lane 0 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_by_cpu.mem_read.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x04,event=0xc0",
	.desc = "Read request for 4 bytes made by the CPU to IIO Part1. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for 4 bytes of data made by a unit on the main die (generally a core) to the MMIO space of a card on IIO Part1. In the general case, Part1 refers to a x4 PCIe card plugged into the second slot of a PCIe riser card, but it could refer to any x4 device attached to the IIO unit using lanes starting at lane 4 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_by_cpu.mem_read.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x04,event=0xc0",
	.desc = "Read request for 4 bytes made by the CPU to IIO Part2. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for 4 bytes of data made by a unit on the main die (generally a core) to the MMIO space of a card on IIO Part2. In the general case, Part2 refers to a x4 or x8 PCIe card plugged into the third slot of a PCIe riser card, but it could refer to any x4 or x8 device attached to the IIO unit and using lanes starting at lane 8 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_by_cpu.mem_read.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x04,event=0xc0",
	.desc = "Read request for 4 bytes made by the CPU to IIO Part3. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for 4 bytes of data made by a unit on the main die (generally a core) to the MMIO space of a card on IIO Part3. In the general case, Part3 refers to a x4 PCIe card plugged into the fourth slot of a PCIe riser card, but it could brefer to  any device attached to the IIO unit using the lanes starting at lane 12 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_by_cpu.mem_write.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x01,event=0xc0",
	.desc = "Write request of 4 bytes made to IIO Part0 by the CPU. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of 4 bytes of data made to the MMIO space of a card on IIO Part0 by a unit on the main die (generally a core). In the general case, Part0 refers to a standard PCIe card of any size (x16,x8,x4) that is plugged directly into one of the PCIe slots. Part0 could also refer to any device plugged into the first slot of a PCIe riser card or to a device attached to the IIO unit which starts its use of the bus using lane 0 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_by_cpu.mem_write.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x01,event=0xc0",
	.desc = "Write request of 4 bytes made to IIO Part1 by the CPU. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of 4 bytes of data made to the MMIO space of a card on IIO Part1 by a unit on the main die (generally a core). In the general case, Part1 refers to a x4 PCIe card plugged into the second slot of a PCIe riser card, but it could refer to any x4 device attached to the IIO unit using lanes starting at lane 4 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_by_cpu.mem_write.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x01,event=0xc0",
	.desc = "Write request of 4 bytes made to IIO Part2 by the CPU . Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of 4 bytes of data made to the MMIO space of a card on IIO Part2 by  a unit on the main die (generally a core). In the general case, Part2 refers to a x4 or x8 PCIe card plugged into the third slot of a PCIe riser card, but it could refer to any x4 or x8 device attached to the IIO unit and using lanes starting at lane 8 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_by_cpu.mem_write.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x01,event=0xc0",
	.desc = "Write request of 4 bytes made to IIO Part3 by the CPU . Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of 4 bytes of data made to the MMIO space of a card on IIO Part3 by  a unit on the main die (generally a core). In the general case, Part3 refers to a x4 PCIe card plugged into the fourth slot of a PCIe riser card, but it could brefer to any device attached to the IIO unit using the lanes starting at lane 12 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_by_cpu.mem_read.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x04,event=0xc1",
	.desc = "Read request for up to a 64 byte transaction is made by the CPU to IIO Part0. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for up to a 64 byte transaction of data made by a unit on the main die (generally a core) to the MMIO space of a card on IIO Part0. In the general case, part0 refers to a standard PCIe card of any size (x16,x8,x4) that is plugged directly into one of the PCIe slots. Part0 could also refer to any device plugged into the first slot of a PCIe riser card or to a device attached to the IIO unit which starts its use of the bus using lane 0 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_by_cpu.mem_read.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x04,event=0xc1",
	.desc = "Read request for up to a 64 byte transaction is made by the CPU to IIO Part1. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for up to a 64 byte transaction of data made by a unit on the main die (generally a core) to the MMIO space of a card on IIO Part1. In the general case, Part1 refers to a x4 PCIe card plugged into the second slot of a PCIe riser card, but it could refer to any x4 device attached to the IIO unit using lanes starting at lane 4 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_by_cpu.mem_read.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x04,event=0xc1",
	.desc = "Read request for up to a 64 byte transaction is made by the CPU to IIO Part2. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for up to a 64 byte transaction of data made by a unit on the main die (generally a core) to the MMIO space of a card on IIO Part2. In the general case, Part2 refers to a x4 or x8 PCIe card plugged into the third slot of a PCIe riser card, but it could refer to any x4 or x8 device attached to the IIO unit and using lanes starting at lane 8 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_by_cpu.mem_read.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x04,event=0xc1",
	.desc = "Read request for up to a 64 byte transaction is made by the CPU to IIO Part3. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for up to a 64 byte transaction of data made by a unit on the main die (generally a core) to the MMIO space of a card on IIO Part3. In the general case, Part3 refers to a x4 PCIe card plugged into the fourth slot of a PCIe riser card, but it could brefer to  any device attached to the IIO unit using the lanes starting at lane 12 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_by_cpu.mem_write.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x01,event=0xc1",
	.desc = "Write request of up to a 64 byte transaction is made to IIO Part0 by the CPU. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of up to a 64 byte transaction of data made to the MMIO space of a card on IIO Part0 by a unit on the main die (generally a core). In the general case, Part0 refers to a standard PCIe card of any size (x16,x8,x4) that is plugged directly into one of the PCIe slots. Part0 could also refer to any device plugged into the first slot of a PCIe riser card or to a device attached to the IIO unit which starts its use of the bus using lane 0 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_by_cpu.mem_write.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x01,event=0xc1",
	.desc = "Write request of up to a 64 byte transaction is made to IIO Part1 by the CPU. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of up to a 64 byte transaction of data made to the MMIO space of a card on IIO Part1 by a unit on the main die (generally a core). In the general case, Part1 refers to a x4 PCIe card plugged into the second slot of a PCIe riser card, but it could refer to any x4 device attached to the IIO unit using lanes starting at lane 4 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_by_cpu.mem_write.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x01,event=0xc1",
	.desc = "Write request of up to a 64 byte transaction is made to IIO Part2 by the CPU . Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of up to a 64 byte transaction of data made to the MMIO space of a card on IIO Part2 by a unit on the main die (generally a core). In the general case, Part2 refers to a x4 or x8 PCIe card plugged into the third slot of a PCIe riser card, but it could refer to any x4 or x8 device attached to the IIO unit and using lanes starting at lane 8 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_by_cpu.mem_write.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x01,event=0xc1",
	.desc = "Write request of up to a 64 byte transaction is made to IIO Part3 by the CPU . Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of up to a 64 byte transaction of data made to the MMIO space of a card on IIO Part3 by a unit on the main die (generally a core). In the general case, Part3 refers to a x4 PCIe card plugged into the fourth slot of a PCIe riser card, but it could brefer to  any device attached to the IIO unit using the lanes starting at lane 12 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_of_cpu.mem_read.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x04,event=0x84",
	.desc = "Read request for up to a 64 byte transaction is made by IIO Part0 to Memory. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for up to a 64 byte transaction of data made by IIO Part0 to a unit on the main die (generally memory). In the general case, Part0 refers to a standard PCIe card of any size (x16,x8,x4) that is plugged directly into one of the PCIe slots. Part0 could also refer to any device plugged into the first slot of a PCIe riser card or to a device attached to the IIO unit which starts its use of the bus using lane 0 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_of_cpu.mem_read.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x04,event=0x84",
	.desc = "Read request for up to a 64 byte transaction is  made by IIO Part1 to Memory. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for up to a 64 byte transaction of data made by IIO Part1 to a unit on the main die (generally memory). In the general case, Part1 refers to a x4 PCIe card plugged into the second slot of a PCIe riser card, but it could refer to any x4 device attached to the IIO unit using lanes starting at lane 4 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_of_cpu.mem_read.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x04,event=0x84",
	.desc = "Read request for up to a 64 byte transaction is made by IIO Part2 to Memory. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for up to a 64 byte transaction of data made by IIO Part2 to a unit on the main die (generally memory). In the general case, Part2 refers to a x4 or x8 PCIe card plugged into the third slot of a PCIe riser card, but it could refer to any x4 or x8 device attached to the IIO unit and using lanes starting at lane 8 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_of_cpu.mem_read.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x04,event=0x84",
	.desc = "Read request for up to a 64 byte transaction is made by IIO Part3 to Memory. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every read request for up to a 64 byte transaction of data made by IIO Part3 to a unit on the main die (generally memory). In the general case, Part3 refers to a x4 PCIe card plugged into the fourth slot of a PCIe riser card, but it could brefer to  any device attached to the IIO unit using the lanes starting at lane 12 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_of_cpu.mem_write.part0",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x01,event=0x84",
	.desc = "Write request of up to a 64 byte transaction is made by IIO Part0 to Memory. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of up to a 64 byte transaction of data made by IIO Part0 to a unit on the main die (generally memory). In the general case, Part0 refers to a standard PCIe card of any size (x16,x8,x4) that is plugged directly into one of the PCIe slots. Part0 could also refer to any device plugged into the first slot of a PCIe riser card or to a device attached to the IIO unit which starts its use of the bus using lane 0 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_of_cpu.mem_write.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x01,event=0x84",
	.desc = "Write request of up to a 64 byte transaction is made by IIO Part1 to Memory. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of up to a 64 byte transaction of data made by IIO Part1 to a unit on the main die (generally memory). In the general case, Part1 refers to a x4 PCIe card plugged into the second slot of a PCIe riser card, but it could refer to any x4 device attached to the IIO unit using lanes starting at lane 4 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_of_cpu.mem_write.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x01,event=0x84",
	.desc = "Write request of up to a 64 byte transaction is made by IIO Part2 to Memory. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of up to a 64 byte transaction of data made by IIO Part2 to a unit on the main die (generally memory). In the general case, Part2 refers to a x4 or x8 PCIe card plugged into the third slot of a PCIe riser card, but it could refer to any x4 or x8 device attached to the IIO unit and using lanes starting at lane 8 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_txn_req_of_cpu.mem_write.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x01,event=0x84",
	.desc = "Write request of up to a 64 byte transaction is made by IIO Part3 to Memory. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Counts every write request of up to a 64 byte transaction of data made by IIO Part3 to a unit on the main die (generally memory). In the general case, Part3 refers to a x4 PCIe card plugged into the fourth slot of a PCIe riser card, but it could brefer to  any device attached to the IIO unit using the lanes starting at lane 12 of the 16 lanes supported by the bus",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_m2m_bypass_m2m_egress.not_taken",
	.event = "umask=0x2,event=0x22",
	.desc = "Traffic in which the M2M to iMC Bypass was not taken. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts traffic in which the M2M (Mesh to Memory) to iMC (Memory Controller) bypass was not taken",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_direct2core_not_taken_dirstate",
	.event = "event=0x24",
	.desc = "Cycles when direct to core mode (which bypasses the CHA) was disabled. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts cycles when direct to core mode (which bypasses the CHA) was disabled",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_direct2core_taken",
	.event = "event=0x23",
	.desc = "Messages sent direct to core (bypassing the CHA). Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when messages were sent direct to core (bypassing the CHA)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_direct2core_txn_override",
	.event = "event=0x25",
	.desc = "Number of reads in which direct to core transaction were overridden. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts reads in which direct to core transactions (which would have bypassed the CHA) were overridden",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_direct2upi_not_taken_credits",
	.event = "event=0x28",
	.desc = "Number of reads in which direct to Intel UPI transactions were overridden. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts reads in which direct to Intel Ultra Path Interconnect (UPI) transactions (which would have bypassed the CHA) were overridden",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_direct2upi_not_taken_dirstate",
	.event = "event=0x27",
	.desc = "Cycles when direct to Intel UPI was disabled. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts cycles when the ability to send messages direct to the Intel Ultra Path Interconnect (bypassing the CHA) was disabled",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_direct2upi_taken",
	.event = "event=0x26",
	.desc = "Messages sent direct to the Intel UPI. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when messages were sent direct to the Intel Ultra Path Interconnect (bypassing the CHA)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_direct2upi_txn_override",
	.event = "event=0x29",
	.desc = "Number of reads that a message sent direct2 Intel UPI was overridden. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when a read message that was sent direct to the Intel Ultra Path Interconnect (bypassing the CHA) was overridden",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_lookup.any",
	.event = "umask=0x1,event=0x2d",
	.desc = "Multi-socket cacheline Directory lookups (any state found). Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) looks into the multi-socket cacheline Directory state, and found the cacheline marked in Any State (A, I, S or unused)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_lookup.state_a",
	.event = "umask=0x8,event=0x2d",
	.desc = "Multi-socket cacheline Directory lookups (cacheline found in A state) . Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) looks into the multi-socket cacheline Directory state, and found the cacheline marked in the A (SnoopAll) state, indicating the cacheline is stored in another socket in any state, and we must snoop the other sockets to make sure we get the latest data.  The data may be stored in any state in the local socket",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_lookup.state_i",
	.event = "umask=0x2,event=0x2d",
	.desc = "Multi-socket cacheline Directory lookup (cacheline found in I state) . Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) looks into the multi-socket cacheline Directory state , and found the cacheline marked in the I (Invalid) state indicating the cacheline is not stored in another socket, and so there is no need to snoop the other sockets for the latest data.  The data may be stored in any state in the local socket",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_lookup.state_s",
	.event = "umask=0x4,event=0x2d",
	.desc = "Multi-socket cacheline Directory lookup (cacheline found in S state) . Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) looks into the multi-socket cacheline Directory state , and found the cacheline marked in the S (Shared) state indicating the cacheline is either stored in another socket in the S(hared) state , and so there is no need to snoop the other sockets for the latest data.  The data may be stored in any state in the local socket",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_update.a2i",
	.event = "umask=0x20,event=0x2e",
	.desc = "Multi-socket cacheline Directory update from A to I. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) updates the multi-socket cacheline Directory state from from A (SnoopAll) to I (Invalid)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_update.a2s",
	.event = "umask=0x40,event=0x2e",
	.desc = "Multi-socket cacheline Directory update from A to S. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) updates the multi-socket cacheline Directory state from from A (SnoopAll) to S (Shared)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_update.any",
	.event = "umask=0x1,event=0x2e",
	.desc = "Multi-socket cacheline Directory update from/to Any state . Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) updates the multi-socket cacheline Directory to a new state",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_update.i2a",
	.event = "umask=0x4,event=0x2e",
	.desc = "Multi-socket cacheline Directory update from I to A. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) updates the multi-socket cacheline Directory state from from I (Invalid) to A (SnoopAll)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_update.i2s",
	.event = "umask=0x2,event=0x2e",
	.desc = "Multi-socket cacheline Directory update from I to S. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) updates the multi-socket cacheline Directory state from from I (Invalid) to S (Shared)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_update.s2a",
	.event = "umask=0x10,event=0x2e",
	.desc = "Multi-socket cacheline Directory update from S to A. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) updates the multi-socket cacheline Directory state from from S (Shared) to A (SnoopAll)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_directory_update.s2i",
	.event = "umask=0x8,event=0x2e",
	.desc = "Multi-socket cacheline Directory update from S to I. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) updates the multi-socket cacheline Directory state from from S (Shared) to I (Invalid)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_imc_reads.all",
	.event = "umask=0x4,event=0x37",
	.desc = "Reads to iMC issued. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) issues reads to the iMC (Memory Controller)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_imc_reads.normal",
	.event = "umask=0x1,event=0x37",
	.desc = "Reads to iMC issued at Normal Priority (Non-Isochronous). Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) issues reads to the iMC (Memory Controller).  It only counts  normal priority non-isochronous reads",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_imc_writes.all",
	.event = "umask=0x10,event=0x38",
	.desc = "Writes to iMC issued. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) issues writes to the iMC (Memory Controller)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_imc_writes.partial",
	.event = "umask=0x2,event=0x38",
	.desc = "Partial Non-Isochronous writes to the iMC. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) issues partial writes to the iMC (Memory Controller).  It only counts normal priority non-isochronous writes",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_prefcam_demand_promotions",
	.event = "event=0x56",
	.desc = "Prefecth requests that got turn into a demand request. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) promotes a outstanding request in the prefetch queue due to a subsequent demand read request that entered the M2M with the same address.  Explanatory Side Note: The Prefecth queue is made of CAM (Content Addressable Memory)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_prefcam_inserts",
	.event = "event=0x57",
	.desc = "Inserts into the Memory Controller Prefetch Queue. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the M2M (Mesh to Memory) receives a prefetch request and inserts it into its outstanding prefetch queue.  Explanatory Side Note: the prefect queue is made from CAM: Content Addressable Memory",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2m_rxc_ad_inserts",
	.event = "event=0x1",
	.desc = "AD Ingress (from CMS) Queue Inserts. Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Counts when the a new entry is Received(RxC) and then added to the AD (Address Ring) Ingress Queue from the CMS (Common Mesh Stop).  This is generally used for reads, and ",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m3upi_upi_prefetch_spawn",
	.event = "event=0x29",
	.desc = "Prefetches generated by the flow control queue of the M3UPI unit. Unit: uncore_m3upi ",
	.topic = "uncore other",
	.long_desc = "Count cases where flow control queue that sits between the Intel Ultra Path Interconnect (UPI) and the mesh spawns a prefetch to the iMC (Memory Controller)",
	.pmu = "uncore_m3upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_clockticks",
	.event = "event=0x1",
	.desc = "Clocks of the Intel Ultra Path Interconnect (UPI). Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts clockticks of the fixed frequency clock controlling the Intel Ultra Path Interconnect (UPI).  This clock runs at1/8th the 'GT/s' speed of the UPI link.  For example, a  9.6GT/s  link will have a fixed Frequency of 1.2 Ghz",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_direct_attempts.d2c",
	.event = "umask=0x1,event=0x12",
	.desc = "Data Response packets that go direct to core. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts Data Response (DRS) packets that attempted to go direct to core bypassing the CHA",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_direct_attempts.d2u",
	.event = "umask=0x2,event=0x12",
	.desc = "Data Response packets that go direct to Intel UPI. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts Data Response (DRS) packets that attempted to go direct to Intel Ultra Path Interconnect (UPI) bypassing the CHA ",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_l1_power_cycles",
	.event = "event=0x21",
	.desc = "Cycles Intel UPI is in L1 power mode (shutdown). Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts cycles when the Intel Ultra Path Interconnect (UPI) is in L1 power mode.  L1 is a mode that totally shuts down the UPI link.  Link power states are per link and per direction, so for example the Tx direction could be in one state while Rx was in another, this event only coutns when both links are shutdown",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_rxl0p_power_cycles",
	.event = "event=0x25",
	.desc = "Cycles the Rx of the Intel UPI is in L0p power mode. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts cycles when the the receive side (Rx) of the Intel Ultra Path Interconnect(UPI) is in L0p power mode. L0p is a mode where we disable 60% of the UPI lanes, decreasing our bandwidth in order to save power",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_rxl_bypassed.slot0",
	.event = "umask=0x1,event=0x31",
	.desc = "FLITs received which bypassed the Slot0 Receive Buffer. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts incoming FLITs (FLow control unITs) which bypassed the slot0 RxQ buffer (Receive Queue) and passed directly to the Egress.  This is a latency optimization, and should generally be the common case.  If this value is less than the number of FLITs transferred, it implies that there was queueing getting onto the ring, and thus the transactions saw higher latency",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_rxl_bypassed.slot1",
	.event = "umask=0x2,event=0x31",
	.desc = "FLITs received which bypassed the Slot0 Receive Buffer. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts incoming FLITs (FLow control unITs) which bypassed the slot1 RxQ buffer  (Receive Queue) and passed directly across the BGF and into the Egress.  This is a latency optimization, and should generally be the common case.  If this value is less than the number of FLITs transferred, it implies that there was queueing getting onto the ring, and thus the transactions saw higher latency",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_rxl_bypassed.slot2",
	.event = "umask=0x4,event=0x31",
	.desc = "FLITs received which bypassed the Slot0 Receive Buffer. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts incoming FLITs (FLow control unITs) which bypassed the slot2 RxQ buffer (Receive Queue)  and passed directly to the Egress.  This is a latency optimization, and should generally be the common case.  If this value is less than the number of FLITs transferred, it implies that there was queueing getting onto the ring, and thus the transactions saw higher latency",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_rxl_flits.all_data",
	.event = "umask=0x0F,event=0x3",
	.desc = "Valid data FLITs received from any slot. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts valid data FLITs  (80 bit FLow control unITs: 64bits of data) received from any of the 3 Intel Ultra Path Interconnect (UPI) Receive Queue slots on this UPI unit",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_rxl_flits.all_null",
	.event = "umask=0x27,event=0x3",
	.desc = "Null FLITs received from any slot. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts null FLITs (80 bit FLow control unITs) received from any of the 3 Intel Ultra Path Interconnect (UPI) Receive Queue slots on this UPI unit",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_rxl_flits.non_data",
	.event = "umask=0x97,event=0x3",
	.desc = "Protocol header and credit FLITs received from any slot. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts protocol header and credit FLITs  (80 bit FLow control unITs) received from any of the 3 UPI slots on this UPI unit",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_txl0p_power_cycles",
	.event = "event=0x27",
	.desc = "Cycles in which the Tx of the Intel Ultra Path Interconnect (UPI) is in L0p power mode. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts cycles when the transmit side (Tx) of the Intel Ultra Path Interconnect(UPI) is in L0p power mode. L0p is a mode where we disable 60% of the UPI lanes, decreasing our bandwidth in order to save power",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_txl_bypassed",
	.event = "event=0x41",
	.desc = "FLITs that bypassed the TxL Buffer. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts incoming FLITs (FLow control unITs) which bypassed the TxL(transmit) FLIT buffer and pass directly out the UPI Link. Generally, when data is transmitted across the Intel Ultra Path Interconnect (UPI), it will bypass the TxQ and pass directly to the link.  However, the TxQ will be used in L0p (Low Power) mode and (Link Layer Retry) LLR  mode, increasing latency to transfer out to the link",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "upi_data_bandwidth_tx",
	.event = "umask=0x0F,event=0x2",
	.desc = "UPI interconnect send bandwidth for payload. Derived from unc_upi_txl_flits.all_data. Unit: uncore_upi ",
	.topic = "uncore other",
	.pmu = "uncore_upi",
	.unit = "7.11E-06Bytes",
	.perpkg = "1",
},
{
	.name = "unc_upi_txl_flits.all_null",
	.event = "umask=0x27,event=0x2",
	.desc = "Null FLITs transmitted from any slot. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts null FLITs (80 bit FLow control unITs) transmitted via any of the 3 Intel Ulra Path Interconnect (UPI) slots on this UPI unit",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_txl_flits.idle",
	.event = "umask=0x47,event=0x2",
	.desc = "Idle FLITs transmitted. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts when the Intel Ultra Path Interconnect(UPI) transmits an idle FLIT(80 bit FLow control unITs).  Every UPI cycle must be sending either data FLITs, protocol/credit FLITs or idle FLITs",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "unc_upi_txl_flits.non_data",
	.event = "umask=0x97,event=0x2",
	.desc = "Protocol header and credit FLITs transmitted across any slot. Unit: uncore_upi ",
	.topic = "uncore other",
	.long_desc = "Counts protocol header and credit FLITs (80 bit FLow control unITs) transmitted across any of the 3 UPI (Ultra Path Interconnect) slots on this UPI unit",
	.pmu = "uncore_upi",
	.perpkg = "1",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts demand data loads that caused a page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels, but the walk need not have completed",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x8",
	.desc = "Page walk completed due to a demand data load to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Page walk completed due to a demand data load to a 2M/4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 2M/4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_1g",
	.event = "umask=0x8,period=2000003,event=0x8",
	.desc = "Page walk completed due to a demand data load to a 1G page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x8",
	.desc = "Load miss in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts demand data loads that caused a completed page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels. The page walk can end with or without a fault",
},
{
	.name = "dtlb_load_misses.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x8",
	.desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a load. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a load. EPT page walk duration are excluded in Skylake microarchitecture",
},
{
	.name = "dtlb_load_misses.walk_active",
	.event = "umask=0x10,cmask=1,period=100003,event=0x8",
	.desc = "Cycles when at least one PMH is busy with a page walk for a load. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts cycles when at least one PMH (Page Miss Handler) is busy with a page walk for a load",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x20,period=2000003,event=0x8",
	.desc = "Loads that miss the DTLB and hit the STLB",
	.topic = "virtual memory",
	.long_desc = "Counts loads that miss the DTLB (Data TLB) and hit the STLB (Second level TLB)",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts demand data stores that caused a page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels, but the walk need not have completed",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 2M/4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 2M/4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 1G page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 1G pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x49",
	.desc = "Store misses in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts demand data stores that caused a completed page walk of any page size (4K/2M/4M/1G). This implies it missed in all TLB levels. The page walk can end with or without a fault",
},
{
	.name = "dtlb_store_misses.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x49",
	.desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a store. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts 1 per cycle for each PMH that is busy with a page walk for a store. EPT page walk duration are excluded in Skylake microarchitecture",
},
{
	.name = "dtlb_store_misses.walk_active",
	.event = "umask=0x10,cmask=1,period=100003,event=0x49",
	.desc = "Cycles when at least one PMH is busy with a page walk for a store. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts cycles when at least one PMH (Page Miss Handler) is busy with a page walk for a store",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x20,period=100003,event=0x49",
	.desc = "Stores that miss the DTLB and hit the STLB",
	.topic = "virtual memory",
	.long_desc = "Stores that miss the DTLB (Data TLB) and hit the STLB (2nd Level TLB)",
},
{
	.name = "ept.walk_pending",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Counts 1 per cycle for each PMH that is busy with a EPT (Extended Page Table) walk for any request type",
	.topic = "virtual memory",
	.long_desc = "Counts cycles for each PMH (Page Miss Handler) that is busy with an EPT (Extended Page Table) walk for any request type",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks",
	.topic = "virtual memory",
	.long_desc = "Counts page walks of any page size (4K/2M/4M/1G) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB, but the walk need not have completed",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (4K)",
	.topic = "virtual memory",
	.long_desc = "Counts completed page walks (4K page size) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB. The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (2M/4M)",
	.topic = "virtual memory",
	.long_desc = "Counts code misses in all ITLB levels that caused a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (1G)",
	.topic = "virtual memory",
	.long_desc = "Counts store misses in all DTLB levels that cause a completed page walk (1G page size). The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (All page sizes)",
	.topic = "virtual memory",
	.long_desc = "Counts completed page walks (2M and 4M page sizes) caused by a code fetch. This implies it missed in the ITLB and further levels of TLB. The page walk can end with or without a fault",
},
{
	.name = "itlb_misses.walk_pending",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Counts 1 per cycle for each PMH that is busy with a page walk for an instruction fetch request. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Counts 1 per cycle for each PMH (Page Miss Handler) that is busy with a page walk for an instruction fetch request. EPT page walk duration are excluded in Skylake michroarchitecture",
},
{
	.name = "itlb_misses.walk_active",
	.event = "umask=0x10,cmask=1,period=100003,event=0x85",
	.desc = "Cycles when at least one PMH is busy with a page walk for code (instruction fetch) request. EPT page walk duration are excluded in Skylake",
	.topic = "virtual memory",
	.long_desc = "Cycles when at least one PMH is busy with a page walk for code (instruction fetch) request. EPT page walk duration are excluded in Skylake microarchitecture",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x20,period=100003,event=0x85",
	.desc = "Instruction fetch requests that miss the ITLB and hit the STLB",
	.topic = "virtual memory",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "Counts the number of flushes of the big or small ITLB pages. Counting include both TLB Flush (covering all sets) and TLB Set Clear (set-specific)",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "Counts the number of DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "Counts the number of any STLB flush attempts (such as entire, VPID, PCID, InvPage, CR3 write, etc.)",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_tremontx[] = {
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=200003,event=0x2e",
	.desc = "Counts memory requests originating from the core that miss in the last level cache. If the platform has an L3 cache, last level cache is the L3, otherwise it is the L2",
	.topic = "cache",
	.long_desc = "Counts cacheable memory requests that miss in the the Last Level Cache.  Requests include Demand Loads, Reads for Ownership(RFO), Instruction fetches and L1 HW prefetches. If the platform has an L3 cache, last level cache is the L3, otherwise it is the L2",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=200003,event=0x2e",
	.desc = "Counts memory requests originating from the core that reference a cache line in the last level cache. If the platform has an L3 cache, last level cache is the L3, otherwise it is the L2",
	.topic = "cache",
	.long_desc = "Counts cacheable memory requests that access the Last Level Cache.  Requests include Demand Loads, Reads for Ownership(RFO), Instruction fetches and L1 HW prefetches. If the platform has an L3 cache, last level cache is the L3, otherwise it is the L2",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=200003,event=0xd0",
	.desc = "Counts the number of load uops retired  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts the number of load uops retired. This event is Precise Event capable  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=200003,event=0xd0",
	.desc = "Counts the number of store uops retired  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "Counts the number of store uops retired. This event is Precise Event capable  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=200003,event=0xd1",
	.desc = "Counts the number of load uops retired that hit the level 1 data cache  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=200003,event=0xd1",
	.desc = "Counts the number of load uops retired that hit in the level 2 cache  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l3_hit",
	.event = "umask=0x4,period=200003,event=0xd1",
	.desc = "Counts the number of load uops retired that miss in the level 3 cache (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=200003,event=0xd1",
	.desc = "Counts the number of load uops retired that miss in the level 1 data cache  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=200003,event=0xd1",
	.desc = "Counts the number of load uops retired that miss in the level 2 cache  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Counts requests to the Instruction Cache (ICache) for one or more bytes in a cache line and they do not hit in the ICache (miss)",
	.topic = "frontend",
	.long_desc = "Counts requests to the Instruction Cache (ICache)  for one or more bytes in an ICache Line and that cache line is not in the ICache (miss).  The event strives to count on a cache line basis, so that multiple accesses which miss in a single cache line count as one ICACHE.MISS.  Specifically, the event counts when straight line code crosses the cache line boundary, or when a branch target is to a new line, and that cache line is not in the ICache",
},
{
	.name = "icache.accesses",
	.event = "umask=0x3,period=200003,event=0x80",
	.desc = "Counts requests to the Instruction Cache (ICache) for one or more bytes cache Line",
	.topic = "frontend",
	.long_desc = "Counts requests to the Instruction Cache (ICache) for one or more bytes in an ICache Line.  The event strives to count on a cache line basis, so that multiple fetches to a single cache line count as one ICACHE.ACCESS.  Specifically, the event counts when accesses from straight line code crosses the cache line boundary, or when a branch target is to a new line",
},
{
	.name = "ocr.demand_data_rd.l3_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000000003F04000001",
	.desc = "Counts demand data reads that was not supplied by the L3 cache",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.l3_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000000003F04000002",
	.desc = "Counts all demand reads for ownership (RFO) requests and software based prefetches for exclusive ownership (PREFETCHW) that was not supplied by the L3 cache",
	.topic = "memory",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000000000000010001",
	.desc = "Counts demand data reads that have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "ocr.demand_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x000000000000010002",
	.desc = "Counts all demand reads for ownership (RFO) requests and software based prefetches for exclusive ownership (PREFETCHW) that have any response type",
	.topic = "other",
	.long_desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Counts the number of instructions retired. (Fixed event) (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions that retire. For instructions that consist of multiple uops, this event counts the retirement of the last uop of the instruction. The counter continues counting during hardware interrupts, traps, and inside interrupt handlers.  This event uses fixed counter 0 (Precise event)",
},
{
	.name = "cpu_clk_unhalted.core",
	.event = "event=0x3c,period=2000003",
	.desc = "Counts the number of unhalted core clock cycles. (Fixed event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of core cycles while the core is not in a halt state.  The core enters the halt state when it is running the HLT instruction. The core frequency may change from time to time. For this reason this event may have a changing ratio with regards to time.  This event uses fixed counter 1",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Counts the number of unhalted reference clock cycles at TSC frequency. (Fixed event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of reference cycles that the core is not in a halt state. The core enters the halt state when it is running the HLT instruction.  The core frequency may change from time.  This event is not affected by core frequency changes and at a fixed frequency.  This event uses fixed counter 2",
},
{
	.name = "cpu_clk_unhalted.core_p",
	.event = "period=2000003,event=0x3c",
	.desc = "Counts the number of unhalted core clock cycles",
	.topic = "pipeline",
	.long_desc = "Counts the number of core cycles while the core is not in a halt state.  The core enters the halt state when it is running the HLT instruction. The core frequency may change from time to time. For this reason this event may have a changing ratio with regards to time.  This event uses a programmable general purpose performance counter",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Counts the number of unhalted reference clock cycles at TSC frequency",
	.topic = "pipeline",
	.long_desc = "Counts reference cycles (at TSC frequency) when core is not halted.  This event uses a programmable general purpose perfmon counter",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Counts the number of instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts the number of instructions that retire execution. For instructions that consist of multiple uops, this event counts the retirement of the last uop of the instruction. The event continues counting during hardware interrupts, traps, and inside interrupt handlers.  This is an architectural performance event.  This event uses a Programmable general purpose perfmon counter. *This event is Precise Event capable:  The EventingRIP field in the PEBS record is precise to the address of the instruction which caused the event (Precise event)",
},
{
	.name = "machine_clears.any",
	.event = "period=20003,event=0xc3",
	.desc = "Counts all machine clears due to, but not limited to memory ordering, memory disambiguation, SMC, page faults and FP assist",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "period=200003,event=0xc4",
	.desc = "Counts the number of branch instructions retired for all branch types (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts branch instructions retired for all branch types. This event is Precise Event capable. This is an architectural event (Precise event)",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "period=200003,event=0xc5",
	.desc = "Counts the number of mispredicted branch instructions retired (Precise event)",
	.topic = "pipeline",
	.long_desc = "Counts mispredicted branch instructions retired for all branch types. This event is Precise Event capable. This is an architectural event (Precise event)",
},
{
	.name = "cycles_div_busy.any",
	.event = "period=2000003,event=0xcd",
	.desc = "Counts cycles the floating point divider or integer divider or both are busy.  Does not imply a stall waiting for either divider",
	.topic = "pipeline",
},
{
	.name = "llc_misses.mem_read",
	.event = "umask=0x0f,event=0x4",
	.desc = "read requests to memory controller. Derived from unc_m_cas_count.rd. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.mem_write",
	.event = "umask=0x30,event=0x4",
	.desc = "write requests to memory controller. Derived from unc_m_cas_count.wr. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "unc_m_clockticks",
	.event = "event=0",
	.desc = "Memory controller clock ticks. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.rd",
	.event = "umask=0x04,event=0x2",
	.desc = "Pre-charge for reads. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.wr",
	.event = "umask=0x08,event=0x2",
	.desc = "Pre-charge for writes. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.all",
	.event = "umask=0x1c,event=0x2",
	.desc = "Precharge due to read on page miss, write on page miss or PGT. Unit: uncore_imc ",
	.topic = "uncore memory",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_m_pre_count.pgt",
	.event = "umask=0x10,event=0x2",
	.desc = "DRAM Precharge commands. : Precharge due to page table. Unit: uncore_imc ",
	.topic = "uncore memory",
	.long_desc = "DRAM Precharge commands. : Precharge due to page table : Counts the number of DRAM Precharge commands sent on this channel",
	.pmu = "uncore_imc",
	.perpkg = "1",
},
{
	.name = "unc_cha_clockticks",
	.event = "event=0",
	.desc = "Uncore cache clock ticks. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_misses.uncacheable",
	.event = "umask=0xC001FE01,event=0x35,config1=0x40e33",
	.desc = "LLC misses - Uncacheable reads (from cpu) . Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_read",
	.event = "umask=0xC001FE01,event=0x35,config1=0x40040e33",
	.desc = "MMIO reads. Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_misses.mmio_write",
	.event = "umask=0xC001FE01,event=0x35,config1=0x40041e33",
	.desc = "MMIO writes. Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_full",
	.event = "umask=0xC001FE01,event=0x35,config1=0x41833",
	.desc = "Streaming stores (full cache line). Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_references.streaming_partial",
	.event = "umask=0xC001FE01,event=0x35,config1=0x41a33",
	.desc = "Streaming stores (partial cache line). Derived from unc_cha_tor_inserts.ia_miss. Unit: uncore_cha ",
	.topic = "uncore other",
	.pmu = "uncore_cha",
	.unit = "64Bytes",
	.perpkg = "1",
},
{
	.name = "llc_misses.pcie_read",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x04,event=0x83,ch_mask=0x1f",
	.desc = "PCI Express bandwidth reading at IIO. Derived from unc_iio_data_req_of_cpu.mem_read.part0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_read.part0 +unc_iio_data_req_of_cpu.mem_read.part1 +unc_iio_data_req_of_cpu.mem_read.part2 +unc_iio_data_req_of_cpu.mem_read.part3",
	.metric_name = "LLC_MISSES.PCIE_READ",
},
{
	.name = "llc_misses.pcie_write",
	.event = "fc_mask=0x07,ch_mask=0x01,umask=0x01,event=0x83,ch_mask=0x1f",
	.desc = "PCI Express bandwidth writing at IIO. Derived from unc_iio_data_req_of_cpu.mem_write.part0. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
	.metric_expr = "unc_iio_data_req_of_cpu.mem_write.part0 +unc_iio_data_req_of_cpu.mem_write.part1 +unc_iio_data_req_of_cpu.mem_write.part2 +unc_iio_data_req_of_cpu.mem_write.part3",
	.metric_name = "LLC_MISSES.PCIE_WRITE",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 1. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 2. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x01,event=0x83",
	.desc = "PCI Express bandwidth writing at IIO, part 3. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part1",
	.event = "fc_mask=0x07,ch_mask=0x02,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 1. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part2",
	.event = "fc_mask=0x07,ch_mask=0x04,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 2. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part3",
	.event = "fc_mask=0x07,ch_mask=0x08,umask=0x04,event=0x83",
	.desc = "PCI Express bandwidth reading at IIO, part 3. Unit: uncore_iio ",
	.topic = "uncore other",
	.pmu = "uncore_iio",
	.unit = "4Bytes",
	.perpkg = "1",
},
{
	.name = "unc_cha_tor_inserts.ia_miss_crd",
	.event = "umask=0xC80FFE01,event=0x35",
	.desc = "TOR Inserts; CRd misses from local IA. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "TOR Inserts; Code read from local IA that misses in the snoop filter",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_tor_inserts.ia_miss_crd_pref",
	.event = "umask=0xC88FFE01,event=0x35",
	.desc = "TOR Inserts; CRd Pref misses from local IA. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "TOR Inserts; Code read prefetch from local IA that misses in the snoop filter",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_tor_inserts.ia_miss_drd_opt",
	.event = "umask=0xC827FE01,event=0x35",
	.desc = "TOR Inserts; DRd Opt misses from local IA. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "TOR Inserts; Data read opt from local IA that misses in the snoop filter",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_tor_inserts.ia_miss_drd_opt_pref",
	.event = "umask=0xC8A7FE01,event=0x35",
	.desc = "TOR Inserts; DRd Opt Pref misses from local IA. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "TOR Inserts; Data read opt prefetch from local IA that misses in the snoop filter",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_tor_inserts.ia_miss_rfo",
	.event = "umask=0xC807FE01,event=0x35",
	.desc = "TOR Inserts; RFO misses from local IA. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "TOR Inserts; Read for ownership from local IA that misses in the snoop filter",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_tor_inserts.ia_miss_rfo_pref",
	.event = "umask=0xC887FE01,event=0x35",
	.desc = "TOR Inserts; RFO pref misses from local IA. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "TOR Inserts; Read for ownership prefetch from local IA that misses in the snoop filter",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_tor_inserts.ia_miss_wcil",
	.event = "umask=0xC86FFE01,event=0x35",
	.desc = "TOR Inserts; WCiL misses from local IA. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "TOR Inserts; Data read from local IA that misses in the snoop filter",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_cha_tor_inserts.ia_miss_wcilf",
	.event = "umask=0xC867FE01,event=0x35",
	.desc = "TOR Inserts; WCiLF misses from local IA. Unit: uncore_cha ",
	.topic = "uncore other",
	.long_desc = "TOR Inserts; Data read from local IA that misses in the snoop filter",
	.pmu = "uncore_cha",
	.perpkg = "1",
},
{
	.name = "unc_iio_clockticks",
	.event = "event=0x1",
	.desc = "Clockticks of the integrated IO (IIO) traffic controller. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Clockticks of the integrated IO (IIO) traffic controller",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part4",
	.event = "fc_mask=0x07,ch_mask=0x10,umask=0x04,event=0x83",
	.desc = "Data requested of the CPU : Card reading from DRAM. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Data requested of the CPU : Card reading from DRAM : Number of DWs (4 bytes) the card requests of the main die.    Includes all requests initiated by the Card, including reads and writes. : x16 card plugged in to stack, Or x8 card plugged in to Lane 0/1, Or x4 card is plugged in to slot 0",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part5",
	.event = "fc_mask=0x07,ch_mask=0x20,umask=0x04,event=0x83",
	.desc = "Data requested of the CPU : Card reading from DRAM. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Data requested of the CPU : Card reading from DRAM : Number of DWs (4 bytes) the card requests of the main die.    Includes all requests initiated by the Card, including reads and writes. : x4 card is plugged in to slot 1",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part6",
	.event = "fc_mask=0x07,ch_mask=0x40,umask=0x04,event=0x83",
	.desc = "Data requested of the CPU : Card reading from DRAM. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Data requested of the CPU : Card reading from DRAM : Number of DWs (4 bytes) the card requests of the main die.    Includes all requests initiated by the Card, including reads and writes. : x8 card plugged in to Lane 2/3, Or x4 card is plugged in to slot 1",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_read.part7",
	.event = "fc_mask=0x07,ch_mask=0x80,umask=0x04,event=0x83",
	.desc = "Data requested of the CPU : Card reading from DRAM. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Data requested of the CPU : Card reading from DRAM : Number of DWs (4 bytes) the card requests of the main die.    Includes all requests initiated by the Card, including reads and writes. : x4 card is plugged in to slot 3",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part4",
	.event = "fc_mask=0x07,ch_mask=0x10,umask=0x01,event=0x83",
	.desc = "Data requested of the CPU : Card writing to DRAM. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Data requested of the CPU : Card writing to DRAM : Number of DWs (4 bytes) the card requests of the main die.    Includes all requests initiated by the Card, including reads and writes. : x16 card plugged in to stack, Or x8 card plugged in to Lane 0/1, Or x4 card is plugged in to slot 0",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part5",
	.event = "fc_mask=0x07,ch_mask=0x20,umask=0x01,event=0x83",
	.desc = "Data requested of the CPU : Card writing to DRAM. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Data requested of the CPU : Card writing to DRAM : Number of DWs (4 bytes) the card requests of the main die.    Includes all requests initiated by the Card, including reads and writes. : x4 card is plugged in to slot 1",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part6",
	.event = "fc_mask=0x07,ch_mask=0x40,umask=0x01,event=0x83",
	.desc = "Data requested of the CPU : Card writing to DRAM. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Data requested of the CPU : Card writing to DRAM : Number of DWs (4 bytes) the card requests of the main die.    Includes all requests initiated by the Card, including reads and writes. : x8 card plugged in to Lane 2/3, Or x4 card is plugged in to slot 1",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_iio_data_req_of_cpu.mem_write.part7",
	.event = "fc_mask=0x07,ch_mask=0x80,umask=0x01,event=0x83",
	.desc = "Data requested of the CPU : Card writing to DRAM. Unit: uncore_iio ",
	.topic = "uncore other",
	.long_desc = "Data requested of the CPU : Card writing to DRAM : Number of DWs (4 bytes) the card requests of the main die.    Includes all requests initiated by the Card, including reads and writes. : x4 card is plugged in to slot 3",
	.pmu = "uncore_iio",
	.perpkg = "1",
},
{
	.name = "unc_i_clockticks",
	.event = "event=0x1",
	.desc = "Clockticks of the IO coherency tracker (IRP). Unit: uncore_irp ",
	.topic = "uncore other",
	.long_desc = "Clockticks of the IO coherency tracker (IRP)",
	.pmu = "uncore_irp",
	.perpkg = "1",
},
{
	.name = "unc_m2m_clockticks",
	.event = "event=0",
	.desc = "Clockticks of the mesh to memory (M2M). Unit: uncore_m2m ",
	.topic = "uncore other",
	.long_desc = "Clockticks of the mesh to memory (M2M)",
	.pmu = "uncore_m2m",
	.perpkg = "1",
},
{
	.name = "unc_m2p_clockticks",
	.event = "event=0x1",
	.desc = "Clockticks of the mesh to PCI (M2P). Unit: uncore_m2pcie ",
	.topic = "uncore other",
	.long_desc = "Clockticks of the mesh to PCI (M2P)",
	.pmu = "uncore_m2pcie",
	.perpkg = "1",
},
{
	.name = "unc_u_clockticks",
	.event = "event=0xff",
	.desc = "Clockticks in the UBOX using a dedicated 48-bit Fixed Counter. Unit: uncore_ubox ",
	.topic = "uncore other",
	.long_desc = "Clockticks in the UBOX using a dedicated 48-bit Fixed Counter",
	.pmu = "uncore_ubox",
	.perpkg = "1",
},
{
	.name = "unc_p_clockticks",
	.event = "event=0",
	.desc = "Clockticks of the power control unit (PCU). Unit: uncore_pcu ",
	.topic = "uncore power",
	.long_desc = "Clockticks of the power control unit (PCU)",
	.pmu = "uncore_pcu",
	.perpkg = "1",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=200003,event=0x8",
	.desc = "Page walk completed due to a demand load to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads (including SW prefetches) whose address translations missed in all TLB levels and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=200003,event=0x8",
	.desc = "Page walk completed due to a demand load to a 2M or 4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data loads (including SW prefetches) whose address translations missed in all TLB levels and were mapped to 2M or 4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x49",
	.desc = "Page walk completed due to a demand data store to a 2M or 4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to demand data stores whose address translations missed in the TLB and were mapped to 2M or 4M pages.  The page walks can end with or without a page fault",
},
{
	.name = "itlb.fills",
	.event = "umask=0x4,period=200003,event=0x81",
	.desc = "Counts the number of times there was an ITLB miss and a new translation was filled into the ITLB",
	.topic = "virtual memory",
	.long_desc = "Counts the number of times the machine was unable to find a translation in the Instruction Translation Lookaside Buffer (ITLB) and new translation was filled into the ITLB.  The event is speculative in nature, but will not count translations (page walks) that are begun and not finished, or translations that are finished but not filled into the ITLB",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x85",
	.desc = "Page walk completed due to an instruction fetch in a 4K page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to instruction fetches whose address translations missed in the TLB and were mapped to 4K pages.  The page walks can end with or without a page fault",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x85",
	.desc = "Page walk completed due to an instruction fetch in a 2M or 4M page",
	.topic = "virtual memory",
	.long_desc = "Counts page walks completed due to instruction fetches whose address translations missed in the TLB and were mapped to 2M or 4M pages.  The page walks can end with or without a page fault",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_westmereep_dp[] = {
{
	.name = "cache_lock_cycles.l1d",
	.event = "umask=0x2,period=2000000,event=0x63",
	.desc = "Cycles L1D locked",
	.topic = "cache",
},
{
	.name = "cache_lock_cycles.l1d_l2",
	.event = "umask=0x1,period=2000000,event=0x63",
	.desc = "Cycles L1D and L2 locked",
	.topic = "cache",
},
{
	.name = "l1d.m_evict",
	.event = "umask=0x4,period=2000000,event=0x51",
	.desc = "L1D cache lines replaced in M state",
	.topic = "cache",
},
{
	.name = "l1d.m_repl",
	.event = "umask=0x2,period=2000000,event=0x51",
	.desc = "L1D cache lines allocated in the M state",
	.topic = "cache",
},
{
	.name = "l1d.m_snoop_evict",
	.event = "umask=0x8,period=2000000,event=0x51",
	.desc = "L1D snoop eviction of cache lines in M state",
	.topic = "cache",
},
{
	.name = "l1d.repl",
	.event = "umask=0x1,period=2000000,event=0x51",
	.desc = "L1 data cache lines allocated",
	.topic = "cache",
},
{
	.name = "l1d_cache_prefetch_lock_fb_hit",
	.event = "umask=0x1,period=2000000,event=0x52",
	.desc = "L1D prefetch load lock accepted in fill buffer",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.miss",
	.event = "umask=0x2,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch misses",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.requests",
	.event = "umask=0x1,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.triggers",
	.event = "umask=0x4,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests triggered",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.e_state",
	.event = "umask=0x4,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in E state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.i_state",
	.event = "umask=0x1,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in I state (misses)",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.m_state",
	.event = "umask=0x8,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in M state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.mesi",
	.event = "umask=0xf,period=100000,event=0x28",
	.desc = "All L1 writebacks to L2",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.s_state",
	.event = "umask=0x2,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.any",
	.event = "umask=0xff,period=200000,event=0x26",
	.desc = "All L2 data requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.e_state",
	.event = "umask=0x4,period=200000,event=0x26",
	.desc = "L2 data demand loads in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.i_state",
	.event = "umask=0x1,period=200000,event=0x26",
	.desc = "L2 data demand loads in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.m_state",
	.event = "umask=0x8,period=200000,event=0x26",
	.desc = "L2 data demand loads in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.mesi",
	.event = "umask=0xf,period=200000,event=0x26",
	.desc = "L2 data demand requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.s_state",
	.event = "umask=0x2,period=200000,event=0x26",
	.desc = "L2 data demand loads in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.e_state",
	.event = "umask=0x40,period=200000,event=0x26",
	.desc = "L2 data prefetches in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.i_state",
	.event = "umask=0x10,period=200000,event=0x26",
	.desc = "L2 data prefetches in the I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.m_state",
	.event = "umask=0x80,period=200000,event=0x26",
	.desc = "L2 data prefetches in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.mesi",
	.event = "umask=0xf0,period=200000,event=0x26",
	.desc = "All L2 data prefetches",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.s_state",
	.event = "umask=0x20,period=200000,event=0x26",
	.desc = "L2 data prefetches in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.any",
	.event = "umask=0x7,period=100000,event=0xf1",
	.desc = "L2 lines alloacated",
	.topic = "cache",
},
{
	.name = "l2_lines_in.e_state",
	.event = "umask=0x4,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the E state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.s_state",
	.event = "umask=0x2,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_out.any",
	.event = "umask=0xf,period=100000,event=0xf2",
	.desc = "L2 lines evicted",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x1,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x2,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_clean",
	.event = "umask=0x4,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_dirty",
	.event = "umask=0x8,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_hit",
	.event = "umask=0x10,period=200000,event=0x24",
	.desc = "L2 instruction fetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_miss",
	.event = "umask=0x20,period=200000,event=0x24",
	.desc = "L2 instruction fetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetches",
	.event = "umask=0x30,period=200000,event=0x24",
	.desc = "L2 instruction fetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_hit",
	.event = "umask=0x1,period=200000,event=0x24",
	.desc = "L2 load hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_miss",
	.event = "umask=0x2,period=200000,event=0x24",
	.desc = "L2 load misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.loads",
	.event = "umask=0x3,period=200000,event=0x24",
	.desc = "L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0xaa,period=200000,event=0x24",
	.desc = "All L2 misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_hit",
	.event = "umask=0x40,period=200000,event=0x24",
	.desc = "L2 prefetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_miss",
	.event = "umask=0x80,period=200000,event=0x24",
	.desc = "L2 prefetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetches",
	.event = "umask=0xc0,period=200000,event=0x24",
	.desc = "All L2 prefetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200000,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x4,period=200000,event=0x24",
	.desc = "L2 RFO hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x8,period=200000,event=0x24",
	.desc = "L2 RFO misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfos",
	.event = "umask=0xc,period=200000,event=0x24",
	.desc = "L2 RFO requests",
	.topic = "cache",
},
{
	.name = "l2_transactions.any",
	.event = "umask=0x80,period=200000,event=0xf0",
	.desc = "All L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.fill",
	.event = "umask=0x20,period=200000,event=0xf0",
	.desc = "L2 fill transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.ifetch",
	.event = "umask=0x4,period=200000,event=0xf0",
	.desc = "L2 instruction fetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.l1d_wb",
	.event = "umask=0x10,period=200000,event=0xf0",
	.desc = "L1D writeback to L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.load",
	.event = "umask=0x1,period=200000,event=0xf0",
	.desc = "L2 Load transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.prefetch",
	.event = "umask=0x8,period=200000,event=0xf0",
	.desc = "L2 prefetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.rfo",
	.event = "umask=0x2,period=200000,event=0xf0",
	.desc = "L2 RFO transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.wb",
	.event = "umask=0x40,period=200000,event=0xf0",
	.desc = "L2 writeback to LLC transactions",
	.topic = "cache",
},
{
	.name = "l2_write.lock.e_state",
	.event = "umask=0x40,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in E state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.hit",
	.event = "umask=0xe0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.lock.i_state",
	.event = "umask=0x10,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.lock.m_state",
	.event = "umask=0x80,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.mesi",
	.event = "umask=0xf0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.lock.s_state",
	.event = "umask=0x20,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in S state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.hit",
	.event = "umask=0xe,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.i_state",
	.event = "umask=0x1,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.m_state",
	.event = "umask=0x8,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.mesi",
	.event = "umask=0xf,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.s_state",
	.event = "umask=0x2,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in S state",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100000,event=0x2e",
	.desc = "Longest latency cache miss",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=200000,event=0x2e",
	.desc = "Longest latency cache reference",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.loads",
	.event = "umask=0x1,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a load (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.stores",
	.event = "umask=0x2,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a store (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.hit_lfb",
	.event = "umask=0x40,period=200000,event=0xcb",
	.desc = "Retired loads that miss L1D and hit an previously allocated LFB (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l1d_hit",
	.event = "umask=0x1,period=2000000,event=0xcb",
	.desc = "Retired loads that hit the L1 data cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x2,period=200000,event=0xcb",
	.desc = "Retired loads that hit the L2 cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_miss",
	.event = "umask=0x10,period=10000,event=0xcb",
	.desc = "Retired loads that miss the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_unshared_hit",
	.event = "umask=0x4,period=40000,event=0xcb",
	.desc = "Retired loads that hit valid versions in the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.other_core_l2_hit_hitm",
	.event = "umask=0x8,period=40000,event=0xcb",
	.desc = "Retired loads that hit sibling core's L2 in modified or unmodified states (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_requests.any",
	.event = "umask=0x80,period=100000,event=0xb0",
	.desc = "All offcore requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.any.read",
	.event = "umask=0x8,period=100000,event=0xb0",
	.desc = "Offcore read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.any.rfo",
	.event = "umask=0x10,period=100000,event=0xb0",
	.desc = "Offcore RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand.read_code",
	.event = "umask=0x2,period=100000,event=0xb0",
	.desc = "Offcore demand code read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand.read_data",
	.event = "umask=0x1,period=100000,event=0xb0",
	.desc = "Offcore demand data read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand.rfo",
	.event = "umask=0x4,period=100000,event=0xb0",
	.desc = "Offcore demand RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.l1d_writeback",
	.event = "umask=0x40,period=100000,event=0xb0",
	.desc = "Offcore L1 data cache writebacks",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.any.read",
	.event = "umask=0x8,period=2000000,event=0x60",
	.desc = "Outstanding offcore reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.any.read_not_empty",
	.event = "umask=0x8,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore reads busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_code",
	.event = "umask=0x2,period=2000000,event=0x60",
	.desc = "Outstanding offcore demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_code_not_empty",
	.event = "umask=0x2,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore demand code read busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_data",
	.event = "umask=0x1,period=2000000,event=0x60",
	.desc = "Outstanding offcore demand data reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_data_not_empty",
	.event = "umask=0x1,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore demand data read busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.rfo",
	.event = "umask=0x4,period=2000000,event=0x60",
	.desc = "Outstanding offcore demand RFOs",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.rfo_not_empty",
	.event = "umask=0x4,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore demand RFOs busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_sq_full",
	.event = "umask=0x1,period=100000,event=0xb2",
	.desc = "Offcore requests blocked due to Super Queue full",
	.topic = "cache",
},
{
	.name = "sq_misc.lru_hints",
	.event = "umask=0x4,period=2000000,event=0xf4",
	.desc = "Super Queue LRU hints sent to LLC",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=2000000,event=0xf4",
	.desc = "Super Queue lock splits across a cache line",
	.topic = "cache",
},
{
	.name = "store_blocks.at_ret",
	.event = "umask=0x4,period=200000,event=0x6",
	.desc = "Loads delayed with at-Retirement block code",
	.topic = "cache",
},
{
	.name = "store_blocks.l1d_block",
	.event = "umask=0x8,period=200000,event=0x6",
	.desc = "Cacheable loads delayed with L1D block code",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_0",
	.event = "umask=0x10,period=2000000,event=0xb,ldlat=0x0",
	.desc = "Memory instructions retired above 0 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_1024",
	.event = "umask=0x10,period=100,event=0xb,ldlat=0x400",
	.desc = "Memory instructions retired above 1024 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_128",
	.event = "umask=0x10,period=1000,event=0xb,ldlat=0x80",
	.desc = "Memory instructions retired above 128 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16",
	.event = "umask=0x10,period=10000,event=0xb,ldlat=0x10",
	.desc = "Memory instructions retired above 16 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16384",
	.event = "umask=0x10,period=5,event=0xb,ldlat=0x4000",
	.desc = "Memory instructions retired above 16384 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_2048",
	.event = "umask=0x10,period=50,event=0xb,ldlat=0x800",
	.desc = "Memory instructions retired above 2048 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_256",
	.event = "umask=0x10,period=500,event=0xb,ldlat=0x100",
	.desc = "Memory instructions retired above 256 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32",
	.event = "umask=0x10,period=5000,event=0xb,ldlat=0x20",
	.desc = "Memory instructions retired above 32 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32768",
	.event = "umask=0x10,period=3,event=0xb,ldlat=0x8000",
	.desc = "Memory instructions retired above 32768 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4",
	.event = "umask=0x10,period=50000,event=0xb,ldlat=0x4",
	.desc = "Memory instructions retired above 4 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4096",
	.event = "umask=0x10,period=20,event=0xb,ldlat=0x1000",
	.desc = "Memory instructions retired above 4096 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_512",
	.event = "umask=0x10,period=200,event=0xb,ldlat=0x200",
	.desc = "Memory instructions retired above 512 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_64",
	.event = "umask=0x10,period=2000,event=0xb,ldlat=0x40",
	.desc = "Memory instructions retired above 64 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8",
	.event = "umask=0x10,period=20000,event=0xb,ldlat=0x8",
	.desc = "Memory instructions retired above 8 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8192",
	.event = "umask=0x10,period=10,event=0xb,ldlat=0x2000",
	.desc = "Memory instructions retired above 8192 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5011",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f11",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff11",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8011",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x111",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x211",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x411",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x711",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1011",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x811",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5044",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f44",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff44",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8044",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x144",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x244",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x444",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x744",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1044",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x844",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x50ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7fff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xffff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x80ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x10ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5022",
	.desc = "REQUEST = ANY RFO and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f22",
	.desc = "REQUEST = ANY RFO and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff22",
	.desc = "REQUEST = ANY RFO and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8022",
	.desc = "REQUEST = ANY RFO and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x122",
	.desc = "REQUEST = ANY RFO and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x222",
	.desc = "REQUEST = ANY RFO and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x422",
	.desc = "REQUEST = ANY RFO and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x722",
	.desc = "REQUEST = ANY RFO and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1022",
	.desc = "REQUEST = ANY RFO and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x822",
	.desc = "REQUEST = ANY RFO and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5008",
	.desc = "REQUEST = CORE_WB and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f08",
	.desc = "REQUEST = CORE_WB and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff08",
	.desc = "REQUEST = CORE_WB and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8008",
	.desc = "REQUEST = CORE_WB and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x108",
	.desc = "REQUEST = CORE_WB and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x208",
	.desc = "REQUEST = CORE_WB and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x408",
	.desc = "REQUEST = CORE_WB and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x708",
	.desc = "REQUEST = CORE_WB and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1008",
	.desc = "REQUEST = CORE_WB and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x808",
	.desc = "REQUEST = CORE_WB and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5077",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f77",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff77",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8077",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x177",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x277",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x477",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x777",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1077",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x877",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5033",
	.desc = "REQUEST = DATA_IN and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f33",
	.desc = "REQUEST = DATA_IN and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff33",
	.desc = "REQUEST = DATA_IN and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8033",
	.desc = "REQUEST = DATA_IN and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x133",
	.desc = "REQUEST = DATA_IN and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x233",
	.desc = "REQUEST = DATA_IN and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x433",
	.desc = "REQUEST = DATA_IN and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x733",
	.desc = "REQUEST = DATA_IN and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1033",
	.desc = "REQUEST = DATA_IN and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x833",
	.desc = "REQUEST = DATA_IN and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5003",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f03",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff03",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8003",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x103",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x203",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x403",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x703",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1003",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x803",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5001",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f01",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff01",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8001",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x101",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x201",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x401",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x701",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1001",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x801",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5004",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f04",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff04",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8004",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x104",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x204",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x404",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x704",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1004",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x804",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5002",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f02",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff02",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8002",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x102",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x202",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x402",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x702",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1002",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x802",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5080",
	.desc = "REQUEST = OTHER and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f80",
	.desc = "REQUEST = OTHER and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff80",
	.desc = "REQUEST = OTHER and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.other.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8080",
	.desc = "REQUEST = OTHER and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x180",
	.desc = "REQUEST = OTHER and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x280",
	.desc = "REQUEST = OTHER and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x480",
	.desc = "REQUEST = OTHER and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x780",
	.desc = "REQUEST = OTHER and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1080",
	.desc = "REQUEST = OTHER and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x880",
	.desc = "REQUEST = OTHER and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5050",
	.desc = "REQUEST = PF_DATA and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f50",
	.desc = "REQUEST = PF_DATA and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff50",
	.desc = "REQUEST = PF_DATA and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8050",
	.desc = "REQUEST = PF_DATA and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x150",
	.desc = "REQUEST = PF_DATA and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x250",
	.desc = "REQUEST = PF_DATA and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x450",
	.desc = "REQUEST = PF_DATA and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x750",
	.desc = "REQUEST = PF_DATA and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1050",
	.desc = "REQUEST = PF_DATA and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x850",
	.desc = "REQUEST = PF_DATA and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5010",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f10",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff10",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8010",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x110",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x210",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x410",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x710",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1010",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x810",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5040",
	.desc = "REQUEST = PF_RFO and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f40",
	.desc = "REQUEST = PF_RFO and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff40",
	.desc = "REQUEST = PF_RFO and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8040",
	.desc = "REQUEST = PF_RFO and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x140",
	.desc = "REQUEST = PF_RFO and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x240",
	.desc = "REQUEST = PF_RFO and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x440",
	.desc = "REQUEST = PF_RFO and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x740",
	.desc = "REQUEST = PF_RFO and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1040",
	.desc = "REQUEST = PF_RFO and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x840",
	.desc = "REQUEST = PF_RFO and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5020",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f20",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff20",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8020",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x120",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x220",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x420",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x720",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1020",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x820",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.all_local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5070",
	.desc = "REQUEST = PREFETCH and RESPONSE = ALL_LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7f70",
	.desc = "REQUEST = PREFETCH and RESPONSE = ANY_CACHE_DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xff70",
	.desc = "REQUEST = PREFETCH and RESPONSE = ANY_LOCATION",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8070",
	.desc = "REQUEST = PREFETCH and RESPONSE = IO_CSR_MMIO",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x170",
	.desc = "REQUEST = PREFETCH and RESPONSE = LLC_HIT_NO_OTHER_CORE",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x270",
	.desc = "REQUEST = PREFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x470",
	.desc = "REQUEST = PREFETCH and RESPONSE = LLC_HIT_OTHER_CORE_HITM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x770",
	.desc = "REQUEST = PREFETCH and RESPONSE = LOCAL_CACHE",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_dram_and_remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1070",
	.desc = "REQUEST = PREFETCH and RESPONSE = LOCAL_DRAM AND REMOTE_CACHE_HIT",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x870",
	.desc = "REQUEST = PREFETCH and RESPONSE = REMOTE_CACHE_HITM",
	.topic = "cache",
},
{
	.name = "fp_assist.all",
	.event = "umask=0x1,period=20000,event=0xf7",
	.desc = "X87 Floating point assists (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.input",
	.event = "umask=0x4,period=20000,event=0xf7",
	.desc = "X87 Floating poiint assists for invalid input value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.output",
	.event = "umask=0x2,period=20000,event=0xf7",
	.desc = "X87 Floating point assists for invalid output value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.mmx",
	.event = "umask=0x2,period=2000000,event=0x10",
	.desc = "MMX Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_double_precision",
	.event = "umask=0x80,period=2000000,event=0x10",
	.desc = "SSE* FP double precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp",
	.event = "umask=0x4,period=2000000,event=0x10",
	.desc = "SSE and SSE2 FP Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_packed",
	.event = "umask=0x10,period=2000000,event=0x10",
	.desc = "SSE FP packed Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_scalar",
	.event = "umask=0x20,period=2000000,event=0x10",
	.desc = "SSE FP scalar Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_single_precision",
	.event = "umask=0x40,period=2000000,event=0x10",
	.desc = "SSE* FP single precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse2_integer",
	.event = "umask=0x8,period=2000000,event=0x10",
	.desc = "SSE2 integer Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.x87",
	.event = "umask=0x1,period=2000000,event=0x10",
	.desc = "Computational floating-point operations executed",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.any",
	.event = "umask=0x3,period=2000000,event=0xcc",
	.desc = "All Floating Point to and from MMX transitions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_fp",
	.event = "umask=0x1,period=2000000,event=0xcc",
	.desc = "Transitions from MMX to Floating Point instructions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_mmx",
	.event = "umask=0x2,period=2000000,event=0xcc",
	.desc = "Transitions from Floating Point to MMX instructions",
	.topic = "floating point",
},
{
	.name = "simd_int_128.pack",
	.event = "umask=0x4,period=200000,event=0x12",
	.desc = "128 bit SIMD integer pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_arith",
	.event = "umask=0x20,period=200000,event=0x12",
	.desc = "128 bit SIMD integer arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_logical",
	.event = "umask=0x10,period=200000,event=0x12",
	.desc = "128 bit SIMD integer logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_mpy",
	.event = "umask=0x1,period=200000,event=0x12",
	.desc = "128 bit SIMD integer multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_shift",
	.event = "umask=0x2,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.shuffle_move",
	.event = "umask=0x40,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.unpack",
	.event = "umask=0x8,period=200000,event=0x12",
	.desc = "128 bit SIMD integer unpack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.pack",
	.event = "umask=0x4,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_arith",
	.event = "umask=0x20,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_logical",
	.event = "umask=0x10,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_mpy",
	.event = "umask=0x1,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit packed multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_shift",
	.event = "umask=0x2,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.shuffle_move",
	.event = "umask=0x40,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.unpack",
	.event = "umask=0x8,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit unpack operations",
	.topic = "floating point",
},
{
	.name = "macro_insts.decoded",
	.event = "umask=0x1,period=2000000,event=0xd0",
	.desc = "Instructions decoded",
	.topic = "frontend",
},
{
	.name = "macro_insts.fusions_decoded",
	.event = "umask=0x1,period=2000000,event=0xa6",
	.desc = "Macro-fused instructions decoded",
	.topic = "frontend",
},
{
	.name = "two_uop_insts_decoded",
	.event = "umask=0x1,period=2000000,event=0x19",
	.desc = "Two Uop instructions decoded",
	.topic = "frontend",
},
{
	.name = "misalign_mem_ref.store",
	.event = "umask=0x2,period=200000,event=0x5",
	.desc = "Misaligned store references",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3011",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf811",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4011",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2011",
	.desc = "REQUEST = ANY_DATA read and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3044",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf844",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4044",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2044",
	.desc = "REQUEST = ANY IFETCH and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x30ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf8ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x40ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x20ff",
	.desc = "REQUEST = ANY_REQUEST and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3022",
	.desc = "REQUEST = ANY RFO and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf822",
	.desc = "REQUEST = ANY RFO and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4022",
	.desc = "REQUEST = ANY RFO and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2022",
	.desc = "REQUEST = ANY RFO and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3008",
	.desc = "REQUEST = CORE_WB and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf808",
	.desc = "REQUEST = CORE_WB and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4008",
	.desc = "REQUEST = CORE_WB and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2008",
	.desc = "REQUEST = CORE_WB and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3077",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf877",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4077",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2077",
	.desc = "REQUEST = DATA_IFETCH and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3033",
	.desc = "REQUEST = DATA_IN and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf833",
	.desc = "REQUEST = DATA_IN and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4033",
	.desc = "REQUEST = DATA_IN and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2033",
	.desc = "REQUEST = DATA_IN and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3003",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf803",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4003",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2003",
	.desc = "REQUEST = DEMAND_DATA and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3001",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf801",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4001",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2001",
	.desc = "REQUEST = DEMAND_DATA_RD and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3004",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf804",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4004",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2004",
	.desc = "REQUEST = DEMAND_IFETCH and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3002",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf802",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4002",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2002",
	.desc = "REQUEST = DEMAND_RFO and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3080",
	.desc = "REQUEST = OTHER and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf880",
	.desc = "REQUEST = OTHER and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.other.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4080",
	.desc = "REQUEST = OTHER and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2080",
	.desc = "REQUEST = OTHER and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3050",
	.desc = "REQUEST = PF_DATA and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf850",
	.desc = "REQUEST = PF_DATA and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4050",
	.desc = "REQUEST = PF_DATA and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2050",
	.desc = "REQUEST = PF_DATA and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3010",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf810",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4010",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2010",
	.desc = "REQUEST = PF_DATA_RD and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3040",
	.desc = "REQUEST = PF_RFO and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf840",
	.desc = "REQUEST = PF_RFO and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4040",
	.desc = "REQUEST = PF_RFO and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2040",
	.desc = "REQUEST = PF_RFO and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3020",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf820",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4020",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2020",
	.desc = "REQUEST = PF_IFETCH and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_dram_and_remote_fwd",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3070",
	.desc = "REQUEST = PREFETCH and RESPONSE = ANY_DRAM AND REMOTE_FWD",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xf870",
	.desc = "REQUEST = PREFETCH and RESPONSE = ANY_LLC_MISS",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.other_local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4070",
	.desc = "REQUEST = PREFETCH and RESPONSE = OTHER_LOCAL_DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2070",
	.desc = "REQUEST = PREFETCH and RESPONSE = REMOTE_DRAM",
	.topic = "memory",
},
{
	.name = "bpu_clears.early",
	.event = "umask=0x1,period=2000000,event=0xe8",
	.desc = "Early Branch Prediciton Unit clears",
	.topic = "other",
},
{
	.name = "bpu_clears.late",
	.event = "umask=0x2,period=2000000,event=0xe8",
	.desc = "Late Branch Prediction Unit clears",
	.topic = "other",
},
{
	.name = "bpu_missed_call_ret",
	.event = "umask=0x1,period=2000000,event=0xe5",
	.desc = "Branch prediction unit missed call or return",
	.topic = "other",
},
{
	.name = "es_reg_renames",
	.event = "umask=0x1,period=2000000,event=0xd5",
	.desc = "ES segment renames",
	.topic = "other",
},
{
	.name = "io_transactions",
	.event = "umask=0x1,period=2000000,event=0x6c",
	.desc = "I/O transactions",
	.topic = "other",
},
{
	.name = "l1i.cycles_stalled",
	.event = "umask=0x4,period=2000000,event=0x80",
	.desc = "L1I instruction fetch stall cycles",
	.topic = "other",
},
{
	.name = "l1i.hits",
	.event = "umask=0x1,period=2000000,event=0x80",
	.desc = "L1I instruction fetch hits",
	.topic = "other",
},
{
	.name = "l1i.misses",
	.event = "umask=0x2,period=2000000,event=0x80",
	.desc = "L1I instruction fetch misses",
	.topic = "other",
},
{
	.name = "l1i.reads",
	.event = "umask=0x3,period=2000000,event=0x80",
	.desc = "L1I Instruction fetches",
	.topic = "other",
},
{
	.name = "large_itlb.hit",
	.event = "umask=0x1,period=200000,event=0x82",
	.desc = "Large ITLB hit",
	.topic = "other",
},
{
	.name = "load_block.overlap_store",
	.event = "umask=0x2,period=200000,event=0x3",
	.desc = "Loads that partially overlap an earlier store",
	.topic = "other",
},
{
	.name = "load_dispatch.any",
	.event = "umask=0x7,period=2000000,event=0x13",
	.desc = "All loads dispatched",
	.topic = "other",
},
{
	.name = "load_dispatch.mob",
	.event = "umask=0x4,period=2000000,event=0x13",
	.desc = "Loads dispatched from the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs",
	.event = "umask=0x1,period=2000000,event=0x13",
	.desc = "Loads dispatched that bypass the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs_delayed",
	.event = "umask=0x2,period=2000000,event=0x13",
	.desc = "Loads dispatched from stage 305",
	.topic = "other",
},
{
	.name = "partial_address_alias",
	.event = "umask=0x1,period=200000,event=0x7",
	.desc = "False dependencies due to partial address aliasing",
	.topic = "other",
},
{
	.name = "rat_stalls.any",
	.event = "umask=0xf,period=2000000,event=0xd2",
	.desc = "All RAT stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.flags",
	.event = "umask=0x1,period=2000000,event=0xd2",
	.desc = "Flag stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.registers",
	.event = "umask=0x2,period=2000000,event=0xd2",
	.desc = "Partial register stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.rob_read_port",
	.event = "umask=0x4,period=2000000,event=0xd2",
	.desc = "ROB read port stalls cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.scoreboard",
	.event = "umask=0x8,period=2000000,event=0xd2",
	.desc = "Scoreboard stall cycles",
	.topic = "other",
},
{
	.name = "sb_drain.any",
	.event = "umask=0x7,period=200000,event=0x4",
	.desc = "All Store buffer stall cycles",
	.topic = "other",
},
{
	.name = "seg_rename_stalls",
	.event = "umask=0x1,period=2000000,event=0xd4",
	.desc = "Segment rename stall cycles",
	.topic = "other",
},
{
	.name = "snoop_response.hit",
	.event = "umask=0x1,period=100000,event=0xb8",
	.desc = "Thread responded HIT to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hite",
	.event = "umask=0x2,period=100000,event=0xb8",
	.desc = "Thread responded HITE to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hitm",
	.event = "umask=0x4,period=100000,event=0xb8",
	.desc = "Thread responded HITM to snoop",
	.topic = "other",
},
{
	.name = "snoopq_requests.code",
	.event = "umask=0x4,period=100000,event=0xb4",
	.desc = "Snoop code requests",
	.topic = "other",
},
{
	.name = "snoopq_requests.data",
	.event = "umask=0x1,period=100000,event=0xb4",
	.desc = "Snoop data requests",
	.topic = "other",
},
{
	.name = "snoopq_requests.invalidate",
	.event = "umask=0x2,period=100000,event=0xb4",
	.desc = "Snoop invalidate requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.code",
	.event = "umask=0x4,period=2000000,event=0xb3",
	.desc = "Outstanding snoop code requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.code_not_empty",
	.event = "umask=0x4,period=2000000,cmask=1,event=0xb3",
	.desc = "Cycles snoop code requests queued",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.data",
	.event = "umask=0x1,period=2000000,event=0xb3",
	.desc = "Outstanding snoop data requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.data_not_empty",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xb3",
	.desc = "Cycles snoop data requests queued",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.invalidate",
	.event = "umask=0x2,period=2000000,event=0xb3",
	.desc = "Outstanding snoop invalidate requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.invalidate_not_empty",
	.event = "umask=0x2,period=2000000,cmask=1,event=0xb3",
	.desc = "Cycles snoop invalidate requests queued",
	.topic = "other",
},
{
	.name = "sq_full_stall_cycles",
	.event = "umask=0x1,period=2000000,event=0xf6",
	.desc = "Super Queue full stall cycles",
	.topic = "other",
},
{
	.name = "arith.cycles_div_busy",
	.event = "umask=0x1,period=2000000,event=0x14",
	.desc = "Cycles the divider is busy",
	.topic = "pipeline",
},
{
	.name = "arith.div",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,edge=1,event=0x14",
	.desc = "Divide Operations executed",
	.topic = "pipeline",
},
{
	.name = "arith.mul",
	.event = "umask=0x2,period=2000000,event=0x14",
	.desc = "Multiply operations executed",
	.topic = "pipeline",
},
{
	.name = "baclear.bad_target",
	.event = "umask=0x2,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted with bad target address",
	.topic = "pipeline",
},
{
	.name = "baclear.clear",
	.event = "umask=0x1,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted, regardless of cause ",
	.topic = "pipeline",
},
{
	.name = "baclear_force_iq",
	.event = "umask=0x1,period=2000000,event=0xa7",
	.desc = "Instruction queue forced BACLEAR",
	.topic = "pipeline",
},
{
	.name = "br_inst_decoded",
	.event = "umask=0x1,period=2000000,event=0xe0",
	.desc = "Branch instructions decoded",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.any",
	.event = "umask=0x7f,period=200000,event=0x88",
	.desc = "Branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.cond",
	.event = "umask=0x1,period=200000,event=0x88",
	.desc = "Conditional branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct",
	.event = "umask=0x2,period=200000,event=0x88",
	.desc = "Unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct_near_call",
	.event = "umask=0x10,period=20000,event=0x88",
	.desc = "Unconditional call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_near_call",
	.event = "umask=0x20,period=20000,event=0x88",
	.desc = "Indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_non_call",
	.event = "umask=0x4,period=20000,event=0x88",
	.desc = "Indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.near_calls",
	.event = "umask=0x30,period=20000,event=0x88",
	.desc = "Call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.non_calls",
	.event = "umask=0x7,period=200000,event=0x88",
	.desc = "All non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.return_near",
	.event = "umask=0x8,period=20000,event=0x88",
	.desc = "Indirect return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken",
	.event = "umask=0x40,period=200000,event=0x88",
	.desc = "Taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x4,period=200000,event=0xc4",
	.desc = "Retired branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=200000,event=0xc4",
	.desc = "Retired conditional branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=20000,event=0xc4",
	.desc = "Retired near call instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.any",
	.event = "umask=0x7f,period=20000,event=0x89",
	.desc = "Mispredicted branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.cond",
	.event = "umask=0x1,period=20000,event=0x89",
	.desc = "Mispredicted conditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct",
	.event = "umask=0x2,period=20000,event=0x89",
	.desc = "Mispredicted unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct_near_call",
	.event = "umask=0x10,period=2000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_near_call",
	.event = "umask=0x20,period=2000,event=0x89",
	.desc = "Mispredicted indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_non_call",
	.event = "umask=0x4,period=2000,event=0x89",
	.desc = "Mispredicted indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.near_calls",
	.event = "umask=0x30,period=2000,event=0x89",
	.desc = "Mispredicted call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.non_calls",
	.event = "umask=0x7,period=20000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.return_near",
	.event = "umask=0x8,period=2000,event=0x89",
	.desc = "Mispredicted return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken",
	.event = "umask=0x40,period=20000,event=0x89",
	.desc = "Mispredicted taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x4,period=20000,event=0xc5",
	.desc = "Mispredicted retired branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=20000,event=0xc5",
	.desc = "Mispredicted conditional retired branches (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=2000,event=0xc5",
	.desc = "Mispredicted near retired calls (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Reference cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_p",
	.event = "umask=0x1,period=100000,event=0x3c",
	.desc = "Reference base clock (133 Mhz) cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000000,event=0x3c",
	.desc = "Cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.total_cycles",
	.event = "inv=1,umask=0x0,period=2000000,cmask=2,event=0x3c",
	.desc = "Total CPU cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.any",
	.event = "umask=0xf,period=2000000,event=0x87",
	.desc = "Any Instruction Length Decoder stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000000,event=0x87",
	.desc = "Instruction Queue full stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000000,event=0x87",
	.desc = "Length Change Prefix stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.mru",
	.event = "umask=0x2,period=2000000,event=0x87",
	.desc = "Stall cycles due to BPU MRU bypass",
	.topic = "pipeline",
},
{
	.name = "ild_stall.regen",
	.event = "umask=0x8,period=2000000,event=0x87",
	.desc = "Regen stall cycles",
	.topic = "pipeline",
},
{
	.name = "inst_decoded.dec0",
	.event = "umask=0x1,period=2000000,event=0x18",
	.desc = "Instructions that must be decoded by decoder 0",
	.topic = "pipeline",
},
{
	.name = "inst_queue_write_cycles",
	.event = "umask=0x1,period=2000000,event=0x1e",
	.desc = "Cycles instructions are written to the instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_queue_writes",
	.event = "umask=0x1,period=2000000,event=0x17",
	.desc = "Instructions written to instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (Programmable counter and Precise Event) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.mmx",
	.event = "umask=0x4,period=2000000,event=0xc0",
	.desc = "Retired MMX instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000000,event=0xc0",
	.desc = "Retired floating-point operations (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre",
	.event = "umask=0x1,period=200000,event=0x4c",
	.desc = "Load operations conflicting with software prefetches",
	.topic = "pipeline",
},
{
	.name = "lsd.active",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles when uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.inactive",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles no uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd_overflow",
	.event = "umask=0x1,period=2000000,event=0x20",
	.desc = "Loops that can't stream from the instruction queue",
	.topic = "pipeline",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=20000,event=0xc3",
	.desc = "Cycles machine clear asserted",
	.topic = "pipeline",
},
{
	.name = "machine_clears.mem_order",
	.event = "umask=0x2,period=20000,event=0xc3",
	.desc = "Execution pipeline restart due to Memory ordering conflicts",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=20000,event=0xc3",
	.desc = "Self-Modifying Code detected",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000000,event=0xa2",
	.desc = "Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.fpcw",
	.event = "umask=0x20,period=2000000,event=0xa2",
	.desc = "FPU control word write stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.load",
	.event = "umask=0x2,period=2000000,event=0xa2",
	.desc = "Load buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.mxcsr",
	.event = "umask=0x40,period=2000000,event=0xa2",
	.desc = "MXCSR rename stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.other",
	.event = "umask=0x80,period=2000000,event=0xa2",
	.desc = "Other Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rob_full",
	.event = "umask=0x10,period=2000000,event=0xa2",
	.desc = "ROB full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rs_full",
	.event = "umask=0x4,period=2000000,event=0xa2",
	.desc = "Reservation Station full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.store",
	.event = "umask=0x8,period=2000000,event=0xa2",
	.desc = "Store buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_double",
	.event = "umask=0x4,period=200000,event=0xc7",
	.desc = "SIMD Packed-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_single",
	.event = "umask=0x1,period=200000,event=0xc7",
	.desc = "SIMD Packed-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_double",
	.event = "umask=0x8,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_single",
	.event = "umask=0x2,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.vector_integer",
	.event = "umask=0x10,period=200000,event=0xc7",
	.desc = "SIMD Vector Integer Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uop_unfusion",
	.event = "umask=0x1,period=2000000,event=0xdb",
	.desc = "Uop unfusions due to FP exceptions",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_folding",
	.event = "umask=0x4,period=2000000,event=0xd1",
	.desc = "Stack pointer instructions decoded",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_sync",
	.event = "umask=0x8,period=2000000,event=0xd1",
	.desc = "Stack pointer sync operations",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.ms_cycles_active",
	.event = "umask=0x2,period=2000000,cmask=1,event=0xd1",
	.desc = "Uops decoded by Microcode Sequencer",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xd1",
	.desc = "Cycles no Uops are decoded",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles",
	.event = "umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles_no_port5",
	.event = "umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count",
	.event = "inv=1,umask=0x3f,any=1,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count_no_port5",
	.event = "inv=1,umask=0x1f,any=1,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles",
	.event = "inv=1,umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles_no_port5",
	.event = "inv=1,umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port0",
	.event = "umask=0x1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015",
	.event = "umask=0x40,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015_stall_cycles",
	.event = "inv=1,umask=0x40,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port1",
	.event = "umask=0x2,period=2000000,event=0xb1",
	.desc = "Uops executed on port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port2_core",
	.event = "umask=0x4,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 2 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port234_core",
	.event = "umask=0x80,any=1,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 2, 3 or 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port3_core",
	.event = "umask=0x8,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 3 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port4_core",
	.event = "umask=0x10,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port5",
	.event = "umask=0x20,period=2000000,event=0xb1",
	.desc = "Uops executed on port 5",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000000,event=0xe",
	.desc = "Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued on any thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.cycles_all_threads",
	.event = "umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles Uops were issued on either thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.fused",
	.event = "umask=0x2,period=2000000,event=0xe",
	.desc = "Fused Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued",
	.topic = "pipeline",
},
{
	.name = "uops_retired.active_cycles",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are being retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.any",
	.event = "umask=0x1,period=2000000,event=0xc2",
	.desc = "Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.macro_fused",
	.event = "umask=0x4,period=2000000,event=0xc2",
	.desc = "Macro-fused Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000000,event=0xc2",
	.desc = "Retirement slots used (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are not retiring (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc2",
	.desc = "Total cycles using precise uop retired event (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles_ps",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "dtlb_load_misses.any",
	.event = "umask=0x1,period=200000,event=0x8",
	.desc = "DTLB load misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.large_walk_completed",
	.event = "umask=0x80,period=200000,event=0x8",
	.desc = "DTLB load miss large page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.pde_miss",
	.event = "umask=0x20,period=200000,event=0x8",
	.desc = "DTLB load miss caused by low part of address",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x10,period=2000000,event=0x8",
	.desc = "DTLB second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x8",
	.desc = "DTLB load miss page walks complete",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_cycles",
	.event = "umask=0x4,period=200000,event=0x8",
	.desc = "DTLB load miss page walk cycles",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x49",
	.desc = "DTLB misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.large_walk_completed",
	.event = "umask=0x80,period=200000,event=0x49",
	.desc = "DTLB miss large page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.pde_miss",
	.event = "umask=0x20,period=200000,event=0x49",
	.desc = "DTLB misses casued by low part of address",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.stlb_hit",
	.event = "umask=0x10,period=200000,event=0x49",
	.desc = "DTLB first level misses but second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x49",
	.desc = "DTLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.walk_cycles",
	.event = "umask=0x4,period=2000000,event=0x49",
	.desc = "DTLB miss page walk cycles",
	.topic = "virtual memory",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000000,event=0x4f",
	.desc = "Extended Page Table walk cycles",
	.topic = "virtual memory",
},
{
	.name = "itlb_flush",
	.event = "umask=0x1,period=2000000,event=0xae",
	.desc = "ITLB flushes",
	.topic = "virtual memory",
},
{
	.name = "itlb_miss_retired",
	.event = "umask=0x20,period=200000,event=0xc8",
	.desc = "Retired instructions that missed the ITLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x85",
	.desc = "ITLB miss",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.large_walk_completed",
	.event = "umask=0x80,period=200000,event=0x85",
	.desc = "ITLB miss large page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x85",
	.desc = "ITLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_cycles",
	.event = "umask=0x4,period=2000000,event=0x85",
	.desc = "ITLB miss page walk cycles",
	.topic = "virtual memory",
},
{
	.name = "mem_load_retired.dtlb_miss",
	.event = "umask=0x80,period=200000,event=0xcb",
	.desc = "Retired loads that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "mem_store_retired.dtlb_miss",
	.event = "umask=0x1,period=200000,event=0xc",
	.desc = "Retired stores that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_westmereep_sp[] = {
{
	.name = "cache_lock_cycles.l1d",
	.event = "umask=0x2,period=2000000,event=0x63",
	.desc = "Cycles L1D locked",
	.topic = "cache",
},
{
	.name = "cache_lock_cycles.l1d_l2",
	.event = "umask=0x1,period=2000000,event=0x63",
	.desc = "Cycles L1D and L2 locked",
	.topic = "cache",
},
{
	.name = "l1d.m_evict",
	.event = "umask=0x4,period=2000000,event=0x51",
	.desc = "L1D cache lines replaced in M state",
	.topic = "cache",
},
{
	.name = "l1d.m_repl",
	.event = "umask=0x2,period=2000000,event=0x51",
	.desc = "L1D cache lines allocated in the M state",
	.topic = "cache",
},
{
	.name = "l1d.m_snoop_evict",
	.event = "umask=0x8,period=2000000,event=0x51",
	.desc = "L1D snoop eviction of cache lines in M state",
	.topic = "cache",
},
{
	.name = "l1d.repl",
	.event = "umask=0x1,period=2000000,event=0x51",
	.desc = "L1 data cache lines allocated",
	.topic = "cache",
},
{
	.name = "l1d_cache_prefetch_lock_fb_hit",
	.event = "umask=0x1,period=2000000,event=0x52",
	.desc = "L1D prefetch load lock accepted in fill buffer",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.miss",
	.event = "umask=0x2,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch misses",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.requests",
	.event = "umask=0x1,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.triggers",
	.event = "umask=0x4,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests triggered",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.e_state",
	.event = "umask=0x4,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in E state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.i_state",
	.event = "umask=0x1,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in I state (misses)",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.m_state",
	.event = "umask=0x8,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in M state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.mesi",
	.event = "umask=0xf,period=100000,event=0x28",
	.desc = "All L1 writebacks to L2",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.s_state",
	.event = "umask=0x2,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.any",
	.event = "umask=0xff,period=200000,event=0x26",
	.desc = "All L2 data requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.e_state",
	.event = "umask=0x4,period=200000,event=0x26",
	.desc = "L2 data demand loads in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.i_state",
	.event = "umask=0x1,period=200000,event=0x26",
	.desc = "L2 data demand loads in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.m_state",
	.event = "umask=0x8,period=200000,event=0x26",
	.desc = "L2 data demand loads in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.mesi",
	.event = "umask=0xf,period=200000,event=0x26",
	.desc = "L2 data demand requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.s_state",
	.event = "umask=0x2,period=200000,event=0x26",
	.desc = "L2 data demand loads in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.e_state",
	.event = "umask=0x40,period=200000,event=0x26",
	.desc = "L2 data prefetches in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.i_state",
	.event = "umask=0x10,period=200000,event=0x26",
	.desc = "L2 data prefetches in the I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.m_state",
	.event = "umask=0x80,period=200000,event=0x26",
	.desc = "L2 data prefetches in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.mesi",
	.event = "umask=0xf0,period=200000,event=0x26",
	.desc = "All L2 data prefetches",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.s_state",
	.event = "umask=0x20,period=200000,event=0x26",
	.desc = "L2 data prefetches in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.any",
	.event = "umask=0x7,period=100000,event=0xf1",
	.desc = "L2 lines alloacated",
	.topic = "cache",
},
{
	.name = "l2_lines_in.e_state",
	.event = "umask=0x4,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the E state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.s_state",
	.event = "umask=0x2,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_out.any",
	.event = "umask=0xf,period=100000,event=0xf2",
	.desc = "L2 lines evicted",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x1,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x2,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_clean",
	.event = "umask=0x4,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_dirty",
	.event = "umask=0x8,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_hit",
	.event = "umask=0x10,period=200000,event=0x24",
	.desc = "L2 instruction fetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_miss",
	.event = "umask=0x20,period=200000,event=0x24",
	.desc = "L2 instruction fetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetches",
	.event = "umask=0x30,period=200000,event=0x24",
	.desc = "L2 instruction fetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_hit",
	.event = "umask=0x1,period=200000,event=0x24",
	.desc = "L2 load hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_miss",
	.event = "umask=0x2,period=200000,event=0x24",
	.desc = "L2 load misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.loads",
	.event = "umask=0x3,period=200000,event=0x24",
	.desc = "L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0xaa,period=200000,event=0x24",
	.desc = "All L2 misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_hit",
	.event = "umask=0x40,period=200000,event=0x24",
	.desc = "L2 prefetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_miss",
	.event = "umask=0x80,period=200000,event=0x24",
	.desc = "L2 prefetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetches",
	.event = "umask=0xc0,period=200000,event=0x24",
	.desc = "All L2 prefetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200000,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x4,period=200000,event=0x24",
	.desc = "L2 RFO hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x8,period=200000,event=0x24",
	.desc = "L2 RFO misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfos",
	.event = "umask=0xc,period=200000,event=0x24",
	.desc = "L2 RFO requests",
	.topic = "cache",
},
{
	.name = "l2_transactions.any",
	.event = "umask=0x80,period=200000,event=0xf0",
	.desc = "All L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.fill",
	.event = "umask=0x20,period=200000,event=0xf0",
	.desc = "L2 fill transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.ifetch",
	.event = "umask=0x4,period=200000,event=0xf0",
	.desc = "L2 instruction fetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.l1d_wb",
	.event = "umask=0x10,period=200000,event=0xf0",
	.desc = "L1D writeback to L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.load",
	.event = "umask=0x1,period=200000,event=0xf0",
	.desc = "L2 Load transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.prefetch",
	.event = "umask=0x8,period=200000,event=0xf0",
	.desc = "L2 prefetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.rfo",
	.event = "umask=0x2,period=200000,event=0xf0",
	.desc = "L2 RFO transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.wb",
	.event = "umask=0x40,period=200000,event=0xf0",
	.desc = "L2 writeback to LLC transactions",
	.topic = "cache",
},
{
	.name = "l2_write.lock.e_state",
	.event = "umask=0x40,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in E state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.hit",
	.event = "umask=0xe0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.lock.i_state",
	.event = "umask=0x10,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.lock.m_state",
	.event = "umask=0x80,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.mesi",
	.event = "umask=0xf0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.lock.s_state",
	.event = "umask=0x20,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in S state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.hit",
	.event = "umask=0xe,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.i_state",
	.event = "umask=0x1,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.m_state",
	.event = "umask=0x8,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.mesi",
	.event = "umask=0xf,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.s_state",
	.event = "umask=0x2,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in S state",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100000,event=0x2e",
	.desc = "Longest latency cache miss",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=200000,event=0x2e",
	.desc = "Longest latency cache reference",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.loads",
	.event = "umask=0x1,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a load (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.stores",
	.event = "umask=0x2,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a store (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.hit_lfb",
	.event = "umask=0x40,period=200000,event=0xcb",
	.desc = "Retired loads that miss L1D and hit an previously allocated LFB (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l1d_hit",
	.event = "umask=0x1,period=2000000,event=0xcb",
	.desc = "Retired loads that hit the L1 data cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x2,period=200000,event=0xcb",
	.desc = "Retired loads that hit the L2 cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_miss",
	.event = "umask=0x10,period=10000,event=0xcb",
	.desc = "Retired loads that miss the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_unshared_hit",
	.event = "umask=0x4,period=40000,event=0xcb",
	.desc = "Retired loads that hit valid versions in the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.other_core_l2_hit_hitm",
	.event = "umask=0x8,period=40000,event=0xcb",
	.desc = "Retired loads that hit sibling core's L2 in modified or unmodified states (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.local_dram",
	.event = "umask=0x10,period=10000,event=0xf",
	.desc = "Load instructions retired with a data source of local DRAM or locally homed remote hitm (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.other_core_l2_hitm",
	.event = "umask=0x2,period=40000,event=0xf",
	.desc = "Load instructions retired that HIT modified data in sibling core (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.remote_cache_local_home_hit",
	.event = "umask=0x8,period=20000,event=0xf",
	.desc = "Load instructions retired remote cache HIT data source (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.remote_dram",
	.event = "umask=0x20,period=10000,event=0xf",
	.desc = "Load instructions retired remote DRAM and remote home-remote cache HITM (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.uncacheable",
	.event = "umask=0x80,period=4000,event=0xf",
	.desc = "Load instructions retired IO (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_requests.any",
	.event = "umask=0x80,period=100000,event=0xb0",
	.desc = "All offcore requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.any.read",
	.event = "umask=0x8,period=100000,event=0xb0",
	.desc = "Offcore read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.any.rfo",
	.event = "umask=0x10,period=100000,event=0xb0",
	.desc = "Offcore RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand.read_code",
	.event = "umask=0x2,period=100000,event=0xb0",
	.desc = "Offcore demand code read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand.read_data",
	.event = "umask=0x1,period=100000,event=0xb0",
	.desc = "Offcore demand data read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand.rfo",
	.event = "umask=0x4,period=100000,event=0xb0",
	.desc = "Offcore demand RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.l1d_writeback",
	.event = "umask=0x40,period=100000,event=0xb0",
	.desc = "Offcore L1 data cache writebacks",
	.topic = "cache",
},
{
	.name = "offcore_requests.uncached_mem",
	.event = "umask=0x20,period=100000,event=0xb0",
	.desc = "Offcore uncached memory accesses",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.any.read",
	.event = "umask=0x8,period=2000000,event=0x60",
	.desc = "Outstanding offcore reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.any.read_not_empty",
	.event = "umask=0x8,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore reads busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_code",
	.event = "umask=0x2,period=2000000,event=0x60",
	.desc = "Outstanding offcore demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_code_not_empty",
	.event = "umask=0x2,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore demand code read busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_data",
	.event = "umask=0x1,period=2000000,event=0x60",
	.desc = "Outstanding offcore demand data reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_data_not_empty",
	.event = "umask=0x1,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore demand data read busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.rfo",
	.event = "umask=0x4,period=2000000,event=0x60",
	.desc = "Outstanding offcore demand RFOs",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.rfo_not_empty",
	.event = "umask=0x4,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore demand RFOs busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_sq_full",
	.event = "umask=0x1,period=100000,event=0xb2",
	.desc = "Offcore requests blocked due to Super Queue full",
	.topic = "cache",
},
{
	.name = "sq_misc.lru_hints",
	.event = "umask=0x4,period=2000000,event=0xf4",
	.desc = "Super Queue LRU hints sent to LLC",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=2000000,event=0xf4",
	.desc = "Super Queue lock splits across a cache line",
	.topic = "cache",
},
{
	.name = "store_blocks.at_ret",
	.event = "umask=0x4,period=200000,event=0x6",
	.desc = "Loads delayed with at-Retirement block code",
	.topic = "cache",
},
{
	.name = "store_blocks.l1d_block",
	.event = "umask=0x8,period=200000,event=0x6",
	.desc = "Cacheable loads delayed with L1D block code",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_0",
	.event = "umask=0x10,period=2000000,event=0xb,ldlat=0x0",
	.desc = "Memory instructions retired above 0 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_1024",
	.event = "umask=0x10,period=100,event=0xb,ldlat=0x400",
	.desc = "Memory instructions retired above 1024 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_128",
	.event = "umask=0x10,period=1000,event=0xb,ldlat=0x80",
	.desc = "Memory instructions retired above 128 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16",
	.event = "umask=0x10,period=10000,event=0xb,ldlat=0x10",
	.desc = "Memory instructions retired above 16 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16384",
	.event = "umask=0x10,period=5,event=0xb,ldlat=0x4000",
	.desc = "Memory instructions retired above 16384 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_2048",
	.event = "umask=0x10,period=50,event=0xb,ldlat=0x800",
	.desc = "Memory instructions retired above 2048 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_256",
	.event = "umask=0x10,period=500,event=0xb,ldlat=0x100",
	.desc = "Memory instructions retired above 256 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32",
	.event = "umask=0x10,period=5000,event=0xb,ldlat=0x20",
	.desc = "Memory instructions retired above 32 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32768",
	.event = "umask=0x10,period=3,event=0xb,ldlat=0x8000",
	.desc = "Memory instructions retired above 32768 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4",
	.event = "umask=0x10,period=50000,event=0xb,ldlat=0x4",
	.desc = "Memory instructions retired above 4 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4096",
	.event = "umask=0x10,period=20,event=0xb,ldlat=0x1000",
	.desc = "Memory instructions retired above 4096 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_512",
	.event = "umask=0x10,period=200,event=0xb,ldlat=0x200",
	.desc = "Memory instructions retired above 512 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_64",
	.event = "umask=0x10,period=2000,event=0xb,ldlat=0x40",
	.desc = "Memory instructions retired above 64 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8",
	.event = "umask=0x10,period=20000,event=0xb,ldlat=0x8",
	.desc = "Memory instructions retired above 8 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8192",
	.event = "umask=0x10,period=10,event=0xb,ldlat=0x2000",
	.desc = "Memory instructions retired above 8192 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F11",
	.desc = "Offcore data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF11",
	.desc = "All offcore data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8011",
	.desc = "Offcore data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x111",
	.desc = "Offcore data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x211",
	.desc = "Offcore data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x411",
	.desc = "Offcore data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x711",
	.desc = "Offcore data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2711",
	.desc = "Offcore data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1811",
	.desc = "Offcore data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5811",
	.desc = "Offcore data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1011",
	.desc = "Offcore data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x811",
	.desc = "Offcore data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F44",
	.desc = "Offcore code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF44",
	.desc = "All offcore code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8044",
	.desc = "Offcore code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x144",
	.desc = "Offcore code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x244",
	.desc = "Offcore code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x444",
	.desc = "Offcore code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x744",
	.desc = "Offcore code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2744",
	.desc = "Offcore code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1844",
	.desc = "Offcore code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5844",
	.desc = "Offcore code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1044",
	.desc = "Offcore code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x844",
	.desc = "Offcore code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7FFF",
	.desc = "Offcore requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFFFF",
	.desc = "All offcore requests",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x80FF",
	.desc = "Offcore requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1FF",
	.desc = "Offcore requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2FF",
	.desc = "Offcore requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4FF",
	.desc = "Offcore requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7FF",
	.desc = "Offcore requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x27FF",
	.desc = "Offcore requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x18FF",
	.desc = "Offcore requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x58FF",
	.desc = "Offcore requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x10FF",
	.desc = "Offcore requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8FF",
	.desc = "Offcore requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F22",
	.desc = "Offcore RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF22",
	.desc = "All offcore RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8022",
	.desc = "Offcore RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x122",
	.desc = "Offcore RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x222",
	.desc = "Offcore RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x422",
	.desc = "Offcore RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x722",
	.desc = "Offcore RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2722",
	.desc = "Offcore RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1822",
	.desc = "Offcore RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5822",
	.desc = "Offcore RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1022",
	.desc = "Offcore RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x822",
	.desc = "Offcore RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F08",
	.desc = "Offcore writebacks to any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF08",
	.desc = "All offcore writebacks",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8008",
	.desc = "Offcore writebacks to the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x108",
	.desc = "Offcore writebacks to the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x408",
	.desc = "Offcore writebacks to the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x708",
	.desc = "Offcore writebacks to the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2708",
	.desc = "Offcore writebacks to the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1808",
	.desc = "Offcore writebacks to a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5808",
	.desc = "Offcore writebacks to a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1008",
	.desc = "Offcore writebacks that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x808",
	.desc = "Offcore writebacks that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F77",
	.desc = "Offcore code or data read requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF77",
	.desc = "All offcore code or data read requests",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8077",
	.desc = "Offcore code or data read requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x177",
	.desc = "Offcore code or data read requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x277",
	.desc = "Offcore code or data read requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x477",
	.desc = "Offcore code or data read requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x777",
	.desc = "Offcore code or data read requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2777",
	.desc = "Offcore code or data read requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1877",
	.desc = "Offcore code or data read requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5877",
	.desc = "Offcore code or data read requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1077",
	.desc = "Offcore code or data read requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x877",
	.desc = "Offcore code or data read requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F33",
	.desc = "Offcore request = all data, response = any cache_dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF33",
	.desc = "Offcore request = all data, response = any location",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8033",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x133",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x233",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x433",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x733",
	.desc = "Offcore request = all data, response = local cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2733",
	.desc = "Offcore request = all data, response = local cache or dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1833",
	.desc = "Offcore request = all data, response = remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5833",
	.desc = "Offcore request = all data, response = remote cache or dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1033",
	.desc = "Offcore data reads, RFO's and prefetches that HIT in a remote cache ",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x833",
	.desc = "Offcore data reads, RFO's and prefetches that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F03",
	.desc = "Offcore demand data requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF03",
	.desc = "All offcore demand data requests",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8003",
	.desc = "Offcore demand data requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x103",
	.desc = "Offcore demand data requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x203",
	.desc = "Offcore demand data requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x403",
	.desc = "Offcore demand data requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x703",
	.desc = "Offcore demand data requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2703",
	.desc = "Offcore demand data requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1803",
	.desc = "Offcore demand data requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5803",
	.desc = "Offcore demand data requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1003",
	.desc = "Offcore demand data requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x803",
	.desc = "Offcore demand data requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F01",
	.desc = "Offcore demand data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF01",
	.desc = "All offcore demand data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8001",
	.desc = "Offcore demand data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x101",
	.desc = "Offcore demand data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x201",
	.desc = "Offcore demand data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x401",
	.desc = "Offcore demand data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x701",
	.desc = "Offcore demand data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2701",
	.desc = "Offcore demand data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1801",
	.desc = "Offcore demand data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5801",
	.desc = "Offcore demand data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1001",
	.desc = "Offcore demand data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x801",
	.desc = "Offcore demand data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F04",
	.desc = "Offcore demand code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF04",
	.desc = "All offcore demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8004",
	.desc = "Offcore demand code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x104",
	.desc = "Offcore demand code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x204",
	.desc = "Offcore demand code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x404",
	.desc = "Offcore demand code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x704",
	.desc = "Offcore demand code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2704",
	.desc = "Offcore demand code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1804",
	.desc = "Offcore demand code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5804",
	.desc = "Offcore demand code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1004",
	.desc = "Offcore demand code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x804",
	.desc = "Offcore demand code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F02",
	.desc = "Offcore demand RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF02",
	.desc = "All offcore demand RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8002",
	.desc = "Offcore demand RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x102",
	.desc = "Offcore demand RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x202",
	.desc = "Offcore demand RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x402",
	.desc = "Offcore demand RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x702",
	.desc = "Offcore demand RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2702",
	.desc = "Offcore demand RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1802",
	.desc = "Offcore demand RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5802",
	.desc = "Offcore demand RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1002",
	.desc = "Offcore demand RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x802",
	.desc = "Offcore demand RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F80",
	.desc = "Offcore other requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF80",
	.desc = "All offcore other requests",
	.topic = "cache",
},
{
	.name = "offcore_response.other.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8080",
	.desc = "Offcore other requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x180",
	.desc = "Offcore other requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x280",
	.desc = "Offcore other requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x480",
	.desc = "Offcore other requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x780",
	.desc = "Offcore other requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2780",
	.desc = "Offcore other requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1880",
	.desc = "Offcore other requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5880",
	.desc = "Offcore other requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1080",
	.desc = "Offcore other requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x880",
	.desc = "Offcore other requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F50",
	.desc = "Offcore prefetch data requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF50",
	.desc = "All offcore prefetch data requests",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8050",
	.desc = "Offcore prefetch data requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x150",
	.desc = "Offcore prefetch data requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x250",
	.desc = "Offcore prefetch data requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x450",
	.desc = "Offcore prefetch data requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x750",
	.desc = "Offcore prefetch data requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2750",
	.desc = "Offcore prefetch data requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1850",
	.desc = "Offcore prefetch data requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5850",
	.desc = "Offcore prefetch data requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1050",
	.desc = "Offcore prefetch data requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x850",
	.desc = "Offcore prefetch data requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F10",
	.desc = "Offcore prefetch data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF10",
	.desc = "All offcore prefetch data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8010",
	.desc = "Offcore prefetch data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x110",
	.desc = "Offcore prefetch data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x210",
	.desc = "Offcore prefetch data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x410",
	.desc = "Offcore prefetch data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x710",
	.desc = "Offcore prefetch data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2710",
	.desc = "Offcore prefetch data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1810",
	.desc = "Offcore prefetch data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5810",
	.desc = "Offcore prefetch data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1010",
	.desc = "Offcore prefetch data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x810",
	.desc = "Offcore prefetch data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F40",
	.desc = "Offcore prefetch code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF40",
	.desc = "All offcore prefetch code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8040",
	.desc = "Offcore prefetch code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x140",
	.desc = "Offcore prefetch code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x240",
	.desc = "Offcore prefetch code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x440",
	.desc = "Offcore prefetch code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x740",
	.desc = "Offcore prefetch code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2740",
	.desc = "Offcore prefetch code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1840",
	.desc = "Offcore prefetch code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5840",
	.desc = "Offcore prefetch code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1040",
	.desc = "Offcore prefetch code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x840",
	.desc = "Offcore prefetch code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F20",
	.desc = "Offcore prefetch RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF20",
	.desc = "All offcore prefetch RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8020",
	.desc = "Offcore prefetch RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x120",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x220",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x420",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x720",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2720",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1820",
	.desc = "Offcore prefetch RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5820",
	.desc = "Offcore prefetch RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1020",
	.desc = "Offcore prefetch RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x820",
	.desc = "Offcore prefetch RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F70",
	.desc = "Offcore prefetch requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF70",
	.desc = "All offcore prefetch requests",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8070",
	.desc = "Offcore prefetch requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x170",
	.desc = "Offcore prefetch requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x270",
	.desc = "Offcore prefetch requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x470",
	.desc = "Offcore prefetch requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x770",
	.desc = "Offcore prefetch requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2770",
	.desc = "Offcore prefetch requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1870",
	.desc = "Offcore prefetch requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x5870",
	.desc = "Offcore prefetch requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1070",
	.desc = "Offcore prefetch requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x870",
	.desc = "Offcore prefetch requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "fp_assist.all",
	.event = "umask=0x1,period=20000,event=0xf7",
	.desc = "X87 Floating point assists (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.input",
	.event = "umask=0x4,period=20000,event=0xf7",
	.desc = "X87 Floating poiint assists for invalid input value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.output",
	.event = "umask=0x2,period=20000,event=0xf7",
	.desc = "X87 Floating point assists for invalid output value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.mmx",
	.event = "umask=0x2,period=2000000,event=0x10",
	.desc = "MMX Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_double_precision",
	.event = "umask=0x80,period=2000000,event=0x10",
	.desc = "SSE* FP double precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp",
	.event = "umask=0x4,period=2000000,event=0x10",
	.desc = "SSE and SSE2 FP Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_packed",
	.event = "umask=0x10,period=2000000,event=0x10",
	.desc = "SSE FP packed Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_scalar",
	.event = "umask=0x20,period=2000000,event=0x10",
	.desc = "SSE FP scalar Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_single_precision",
	.event = "umask=0x40,period=2000000,event=0x10",
	.desc = "SSE* FP single precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse2_integer",
	.event = "umask=0x8,period=2000000,event=0x10",
	.desc = "SSE2 integer Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.x87",
	.event = "umask=0x1,period=2000000,event=0x10",
	.desc = "Computational floating-point operations executed",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.any",
	.event = "umask=0x3,period=2000000,event=0xcc",
	.desc = "All Floating Point to and from MMX transitions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_fp",
	.event = "umask=0x1,period=2000000,event=0xcc",
	.desc = "Transitions from MMX to Floating Point instructions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_mmx",
	.event = "umask=0x2,period=2000000,event=0xcc",
	.desc = "Transitions from Floating Point to MMX instructions",
	.topic = "floating point",
},
{
	.name = "simd_int_128.pack",
	.event = "umask=0x4,period=200000,event=0x12",
	.desc = "128 bit SIMD integer pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_arith",
	.event = "umask=0x20,period=200000,event=0x12",
	.desc = "128 bit SIMD integer arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_logical",
	.event = "umask=0x10,period=200000,event=0x12",
	.desc = "128 bit SIMD integer logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_mpy",
	.event = "umask=0x1,period=200000,event=0x12",
	.desc = "128 bit SIMD integer multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_shift",
	.event = "umask=0x2,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.shuffle_move",
	.event = "umask=0x40,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.unpack",
	.event = "umask=0x8,period=200000,event=0x12",
	.desc = "128 bit SIMD integer unpack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.pack",
	.event = "umask=0x4,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_arith",
	.event = "umask=0x20,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_logical",
	.event = "umask=0x10,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_mpy",
	.event = "umask=0x1,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit packed multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_shift",
	.event = "umask=0x2,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.shuffle_move",
	.event = "umask=0x40,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.unpack",
	.event = "umask=0x8,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit unpack operations",
	.topic = "floating point",
},
{
	.name = "macro_insts.decoded",
	.event = "umask=0x1,period=2000000,event=0xd0",
	.desc = "Instructions decoded",
	.topic = "frontend",
},
{
	.name = "macro_insts.fusions_decoded",
	.event = "umask=0x1,period=2000000,event=0xa6",
	.desc = "Macro-fused instructions decoded",
	.topic = "frontend",
},
{
	.name = "two_uop_insts_decoded",
	.event = "umask=0x1,period=2000000,event=0x19",
	.desc = "Two Uop instructions decoded",
	.topic = "frontend",
},
{
	.name = "offcore_response.any_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6011",
	.desc = "Offcore data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF811",
	.desc = "Offcore data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2011",
	.desc = "Offcore data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4011",
	.desc = "Offcore data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6044",
	.desc = "Offcore code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF844",
	.desc = "Offcore code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2044",
	.desc = "Offcore code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4044",
	.desc = "Offcore code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x60FF",
	.desc = "Offcore requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF8FF",
	.desc = "Offcore requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x20FF",
	.desc = "Offcore requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x40FF",
	.desc = "Offcore requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6022",
	.desc = "Offcore RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF822",
	.desc = "Offcore RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2022",
	.desc = "Offcore RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4022",
	.desc = "Offcore RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6008",
	.desc = "Offcore writebacks to any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF808",
	.desc = "Offcore writebacks that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2008",
	.desc = "Offcore writebacks to the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4008",
	.desc = "Offcore writebacks to a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6077",
	.desc = "Offcore code or data read requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF877",
	.desc = "Offcore code or data read requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2077",
	.desc = "Offcore code or data read requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4077",
	.desc = "Offcore code or data read requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6033",
	.desc = "Offcore request = all data, response = any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF833",
	.desc = "Offcore request = all data, response = any LLC miss",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2033",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4033",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6003",
	.desc = "Offcore demand data requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF803",
	.desc = "Offcore demand data requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2003",
	.desc = "Offcore demand data requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4003",
	.desc = "Offcore demand data requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6001",
	.desc = "Offcore demand data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF801",
	.desc = "Offcore demand data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2001",
	.desc = "Offcore demand data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4001",
	.desc = "Offcore demand data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6004",
	.desc = "Offcore demand code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF804",
	.desc = "Offcore demand code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2004",
	.desc = "Offcore demand code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4004",
	.desc = "Offcore demand code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6002",
	.desc = "Offcore demand RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF802",
	.desc = "Offcore demand RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2002",
	.desc = "Offcore demand RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4002",
	.desc = "Offcore demand RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6080",
	.desc = "Offcore other requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF880",
	.desc = "Offcore other requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.other.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4080",
	.desc = "Offcore other requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6050",
	.desc = "Offcore prefetch data requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF850",
	.desc = "Offcore prefetch data requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2050",
	.desc = "Offcore prefetch data requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4050",
	.desc = "Offcore prefetch data requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6010",
	.desc = "Offcore prefetch data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF810",
	.desc = "Offcore prefetch data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2010",
	.desc = "Offcore prefetch data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4010",
	.desc = "Offcore prefetch data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6040",
	.desc = "Offcore prefetch code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF840",
	.desc = "Offcore prefetch code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2040",
	.desc = "Offcore prefetch code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4040",
	.desc = "Offcore prefetch code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6020",
	.desc = "Offcore prefetch RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF820",
	.desc = "Offcore prefetch RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2020",
	.desc = "Offcore prefetch RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4020",
	.desc = "Offcore prefetch RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6070",
	.desc = "Offcore prefetch requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF870",
	.desc = "Offcore prefetch requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2070",
	.desc = "Offcore prefetch requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4070",
	.desc = "Offcore prefetch requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "bpu_clears.early",
	.event = "umask=0x1,period=2000000,event=0xe8",
	.desc = "Early Branch Prediciton Unit clears",
	.topic = "other",
},
{
	.name = "bpu_clears.late",
	.event = "umask=0x2,period=2000000,event=0xe8",
	.desc = "Late Branch Prediction Unit clears",
	.topic = "other",
},
{
	.name = "bpu_missed_call_ret",
	.event = "umask=0x1,period=2000000,event=0xe5",
	.desc = "Branch prediction unit missed call or return",
	.topic = "other",
},
{
	.name = "es_reg_renames",
	.event = "umask=0x1,period=2000000,event=0xd5",
	.desc = "ES segment renames",
	.topic = "other",
},
{
	.name = "io_transactions",
	.event = "umask=0x1,period=2000000,event=0x6c",
	.desc = "I/O transactions",
	.topic = "other",
},
{
	.name = "l1i.cycles_stalled",
	.event = "umask=0x4,period=2000000,event=0x80",
	.desc = "L1I instruction fetch stall cycles",
	.topic = "other",
},
{
	.name = "l1i.hits",
	.event = "umask=0x1,period=2000000,event=0x80",
	.desc = "L1I instruction fetch hits",
	.topic = "other",
},
{
	.name = "l1i.misses",
	.event = "umask=0x2,period=2000000,event=0x80",
	.desc = "L1I instruction fetch misses",
	.topic = "other",
},
{
	.name = "l1i.reads",
	.event = "umask=0x3,period=2000000,event=0x80",
	.desc = "L1I Instruction fetches",
	.topic = "other",
},
{
	.name = "large_itlb.hit",
	.event = "umask=0x1,period=200000,event=0x82",
	.desc = "Large ITLB hit",
	.topic = "other",
},
{
	.name = "load_block.overlap_store",
	.event = "umask=0x2,period=200000,event=0x3",
	.desc = "Loads that partially overlap an earlier store",
	.topic = "other",
},
{
	.name = "load_dispatch.any",
	.event = "umask=0x7,period=2000000,event=0x13",
	.desc = "All loads dispatched",
	.topic = "other",
},
{
	.name = "load_dispatch.mob",
	.event = "umask=0x4,period=2000000,event=0x13",
	.desc = "Loads dispatched from the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs",
	.event = "umask=0x1,period=2000000,event=0x13",
	.desc = "Loads dispatched that bypass the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs_delayed",
	.event = "umask=0x2,period=2000000,event=0x13",
	.desc = "Loads dispatched from stage 305",
	.topic = "other",
},
{
	.name = "partial_address_alias",
	.event = "umask=0x1,period=200000,event=0x7",
	.desc = "False dependencies due to partial address aliasing",
	.topic = "other",
},
{
	.name = "rat_stalls.any",
	.event = "umask=0xf,period=2000000,event=0xd2",
	.desc = "All RAT stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.flags",
	.event = "umask=0x1,period=2000000,event=0xd2",
	.desc = "Flag stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.registers",
	.event = "umask=0x2,period=2000000,event=0xd2",
	.desc = "Partial register stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.rob_read_port",
	.event = "umask=0x4,period=2000000,event=0xd2",
	.desc = "ROB read port stalls cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.scoreboard",
	.event = "umask=0x8,period=2000000,event=0xd2",
	.desc = "Scoreboard stall cycles",
	.topic = "other",
},
{
	.name = "sb_drain.any",
	.event = "umask=0x7,period=200000,event=0x4",
	.desc = "All Store buffer stall cycles",
	.topic = "other",
},
{
	.name = "seg_rename_stalls",
	.event = "umask=0x1,period=2000000,event=0xd4",
	.desc = "Segment rename stall cycles",
	.topic = "other",
},
{
	.name = "snoop_response.hit",
	.event = "umask=0x1,period=100000,event=0xb8",
	.desc = "Thread responded HIT to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hite",
	.event = "umask=0x2,period=100000,event=0xb8",
	.desc = "Thread responded HITE to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hitm",
	.event = "umask=0x4,period=100000,event=0xb8",
	.desc = "Thread responded HITM to snoop",
	.topic = "other",
},
{
	.name = "snoopq_requests.code",
	.event = "umask=0x4,period=100000,event=0xb4",
	.desc = "Snoop code requests",
	.topic = "other",
},
{
	.name = "snoopq_requests.data",
	.event = "umask=0x1,period=100000,event=0xb4",
	.desc = "Snoop data requests",
	.topic = "other",
},
{
	.name = "snoopq_requests.invalidate",
	.event = "umask=0x2,period=100000,event=0xb4",
	.desc = "Snoop invalidate requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.code",
	.event = "umask=0x4,period=2000000,event=0xb3",
	.desc = "Outstanding snoop code requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.code_not_empty",
	.event = "umask=0x4,period=2000000,cmask=1,event=0xb3",
	.desc = "Cycles snoop code requests queued",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.data",
	.event = "umask=0x1,period=2000000,event=0xb3",
	.desc = "Outstanding snoop data requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.data_not_empty",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xb3",
	.desc = "Cycles snoop data requests queued",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.invalidate",
	.event = "umask=0x2,period=2000000,event=0xb3",
	.desc = "Outstanding snoop invalidate requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.invalidate_not_empty",
	.event = "umask=0x2,period=2000000,cmask=1,event=0xb3",
	.desc = "Cycles snoop invalidate requests queued",
	.topic = "other",
},
{
	.name = "sq_full_stall_cycles",
	.event = "umask=0x1,period=2000000,event=0xf6",
	.desc = "Super Queue full stall cycles",
	.topic = "other",
},
{
	.name = "arith.cycles_div_busy",
	.event = "umask=0x1,period=2000000,event=0x14",
	.desc = "Cycles the divider is busy",
	.topic = "pipeline",
},
{
	.name = "arith.div",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,edge=1,event=0x14",
	.desc = "Divide Operations executed",
	.topic = "pipeline",
},
{
	.name = "arith.mul",
	.event = "umask=0x2,period=2000000,event=0x14",
	.desc = "Multiply operations executed",
	.topic = "pipeline",
},
{
	.name = "baclear.bad_target",
	.event = "umask=0x2,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted with bad target address",
	.topic = "pipeline",
},
{
	.name = "baclear.clear",
	.event = "umask=0x1,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted, regardless of cause ",
	.topic = "pipeline",
},
{
	.name = "baclear_force_iq",
	.event = "umask=0x1,period=2000000,event=0xa7",
	.desc = "Instruction queue forced BACLEAR",
	.topic = "pipeline",
},
{
	.name = "br_inst_decoded",
	.event = "umask=0x1,period=2000000,event=0xe0",
	.desc = "Branch instructions decoded",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.any",
	.event = "umask=0x7f,period=200000,event=0x88",
	.desc = "Branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.cond",
	.event = "umask=0x1,period=200000,event=0x88",
	.desc = "Conditional branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct",
	.event = "umask=0x2,period=200000,event=0x88",
	.desc = "Unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct_near_call",
	.event = "umask=0x10,period=20000,event=0x88",
	.desc = "Unconditional call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_near_call",
	.event = "umask=0x20,period=20000,event=0x88",
	.desc = "Indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_non_call",
	.event = "umask=0x4,period=20000,event=0x88",
	.desc = "Indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.near_calls",
	.event = "umask=0x30,period=20000,event=0x88",
	.desc = "Call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.non_calls",
	.event = "umask=0x7,period=200000,event=0x88",
	.desc = "All non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.return_near",
	.event = "umask=0x8,period=20000,event=0x88",
	.desc = "Indirect return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken",
	.event = "umask=0x40,period=200000,event=0x88",
	.desc = "Taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x4,period=200000,event=0xc4",
	.desc = "Retired branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=200000,event=0xc4",
	.desc = "Retired conditional branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=20000,event=0xc4",
	.desc = "Retired near call instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.any",
	.event = "umask=0x7f,period=20000,event=0x89",
	.desc = "Mispredicted branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.cond",
	.event = "umask=0x1,period=20000,event=0x89",
	.desc = "Mispredicted conditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct",
	.event = "umask=0x2,period=20000,event=0x89",
	.desc = "Mispredicted unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct_near_call",
	.event = "umask=0x10,period=2000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_near_call",
	.event = "umask=0x20,period=2000,event=0x89",
	.desc = "Mispredicted indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_non_call",
	.event = "umask=0x4,period=2000,event=0x89",
	.desc = "Mispredicted indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.near_calls",
	.event = "umask=0x30,period=2000,event=0x89",
	.desc = "Mispredicted call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.non_calls",
	.event = "umask=0x7,period=20000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.return_near",
	.event = "umask=0x8,period=2000,event=0x89",
	.desc = "Mispredicted return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken",
	.event = "umask=0x40,period=20000,event=0x89",
	.desc = "Mispredicted taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x4,period=20000,event=0xc5",
	.desc = "Mispredicted retired branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=20000,event=0xc5",
	.desc = "Mispredicted conditional retired branches (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=2000,event=0xc5",
	.desc = "Mispredicted near retired calls (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Reference cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_p",
	.event = "umask=0x1,period=100000,event=0x3c",
	.desc = "Reference base clock (133 Mhz) cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000000,event=0x3c",
	.desc = "Cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.total_cycles",
	.event = "inv=1,umask=0x0,period=2000000,cmask=2,event=0x3c",
	.desc = "Total CPU cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.any",
	.event = "umask=0xf,period=2000000,event=0x87",
	.desc = "Any Instruction Length Decoder stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000000,event=0x87",
	.desc = "Instruction Queue full stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000000,event=0x87",
	.desc = "Length Change Prefix stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.mru",
	.event = "umask=0x2,period=2000000,event=0x87",
	.desc = "Stall cycles due to BPU MRU bypass",
	.topic = "pipeline",
},
{
	.name = "ild_stall.regen",
	.event = "umask=0x8,period=2000000,event=0x87",
	.desc = "Regen stall cycles",
	.topic = "pipeline",
},
{
	.name = "inst_decoded.dec0",
	.event = "umask=0x1,period=2000000,event=0x18",
	.desc = "Instructions that must be decoded by decoder 0",
	.topic = "pipeline",
},
{
	.name = "inst_queue_write_cycles",
	.event = "umask=0x1,period=2000000,event=0x1e",
	.desc = "Cycles instructions are written to the instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_queue_writes",
	.event = "umask=0x1,period=2000000,event=0x17",
	.desc = "Instructions written to instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (Programmable counter and Precise Event) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.mmx",
	.event = "umask=0x4,period=2000000,event=0xc0",
	.desc = "Retired MMX instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000000,event=0xc0",
	.desc = "Retired floating-point operations (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre",
	.event = "umask=0x1,period=200000,event=0x4c",
	.desc = "Load operations conflicting with software prefetches",
	.topic = "pipeline",
},
{
	.name = "lsd.active",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles when uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.inactive",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles no uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd_overflow",
	.event = "umask=0x1,period=2000000,event=0x20",
	.desc = "Loops that can't stream from the instruction queue",
	.topic = "pipeline",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=20000,event=0xc3",
	.desc = "Cycles machine clear asserted",
	.topic = "pipeline",
},
{
	.name = "machine_clears.mem_order",
	.event = "umask=0x2,period=20000,event=0xc3",
	.desc = "Execution pipeline restart due to Memory ordering conflicts",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=20000,event=0xc3",
	.desc = "Self-Modifying Code detected",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000000,event=0xa2",
	.desc = "Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.fpcw",
	.event = "umask=0x20,period=2000000,event=0xa2",
	.desc = "FPU control word write stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.load",
	.event = "umask=0x2,period=2000000,event=0xa2",
	.desc = "Load buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.mxcsr",
	.event = "umask=0x40,period=2000000,event=0xa2",
	.desc = "MXCSR rename stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.other",
	.event = "umask=0x80,period=2000000,event=0xa2",
	.desc = "Other Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rob_full",
	.event = "umask=0x10,period=2000000,event=0xa2",
	.desc = "ROB full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rs_full",
	.event = "umask=0x4,period=2000000,event=0xa2",
	.desc = "Reservation Station full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.store",
	.event = "umask=0x8,period=2000000,event=0xa2",
	.desc = "Store buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_double",
	.event = "umask=0x4,period=200000,event=0xc7",
	.desc = "SIMD Packed-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_single",
	.event = "umask=0x1,period=200000,event=0xc7",
	.desc = "SIMD Packed-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_double",
	.event = "umask=0x8,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_single",
	.event = "umask=0x2,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.vector_integer",
	.event = "umask=0x10,period=200000,event=0xc7",
	.desc = "SIMD Vector Integer Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uop_unfusion",
	.event = "umask=0x1,period=2000000,event=0xdb",
	.desc = "Uop unfusions due to FP exceptions",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_folding",
	.event = "umask=0x4,period=2000000,event=0xd1",
	.desc = "Stack pointer instructions decoded",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_sync",
	.event = "umask=0x8,period=2000000,event=0xd1",
	.desc = "Stack pointer sync operations",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.ms_cycles_active",
	.event = "umask=0x2,period=2000000,cmask=1,event=0xd1",
	.desc = "Uops decoded by Microcode Sequencer",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xd1",
	.desc = "Cycles no Uops are decoded",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles",
	.event = "umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles_no_port5",
	.event = "umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count",
	.event = "inv=1,umask=0x3f,any=1,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count_no_port5",
	.event = "inv=1,umask=0x1f,any=1,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles",
	.event = "inv=1,umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles_no_port5",
	.event = "inv=1,umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port0",
	.event = "umask=0x1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015",
	.event = "umask=0x40,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015_stall_cycles",
	.event = "inv=1,umask=0x40,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port1",
	.event = "umask=0x2,period=2000000,event=0xb1",
	.desc = "Uops executed on port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port2_core",
	.event = "umask=0x4,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 2 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port234_core",
	.event = "umask=0x80,any=1,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 2, 3 or 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port3_core",
	.event = "umask=0x8,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 3 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port4_core",
	.event = "umask=0x10,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port5",
	.event = "umask=0x20,period=2000000,event=0xb1",
	.desc = "Uops executed on port 5",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000000,event=0xe",
	.desc = "Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued on any thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.cycles_all_threads",
	.event = "umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles Uops were issued on either thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.fused",
	.event = "umask=0x2,period=2000000,event=0xe",
	.desc = "Fused Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued",
	.topic = "pipeline",
},
{
	.name = "uops_retired.active_cycles",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are being retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.any",
	.event = "umask=0x1,period=2000000,event=0xc2",
	.desc = "Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.macro_fused",
	.event = "umask=0x4,period=2000000,event=0xc2",
	.desc = "Macro-fused Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000000,event=0xc2",
	.desc = "Retirement slots used (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are not retiring (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc2",
	.desc = "Total cycles using precise uop retired event (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles_ps",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "dtlb_load_misses.any",
	.event = "umask=0x1,period=200000,event=0x8",
	.desc = "DTLB load misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.pde_miss",
	.event = "umask=0x20,period=200000,event=0x8",
	.desc = "DTLB load miss caused by low part of address",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x10,period=2000000,event=0x8",
	.desc = "DTLB second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x8",
	.desc = "DTLB load miss page walks complete",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_cycles",
	.event = "umask=0x4,period=200000,event=0x8",
	.desc = "DTLB load miss page walk cycles",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x49",
	.desc = "DTLB misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.large_walk_completed",
	.event = "umask=0x80,period=200000,event=0x49",
	.desc = "DTLB miss large page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.stlb_hit",
	.event = "umask=0x10,period=200000,event=0x49",
	.desc = "DTLB first level misses but second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x49",
	.desc = "DTLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.walk_cycles",
	.event = "umask=0x4,period=2000000,event=0x49",
	.desc = "DTLB miss page walk cycles",
	.topic = "virtual memory",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000000,event=0x4f",
	.desc = "Extended Page Table walk cycles",
	.topic = "virtual memory",
},
{
	.name = "itlb_flush",
	.event = "umask=0x1,period=2000000,event=0xae",
	.desc = "ITLB flushes",
	.topic = "virtual memory",
},
{
	.name = "itlb_miss_retired",
	.event = "umask=0x20,period=200000,event=0xc8",
	.desc = "Retired instructions that missed the ITLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x85",
	.desc = "ITLB miss",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x85",
	.desc = "ITLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_cycles",
	.event = "umask=0x4,period=2000000,event=0x85",
	.desc = "ITLB miss page walk cycles",
	.topic = "virtual memory",
},
{
	.name = "mem_load_retired.dtlb_miss",
	.event = "umask=0x80,period=200000,event=0xcb",
	.desc = "Retired loads that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "mem_store_retired.dtlb_miss",
	.event = "umask=0x1,period=200000,event=0xc",
	.desc = "Retired stores that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_westmereex[] = {
{
	.name = "cache_lock_cycles.l1d",
	.event = "umask=0x2,period=2000000,event=0x63",
	.desc = "Cycles L1D locked",
	.topic = "cache",
},
{
	.name = "cache_lock_cycles.l1d_l2",
	.event = "umask=0x1,period=2000000,event=0x63",
	.desc = "Cycles L1D and L2 locked",
	.topic = "cache",
},
{
	.name = "l1d.m_evict",
	.event = "umask=0x4,period=2000000,event=0x51",
	.desc = "L1D cache lines replaced in M state",
	.topic = "cache",
},
{
	.name = "l1d.m_repl",
	.event = "umask=0x2,period=2000000,event=0x51",
	.desc = "L1D cache lines allocated in the M state",
	.topic = "cache",
},
{
	.name = "l1d.m_snoop_evict",
	.event = "umask=0x8,period=2000000,event=0x51",
	.desc = "L1D snoop eviction of cache lines in M state",
	.topic = "cache",
},
{
	.name = "l1d.repl",
	.event = "umask=0x1,period=2000000,event=0x51",
	.desc = "L1 data cache lines allocated",
	.topic = "cache",
},
{
	.name = "l1d_cache_prefetch_lock_fb_hit",
	.event = "umask=0x1,period=2000000,event=0x52",
	.desc = "L1D prefetch load lock accepted in fill buffer",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.miss",
	.event = "umask=0x2,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch misses",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.requests",
	.event = "umask=0x1,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests",
	.topic = "cache",
},
{
	.name = "l1d_prefetch.triggers",
	.event = "umask=0x4,period=200000,event=0x4e",
	.desc = "L1D hardware prefetch requests triggered",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.e_state",
	.event = "umask=0x4,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in E state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.i_state",
	.event = "umask=0x1,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in I state (misses)",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.m_state",
	.event = "umask=0x8,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in M state",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.mesi",
	.event = "umask=0xf,period=100000,event=0x28",
	.desc = "All L1 writebacks to L2",
	.topic = "cache",
},
{
	.name = "l1d_wb_l2.s_state",
	.event = "umask=0x2,period=100000,event=0x28",
	.desc = "L1 writebacks to L2 in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.any",
	.event = "umask=0xff,period=200000,event=0x26",
	.desc = "All L2 data requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.e_state",
	.event = "umask=0x4,period=200000,event=0x26",
	.desc = "L2 data demand loads in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.i_state",
	.event = "umask=0x1,period=200000,event=0x26",
	.desc = "L2 data demand loads in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.m_state",
	.event = "umask=0x8,period=200000,event=0x26",
	.desc = "L2 data demand loads in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.mesi",
	.event = "umask=0xf,period=200000,event=0x26",
	.desc = "L2 data demand requests",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.demand.s_state",
	.event = "umask=0x2,period=200000,event=0x26",
	.desc = "L2 data demand loads in S state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.e_state",
	.event = "umask=0x40,period=200000,event=0x26",
	.desc = "L2 data prefetches in E state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.i_state",
	.event = "umask=0x10,period=200000,event=0x26",
	.desc = "L2 data prefetches in the I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.m_state",
	.event = "umask=0x80,period=200000,event=0x26",
	.desc = "L2 data prefetches in M state",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.mesi",
	.event = "umask=0xf0,period=200000,event=0x26",
	.desc = "All L2 data prefetches",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.prefetch.s_state",
	.event = "umask=0x20,period=200000,event=0x26",
	.desc = "L2 data prefetches in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.any",
	.event = "umask=0x7,period=100000,event=0xf1",
	.desc = "L2 lines alloacated",
	.topic = "cache",
},
{
	.name = "l2_lines_in.e_state",
	.event = "umask=0x4,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the E state",
	.topic = "cache",
},
{
	.name = "l2_lines_in.s_state",
	.event = "umask=0x2,period=100000,event=0xf1",
	.desc = "L2 lines allocated in the S state",
	.topic = "cache",
},
{
	.name = "l2_lines_out.any",
	.event = "umask=0xf,period=100000,event=0xf2",
	.desc = "L2 lines evicted",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x1,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.demand_dirty",
	.event = "umask=0x2,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a demand request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_clean",
	.event = "umask=0x4,period=100000,event=0xf2",
	.desc = "L2 lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_lines_out.prefetch_dirty",
	.event = "umask=0x8,period=100000,event=0xf2",
	.desc = "L2 modified lines evicted by a prefetch request",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_hit",
	.event = "umask=0x10,period=200000,event=0x24",
	.desc = "L2 instruction fetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetch_miss",
	.event = "umask=0x20,period=200000,event=0x24",
	.desc = "L2 instruction fetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ifetches",
	.event = "umask=0x30,period=200000,event=0x24",
	.desc = "L2 instruction fetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_hit",
	.event = "umask=0x1,period=200000,event=0x24",
	.desc = "L2 load hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.ld_miss",
	.event = "umask=0x2,period=200000,event=0x24",
	.desc = "L2 load misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.loads",
	.event = "umask=0x3,period=200000,event=0x24",
	.desc = "L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0xaa,period=200000,event=0x24",
	.desc = "All L2 misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_hit",
	.event = "umask=0x40,period=200000,event=0x24",
	.desc = "L2 prefetch hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetch_miss",
	.event = "umask=0x80,period=200000,event=0x24",
	.desc = "L2 prefetch misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.prefetches",
	.event = "umask=0xc0,period=200000,event=0x24",
	.desc = "All L2 prefetches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200000,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0x4,period=200000,event=0x24",
	.desc = "L2 RFO hits",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x8,period=200000,event=0x24",
	.desc = "L2 RFO misses",
	.topic = "cache",
},
{
	.name = "l2_rqsts.rfos",
	.event = "umask=0xc,period=200000,event=0x24",
	.desc = "L2 RFO requests",
	.topic = "cache",
},
{
	.name = "l2_transactions.any",
	.event = "umask=0x80,period=200000,event=0xf0",
	.desc = "All L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.fill",
	.event = "umask=0x20,period=200000,event=0xf0",
	.desc = "L2 fill transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.ifetch",
	.event = "umask=0x4,period=200000,event=0xf0",
	.desc = "L2 instruction fetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.l1d_wb",
	.event = "umask=0x10,period=200000,event=0xf0",
	.desc = "L1D writeback to L2 transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.load",
	.event = "umask=0x1,period=200000,event=0xf0",
	.desc = "L2 Load transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.prefetch",
	.event = "umask=0x8,period=200000,event=0xf0",
	.desc = "L2 prefetch transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.rfo",
	.event = "umask=0x2,period=200000,event=0xf0",
	.desc = "L2 RFO transactions",
	.topic = "cache",
},
{
	.name = "l2_transactions.wb",
	.event = "umask=0x40,period=200000,event=0xf0",
	.desc = "L2 writeback to LLC transactions",
	.topic = "cache",
},
{
	.name = "l2_write.lock.e_state",
	.event = "umask=0x40,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in E state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.hit",
	.event = "umask=0xe0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.lock.i_state",
	.event = "umask=0x10,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.lock.m_state",
	.event = "umask=0x80,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.lock.mesi",
	.event = "umask=0xf0,period=100000,event=0x27",
	.desc = "All demand L2 lock RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.lock.s_state",
	.event = "umask=0x20,period=100000,event=0x27",
	.desc = "L2 demand lock RFOs in S state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.hit",
	.event = "umask=0xe,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs that hit the cache",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.i_state",
	.event = "umask=0x1,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in I state (misses)",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.m_state",
	.event = "umask=0x8,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in M state",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.mesi",
	.event = "umask=0xf,period=100000,event=0x27",
	.desc = "All L2 demand store RFOs",
	.topic = "cache",
},
{
	.name = "l2_write.rfo.s_state",
	.event = "umask=0x2,period=100000,event=0x27",
	.desc = "L2 demand store RFOs in S state",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100000,event=0x2e",
	.desc = "Longest latency cache miss",
	.topic = "cache",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=200000,event=0x2e",
	.desc = "Longest latency cache reference",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.loads",
	.event = "umask=0x1,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a load (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.stores",
	.event = "umask=0x2,period=2000000,event=0xb",
	.desc = "Instructions retired which contains a store (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.hit_lfb",
	.event = "umask=0x40,period=200000,event=0xcb",
	.desc = "Retired loads that miss L1D and hit an previously allocated LFB (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l1d_hit",
	.event = "umask=0x1,period=2000000,event=0xcb",
	.desc = "Retired loads that hit the L1 data cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x2,period=200000,event=0xcb",
	.desc = "Retired loads that hit the L2 cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_miss",
	.event = "umask=0x10,period=10000,event=0xcb",
	.desc = "Retired loads that miss the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.llc_unshared_hit",
	.event = "umask=0x4,period=40000,event=0xcb",
	.desc = "Retired loads that hit valid versions in the LLC cache (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.other_core_l2_hit_hitm",
	.event = "umask=0x8,period=40000,event=0xcb",
	.desc = "Retired loads that hit sibling core's L2 in modified or unmodified states (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.local_hitm",
	.event = "umask=0x2,period=40000,event=0xf",
	.desc = "Load instructions retired that HIT modified data in sibling core (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.local_dram_and_remote_cache_hit",
	.event = "umask=0x8,period=20000,event=0xf",
	.desc = "Load instructions retired local dram and remote cache HIT data sources (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.remote_dram",
	.event = "umask=0x20,period=10000,event=0xf",
	.desc = "Load instructions retired remote DRAM and remote home-remote cache HITM (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.uncacheable",
	.event = "umask=0x80,period=4000,event=0xf",
	.desc = "Load instructions retired IO (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_uncore_retired.remote_hitm",
	.event = "umask=0x4,period=40000,event=0xf",
	.desc = "Retired loads that hit remote socket in modified state (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_requests.any",
	.event = "umask=0x80,period=100000,event=0xb0",
	.desc = "All offcore requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.any.read",
	.event = "umask=0x8,period=100000,event=0xb0",
	.desc = "Offcore read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.any.rfo",
	.event = "umask=0x10,period=100000,event=0xb0",
	.desc = "Offcore RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand.read_code",
	.event = "umask=0x2,period=100000,event=0xb0",
	.desc = "Offcore demand code read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand.read_data",
	.event = "umask=0x1,period=100000,event=0xb0",
	.desc = "Offcore demand data read requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.demand.rfo",
	.event = "umask=0x4,period=100000,event=0xb0",
	.desc = "Offcore demand RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_requests.l1d_writeback",
	.event = "umask=0x40,period=100000,event=0xb0",
	.desc = "Offcore L1 data cache writebacks",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.any.read",
	.event = "umask=0x8,period=2000000,event=0x60",
	.desc = "Outstanding offcore reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.any.read_not_empty",
	.event = "umask=0x8,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore reads busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_code",
	.event = "umask=0x2,period=2000000,event=0x60",
	.desc = "Outstanding offcore demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_code_not_empty",
	.event = "umask=0x2,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore demand code read busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_data",
	.event = "umask=0x1,period=2000000,event=0x60",
	.desc = "Outstanding offcore demand data reads",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.read_data_not_empty",
	.event = "umask=0x1,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore demand data read busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.rfo",
	.event = "umask=0x4,period=2000000,event=0x60",
	.desc = "Outstanding offcore demand RFOs",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand.rfo_not_empty",
	.event = "umask=0x4,period=2000000,cmask=1,event=0x60",
	.desc = "Cycles offcore demand RFOs busy",
	.topic = "cache",
},
{
	.name = "offcore_requests_sq_full",
	.event = "umask=0x1,period=100000,event=0xb2",
	.desc = "Offcore requests blocked due to Super Queue full",
	.topic = "cache",
},
{
	.name = "sq_misc.lru_hints",
	.event = "umask=0x4,period=2000000,event=0xf4",
	.desc = "Super Queue LRU hints sent to LLC",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=2000000,event=0xf4",
	.desc = "Super Queue lock splits across a cache line",
	.topic = "cache",
},
{
	.name = "store_blocks.at_ret",
	.event = "umask=0x4,period=200000,event=0x6",
	.desc = "Loads delayed with at-Retirement block code",
	.topic = "cache",
},
{
	.name = "store_blocks.l1d_block",
	.event = "umask=0x8,period=200000,event=0x6",
	.desc = "Cacheable loads delayed with L1D block code",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_0",
	.event = "umask=0x10,period=2000000,event=0xb,ldlat=0x0",
	.desc = "Memory instructions retired above 0 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_1024",
	.event = "umask=0x10,period=100,event=0xb,ldlat=0x400",
	.desc = "Memory instructions retired above 1024 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_128",
	.event = "umask=0x10,period=1000,event=0xb,ldlat=0x80",
	.desc = "Memory instructions retired above 128 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16",
	.event = "umask=0x10,period=10000,event=0xb,ldlat=0x10",
	.desc = "Memory instructions retired above 16 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_16384",
	.event = "umask=0x10,period=5,event=0xb,ldlat=0x4000",
	.desc = "Memory instructions retired above 16384 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_2048",
	.event = "umask=0x10,period=50,event=0xb,ldlat=0x800",
	.desc = "Memory instructions retired above 2048 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_256",
	.event = "umask=0x10,period=500,event=0xb,ldlat=0x100",
	.desc = "Memory instructions retired above 256 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32",
	.event = "umask=0x10,period=5000,event=0xb,ldlat=0x20",
	.desc = "Memory instructions retired above 32 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_32768",
	.event = "umask=0x10,period=3,event=0xb,ldlat=0x8000",
	.desc = "Memory instructions retired above 32768 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4",
	.event = "umask=0x10,period=50000,event=0xb,ldlat=0x4",
	.desc = "Memory instructions retired above 4 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_4096",
	.event = "umask=0x10,period=20,event=0xb,ldlat=0x1000",
	.desc = "Memory instructions retired above 4096 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_512",
	.event = "umask=0x10,period=200,event=0xb,ldlat=0x200",
	.desc = "Memory instructions retired above 512 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_64",
	.event = "umask=0x10,period=2000,event=0xb,ldlat=0x40",
	.desc = "Memory instructions retired above 64 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8",
	.event = "umask=0x10,period=20000,event=0xb,ldlat=0x8",
	.desc = "Memory instructions retired above 8 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "mem_inst_retired.latency_above_threshold_8192",
	.event = "umask=0x10,period=10,event=0xb,ldlat=0x2000",
	.desc = "Memory instructions retired above 8192 clocks (Precise Event)",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F11",
	.desc = "Offcore data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF11",
	.desc = "All offcore data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8011",
	.desc = "Offcore data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x111",
	.desc = "Offcore data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x211",
	.desc = "Offcore data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x411",
	.desc = "Offcore data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x711",
	.desc = "Offcore data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4711",
	.desc = "Offcore data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1811",
	.desc = "Offcore data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3811",
	.desc = "Offcore data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1011",
	.desc = "Offcore data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x811",
	.desc = "Offcore data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F44",
	.desc = "Offcore code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF44",
	.desc = "All offcore code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8044",
	.desc = "Offcore code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x144",
	.desc = "Offcore code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x244",
	.desc = "Offcore code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x444",
	.desc = "Offcore code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x744",
	.desc = "Offcore code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4744",
	.desc = "Offcore code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1844",
	.desc = "Offcore code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3844",
	.desc = "Offcore code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1044",
	.desc = "Offcore code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x844",
	.desc = "Offcore code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7FFF",
	.desc = "Offcore requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFFFF",
	.desc = "All offcore requests",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x80FF",
	.desc = "Offcore requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1FF",
	.desc = "Offcore requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2FF",
	.desc = "Offcore requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4FF",
	.desc = "Offcore requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7FF",
	.desc = "Offcore requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x47FF",
	.desc = "Offcore requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x18FF",
	.desc = "Offcore requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x38FF",
	.desc = "Offcore requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x10FF",
	.desc = "Offcore requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_request.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8FF",
	.desc = "Offcore requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F22",
	.desc = "Offcore RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF22",
	.desc = "All offcore RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8022",
	.desc = "Offcore RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x122",
	.desc = "Offcore RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x222",
	.desc = "Offcore RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x422",
	.desc = "Offcore RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x722",
	.desc = "Offcore RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4722",
	.desc = "Offcore RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1822",
	.desc = "Offcore RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3822",
	.desc = "Offcore RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1022",
	.desc = "Offcore RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.any_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x822",
	.desc = "Offcore RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F08",
	.desc = "Offcore writebacks to any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF08",
	.desc = "All offcore writebacks",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8008",
	.desc = "Offcore writebacks to the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x108",
	.desc = "Offcore writebacks to the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x408",
	.desc = "Offcore writebacks to the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x708",
	.desc = "Offcore writebacks to the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4708",
	.desc = "Offcore writebacks to the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1808",
	.desc = "Offcore writebacks to a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3808",
	.desc = "Offcore writebacks to a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1008",
	.desc = "Offcore writebacks that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.corewb.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x808",
	.desc = "Offcore writebacks that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F77",
	.desc = "Offcore code or data read requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF77",
	.desc = "All offcore code or data read requests",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8077",
	.desc = "Offcore code or data read requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x177",
	.desc = "Offcore code or data read requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x277",
	.desc = "Offcore code or data read requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x477",
	.desc = "Offcore code or data read requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x777",
	.desc = "Offcore code or data read requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4777",
	.desc = "Offcore code or data read requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1877",
	.desc = "Offcore code or data read requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3877",
	.desc = "Offcore code or data read requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1077",
	.desc = "Offcore code or data read requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x877",
	.desc = "Offcore code or data read requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F33",
	.desc = "Offcore request = all data, response = any cache_dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF33",
	.desc = "Offcore request = all data, response = any location",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8033",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x133",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x233",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x433",
	.desc = "Offcore data reads, RFO's and prefetches satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x733",
	.desc = "Offcore request = all data, response = local cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4733",
	.desc = "Offcore request = all data, response = local cache or dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1833",
	.desc = "Offcore request = all data, response = remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3833",
	.desc = "Offcore request = all data, response = remote cache or dram",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1033",
	.desc = "Offcore data reads, RFO's and prefetches that HIT in a remote cache ",
	.topic = "cache",
},
{
	.name = "offcore_response.data_in.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x833",
	.desc = "Offcore data reads, RFO's and prefetches that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F03",
	.desc = "Offcore demand data requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF03",
	.desc = "All offcore demand data requests",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8003",
	.desc = "Offcore demand data requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x103",
	.desc = "Offcore demand data requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x203",
	.desc = "Offcore demand data requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x403",
	.desc = "Offcore demand data requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x703",
	.desc = "Offcore demand data requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4703",
	.desc = "Offcore demand data requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1803",
	.desc = "Offcore demand data requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3803",
	.desc = "Offcore demand data requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1003",
	.desc = "Offcore demand data requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x803",
	.desc = "Offcore demand data requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F01",
	.desc = "Offcore demand data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF01",
	.desc = "All offcore demand data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8001",
	.desc = "Offcore demand data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x101",
	.desc = "Offcore demand data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x201",
	.desc = "Offcore demand data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x401",
	.desc = "Offcore demand data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x701",
	.desc = "Offcore demand data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4701",
	.desc = "Offcore demand data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1801",
	.desc = "Offcore demand data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3801",
	.desc = "Offcore demand data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1001",
	.desc = "Offcore demand data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x801",
	.desc = "Offcore demand data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F04",
	.desc = "Offcore demand code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF04",
	.desc = "All offcore demand code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8004",
	.desc = "Offcore demand code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x104",
	.desc = "Offcore demand code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x204",
	.desc = "Offcore demand code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x404",
	.desc = "Offcore demand code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x704",
	.desc = "Offcore demand code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4704",
	.desc = "Offcore demand code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1804",
	.desc = "Offcore demand code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3804",
	.desc = "Offcore demand code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1004",
	.desc = "Offcore demand code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x804",
	.desc = "Offcore demand code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F02",
	.desc = "Offcore demand RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF02",
	.desc = "All offcore demand RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8002",
	.desc = "Offcore demand RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x102",
	.desc = "Offcore demand RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x202",
	.desc = "Offcore demand RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x402",
	.desc = "Offcore demand RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x702",
	.desc = "Offcore demand RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4702",
	.desc = "Offcore demand RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1802",
	.desc = "Offcore demand RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3802",
	.desc = "Offcore demand RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1002",
	.desc = "Offcore demand RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.demand_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x802",
	.desc = "Offcore demand RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F80",
	.desc = "Offcore other requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF80",
	.desc = "All offcore other requests",
	.topic = "cache",
},
{
	.name = "offcore_response.other.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8080",
	.desc = "Offcore other requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x180",
	.desc = "Offcore other requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x280",
	.desc = "Offcore other requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x480",
	.desc = "Offcore other requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x780",
	.desc = "Offcore other requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.other.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4780",
	.desc = "Offcore other requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1880",
	.desc = "Offcore other requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3880",
	.desc = "Offcore other requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1080",
	.desc = "Offcore other requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.other.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x880",
	.desc = "Offcore other requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F30",
	.desc = "Offcore prefetch data requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF30",
	.desc = "All offcore prefetch data requests",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8030",
	.desc = "Offcore prefetch data requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x130",
	.desc = "Offcore prefetch data requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x230",
	.desc = "Offcore prefetch data requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x430",
	.desc = "Offcore prefetch data requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x730",
	.desc = "Offcore prefetch data requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4730",
	.desc = "Offcore prefetch data requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1830",
	.desc = "Offcore prefetch data requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3830",
	.desc = "Offcore prefetch data requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1030",
	.desc = "Offcore prefetch data requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x830",
	.desc = "Offcore prefetch data requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F10",
	.desc = "Offcore prefetch data reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF10",
	.desc = "All offcore prefetch data reads",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8010",
	.desc = "Offcore prefetch data reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x110",
	.desc = "Offcore prefetch data reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x210",
	.desc = "Offcore prefetch data reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x410",
	.desc = "Offcore prefetch data reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x710",
	.desc = "Offcore prefetch data reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4710",
	.desc = "Offcore prefetch data reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1810",
	.desc = "Offcore prefetch data reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3810",
	.desc = "Offcore prefetch data reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1010",
	.desc = "Offcore prefetch data reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_data_rd.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x810",
	.desc = "Offcore prefetch data reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F40",
	.desc = "Offcore prefetch code reads satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF40",
	.desc = "All offcore prefetch code reads",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8040",
	.desc = "Offcore prefetch code reads satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x140",
	.desc = "Offcore prefetch code reads satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x240",
	.desc = "Offcore prefetch code reads satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x440",
	.desc = "Offcore prefetch code reads satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x740",
	.desc = "Offcore prefetch code reads satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4740",
	.desc = "Offcore prefetch code reads satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1840",
	.desc = "Offcore prefetch code reads satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3840",
	.desc = "Offcore prefetch code reads satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1040",
	.desc = "Offcore prefetch code reads that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_ifetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x840",
	.desc = "Offcore prefetch code reads that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F20",
	.desc = "Offcore prefetch RFO requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF20",
	.desc = "All offcore prefetch RFO requests",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8020",
	.desc = "Offcore prefetch RFO requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x120",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x220",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x420",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x720",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4720",
	.desc = "Offcore prefetch RFO requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1820",
	.desc = "Offcore prefetch RFO requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3820",
	.desc = "Offcore prefetch RFO requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1020",
	.desc = "Offcore prefetch RFO requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.pf_rfo.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x820",
	.desc = "Offcore prefetch RFO requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x7F70",
	.desc = "Offcore prefetch requests satisfied by any cache or DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.any_location",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xFF70",
	.desc = "All offcore prefetch requests",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.io_csr_mmio",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x8070",
	.desc = "Offcore prefetch requests satisfied by the IO, CSR, MMIO unit",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_no_other_core",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x170",
	.desc = "Offcore prefetch requests satisfied by the LLC and not found in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x270",
	.desc = "Offcore prefetch requests satisfied by the LLC and HIT in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.llc_hit_other_core_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x470",
	.desc = "Offcore prefetch requests satisfied by the LLC  and HITM in a sibling core",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x770",
	.desc = "Offcore prefetch requests satisfied by the LLC",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.local_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4770",
	.desc = "Offcore prefetch requests satisfied by the LLC or local DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1870",
	.desc = "Offcore prefetch requests satisfied by a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x3870",
	.desc = "Offcore prefetch requests satisfied by a remote cache or remote DRAM",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_hit",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x1070",
	.desc = "Offcore prefetch requests that HIT in a remote cache",
	.topic = "cache",
},
{
	.name = "offcore_response.prefetch.remote_cache_hitm",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x870",
	.desc = "Offcore prefetch requests that HITM in a remote cache",
	.topic = "cache",
},
{
	.name = "fp_assist.all",
	.event = "umask=0x1,period=20000,event=0xf7",
	.desc = "X87 Floating point assists (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.input",
	.event = "umask=0x4,period=20000,event=0xf7",
	.desc = "X87 Floating poiint assists for invalid input value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_assist.output",
	.event = "umask=0x2,period=20000,event=0xf7",
	.desc = "X87 Floating point assists for invalid output value (Precise Event)",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.mmx",
	.event = "umask=0x2,period=2000000,event=0x10",
	.desc = "MMX Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_double_precision",
	.event = "umask=0x80,period=2000000,event=0x10",
	.desc = "SSE* FP double precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp",
	.event = "umask=0x4,period=2000000,event=0x10",
	.desc = "SSE and SSE2 FP Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_packed",
	.event = "umask=0x10,period=2000000,event=0x10",
	.desc = "SSE FP packed Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_fp_scalar",
	.event = "umask=0x20,period=2000000,event=0x10",
	.desc = "SSE FP scalar Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse_single_precision",
	.event = "umask=0x40,period=2000000,event=0x10",
	.desc = "SSE* FP single precision Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.sse2_integer",
	.event = "umask=0x8,period=2000000,event=0x10",
	.desc = "SSE2 integer Uops",
	.topic = "floating point",
},
{
	.name = "fp_comp_ops_exe.x87",
	.event = "umask=0x1,period=2000000,event=0x10",
	.desc = "Computational floating-point operations executed",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.any",
	.event = "umask=0x3,period=2000000,event=0xcc",
	.desc = "All Floating Point to and from MMX transitions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_fp",
	.event = "umask=0x1,period=2000000,event=0xcc",
	.desc = "Transitions from MMX to Floating Point instructions",
	.topic = "floating point",
},
{
	.name = "fp_mmx_trans.to_mmx",
	.event = "umask=0x2,period=2000000,event=0xcc",
	.desc = "Transitions from Floating Point to MMX instructions",
	.topic = "floating point",
},
{
	.name = "simd_int_128.pack",
	.event = "umask=0x4,period=200000,event=0x12",
	.desc = "128 bit SIMD integer pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_arith",
	.event = "umask=0x20,period=200000,event=0x12",
	.desc = "128 bit SIMD integer arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_logical",
	.event = "umask=0x10,period=200000,event=0x12",
	.desc = "128 bit SIMD integer logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_mpy",
	.event = "umask=0x1,period=200000,event=0x12",
	.desc = "128 bit SIMD integer multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.packed_shift",
	.event = "umask=0x2,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.shuffle_move",
	.event = "umask=0x40,period=200000,event=0x12",
	.desc = "128 bit SIMD integer shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_128.unpack",
	.event = "umask=0x8,period=200000,event=0x12",
	.desc = "128 bit SIMD integer unpack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.pack",
	.event = "umask=0x4,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit pack operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_arith",
	.event = "umask=0x20,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit arithmetic operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_logical",
	.event = "umask=0x10,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit logical operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_mpy",
	.event = "umask=0x1,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit packed multiply operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.packed_shift",
	.event = "umask=0x2,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shift operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.shuffle_move",
	.event = "umask=0x40,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit shuffle/move operations",
	.topic = "floating point",
},
{
	.name = "simd_int_64.unpack",
	.event = "umask=0x8,period=200000,event=0xfd",
	.desc = "SIMD integer 64 bit unpack operations",
	.topic = "floating point",
},
{
	.name = "macro_insts.decoded",
	.event = "umask=0x1,period=2000000,event=0xd0",
	.desc = "Instructions decoded",
	.topic = "frontend",
},
{
	.name = "macro_insts.fusions_decoded",
	.event = "umask=0x1,period=2000000,event=0xa6",
	.desc = "Macro-fused instructions decoded",
	.topic = "frontend",
},
{
	.name = "two_uop_insts_decoded",
	.event = "umask=0x1,period=2000000,event=0x19",
	.desc = "Two Uop instructions decoded",
	.topic = "frontend",
},
{
	.name = "misalign_mem_ref.store",
	.event = "umask=0x2,period=200000,event=0x5",
	.desc = "Misaligned store references",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6011",
	.desc = "Offcore data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF811",
	.desc = "Offcore data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4011",
	.desc = "Offcore data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2011",
	.desc = "Offcore data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6044",
	.desc = "Offcore code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF844",
	.desc = "Offcore code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4044",
	.desc = "Offcore code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2044",
	.desc = "Offcore code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x60FF",
	.desc = "Offcore requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF8FF",
	.desc = "Offcore requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x40FF",
	.desc = "Offcore requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_request.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x20FF",
	.desc = "Offcore requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6022",
	.desc = "Offcore RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF822",
	.desc = "Offcore RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4022",
	.desc = "Offcore RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.any_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2022",
	.desc = "Offcore RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6008",
	.desc = "Offcore writebacks to any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF808",
	.desc = "Offcore writebacks that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4008",
	.desc = "Offcore writebacks to the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.corewb.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2008",
	.desc = "Offcore writebacks to a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6077",
	.desc = "Offcore code or data read requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF877",
	.desc = "Offcore code or data read requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4077",
	.desc = "Offcore code or data read requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2077",
	.desc = "Offcore code or data read requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6033",
	.desc = "Offcore request = all data, response = any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF833",
	.desc = "Offcore request = all data, response = any LLC miss",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4033",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.data_in.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2033",
	.desc = "Offcore data reads, RFO's and prefetches statisfied by the remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6003",
	.desc = "Offcore demand data requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF803",
	.desc = "Offcore demand data requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4003",
	.desc = "Offcore demand data requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2003",
	.desc = "Offcore demand data requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6001",
	.desc = "Offcore demand data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF801",
	.desc = "Offcore demand data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4001",
	.desc = "Offcore demand data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2001",
	.desc = "Offcore demand data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6004",
	.desc = "Offcore demand code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF804",
	.desc = "Offcore demand code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4004",
	.desc = "Offcore demand code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2004",
	.desc = "Offcore demand code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6002",
	.desc = "Offcore demand RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF802",
	.desc = "Offcore demand RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4002",
	.desc = "Offcore demand RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.demand_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2002",
	.desc = "Offcore demand RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6080",
	.desc = "Offcore other requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.other.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF880",
	.desc = "Offcore other requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.other.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2080",
	.desc = "Offcore other requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6030",
	.desc = "Offcore prefetch data requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF830",
	.desc = "Offcore prefetch data requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4030",
	.desc = "Offcore prefetch data requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2030",
	.desc = "Offcore prefetch data requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6010",
	.desc = "Offcore prefetch data reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF810",
	.desc = "Offcore prefetch data reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4010",
	.desc = "Offcore prefetch data reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_data_rd.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2010",
	.desc = "Offcore prefetch data reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6040",
	.desc = "Offcore prefetch code reads satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF840",
	.desc = "Offcore prefetch code reads that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4040",
	.desc = "Offcore prefetch code reads satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_ifetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2040",
	.desc = "Offcore prefetch code reads satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6020",
	.desc = "Offcore prefetch RFO requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF820",
	.desc = "Offcore prefetch RFO requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4020",
	.desc = "Offcore prefetch RFO requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.pf_rfo.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2020",
	.desc = "Offcore prefetch RFO requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x6070",
	.desc = "Offcore prefetch requests satisfied by any DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.any_llc_miss",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0xF870",
	.desc = "Offcore prefetch requests that missed the LLC",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.local_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x4070",
	.desc = "Offcore prefetch requests satisfied by the local DRAM",
	.topic = "memory",
},
{
	.name = "offcore_response.prefetch.remote_dram",
	.event = "umask=0x1,period=100000,event=0xb7,offcore_rsp=0x2070",
	.desc = "Offcore prefetch requests satisfied by a remote DRAM",
	.topic = "memory",
},
{
	.name = "bpu_clears.early",
	.event = "umask=0x1,period=2000000,event=0xe8",
	.desc = "Early Branch Prediciton Unit clears",
	.topic = "other",
},
{
	.name = "bpu_clears.late",
	.event = "umask=0x2,period=2000000,event=0xe8",
	.desc = "Late Branch Prediction Unit clears",
	.topic = "other",
},
{
	.name = "bpu_missed_call_ret",
	.event = "umask=0x1,period=2000000,event=0xe5",
	.desc = "Branch prediction unit missed call or return",
	.topic = "other",
},
{
	.name = "es_reg_renames",
	.event = "umask=0x1,period=2000000,event=0xd5",
	.desc = "ES segment renames",
	.topic = "other",
},
{
	.name = "io_transactions",
	.event = "umask=0x1,period=2000000,event=0x6c",
	.desc = "I/O transactions",
	.topic = "other",
},
{
	.name = "l1i.cycles_stalled",
	.event = "umask=0x4,period=2000000,event=0x80",
	.desc = "L1I instruction fetch stall cycles",
	.topic = "other",
},
{
	.name = "l1i.hits",
	.event = "umask=0x1,period=2000000,event=0x80",
	.desc = "L1I instruction fetch hits",
	.topic = "other",
},
{
	.name = "l1i.misses",
	.event = "umask=0x2,period=2000000,event=0x80",
	.desc = "L1I instruction fetch misses",
	.topic = "other",
},
{
	.name = "l1i.reads",
	.event = "umask=0x3,period=2000000,event=0x80",
	.desc = "L1I Instruction fetches",
	.topic = "other",
},
{
	.name = "large_itlb.hit",
	.event = "umask=0x1,period=200000,event=0x82",
	.desc = "Large ITLB hit",
	.topic = "other",
},
{
	.name = "load_block.overlap_store",
	.event = "umask=0x2,period=200000,event=0x3",
	.desc = "Loads that partially overlap an earlier store",
	.topic = "other",
},
{
	.name = "load_dispatch.any",
	.event = "umask=0x7,period=2000000,event=0x13",
	.desc = "All loads dispatched",
	.topic = "other",
},
{
	.name = "load_dispatch.mob",
	.event = "umask=0x4,period=2000000,event=0x13",
	.desc = "Loads dispatched from the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs",
	.event = "umask=0x1,period=2000000,event=0x13",
	.desc = "Loads dispatched that bypass the MOB",
	.topic = "other",
},
{
	.name = "load_dispatch.rs_delayed",
	.event = "umask=0x2,period=2000000,event=0x13",
	.desc = "Loads dispatched from stage 305",
	.topic = "other",
},
{
	.name = "partial_address_alias",
	.event = "umask=0x1,period=200000,event=0x7",
	.desc = "False dependencies due to partial address aliasing",
	.topic = "other",
},
{
	.name = "rat_stalls.any",
	.event = "umask=0xf,period=2000000,event=0xd2",
	.desc = "All RAT stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.flags",
	.event = "umask=0x1,period=2000000,event=0xd2",
	.desc = "Flag stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.registers",
	.event = "umask=0x2,period=2000000,event=0xd2",
	.desc = "Partial register stall cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.rob_read_port",
	.event = "umask=0x4,period=2000000,event=0xd2",
	.desc = "ROB read port stalls cycles",
	.topic = "other",
},
{
	.name = "rat_stalls.scoreboard",
	.event = "umask=0x8,period=2000000,event=0xd2",
	.desc = "Scoreboard stall cycles",
	.topic = "other",
},
{
	.name = "sb_drain.any",
	.event = "umask=0x7,period=200000,event=0x4",
	.desc = "All Store buffer stall cycles",
	.topic = "other",
},
{
	.name = "seg_rename_stalls",
	.event = "umask=0x1,period=2000000,event=0xd4",
	.desc = "Segment rename stall cycles",
	.topic = "other",
},
{
	.name = "snoop_response.hit",
	.event = "umask=0x1,period=100000,event=0xb8",
	.desc = "Thread responded HIT to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hite",
	.event = "umask=0x2,period=100000,event=0xb8",
	.desc = "Thread responded HITE to snoop",
	.topic = "other",
},
{
	.name = "snoop_response.hitm",
	.event = "umask=0x4,period=100000,event=0xb8",
	.desc = "Thread responded HITM to snoop",
	.topic = "other",
},
{
	.name = "snoopq_requests.code",
	.event = "umask=0x4,period=100000,event=0xb4",
	.desc = "Snoop code requests",
	.topic = "other",
},
{
	.name = "snoopq_requests.data",
	.event = "umask=0x1,period=100000,event=0xb4",
	.desc = "Snoop data requests",
	.topic = "other",
},
{
	.name = "snoopq_requests.invalidate",
	.event = "umask=0x2,period=100000,event=0xb4",
	.desc = "Snoop invalidate requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.code",
	.event = "umask=0x4,period=2000000,event=0xb3",
	.desc = "Outstanding snoop code requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.code_not_empty",
	.event = "umask=0x4,period=2000000,cmask=1,event=0xb3",
	.desc = "Cycles snoop code requests queued",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.data",
	.event = "umask=0x1,period=2000000,event=0xb3",
	.desc = "Outstanding snoop data requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.data_not_empty",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xb3",
	.desc = "Cycles snoop data requests queued",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.invalidate",
	.event = "umask=0x2,period=2000000,event=0xb3",
	.desc = "Outstanding snoop invalidate requests",
	.topic = "other",
},
{
	.name = "snoopq_requests_outstanding.invalidate_not_empty",
	.event = "umask=0x2,period=2000000,cmask=1,event=0xb3",
	.desc = "Cycles snoop invalidate requests queued",
	.topic = "other",
},
{
	.name = "sq_full_stall_cycles",
	.event = "umask=0x1,period=2000000,event=0xf6",
	.desc = "Super Queue full stall cycles",
	.topic = "other",
},
{
	.name = "arith.cycles_div_busy",
	.event = "umask=0x1,period=2000000,event=0x14",
	.desc = "Cycles the divider is busy",
	.topic = "pipeline",
},
{
	.name = "arith.div",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,edge=1,event=0x14",
	.desc = "Divide Operations executed",
	.topic = "pipeline",
},
{
	.name = "arith.mul",
	.event = "umask=0x2,period=2000000,event=0x14",
	.desc = "Multiply operations executed",
	.topic = "pipeline",
},
{
	.name = "baclear.bad_target",
	.event = "umask=0x2,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted with bad target address",
	.topic = "pipeline",
},
{
	.name = "baclear.clear",
	.event = "umask=0x1,period=2000000,event=0xe6",
	.desc = "BACLEAR asserted, regardless of cause ",
	.topic = "pipeline",
},
{
	.name = "baclear_force_iq",
	.event = "umask=0x1,period=2000000,event=0xa7",
	.desc = "Instruction queue forced BACLEAR",
	.topic = "pipeline",
},
{
	.name = "br_inst_decoded",
	.event = "umask=0x1,period=2000000,event=0xe0",
	.desc = "Branch instructions decoded",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.any",
	.event = "umask=0x7f,period=200000,event=0x88",
	.desc = "Branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.cond",
	.event = "umask=0x1,period=200000,event=0x88",
	.desc = "Conditional branch instructions executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct",
	.event = "umask=0x2,period=200000,event=0x88",
	.desc = "Unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.direct_near_call",
	.event = "umask=0x10,period=20000,event=0x88",
	.desc = "Unconditional call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_near_call",
	.event = "umask=0x20,period=20000,event=0x88",
	.desc = "Indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.indirect_non_call",
	.event = "umask=0x4,period=20000,event=0x88",
	.desc = "Indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.near_calls",
	.event = "umask=0x30,period=20000,event=0x88",
	.desc = "Call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.non_calls",
	.event = "umask=0x7,period=200000,event=0x88",
	.desc = "All non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.return_near",
	.event = "umask=0x8,period=20000,event=0x88",
	.desc = "Indirect return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_exec.taken",
	.event = "umask=0x40,period=200000,event=0x88",
	.desc = "Taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x4,period=200000,event=0xc4",
	.desc = "Retired branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=200000,event=0xc4",
	.desc = "Retired conditional branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=20000,event=0xc4",
	.desc = "Retired near call instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.any",
	.event = "umask=0x7f,period=20000,event=0x89",
	.desc = "Mispredicted branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.cond",
	.event = "umask=0x1,period=20000,event=0x89",
	.desc = "Mispredicted conditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct",
	.event = "umask=0x2,period=20000,event=0x89",
	.desc = "Mispredicted unconditional branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.direct_near_call",
	.event = "umask=0x10,period=2000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_near_call",
	.event = "umask=0x20,period=2000,event=0x89",
	.desc = "Mispredicted indirect call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.indirect_non_call",
	.event = "umask=0x4,period=2000,event=0x89",
	.desc = "Mispredicted indirect non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.near_calls",
	.event = "umask=0x30,period=2000,event=0x89",
	.desc = "Mispredicted call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.non_calls",
	.event = "umask=0x7,period=20000,event=0x89",
	.desc = "Mispredicted non call branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.return_near",
	.event = "umask=0x8,period=2000,event=0x89",
	.desc = "Mispredicted return branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.taken",
	.event = "umask=0x40,period=20000,event=0x89",
	.desc = "Mispredicted taken branches executed",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x4,period=20000,event=0xc5",
	.desc = "Mispredicted retired branch instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=20000,event=0xc5",
	.desc = "Mispredicted conditional retired branches (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "br_misp_retired.near_call",
	.event = "umask=0x2,period=2000,event=0xc5",
	.desc = "Mispredicted near retired calls (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Reference cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_p",
	.event = "umask=0x1,period=100000,event=0x3c",
	.desc = "Reference base clock (133 Mhz) cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Cycles when thread is not halted (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000000,event=0x3c",
	.desc = "Cycles when thread is not halted (programmable counter)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.total_cycles",
	.event = "inv=1,umask=0x0,period=2000000,cmask=2,event=0x3c",
	.desc = "Total CPU cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.any",
	.event = "umask=0xf,period=2000000,event=0x87",
	.desc = "Any Instruction Length Decoder stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.iq_full",
	.event = "umask=0x4,period=2000000,event=0x87",
	.desc = "Instruction Queue full stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000000,event=0x87",
	.desc = "Length Change Prefix stall cycles",
	.topic = "pipeline",
},
{
	.name = "ild_stall.mru",
	.event = "umask=0x2,period=2000000,event=0x87",
	.desc = "Stall cycles due to BPU MRU bypass",
	.topic = "pipeline",
},
{
	.name = "ild_stall.regen",
	.event = "umask=0x8,period=2000000,event=0x87",
	.desc = "Regen stall cycles",
	.topic = "pipeline",
},
{
	.name = "inst_decoded.dec0",
	.event = "umask=0x1,period=2000000,event=0x18",
	.desc = "Instructions that must be decoded by decoder 0",
	.topic = "pipeline",
},
{
	.name = "inst_queue_write_cycles",
	.event = "umask=0x1,period=2000000,event=0x1e",
	.desc = "Cycles instructions are written to the instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_queue_writes",
	.event = "umask=0x1,period=2000000,event=0x17",
	.desc = "Instructions written to instruction queue",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (fixed counter)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (Programmable counter and Precise Event) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.mmx",
	.event = "umask=0x4,period=2000000,event=0xc0",
	.desc = "Retired MMX instructions (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000000,event=0xc0",
	.desc = "Retired floating-point operations (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre",
	.event = "umask=0x1,period=200000,event=0x4c",
	.desc = "Load operations conflicting with software prefetches",
	.topic = "pipeline",
},
{
	.name = "lsd.active",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles when uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd.inactive",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xa8",
	.desc = "Cycles no uops were delivered by the LSD",
	.topic = "pipeline",
},
{
	.name = "lsd_overflow",
	.event = "umask=0x1,period=2000000,event=0x20",
	.desc = "Loops that can't stream from the instruction queue",
	.topic = "pipeline",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=20000,event=0xc3",
	.desc = "Cycles machine clear asserted",
	.topic = "pipeline",
},
{
	.name = "machine_clears.mem_order",
	.event = "umask=0x2,period=20000,event=0xc3",
	.desc = "Execution pipeline restart due to Memory ordering conflicts",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=20000,event=0xc3",
	.desc = "Self-Modifying Code detected",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000000,event=0xa2",
	.desc = "Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.fpcw",
	.event = "umask=0x20,period=2000000,event=0xa2",
	.desc = "FPU control word write stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.load",
	.event = "umask=0x2,period=2000000,event=0xa2",
	.desc = "Load buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.mxcsr",
	.event = "umask=0x40,period=2000000,event=0xa2",
	.desc = "MXCSR rename stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.other",
	.event = "umask=0x80,period=2000000,event=0xa2",
	.desc = "Other Resource related stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rob_full",
	.event = "umask=0x10,period=2000000,event=0xa2",
	.desc = "ROB full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.rs_full",
	.event = "umask=0x4,period=2000000,event=0xa2",
	.desc = "Reservation Station full stall cycles",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.store",
	.event = "umask=0x8,period=2000000,event=0xa2",
	.desc = "Store buffer stall cycles",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_double",
	.event = "umask=0x4,period=200000,event=0xc7",
	.desc = "SIMD Packed-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.packed_single",
	.event = "umask=0x1,period=200000,event=0xc7",
	.desc = "SIMD Packed-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_double",
	.event = "umask=0x8,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Double Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.scalar_single",
	.event = "umask=0x2,period=200000,event=0xc7",
	.desc = "SIMD Scalar-Single Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "ssex_uops_retired.vector_integer",
	.event = "umask=0x10,period=200000,event=0xc7",
	.desc = "SIMD Vector Integer Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000000,event=0x3c",
	.desc = "Cycles thread is active",
	.topic = "pipeline",
},
{
	.name = "uop_unfusion",
	.event = "umask=0x1,period=2000000,event=0xdb",
	.desc = "Uop unfusions due to FP exceptions",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_folding",
	.event = "umask=0x4,period=2000000,event=0xd1",
	.desc = "Stack pointer instructions decoded",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.esp_sync",
	.event = "umask=0x8,period=2000000,event=0xd1",
	.desc = "Stack pointer sync operations",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.ms_cycles_active",
	.event = "umask=0x2,period=2000000,cmask=1,event=0xd1",
	.desc = "Uops decoded by Microcode Sequencer",
	.topic = "pipeline",
},
{
	.name = "uops_decoded.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xd1",
	.desc = "Cycles no Uops are decoded",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles",
	.event = "umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_active_cycles_no_port5",
	.event = "umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count",
	.event = "inv=1,umask=0x3f,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_count_no_port5",
	.event = "inv=1,umask=0x1f,period=2000000,cmask=1,edge=1,event=0xb1",
	.desc = "Uops executed on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles",
	.event = "inv=1,umask=0x3f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on any port (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_stall_cycles_no_port5",
	.event = "inv=1,umask=0x1f,any=1,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0-4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port0",
	.event = "umask=0x1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015",
	.event = "umask=0x40,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port015_stall_cycles",
	.event = "inv=1,umask=0x40,period=2000000,cmask=1,event=0xb1",
	.desc = "Cycles no Uops issued on ports 0, 1 or 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port1",
	.event = "umask=0x2,period=2000000,event=0xb1",
	.desc = "Uops executed on port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port2_core",
	.event = "umask=0x4,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 2 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port234_core",
	.event = "umask=0x80,any=1,period=2000000,event=0xb1",
	.desc = "Uops issued on ports 2, 3 or 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port3_core",
	.event = "umask=0x8,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 3 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port4_core",
	.event = "umask=0x10,any=1,period=2000000,event=0xb1",
	.desc = "Uops executed on port 4 (core count)",
	.topic = "pipeline",
},
{
	.name = "uops_executed.port5",
	.event = "umask=0x20,period=2000000,event=0xb1",
	.desc = "Uops executed on port 5",
	.topic = "pipeline",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000000,event=0xe",
	.desc = "Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.core_stall_cycles",
	.event = "inv=1,umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued on any thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.cycles_all_threads",
	.event = "umask=0x1,any=1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles Uops were issued on either thread",
	.topic = "pipeline",
},
{
	.name = "uops_issued.fused",
	.event = "umask=0x2,period=2000000,event=0xe",
	.desc = "Fused Uops issued",
	.topic = "pipeline",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xe",
	.desc = "Cycles no Uops were issued",
	.topic = "pipeline",
},
{
	.name = "uops_retired.active_cycles",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are being retired (Precise event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.any",
	.event = "umask=0x1,period=2000000,event=0xc2",
	.desc = "Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.macro_fused",
	.event = "umask=0x4,period=2000000,event=0xc2",
	.desc = "Macro-fused Uops retired (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000000,event=0xc2",
	.desc = "Retirement slots used (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=1,event=0xc2",
	.desc = "Cycles Uops are not retiring (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc2",
	.desc = "Total cycles using precise uop retired event (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.total_cycles_ps",
	.event = "inv=1,umask=0x1,period=2000000,cmask=16,event=0xc0",
	.desc = "Total cycles (Precise Event)",
	.topic = "pipeline",
},
{
	.name = "dtlb_load_misses.any",
	.event = "umask=0x1,period=200000,event=0x8",
	.desc = "DTLB load misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.large_walk_completed",
	.event = "umask=0x80,period=200000,event=0x8",
	.desc = "DTLB load miss large page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.pde_miss",
	.event = "umask=0x20,period=200000,event=0x8",
	.desc = "DTLB load miss caused by low part of address",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x10,period=2000000,event=0x8",
	.desc = "DTLB second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x8",
	.desc = "DTLB load miss page walks complete",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_cycles",
	.event = "umask=0x4,period=200000,event=0x8",
	.desc = "DTLB load miss page walk cycles",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x49",
	.desc = "DTLB misses",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.large_walk_completed",
	.event = "umask=0x80,period=200000,event=0x49",
	.desc = "DTLB miss large page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.pde_miss",
	.event = "umask=0x20,period=200000,event=0x49",
	.desc = "DTLB misses caused by low part of address. Count also includes 2M page references because 2M pages do not use the PDE",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.stlb_hit",
	.event = "umask=0x10,period=200000,event=0x49",
	.desc = "DTLB first level misses but second level hit",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x49",
	.desc = "DTLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_misses.walk_cycles",
	.event = "umask=0x4,period=2000000,event=0x49",
	.desc = "DTLB miss page walk cycles",
	.topic = "virtual memory",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000000,event=0x4f",
	.desc = "Extended Page Table walk cycles",
	.topic = "virtual memory",
},
{
	.name = "itlb_flush",
	.event = "umask=0x1,period=2000000,event=0xae",
	.desc = "ITLB flushes",
	.topic = "virtual memory",
},
{
	.name = "itlb_miss_retired",
	.event = "umask=0x20,period=200000,event=0xc8",
	.desc = "Retired instructions that missed the ITLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.any",
	.event = "umask=0x1,period=200000,event=0x85",
	.desc = "ITLB miss",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.large_walk_completed",
	.event = "umask=0x80,period=200000,event=0x85",
	.desc = "ITLB miss large page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0x2,period=200000,event=0x85",
	.desc = "ITLB miss page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_cycles",
	.event = "umask=0x4,period=2000000,event=0x85",
	.desc = "ITLB miss page walk cycles",
	.topic = "virtual memory",
},
{
	.name = "mem_load_retired.dtlb_miss",
	.event = "umask=0x80,period=200000,event=0xcb",
	.desc = "Retired loads that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = "mem_store_retired.dtlb_miss",
	.event = "umask=0x1,period=200000,event=0xc",
	.desc = "Retired stores that miss the DTLB (Precise Event)",
	.topic = "virtual memory",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_amdzen1[] = {
{
	.name = "bp_l1_btb_correct",
	.event = "event=0x8a",
	.desc = "L1 BTB Correction",
	.topic = "branch",
},
{
	.name = "bp_l2_btb_correct",
	.event = "event=0x8b",
	.desc = "L2 BTB Correction",
	.topic = "branch",
},
{
	.name = "bp_dyn_ind_pred",
	.event = "event=0x8e",
	.desc = "Dynamic Indirect Predictions",
	.topic = "branch",
	.long_desc = "Indirect Branch Prediction for potential multi-target branch (speculative)",
},
{
	.name = "bp_de_redirect",
	.event = "event=0x91",
	.desc = "Decoder Overrides Existing Branch Prediction (speculative)",
	.topic = "branch",
},
{
	.name = "ic_fw32",
	.event = "event=0x80",
	.desc = "The number of 32B fetch windows transferred from IC pipe to DE instruction decoder (includes non-cacheable and cacheable fill responses)",
	.topic = "cache",
},
{
	.name = "ic_fw32_miss",
	.event = "event=0x81",
	.desc = "The number of 32B fetch windows tried to read the L1 IC and missed in the full tag",
	.topic = "cache",
},
{
	.name = "ic_cache_fill_l2",
	.event = "event=0x82",
	.desc = "The number of 64 byte instruction cache line was fulfilled from the L2 cache",
	.topic = "cache",
},
{
	.name = "ic_cache_fill_sys",
	.event = "event=0x83",
	.desc = "The number of 64 byte instruction cache line fulfilled from system memory or another cache",
	.topic = "cache",
},
{
	.name = "bp_l1_tlb_miss_l2_hit",
	.event = "event=0x84",
	.desc = "The number of instruction fetches that miss in the L1 ITLB but hit in the L2 ITLB",
	.topic = "cache",
},
{
	.name = "bp_l1_tlb_miss_l2_miss",
	.event = "event=0x85",
	.desc = "The number of instruction fetches that miss in both the L1 and L2 TLBs",
	.topic = "cache",
},
{
	.name = "bp_snp_re_sync",
	.event = "event=0x86",
	.desc = "The number of pipeline restarts caused by invalidating probes that hit on the instruction stream currently being executed. This would happen if the active instruction stream was being modified by another processor in an MP system - typically a highly unlikely event",
	.topic = "cache",
},
{
	.name = "ic_fetch_stall.ic_stall_any",
	.event = "umask=0x4,event=0x87",
	.desc = "Instruction Pipe Stall. IC pipe was stalled during this clock cycle for any reason (nothing valid in pipe ICM1)",
	.topic = "cache",
},
{
	.name = "ic_fetch_stall.ic_stall_dq_empty",
	.event = "umask=0x2,event=0x87",
	.desc = "Instruction Pipe Stall. IC pipe was stalled during this clock cycle (including IC to OC fetches) due to DQ empty",
	.topic = "cache",
},
{
	.name = "ic_fetch_stall.ic_stall_back_pressure",
	.event = "umask=0x1,event=0x87",
	.desc = "Instruction Pipe Stall. IC pipe was stalled during this clock cycle (including IC to OC fetches) due to back-pressure",
	.topic = "cache",
},
{
	.name = "ic_cache_inval.l2_invalidating_probe",
	.event = "umask=0x2,event=0x8c",
	.desc = "IC line invalidated due to L2 invalidating probe (external or LS). The number of instruction cache lines invalidated. A non-SMC event is CMC (cross modifying code), either from the other thread of the core or another core",
	.topic = "cache",
},
{
	.name = "ic_cache_inval.fill_invalidated",
	.event = "umask=0x1,event=0x8c",
	.desc = "IC line invalidated due to overwriting fill response. The number of instruction cache lines invalidated. A non-SMC event is CMC (cross modifying code), either from the other thread of the core or another core",
	.topic = "cache",
},
{
	.name = "bp_tlb_rel",
	.event = "event=0x99",
	.desc = "The number of ITLB reload requests",
	.topic = "cache",
},
{
	.name = "l2_request_g1.rd_blk_l",
	.event = "umask=0x80,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Data cache reads (including hardware and software prefetch)",
	.topic = "cache",
},
{
	.name = "l2_request_g1.rd_blk_x",
	.event = "umask=0x40,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Data cache stores",
	.topic = "cache",
},
{
	.name = "l2_request_g1.ls_rd_blk_c_s",
	.event = "umask=0x20,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Data cache shared reads",
	.topic = "cache",
},
{
	.name = "l2_request_g1.cacheable_ic_read",
	.event = "umask=0x10,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Instruction cache reads",
	.topic = "cache",
},
{
	.name = "l2_request_g1.change_to_x",
	.event = "umask=0x8,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Data cache state change requests. Request change to writable, check L2 for current state",
	.topic = "cache",
},
{
	.name = "l2_request_g1.prefetch_l2_cmd",
	.event = "umask=0x4,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). PrefetchL2Cmd",
	.topic = "cache",
},
{
	.name = "l2_request_g1.l2_hw_pf",
	.event = "umask=0x2,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). L2 Prefetcher. All prefetches accepted by L2 pipeline, hit or miss. Types of PF and L2 hit/miss broken out in a separate perfmon event",
	.topic = "cache",
},
{
	.name = "l2_request_g1.group2",
	.event = "umask=0x1,event=0x60",
	.desc = "Miscellaneous events covered in more detail by l2_request_g2 (PMCx061)",
	.topic = "cache",
},
{
	.name = "l2_request_g2.group1",
	.event = "umask=0x80,event=0x61",
	.desc = "Miscellaneous events covered in more detail by l2_request_g1 (PMCx060)",
	.topic = "cache",
},
{
	.name = "l2_request_g2.ls_rd_sized",
	.event = "umask=0x40,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Data cache read sized",
	.topic = "cache",
},
{
	.name = "l2_request_g2.ls_rd_sized_nc",
	.event = "umask=0x20,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Data cache read sized non-cacheable",
	.topic = "cache",
},
{
	.name = "l2_request_g2.ic_rd_sized",
	.event = "umask=0x10,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Instruction cache read sized",
	.topic = "cache",
},
{
	.name = "l2_request_g2.ic_rd_sized_nc",
	.event = "umask=0x8,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Instruction cache read sized non-cacheable",
	.topic = "cache",
},
{
	.name = "l2_request_g2.smc_inval",
	.event = "umask=0x4,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Self-modifying code invalidates",
	.topic = "cache",
},
{
	.name = "l2_request_g2.bus_locks_originator",
	.event = "umask=0x2,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Bus locks",
	.topic = "cache",
},
{
	.name = "l2_request_g2.bus_locks_responses",
	.event = "umask=0x1,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Bus lock response",
	.topic = "cache",
},
{
	.name = "l2_latency.l2_cycles_waiting_on_fills",
	.event = "umask=0x1,event=0x62",
	.desc = "Total cycles spent waiting for L2 fills to complete from L3 or memory, divided by four. Event counts are for both threads. To calculate average latency, the number of fills from both threads must be used",
	.topic = "cache",
},
{
	.name = "l2_wcb_req.wcb_write",
	.event = "umask=0x40,event=0x63",
	.desc = "LS to L2 WCB write requests. LS (Load/Store unit) to L2 WCB (Write Combining Buffer) write requests",
	.topic = "cache",
},
{
	.name = "l2_wcb_req.wcb_close",
	.event = "umask=0x20,event=0x63",
	.desc = "LS to L2 WCB close requests. LS (Load/Store unit) to L2 WCB (Write Combining Buffer) close requests",
	.topic = "cache",
},
{
	.name = "l2_wcb_req.zero_byte_store",
	.event = "umask=0x4,event=0x63",
	.desc = "LS to L2 WCB zero byte store requests. LS (Load/Store unit) to L2 WCB (Write Combining Buffer) zero byte store requests",
	.topic = "cache",
},
{
	.name = "l2_wcb_req.cl_zero",
	.event = "umask=0x1,event=0x63",
	.desc = "LS to L2 WCB cache line zeroing requests. LS (Load/Store unit) to L2 WCB (Write Combining Buffer) cache line zeroing requests",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_cs",
	.event = "umask=0x80,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache shared read hit in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_l_hit_x",
	.event = "umask=0x40,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache read hit in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_l_hit_s",
	.event = "umask=0x20,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache read hit on shared line in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_x",
	.event = "umask=0x10,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache store or state change hit in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_c",
	.event = "umask=0x8,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache request miss in L2 (all types)",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ic_fill_hit_x",
	.event = "umask=0x4,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Instruction cache hit modifiable line in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ic_fill_hit_s",
	.event = "umask=0x2,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Instruction cache hit clean line in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ic_fill_miss",
	.event = "umask=0x1,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Instruction cache request miss in L2",
	.topic = "cache",
},
{
	.name = "l2_fill_pending.l2_fill_busy",
	.event = "umask=0x1,event=0x6d",
	.desc = "Cycles with fill pending from L2. Total cycles spent with one or more fill requests in flight from L2",
	.topic = "cache",
},
{
	.name = "l3_request_g1.caching_l3_cache_accesses",
	.event = "umask=0x80,event=0x1",
	.desc = "Caching: L3 cache accesses. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "l3_lookup_state.all_l3_req_typs",
	.event = "umask=0xff,event=0x4",
	.desc = "All L3 Request Types. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "l3_comb_clstr_state.other_l3_miss_typs",
	.event = "umask=0xfe,event=0x6",
	.desc = "Other L3 Miss Request Types. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "l3_comb_clstr_state.request_miss",
	.event = "umask=0x01,event=0x6",
	.desc = "L3 cache misses. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "xi_sys_fill_latency",
	.event = "umask=0x00,event=0x90",
	.desc = "L3 Cache Miss Latency. Total cycles for all transactions divided by 16. Ignores SliceMask and ThreadMask. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "xi_ccx_sdp_req1.all_l3_miss_req_typs",
	.event = "umask=0x3f,event=0x9a",
	.desc = "All L3 Miss Request Types. Ignores SliceMask and ThreadMask. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "ex_ret_instr",
	.event = "event=0xc0",
	.desc = "Retired Instructions",
	.topic = "core",
},
{
	.name = "ex_ret_cops",
	.event = "event=0xc1",
	.desc = "Retired Uops",
	.topic = "core",
	.long_desc = "The number of uOps retired. This includes all processor activity (instructions, exceptions, interrupts, microcode assists, etc.). The number of events logged per cycle can vary from 0 to 4",
},
{
	.name = "ex_ret_brn",
	.event = "event=0xc2",
	.desc = "Retired Branch Instructions",
	.topic = "core",
	.long_desc = "The number of branch instructions retired. This includes all types of architectural control flow changes, including exceptions and interrupts",
},
{
	.name = "ex_ret_brn_misp",
	.event = "event=0xc3",
	.desc = "Retired Branch Instructions Mispredicted",
	.topic = "core",
	.long_desc = "The number of branch instructions retired, of any type, that were not correctly predicted. This includes those for which prediction is not attempted (far control transfers, exceptions and interrupts)",
},
{
	.name = "ex_ret_brn_tkn",
	.event = "event=0xc4",
	.desc = "Retired Taken Branch Instructions",
	.topic = "core",
	.long_desc = "The number of taken branches that were retired. This includes all types of architectural control flow changes, including exceptions and interrupts",
},
{
	.name = "ex_ret_brn_tkn_misp",
	.event = "event=0xc5",
	.desc = "Retired Taken Branch Instructions Mispredicted",
	.topic = "core",
	.long_desc = "The number of retired taken branch instructions that were mispredicted",
},
{
	.name = "ex_ret_brn_far",
	.event = "event=0xc6",
	.desc = "Retired Far Control Transfers",
	.topic = "core",
	.long_desc = "The number of far control transfers retired including far call/jump/return, IRET, SYSCALL and SYSRET, plus exceptions and interrupts. Far control transfers are not subject to branch prediction",
},
{
	.name = "ex_ret_brn_resync",
	.event = "event=0xc7",
	.desc = "Retired Branch Resyncs",
	.topic = "core",
	.long_desc = "The number of resync branches. These reflect pipeline restarts due to certain microcode assists and events such as writes to the active instruction stream, among other things. Each occurrence reflects a restart penalty similar to a branch mispredict. This is relatively rare",
},
{
	.name = "ex_ret_near_ret",
	.event = "event=0xc8",
	.desc = "Retired Near Returns",
	.topic = "core",
	.long_desc = "The number of near return instructions (RET or RET Iw) retired",
},
{
	.name = "ex_ret_near_ret_mispred",
	.event = "event=0xc9",
	.desc = "Retired Near Returns Mispredicted",
	.topic = "core",
	.long_desc = "The number of near returns retired that were not correctly predicted by the return address predictor. Each such mispredict incurs the same penalty as a mispredicted conditional branch instruction",
},
{
	.name = "ex_ret_brn_ind_misp",
	.event = "event=0xca",
	.desc = "Retired Indirect Branch Instructions Mispredicted",
	.topic = "core",
},
{
	.name = "ex_ret_mmx_fp_instr.sse_instr",
	.event = "umask=0x4,event=0xcb",
	.desc = "SSE instructions (SSE, SSE2, SSE3, SSSE3, SSE4A, SSE41, SSE42, AVX)",
	.topic = "core",
	.long_desc = "The number of MMX, SSE or x87 instructions retired. The UnitMask allows the selection of the individual classes of instructions as given in the table. Each increment represents one complete instruction. Since this event includes non-numeric instructions it is not suitable for measuring MFLOPS. SSE instructions (SSE, SSE2, SSE3, SSSE3, SSE4A, SSE41, SSE42, AVX)",
},
{
	.name = "ex_ret_mmx_fp_instr.mmx_instr",
	.event = "umask=0x2,event=0xcb",
	.desc = "MMX instructions",
	.topic = "core",
	.long_desc = "The number of MMX, SSE or x87 instructions retired. The UnitMask allows the selection of the individual classes of instructions as given in the table. Each increment represents one complete instruction. Since this event includes non-numeric instructions it is not suitable for measuring MFLOPS. MMX instructions",
},
{
	.name = "ex_ret_mmx_fp_instr.x87_instr",
	.event = "umask=0x1,event=0xcb",
	.desc = "x87 instructions",
	.topic = "core",
	.long_desc = "The number of MMX, SSE or x87 instructions retired. The UnitMask allows the selection of the individual classes of instructions as given in the table. Each increment represents one complete instruction. Since this event includes non-numeric instructions it is not suitable for measuring MFLOPS. x87 instructions",
},
{
	.name = "ex_ret_cond",
	.event = "event=0xd1",
	.desc = "Retired Conditional Branch Instructions",
	.topic = "core",
},
{
	.name = "ex_div_busy",
	.event = "event=0xd3",
	.desc = "Div Cycles Busy count",
	.topic = "core",
},
{
	.name = "ex_div_count",
	.event = "event=0xd4",
	.desc = "Div Op Count",
	.topic = "core",
},
{
	.name = "ex_tagged_ibs_ops.ibs_count_rollover",
	.event = "umask=0x4,event=0x1cf",
	.desc = "Tagged IBS Ops. Number of times an op could not be tagged by IBS because of a previous tagged op that has not retired",
	.topic = "core",
},
{
	.name = "ex_tagged_ibs_ops.ibs_tagged_ops_ret",
	.event = "umask=0x2,event=0x1cf",
	.desc = "Tagged IBS Ops. Number of Ops tagged by IBS that retired",
	.topic = "core",
},
{
	.name = "ex_tagged_ibs_ops.ibs_tagged_ops",
	.event = "umask=0x1,event=0x1cf",
	.desc = "Tagged IBS Ops. Number of Ops tagged by IBS",
	.topic = "core",
},
{
	.name = "ex_ret_fus_brnch_inst",
	.event = "event=0x1d0",
	.desc = "The number of fused retired branch instructions retired per cycle. The number of events logged per cycle can vary from 0 to 3",
	.topic = "core",
},
{
	.name = "fpu_pipe_assignment.dual",
	.event = "umask=0xf0,event=0",
	.desc = "Total number multi-pipe uOps assigned to all pipes",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) and dual-pipe uOps dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number multi-pipe uOps assigned to all pipes",
},
{
	.name = "fpu_pipe_assignment.dual3",
	.event = "umask=0x80,event=0",
	.desc = "Total number multi-pipe uOps assigned to pipe 3",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) and dual-pipe uOps dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number multi-pipe uOps assigned to pipe 3",
},
{
	.name = "fpu_pipe_assignment.dual2",
	.event = "umask=0x40,event=0",
	.desc = "Total number multi-pipe uOps assigned to pipe 2",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) and dual-pipe uOps dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number multi-pipe uOps assigned to pipe 2",
},
{
	.name = "fpu_pipe_assignment.dual1",
	.event = "umask=0x20,event=0",
	.desc = "Total number multi-pipe uOps assigned to pipe 1",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) and dual-pipe uOps dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number multi-pipe uOps assigned to pipe 1",
},
{
	.name = "fpu_pipe_assignment.dual0",
	.event = "umask=0x10,event=0",
	.desc = "Total number multi-pipe uOps assigned to pipe 0",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) and dual-pipe uOps dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number multi-pipe uOps assigned to pipe 0",
},
{
	.name = "fpu_pipe_assignment.total",
	.event = "umask=0xf,event=0",
	.desc = "Total number uOps assigned to all fpu pipes",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) and dual-pipe uOps dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number uOps assigned to all pipes",
},
{
	.name = "fpu_pipe_assignment.total3",
	.event = "umask=0x8,event=0",
	.desc = "Total number of fp uOps on pipe 3",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one-cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number uOps assigned to pipe 3",
},
{
	.name = "fpu_pipe_assignment.total2",
	.event = "umask=0x4,event=0",
	.desc = "Total number of fp uOps on pipe 2",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number uOps assigned to pipe 2",
},
{
	.name = "fpu_pipe_assignment.total1",
	.event = "umask=0x2,event=0",
	.desc = "Total number of fp uOps on pipe 1",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number uOps assigned to pipe 1",
},
{
	.name = "fpu_pipe_assignment.total0",
	.event = "umask=0x1,event=0",
	.desc = "Total number of fp uOps  on pipe 0",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number uOps assigned to pipe 0",
},
{
	.name = "fp_sched_empty",
	.event = "event=0x1",
	.desc = "This is a speculative event. The number of cycles in which the FPU scheduler is empty. Note that some Ops like FP loads bypass the scheduler",
	.topic = "floating point",
},
{
	.name = "fp_retx87_fp_ops.all",
	.event = "umask=0x7,event=0x2",
	.desc = "All Ops",
	.topic = "floating point",
	.long_desc = "The number of x87 floating-point Ops that have retired. The number of events logged per cycle can vary from 0 to 8",
},
{
	.name = "fp_retx87_fp_ops.div_sqr_r_ops",
	.event = "umask=0x4,event=0x2",
	.desc = "Divide and square root Ops",
	.topic = "floating point",
	.long_desc = "The number of x87 floating-point Ops that have retired. The number of events logged per cycle can vary from 0 to 8. Divide and square root Ops",
},
{
	.name = "fp_retx87_fp_ops.mul_ops",
	.event = "umask=0x2,event=0x2",
	.desc = "Multiply Ops",
	.topic = "floating point",
	.long_desc = "The number of x87 floating-point Ops that have retired. The number of events logged per cycle can vary from 0 to 8. Multiply Ops",
},
{
	.name = "fp_retx87_fp_ops.add_sub_ops",
	.event = "umask=0x1,event=0x2",
	.desc = "Add/subtract Ops",
	.topic = "floating point",
	.long_desc = "The number of x87 floating-point Ops that have retired. The number of events logged per cycle can vary from 0 to 8. Add/subtract Ops",
},
{
	.name = "fp_ret_sse_avx_ops.all",
	.event = "umask=0xff,event=0x3",
	.desc = "All FLOPS",
	.topic = "floating point",
	.long_desc = "This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15",
},
{
	.name = "fp_ret_sse_avx_ops.dp_mult_add_flops",
	.event = "umask=0x80,event=0x3",
	.desc = "Double precision multiply-add FLOPS. Multiply-add counts as 2 FLOPS",
	.topic = "floating point",
	.long_desc = "This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15. Double precision multiply-add FLOPS. Multiply-add counts as 2 FLOPS",
},
{
	.name = "fp_ret_sse_avx_ops.dp_div_flops",
	.event = "umask=0x40,event=0x3",
	.desc = "Double precision divide/square root FLOPS",
	.topic = "floating point",
	.long_desc = "This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15. Double precision divide/square root FLOPS",
},
{
	.name = "fp_ret_sse_avx_ops.dp_mult_flops",
	.event = "umask=0x20,event=0x3",
	.desc = "Double precision multiply FLOPS",
	.topic = "floating point",
	.long_desc = "This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15. Double precision multiply FLOPS",
},
{
	.name = "fp_ret_sse_avx_ops.dp_add_sub_flops",
	.event = "umask=0x10,event=0x3",
	.desc = "Double precision add/subtract FLOPS",
	.topic = "floating point",
	.long_desc = "This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15. Double precision add/subtract FLOPS",
},
{
	.name = "fp_ret_sse_avx_ops.sp_mult_add_flops",
	.event = "umask=0x8,event=0x3",
	.desc = "Single precision multiply-add FLOPS. Multiply-add counts as 2 FLOPS",
	.topic = "floating point",
	.long_desc = "This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15. Single precision multiply-add FLOPS. Multiply-add counts as 2 FLOPS",
},
{
	.name = "fp_ret_sse_avx_ops.sp_div_flops",
	.event = "umask=0x4,event=0x3",
	.desc = "Single-precision divide/square root FLOPS",
	.topic = "floating point",
	.long_desc = "This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15. Single-precision divide/square root FLOPS",
},
{
	.name = "fp_ret_sse_avx_ops.sp_mult_flops",
	.event = "umask=0x2,event=0x3",
	.desc = "Single-precision multiply FLOPS",
	.topic = "floating point",
	.long_desc = "This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15. Single-precision multiply FLOPS",
},
{
	.name = "fp_ret_sse_avx_ops.sp_add_sub_flops",
	.event = "umask=0x1,event=0x3",
	.desc = "Single-precision add/subtract FLOPS",
	.topic = "floating point",
	.long_desc = "This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15. Single-precision add/subtract FLOPS",
},
{
	.name = "fp_num_mov_elim_scal_op.optimized",
	.event = "umask=0x8,event=0x4",
	.desc = "Number of Scalar Ops optimized",
	.topic = "floating point",
	.long_desc = "This is a dispatch based speculative event, and is useful for measuring the effectiveness of the Move elimination and Scalar code optimization schemes. Number of Scalar Ops optimized",
},
{
	.name = "fp_num_mov_elim_scal_op.opt_potential",
	.event = "umask=0x4,event=0x4",
	.desc = "Number of Ops that are candidates for optimization (have Z-bit either set or pass)",
	.topic = "floating point",
	.long_desc = "This is a dispatch based speculative event, and is useful for measuring the effectiveness of the Move elimination and Scalar code optimization schemes. Number of Ops that are candidates for optimization (have Z-bit either set or pass)",
},
{
	.name = "fp_num_mov_elim_scal_op.sse_mov_ops_elim",
	.event = "umask=0x2,event=0x4",
	.desc = "Number of SSE Move Ops eliminated",
	.topic = "floating point",
	.long_desc = "This is a dispatch based speculative event, and is useful for measuring the effectiveness of the Move elimination and Scalar code optimization schemes. Number of SSE Move Ops eliminated",
},
{
	.name = "fp_num_mov_elim_scal_op.sse_mov_ops",
	.event = "umask=0x1,event=0x4",
	.desc = "Number of SSE Move Ops",
	.topic = "floating point",
	.long_desc = "This is a dispatch based speculative event, and is useful for measuring the effectiveness of the Move elimination and Scalar code optimization schemes. Number of SSE Move Ops",
},
{
	.name = "fp_retired_ser_ops.x87_ctrl_ret",
	.event = "umask=0x8,event=0x5",
	.desc = "x87 control word mispredict traps due to mispredictions in RC or PC, or changes in mask bits",
	.topic = "floating point",
	.long_desc = "The number of serializing Ops retired. x87 control word mispredict traps due to mispredictions in RC or PC, or changes in mask bits",
},
{
	.name = "fp_retired_ser_ops.x87_bot_ret",
	.event = "umask=0x4,event=0x5",
	.desc = "x87 bottom-executing uOps retired",
	.topic = "floating point",
	.long_desc = "The number of serializing Ops retired. x87 bottom-executing uOps retired",
},
{
	.name = "fp_retired_ser_ops.sse_ctrl_ret",
	.event = "umask=0x2,event=0x5",
	.desc = "SSE control word mispredict traps due to mispredictions in RC, FTZ or DAZ, or changes in mask bits",
	.topic = "floating point",
	.long_desc = "The number of serializing Ops retired. SSE control word mispredict traps due to mispredictions in RC, FTZ or DAZ, or changes in mask bits",
},
{
	.name = "fp_retired_ser_ops.sse_bot_ret",
	.event = "umask=0x1,event=0x5",
	.desc = "SSE bottom-executing uOps retired",
	.topic = "floating point",
	.long_desc = "The number of serializing Ops retired. SSE bottom-executing uOps retired",
},
{
	.name = "ls_locks.bus_lock",
	.event = "umask=0x1,event=0x25",
	.desc = "Bus lock when a locked operations crosses a cache boundary or is done on an uncacheable memory type",
	.topic = "memory",
},
{
	.name = "ls_dispatch.ld_st_dispatch",
	.event = "umask=0x4,event=0x29",
	.desc = "Counts the number of operations dispatched to the LS unit. Unit Masks ADDed. Load-op-Stores",
	.topic = "memory",
},
{
	.name = "ls_dispatch.store_dispatch",
	.event = "umask=0x2,event=0x29",
	.desc = "Counts the number of stores dispatched to the LS unit. Unit Masks ADDed",
	.topic = "memory",
},
{
	.name = "ls_dispatch.ld_dispatch",
	.event = "umask=0x1,event=0x29",
	.desc = "Counts the number of loads dispatched to the LS unit. Unit Masks ADDed",
	.topic = "memory",
},
{
	.name = "ls_stlf",
	.event = "event=0x35",
	.desc = "Number of STLF hits",
	.topic = "memory",
},
{
	.name = "ls_dc_accesses",
	.event = "event=0x40",
	.desc = "The number of accesses to the data cache for load and store references. This may include certain microcode scratchpad accesses, although these are generally rare. Each increment represents an eight-byte access, although the instruction may only be accessing a portion of that. This event is a speculative event",
	.topic = "memory",
},
{
	.name = "ls_mab_alloc.dc_prefetcher",
	.event = "umask=0x8,event=0x41",
	.desc = "LS MAB allocates by type - DC prefetcher",
	.topic = "memory",
},
{
	.name = "ls_mab_alloc.stores",
	.event = "umask=0x2,event=0x41",
	.desc = "LS MAB allocates by type - stores",
	.topic = "memory",
},
{
	.name = "ls_mab_alloc.loads",
	.event = "umask=0x01,event=0x41",
	.desc = "LS MAB allocates by type - loads",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.all",
	.event = "umask=0xff,event=0x45",
	.desc = "L1 DTLB Miss or Reload off all sizes",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_1g_l2_miss",
	.event = "umask=0x80,event=0x45",
	.desc = "L1 DTLB Miss of a page of 1G size",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_2m_l2_miss",
	.event = "umask=0x40,event=0x45",
	.desc = "L1 DTLB Miss of a page of 2M size",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_32k_l2_miss",
	.event = "umask=0x20,event=0x45",
	.desc = "L1 DTLB Miss of a page of 32K size",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_4k_l2_miss",
	.event = "umask=0x10,event=0x45",
	.desc = "L1 DTLB Miss of a page of 4K size",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_1g_l2_hit",
	.event = "umask=0x8,event=0x45",
	.desc = "L1 DTLB Reload of a page of 1G size",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_2m_l2_hit",
	.event = "umask=0x4,event=0x45",
	.desc = "L1 DTLB Reload of a page of 2M size",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_32k_l2_hit",
	.event = "umask=0x2,event=0x45",
	.desc = "L1 DTLB Reload of a page of 32K size",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_4k_l2_hit",
	.event = "umask=0x1,event=0x45",
	.desc = "L1 DTLB Reload of a page of 4K size",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.iside",
	.event = "umask=0xc,event=0x46",
	.desc = "Total Page Table Walks on I-side",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.ic_type1",
	.event = "umask=0x8,event=0x46",
	.desc = "Total Page Table Walks IC Type 1",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.ic_type0",
	.event = "umask=0x4,event=0x46",
	.desc = "Total Page Table Walks IC Type 0",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.dside",
	.event = "umask=0x3,event=0x46",
	.desc = "Total Page Table Walks on D-side",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.dc_type1",
	.event = "umask=0x2,event=0x46",
	.desc = "Total Page Table Walks DC Type 1",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.dc_type0",
	.event = "umask=0x1,event=0x46",
	.desc = "Total Page Table Walks DC Type 0",
	.topic = "memory",
},
{
	.name = "ls_misal_accesses",
	.event = "event=0x47",
	.desc = "Misaligned loads",
	.topic = "memory",
},
{
	.name = "ls_pref_instr_disp.prefetch_nta",
	.event = "umask=0x4,event=0x4b",
	.desc = "Software Prefetch Instructions (PREFETCHNTA instruction) Dispatched",
	.topic = "memory",
},
{
	.name = "ls_pref_instr_disp.store_prefetch_w",
	.event = "umask=0x2,event=0x4b",
	.desc = "Software Prefetch Instructions (3DNow PREFETCHW instruction) Dispatched",
	.topic = "memory",
},
{
	.name = "ls_pref_instr_disp.load_prefetch_w",
	.event = "umask=0x1,event=0x4b",
	.desc = "Software Prefetch Instructions Dispatched. Prefetch, Prefetch_T0_T1_T2",
	.topic = "memory",
},
{
	.name = "ls_inef_sw_pref.mab_mch_cnt",
	.event = "umask=0x2,event=0x52",
	.desc = "The number of software prefetches that did not fetch data outside of the processor core. Software PREFETCH instruction saw a match on an already-allocated miss request buffer",
	.topic = "memory",
},
{
	.name = "ls_inef_sw_pref.data_pipe_sw_pf_dc_hit",
	.event = "umask=0x1,event=0x52",
	.desc = "The number of software prefetches that did not fetch data outside of the processor core. Software PREFETCH instruction saw a DC hit",
	.topic = "memory",
},
{
	.name = "ls_not_halted_cyc",
	.event = "event=0x76",
	.desc = "Cycles not in Halt",
	.topic = "memory",
},
{
	.name = "ic_oc_mode_switch.oc_ic_mode_switch",
	.event = "umask=0x2,event=0x28a",
	.desc = "OC Mode Switch. OC to IC mode switch",
	.topic = "other",
},
{
	.name = "ic_oc_mode_switch.ic_oc_mode_switch",
	.event = "umask=0x1,event=0x28a",
	.desc = "OC Mode Switch. IC to OC mode switch",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.retire_token_stall",
	.event = "umask=0x40,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. RETIRE Tokens unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.agsq_token_stall",
	.event = "umask=0x20,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. AGSQ Tokens unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.alu_token_stall",
	.event = "umask=0x10,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. ALU tokens total unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.alsq3_0_token_stall",
	.event = "umask=0x8,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. ALSQ 3_0 Tokens unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.alsq3_token_stall",
	.event = "umask=0x4,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. ALSQ 3 Tokens unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.alsq2_token_stall",
	.event = "umask=0x2,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. ALSQ 2 Tokens unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.alsq1_token_stall",
	.event = "umask=0x1,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. ALSQ 1 Tokens unavailable",
	.topic = "other",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_amdzen2[] = {
{
	.name = "bp_l1_btb_correct",
	.event = "event=0x8a",
	.desc = "L1 Branch Prediction Overrides Existing Prediction (speculative)",
	.topic = "branch",
},
{
	.name = "bp_l2_btb_correct",
	.event = "event=0x8b",
	.desc = "L2 Branch Prediction Overrides Existing Prediction (speculative)",
	.topic = "branch",
},
{
	.name = "bp_dyn_ind_pred",
	.event = "event=0x8e",
	.desc = "Dynamic Indirect Predictions",
	.topic = "branch",
	.long_desc = "Indirect Branch Prediction for potential multi-target branch (speculative)",
},
{
	.name = "bp_de_redirect",
	.event = "event=0x91",
	.desc = "Decoder Overrides Existing Branch Prediction (speculative)",
	.topic = "branch",
},
{
	.name = "bp_l1_tlb_fetch_hit",
	.event = "umask=0xFF,event=0x94",
	.desc = "The number of instruction fetches that hit in the L1 ITLB",
	.topic = "branch",
},
{
	.name = "bp_l1_tlb_fetch_hit.if1g",
	.event = "umask=0x4,event=0x94",
	.desc = "The number of instruction fetches that hit in the L1 ITLB. Instruction fetches to a 1GB page",
	.topic = "branch",
},
{
	.name = "bp_l1_tlb_fetch_hit.if2m",
	.event = "umask=0x2,event=0x94",
	.desc = "The number of instruction fetches that hit in the L1 ITLB. Instruction fetches to a 2MB page",
	.topic = "branch",
},
{
	.name = "bp_l1_tlb_fetch_hit.if4k",
	.event = "umask=0x1,event=0x94",
	.desc = "The number of instruction fetches that hit in the L1 ITLB. Instruction fetches to a 4KB page",
	.topic = "branch",
},
{
	.name = "bp_tlb_rel",
	.event = "event=0x99",
	.desc = "The number of ITLB reload requests",
	.topic = "branch",
},
{
	.name = "l2_request_g1.rd_blk_l",
	.event = "umask=0x80,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Data cache reads (including hardware and software prefetch)",
	.topic = "cache",
},
{
	.name = "l2_request_g1.rd_blk_x",
	.event = "umask=0x40,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Data cache stores",
	.topic = "cache",
},
{
	.name = "l2_request_g1.ls_rd_blk_c_s",
	.event = "umask=0x20,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Data cache shared reads",
	.topic = "cache",
},
{
	.name = "l2_request_g1.cacheable_ic_read",
	.event = "umask=0x10,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Instruction cache reads",
	.topic = "cache",
},
{
	.name = "l2_request_g1.change_to_x",
	.event = "umask=0x8,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). Data cache state change requests. Request change to writable, check L2 for current state",
	.topic = "cache",
},
{
	.name = "l2_request_g1.prefetch_l2_cmd",
	.event = "umask=0x4,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). PrefetchL2Cmd",
	.topic = "cache",
},
{
	.name = "l2_request_g1.l2_hw_pf",
	.event = "umask=0x2,event=0x60",
	.desc = "All L2 Cache Requests (Breakdown 1 - Common). L2 Prefetcher. All prefetches accepted by L2 pipeline, hit or miss. Types of PF and L2 hit/miss broken out in a separate perfmon event",
	.topic = "cache",
},
{
	.name = "l2_request_g1.group2",
	.event = "umask=0x1,event=0x60",
	.desc = "Miscellaneous events covered in more detail by l2_request_g2 (PMCx061)",
	.topic = "cache",
},
{
	.name = "l2_request_g2.group1",
	.event = "umask=0x80,event=0x61",
	.desc = "Miscellaneous events covered in more detail by l2_request_g1 (PMCx060)",
	.topic = "cache",
},
{
	.name = "l2_request_g2.ls_rd_sized",
	.event = "umask=0x40,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Data cache read sized",
	.topic = "cache",
},
{
	.name = "l2_request_g2.ls_rd_sized_nc",
	.event = "umask=0x20,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Data cache read sized non-cacheable",
	.topic = "cache",
},
{
	.name = "l2_request_g2.ic_rd_sized",
	.event = "umask=0x10,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Instruction cache read sized",
	.topic = "cache",
},
{
	.name = "l2_request_g2.ic_rd_sized_nc",
	.event = "umask=0x8,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Instruction cache read sized non-cacheable",
	.topic = "cache",
},
{
	.name = "l2_request_g2.smc_inval",
	.event = "umask=0x4,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Self-modifying code invalidates",
	.topic = "cache",
},
{
	.name = "l2_request_g2.bus_locks_originator",
	.event = "umask=0x2,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Bus locks",
	.topic = "cache",
},
{
	.name = "l2_request_g2.bus_locks_responses",
	.event = "umask=0x1,event=0x61",
	.desc = "All L2 Cache Requests (Breakdown 2 - Rare). Bus lock response",
	.topic = "cache",
},
{
	.name = "l2_latency.l2_cycles_waiting_on_fills",
	.event = "umask=0x1,event=0x62",
	.desc = "Total cycles spent waiting for L2 fills to complete from L3 or memory, divided by four. Event counts are for both threads. To calculate average latency, the number of fills from both threads must be used",
	.topic = "cache",
},
{
	.name = "l2_wcb_req.wcb_write",
	.event = "umask=0x40,event=0x63",
	.desc = "LS to L2 WCB write requests. LS (Load/Store unit) to L2 WCB (Write Combining Buffer) write requests",
	.topic = "cache",
},
{
	.name = "l2_wcb_req.wcb_close",
	.event = "umask=0x20,event=0x63",
	.desc = "LS to L2 WCB close requests. LS (Load/Store unit) to L2 WCB (Write Combining Buffer) close requests",
	.topic = "cache",
},
{
	.name = "l2_wcb_req.zero_byte_store",
	.event = "umask=0x4,event=0x63",
	.desc = "LS to L2 WCB zero byte store requests. LS (Load/Store unit) to L2 WCB (Write Combining Buffer) zero byte store requests",
	.topic = "cache",
},
{
	.name = "l2_wcb_req.cl_zero",
	.event = "umask=0x1,event=0x63",
	.desc = "LS to L2 WCB cache line zeroing requests. LS (Load/Store unit) to L2 WCB (Write Combining Buffer) cache line zeroing requests",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_cs",
	.event = "umask=0x80,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache shared read hit in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_l_hit_x",
	.event = "umask=0x40,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache read hit in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_l_hit_s",
	.event = "umask=0x20,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache read hit on shared line in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_x",
	.event = "umask=0x10,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache store or state change hit in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ls_rd_blk_c",
	.event = "umask=0x8,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Data cache request miss in L2 (all types)",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ic_fill_hit_x",
	.event = "umask=0x4,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Instruction cache hit modifiable line in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ic_fill_hit_s",
	.event = "umask=0x2,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Instruction cache hit clean line in L2",
	.topic = "cache",
},
{
	.name = "l2_cache_req_stat.ic_fill_miss",
	.event = "umask=0x1,event=0x64",
	.desc = "Core to L2 cacheable request access status (not including L2 Prefetch). Instruction cache request miss in L2",
	.topic = "cache",
},
{
	.name = "l2_fill_pending.l2_fill_busy",
	.event = "umask=0x1,event=0x6d",
	.desc = "Cycles with fill pending from L2. Total cycles spent with one or more fill requests in flight from L2",
	.topic = "cache",
},
{
	.name = "l2_pf_hit_l2",
	.event = "umask=0xff,event=0x70",
	.desc = "L2 prefetch hit in L2",
	.topic = "cache",
},
{
	.name = "l2_pf_miss_l2_hit_l3",
	.event = "umask=0xff,event=0x71",
	.desc = "L2 prefetcher hits in L3. Counts all L2 prefetches accepted by the L2 pipeline which miss the L2 cache and hit the L3",
	.topic = "cache",
},
{
	.name = "l2_pf_miss_l2_l3",
	.event = "umask=0xff,event=0x72",
	.desc = "L2 prefetcher misses in L3. All L2 prefetches accepted by the L2 pipeline which miss the L2 and the L3 caches",
	.topic = "cache",
},
{
	.name = "ic_fw32",
	.event = "event=0x80",
	.desc = "The number of 32B fetch windows transferred from IC pipe to DE instruction decoder (includes non-cacheable and cacheable fill responses)",
	.topic = "cache",
},
{
	.name = "ic_fw32_miss",
	.event = "event=0x81",
	.desc = "The number of 32B fetch windows tried to read the L1 IC and missed in the full tag",
	.topic = "cache",
},
{
	.name = "ic_cache_fill_l2",
	.event = "event=0x82",
	.desc = "The number of 64 byte instruction cache line was fulfilled from the L2 cache",
	.topic = "cache",
},
{
	.name = "ic_cache_fill_sys",
	.event = "event=0x83",
	.desc = "The number of 64 byte instruction cache line fulfilled from system memory or another cache",
	.topic = "cache",
},
{
	.name = "bp_l1_tlb_miss_l2_hit",
	.event = "event=0x84",
	.desc = "The number of instruction fetches that miss in the L1 ITLB but hit in the L2 ITLB",
	.topic = "cache",
},
{
	.name = "bp_l1_tlb_miss_l2_tlb_miss",
	.event = "umask=0xff,event=0x85",
	.desc = "The number of instruction fetches that miss in both the L1 and L2 TLBs",
	.topic = "cache",
},
{
	.name = "bp_l1_tlb_miss_l2_tlb_miss.if1g",
	.event = "umask=0x4,event=0x85",
	.desc = "The number of instruction fetches that miss in both the L1 and L2 TLBs. Instruction fetches to a 1GB page",
	.topic = "cache",
},
{
	.name = "bp_l1_tlb_miss_l2_tlb_miss.if2m",
	.event = "umask=0x2,event=0x85",
	.desc = "The number of instruction fetches that miss in both the L1 and L2 TLBs. Instruction fetches to a 2MB page",
	.topic = "cache",
},
{
	.name = "bp_l1_tlb_miss_l2_tlb_miss.if4k",
	.event = "umask=0x1,event=0x85",
	.desc = "The number of instruction fetches that miss in both the L1 and L2 TLBs. Instruction fetches to a 4KB page",
	.topic = "cache",
},
{
	.name = "bp_snp_re_sync",
	.event = "event=0x86",
	.desc = "The number of pipeline restarts caused by invalidating probes that hit on the instruction stream currently being executed. This would happen if the active instruction stream was being modified by another processor in an MP system - typically a highly unlikely event",
	.topic = "cache",
},
{
	.name = "ic_fetch_stall.ic_stall_any",
	.event = "umask=0x4,event=0x87",
	.desc = "Instruction Pipe Stall. IC pipe was stalled during this clock cycle for any reason (nothing valid in pipe ICM1)",
	.topic = "cache",
},
{
	.name = "ic_fetch_stall.ic_stall_dq_empty",
	.event = "umask=0x2,event=0x87",
	.desc = "Instruction Pipe Stall. IC pipe was stalled during this clock cycle (including IC to OC fetches) due to DQ empty",
	.topic = "cache",
},
{
	.name = "ic_fetch_stall.ic_stall_back_pressure",
	.event = "umask=0x1,event=0x87",
	.desc = "Instruction Pipe Stall. IC pipe was stalled during this clock cycle (including IC to OC fetches) due to back-pressure",
	.topic = "cache",
},
{
	.name = "ic_cache_inval.l2_invalidating_probe",
	.event = "umask=0x2,event=0x8c",
	.desc = "IC line invalidated due to L2 invalidating probe (external or LS). The number of instruction cache lines invalidated. A non-SMC event is CMC (cross modifying code), either from the other thread of the core or another core",
	.topic = "cache",
},
{
	.name = "ic_cache_inval.fill_invalidated",
	.event = "umask=0x1,event=0x8c",
	.desc = "IC line invalidated due to overwriting fill response. The number of instruction cache lines invalidated. A non-SMC event is CMC (cross modifying code), either from the other thread of the core or another core",
	.topic = "cache",
},
{
	.name = "ic_oc_mode_switch.oc_ic_mode_switch",
	.event = "umask=0x2,event=0x28a",
	.desc = "OC Mode Switch. OC to IC mode switch",
	.topic = "cache",
},
{
	.name = "ic_oc_mode_switch.ic_oc_mode_switch",
	.event = "umask=0x1,event=0x28a",
	.desc = "OC Mode Switch. IC to OC mode switch",
	.topic = "cache",
},
{
	.name = "l3_request_g1.caching_l3_cache_accesses",
	.event = "umask=0x80,event=0x1",
	.desc = "Caching: L3 cache accesses. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "l3_lookup_state.all_l3_req_typs",
	.event = "umask=0xff,event=0x4",
	.desc = "All L3 Request Types. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "l3_comb_clstr_state.other_l3_miss_typs",
	.event = "umask=0xfe,event=0x6",
	.desc = "Other L3 Miss Request Types. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "l3_comb_clstr_state.request_miss",
	.event = "umask=0x01,event=0x6",
	.desc = "L3 cache misses. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "xi_sys_fill_latency",
	.event = "umask=0x00,event=0x90",
	.desc = "L3 Cache Miss Latency. Total cycles for all transactions divided by 16. Ignores SliceMask and ThreadMask. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "xi_ccx_sdp_req1.all_l3_miss_req_typs",
	.event = "umask=0x3f,event=0x9a",
	.desc = "All L3 Miss Request Types. Ignores SliceMask and ThreadMask. Unit: amd_l3 ",
	.topic = "cache",
	.pmu = "amd_l3",
},
{
	.name = "ex_ret_instr",
	.event = "event=0xc0",
	.desc = "Retired Instructions",
	.topic = "core",
},
{
	.name = "ex_ret_cops",
	.event = "event=0xc1",
	.desc = "Retired Uops",
	.topic = "core",
	.long_desc = "The number of micro-ops retired. This count includes all processor activity (instructions, exceptions, interrupts, microcode assists, etc.). The number of events logged per cycle can vary from 0 to 8",
},
{
	.name = "ex_ret_brn",
	.event = "event=0xc2",
	.desc = "Retired Branch Instructions",
	.topic = "core",
	.long_desc = "The number of branch instructions retired. This includes all types of architectural control flow changes, including exceptions and interrupts",
},
{
	.name = "ex_ret_brn_misp",
	.event = "event=0xc3",
	.desc = "Retired Branch Instructions Mispredicted",
	.topic = "core",
	.long_desc = "The number of branch instructions retired, of any type, that were not correctly predicted. This includes those for which prediction is not attempted (far control transfers, exceptions and interrupts)",
},
{
	.name = "ex_ret_brn_tkn",
	.event = "event=0xc4",
	.desc = "Retired Taken Branch Instructions",
	.topic = "core",
	.long_desc = "The number of taken branches that were retired. This includes all types of architectural control flow changes, including exceptions and interrupts",
},
{
	.name = "ex_ret_brn_tkn_misp",
	.event = "event=0xc5",
	.desc = "Retired Taken Branch Instructions Mispredicted",
	.topic = "core",
	.long_desc = "The number of retired taken branch instructions that were mispredicted",
},
{
	.name = "ex_ret_brn_far",
	.event = "event=0xc6",
	.desc = "Retired Far Control Transfers",
	.topic = "core",
	.long_desc = "The number of far control transfers retired including far call/jump/return, IRET, SYSCALL and SYSRET, plus exceptions and interrupts. Far control transfers are not subject to branch prediction",
},
{
	.name = "ex_ret_brn_resync",
	.event = "event=0xc7",
	.desc = "Retired Branch Resyncs",
	.topic = "core",
	.long_desc = "The number of resync branches. These reflect pipeline restarts due to certain microcode assists and events such as writes to the active instruction stream, among other things. Each occurrence reflects a restart penalty similar to a branch mispredict. This is relatively rare",
},
{
	.name = "ex_ret_near_ret",
	.event = "event=0xc8",
	.desc = "Retired Near Returns",
	.topic = "core",
	.long_desc = "The number of near return instructions (RET or RET Iw) retired",
},
{
	.name = "ex_ret_near_ret_mispred",
	.event = "event=0xc9",
	.desc = "Retired Near Returns Mispredicted",
	.topic = "core",
	.long_desc = "The number of near returns retired that were not correctly predicted by the return address predictor. Each such mispredict incurs the same penalty as a mispredicted conditional branch instruction",
},
{
	.name = "ex_ret_brn_ind_misp",
	.event = "event=0xca",
	.desc = "Retired Indirect Branch Instructions Mispredicted",
	.topic = "core",
},
{
	.name = "ex_ret_mmx_fp_instr.sse_instr",
	.event = "umask=0x4,event=0xcb",
	.desc = "SSE instructions (SSE, SSE2, SSE3, SSSE3, SSE4A, SSE41, SSE42, AVX)",
	.topic = "core",
	.long_desc = "The number of MMX, SSE or x87 instructions retired. The UnitMask allows the selection of the individual classes of instructions as given in the table. Each increment represents one complete instruction. Since this event includes non-numeric instructions it is not suitable for measuring MFLOPS. SSE instructions (SSE, SSE2, SSE3, SSSE3, SSE4A, SSE41, SSE42, AVX)",
},
{
	.name = "ex_ret_mmx_fp_instr.mmx_instr",
	.event = "umask=0x2,event=0xcb",
	.desc = "MMX instructions",
	.topic = "core",
	.long_desc = "The number of MMX, SSE or x87 instructions retired. The UnitMask allows the selection of the individual classes of instructions as given in the table. Each increment represents one complete instruction. Since this event includes non-numeric instructions it is not suitable for measuring MFLOPS. MMX instructions",
},
{
	.name = "ex_ret_mmx_fp_instr.x87_instr",
	.event = "umask=0x1,event=0xcb",
	.desc = "x87 instructions",
	.topic = "core",
	.long_desc = "The number of MMX, SSE or x87 instructions retired. The UnitMask allows the selection of the individual classes of instructions as given in the table. Each increment represents one complete instruction. Since this event includes non-numeric instructions it is not suitable for measuring MFLOPS. x87 instructions",
},
{
	.name = "ex_ret_cond",
	.event = "event=0xd1",
	.desc = "Retired Conditional Branch Instructions",
	.topic = "core",
},
{
	.name = "ex_ret_cond_misp",
	.event = "event=0xd2",
	.desc = "Retired Conditional Branch Instructions Mispredicted",
	.topic = "core",
},
{
	.name = "ex_div_busy",
	.event = "event=0xd3",
	.desc = "Div Cycles Busy count",
	.topic = "core",
},
{
	.name = "ex_div_count",
	.event = "event=0xd4",
	.desc = "Div Op Count",
	.topic = "core",
},
{
	.name = "ex_tagged_ibs_ops.ibs_count_rollover",
	.event = "umask=0x4,event=0x1cf",
	.desc = "Tagged IBS Ops. Number of times an op could not be tagged by IBS because of a previous tagged op that has not retired",
	.topic = "core",
},
{
	.name = "ex_tagged_ibs_ops.ibs_tagged_ops_ret",
	.event = "umask=0x2,event=0x1cf",
	.desc = "Tagged IBS Ops. Number of Ops tagged by IBS that retired",
	.topic = "core",
},
{
	.name = "ex_tagged_ibs_ops.ibs_tagged_ops",
	.event = "umask=0x1,event=0x1cf",
	.desc = "Tagged IBS Ops. Number of Ops tagged by IBS",
	.topic = "core",
},
{
	.name = "ex_ret_fus_brnch_inst",
	.event = "event=0x1d0",
	.desc = "Retired Fused Instructions. The number of fuse-branch instructions retired per cycle. The number of events logged per cycle can vary from 0-8",
	.topic = "core",
},
{
	.name = "fpu_pipe_assignment.total",
	.event = "umask=0xf,event=0",
	.desc = "Total number of fp uOps",
	.topic = "floating point",
	.long_desc = "Total number of fp uOps. The number of operations (uOps) dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS",
},
{
	.name = "fpu_pipe_assignment.total3",
	.event = "umask=0x8,event=0",
	.desc = "Total number uOps assigned to pipe 3",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one-cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number uOps assigned to pipe 3",
},
{
	.name = "fpu_pipe_assignment.total2",
	.event = "umask=0x4,event=0",
	.desc = "Total number uOps assigned to pipe 2",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number uOps assigned to pipe 2",
},
{
	.name = "fpu_pipe_assignment.total1",
	.event = "umask=0x2,event=0",
	.desc = "Total number uOps assigned to pipe 1",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number uOps assigned to pipe 1",
},
{
	.name = "fpu_pipe_assignment.total0",
	.event = "umask=0x1,event=0",
	.desc = "Total number of fp uOps  on pipe 0",
	.topic = "floating point",
	.long_desc = "The number of operations (uOps) dispatched to each of the 4 FPU execution pipelines. This event reflects how busy the FPU pipelines are and may be used for workload characterization. This includes all operations performed by x87, MMX, and SSE instructions, including moves. Each increment represents a one- cycle dispatch event. This event is a speculative event. Since this event includes non-numeric operations it is not suitable for measuring MFLOPS. Total number uOps assigned to pipe 0",
},
{
	.name = "fp_ret_sse_avx_ops.all",
	.event = "umask=0xff,event=0x3",
	.desc = "All FLOPS. This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15",
	.topic = "floating point",
},
{
	.name = "fp_ret_sse_avx_ops.mac_flops",
	.event = "umask=0x8,event=0x3",
	.desc = "Multiply-add FLOPS. Multiply-add counts as 2 FLOPS. This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15",
	.topic = "floating point",
},
{
	.name = "fp_ret_sse_avx_ops.div_flops",
	.event = "umask=0x4,event=0x3",
	.desc = "Divide/square root FLOPS. This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15",
	.topic = "floating point",
},
{
	.name = "fp_ret_sse_avx_ops.mult_flops",
	.event = "umask=0x2,event=0x3",
	.desc = "Multiply FLOPS. This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15",
	.topic = "floating point",
},
{
	.name = "fp_ret_sse_avx_ops.add_sub_flops",
	.event = "umask=0x1,event=0x3",
	.desc = "Add/subtract FLOPS. This is a retire-based event. The number of retired SSE/AVX FLOPS. The number of events logged per cycle can vary from 0 to 64. This event can count above 15",
	.topic = "floating point",
},
{
	.name = "fp_num_mov_elim_scal_op.optimized",
	.event = "umask=0x8,event=0x4",
	.desc = "Number of Scalar Ops optimized. This is a dispatch based speculative event, and is useful for measuring the effectiveness of the Move elimination and Scalar code optimization schemes",
	.topic = "floating point",
},
{
	.name = "fp_num_mov_elim_scal_op.opt_potential",
	.event = "umask=0x4,event=0x4",
	.desc = "Number of Ops that are candidates for optimization (have Z-bit either set or pass). This is a dispatch based speculative event, and is useful for measuring the effectiveness of the Move elimination and Scalar code optimization schemes",
	.topic = "floating point",
},
{
	.name = "fp_num_mov_elim_scal_op.sse_mov_ops_elim",
	.event = "umask=0x2,event=0x4",
	.desc = "Number of SSE Move Ops eliminated. This is a dispatch based speculative event, and is useful for measuring the effectiveness of the Move elimination and Scalar code optimization schemes",
	.topic = "floating point",
},
{
	.name = "fp_num_mov_elim_scal_op.sse_mov_ops",
	.event = "umask=0x1,event=0x4",
	.desc = "Number of SSE Move Ops. This is a dispatch based speculative event, and is useful for measuring the effectiveness of the Move elimination and Scalar code optimization schemes",
	.topic = "floating point",
},
{
	.name = "fp_retired_ser_ops.sse_bot_ret",
	.event = "umask=0x8,event=0x5",
	.desc = "SSE bottom-executing uOps retired. The number of serializing Ops retired",
	.topic = "floating point",
},
{
	.name = "fp_retired_ser_ops.sse_ctrl_ret",
	.event = "umask=0x4,event=0x5",
	.desc = "The number of serializing Ops retired. SSE control word mispredict traps due to mispredictions in RC, FTZ or DAZ, or changes in mask bits",
	.topic = "floating point",
},
{
	.name = "fp_retired_ser_ops.x87_bot_ret",
	.event = "umask=0x2,event=0x5",
	.desc = "x87 bottom-executing uOps retired. The number of serializing Ops retired",
	.topic = "floating point",
},
{
	.name = "fp_retired_ser_ops.x87_ctrl_ret",
	.event = "umask=0x1,event=0x5",
	.desc = "x87 control word mispredict traps due to mispredictions in RC or PC, or changes in mask bits. The number of serializing Ops retired",
	.topic = "floating point",
},
{
	.name = "fp_disp_faults.ymm_spill_fault",
	.event = "umask=0x8,event=0xe",
	.desc = "Floating Point Dispatch Faults. YMM spill fault",
	.topic = "floating point",
},
{
	.name = "fp_disp_faults.ymm_fill_fault",
	.event = "umask=0x4,event=0xe",
	.desc = "Floating Point Dispatch Faults. YMM fill fault",
	.topic = "floating point",
},
{
	.name = "fp_disp_faults.xmm_fill_fault",
	.event = "umask=0x2,event=0xe",
	.desc = "Floating Point Dispatch Faults. XMM fill fault",
	.topic = "floating point",
},
{
	.name = "fp_disp_faults.x87_fill_fault",
	.event = "umask=0x1,event=0xe",
	.desc = "Floating Point Dispatch Faults. x87 fill fault",
	.topic = "floating point",
},
{
	.name = "ls_bad_status2.stli_other",
	.event = "umask=0x2,event=0x24",
	.desc = "Non-forwardable conflict; used to reduce STLI's via software. All reasons. Store To Load Interlock (STLI) are loads that were unable to complete because of a possible match with an older store, and the older store could not do STLF for some reason",
	.topic = "memory",
	.long_desc = "Store-to-load conflicts: A load was unable to complete due to a non-forwardable conflict with an older store. Most commonly, a load's address range partially but not completely overlaps with an uncompleted older store. Software can avoid this problem by using same-size and same-alignment loads and stores when accessing the same data. Vector/SIMD code is particularly susceptible to this problem; software should construct wide vector stores by manipulating vector elements in registers using shuffle/blend/swap instructions prior to storing to memory, instead of using narrow element-by-element stores",
},
{
	.name = "ls_locks.spec_lock_hi_spec",
	.event = "umask=0x8,event=0x25",
	.desc = "Retired lock instructions. High speculative cacheable lock speculation succeeded",
	.topic = "memory",
},
{
	.name = "ls_locks.spec_lock_lo_spec",
	.event = "umask=0x4,event=0x25",
	.desc = "Retired lock instructions. Low speculative cacheable lock speculation succeeded",
	.topic = "memory",
},
{
	.name = "ls_locks.non_spec_lock",
	.event = "umask=0x2,event=0x25",
	.desc = "Retired lock instructions. Non-speculative lock succeeded",
	.topic = "memory",
},
{
	.name = "ls_locks.bus_lock",
	.event = "umask=0x1,event=0x25",
	.desc = "Retired lock instructions. Bus lock when a locked operations crosses a cache boundary or is done on an uncacheable memory type. Comparable to legacy bus lock",
	.topic = "memory",
},
{
	.name = "ls_ret_cl_flush",
	.event = "event=0x26",
	.desc = "Number of retired CLFLUSH instructions",
	.topic = "memory",
},
{
	.name = "ls_ret_cpuid",
	.event = "event=0x27",
	.desc = "Number of retired CPUID instructions",
	.topic = "memory",
},
{
	.name = "ls_dispatch.ld_st_dispatch",
	.event = "umask=0x4,event=0x29",
	.desc = "Dispatch of a single op that performs a load from and store to the same memory address. Number of single ops that do load/store to an address",
	.topic = "memory",
},
{
	.name = "ls_dispatch.store_dispatch",
	.event = "umask=0x2,event=0x29",
	.desc = "Number of stores dispatched. Counts the number of operations dispatched to the LS unit. Unit Masks ADDed",
	.topic = "memory",
},
{
	.name = "ls_dispatch.ld_dispatch",
	.event = "umask=0x1,event=0x29",
	.desc = "Number of loads dispatched. Counts the number of operations dispatched to the LS unit. Unit Masks ADDed",
	.topic = "memory",
},
{
	.name = "ls_smi_rx",
	.event = "event=0x2b",
	.desc = "Number of SMIs received",
	.topic = "memory",
},
{
	.name = "ls_int_taken",
	.event = "event=0x2c",
	.desc = "Number of interrupts taken",
	.topic = "memory",
},
{
	.name = "ls_rdtsc",
	.event = "event=0x2d",
	.desc = "Number of reads of the TSC (RDTSC instructions). The count is speculative",
	.topic = "memory",
},
{
	.name = "ls_stlf",
	.event = "event=0x35",
	.desc = "Number of STLF hits",
	.topic = "memory",
},
{
	.name = "ls_st_commit_cancel2.st_commit_cancel_wcb_full",
	.event = "event=0x37",
	.desc = "A non-cacheable store and the non-cacheable commit buffer is full",
	.topic = "memory",
},
{
	.name = "ls_dc_accesses",
	.event = "event=0x40",
	.desc = "Number of accesses to the dcache for load/store references",
	.topic = "memory",
	.long_desc = "The number of accesses to the data cache for load and store references. This may include certain microcode scratchpad accesses, although these are generally rare. Each increment represents an eight-byte access, although the instruction may only be accessing a portion of that. This event is a speculative event",
},
{
	.name = "ls_mab_alloc.dc_prefetcher",
	.event = "umask=0x8,event=0x41",
	.desc = "LS MAB Allocates by Type. DC prefetcher",
	.topic = "memory",
},
{
	.name = "ls_mab_alloc.stores",
	.event = "umask=0x2,event=0x41",
	.desc = "LS MAB Allocates by Type. Stores",
	.topic = "memory",
},
{
	.name = "ls_mab_alloc.loads",
	.event = "umask=0x1,event=0x41",
	.desc = "LS MAB Allocates by Type. Loads",
	.topic = "memory",
},
{
	.name = "ls_refills_from_sys.ls_mabresp_rmt_dram",
	.event = "umask=0x40,event=0x43",
	.desc = "Demand Data Cache Fills by Data Source. DRAM or IO from different die",
	.topic = "memory",
},
{
	.name = "ls_refills_from_sys.ls_mabresp_rmt_cache",
	.event = "umask=0x10,event=0x43",
	.desc = "Demand Data Cache Fills by Data Source. Hit in cache; Remote CCX and the address's Home Node is on a different die",
	.topic = "memory",
},
{
	.name = "ls_refills_from_sys.ls_mabresp_lcl_dram",
	.event = "umask=0x8,event=0x43",
	.desc = "Demand Data Cache Fills by Data Source. DRAM or IO from this thread's die",
	.topic = "memory",
},
{
	.name = "ls_refills_from_sys.ls_mabresp_lcl_cache",
	.event = "umask=0x2,event=0x43",
	.desc = "Demand Data Cache Fills by Data Source. Hit in cache; local CCX (not Local L2), or Remote CCX and the address's Home Node is on this thread's die",
	.topic = "memory",
},
{
	.name = "ls_refills_from_sys.ls_mabresp_lcl_l2",
	.event = "umask=0x1,event=0x43",
	.desc = "Demand Data Cache Fills by Data Source. Local L2 hit",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.all",
	.event = "umask=0xff,event=0x45",
	.desc = "All L1 DTLB Misses or Reloads",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_1g_l2_miss",
	.event = "umask=0x80,event=0x45",
	.desc = "L1 DTLB Miss. DTLB reload to a 1G page that miss in the L2 TLB",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_2m_l2_miss",
	.event = "umask=0x40,event=0x45",
	.desc = "L1 DTLB Miss. DTLB reload to a 2M page that miss in the L2 TLB",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_coalesced_page_miss",
	.event = "umask=0x20,event=0x45",
	.desc = "L1 DTLB Miss. DTLB reload coalesced page miss",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_4k_l2_miss",
	.event = "umask=0x10,event=0x45",
	.desc = "L1 DTLB Miss. DTLB reload to a 4K page that miss the L2 TLB",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_1g_l2_hit",
	.event = "umask=0x8,event=0x45",
	.desc = "L1 DTLB Miss. DTLB reload to a 1G page that hit in the L2 TLB",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_2m_l2_hit",
	.event = "umask=0x4,event=0x45",
	.desc = "L1 DTLB Miss. DTLB reload to a 2M page that hit in the L2 TLB",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_coalesced_page_hit",
	.event = "umask=0x2,event=0x45",
	.desc = "L1 DTLB Miss. DTLB reload hit a coalesced page",
	.topic = "memory",
},
{
	.name = "ls_l1_d_tlb_miss.tlb_reload_4k_l2_hit",
	.event = "umask=0x1,event=0x45",
	.desc = "L1 DTLB Miss. DTLB reload to a 4K page that hit in the L2 TLB",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.iside",
	.event = "umask=0xc,event=0x46",
	.desc = "Total Page Table Walks on I-side",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.ic_type1",
	.event = "umask=0x8,event=0x46",
	.desc = "Total Page Table Walks IC Type 1",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.ic_type0",
	.event = "umask=0x4,event=0x46",
	.desc = "Total Page Table Walks IC Type 0",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.dside",
	.event = "umask=0x3,event=0x46",
	.desc = "Total Page Table Walks on D-side",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.dc_type1",
	.event = "umask=0x2,event=0x46",
	.desc = "Total Page Table Walks DC Type 1",
	.topic = "memory",
},
{
	.name = "ls_tablewalker.dc_type0",
	.event = "umask=0x1,event=0x46",
	.desc = "Total Page Table Walks DC Type 0",
	.topic = "memory",
},
{
	.name = "ls_misal_accesses",
	.event = "event=0x47",
	.desc = "Misaligned loads",
	.topic = "memory",
},
{
	.name = "ls_pref_instr_disp",
	.event = "umask=0xff,event=0x4b",
	.desc = "Software Prefetch Instructions Dispatched (Speculative)",
	.topic = "memory",
},
{
	.name = "ls_pref_instr_disp.prefetch_nta",
	.event = "umask=0x4,event=0x4b",
	.desc = "Software Prefetch Instructions Dispatched (Speculative). PrefetchNTA instruction. See docAPM3 PREFETCHlevel",
	.topic = "memory",
},
{
	.name = "ls_pref_instr_disp.prefetch_w",
	.event = "umask=0x2,event=0x4b",
	.desc = "Software Prefetch Instructions Dispatched (Speculative). See docAPM3 PREFETCHW",
	.topic = "memory",
},
{
	.name = "ls_pref_instr_disp.prefetch",
	.event = "umask=0x1,event=0x4b",
	.desc = "Software Prefetch Instructions Dispatched (Speculative). Prefetch_T0_T1_T2. PrefetchT0, T1 and T2 instructions. See docAPM3 PREFETCHlevel",
	.topic = "memory",
},
{
	.name = "ls_inef_sw_pref.mab_mch_cnt",
	.event = "umask=0x2,event=0x52",
	.desc = "The number of software prefetches that did not fetch data outside of the processor core. Software PREFETCH instruction saw a match on an already-allocated miss request buffer",
	.topic = "memory",
},
{
	.name = "ls_inef_sw_pref.data_pipe_sw_pf_dc_hit",
	.event = "umask=0x1,event=0x52",
	.desc = "The number of software prefetches that did not fetch data outside of the processor core. Software PREFETCH instruction saw a DC hit",
	.topic = "memory",
},
{
	.name = "ls_sw_pf_dc_fill.ls_mabresp_rmt_dram",
	.event = "umask=0x40,event=0x59",
	.desc = "Software Prefetch Data Cache Fills by Data Source. From DRAM (home node remote)",
	.topic = "memory",
},
{
	.name = "ls_sw_pf_dc_fill.ls_mabresp_rmt_cache",
	.event = "umask=0x10,event=0x59",
	.desc = "Software Prefetch Data Cache Fills by Data Source. From another cache (home node remote)",
	.topic = "memory",
},
{
	.name = "ls_sw_pf_dc_fill.ls_mabresp_lcl_dram",
	.event = "umask=0x8,event=0x59",
	.desc = "Software Prefetch Data Cache Fills by Data Source. DRAM or IO from this thread's die.  From DRAM (home node local)",
	.topic = "memory",
},
{
	.name = "ls_sw_pf_dc_fill.ls_mabresp_lcl_cache",
	.event = "umask=0x2,event=0x59",
	.desc = "Software Prefetch Data Cache Fills by Data Source. From another cache (home node local)",
	.topic = "memory",
},
{
	.name = "ls_sw_pf_dc_fill.ls_mabresp_lcl_l2",
	.event = "umask=0x1,event=0x59",
	.desc = "Software Prefetch Data Cache Fills by Data Source. Local L2 hit",
	.topic = "memory",
},
{
	.name = "ls_hw_pf_dc_fill.ls_mabresp_rmt_dram",
	.event = "umask=0x40,event=0x5a",
	.desc = "Hardware Prefetch Data Cache Fills by Data Source. From DRAM (home node remote)",
	.topic = "memory",
},
{
	.name = "ls_hw_pf_dc_fill.ls_mabresp_rmt_cache",
	.event = "umask=0x10,event=0x5a",
	.desc = "Hardware Prefetch Data Cache Fills by Data Source. From another cache (home node remote)",
	.topic = "memory",
},
{
	.name = "ls_hw_pf_dc_fill.ls_mabresp_lcl_dram",
	.event = "umask=0x8,event=0x5a",
	.desc = "Hardware Prefetch Data Cache Fills by Data Source. From DRAM (home node local)",
	.topic = "memory",
},
{
	.name = "ls_hw_pf_dc_fill.ls_mabresp_lcl_cache",
	.event = "umask=0x2,event=0x5a",
	.desc = "Hardware Prefetch Data Cache Fills by Data Source. From another cache (home node local)",
	.topic = "memory",
},
{
	.name = "ls_hw_pf_dc_fill.ls_mabresp_lcl_l2",
	.event = "umask=0x1,event=0x5a",
	.desc = "Hardware Prefetch Data Cache Fills by Data Source. Local L2 hit",
	.topic = "memory",
},
{
	.name = "ls_not_halted_cyc",
	.event = "event=0x76",
	.desc = "Cycles not in Halt",
	.topic = "memory",
},
{
	.name = "ls_tlb_flush",
	.event = "event=0x78",
	.desc = "All TLB Flushes",
	.topic = "memory",
},
{
	.name = "de_dis_uop_queue_empty_di0",
	.event = "event=0xa9",
	.desc = "Cycles where the Micro-Op Queue is empty",
	.topic = "other",
},
{
	.name = "de_dis_uops_from_decoder",
	.event = "umask=0xff,event=0xaa",
	.desc = "Ops dispatched from either the decoders, OpCache or both",
	.topic = "other",
},
{
	.name = "de_dis_uops_from_decoder.opcache_dispatched",
	.event = "umask=0x2,event=0xaa",
	.desc = "Count of dispatched Ops from OpCache",
	.topic = "other",
},
{
	.name = "de_dis_uops_from_decoder.decoder_dispatched",
	.event = "umask=0x1,event=0xaa",
	.desc = "Count of dispatched Ops from Decoder",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls1.fp_misc_rsrc_stall",
	.event = "umask=0x80,event=0xae",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. FP Miscellaneous resource unavailable. Applies to the recovery of mispredicts with FP ops",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls1.fp_sch_rsrc_stall",
	.event = "umask=0x40,event=0xae",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. FP scheduler resource stall. Applies to ops that use the FP scheduler",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls1.fp_reg_file_rsrc_stall",
	.event = "umask=0x20,event=0xae",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. Floating point register file resource stall. Applies to all FP ops that have a destination register",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls1.taken_branch_buffer_rsrc_stall",
	.event = "umask=0x10,event=0xae",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. Taken branch buffer resource stall",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls1.int_sched_misc_token_stall",
	.event = "umask=0x8,event=0xae",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. Integer Scheduler miscellaneous resource stall",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls1.store_queue_token_stall",
	.event = "umask=0x4,event=0xae",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. Store queue resource stall. Applies to all ops with store semantics",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls1.load_queue_token_stall",
	.event = "umask=0x2,event=0xae",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. Load queue resource stall. Applies to all ops with load semantics",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls1.int_phy_reg_file_token_stall",
	.event = "umask=0x1,event=0xae",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. Integer Physical Register File resource stall. Applies to all ops that have an integer destination register",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.sc_agu_dispatch_stall",
	.event = "umask=0x40,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. SC AGU dispatch stall",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.retire_token_stall",
	.event = "umask=0x20,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. RETIRE Tokens unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.agsq_token_stall",
	.event = "umask=0x10,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. AGSQ Tokens unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.alu_token_stall",
	.event = "umask=0x8,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. ALU tokens total unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.alsq3_0_token_stall",
	.event = "umask=0x4,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. ALSQ3_0_TokenStall",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.alsq2_token_stall",
	.event = "umask=0x2,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. ALSQ 2 Tokens unavailable",
	.topic = "other",
},
{
	.name = "de_dis_dispatch_token_stalls0.alsq1_token_stall",
	.event = "umask=0x1,event=0xaf",
	.desc = "Cycles where a dispatch group is valid but does not get dispatched due to a token stall. ALSQ 1 Tokens unavailable",
	.topic = "other",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_bonnell[] = {
{
	.name = "l2_ads.self",
	.event = "umask=0x40,period=200000,event=0x21",
	.desc = "Cycles L2 address bus is in use",
	.topic = "cache",
},
{
	.name = "l2_dbus_busy.self",
	.event = "umask=0x40,period=200000,event=0x22",
	.desc = "Cycles the L2 cache data bus is busy",
	.topic = "cache",
},
{
	.name = "l2_dbus_busy_rd.self",
	.event = "umask=0x40,period=200000,event=0x23",
	.desc = "Cycles the L2 transfers data to the core",
	.topic = "cache",
},
{
	.name = "l2_lines_in.self.any",
	.event = "umask=0x70,period=200000,event=0x24",
	.desc = "L2 cache misses",
	.topic = "cache",
},
{
	.name = "l2_lines_in.self.demand",
	.event = "umask=0x40,period=200000,event=0x24",
	.desc = "L2 cache misses",
	.topic = "cache",
},
{
	.name = "l2_lines_in.self.prefetch",
	.event = "umask=0x50,period=200000,event=0x24",
	.desc = "L2 cache misses",
	.topic = "cache",
},
{
	.name = "l2_m_lines_in.self",
	.event = "umask=0x40,period=200000,event=0x25",
	.desc = "L2 cache line modifications",
	.topic = "cache",
},
{
	.name = "l2_lines_out.self.any",
	.event = "umask=0x70,period=200000,event=0x26",
	.desc = "L2 cache lines evicted",
	.topic = "cache",
},
{
	.name = "l2_lines_out.self.demand",
	.event = "umask=0x40,period=200000,event=0x26",
	.desc = "L2 cache lines evicted",
	.topic = "cache",
},
{
	.name = "l2_lines_out.self.prefetch",
	.event = "umask=0x50,period=200000,event=0x26",
	.desc = "L2 cache lines evicted",
	.topic = "cache",
},
{
	.name = "l2_m_lines_out.self.any",
	.event = "umask=0x70,period=200000,event=0x27",
	.desc = "Modified lines evicted from the L2 cache",
	.topic = "cache",
},
{
	.name = "l2_m_lines_out.self.demand",
	.event = "umask=0x40,period=200000,event=0x27",
	.desc = "Modified lines evicted from the L2 cache",
	.topic = "cache",
},
{
	.name = "l2_m_lines_out.self.prefetch",
	.event = "umask=0x50,period=200000,event=0x27",
	.desc = "Modified lines evicted from the L2 cache",
	.topic = "cache",
},
{
	.name = "l2_ifetch.self.e_state",
	.event = "umask=0x44,period=200000,event=0x28",
	.desc = "L2 cacheable instruction fetch requests",
	.topic = "cache",
},
{
	.name = "l2_ifetch.self.i_state",
	.event = "umask=0x41,period=200000,event=0x28",
	.desc = "L2 cacheable instruction fetch requests",
	.topic = "cache",
},
{
	.name = "l2_ifetch.self.m_state",
	.event = "umask=0x48,period=200000,event=0x28",
	.desc = "L2 cacheable instruction fetch requests",
	.topic = "cache",
},
{
	.name = "l2_ifetch.self.s_state",
	.event = "umask=0x42,period=200000,event=0x28",
	.desc = "L2 cacheable instruction fetch requests",
	.topic = "cache",
},
{
	.name = "l2_ifetch.self.mesi",
	.event = "umask=0x4f,period=200000,event=0x28",
	.desc = "L2 cacheable instruction fetch requests",
	.topic = "cache",
},
{
	.name = "l2_ld.self.any.e_state",
	.event = "umask=0x74,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.any.i_state",
	.event = "umask=0x71,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.any.m_state",
	.event = "umask=0x78,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.any.s_state",
	.event = "umask=0x72,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.any.mesi",
	.event = "umask=0x7f,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.demand.e_state",
	.event = "umask=0x44,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.demand.i_state",
	.event = "umask=0x41,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.demand.m_state",
	.event = "umask=0x48,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.demand.s_state",
	.event = "umask=0x42,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.demand.mesi",
	.event = "umask=0x4f,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.prefetch.e_state",
	.event = "umask=0x54,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.prefetch.i_state",
	.event = "umask=0x51,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.prefetch.m_state",
	.event = "umask=0x58,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.prefetch.s_state",
	.event = "umask=0x52,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_ld.self.prefetch.mesi",
	.event = "umask=0x5f,period=200000,event=0x29",
	.desc = "L2 cache reads",
	.topic = "cache",
},
{
	.name = "l2_st.self.e_state",
	.event = "umask=0x44,period=200000,event=0x2a",
	.desc = "L2 store requests",
	.topic = "cache",
},
{
	.name = "l2_st.self.i_state",
	.event = "umask=0x41,period=200000,event=0x2a",
	.desc = "L2 store requests",
	.topic = "cache",
},
{
	.name = "l2_st.self.m_state",
	.event = "umask=0x48,period=200000,event=0x2a",
	.desc = "L2 store requests",
	.topic = "cache",
},
{
	.name = "l2_st.self.s_state",
	.event = "umask=0x42,period=200000,event=0x2a",
	.desc = "L2 store requests",
	.topic = "cache",
},
{
	.name = "l2_st.self.mesi",
	.event = "umask=0x4f,period=200000,event=0x2a",
	.desc = "L2 store requests",
	.topic = "cache",
},
{
	.name = "l2_lock.self.e_state",
	.event = "umask=0x44,period=200000,event=0x2b",
	.desc = "L2 locked accesses",
	.topic = "cache",
},
{
	.name = "l2_lock.self.i_state",
	.event = "umask=0x41,period=200000,event=0x2b",
	.desc = "L2 locked accesses",
	.topic = "cache",
},
{
	.name = "l2_lock.self.m_state",
	.event = "umask=0x48,period=200000,event=0x2b",
	.desc = "L2 locked accesses",
	.topic = "cache",
},
{
	.name = "l2_lock.self.s_state",
	.event = "umask=0x42,period=200000,event=0x2b",
	.desc = "L2 locked accesses",
	.topic = "cache",
},
{
	.name = "l2_lock.self.mesi",
	.event = "umask=0x4f,period=200000,event=0x2b",
	.desc = "L2 locked accesses",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.self.e_state",
	.event = "umask=0x44,period=200000,event=0x2c",
	.desc = "All data requests from the L1 data cache",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.self.i_state",
	.event = "umask=0x41,period=200000,event=0x2c",
	.desc = "All data requests from the L1 data cache",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.self.m_state",
	.event = "umask=0x48,period=200000,event=0x2c",
	.desc = "All data requests from the L1 data cache",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.self.s_state",
	.event = "umask=0x42,period=200000,event=0x2c",
	.desc = "All data requests from the L1 data cache",
	.topic = "cache",
},
{
	.name = "l2_data_rqsts.self.mesi",
	.event = "umask=0x4f,period=200000,event=0x2c",
	.desc = "All data requests from the L1 data cache",
	.topic = "cache",
},
{
	.name = "l2_ld_ifetch.self.e_state",
	.event = "umask=0x44,period=200000,event=0x2d",
	.desc = "All read requests from L1 instruction and data caches",
	.topic = "cache",
},
{
	.name = "l2_ld_ifetch.self.i_state",
	.event = "umask=0x41,period=200000,event=0x2d",
	.desc = "All read requests from L1 instruction and data caches",
	.topic = "cache",
},
{
	.name = "l2_ld_ifetch.self.m_state",
	.event = "umask=0x48,period=200000,event=0x2d",
	.desc = "All read requests from L1 instruction and data caches",
	.topic = "cache",
},
{
	.name = "l2_ld_ifetch.self.s_state",
	.event = "umask=0x42,period=200000,event=0x2d",
	.desc = "All read requests from L1 instruction and data caches",
	.topic = "cache",
},
{
	.name = "l2_ld_ifetch.self.mesi",
	.event = "umask=0x4f,period=200000,event=0x2d",
	.desc = "All read requests from L1 instruction and data caches",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.any.e_state",
	.event = "umask=0x74,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.any.i_state",
	.event = "umask=0x71,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.any.m_state",
	.event = "umask=0x78,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.any.s_state",
	.event = "umask=0x72,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.any.mesi",
	.event = "umask=0x7f,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.demand.e_state",
	.event = "umask=0x44,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.demand.m_state",
	.event = "umask=0x48,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.demand.s_state",
	.event = "umask=0x42,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.prefetch.e_state",
	.event = "umask=0x54,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.prefetch.i_state",
	.event = "umask=0x51,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.prefetch.m_state",
	.event = "umask=0x58,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.prefetch.s_state",
	.event = "umask=0x52,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.prefetch.mesi",
	.event = "umask=0x5f,period=200000,event=0x2e",
	.desc = "L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.demand.i_state",
	.event = "umask=0x41,period=200000,event=0x2e",
	.desc = "L2 cache demand requests from this core that missed the L2",
	.topic = "cache",
},
{
	.name = "l2_rqsts.self.demand.mesi",
	.event = "umask=0x4f,period=200000,event=0x2e",
	.desc = "L2 cache demand requests from this core",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.any.e_state",
	.event = "umask=0x74,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.any.i_state",
	.event = "umask=0x71,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.any.m_state",
	.event = "umask=0x78,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.any.s_state",
	.event = "umask=0x72,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.any.mesi",
	.event = "umask=0x7f,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.demand.e_state",
	.event = "umask=0x44,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.demand.i_state",
	.event = "umask=0x41,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.demand.m_state",
	.event = "umask=0x48,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.demand.s_state",
	.event = "umask=0x42,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.demand.mesi",
	.event = "umask=0x4f,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.prefetch.e_state",
	.event = "umask=0x54,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.prefetch.i_state",
	.event = "umask=0x51,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.prefetch.m_state",
	.event = "umask=0x58,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.prefetch.s_state",
	.event = "umask=0x52,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_reject_busq.self.prefetch.mesi",
	.event = "umask=0x5f,period=200000,event=0x30",
	.desc = "Rejected L2 cache requests",
	.topic = "cache",
},
{
	.name = "l2_no_req.self",
	.event = "umask=0x40,period=200000,event=0x32",
	.desc = "Cycles no L2 cache requests are pending",
	.topic = "cache",
},
{
	.name = "l1d_cache.ld",
	.event = "umask=0xa1,period=2000000,event=0x40",
	.desc = "L1 Cacheable Data Reads",
	.topic = "cache",
},
{
	.name = "l1d_cache.st",
	.event = "umask=0xa2,period=2000000,event=0x40",
	.desc = "L1 Cacheable Data Writes",
	.topic = "cache",
},
{
	.name = "l1d_cache.all_ref",
	.event = "umask=0x83,period=2000000,event=0x40",
	.desc = "L1 Data reads and writes",
	.topic = "cache",
},
{
	.name = "l1d_cache.all_cache_ref",
	.event = "umask=0xa3,period=2000000,event=0x40",
	.desc = "L1 Data Cacheable reads and writes",
	.topic = "cache",
},
{
	.name = "l1d_cache.repl",
	.event = "umask=0x8,period=200000,event=0x40",
	.desc = "L1 Data line replacements",
	.topic = "cache",
},
{
	.name = "l1d_cache.replm",
	.event = "umask=0x48,period=200000,event=0x40",
	.desc = "Modified cache lines allocated in the L1 data cache",
	.topic = "cache",
},
{
	.name = "l1d_cache.evict",
	.event = "umask=0x10,period=200000,event=0x40",
	.desc = "Modified cache lines evicted from the L1 data cache",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l2_hit",
	.event = "umask=0x1,period=200000,event=0xcb",
	.desc = "Retired loads that hit the L2 cache (precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_retired.l2_miss",
	.event = "umask=0x2,period=10000,event=0xcb",
	.desc = "Retired loads that miss the L2 cache",
	.topic = "cache",
},
{
	.name = "x87_comp_ops_exe.any.s",
	.event = "umask=0x1,period=2000000,event=0x10",
	.desc = "Floating point computational micro-ops executed",
	.topic = "floating point",
},
{
	.name = "x87_comp_ops_exe.any.ar",
	.event = "umask=0x81,period=2000000,event=0x10",
	.desc = "Floating point computational micro-ops retired (Must be precise)",
	.topic = "floating point",
},
{
	.name = "x87_comp_ops_exe.fxch.s",
	.event = "umask=0x2,period=2000000,event=0x10",
	.desc = "FXCH uops executed",
	.topic = "floating point",
},
{
	.name = "x87_comp_ops_exe.fxch.ar",
	.event = "umask=0x82,period=2000000,event=0x10",
	.desc = "FXCH uops retired (Must be precise)",
	.topic = "floating point",
},
{
	.name = "fp_assist.s",
	.event = "umask=0x1,period=10000,event=0x11",
	.desc = "Floating point assists",
	.topic = "floating point",
},
{
	.name = "fp_assist.ar",
	.event = "umask=0x81,period=10000,event=0x11",
	.desc = "Floating point assists for retired operations",
	.topic = "floating point",
},
{
	.name = "simd_uops_exec.s",
	.event = "umask=0x0,period=2000000,event=0xb0",
	.desc = "SIMD micro-ops executed (excluding stores)",
	.topic = "floating point",
},
{
	.name = "simd_uops_exec.ar",
	.event = "umask=0x80,period=2000000,event=0xb0",
	.desc = "SIMD micro-ops retired (excluding stores) (Must be precise)",
	.topic = "floating point",
},
{
	.name = "simd_sat_uop_exec.s",
	.event = "umask=0x0,period=2000000,event=0xb1",
	.desc = "SIMD saturated arithmetic micro-ops executed",
	.topic = "floating point",
},
{
	.name = "simd_sat_uop_exec.ar",
	.event = "umask=0x80,period=2000000,event=0xb1",
	.desc = "SIMD saturated arithmetic micro-ops retired",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.mul.s",
	.event = "umask=0x1,period=2000000,event=0xb3",
	.desc = "SIMD packed multiply micro-ops executed",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.mul.ar",
	.event = "umask=0x81,period=2000000,event=0xb3",
	.desc = "SIMD packed multiply micro-ops retired",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.shift.s",
	.event = "umask=0x2,period=2000000,event=0xb3",
	.desc = "SIMD packed shift micro-ops executed",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.shift.ar",
	.event = "umask=0x82,period=2000000,event=0xb3",
	.desc = "SIMD packed shift micro-ops retired",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.pack.s",
	.event = "umask=0x4,period=2000000,event=0xb3",
	.desc = "SIMD packed micro-ops executed",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.pack.ar",
	.event = "umask=0x84,period=2000000,event=0xb3",
	.desc = "SIMD packed micro-ops retired",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.unpack.s",
	.event = "umask=0x8,period=2000000,event=0xb3",
	.desc = "SIMD unpacked micro-ops executed",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.unpack.ar",
	.event = "umask=0x88,period=2000000,event=0xb3",
	.desc = "SIMD unpacked micro-ops retired",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.logical.s",
	.event = "umask=0x10,period=2000000,event=0xb3",
	.desc = "SIMD packed logical micro-ops executed",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.logical.ar",
	.event = "umask=0x90,period=2000000,event=0xb3",
	.desc = "SIMD packed logical micro-ops retired",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.arithmetic.s",
	.event = "umask=0x20,period=2000000,event=0xb3",
	.desc = "SIMD packed arithmetic micro-ops executed",
	.topic = "floating point",
},
{
	.name = "simd_uop_type_exec.arithmetic.ar",
	.event = "umask=0xa0,period=2000000,event=0xb3",
	.desc = "SIMD packed arithmetic micro-ops retired",
	.topic = "floating point",
},
{
	.name = "simd_inst_retired.packed_single",
	.event = "umask=0x1,period=2000000,event=0xc7",
	.desc = "Retired Streaming SIMD Extensions (SSE) packed-single instructions",
	.topic = "floating point",
},
{
	.name = "simd_inst_retired.scalar_single",
	.event = "umask=0x2,period=2000000,event=0xc7",
	.desc = "Retired Streaming SIMD Extensions (SSE) scalar-single instructions",
	.topic = "floating point",
},
{
	.name = "simd_inst_retired.scalar_double",
	.event = "umask=0x8,period=2000000,event=0xc7",
	.desc = "Retired Streaming SIMD Extensions 2 (SSE2) scalar-double instructions",
	.topic = "floating point",
},
{
	.name = "simd_inst_retired.vector",
	.event = "umask=0x10,period=2000000,event=0xc7",
	.desc = "Retired Streaming SIMD Extensions 2 (SSE2) vector instructions",
	.topic = "floating point",
},
{
	.name = "simd_comp_inst_retired.packed_single",
	.event = "umask=0x1,period=2000000,event=0xca",
	.desc = "Retired computational Streaming SIMD Extensions (SSE) packed-single instructions",
	.topic = "floating point",
},
{
	.name = "simd_comp_inst_retired.scalar_single",
	.event = "umask=0x2,period=2000000,event=0xca",
	.desc = "Retired computational Streaming SIMD Extensions (SSE) scalar-single instructions",
	.topic = "floating point",
},
{
	.name = "simd_comp_inst_retired.scalar_double",
	.event = "umask=0x8,period=2000000,event=0xca",
	.desc = "Retired computational Streaming SIMD Extensions 2 (SSE2) scalar-double instructions",
	.topic = "floating point",
},
{
	.name = "simd_assist",
	.event = "umask=0x0,period=100000,event=0xcd",
	.desc = "SIMD assists invoked",
	.topic = "floating point",
},
{
	.name = "simd_instr_retired",
	.event = "umask=0x0,period=2000000,event=0xce",
	.desc = "SIMD Instructions retired",
	.topic = "floating point",
},
{
	.name = "simd_sat_instr_retired",
	.event = "umask=0x0,period=2000000,event=0xcf",
	.desc = "Saturated arithmetic instructions retired",
	.topic = "floating point",
},
{
	.name = "icache.accesses",
	.event = "umask=0x3,period=200000,event=0x80",
	.desc = "Instruction fetches",
	.topic = "frontend",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=200000,event=0x80",
	.desc = "Icache hit",
	.topic = "frontend",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200000,event=0x80",
	.desc = "Icache miss",
	.topic = "frontend",
},
{
	.name = "cycles_icache_mem_stalled.icache_mem_stalled",
	.event = "umask=0x1,period=2000000,event=0x86",
	.desc = "Cycles during which instruction fetches are  stalled",
	.topic = "frontend",
},
{
	.name = "decode_stall.pfb_empty",
	.event = "umask=0x1,period=2000000,event=0x87",
	.desc = "Decode stall due to PFB empty",
	.topic = "frontend",
},
{
	.name = "decode_stall.iq_full",
	.event = "umask=0x2,period=2000000,event=0x87",
	.desc = "Decode stall due to IQ full",
	.topic = "frontend",
},
{
	.name = "macro_insts.non_cisc_decoded",
	.event = "umask=0x1,period=2000000,event=0xaa",
	.desc = "Non-CISC nacro instructions decoded",
	.topic = "frontend",
},
{
	.name = "macro_insts.cisc_decoded",
	.event = "umask=0x2,period=2000000,event=0xaa",
	.desc = "CISC macro instructions decoded",
	.topic = "frontend",
},
{
	.name = "macro_insts.all_decoded",
	.event = "umask=0x3,period=2000000,event=0xaa",
	.desc = "All Instructions decoded",
	.topic = "frontend",
},
{
	.name = "uops.ms_cycles",
	.event = "umask=0x1,period=2000000,cmask=1,event=0xa9",
	.desc = "This event counts the cycles where 1 or more uops are issued by the micro-sequencer (MS), including microcode assists and inserted flows, and written to the IQ",
	.topic = "frontend",
},
{
	.name = "misalign_mem_ref.split",
	.event = "umask=0xf,period=200000,event=0x5",
	.desc = "Memory references that cross an 8-byte boundary",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.ld_split",
	.event = "umask=0x9,period=200000,event=0x5",
	.desc = "Load splits",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.st_split",
	.event = "umask=0xa,period=200000,event=0x5",
	.desc = "Store splits",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.split.ar",
	.event = "umask=0x8f,period=200000,event=0x5",
	.desc = "Memory references that cross an 8-byte boundary (At Retirement)",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.ld_split.ar",
	.event = "umask=0x89,period=200000,event=0x5",
	.desc = "Load splits (At Retirement)",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.st_split.ar",
	.event = "umask=0x8a,period=200000,event=0x5",
	.desc = "Store splits (Ar Retirement)",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.rmw_split",
	.event = "umask=0x8c,period=200000,event=0x5",
	.desc = "ld-op-st splits",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.bubble",
	.event = "umask=0x97,period=200000,event=0x5",
	.desc = "Nonzero segbase 1 bubble",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.ld_bubble",
	.event = "umask=0x91,period=200000,event=0x5",
	.desc = "Nonzero segbase load 1 bubble",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.st_bubble",
	.event = "umask=0x92,period=200000,event=0x5",
	.desc = "Nonzero segbase store 1 bubble",
	.topic = "memory",
},
{
	.name = "misalign_mem_ref.rmw_bubble",
	.event = "umask=0x94,period=200000,event=0x5",
	.desc = "Nonzero segbase ld-op-st 1 bubble",
	.topic = "memory",
},
{
	.name = "prefetch.prefetcht0",
	.event = "umask=0x81,period=200000,event=0x7",
	.desc = "Streaming SIMD Extensions (SSE) PrefetchT0 instructions executed",
	.topic = "memory",
},
{
	.name = "prefetch.prefetcht1",
	.event = "umask=0x82,period=200000,event=0x7",
	.desc = "Streaming SIMD Extensions (SSE) PrefetchT1 instructions executed",
	.topic = "memory",
},
{
	.name = "prefetch.prefetcht2",
	.event = "umask=0x84,period=200000,event=0x7",
	.desc = "Streaming SIMD Extensions (SSE) PrefetchT2 instructions executed",
	.topic = "memory",
},
{
	.name = "prefetch.sw_l2",
	.event = "umask=0x86,period=200000,event=0x7",
	.desc = "Streaming SIMD Extensions (SSE) PrefetchT1 and PrefetchT2 instructions executed",
	.topic = "memory",
},
{
	.name = "prefetch.prefetchnta",
	.event = "umask=0x88,period=200000,event=0x7",
	.desc = "Streaming SIMD Extensions (SSE) Prefetch NTA instructions executed",
	.topic = "memory",
},
{
	.name = "prefetch.hw_prefetch",
	.event = "umask=0x10,period=2000000,event=0x7",
	.desc = "L1 hardware prefetch request",
	.topic = "memory",
},
{
	.name = "prefetch.software_prefetch",
	.event = "umask=0xf,period=200000,event=0x7",
	.desc = "Any Software prefetch",
	.topic = "memory",
},
{
	.name = "prefetch.software_prefetch.ar",
	.event = "umask=0x8f,period=200000,event=0x7",
	.desc = "Any Software prefetch",
	.topic = "memory",
},
{
	.name = "segment_reg_loads.any",
	.event = "umask=0x80,period=200000,event=0x6",
	.desc = "Number of segment register loads",
	.topic = "other",
},
{
	.name = "dispatch_blocked.any",
	.event = "umask=0x20,period=200000,event=0x9",
	.desc = "Memory cluster signals to block micro-op dispatch for any reason",
	.topic = "other",
},
{
	.name = "eist_trans",
	.event = "umask=0x0,period=200000,event=0x3a",
	.desc = "Number of Enhanced Intel SpeedStep(R) Technology (EIST) transitions",
	.topic = "other",
},
{
	.name = "thermal_trip",
	.event = "umask=0xc0,period=200000,event=0x3b",
	.desc = "Number of thermal trips",
	.topic = "other",
},
{
	.name = "bus_request_outstanding.all_agents",
	.event = "umask=0xe0,period=200000,event=0x60",
	.desc = "Outstanding cacheable data read bus requests duration",
	.topic = "other",
},
{
	.name = "bus_request_outstanding.self",
	.event = "umask=0x40,period=200000,event=0x60",
	.desc = "Outstanding cacheable data read bus requests duration",
	.topic = "other",
},
{
	.name = "bus_bnr_drv.all_agents",
	.event = "umask=0x20,period=200000,event=0x61",
	.desc = "Number of Bus Not Ready signals asserted",
	.topic = "other",
},
{
	.name = "bus_bnr_drv.this_agent",
	.event = "umask=0x0,period=200000,event=0x61",
	.desc = "Number of Bus Not Ready signals asserted",
	.topic = "other",
},
{
	.name = "bus_drdy_clocks.all_agents",
	.event = "umask=0x20,period=200000,event=0x62",
	.desc = "Bus cycles when data is sent on the bus",
	.topic = "other",
},
{
	.name = "bus_drdy_clocks.this_agent",
	.event = "umask=0x0,period=200000,event=0x62",
	.desc = "Bus cycles when data is sent on the bus",
	.topic = "other",
},
{
	.name = "bus_lock_clocks.all_agents",
	.event = "umask=0xe0,period=200000,event=0x63",
	.desc = "Bus cycles when a LOCK signal is asserted",
	.topic = "other",
},
{
	.name = "bus_lock_clocks.self",
	.event = "umask=0x40,period=200000,event=0x63",
	.desc = "Bus cycles when a LOCK signal is asserted",
	.topic = "other",
},
{
	.name = "bus_data_rcv.self",
	.event = "umask=0x40,period=200000,event=0x64",
	.desc = "Bus cycles while processor receives data",
	.topic = "other",
},
{
	.name = "bus_trans_brd.all_agents",
	.event = "umask=0xe0,period=200000,event=0x65",
	.desc = "Burst read bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_brd.self",
	.event = "umask=0x40,period=200000,event=0x65",
	.desc = "Burst read bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_rfo.all_agents",
	.event = "umask=0xe0,period=200000,event=0x66",
	.desc = "RFO bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_rfo.self",
	.event = "umask=0x40,period=200000,event=0x66",
	.desc = "RFO bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_wb.all_agents",
	.event = "umask=0xe0,period=200000,event=0x67",
	.desc = "Explicit writeback bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_wb.self",
	.event = "umask=0x40,period=200000,event=0x67",
	.desc = "Explicit writeback bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_ifetch.all_agents",
	.event = "umask=0xe0,period=200000,event=0x68",
	.desc = "Instruction-fetch bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_ifetch.self",
	.event = "umask=0x40,period=200000,event=0x68",
	.desc = "Instruction-fetch bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_inval.all_agents",
	.event = "umask=0xe0,period=200000,event=0x69",
	.desc = "Invalidate bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_inval.self",
	.event = "umask=0x40,period=200000,event=0x69",
	.desc = "Invalidate bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_pwr.all_agents",
	.event = "umask=0xe0,period=200000,event=0x6a",
	.desc = "Partial write bus transaction",
	.topic = "other",
},
{
	.name = "bus_trans_pwr.self",
	.event = "umask=0x40,period=200000,event=0x6a",
	.desc = "Partial write bus transaction",
	.topic = "other",
},
{
	.name = "bus_trans_p.all_agents",
	.event = "umask=0xe0,period=200000,event=0x6b",
	.desc = "Partial bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_p.self",
	.event = "umask=0x40,period=200000,event=0x6b",
	.desc = "Partial bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_io.all_agents",
	.event = "umask=0xe0,period=200000,event=0x6c",
	.desc = "IO bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_io.self",
	.event = "umask=0x40,period=200000,event=0x6c",
	.desc = "IO bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_def.all_agents",
	.event = "umask=0xe0,period=200000,event=0x6d",
	.desc = "Deferred bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_def.self",
	.event = "umask=0x40,period=200000,event=0x6d",
	.desc = "Deferred bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_burst.all_agents",
	.event = "umask=0xe0,period=200000,event=0x6e",
	.desc = "Burst (full cache-line) bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_burst.self",
	.event = "umask=0x40,period=200000,event=0x6e",
	.desc = "Burst (full cache-line) bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_mem.all_agents",
	.event = "umask=0xe0,period=200000,event=0x6f",
	.desc = "Memory bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_mem.self",
	.event = "umask=0x40,period=200000,event=0x6f",
	.desc = "Memory bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_any.all_agents",
	.event = "umask=0xe0,period=200000,event=0x70",
	.desc = "All bus transactions",
	.topic = "other",
},
{
	.name = "bus_trans_any.self",
	.event = "umask=0x40,period=200000,event=0x70",
	.desc = "All bus transactions",
	.topic = "other",
},
{
	.name = "ext_snoop.this_agent.any",
	.event = "umask=0xb,period=200000,event=0x77",
	.desc = "External snoops",
	.topic = "other",
},
{
	.name = "ext_snoop.this_agent.clean",
	.event = "umask=0x1,period=200000,event=0x77",
	.desc = "External snoops",
	.topic = "other",
},
{
	.name = "ext_snoop.this_agent.hit",
	.event = "umask=0x2,period=200000,event=0x77",
	.desc = "External snoops",
	.topic = "other",
},
{
	.name = "ext_snoop.this_agent.hitm",
	.event = "umask=0x8,period=200000,event=0x77",
	.desc = "External snoops",
	.topic = "other",
},
{
	.name = "ext_snoop.all_agents.any",
	.event = "umask=0x2b,period=200000,event=0x77",
	.desc = "External snoops",
	.topic = "other",
},
{
	.name = "ext_snoop.all_agents.clean",
	.event = "umask=0x21,period=200000,event=0x77",
	.desc = "External snoops",
	.topic = "other",
},
{
	.name = "ext_snoop.all_agents.hit",
	.event = "umask=0x22,period=200000,event=0x77",
	.desc = "External snoops",
	.topic = "other",
},
{
	.name = "ext_snoop.all_agents.hitm",
	.event = "umask=0x28,period=200000,event=0x77",
	.desc = "External snoops",
	.topic = "other",
},
{
	.name = "bus_hit_drv.all_agents",
	.event = "umask=0x20,period=200000,event=0x7a",
	.desc = "HIT signal asserted",
	.topic = "other",
},
{
	.name = "bus_hit_drv.this_agent",
	.event = "umask=0x0,period=200000,event=0x7a",
	.desc = "HIT signal asserted",
	.topic = "other",
},
{
	.name = "bus_hitm_drv.all_agents",
	.event = "umask=0x20,period=200000,event=0x7b",
	.desc = "HITM signal asserted",
	.topic = "other",
},
{
	.name = "bus_hitm_drv.this_agent",
	.event = "umask=0x0,period=200000,event=0x7b",
	.desc = "HITM signal asserted",
	.topic = "other",
},
{
	.name = "busq_empty.self",
	.event = "umask=0x40,period=200000,event=0x7d",
	.desc = "Bus queue is empty",
	.topic = "other",
},
{
	.name = "snoop_stall_drv.all_agents",
	.event = "umask=0xe0,period=200000,event=0x7e",
	.desc = "Bus stalled for snoops",
	.topic = "other",
},
{
	.name = "snoop_stall_drv.self",
	.event = "umask=0x40,period=200000,event=0x7e",
	.desc = "Bus stalled for snoops",
	.topic = "other",
},
{
	.name = "bus_io_wait.self",
	.event = "umask=0x40,period=200000,event=0x7f",
	.desc = "IO requests waiting in the bus queue",
	.topic = "other",
},
{
	.name = "cycles_int_masked.cycles_int_masked",
	.event = "umask=0x1,period=2000000,event=0xc6",
	.desc = "Cycles during which interrupts are disabled",
	.topic = "other",
},
{
	.name = "cycles_int_masked.cycles_int_pending_and_masked",
	.event = "umask=0x2,period=2000000,event=0xc6",
	.desc = "Cycles during which interrupts are pending and disabled",
	.topic = "other",
},
{
	.name = "hw_int_rcv",
	.event = "umask=0x0,period=200000,event=0xc8",
	.desc = "Hardware interrupts received",
	.topic = "other",
},
{
	.name = "store_forwards.any",
	.event = "umask=0x83,period=200000,event=0x2",
	.desc = "All store forwards",
	.topic = "pipeline",
},
{
	.name = "store_forwards.good",
	.event = "umask=0x81,period=200000,event=0x2",
	.desc = "Good store forwards",
	.topic = "pipeline",
},
{
	.name = "reissue.any",
	.event = "umask=0x7f,period=200000,event=0x3",
	.desc = "Micro-op reissues for any cause",
	.topic = "pipeline",
},
{
	.name = "reissue.any.ar",
	.event = "umask=0xff,period=200000,event=0x3",
	.desc = "Micro-op reissues for any cause (At Retirement)",
	.topic = "pipeline",
},
{
	.name = "mul.s",
	.event = "umask=0x1,period=2000000,event=0x12",
	.desc = "Multiply operations executed",
	.topic = "pipeline",
},
{
	.name = "mul.ar",
	.event = "umask=0x81,period=2000000,event=0x12",
	.desc = "Multiply operations retired",
	.topic = "pipeline",
},
{
	.name = "div.s",
	.event = "umask=0x1,period=2000000,event=0x13",
	.desc = "Divide operations executed",
	.topic = "pipeline",
},
{
	.name = "div.ar",
	.event = "umask=0x81,period=2000000,event=0x13",
	.desc = "Divide operations retired",
	.topic = "pipeline",
},
{
	.name = "cycles_div_busy",
	.event = "umask=0x1,period=2000000,event=0x14",
	.desc = "Cycles the divider is busy",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.core_p",
	.event = "umask=0x0,period=2000000,event=0x3c",
	.desc = "Core cycles when core is not halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.bus",
	.event = "umask=0x1,period=200000,event=0x3c",
	.desc = "Bus cycles when core is not halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.core",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when core is not halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref",
	.event = "event=0x0,umask=0x03,period=2000003",
	.desc = "Reference cycles when core is not halted",
	.topic = "pipeline",
},
{
	.name = "br_inst_type_retired.cond",
	.event = "umask=0x1,period=2000000,event=0x88",
	.desc = "All macro conditional branch instructions",
	.topic = "pipeline",
},
{
	.name = "br_inst_type_retired.uncond",
	.event = "umask=0x2,period=2000000,event=0x88",
	.desc = "All macro unconditional branch instructions, excluding calls and indirects",
	.topic = "pipeline",
},
{
	.name = "br_inst_type_retired.ind",
	.event = "umask=0x4,period=2000000,event=0x88",
	.desc = "All indirect branches that are not calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_type_retired.ret",
	.event = "umask=0x8,period=2000000,event=0x88",
	.desc = "All indirect branches that have a return mnemonic",
	.topic = "pipeline",
},
{
	.name = "br_inst_type_retired.dir_call",
	.event = "umask=0x10,period=2000000,event=0x88",
	.desc = "All non-indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_inst_type_retired.ind_call",
	.event = "umask=0x20,period=2000000,event=0x88",
	.desc = "All indirect calls, including both register and memory indirect",
	.topic = "pipeline",
},
{
	.name = "br_inst_type_retired.cond_taken",
	.event = "umask=0x41,period=2000000,event=0x88",
	.desc = "Only taken macro conditional branch instructions",
	.topic = "pipeline",
},
{
	.name = "br_missp_type_retired.cond",
	.event = "umask=0x1,period=200000,event=0x89",
	.desc = "Mispredicted cond branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "br_missp_type_retired.ind",
	.event = "umask=0x2,period=200000,event=0x89",
	.desc = "Mispredicted ind branches that are not calls",
	.topic = "pipeline",
},
{
	.name = "br_missp_type_retired.return",
	.event = "umask=0x4,period=200000,event=0x89",
	.desc = "Mispredicted return branches",
	.topic = "pipeline",
},
{
	.name = "br_missp_type_retired.ind_call",
	.event = "umask=0x8,period=200000,event=0x89",
	.desc = "Mispredicted indirect calls, including both register and memory indirect",
	.topic = "pipeline",
},
{
	.name = "br_missp_type_retired.cond_taken",
	.event = "umask=0x11,period=200000,event=0x89",
	.desc = "Mispredicted and taken cond branch instructions retired",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired (precise event) (Must be precise)",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired",
	.topic = "pipeline",
},
{
	.name = "uops_retired.any",
	.event = "umask=0x10,period=2000000,event=0xc2",
	.desc = "Micro-ops retired",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stalled_cycles",
	.event = "umask=0x10,period=2000000,event=0xc2",
	.desc = "Cycles no micro-ops retired",
	.topic = "pipeline",
},
{
	.name = "uops_retired.stalls",
	.event = "umask=0x10,period=2000000,event=0xc2",
	.desc = "Periods no micro-ops retired",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x1,period=200000,event=0xc3",
	.desc = "Self-Modifying Code detected",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.any",
	.event = "umask=0x0,period=2000000,event=0xc4",
	.desc = "Retired branch instructions",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.pred_not_taken",
	.event = "umask=0x1,period=2000000,event=0xc4",
	.desc = "Retired branch instructions that were predicted not-taken",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.mispred_not_taken",
	.event = "umask=0x2,period=200000,event=0xc4",
	.desc = "Retired branch instructions that were mispredicted not-taken",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.pred_taken",
	.event = "umask=0x4,period=2000000,event=0xc4",
	.desc = "Retired branch instructions that were predicted taken",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.mispred_taken",
	.event = "umask=0x8,period=200000,event=0xc4",
	.desc = "Retired branch instructions that were mispredicted taken",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.taken",
	.event = "umask=0xc,period=2000000,event=0xc4",
	.desc = "Retired taken branch instructions",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.any1",
	.event = "umask=0xf,period=2000000,event=0xc4",
	.desc = "Retired branch instructions",
	.topic = "pipeline",
},
{
	.name = "br_inst_retired.mispred",
	.event = "umask=0x0,period=200000,event=0xc5",
	.desc = "Retired mispredicted branch instructions (precise event) (Precise event)",
	.topic = "pipeline",
},
{
	.name = "resource_stalls.div_busy",
	.event = "umask=0x2,period=2000000,event=0xdc",
	.desc = "Cycles issue is stalled due to div busy",
	.topic = "pipeline",
},
{
	.name = "br_inst_decoded",
	.event = "umask=0x1,period=2000000,event=0xe0",
	.desc = "Branch instructions decoded",
	.topic = "pipeline",
},
{
	.name = "bogus_br",
	.event = "umask=0x1,period=2000000,event=0xe4",
	.desc = "Bogus branches",
	.topic = "pipeline",
},
{
	.name = "baclears.any",
	.event = "umask=0x1,period=2000000,event=0xe6",
	.desc = "BACLEARS asserted",
	.topic = "pipeline",
},
{
	.name = "reissue.overlap_store",
	.event = "umask=0x1,period=200000,event=0x3",
	.desc = "Micro-op reissues on a store-load collision",
	.topic = "pipeline",
},
{
	.name = "reissue.overlap_store.ar",
	.event = "umask=0x81,period=200000,event=0x3",
	.desc = "Micro-op reissues on a store-load collision (At Retirement)",
	.topic = "pipeline",
},
{
	.name = "data_tlb_misses.dtlb_miss",
	.event = "umask=0x7,period=200000,event=0x8",
	.desc = "Memory accesses that missed the DTLB",
	.topic = "virtual memory",
},
{
	.name = "data_tlb_misses.dtlb_miss_ld",
	.event = "umask=0x5,period=200000,event=0x8",
	.desc = "DTLB misses due to load operations",
	.topic = "virtual memory",
},
{
	.name = "data_tlb_misses.l0_dtlb_miss_ld",
	.event = "umask=0x9,period=200000,event=0x8",
	.desc = "L0 DTLB misses due to load operations",
	.topic = "virtual memory",
},
{
	.name = "data_tlb_misses.dtlb_miss_st",
	.event = "umask=0x6,period=200000,event=0x8",
	.desc = "DTLB misses due to store operations",
	.topic = "virtual memory",
},
{
	.name = "data_tlb_misses.l0_dtlb_miss_st",
	.event = "umask=0xa,period=200000,event=0x8",
	.desc = "L0 DTLB misses due to store operations",
	.topic = "virtual memory",
},
{
	.name = "page_walks.walks",
	.event = "umask=0x3,period=200000,event=0xc",
	.desc = "Number of page-walks executed",
	.topic = "virtual memory",
},
{
	.name = "page_walks.cycles",
	.event = "umask=0x3,period=2000000,event=0xc",
	.desc = "Duration of page-walks in core cycles",
	.topic = "virtual memory",
},
{
	.name = "page_walks.d_side_walks",
	.event = "umask=0x1,period=200000,event=0xc",
	.desc = "Number of D-side only page walks",
	.topic = "virtual memory",
},
{
	.name = "page_walks.d_side_cycles",
	.event = "umask=0x1,period=2000000,event=0xc",
	.desc = "Duration of D-side only page walks",
	.topic = "virtual memory",
},
{
	.name = "page_walks.i_side_walks",
	.event = "umask=0x2,period=200000,event=0xc",
	.desc = "Number of I-Side page walks",
	.topic = "virtual memory",
},
{
	.name = "page_walks.i_side_cycles",
	.event = "umask=0x2,period=2000000,event=0xc",
	.desc = "Duration of I-Side page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb.hit",
	.event = "umask=0x1,period=200000,event=0x82",
	.desc = "ITLB hits",
	.topic = "virtual memory",
},
{
	.name = "itlb.flush",
	.event = "umask=0x4,period=200000,event=0x82",
	.desc = "ITLB flushes",
	.topic = "virtual memory",
},
{
	.name = "itlb.misses",
	.event = "umask=0x2,period=200000,event=0x82",
	.desc = "ITLB misses (Must be precise)",
	.topic = "virtual memory",
},
{
	.name = "mem_load_retired.dtlb_miss",
	.event = "umask=0x4,period=200000,event=0xcb",
	.desc = "Retired loads that miss the DTLB (precise event) (Precise event)",
	.topic = "virtual memory",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_event pme_broadwell[] = {
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend",
	.topic = "bdw metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound",
	.metric_expr = "idq_uops_not_delivered.core / (4 * cycles)",
	.metric_name = "Frontend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "bdw metrics",
	.long_desc = "This category represents fraction of slots where the processor's Frontend undersupplies its Backend. Frontend denotes the first part of the processor core responsible to fetch operations that are executed later on by the Backend part. Within the Frontend; a branch predictor predicts the next address to fetch; cache-lines are fetched from the memory subsystem; parsed into instructions; and lastly decoded into micro-ops (uops). Ideally the Frontend can issue 4 uops every cycle to the Backend. Frontend Bound denotes unutilized issue-slots when there is no Backend stall; i.e. bubbles where Frontend delivered no uops while Backend could have accepted them. For example; stalls due to instruction-cache misses would be categorized under Frontend Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Frontend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations",
	.topic = "bdw metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)",
	.metric_name = "Bad_Speculation",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots wasted due to incorrect speculations. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "bdw metrics",
	.long_desc = "This category represents fraction of slots wasted due to incorrect speculations. This include slots used to issue uops that do not eventually get retired and slots for which the issue-pipeline was blocked due to recovery from earlier incorrect speculation. For example; wasted work due to miss-predicted branches are categorized under Bad Speculation category. Incorrect data speculation followed by Memory Ordering Nukes is another example. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Bad_Speculation_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend",
	.topic = "bdw metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * cycles)) + (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles)) + (uops_retired.retire_slots / (4 * cycles)) )",
	.metric_name = "Backend_Bound",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "bdw metrics",
	.long_desc = "This category represents fraction of slots where no uops are being delivered due to a lack of required resources for accepting new uops in the Backend. Backend is the portion of the processor core where the out-of-order scheduler dispatches ready uops into their respective execution units; and once completed these uops get retired according to program order. For example; stalls due to data-cache misses or stalls due to the divider unit being overloaded are both categorized under Backend Bound. Backend Bound is further divided into two main categories: Memory Bound and Core Bound. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "1 - ( (idq_uops_not_delivered.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) + (uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) )",
	.metric_name = "Backend_Bound_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired",
	.topic = "bdw metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided",
	.metric_expr = "uops_retired.retire_slots / (4 * cycles)",
	.metric_name = "Retiring",
	.metric_group = "TopdownL1",
},
{
	.desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. SMT version; use when SMT is enabled and measuring per logical CPU",
	.topic = "bdw metrics",
	.long_desc = "This category represents fraction of slots utilized by useful work i.e. issued uops that eventually get retired. Ideally; all pipeline slots would be attributed to the Retiring category.  Retiring of 100% would indicate the maximum 4 uops retired per cycle has been achieved.  Maximizing Retiring typically increases the Instruction-Per-Cycle metric. Note that a high Retiring value does not necessary mean there is no room for more performance.  For example; Microcode assists are categorized under Retiring. They hurt performance and can often be avoided. SMT version; use when SMT is enabled and measuring per logical CPU",
	.metric_expr = "uops_retired.retire_slots / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))",
	.metric_name = "Retiring_SMT",
	.metric_group = "TopdownL1_SMT",
},
{
	.desc = "Instructions Per Cycle (per logical thread)",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any / cpu_clk_unhalted.thread",
	.metric_name = "IPC",
	.metric_group = "TopDownL1",
},
{
	.desc = "Uops Per Instruction",
	.topic = "bdw metrics",
	.metric_expr = "uops_retired.retire_slots / inst_retired.any",
	.metric_name = "UPI",
	.metric_group = "Pipeline;Retiring",
},
{
	.desc = "Instruction per taken branch",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_taken",
	.metric_name = "IpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Branch instructions per taken branch",
	.topic = "bdw metrics",
	.metric_expr = "br_inst_retired.all_branches / br_inst_retired.near_taken",
	.metric_name = "BpTB",
	.metric_group = "Branches;PGO",
},
{
	.desc = "Rough Estimation of fraction of fetched lines bytes that were likely (includes speculatively fetches) consumed by program instructions",
	.topic = "bdw metrics",
	.metric_expr = "min( 1 , idq.mite_uops / ( (uops_retired.retire_slots / inst_retired.any) * 16 * ( icache.hit + icache.misses ) / 4.0 ) )",
	.metric_name = "IFetch_Line_Utilization",
	.metric_group = "PGO",
},
{
	.desc = "Fraction of Uops delivered by the DSB (aka Decoded ICache; or Uop Cache)",
	.topic = "bdw metrics",
	.metric_expr = "idq.dsb_uops / (( idq.dsb_uops + lsd.uops + idq.mite_uops + idq.ms_uops ) )",
	.metric_name = "DSB_Coverage",
	.metric_group = "DSB;Frontend_Bandwidth",
},
{
	.desc = "Cycles Per Instruction (threaded)",
	.topic = "bdw metrics",
	.metric_expr = "1 / (inst_retired.any / cycles)",
	.metric_name = "CPI",
	.metric_group = "Pipeline;Summary",
},
{
	.desc = "Per-thread actual clocks when the logical processor is active",
	.topic = "bdw metrics",
	.metric_expr = "cpu_clk_unhalted.thread",
	.metric_name = "CLKS",
	.metric_group = "Summary",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "bdw metrics",
	.metric_expr = "4 * cycles",
	.metric_name = "SLOTS",
	.metric_group = "TopDownL1",
},
{
	.desc = "Total issue-pipeline slots (per core)",
	.topic = "bdw metrics",
	.metric_expr = "4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "SLOTS_SMT",
	.metric_group = "TopDownL1_SMT",
},
{
	.desc = "Instructions per Load (lower number means loads are more frequent)",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_loads",
	.metric_name = "IpL",
	.metric_group = "Instruction_Type;L1_Bound",
},
{
	.desc = "Instructions per Store",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any / mem_uops_retired.all_stores",
	.metric_name = "IpS",
	.metric_group = "Instruction_Type;Store_Bound",
},
{
	.desc = "Instructions per Branch",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.all_branches",
	.metric_name = "IpB",
	.metric_group = "Branches;Instruction_Type;Port_5;Port_6",
},
{
	.desc = "Instruction per (near) call",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any / br_inst_retired.near_call",
	.metric_name = "IpCall",
	.metric_group = "Branches",
},
{
	.desc = "Total number of retired Instructions",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any",
	.metric_name = "Instructions",
	.metric_group = "Summary",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any / cycles",
	.metric_name = "CoreIPC",
	.metric_group = "SMT",
},
{
	.desc = "Instructions Per Cycle (per physical core)",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "CoreIPC_SMT",
	.metric_group = "SMT",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "bdw metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * fp_arith_inst_retired.256b_packed_single )) / cycles",
	.metric_name = "FLOPc",
	.metric_group = "FLOPS",
},
{
	.desc = "Floating Point Operations Per Cycle",
	.topic = "bdw metrics",
	.metric_expr = "(( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * fp_arith_inst_retired.256b_packed_single )) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "FLOPc_SMT",
	.metric_group = "FLOPS_SMT",
},
{
	.desc = "Instruction-Level-Parallelism (average number of uops executed when there is at least 1 uop executed)",
	.topic = "bdw metrics",
	.metric_expr = "uops_executed.thread / (( cpu@uops_executed.core\\,cmask\\=1@ / 2 ) if #smt_on else uops_executed.cycles_ge_1_uop_exec)",
	.metric_name = "ILP",
	.metric_group = "Pipeline;Ports_Utilization",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "bdw metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * int_misc.recovery_cycles ) / (4 * cycles))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) * (12 * ( br_misp_retired.all_branches + machine_clears.count + baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * cycles)) ) * (4 * cycles) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Branch Misprediction Cost: Fraction of TopDown slots wasted per branch misprediction (jeclear and baclear)",
	.topic = "bdw metrics",
	.metric_expr = "( ((br_misp_retired.all_branches / ( br_misp_retired.all_branches + machine_clears.count )) * (( uops_issued.any - uops_retired.retire_slots + 4 * (( int_misc.recovery_cycles_any / 2 )) ) / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))))) + (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) * (12 * ( br_misp_retired.all_branches + machine_clears.count + baclears.any ) / cycles) / (4 * idq_uops_not_delivered.cycles_0_uops_deliv.core / (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )))) ) * (4 * (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))) / br_misp_retired.all_branches",
	.metric_name = "Branch_Misprediction_Cost_SMT",
	.metric_group = "Branch_Mispredicts_SMT",
},
{
	.desc = "Number of Instructions per non-speculative Branch Misprediction (JEClear)",
	.topic = "bdw metrics",
	.metric_expr = "inst_retired.any / br_misp_retired.all_branches",
	.metric_name = "IpMispredict",
	.metric_group = "Branch_Mispredicts",
},
{
	.desc = "Core actual clocks when any thread is active on the physical core",
	.topic = "bdw metrics",
	.metric_expr = "( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) )",
	.metric_name = "CORE_CLKS",
	.metric_group = "SMT",
},
{
	.desc = "Actual Average Latency for L1 data-cache miss demand loads (in core cycles)",
	.topic = "bdw metrics",
	.metric_expr = "l1d_pend_miss.pending / ( mem_load_uops_retired.l1_miss + mem_load_uops_retired.hit_lfb )",
	.metric_name = "Load_Miss_Real_Latency",
	.metric_group = "Memory_Bound;Memory_Lat",
},
{
	.desc = "Memory-Level-Parallelism (average number of L1 miss demand load when there is at least one such miss. Per-thread)",
	.topic = "bdw metrics",
	.metric_expr = "l1d_pend_miss.pending / l1d_pend_miss.pending_cycles",
	.metric_name = "MLP",
	.metric_group = "Memory_Bound;Memory_BW",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "bdw metrics",
	.metric_expr = "( cpu@itlb_misses.walk_duration\\,cmask\\=1@ + cpu@dtlb_load_misses.walk_duration\\,cmask\\=1@ + cpu@dtlb_store_misses.walk_duration\\,cmask\\=1@ + 7 * ( dtlb_store_misses.walk_completed + dtlb_load_misses.walk_completed + itlb_misses.walk_completed ) ) / cycles",
	.metric_name = "Page_Walks_Utilization",
	.metric_group = "TLB",
},
{
	.desc = "Utilization of the core's Page Walker(s) serving STLB misses triggered by instruction/Load/Store accesses",
	.topic = "bdw metrics",
	.metric_expr = "( cpu@itlb_misses.walk_duration\\,cmask\\=1@ + cpu@dtlb_load_misses.walk_duration\\,cmask\\=1@ + cpu@dtlb_store_misses.walk_duration\\,cmask\\=1@ + 7 * ( dtlb_store_misses.walk_completed + dtlb_load_misses.walk_completed + itlb_misses.walk_completed ) ) / (( ( cpu_clk_unhalted.thread / 2 ) * ( 1 + cpu_clk_unhalted.one_thread_active / cpu_clk_unhalted.ref_xclk ) ))",
	.metric_name = "Page_Walks_Utilization_SMT",
	.metric_group = "TLB_SMT",
},
{
	.desc = "Average data fill bandwidth to the L1 data cache [GB / sec]",
	.topic = "bdw metrics",
	.metric_expr = "64 * l1d.replacement / 1000000000 / duration_time",
	.metric_name = "L1D_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average data fill bandwidth to the L2 cache [GB / sec]",
	.topic = "bdw metrics",
	.metric_expr = "64 * l2_lines_in.all / 1000000000 / duration_time",
	.metric_name = "L2_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "Average per-core data fill bandwidth to the L3 cache [GB / sec]",
	.topic = "bdw metrics",
	.metric_expr = "64 * longest_lat_cache.miss / 1000000000 / duration_time",
	.metric_name = "L3_Cache_Fill_BW",
	.metric_group = "Memory_BW",
},
{
	.desc = "L1 cache true misses per kilo instruction for retired demand loads",
	.topic = "bdw metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l1_miss / inst_retired.any",
	.metric_name = "L1MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache true misses per kilo instruction for retired demand loads",
	.topic = "bdw metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l2_miss / inst_retired.any",
	.metric_name = "L2MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache misses per kilo instruction for all request types (including speculative)",
	.topic = "bdw metrics",
	.metric_expr = "1000 * l2_rqsts.miss / inst_retired.any",
	.metric_name = "L2MPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L2 cache hits per kilo instruction for all request types (including speculative)",
	.topic = "bdw metrics",
	.metric_expr = "1000 * ( l2_rqsts.references - l2_rqsts.miss ) / inst_retired.any",
	.metric_name = "L2HPKI_All",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "L3 cache true misses per kilo instruction for retired demand loads",
	.topic = "bdw metrics",
	.metric_expr = "1000 * mem_load_uops_retired.l3_miss / inst_retired.any",
	.metric_name = "L3MPKI",
	.metric_group = "Cache_Misses;",
},
{
	.desc = "Average CPU Utilization",
	.topic = "bdw metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc / msr@tsc@",
	.metric_name = "CPU_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Giga Floating Point Operations Per Second",
	.topic = "bdw metrics",
	.metric_expr = "( (( 1 * ( fp_arith_inst_retired.scalar_single + fp_arith_inst_retired.scalar_double ) + 2 * fp_arith_inst_retired.128b_packed_double + 4 * ( fp_arith_inst_retired.128b_packed_single + fp_arith_inst_retired.256b_packed_double ) + 8 * fp_arith_inst_retired.256b_packed_single )) / 1000000000 ) / duration_time",
	.metric_name = "GFLOPs",
	.metric_group = "FLOPS;Summary",
},
{
	.desc = "Average Frequency Utilization relative nominal frequency",
	.topic = "bdw metrics",
	.metric_expr = "cpu_clk_unhalted.thread / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Turbo_Utilization",
	.metric_group = "Power",
},
{
	.desc = "Fraction of cycles where both hardware threads were active",
	.topic = "bdw metrics",
	.metric_expr = "1 - cpu_clk_thread_unhalted.one_thread_active / ( cpu_clk_thread_unhalted.ref_xclk_any / 2 ) if #smt_on else 0",
	.metric_name = "SMT_2T_Utilization",
	.metric_group = "SMT;Summary",
},
{
	.desc = "Fraction of cycles spent in Kernel mode",
	.topic = "bdw metrics",
	.metric_expr = "cpu_clk_unhalted.ref_tsc:u / cpu_clk_unhalted.ref_tsc",
	.metric_name = "Kernel_Utilization",
	.metric_group = "Summary",
},
{
	.desc = "Average external Memory Bandwidth Use for reads and writes [GB / sec]",
	.topic = "bdw metrics",
	.metric_expr = "64 * ( arb@event\\=0x81\\,umask\\=0x1@ + arb@event\\=0x84\\,umask\\=0x1@ ) / 1000000 / duration_time / 1000",
	.metric_name = "DRAM_BW_Use",
	.metric_group = "Memory_BW",
},
{
	.desc = "C3 residency percent per core",
	.topic = "bdw metrics",
	.metric_expr = "(cstate_core@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per core",
	.topic = "bdw metrics",
	.metric_expr = "(cstate_core@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per core",
	.topic = "bdw metrics",
	.metric_expr = "(cstate_core@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Core_Residency",
	.metric_group = "Power",
},
{
	.desc = "C2 residency percent per package",
	.topic = "bdw metrics",
	.metric_expr = "(cstate_pkg@c2\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C2_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C3 residency percent per package",
	.topic = "bdw metrics",
	.metric_expr = "(cstate_pkg@c3\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C3_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C6 residency percent per package",
	.topic = "bdw metrics",
	.metric_expr = "(cstate_pkg@c6\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C6_Pkg_Residency",
	.metric_group = "Power",
},
{
	.desc = "C7 residency percent per package",
	.topic = "bdw metrics",
	.metric_expr = "(cstate_pkg@c7\\-residency@ / msr@tsc@) * 100",
	.metric_name = "C7_Pkg_Residency",
	.metric_group = "Power",
},
{
	.name = "l2_rqsts.demand_data_rd_miss",
	.event = "umask=0x21,period=200003,event=0x24",
	.desc = "Demand Data Read miss L2, no rejects",
	.topic = "cache",
	.long_desc = "This event counts the number of demand Data Read requests that miss L2 cache. Only not rejected loads are counted",
},
{
	.name = "l2_rqsts.rfo_miss",
	.event = "umask=0x22,period=200003,event=0x24",
	.desc = "RFO requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_miss",
	.event = "umask=0x24,period=200003,event=0x24",
	.desc = "L2 cache misses when fetching instructions",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_demand_miss",
	.event = "umask=0x27,period=200003,event=0x24",
	.desc = "Demand requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.l2_pf_miss",
	.event = "umask=0x30,period=200003,event=0x24",
	.desc = "L2 prefetch requests that miss L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of requests from the L2 hardware prefetchers that miss L2 cache",
},
{
	.name = "l2_rqsts.miss",
	.event = "umask=0x3f,period=200003,event=0x24",
	.desc = "All requests that miss L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.demand_data_rd_hit",
	.event = "umask=0xc1,period=200003,event=0x24",
	.desc = "Demand Data Read requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "Counts the number of demand Data Read requests, initiated by load instructions, that hit L2 cache",
},
{
	.name = "l2_rqsts.rfo_hit",
	.event = "umask=0xc2,period=200003,event=0x24",
	.desc = "RFO requests that hit L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.code_rd_hit",
	.event = "umask=0xc4,period=200003,event=0x24",
	.desc = "L2 cache hits when fetching instructions, code reads",
	.topic = "cache",
},
{
	.name = "l2_rqsts.l2_pf_hit",
	.event = "umask=0xd0,period=200003,event=0x24",
	.desc = "L2 prefetch requests that hit L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of requests from the L2 hardware prefetchers that hit L2 cache. L3 prefetch new types",
},
{
	.name = "l2_rqsts.all_demand_data_rd",
	.event = "umask=0xe1,period=200003,event=0x24",
	.desc = "Demand Data Read requests",
	.topic = "cache",
	.long_desc = "This event counts the number of demand Data Read requests (including requests from L1D hardware prefetchers). These loads may hit or miss L2 cache. Only non rejected loads are counted",
},
{
	.name = "l2_rqsts.all_rfo",
	.event = "umask=0xe2,period=200003,event=0x24",
	.desc = "RFO requests to L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the total number of RFO (read for ownership) requests to L2 cache. L2 RFO requests include both L1D demand RFO misses as well as L1D RFO prefetches",
},
{
	.name = "l2_rqsts.all_code_rd",
	.event = "umask=0xe4,period=200003,event=0x24",
	.desc = "L2 code requests",
	.topic = "cache",
	.long_desc = "This event counts the total number of L2 code requests",
},
{
	.name = "l2_rqsts.all_demand_references",
	.event = "umask=0xe7,period=200003,event=0x24",
	.desc = "Demand requests to L2 cache",
	.topic = "cache",
},
{
	.name = "l2_rqsts.all_pf",
	.event = "umask=0xf8,period=200003,event=0x24",
	.desc = "Requests from L2 hardware prefetchers",
	.topic = "cache",
	.long_desc = "This event counts the total number of requests from the L2 hardware prefetchers",
},
{
	.name = "l2_rqsts.references",
	.event = "umask=0xff,period=200003,event=0x24",
	.desc = "All L2 requests",
	.topic = "cache",
},
{
	.name = "l2_demand_rqsts.wb_hit",
	.event = "umask=0x50,period=200003,event=0x27",
	.desc = "Not rejected writebacks that hit L2 cache",
	.topic = "cache",
	.long_desc = "This event counts the number of WB requests that hit L2 cache",
},
{
	.name = "longest_lat_cache.miss",
	.event = "umask=0x41,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests missed L3",
	.topic = "cache",
	.long_desc = "This event counts core-originated cacheable demand requests that miss the last level cache (LLC). Demand requests include loads, RFOs, and hardware prefetches from L1D, and instruction fetches from IFU",
},
{
	.name = "longest_lat_cache.reference",
	.event = "umask=0x4f,period=100003,event=0x2e",
	.desc = "Core-originated cacheable demand requests that refer to L3",
	.topic = "cache",
	.long_desc = "This event counts core-originated cacheable demand requests that refer to the last level cache (LLC). Demand requests include loads, RFOs, and hardware prefetches from L1D, and instruction fetches from IFU",
},
{
	.name = "l1d_pend_miss.pending",
	.event = "umask=0x1,period=2000003,event=0x48",
	.desc = "L1D miss oustandings duration in cycles",
	.topic = "cache",
	.long_desc = "This event counts duration of L1D miss outstanding, that is each cycle number of Fill Buffers (FB) outstanding required by Demand Reads. FB either is held by demand loads, or it is held by non-demand loads and gets hit at least once by demand. The valid outstanding interval is defined until the FB deallocation by one of the following ways: from FB allocation, if FB is allocated by demand; from the demand Hit FB, if it is allocated by hardware or software prefetch.\nNote: In the L1D, a Demand Read contains cacheable or noncacheable demand loads, including ones causing cache-line splits and reads due to page walks resulted from any request type",
},
{
	.name = "l1d_pend_miss.pending_cycles",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding",
	.topic = "cache",
	.long_desc = "This event counts duration of L1D miss outstanding in cycles",
},
{
	.name = "l1d_pend_miss.pending_cycles_any",
	.event = "umask=0x1,any=1,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles with L1D load Misses outstanding from any thread on physical core",
	.topic = "cache",
},
{
	.name = "l1d_pend_miss.fb_full",
	.event = "umask=0x2,period=2000003,cmask=1,event=0x48",
	.desc = "Cycles a demand request was blocked due to Fill Buffers inavailability",
	.topic = "cache",
},
{
	.name = "l1d.replacement",
	.event = "umask=0x1,period=2000003,event=0x51",
	.desc = "L1D data line replacements",
	.topic = "cache",
	.long_desc = "This event counts L1D data line replacements including opportunistic replacements, and replacements that require stall-for-replace or block-for-replace",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd",
	.event = "umask=0x1,period=2000003,event=0x60",
	.desc = "Offcore outstanding Demand Data Read transactions in uncore queue  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding Demand Data Read transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor. See the corresponding Umask under OFFCORE_REQUESTS.\nNote: A prefetch promoted to Demand is counted from the promotion point  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_data_rd",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding Demand Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts cycles when offcore outstanding Demand Data Read transactions are present in the super queue (SQ). A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation)  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.demand_data_rd_ge_6",
	.event = "umask=0x1,period=2000003,cmask=6,event=0x60",
	.desc = "Cycles with at least 6 offcore outstanding Demand Data Read transactions in uncore queue  Spec update: BDM76",
	.topic = "cache",
},
{
	.name = "offcore_requests_outstanding.demand_code_rd",
	.event = "umask=0x2,period=2000003,event=0x60",
	.desc = "Offcore outstanding code reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding Code Reads transactions in the super queue every cycle. The Offcore outstanding state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.demand_rfo",
	.event = "umask=0x4,period=2000003,event=0x60",
	.desc = "Offcore outstanding RFO store transactions in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding RFO (store) transactions in the super queue (SQ) every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.cycles_with_demand_rfo",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x60",
	.desc = "Offcore outstanding demand rfo reads transactions in SuperQueue (SQ), queue to uncore, every cycle  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding demand rfo Reads transactions in the super queue every cycle. The Offcore outstanding state of the transaction lasts from the L2 miss until the sending transaction completion to requestor (SQ deallocation). See the corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.all_data_rd",
	.event = "umask=0x8,period=2000003,event=0x60",
	.desc = "Offcore outstanding cacheable Core Data Read transactions in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts the number of offcore outstanding cacheable Core Data Read transactions in the super queue every cycle. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "offcore_requests_outstanding.cycles_with_data_rd",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x60",
	.desc = "Cycles when offcore outstanding cacheable Core Data Read transactions are present in SuperQueue (SQ), queue to uncore  Spec update: BDM76",
	.topic = "cache",
	.long_desc = "This event counts cycles when offcore outstanding cacheable Core Data Read transactions are present in the super queue. A transaction is considered to be in the Offcore outstanding state between L2 miss and transaction completion sent to requestor (SQ de-allocation). See corresponding Umask under OFFCORE_REQUESTS  Spec update: BDM76",
},
{
	.name = "lock_cycles.cache_lock_duration",
	.event = "umask=0x2,period=2000003,event=0x63",
	.desc = "Cycles when L1D is locked",
	.topic = "cache",
	.long_desc = "This event counts the number of cycles when the L1D is locked. It is a superset of the 0x1 mask (BUS_LOCK_CLOCKS.BUS_LOCK_DURATION)",
},
{
	.name = "offcore_requests.demand_data_rd",
	.event = "umask=0x1,period=100003,event=0xb0",
	.desc = "Demand Data Read requests sent to uncore",
	.topic = "cache",
	.long_desc = "This event counts the Demand Data Read requests sent to uncore. Use it in conjunction with OFFCORE_REQUESTS_OUTSTANDING to determine average latency in the uncore",
},
{
	.name = "offcore_requests.demand_code_rd",
	.event = "umask=0x2,period=100003,event=0xb0",
	.desc = "Cacheable and noncachaeble code read requests",
	.topic = "cache",
	.long_desc = "This event counts both cacheable and noncachaeble code read requests",
},
{
	.name = "offcore_requests.demand_rfo",
	.event = "umask=0x4,period=100003,event=0xb0",
	.desc = "Demand RFO requests including regular RFOs, locks, ItoM",
	.topic = "cache",
	.long_desc = "This event counts the demand RFO (read for ownership) requests including regular RFOs, locks, ItoM",
},
{
	.name = "offcore_requests.all_data_rd",
	.event = "umask=0x8,period=100003,event=0xb0",
	.desc = "Demand and prefetch data reads",
	.topic = "cache",
	.long_desc = "This event counts the demand and prefetch data reads. All Core Data Reads include cacheable Demands and L2 prefetchers (not L3 prefetchers). Counting also covers reads due to page walks resulted from any request type",
},
{
	.name = "offcore_requests_buffer.sq_full",
	.event = "umask=0x1,period=2000003,event=0xb2",
	.desc = "Offcore requests buffer cannot take more entries for this thread core",
	.topic = "cache",
	.long_desc = "This event counts the number of cases when the offcore requests buffer cannot take more entries for the core. This can happen when the superqueue does not contain eligible entries, or when L1D writeback pending FIFO requests is full.\nNote: Writeback pending FIFO has six entries",
},
{
	.name = "offcore_response",
	.event = "umask=0x1,period=100003,event=0xb7",
	.desc = "Offcore response can be programmed only with a specific pair of event select and counter MSR, and with specific event codes and predefine mask bit value in a dedicated MSR to specify attributes of the offcore transaction",
	.topic = "cache",
},
{
	.name = "mem_uops_retired.stlb_miss_loads",
	.event = "umask=0x11,period=100003,event=0xd0",
	.desc = "Retired load uops that miss the STLB. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts load uops with true STLB miss retired to the architected path. True STLB miss is an uop triggering page walk that gets completed without blocks, and later gets retired. This page walk can end up with or without a fault  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.stlb_miss_stores",
	.event = "umask=0x12,period=100003,event=0xd0",
	.desc = "Retired store uops that miss the STLB. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts store uops true STLB miss retired to the architected path. True STLB miss is an uop triggering page walk that gets completed without blocks, and later gets retired. This page walk can end up with or without a fault  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.lock_loads",
	.event = "umask=0x21,period=100007,event=0xd0",
	.desc = "Retired load uops with locked access. (Precise Event - PEBS)  Spec update: BDM35.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts load uops with locked access retired to the architected path  Spec update: BDM35.  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.split_loads",
	.event = "umask=0x41,period=100003,event=0xd0",
	.desc = "Retired load uops that split across a cacheline boundary.(Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts line-splitted load uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K)  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.split_stores",
	.event = "umask=0x42,period=100003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts line-splitted store uops retired to the architected path. A line split is across 64B cache-line which includes a page split (4K)  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.all_loads",
	.event = "umask=0x81,period=2000003,event=0xd0",
	.desc = "All retired load uops. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts load uops retired to the architected path with a filter on bits 0 and 1 applied.\nNote: This event ?ounts AVX-256bit load/store double-pump memory uops as a single uop at retirement. This event also counts SW prefetches  Supports address when precise (Precise event)",
},
{
	.name = "mem_uops_retired.all_stores",
	.event = "umask=0x82,period=2000003,event=0xd0",
	.desc = "Retired store uops that split across a cacheline boundary. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts store uops retired to the architected path with a filter on bits 0 and 1 applied.\nNote: This event ?ounts AVX-256bit load/store double-pump memory uops as a single uop at retirement  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_hit",
	.event = "umask=0x1,period=2000003,event=0xd1",
	.desc = "Retired load uops with L1 cache hits as data sources. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data source were hits in the nearest-level (L1) cache.\nNote: Only two data-sources of L1/FB are applicable for AVX-256bit  even though the corresponding AVX load could be serviced by a deeper level in the memory hierarchy. Data source is reported for the Low-half load. This event also counts SW prefetches independent of the actual data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l2_hit",
	.event = "umask=0x2,period=100003,event=0xd1",
	.desc = "Retired load uops with L2 cache hits as data sources. (Precise Event - PEBS)  Spec update: BDM35.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were hits in the mid-level (L2) cache  Spec update: BDM35.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l3_hit",
	.event = "umask=0x4,period=50021,event=0xd1",
	.desc = "Hit in last-level (L3) cache. Excludes Unknown data-source. (Precise Event - PEBS)  Spec update: BDM100.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were data hits in the last-level (L3) cache without snoops required  Spec update: BDM100.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l1_miss",
	.event = "umask=0x8,period=100003,event=0xd1",
	.desc = "Retired load uops misses in L1 cache as data sources. Uses PEBS  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were misses in the nearest-level (L1) cache. Counting excludes unknown and UC data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l2_miss",
	.event = "umask=0x10,period=50021,event=0xd1",
	.desc = "Retired load uops with L2 cache misses as data sources. Uses PEBS  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were misses in the mid-level (L2) cache. Counting excludes unknown and UC data source  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_retired.l3_miss",
	.event = "umask=0x20,period=100007,event=0xd1",
	.desc = "Miss in last-level (L3) cache. Excludes Unknown data-source. (Precise Event - PEBS)  Spec update: BDM100, BDE70.  Supports address when precise (Precise event)",
	.topic = "cache",
},
{
	.name = "mem_load_uops_retired.hit_lfb",
	.event = "umask=0x40,period=100003,event=0xd1",
	.desc = "Retired load uops which data sources were load uops missed L1 but hit FB due to preceding miss to the same cache line with data not ready. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were load uops missed L1 but hit a fill buffer due to a preceding miss to the same cache line with the data not ready.\nNote: Only two data-sources of L1/FB are applicable for AVX-256bit  even though the corresponding AVX load could be serviced by a deeper level in the memory hierarchy. Data source is reported for the Low-half load  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_miss",
	.event = "umask=0x1,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 hit and cross-core snoop missed in on-pkg core cache. (Precise Event - PEBS)  Spec update: BDM100.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were L3 Hit and a cross-core snoop missed in the on-pkg core cache  Spec update: BDM100.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hit",
	.event = "umask=0x2,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were L3 and cross-core snoop hits in on-pkg core cache. (Precise Event - PEBS)  Spec update: BDM100.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were L3 hit and a cross-core snoop hit in the on-pkg core cache  Spec update: BDM100.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_hitm",
	.event = "umask=0x4,period=20011,event=0xd2",
	.desc = "Retired load uops which data sources were HitM responses from shared L3. (Precise Event - PEBS)  Spec update: BDM100.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were HitM responses from a core on same socket (shared L3)  Spec update: BDM100.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_l3_hit_retired.xsnp_none",
	.event = "umask=0x8,period=100003,event=0xd2",
	.desc = "Retired load uops which data sources were hits in L3 without snoops required. (Precise Event - PEBS)  Spec update: BDM100.  Supports address when precise (Precise event)",
	.topic = "cache",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts retired load uops which data sources were hits in the last-level (L3) cache without snoops required  Spec update: BDM100.  Supports address when precise (Precise event)",
},
{
	.name = "mem_load_uops_l3_miss_retired.local_dram",
	.event = "umask=0x1,period=100007,event=0xd3",
	.desc = "(null)",
	.topic = "cache",
	.long_desc = "This event counts retired load uops where the data came from local DRAM. This does not include hardware prefetches. This is a precise event  Spec update: BDE70, BDM100.  Supports address when precise",
},
{
	.name = "l2_trans.demand_data_rd",
	.event = "umask=0x1,period=200003,event=0xf0",
	.desc = "Demand Data Read requests that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts Demand Data Read requests that access L2 cache, including rejects",
},
{
	.name = "l2_trans.rfo",
	.event = "umask=0x2,period=200003,event=0xf0",
	.desc = "RFO requests that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts Read for Ownership (RFO) requests that access L2 cache",
},
{
	.name = "l2_trans.code_rd",
	.event = "umask=0x4,period=200003,event=0xf0",
	.desc = "L2 cache accesses when fetching instructions",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache accesses when fetching instructions",
},
{
	.name = "l2_trans.all_pf",
	.event = "umask=0x8,period=200003,event=0xf0",
	.desc = "L2 or L3 HW prefetches that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L2 or L3 HW prefetches that access L2 cache including rejects",
},
{
	.name = "l2_trans.l1d_wb",
	.event = "umask=0x10,period=200003,event=0xf0",
	.desc = "L1D writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L1D writebacks that access L2 cache",
},
{
	.name = "l2_trans.l2_fill",
	.event = "umask=0x20,period=200003,event=0xf0",
	.desc = "L2 fill requests that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L2 fill requests that access L2 cache",
},
{
	.name = "l2_trans.l2_wb",
	.event = "umask=0x40,period=200003,event=0xf0",
	.desc = "L2 writebacks that access L2 cache",
	.topic = "cache",
	.long_desc = "This event counts L2 writebacks that access L2 cache",
},
{
	.name = "l2_trans.all_requests",
	.event = "umask=0x80,period=200003,event=0xf0",
	.desc = "Transactions accessing L2 pipe",
	.topic = "cache",
	.long_desc = "This event counts transactions that access the L2 pipe including snoops, pagewalks, and so on",
},
{
	.name = "l2_lines_in.i",
	.event = "umask=0x1,period=100003,event=0xf1",
	.desc = "L2 cache lines in I state filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines in the Invalidate state filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_in.s",
	.event = "umask=0x2,period=100003,event=0xf1",
	.desc = "L2 cache lines in S state filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines in the Shared state filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_in.e",
	.event = "umask=0x4,period=100003,event=0xf1",
	.desc = "L2 cache lines in E state filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines in the Exclusive state filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_in.all",
	.event = "umask=0x7,period=100003,event=0xf1",
	.desc = "L2 cache lines filling L2",
	.topic = "cache",
	.long_desc = "This event counts the number of L2 cache lines filling the L2. Counting does not cover rejects",
},
{
	.name = "l2_lines_out.demand_clean",
	.event = "umask=0x5,period=100003,event=0xf2",
	.desc = "Clean L2 cache lines evicted by demand",
	.topic = "cache",
},
{
	.name = "sq_misc.split_lock",
	.event = "umask=0x10,period=100003,event=0xf4",
	.desc = "Split locks in SQ",
	.topic = "cache",
	.long_desc = "This event counts the number of split locks in the super queue",
},
{
	.name = "offcore_response.demand_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010001",
	.desc = "Counts demand data reads have any response type",
	.topic = "cache",
	.long_desc = "Counts demand data reads have any response type",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0001",
	.desc = "Counts demand data reads",
	.topic = "cache",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010002",
	.desc = "Counts all demand data writes (RFOs) have any response type",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs) have any response type",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "cache",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010004",
	.desc = "Counts all demand code reads have any response type",
	.topic = "cache",
	.long_desc = "Counts all demand code reads have any response type",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0004",
	.desc = "Counts all demand code reads",
	.topic = "cache",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.corewb.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010008",
	.desc = "Counts writebacks (modified to exclusive) have any response type",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive) have any response type",
},
{
	.name = "offcore_response.corewb.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "cache",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.pf_l2_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010010",
	.desc = "Counts prefetch (that bring data to L2) data reads have any response type",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads have any response type",
},
{
	.name = "offcore_response.pf_l2_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs have any response type",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs have any response type",
},
{
	.name = "offcore_response.pf_l2_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads have any response type",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads have any response type",
},
{
	.name = "offcore_response.pf_l2_code_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads have any response type",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads have any response type",
},
{
	.name = "offcore_response.pf_l3_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs have any response type",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs have any response type",
},
{
	.name = "offcore_response.pf_l3_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "cache",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads have any response type",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads have any response type",
},
{
	.name = "offcore_response.pf_l3_code_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "cache",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.other.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000018000",
	.desc = "Counts any other requests have any response type",
	.topic = "cache",
	.long_desc = "Counts any other requests have any response type",
},
{
	.name = "offcore_response.other.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80028000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C8000",
	.desc = "Counts any other requests",
	.topic = "cache",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.all_pf_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010090",
	.desc = "Counts all prefetch data reads have any response type",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads have any response type",
},
{
	.name = "offcore_response.all_pf_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0090",
	.desc = "Counts all prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010120",
	.desc = "Counts prefetch RFOs have any response type",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs have any response type",
},
{
	.name = "offcore_response.all_pf_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0120",
	.desc = "Counts prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_code_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010240",
	.desc = "Counts all prefetch code reads have any response type",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads have any response type",
},
{
	.name = "offcore_response.all_pf_code_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0240",
	.desc = "Counts all prefetch code reads",
	.topic = "cache",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_data_rd.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010091",
	.desc = "Counts all demand & prefetch data reads have any response type",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads have any response type",
},
{
	.name = "offcore_response.all_data_rd.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_rfo.any_response",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0000010122",
	.desc = "Counts all demand & prefetch RFOs have any response type",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs have any response type",
},
{
	.name = "offcore_response.all_rfo.supplier_none.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0080020122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.supplier_none.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0100020122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.supplier_none.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0200020122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.supplier_none.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0400020122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.supplier_none.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1000020122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.supplier_none.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F80020122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_hit.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00803C0122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_hit.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x01003C0122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_hit.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x02003C0122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_hit.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x04003C0122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_hit.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x10003C0122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_hit.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F803C0122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "cache",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "other_assists.avx_to_sse",
	.event = "umask=0x8,period=100003,event=0xc1",
	.desc = "Number of transitions from AVX-256 to legacy SSE when penalty applicable (Precise Event)  Spec update: BDM30",
	.topic = "floating point",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts the number of transitions from AVX-256 to legacy SSE when penalty is applicable  Spec update: BDM30",
},
{
	.name = "other_assists.sse_to_avx",
	.event = "umask=0x10,period=100003,event=0xc1",
	.desc = "Number of transitions from legacy SSE to AVX-256 when penalty applicable (Precise Event)  Spec update: BDM30",
	.topic = "floating point",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts the number of transitions from legacy SSE to AVX-256 when penalty is applicable  Spec update: BDM30",
},
{
	.name = "fp_arith_inst_retired.scalar_double",
	.event = "umask=0x1,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar double precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar double precision floating-point instructions: ADD SUB MUL DIV MIN MAX SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.scalar_single",
	.event = "umask=0x2,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar single precision floating-point instructions retired.  Each count represents 1 computation. Applies to SSE* and AVX* scalar single precision floating-point instructions: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT FM(N)ADD/SUB.  FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.scalar",
	.event = "umask=0x3,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational scalar floating-point instructions retired. Applies to SSE* and AVX* scalar, double and single precision floating-point: ADD SUB MUL DIV MIN MAX RSQRT RCP SQRT FM(N)ADD/SUB. FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element. (RSQRT for single precision?)",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_double",
	.event = "umask=0x4,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed double precision floating-point instructions retired.  Each count represents 2 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.128b_packed_single",
	.event = "umask=0x8,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 128-bit packed single precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT RSQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_double",
	.event = "umask=0x10,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed double precision floating-point instructions retired.  Each count represents 4 computations. Applies to SSE* and AVX* packed double precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.double",
	.event = "umask=0x15,period=2000006,event=0xc7",
	.desc = "Number of SSE/AVX computational double precision floating-point instructions retired. Applies to SSE* and AVX*scalar, double and single precision floating-point: ADD SUB MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.256b_packed_single",
	.event = "umask=0x20,period=2000003,event=0xc7",
	.desc = "Number of SSE/AVX computational 256-bit packed single precision floating-point instructions retired.  Each count represents 8 computations. Applies to SSE* and AVX* packed single precision floating-point instructions: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT RSQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.single",
	.event = "umask=0x2a,period=2000005,event=0xc7",
	.desc = "Number of SSE/AVX computational single precision floating-point instructions retired. Applies to SSE* and AVX*scalar, double and single precision floating-point: ADD SUB MUL DIV MIN MAX RCP RSQRT SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element",
	.topic = "floating point",
},
{
	.name = "fp_arith_inst_retired.packed",
	.event = "umask=0x3c,period=2000004,event=0xc7",
	.desc = "Number of SSE/AVX computational packed floating-point instructions retired. Applies to SSE* and AVX*, packed, double and single precision floating-point: ADD SUB HADD HSUB SUBADD MUL DIV MIN MAX SQRT DPP FM(N)ADD/SUB.  DPP and FM(N)ADD/SUB instructions count twice as they perform multiple calculations per element. (RSQRT for single-precision?)",
	.topic = "floating point",
},
{
	.name = "fp_assist.x87_output",
	.event = "umask=0x2,period=100003,event=0xca",
	.desc = "output - Numeric Overflow, Numeric Underflow, Inexact Result  (Precise Event)",
	.topic = "floating point",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts the number of x87 floating point (FP) micro-code assist (numeric overflow/underflow, inexact result) when the output value (destination register) is invalid",
},
{
	.name = "fp_assist.x87_input",
	.event = "umask=0x4,period=100003,event=0xca",
	.desc = "input - Invalid Operation, Denormal Operand, SNaN Operand  (Precise Event)",
	.topic = "floating point",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts x87 floating point (FP) micro-code assist (invalid operation, denormal operand, SNaN operand) when the input value (one of the source operands to an FP instruction) is invalid",
},
{
	.name = "fp_assist.simd_output",
	.event = "umask=0x8,period=100003,event=0xca",
	.desc = "SSE* FP micro-code assist when output value is invalid. (Precise Event)",
	.topic = "floating point",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts the number of SSE* floating point (FP) micro-code assist (numeric overflow/underflow) when the output value (destination register) is invalid. Counting covers only cases involving penalties that require micro-code assist intervention",
},
{
	.name = "fp_assist.simd_input",
	.event = "umask=0x10,period=100003,event=0xca",
	.desc = "Any input SSE* FP Assist -   (Precise Event)",
	.topic = "floating point",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts any input SSE* floating-point (FP) assist - invalid operation, denormal operand, dividing by zero, SNaN operand. Counting includes only cases involving penalties that required micro-code assist intervention",
},
{
	.name = "fp_assist.any",
	.event = "umask=0x1e,period=100003,cmask=1,event=0xca",
	.desc = "Counts any FP_ASSIST umask was incrementing   (Precise Event)",
	.topic = "floating point",
	.long_desc = "This event counts cycles with any input and output SSE or x87 FP assist. If an input and output assist are detected on the same cycle the event increments by 1. Uses PEBS",
},
{
	.name = "idq.empty",
	.event = "umask=0x2,period=2000003,event=0x79",
	.desc = "Instruction Decode Queue (IDQ) empty cycles",
	.topic = "frontend",
	.long_desc = "This counts the number of cycles that the instruction decoder queue is empty and can indicate that the application may be bound in the front end.  It does not determine whether there are uops being delivered to the Alloc stage since uops can be delivered by bypass skipping the Instruction Decode Queue (IDQ) when it is empty",
},
{
	.name = "idq.mite_uops",
	.event = "umask=0x4,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.mite_cycles",
	.event = "umask=0x4,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.dsb_uops",
	.event = "umask=0x8,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.dsb_cycles",
	.event = "umask=0x8,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) from Decode Stream Buffer (DSB) path",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_dsb_uops",
	.event = "umask=0x10,period=2000003,event=0x79",
	.desc = "Uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops initiated by Decode Stream Buffer (DSB) that are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_dsb_cycles",
	.event = "umask=0x10,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops initiated by Decode Stream Buffer (DSB) are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequencer (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_dsb_occur",
	.event = "umask=0x10,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the number of deliveries to Instruction Decode Queue (IDQ) initiated by Decode Stream Buffer (DSB) while the Microcode Sequencer (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.all_dsb_cycles_4_uops",
	.event = "umask=0x18,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles 4  uops were  delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.all_dsb_cycles_any_uops",
	.event = "umask=0x18,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles Decode Stream Buffer (DSB) is delivering any Uop",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles  uops were  delivered to Instruction Decode Queue (IDQ) from the Decode Stream Buffer (DSB) path. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.ms_mite_uops",
	.event = "umask=0x20,period=2000003,event=0x79",
	.desc = "Uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops initiated by MITE and delivered to Instruction Decode Queue (IDQ) while the Microcode Sequenser (MS) is busy. Counting includes uops that may bypass the IDQ",
},
{
	.name = "idq.all_mite_cycles_4_uops",
	.event = "umask=0x24,period=2000003,cmask=4,event=0x79",
	.desc = "Cycles MITE is delivering 4 Uops",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles 4  uops were  delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.all_mite_cycles_any_uops",
	.event = "umask=0x24,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles MITE is delivering any Uop",
	.topic = "frontend",
	.long_desc = "This event counts the number of cycles  uops were delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "idq.ms_uops",
	.event = "umask=0x30,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts the total number of uops delivered to Instruction Decode Queue (IDQ) while the Microcode Sequenser (MS) is busy. Counting includes uops that may bypass the IDQ. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "idq.ms_cycles",
	.event = "umask=0x30,period=2000003,cmask=1,event=0x79",
	.desc = "Cycles when uops are being delivered to Instruction Decode Queue (IDQ) while Microcode Sequenser (MS) is busy",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which uops are being delivered to Instruction Decode Queue (IDQ) while the Microcode Sequenser (MS) is busy. Counting includes uops that may bypass the IDQ. Uops maybe initiated by Decode Stream Buffer (DSB) or MITE",
},
{
	.name = "idq.ms_switches",
	.event = "umask=0x30,edge=1,period=2000003,cmask=1,event=0x79",
	.desc = "Number of switches from DSB (Decode Stream Buffer) or MITE (legacy decode pipeline) to the Microcode Sequencer",
	.topic = "frontend",
},
{
	.name = "idq.mite_all_uops",
	.event = "umask=0x3c,period=2000003,event=0x79",
	.desc = "Uops delivered to Instruction Decode Queue (IDQ) from MITE path",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops delivered to Instruction Decode Queue (IDQ) from the MITE path. Counting includes uops that may bypass the IDQ. This also means that uops are not being delivered from the Decode Stream Buffer (DSB)",
},
{
	.name = "icache.hit",
	.event = "umask=0x1,period=2000003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Reads. both cacheable and noncacheable, including UC fetches",
	.topic = "frontend",
	.long_desc = "This event counts the number of both cacheable and noncacheable Instruction Cache, Streaming Buffer and Victim Cache Reads including UC fetches",
},
{
	.name = "icache.misses",
	.event = "umask=0x2,period=200003,event=0x80",
	.desc = "Number of Instruction Cache, Streaming Buffer and Victim Cache Misses. Includes Uncacheable accesses",
	.topic = "frontend",
	.long_desc = "This event counts the number of instruction cache, streaming buffer and victim cache misses. Counting includes UC accesses",
},
{
	.name = "icache.ifdata_stall",
	.event = "umask=0x4,period=2000003,event=0x80",
	.desc = "Cycles where a code fetch is stalled due to L1 instruction-cache miss",
	.topic = "frontend",
	.long_desc = "This event counts cycles during which the demand fetch waits for data (wfdM104H) from L2 or iSB (opportunistic hit)",
},
{
	.name = "idq_uops_not_delivered.core",
	.event = "umask=0x1,period=2000003,event=0x9c",
	.desc = "Uops not delivered to Resource Allocation Table (RAT) per thread when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "This event counts the number of uops not delivered to Resource Allocation Table (RAT) per thread adding \u201c4 \u2013 x\u201d when Resource Allocation Table (RAT) is not stalled and Instruction Decode Queue (IDQ) delivers x uops to Resource Allocation Table (RAT) (where x belongs to {0,1,2,3}). Counting does not cover cases when:\n a. IDQ-Resource Allocation Table (RAT) pipe serves the other thread;\n b. Resource Allocation Table (RAT) is stalled for the thread (including uop drops and clear BE conditions); \n c. Instruction Decode Queue (IDQ) delivers four uops",
},
{
	.name = "idq_uops_not_delivered.cycles_0_uops_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=4,event=0x9c",
	.desc = "Cycles per thread when 4 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "This event counts, on the per-thread basis, cycles when no uops are delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core =4",
},
{
	.name = "idq_uops_not_delivered.cycles_le_1_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=3,event=0x9c",
	.desc = "Cycles per thread when 3 or more uops are not delivered to Resource Allocation Table (RAT) when backend of the machine is not stalled",
	.topic = "frontend",
	.long_desc = "This event counts, on the per-thread basis, cycles when less than 1 uop is  delivered to Resource Allocation Table (RAT). IDQ_Uops_Not_Delivered.core >=3",
},
{
	.name = "idq_uops_not_delivered.cycles_le_2_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=2,event=0x9c",
	.desc = "Cycles with less than 2 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_le_3_uop_deliv.core",
	.event = "umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Cycles with less than 3 uops delivered by the front end",
	.topic = "frontend",
},
{
	.name = "idq_uops_not_delivered.cycles_fe_was_ok",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0x9c",
	.desc = "Counts cycles FE delivered 4 uops or Resource Allocation Table (RAT) was stalling FE",
	.topic = "frontend",
},
{
	.name = "dsb2mite_switches.penalty_cycles",
	.event = "umask=0x2,period=2000003,event=0xab",
	.desc = "Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles",
	.topic = "frontend",
	.long_desc = "This event counts Decode Stream Buffer (DSB)-to-MITE switch true penalty cycles. These cycles do not include uops routed through because of the switch itself, for example, when Instruction Decode Queue (IDQ) pre-allocation is unavailable, or Instruction Decode Queue (IDQ) is full. SBD-to-MITE switch true penalty cycles happen after the merge mux (MM) receives Decode Stream Buffer (DSB) Sync-indication until receiving the first MITE uop. \nMM is placed before Instruction Decode Queue (IDQ) to merge uops being fed from the MITE and Decode Stream Buffer (DSB) paths. Decode Stream Buffer (DSB) inserts the Sync-indication whenever a Decode Stream Buffer (DSB)-to-MITE switch occurs.\nPenalty: A Decode Stream Buffer (DSB) hit followed by a Decode Stream Buffer (DSB) miss can cost up to six cycles in which no uops are delivered to the IDQ. Most often, such switches from the Decode Stream Buffer (DSB) to the legacy pipeline cost 0\u20132 cycles",
},
{
	.name = "misalign_mem_ref.loads",
	.event = "umask=0x1,period=2000003,event=0x5",
	.desc = "Speculative cache line split load uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "This event counts speculative cache-line split load uops dispatched to the L1 cache",
},
{
	.name = "misalign_mem_ref.stores",
	.event = "umask=0x2,period=2000003,event=0x5",
	.desc = "Speculative cache line split STA uops dispatched to L1 cache",
	.topic = "memory",
	.long_desc = "This event counts speculative cache line split store-address (STA) uops dispatched to the L1 cache",
},
{
	.name = "tx_mem.abort_conflict",
	.event = "umask=0x1,period=2000003,event=0x54",
	.desc = "Number of times a TSX line had a cache conflict",
	.topic = "memory",
	.long_desc = "Number of times a TSX line had a cache conflict",
},
{
	.name = "tx_mem.abort_capacity_write",
	.event = "umask=0x2,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to an evicted line caused by a transaction overflow",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to an evicted line caused by a transaction overflow",
},
{
	.name = "tx_mem.abort_hle_store_to_elided_lock",
	.event = "umask=0x4,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to a non-release/commit store to lock",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to a non-release/commit store to lock",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_not_empty",
	.event = "umask=0x8,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to commit but Lock Buffer not empty",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_mismatch",
	.event = "umask=0x10,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to release/commit but data and address mismatch",
},
{
	.name = "tx_mem.abort_hle_elision_buffer_unsupported_alignment",
	.event = "umask=0x20,period=2000003,event=0x54",
	.desc = "Number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
	.topic = "memory",
	.long_desc = "Number of times a TSX Abort was triggered due to attempting an unsupported alignment from Lock Buffer",
},
{
	.name = "tx_mem.hle_elision_buffer_full",
	.event = "umask=0x40,period=2000003,event=0x54",
	.desc = "Number of times we could not allocate Lock Buffer",
	.topic = "memory",
	.long_desc = "Number of times we could not allocate Lock Buffer",
},
{
	.name = "tx_exec.misc1",
	.event = "umask=0x1,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions that may cause a transactional abort was executed. Since this is the count of execution, it may not always cause a transactional abort",
	.topic = "memory",
},
{
	.name = "tx_exec.misc2",
	.event = "umask=0x2,period=2000003,event=0x5d",
	.desc = "Counts the number of times a class of instructions (e.g., vzeroupper) that may cause a transactional abort was executed inside a transactional region",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by  a vzeroupper instruction",
},
{
	.name = "tx_exec.misc3",
	.event = "umask=0x4,period=2000003,event=0x5d",
	.desc = "Counts the number of times an instruction execution caused the transactional nest count supported to be exceeded",
	.topic = "memory",
	.long_desc = "Unfriendly TSX abort triggered by a nest count that is too deep",
},
{
	.name = "tx_exec.misc4",
	.event = "umask=0x8,period=2000003,event=0x5d",
	.desc = "Counts the number of times a XBEGIN instruction was executed inside an HLE transactional region",
	.topic = "memory",
	.long_desc = "RTM region detected inside HLE",
},
{
	.name = "tx_exec.misc5",
	.event = "umask=0x10,period=2000003,event=0x5d",
	.desc = "Counts the number of times an HLE XACQUIRE instruction was executed inside an RTM transactional region",
	.topic = "memory",
},
{
	.name = "machine_clears.memory_ordering",
	.event = "umask=0x2,period=100003,event=0xc3",
	.desc = "Counts the number of machine clears due to memory order conflicts",
	.topic = "memory",
	.long_desc = "This event counts the number of memory ordering Machine Clears detected. Memory Ordering Machine Clears can result from one of the following:\n1. memory disambiguation,\n2. external snoop, or\n3. cross SMT-HW-thread snoop (stores) hitting load buffer",
},
{
	.name = "hle_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc8",
	.desc = "Number of times we entered an HLE region; does not count nested transactions",
	.topic = "memory",
	.long_desc = "Number of times we entered an HLE region\n does not count nested transactions",
},
{
	.name = "hle_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc8",
	.desc = "Number of times HLE commit succeeded",
	.topic = "memory",
	.long_desc = "Number of times HLE commit succeeded",
},
{
	.name = "hle_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc8",
	.desc = "Number of times HLE abort was triggered (PEBS) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times HLE abort was triggered (PEBS) (Precise event)",
},
{
	.name = "hle_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an HLE abort was attributed to a Memory condition (See TSX_Memory event for additional details)",
},
{
	.name = "hle_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to uncommon conditions",
	.topic = "memory",
	.long_desc = "Number of times the TSX watchdog signaled an HLE abort",
},
{
	.name = "hle_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Number of times a disallowed operation caused an HLE abort",
},
{
	.name = "hle_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times HLE caused a fault",
},
{
	.name = "hle_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc8",
	.desc = "Number of times an HLE execution aborted due to none of the previous 4 categories (e.g. interrupts)",
	.topic = "memory",
	.long_desc = "Number of times HLE aborted and was not due to the abort conditions in subevents 3-6",
},
{
	.name = "rtm_retired.start",
	.event = "umask=0x1,period=2000003,event=0xc9",
	.desc = "Number of times we entered an RTM region; does not count nested transactions",
	.topic = "memory",
	.long_desc = "Number of times we entered an RTM region\n does not count nested transactions",
},
{
	.name = "rtm_retired.commit",
	.event = "umask=0x2,period=2000003,event=0xc9",
	.desc = "Number of times RTM commit succeeded",
	.topic = "memory",
	.long_desc = "Number of times RTM commit succeeded",
},
{
	.name = "rtm_retired.aborted",
	.event = "umask=0x4,period=2000003,event=0xc9",
	.desc = "Number of times RTM abort was triggered (PEBS) (Precise event)",
	.topic = "memory",
	.long_desc = "Number of times RTM abort was triggered (PEBS) (Precise event)",
},
{
	.name = "rtm_retired.aborted_misc1",
	.event = "umask=0x8,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g. read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times an RTM abort was attributed to a Memory condition (See TSX_Memory event for additional details)",
},
{
	.name = "rtm_retired.aborted_misc2",
	.event = "umask=0x10,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to various memory events (e.g., read/write capacity and conflicts)",
	.topic = "memory",
	.long_desc = "Number of times the TSX watchdog signaled an RTM abort",
},
{
	.name = "rtm_retired.aborted_misc3",
	.event = "umask=0x20,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to HLE-unfriendly instructions",
	.topic = "memory",
	.long_desc = "Number of times a disallowed operation caused an RTM abort",
},
{
	.name = "rtm_retired.aborted_misc4",
	.event = "umask=0x40,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to incompatible memory type",
	.topic = "memory",
	.long_desc = "Number of times a RTM caused a fault",
},
{
	.name = "rtm_retired.aborted_misc5",
	.event = "umask=0x80,period=2000003,event=0xc9",
	.desc = "Number of times an RTM execution aborted due to none of the previous 4 categories (e.g. interrupt)",
	.topic = "memory",
	.long_desc = "Number of times RTM aborted and was not due to the abort conditions in subevents 3-6",
},
{
	.name = "mem_trans_retired.load_latency_gt_4",
	.event = "umask=0x1,period=100003,event=0xcd,ldlat=0x4",
	.desc = "Randomly selected loads with latency value being above 4  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above four  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_8",
	.event = "umask=0x1,period=50021,event=0xcd,ldlat=0x8",
	.desc = "Randomly selected loads with latency value being above 8  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above eight  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_16",
	.event = "umask=0x1,period=20011,event=0xcd,ldlat=0x10",
	.desc = "Randomly selected loads with latency value being above 16  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 16  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_32",
	.event = "umask=0x1,period=100007,event=0xcd,ldlat=0x20",
	.desc = "Randomly selected loads with latency value being above 32  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 32  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_64",
	.event = "umask=0x1,period=2003,event=0xcd,ldlat=0x40",
	.desc = "Randomly selected loads with latency value being above 64  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 64  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_128",
	.event = "umask=0x1,period=1009,event=0xcd,ldlat=0x80",
	.desc = "Randomly selected loads with latency value being above 128  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 128  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_256",
	.event = "umask=0x1,period=503,event=0xcd,ldlat=0x100",
	.desc = "Randomly selected loads with latency value being above 256  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 256  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "mem_trans_retired.load_latency_gt_512",
	.event = "umask=0x1,period=101,event=0xcd,ldlat=0x200",
	.desc = "Randomly selected loads with latency value being above 512  Spec update: BDM100, BDM35 (Must be precise)",
	.topic = "memory",
	.long_desc = "Counts randomly selected loads with latency value being above 512  Spec update: BDM100, BDM35 (Must be precise)",
},
{
	.name = "offcore_response.demand_data_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_data_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000001",
	.desc = "Counts demand data reads",
	.topic = "memory",
	.long_desc = "Counts demand data reads",
},
{
	.name = "offcore_response.demand_rfo.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_rfo.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000002",
	.desc = "Counts all demand data writes (RFOs)",
	.topic = "memory",
	.long_desc = "Counts all demand data writes (RFOs)",
},
{
	.name = "offcore_response.demand_code_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.demand_code_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000004",
	.desc = "Counts all demand code reads",
	.topic = "memory",
	.long_desc = "Counts all demand code reads",
},
{
	.name = "offcore_response.corewb.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.corewb.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000008",
	.desc = "Counts writebacks (modified to exclusive)",
	.topic = "memory",
	.long_desc = "Counts writebacks (modified to exclusive)",
},
{
	.name = "offcore_response.pf_l2_data_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_data_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000010",
	.desc = "Counts prefetch (that bring data to L2) data reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to L2) data reads",
},
{
	.name = "offcore_response.pf_l2_rfo.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_rfo.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000020",
	.desc = "Counts all prefetch (that bring data to L2) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to L2) RFOs",
},
{
	.name = "offcore_response.pf_l2_code_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l2_code_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000040",
	.desc = "Counts all prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_data_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000080",
	.desc = "Counts all prefetch (that bring data to LLC only) data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) data reads",
},
{
	.name = "offcore_response.pf_l3_rfo.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_rfo.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000100",
	.desc = "Counts all prefetch (that bring data to LLC only) RFOs",
	.topic = "memory",
	.long_desc = "Counts all prefetch (that bring data to LLC only) RFOs",
},
{
	.name = "offcore_response.pf_l3_code_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.pf_l3_code_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000200",
	.desc = "Counts prefetch (that bring data to LLC only) code reads",
	.topic = "memory",
	.long_desc = "Counts prefetch (that bring data to LLC only) code reads",
},
{
	.name = "offcore_response.other.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000028000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C8000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.other.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C008000",
	.desc = "Counts any other requests",
	.topic = "memory",
	.long_desc = "Counts any other requests",
},
{
	.name = "offcore_response.all_pf_data_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_data_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000090",
	.desc = "Counts all prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch data reads",
},
{
	.name = "offcore_response.all_pf_rfo.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_rfo.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000120",
	.desc = "Counts prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts prefetch RFOs",
},
{
	.name = "offcore_response.all_pf_code_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_pf_code_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000240",
	.desc = "Counts all prefetch code reads",
	.topic = "memory",
	.long_desc = "Counts all prefetch code reads",
},
{
	.name = "offcore_response.all_data_rd.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_data_rd.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000091",
	.desc = "Counts all demand & prefetch data reads",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch data reads",
},
{
	.name = "offcore_response.all_rfo.supplier_none.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2000020122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_hit.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x20003C0122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss_local_dram.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0084000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss_local_dram.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0104000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss_local_dram.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0204000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss_local_dram.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x0404000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss_local_dram.snoop_hitm",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x1004000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss_local_dram.snoop_non_dram",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x2004000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss_local_dram.any_snoop",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x3F84000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss.snoop_none",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x00BC000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss.snoop_not_needed",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x013C000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss.snoop_miss",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x023C000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "offcore_response.all_rfo.l3_miss.snoop_hit_no_fwd",
	.event = "umask=0x1,period=100003,event=0xb7,offcore_rsp=0x043C000122",
	.desc = "Counts all demand & prefetch RFOs",
	.topic = "memory",
	.long_desc = "Counts all demand & prefetch RFOs",
},
{
	.name = "cpl_cycles.ring0",
	.event = "umask=0x1,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when the thread is in ring 0",
	.topic = "other",
	.long_desc = "This event counts the unhalted core cycles during which the thread is in the ring 0 privileged mode",
},
{
	.name = "cpl_cycles.ring0_trans",
	.event = "umask=0x1,edge=1,period=100007,cmask=1,event=0x5c",
	.desc = "Number of intervals between processor halts while thread is in ring 0",
	.topic = "other",
	.long_desc = "This event counts when there is a transition from ring 1,2 or 3 to ring0",
},
{
	.name = "cpl_cycles.ring123",
	.event = "umask=0x2,period=2000003,event=0x5c",
	.desc = "Unhalted core cycles when thread is in rings 1, 2, or 3",
	.topic = "other",
	.long_desc = "This event counts unhalted core cycles during which the thread is in rings 1, 2, or 3",
},
{
	.name = "lock_cycles.split_lock_uc_lock_duration",
	.event = "umask=0x1,period=2000003,event=0x63",
	.desc = "Cycles when L1 and L2 are locked due to UC or split lock",
	.topic = "other",
	.long_desc = "This event counts cycles in which the L1 and L2 are locked due to a UC lock or split lock. A lock is asserted in case of locked memory access, due to noncacheable memory, locked operation that spans two cache lines, or a page walk from the noncacheable page table. L1D and L2 locks have a very high performance penalty and it is highly recommended to avoid such access",
},
{
	.name = "inst_retired.any",
	.event = "event=0xc0,period=2000003",
	.desc = "Instructions retired from execution",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions retired from execution. For instructions that consist of multiple micro-ops, this event counts the retirement of the last micro-op of the instruction. Counting continues during hardware interrupts, traps, and inside interrupt handlers. \nNotes: INST_RETIRED.ANY is counted by a designated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. INST_RETIRED.ANY_P is counted by a programmable counter and it is an architectural performance event. \nCounting: Faulting executions of GETSEC/VM entry/VM Exit/MWait will not count as retired instructions",
},
{
	.name = "cpu_clk_unhalted.thread",
	.event = "event=0x3c,period=2000003",
	.desc = "Core cycles when the thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of core cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. This event is a component in many key event ratios. The core frequency may change from time to time due to transitions associated with Enhanced Intel SpeedStep Technology or TM2. For this reason this event may have a changing ratio with regards to time. When the core frequency is constant, this event can approximate elapsed time while the core was not in the halt state. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events",
},
{
	.name = "cpu_clk_unhalted.thread_any",
	.event = "event=0x3c,any=1,period=2000003",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_tsc",
	.event = "umask=0x3,period=2000003,event=0",
	.desc = "Reference cycles when the core is not in halt state",
	.topic = "pipeline",
	.long_desc = "This event counts the number of reference cycles when the core is not in a halt state. The core enters the halt state when it is running the HLT instruction or the MWAIT instruction. This event is not affected by core frequency changes (for example, P states, TM2 transitions) but has the same incrementing frequency as the time stamp counter. This event can approximate elapsed time while the core was not in a halt state. This event has a constant ratio with the CPU_CLK_UNHALTED.REF_XCLK event. It is counted on a dedicated fixed counter, leaving the four (eight when Hyperthreading is disabled) programmable counters available for other events. \nNote: On all current platforms this event stops counting during 'throttling (TM)' states duty off periods the processor is 'halted'.  This event is clocked by base clock (100 Mhz) on Sandy Bridge. The counter update is done at a lower clock rate then the core clock the overflow status bit for this counter may appear 'sticky'.  After the counter has overflowed and software clears the overflow status bit and resets the counter to less than MAX. The reset value to the counter is not clocked immediately so the overflow status bit will flip 'high (1)' and generate another PMI (if enabled) after which the reset value gets clocked into the counter. Therefore, software will get the interrupt, read the overflow status bit '1 for bit 34 while the counter value is less than MAX. Software should ignore this case",
},
{
	.name = "ld_blocks.store_forward",
	.event = "umask=0x2,period=100003,event=0x3",
	.desc = "Cases when loads get true Block-on-Store blocking code preventing store forwarding",
	.topic = "pipeline",
	.long_desc = "This event counts how many times the load operation got the true Block-on-Store blocking code preventing store forwarding. This includes cases when:\n - preceding store conflicts with the load (incomplete overlap);\n - store forwarding is impossible due to u-arch limitations;\n - preceding lock RMW operations are not forwarded;\n - store has the no-forward bit set (uncacheable/page-split/masked stores);\n - all-blocking stores are used (mostly, fences and port I/O);\nand others.\nThe most common case is a load blocked due to its address range overlapping with a preceding smaller uncompleted store. Note: This event does not take into account cases of out-of-SW-control (for example, SbTailHit), unknown physical STA, and cases of blocking loads on store due to being non-WB memory type or a lock. These cases are covered by other events.\nSee the table of not supported store forwards in the Optimization Guide",
},
{
	.name = "ld_blocks.no_sr",
	.event = "umask=0x8,period=100003,event=0x3",
	.desc = "This event counts the number of times that split load operations are temporarily blocked because all resources for handling the split accesses are in use",
	.topic = "pipeline",
},
{
	.name = "ld_blocks_partial.address_alias",
	.event = "umask=0x1,period=100003,event=0x7",
	.desc = "False dependencies in MOB due to partial compare",
	.topic = "pipeline",
	.long_desc = "This event counts false dependencies in MOB when the partial comparison upon loose net check and dependency was resolved by the Enhanced Loose net mechanism. This may not result in high performance penalties. Loose net checks can fail when loads and stores are 4k aliased",
},
{
	.name = "int_misc.recovery_cycles",
	.event = "umask=0x3,period=2000003,cmask=1,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for this thread (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
	.long_desc = "Cycles checkpoints in Resource Allocation Table (RAT) are recovering from JEClear or machine clear",
},
{
	.name = "int_misc.recovery_cycles_any",
	.event = "umask=0x3,any=1,period=2000003,cmask=1,event=0xd",
	.desc = "Core cycles the allocator was stalled due to recovery from earlier clear event for any thread running on the physical core (e.g. misprediction or memory nuke)",
	.topic = "pipeline",
},
{
	.name = "int_misc.rat_stall_cycles",
	.event = "umask=0x8,period=2000003,event=0xd",
	.desc = "Cycles when Resource Allocation Table (RAT) external stall is sent to Instruction Decode Queue (IDQ) for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts the number of cycles during which Resource Allocation Table (RAT) external stall is sent to Instruction Decode Queue (IDQ) for the current thread. This also includes the cycles during which the Allocator is serving another thread",
},
{
	.name = "uops_issued.any",
	.event = "umask=0x1,period=2000003,event=0xe",
	.desc = "Uops that Resource Allocation Table (RAT) issues to Reservation Station (RS)",
	.topic = "pipeline",
	.long_desc = "This event counts the number of Uops issued by the Resource Allocation Table (RAT) to the reservation station (RS)",
},
{
	.name = "uops_issued.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xe",
	.desc = "Cycles when Resource Allocation Table (RAT) does not issue Uops to Reservation Station (RS) for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which the Resource Allocation Table (RAT) does not issue any Uops to the reservation station (RS) for the current thread",
},
{
	.name = "uops_issued.flags_merge",
	.event = "umask=0x10,period=2000003,event=0xe",
	.desc = "Number of flags-merge uops being allocated. Such uops considered perf sensitive; added by GSR u-arch",
	.topic = "pipeline",
	.long_desc = "Number of flags-merge uops being allocated. Such uops considered perf sensitive\n added by GSR u-arch",
},
{
	.name = "uops_issued.slow_lea",
	.event = "umask=0x20,period=2000003,event=0xe",
	.desc = "Number of slow LEA uops being allocated. A uop is generally considered SlowLea if it has 3 sources (e.g. 2 sources + immediate) regardless if as a result of LEA instruction or not",
	.topic = "pipeline",
},
{
	.name = "uops_issued.single_mul",
	.event = "umask=0x40,period=2000003,event=0xe",
	.desc = "Number of Multiply packed/scalar single precision uops allocated",
	.topic = "pipeline",
},
{
	.name = "arith.fpu_div_active",
	.event = "umask=0x1,period=2000003,event=0x14",
	.desc = "Cycles when divider is busy executing divide operations",
	.topic = "pipeline",
	.long_desc = "This event counts the number of the divide operations executed. Uses edge-detect and a cmask value of 1 on ARITH.FPU_DIV_ACTIVE to get the number of the divide operations executed",
},
{
	.name = "cpu_clk_unhalted.thread_p",
	.event = "umask=0x0,period=2000003,event=0x3c",
	.desc = "Thread cycles when thread is not in halt state",
	.topic = "pipeline",
	.long_desc = "This is an architectural event that counts the number of thread cycles while the thread is not in a halt state. The thread enters the halt state when it is running the HLT instruction. The core frequency may change from time to time due to power or thermal throttling. For this reason, this event may have a changing ratio with regards to wall clock time",
},
{
	.name = "cpu_clk_unhalted.thread_p_any",
	.event = "umask=0x0,any=1,period=2000003,event=0x3c",
	.desc = "Core cycles when at least one thread on the physical core is not in halt state",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "This is a fixed-frequency event programmed to general counters. It counts when the core is unhalted at 100 Mhz",
},
{
	.name = "cpu_clk_thread_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.ref_xclk",
	.event = "umask=0x1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
	.long_desc = "Reference cycles when the thread is unhalted (counts at 100 MHz rate)",
},
{
	.name = "cpu_clk_unhalted.ref_xclk_any",
	.event = "umask=0x1,any=1,period=2000003,event=0x3c",
	.desc = "Reference cycles when the at least one thread on the physical core is unhalted (counts at 100 MHz rate)",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_thread_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "cpu_clk_unhalted.one_thread_active",
	.event = "umask=0x2,period=2000003,event=0x3c",
	.desc = "Count XClk pulses when this thread is unhalted and the other thread is halted",
	.topic = "pipeline",
},
{
	.name = "load_hit_pre.sw_pf",
	.event = "umask=0x1,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for software prefetch",
	.topic = "pipeline",
	.long_desc = "This event counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the software prefetch. It can also be incremented by some lock instructions. So it should only be used with profiling so that the locks can be excluded by asm inspection of the nearby instructions",
},
{
	.name = "load_hit_pre.hw_pf",
	.event = "umask=0x2,period=100003,event=0x4c",
	.desc = "Not software-prefetch load dispatches that hit FB allocated for hardware prefetch",
	.topic = "pipeline",
	.long_desc = "This event counts all not software-prefetch load dispatches that hit the fill buffer (FB) allocated for the hardware prefetch",
},
{
	.name = "move_elimination.int_eliminated",
	.event = "umask=0x1,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_eliminated",
	.event = "umask=0x2,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.int_not_eliminated",
	.event = "umask=0x4,period=1000003,event=0x58",
	.desc = "Number of integer Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "move_elimination.simd_not_eliminated",
	.event = "umask=0x8,period=1000003,event=0x58",
	.desc = "Number of SIMD Move Elimination candidate uops that were not eliminated",
	.topic = "pipeline",
},
{
	.name = "rs_events.empty_cycles",
	.event = "umask=0x1,period=2000003,event=0x5e",
	.desc = "Cycles when Reservation Station (RS) is empty for the thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which the reservation station (RS) is empty for the thread.\nNote: In ST-mode, not active thread should drive 0. This is usually caused by severely costly branch mispredictions, or allocator/FE issues",
},
{
	.name = "rs_events.empty_end",
	.event = "inv=1,umask=0x1,edge=1,period=200003,cmask=1,event=0x5e",
	.desc = "Counts end of periods where the Reservation Station (RS) was empty. Could be useful to precisely locate Frontend Latency Bound issues",
	.topic = "pipeline",
},
{
	.name = "ild_stall.lcp",
	.event = "umask=0x1,period=2000003,event=0x87",
	.desc = "Stalls caused by changing prefix length of the instruction",
	.topic = "pipeline",
	.long_desc = "This event counts stalls occured due to changing prefix length (66, 67 or REX.W when they change the length of the decoded instruction). Occurrences counting is proportional to the number of prefixes in a 16B-line. This may result in the following penalties: three-cycle penalty for each LCP in a 16-byte chunk",
},
{
	.name = "br_inst_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x88",
	.desc = "Not taken macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts not taken macro-conditional branch instructions",
},
{
	.name = "br_inst_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired macro-conditional branch instructions",
},
{
	.name = "br_inst_exec.taken_direct_jump",
	.event = "umask=0x82,period=200003,event=0x88",
	.desc = "Taken speculative and retired macro-conditional branch instructions excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired macro-conditional branch instructions excluding calls and indirect branches",
},
{
	.name = "br_inst_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired indirect branches excluding calls and return branches",
},
{
	.name = "br_inst_exec.taken_indirect_near_return",
	.event = "umask=0x88,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired indirect branches that have a return mnemonic",
},
{
	.name = "br_inst_exec.taken_direct_near_call",
	.event = "umask=0x90,period=200003,event=0x88",
	.desc = "Taken speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x88",
	.desc = "Taken speculative and retired indirect calls",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired indirect calls including both register and memory indirect",
},
{
	.name = "br_inst_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x88",
	.desc = "Speculative and retired macro-conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired macro-conditional branch instructions",
},
{
	.name = "br_inst_exec.all_direct_jmp",
	.event = "umask=0xc2,period=200003,event=0x88",
	.desc = "Speculative and retired macro-unconditional branches excluding calls and indirects",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired macro-unconditional branch instructions, excluding calls and indirects",
},
{
	.name = "br_inst_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x88",
	.desc = "Speculative and retired indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired indirect branches excluding calls and return branches",
},
{
	.name = "br_inst_exec.all_indirect_near_return",
	.event = "umask=0xc8,period=200003,event=0x88",
	.desc = "Speculative and retired indirect return branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired indirect branches that have a return mnemonic",
},
{
	.name = "br_inst_exec.all_direct_near_call",
	.event = "umask=0xd0,period=200003,event=0x88",
	.desc = "Speculative and retired direct near calls",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired direct near calls",
},
{
	.name = "br_inst_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x88",
	.desc = "Speculative and retired  branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired branch instructions",
},
{
	.name = "br_misp_exec.nontaken_conditional",
	.event = "umask=0x41,period=200003,event=0x89",
	.desc = "Not taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts not taken speculative and retired mispredicted macro conditional branch instructions",
},
{
	.name = "br_misp_exec.taken_conditional",
	.event = "umask=0x81,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired mispredicted macro conditional branch instructions",
},
{
	.name = "br_misp_exec.taken_indirect_jump_non_call_ret",
	.event = "umask=0x84,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.taken_return_near",
	.event = "umask=0x88,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect branches with return mnemonic",
	.topic = "pipeline",
	.long_desc = "This event counts taken speculative and retired mispredicted indirect branches that have a return mnemonic",
},
{
	.name = "br_misp_exec.taken_indirect_near_call",
	.event = "umask=0xa0,period=200003,event=0x89",
	.desc = "Taken speculative and retired mispredicted indirect calls",
	.topic = "pipeline",
},
{
	.name = "br_misp_exec.all_conditional",
	.event = "umask=0xc1,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired mispredicted macro conditional branch instructions",
},
{
	.name = "br_misp_exec.all_indirect_jump_non_call_ret",
	.event = "umask=0xc4,period=200003,event=0x89",
	.desc = "Mispredicted indirect branches excluding calls and returns",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken mispredicted indirect branches excluding calls and returns",
},
{
	.name = "br_misp_exec.all_branches",
	.event = "umask=0xff,period=200003,event=0x89",
	.desc = "Speculative and retired mispredicted macro conditional branches",
	.topic = "pipeline",
	.long_desc = "This event counts both taken and not taken speculative and retired mispredicted branch instructions",
},
{
	.name = "uop_dispatches_cancelled.simd_prf",
	.event = "umask=0x3,period=2000003,event=0xa0",
	.desc = "Micro-op dispatches cancelled due to insufficient SIMD physical register file read ports",
	.topic = "pipeline",
	.long_desc = "This event counts the number of micro-operations cancelled after they were dispatched from the scheduler to the execution units when the total number of physical register read ports across all dispatch ports exceeds the read bandwidth of the physical register file.  The SIMD_PRF subevent applies to the following instructions: VDPPS, DPPS, VPCMPESTRI, PCMPESTRI, VPCMPESTRM, PCMPESTRM, VFMADD*, VFMADDSUB*, VFMSUB*, VMSUBADD*, VFNMADD*, VFNMSUB*.  See the Broadwell Optimization Guide for more information",
},
{
	.name = "uops_dispatched_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_executed_port.port_0_core",
	.event = "umask=0x1,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 0",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_0",
	.event = "umask=0x1,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 0",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 0",
},
{
	.name = "uops_dispatched_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_executed_port.port_1_core",
	.event = "umask=0x2,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 1",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_1",
	.event = "umask=0x2,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 1",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 1",
},
{
	.name = "uops_dispatched_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 2",
},
{
	.name = "uops_executed_port.port_2_core",
	.event = "umask=0x4,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 2",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_2",
	.event = "umask=0x4,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 2",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 2",
},
{
	.name = "uops_dispatched_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 3",
},
{
	.name = "uops_executed_port.port_3_core",
	.event = "umask=0x8,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 3",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_3",
	.event = "umask=0x8,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 3",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 3",
},
{
	.name = "uops_dispatched_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 4",
},
{
	.name = "uops_executed_port.port_4_core",
	.event = "umask=0x10,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 4",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_4",
	.event = "umask=0x10,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 4",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 4",
},
{
	.name = "uops_dispatched_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_executed_port.port_5_core",
	.event = "umask=0x20,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 5",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_5",
	.event = "umask=0x20,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 5",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 5",
},
{
	.name = "uops_dispatched_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_executed_port.port_6_core",
	.event = "umask=0x40,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are exectuted in port 6",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_6",
	.event = "umask=0x40,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 6",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 6",
},
{
	.name = "uops_dispatched_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 7",
},
{
	.name = "uops_executed_port.port_7_core",
	.event = "umask=0x80,any=1,period=2000003,event=0xa1",
	.desc = "Cycles per core when uops are dispatched to port 7",
	.topic = "pipeline",
},
{
	.name = "uops_executed_port.port_7",
	.event = "umask=0x80,period=2000003,event=0xa1",
	.desc = "Cycles per thread when uops are executed in port 7",
	.topic = "pipeline",
	.long_desc = "This event counts, on the per-thread basis, cycles during which uops are dispatched from the Reservation Station (RS) to port 7",
},
{
	.name = "resource_stalls.any",
	.event = "umask=0x1,period=2000003,event=0xa2",
	.desc = "Resource-related stall cycles",
	.topic = "pipeline",
	.long_desc = "This event counts resource-related stall cycles",
},
{
	.name = "resource_stalls.rs",
	.event = "umask=0x4,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no eligible RS entry available",
	.topic = "pipeline",
	.long_desc = "This event counts stall cycles caused by absence of eligible entries in the reservation station (RS). This may result from RS overflow, or from RS deallocation because of the RS array Write Port allocation scheme (each RS entry has two write ports instead of four. As a result, empty entries could not be used, although RS is not really full). This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "resource_stalls.sb",
	.event = "umask=0x8,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to no store buffers available. (not including draining form sync)",
	.topic = "pipeline",
	.long_desc = "This event counts stall cycles caused by the store buffer (SB) overflow (excluding draining from synch). This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "resource_stalls.rob",
	.event = "umask=0x10,period=2000003,event=0xa2",
	.desc = "Cycles stalled due to re-order buffer full",
	.topic = "pipeline",
	.long_desc = "This event counts ROB full stall cycles. This counts cycles that the pipeline backend blocked uop delivery from the front end",
},
{
	.name = "cycle_activity.cycles_l2_pending",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles the CPU has at least one pending  demand* load request missing the L2 cache",
},
{
	.name = "cycle_activity.cycles_l2_miss",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa3",
	.desc = "Cycles while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_ldm_pending",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles the CPU has at least one pending  demand load request (that is cycles with non-completed load waiting for its data from memory subsystem)",
},
{
	.name = "cycle_activity.cycles_mem_any",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xa3",
	.desc = "Cycles while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_no_execute",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "This event increments by 1 for every cycle where there was no execute for this thread",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port",
},
{
	.name = "cycle_activity.stalls_total",
	.event = "umask=0x4,period=2000003,cmask=4,event=0xa3",
	.desc = "Total execution stalls",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l2_pending",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port, while there was at least one pending demand* load request missing the L2 cache.(as a footprint) * includes also L1 HW prefetch requests that may or may not be required by demands",
},
{
	.name = "cycle_activity.stalls_l2_miss",
	.event = "umask=0x5,period=2000003,cmask=5,event=0xa3",
	.desc = "Execution stalls while L2 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_ldm_pending",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port, while there was at least one pending demand load request",
},
{
	.name = "cycle_activity.stalls_mem_any",
	.event = "umask=0x6,period=2000003,cmask=6,event=0xa3",
	.desc = "Execution stalls while memory subsystem has an outstanding load",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.cycles_l1d_pending",
	.event = "umask=0x8,period=2000003,cmask=8,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles the CPU has at least one pending  demand load request missing the L1 data cache",
},
{
	.name = "cycle_activity.cycles_l1d_miss",
	.event = "umask=0x8,period=2000003,cmask=8,event=0xa3",
	.desc = "Cycles while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "cycle_activity.stalls_l1d_pending",
	.event = "umask=0xc,period=2000003,cmask=12,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
	.long_desc = "Counts number of cycles nothing is executed on any execution port, while there was at least one pending demand load request missing the L1 data cache",
},
{
	.name = "cycle_activity.stalls_l1d_miss",
	.event = "umask=0xc,period=2000003,cmask=12,event=0xa3",
	.desc = "Execution stalls while L1 cache miss demand load is outstanding",
	.topic = "pipeline",
},
{
	.name = "lsd.uops",
	.event = "umask=0x1,period=2000003,event=0xa8",
	.desc = "Number of Uops delivered by the LSD",
	.topic = "pipeline",
	.long_desc = "Number of Uops delivered by the LSD",
},
{
	.name = "lsd.cycles_4_uops",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xa8",
	.desc = "Cycles 4 Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "lsd.cycles_active",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xa8",
	.desc = "Cycles Uops delivered by the LSD, but didn't come from the decoder",
	.topic = "pipeline",
},
{
	.name = "uops_executed.thread",
	.event = "umask=0x1,period=2000003,event=0xb1",
	.desc = "Counts the number of uops to be executed per-thread each cycle",
	.topic = "pipeline",
	.long_desc = "Number of uops to be executed per-thread each cycle",
},
{
	.name = "uops_executed.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Counts number of cycles no uops were dispatched to be executed on this thread",
	.topic = "pipeline",
	.long_desc = "This event counts cycles during which no uops were dispatched from the Reservation Station (RS) per thread",
},
{
	.name = "uops_executed.cycles_ge_1_uop_exec",
	.event = "umask=0x1,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles where at least 1 uop was executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_2_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles where at least 2 uops were executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_3_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles where at least 3 uops were executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.cycles_ge_4_uops_exec",
	.event = "umask=0x1,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles where at least 4 uops were executed per-thread",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core",
	.event = "umask=0x2,period=2000003,event=0xb1",
	.desc = "Number of uops executed on the core",
	.topic = "pipeline",
	.long_desc = "Number of uops executed from any thread",
},
{
	.name = "uops_executed.core_cycles_ge_1",
	.event = "umask=0x2,period=2000003,cmask=1,event=0xb1",
	.desc = "Cycles at least 1 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_2",
	.event = "umask=0x2,period=2000003,cmask=2,event=0xb1",
	.desc = "Cycles at least 2 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_3",
	.event = "umask=0x2,period=2000003,cmask=3,event=0xb1",
	.desc = "Cycles at least 3 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_ge_4",
	.event = "umask=0x2,period=2000003,cmask=4,event=0xb1",
	.desc = "Cycles at least 4 micro-op is executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "uops_executed.core_cycles_none",
	.event = "inv=1,umask=0x2,period=2000003,event=0xb1",
	.desc = "Cycles with no micro-ops executed from any thread on physical core",
	.topic = "pipeline",
},
{
	.name = "inst_retired.any_p",
	.event = "event=0xc0,period=2000003",
	.desc = "Number of instructions retired. General Counter   - architectural event  Spec update: BDM61",
	.topic = "pipeline",
	.long_desc = "This event counts the number of instructions (EOMs) retired. Counting covers macro-fused instructions individually (that is, increments by two)  Spec update: BDM61",
},
{
	.name = "inst_retired.prec_dist",
	.event = "umask=0x1,period=2000003,event=0xc0",
	.desc = "Precise instruction retired event with HW to reduce effect of PEBS shadow in IP distribution  Spec update: BDM11, BDM55 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts instructions retired  Spec update: BDM11, BDM55 (Must be precise)",
},
{
	.name = "inst_retired.x87",
	.event = "umask=0x2,period=2000003,event=0xc0",
	.desc = "FP operations  retired. X87 FP operations that have no exceptions: (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts FP operations retired. For X87 FP operations that have no exceptions counting also includes flows that have several X87, or flows that use X87 uops in the exception handling (Precise event)",
},
{
	.name = "other_assists.any_wb_assist",
	.event = "umask=0x40,period=100003,event=0xc1",
	.desc = "(null)",
	.topic = "pipeline",
},
{
	.name = "uops_retired.all",
	.event = "umask=0x1,period=2000003,event=0xc2",
	.desc = "Actually retired uops. (Precise Event - PEBS)  Supports address when precise (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts all actually retired uops. Counting increments by two for micro-fused uops, and by one for macro-fused and other uops. Maximal increment value for one cycle is eight  Supports address when precise (Precise event)",
},
{
	.name = "uops_retired.stall_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=1,event=0xc2",
	.desc = "Cycles no executable uops retired (Precise Event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts cycles without actually retired uops",
},
{
	.name = "uops_retired.total_cycles",
	.event = "inv=1,umask=0x1,period=2000003,cmask=10,event=0xc2",
	.desc = "Number of cycles using always true condition applied to  PEBS uops retired event (Precise event)",
	.topic = "pipeline",
	.long_desc = "Number of cycles using always true condition (uops_ret < 16) applied to  PEBS uops retired event (Precise event)",
},
{
	.name = "uops_retired.retire_slots",
	.event = "umask=0x2,period=2000003,event=0xc2",
	.desc = "Retirement slots used. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts the number of retirement slots used (Precise event)",
},
{
	.name = "machine_clears.cycles",
	.event = "umask=0x1,period=2000003,event=0xc3",
	.desc = "Cycles there was a Nuke. Account for both thread-specific and All Thread Nukes",
	.topic = "pipeline",
	.long_desc = "This event counts both thread-specific (TS) and all-thread (AT) nukes",
},
{
	.name = "machine_clears.count",
	.event = "umask=0x1,edge=1,period=100003,cmask=1,event=0xc3",
	.desc = "Number of machine clears (nukes) of any type",
	.topic = "pipeline",
},
{
	.name = "machine_clears.smc",
	.event = "umask=0x4,period=100003,event=0xc3",
	.desc = "Self-modifying code (SMC) detected",
	.topic = "pipeline",
	.long_desc = "This event counts self-modifying code (SMC) detected, which causes a machine clear",
},
{
	.name = "machine_clears.maskmov",
	.event = "umask=0x20,period=100003,event=0xc3",
	.desc = "This event counts the number of executed Intel AVX masked load operations that refer to an illegal address range with the mask bits set to 0",
	.topic = "pipeline",
	.long_desc = "Maskmov false fault - counts number of time ucode passes through Maskmov flow due to instruction's mask being 0 while the flow was completed without raising a fault",
},
{
	.name = "br_inst_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts all (macro) branch instructions retired",
},
{
	.name = "br_inst_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc4",
	.desc = "Conditional branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts conditional branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_call",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect near call instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts both direct and indirect near call instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.near_call_r3",
	.event = "umask=0x2,period=100007,event=0xc4",
	.desc = "Direct and indirect macro near call instructions retired (captured in ring 3). (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts both direct and indirect macro near call instructions retired (captured in ring 3) (Precise event)",
},
{
	.name = "br_inst_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc4",
	.desc = "All (macro) branch instructions retired. (Precise Event - PEBS)  Spec update: BDW98 (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_INST_RETIRED.ALL_BRANCHES that counts all (macro) branch instructions retired  Spec update: BDW98 (Must be precise)",
},
{
	.name = "br_inst_retired.near_return",
	.event = "umask=0x8,period=100007,event=0xc4",
	.desc = "Return instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts return instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.not_taken",
	.event = "umask=0x10,period=400009,event=0xc4",
	.desc = "Counts all not taken macro branch instructions retired. (Precise Event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts not taken branch instructions retired",
},
{
	.name = "br_inst_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc4",
	.desc = "Taken branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts taken branch instructions retired (Precise event)",
},
{
	.name = "br_inst_retired.far_branch",
	.event = "umask=0x40,period=100007,event=0xc4",
	.desc = "Counts the number of far branch instructions retired.(Precise Event)  Spec update: BDW98",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts far branch instructions retired  Spec update: BDW98",
},
{
	.name = "br_misp_retired.all_branches",
	.event = "umask=0x0,period=400009,event=0xc5",
	.desc = "All mispredicted macro branch instructions retired",
	.topic = "pipeline",
	.long_desc = "This event counts all mispredicted macro branch instructions retired",
},
{
	.name = "br_misp_retired.conditional",
	.event = "umask=0x1,period=400009,event=0xc5",
	.desc = "Mispredicted conditional branch instructions retired. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts mispredicted conditional branch instructions retired (Precise event)",
},
{
	.name = "br_misp_retired.all_branches_pebs",
	.event = "umask=0x4,period=400009,event=0xc5",
	.desc = "Mispredicted macro branch instructions retired. (Precise Event - PEBS) (Must be precise)",
	.topic = "pipeline",
	.long_desc = "This is a precise version of BR_MISP_RETIRED.ALL_BRANCHES that counts all mispredicted macro branch instructions retired (Must be precise)",
},
{
	.name = "br_misp_retired.ret",
	.event = "umask=0x8,period=100007,event=0xc5",
	.desc = "This event counts the number of mispredicted ret instructions retired.(Precise Event)",
	.topic = "pipeline",
	.long_desc = "This is a precise version (that is, uses PEBS) of the event that counts mispredicted return instructions retired",
},
{
	.name = "br_misp_retired.near_taken",
	.event = "umask=0x20,period=400009,event=0xc5",
	.desc = "number of near branch instructions retired that were mispredicted and taken. (Precise Event - PEBS) (Precise event)",
	.topic = "pipeline",
	.long_desc = "Number of near branch instructions retired that were mispredicted and taken. (Precise Event - PEBS) (Precise event)",
},
{
	.name = "rob_misc_events.lbr_inserts",
	.event = "umask=0x20,period=2000003,event=0xcc",
	.desc = "Count cases of saving new LBR",
	.topic = "pipeline",
	.long_desc = "This event counts cases of saving new LBR records by hardware. This assumes proper enabling of LBRs and takes into account LBR filtering done by the LBR_SELECT register",
},
{
	.name = "baclears.any",
	.event = "umask=0x1f,period=100003,event=0xe6",
	.desc = "Counts the total number when the front end is resteered, mainly when the BPU cannot provide a correct prediction and this is corrected by other branch handling mechanisms at the front end",
	.topic = "pipeline",
},
{
	.name = "unc_cbo_xsnp_response.miss_xcore",
	.event = "umask=0x41,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop initiated by this Cbox due to processor core memory request which misses in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop initiated by this Cbox due to processor core memory request which misses in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.miss_eviction",
	.event = "umask=0x81,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop resulted from L3 Eviction which misses in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop resulted from L3 Eviction which misses in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hit_xcore",
	.event = "umask=0x44,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop initiated by this Cbox due to processor core memory request which hits a non-modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop initiated by this Cbox due to processor core memory request which hits a non-modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_xsnp_response.hitm_xcore",
	.event = "umask=0x48,event=0x22",
	.desc = "Unit: uncore_cbox A cross-core snoop initiated by this Cbox due to processor core memory request which hits a modified line in some processor core",
	.topic = "uncore",
	.long_desc = "A cross-core snoop initiated by this Cbox due to processor core memory request which hits a modified line in some processor core",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_m",
	.event = "umask=0x11,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_m",
	.event = "umask=0x21,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_m",
	.event = "umask=0x81,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in M-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in M-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_i",
	.event = "umask=0x18,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_i",
	.event = "umask=0x88,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in I-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in I-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_mesi",
	.event = "umask=0x1f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in any MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in any MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_mesi",
	.event = "umask=0x2f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_mesi",
	.event = "umask=0x8f,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in MESI-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in MESI-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.any_es",
	.event = "umask=0x86,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup any request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup any request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.read_es",
	.event = "umask=0x16,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup read request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup read request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_cbo_cache_lookup.write_es",
	.event = "umask=0x26,event=0x34",
	.desc = "Unit: uncore_cbox L3 Lookup write request that access cache and found line in E or S-state",
	.topic = "uncore",
	.long_desc = "L3 Lookup write request that access cache and found line in E or S-state",
	.pmu = "uncore_cbox",
},
{
	.name = "unc_arb_trk_occupancy.all",
	.event = "umask=0x01,event=0x80",
	.desc = "Unit: uncore_arb Each cycle count number of all Core outgoing valid entries. Such entry is defined as valid from it's allocation till first of IDI0 or DRS0 messages is sent out. Accounts for Coherent and non-coherent traffic",
	.topic = "uncore",
	.long_desc = "Each cycle count number of all Core outgoing valid entries. Such entry is defined as valid from it's allocation till first of IDI0 or DRS0 messages is sent out. Accounts for Coherent and non-coherent traffic",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_occupancy.drd_direct",
	.event = "umask=0x02,event=0x80",
	.desc = "Unit: uncore_arb Each cycle count number of 'valid' coherent Data Read entries that are in DirectData mode. Such entry is defined as valid when it is allocated till data sent to Core (first chunk, IDI0). Applicable for IA Cores' requests in normal case",
	.topic = "uncore",
	.long_desc = "Each cycle count number of 'valid' coherent Data Read entries that are in DirectData mode. Such entry is defined as valid when it is allocated till data sent to Core (first chunk, IDI0). Applicable for IA Cores' requests in normal case",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.all",
	.event = "umask=0x01,event=0x81",
	.desc = "Unit: uncore_arb Total number of Core outgoing entries allocated. Accounts for Coherent and non-coherent traffic",
	.topic = "uncore",
	.long_desc = "Total number of Core outgoing entries allocated. Accounts for Coherent and non-coherent traffic",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.drd_direct",
	.event = "umask=0x02,event=0x81",
	.desc = "Unit: uncore_arb Number of Core coherent Data Read entries allocated in DirectData mode",
	.topic = "uncore",
	.long_desc = "Number of Core coherent Data Read entries allocated in DirectData mode",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_requests.writes",
	.event = "umask=0x20,event=0x81",
	.desc = "Unit: uncore_arb Number of Writes allocated - any write transactions: full/partials writes and evictions",
	.topic = "uncore",
	.long_desc = "Number of Writes allocated - any write transactions: full/partials writes and evictions",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_coh_trk_requests.all",
	.event = "umask=0x01,event=0x84",
	.desc = "Unit: uncore_arb Number of entries allocated. Account for Any type: e.g. Snoop, Core aperture, etc",
	.topic = "uncore",
	.long_desc = "Number of entries allocated. Account for Any type: e.g. Snoop, Core aperture, etc",
	.pmu = "uncore_arb",
},
{
	.name = "unc_arb_trk_occupancy.cycles_with_any_request",
	.event = "umask=0x01,cmask=1,event=0x80",
	.desc = "Unit: uncore_arb Cycles with at least one request outstanding is waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC.;",
	.topic = "uncore",
	.long_desc = "Cycles with at least one request outstanding is waiting for data return from memory controller. Account for coherent and non-coherent requests initiated by IA Cores, Processor Graphics Unit, or LLC",
	.pmu = "uncore_arb",
},
{
	.name = "unc_clock.socket",
	.event = "umask=0x01,event=0",
	.desc = "Unit: uncore_ncu This 48-bit fixed counter counts the UCLK cycles",
	.topic = "uncore",
	.long_desc = "This 48-bit fixed counter counts the UCLK cycles",
	.pmu = "uncore_ncu",
},
{
	.name = "dtlb_load_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x8",
	.desc = "Load misses in all DTLB levels that cause page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause page walks of any page size (4K/2M/4M/1G)  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed_4k",
	.event = "umask=0x2,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (4K)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause a completed page walk (4K page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=2000003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes (2M/4M)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed_1g",
	.event = "umask=0x8,period=2000003,event=0x8",
	.desc = "Load miss in all TLB levels causes a page walk that completes. (1G)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts load misses in all DTLB levels that cause a completed page walk (1G  page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x8",
	.desc = "Demand load Miss in all translation lookaside buffer (TLB) levels causes a page walk that completes of any page size  Spec update: BDM69",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.walk_duration",
	.event = "umask=0x10,period=2000003,event=0x8",
	.desc = "Cycles when PMH is busy with page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of cycles while PMH is busy with the page walk  Spec update: BDM69",
},
{
	.name = "dtlb_load_misses.stlb_hit_4k",
	.event = "umask=0x20,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit_2m",
	.event = "umask=0x40,period=2000003,event=0x8",
	.desc = "Load misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_load_misses.stlb_hit",
	.event = "umask=0x60,period=2000003,event=0x8",
	.desc = "Load operations that miss the first DTLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause page walks of any page size (4K/2M/4M/1G)  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x49",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (4K)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (4K page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks (2M/4M)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks (1G)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (1G  page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x49",
	.desc = "Store misses in all DTLB levels that cause completed page walks  Spec update: BDM69",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x49",
	.desc = "Cycles when PMH is busy with page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of cycles while PMH is busy with the page walk  Spec update: BDM69",
},
{
	.name = "dtlb_store_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x49",
	.desc = "Store misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
},
{
	.name = "dtlb_store_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x49",
	.desc = "Store operations that miss the first TLB level but hit the second and do not cause page walks",
	.topic = "virtual memory",
},
{
	.name = "ept.walk_cycles",
	.event = "umask=0x10,period=2000003,event=0x4f",
	.desc = "Cycle count for an Extended Page table walk",
	.topic = "virtual memory",
	.long_desc = "This event counts cycles for an extended page table walk. The Extended Page directory cache differs from standard TLB caches by the operating system that use it. Virtual machine operating systems use the extended page directory cache, while guest operating systems use the standard TLB caches",
},
{
	.name = "itlb_misses.miss_causes_a_walk",
	.event = "umask=0x1,period=100003,event=0x85",
	.desc = "Misses at all ITLB levels that cause page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause page walks of any page size (4K/2M/4M/1G)  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed_4k",
	.event = "umask=0x2,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (4K)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (4K page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed_2m_4m",
	.event = "umask=0x4,period=100003,event=0x85",
	.desc = "Code miss in all TLB levels causes a page walk that completes. (2M/4M)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (2M and 4M page sizes). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed_1g",
	.event = "umask=0x8,period=100003,event=0x85",
	.desc = "Store miss in all TLB levels causes a page walk that completes. (1G)  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts store misses in all DTLB levels that cause a completed page walk (1G  page size). The page walk can end with or without a fault  Spec update: BDM69",
},
{
	.name = "itlb_misses.walk_completed",
	.event = "umask=0xe,period=100003,event=0x85",
	.desc = "Misses in all ITLB levels that cause completed page walks  Spec update: BDM69",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.walk_duration",
	.event = "umask=0x10,period=100003,event=0x85",
	.desc = "Cycles when PMH is busy with page walks  Spec update: BDM69",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of cycles while PMH is busy with the page walk  Spec update: BDM69",
},
{
	.name = "itlb_misses.stlb_hit_4k",
	.event = "umask=0x20,period=100003,event=0x85",
	.desc = "Core misses that miss the  DTLB and hit the STLB (4K)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.stlb_hit_2m",
	.event = "umask=0x40,period=100003,event=0x85",
	.desc = "Code misses that miss the  DTLB and hit the STLB (2M)",
	.topic = "virtual memory",
},
{
	.name = "itlb_misses.stlb_hit",
	.event = "umask=0x60,period=100003,event=0x85",
	.desc = "Operations that miss the first ITLB level but hit the second and do not cause any page walks",
	.topic = "virtual memory",
},
{
	.name = "itlb.itlb_flush",
	.event = "umask=0x1,period=100007,event=0xae",
	.desc = "Flushing of the Instruction TLB (ITLB) pages, includes 4k/2M/4M pages",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of flushes of the big or small ITLB pages. Counting include both TLB Flush (covering all sets) and TLB Set Clear (set-specific)",
},
{
	.name = "page_walker_loads.dtlb_l1",
	.event = "umask=0x11,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L1+FB  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.dtlb_l2",
	.event = "umask=0x12,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L2  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.dtlb_l3",
	.event = "umask=0x14,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in the L3 + XSNP  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.dtlb_memory",
	.event = "umask=0x18,period=2000003,event=0xbc",
	.desc = "Number of DTLB page walker hits in Memory  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.itlb_l1",
	.event = "umask=0x21,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L1+FB  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.itlb_l2",
	.event = "umask=0x22,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L2  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "page_walker_loads.itlb_l3",
	.event = "umask=0x24,period=2000003,event=0xbc",
	.desc = "Number of ITLB page walker hits in the L3 + XSNP  Spec update: BDM69, BDM98",
	.topic = "virtual memory",
},
{
	.name = "tlb_flush.dtlb_thread",
	.event = "umask=0x1,period=100007,event=0xbd",
	.desc = "DTLB flush attempts of the thread-specific entries",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of DTLB flush attempts of the thread-specific entries",
},
{
	.name = "tlb_flush.stlb_any",
	.event = "umask=0x20,period=100007,event=0xbd",
	.desc = "STLB flush attempts",
	.topic = "virtual memory",
	.long_desc = "This event counts the number of any STLB flush attempts (such as entire, VPID, PCID, InvPage, CR3 write, and so on)",
},
{
	.name = 0,
	.event = 0,
	.desc = 0,
},
};
struct pmu_events_map pmu_events_map[] = {
{
	.cpuid = "GenuineIntel-6-56",
	.version = "v5",
	.type = "core",
	.table = pme_broadwellde
},
{
	.cpuid = "GenuineIntel-6-3D",
	.version = "v17",
	.type = "core",
	.table = pme_broadwell
},
{
	.cpuid = "GenuineIntel-6-47",
	.version = "v17",
	.type = "core",
	.table = pme_broadwell
},
{
	.cpuid = "GenuineIntel-6-4F",
	.version = "v10",
	.type = "core",
	.table = pme_broadwellx
},
{
	.cpuid = "GenuineIntel-6-1C",
	.version = "v4",
	.type = "core",
	.table = pme_bonnell
},
{
	.cpuid = "GenuineIntel-6-26",
	.version = "v4",
	.type = "core",
	.table = pme_bonnell
},
{
	.cpuid = "GenuineIntel-6-27",
	.version = "v4",
	.type = "core",
	.table = pme_bonnell
},
{
	.cpuid = "GenuineIntel-6-36",
	.version = "v4",
	.type = "core",
	.table = pme_bonnell
},
{
	.cpuid = "GenuineIntel-6-35",
	.version = "v4",
	.type = "core",
	.table = pme_bonnell
},
{
	.cpuid = "GenuineIntel-6-5C",
	.version = "v8",
	.type = "core",
	.table = pme_goldmont
},
{
	.cpuid = "GenuineIntel-6-7A",
	.version = "v1",
	.type = "core",
	.table = pme_goldmontplus
},
{
	.cpuid = "GenuineIntel-6-3C",
	.version = "v24",
	.type = "core",
	.table = pme_haswell
},
{
	.cpuid = "GenuineIntel-6-45",
	.version = "v24",
	.type = "core",
	.table = pme_haswell
},
{
	.cpuid = "GenuineIntel-6-46",
	.version = "v24",
	.type = "core",
	.table = pme_haswell
},
{
	.cpuid = "GenuineIntel-6-3F",
	.version = "v17",
	.type = "core",
	.table = pme_haswellx
},
{
	.cpuid = "GenuineIntel-6-3A",
	.version = "v18",
	.type = "core",
	.table = pme_ivybridge
},
{
	.cpuid = "GenuineIntel-6-3E",
	.version = "v19",
	.type = "core",
	.table = pme_ivytown
},
{
	.cpuid = "GenuineIntel-6-2D",
	.version = "v20",
	.type = "core",
	.table = pme_jaketown
},
{
	.cpuid = "GenuineIntel-6-57",
	.version = "v9",
	.type = "core",
	.table = pme_knightslanding
},
{
	.cpuid = "GenuineIntel-6-85",
	.version = "v9",
	.type = "core",
	.table = pme_knightslanding
},
{
	.cpuid = "GenuineIntel-6-1E",
	.version = "v2",
	.type = "core",
	.table = pme_nehalemep
},
{
	.cpuid = "GenuineIntel-6-1F",
	.version = "v2",
	.type = "core",
	.table = pme_nehalemep
},
{
	.cpuid = "GenuineIntel-6-1A",
	.version = "v2",
	.type = "core",
	.table = pme_nehalemep
},
{
	.cpuid = "GenuineIntel-6-2E",
	.version = "v2",
	.type = "core",
	.table = pme_nehalemex
},
{
	.cpuid = "GenuineIntel-6-[4589]E",
	.version = "v24",
	.type = "core",
	.table = pme_skylake
},
{
	.cpuid = "GenuineIntel-6-37",
	.version = "v13",
	.type = "core",
	.table = pme_silvermont
},
{
	.cpuid = "GenuineIntel-6-4D",
	.version = "v13",
	.type = "core",
	.table = pme_silvermont
},
{
	.cpuid = "GenuineIntel-6-4C",
	.version = "v13",
	.type = "core",
	.table = pme_silvermont
},
{
	.cpuid = "GenuineIntel-6-2A",
	.version = "v15",
	.type = "core",
	.table = pme_sandybridge
},
{
	.cpuid = "GenuineIntel-6-2C",
	.version = "v2",
	.type = "core",
	.table = pme_westmereep_dp
},
{
	.cpuid = "GenuineIntel-6-25",
	.version = "v2",
	.type = "core",
	.table = pme_westmereep_sp
},
{
	.cpuid = "GenuineIntel-6-2F",
	.version = "v2",
	.type = "core",
	.table = pme_westmereex
},
{
	.cpuid = "GenuineIntel-6-55-[01234]",
	.version = "v1",
	.type = "core",
	.table = pme_skylakex
},
{
	.cpuid = "GenuineIntel-6-55-[56789ABCDEF]",
	.version = "v1",
	.type = "core",
	.table = pme_cascadelakex
},
{
	.cpuid = "GenuineIntel-6-7D",
	.version = "v1",
	.type = "core",
	.table = pme_icelake
},
{
	.cpuid = "GenuineIntel-6-7E",
	.version = "v1",
	.type = "core",
	.table = pme_icelake
},
{
	.cpuid = "GenuineIntel-6-86",
	.version = "v1",
	.type = "core",
	.table = pme_tremontx
},
{
	.cpuid = "AuthenticAMD-23-([12][0-9A-F]|[0-9A-F])",
	.version = "v2",
	.type = "core",
	.table = pme_amdzen1
},
{
	.cpuid = "AuthenticAMD-23-[[:xdigit:]]+",
	.version = "v1",
	.type = "core",
	.table = pme_amdzen2
},
{
	.cpuid = 0,
	.version = 0,
	.type = 0,
	.table = 0,
},
};
